/*
 * (C) Copyright 2012
 * DSPG Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include "version.h"
#include "common.h"
#include "bootrom.h"
#include "hardware.h"
#include "romapi.h"
#include "serial.h"
#include "syscfg.h"
#include "cmu.h"
#include "ddr.h"
#include "load.h"
#include "dvf9a.h"
#include "pmu.h"
#include "gpio.h"
#include "command.h"
#include "command_memory.h"
#include "command_dvf9a.h"
#include "factory.h"

#if CPU > 480
	#define VCORE 1300
#else
	#define VCORE 1200
#endif

#define FEATURE_ACTIVATION_KEY              13 /* Enter */

#if CONSOLE == 1
Command commands[] =
{
	COMMAND_BASE
	COMMAND_MEMORY
	COMMAND_DVF9A
	{0} /* terminate list */
};
#endif

void die(void)
{
	serial_puts("error\n");
	serial_exit();

	writel(0x900db00d, DVF99_ROM_BASE);
	while (1);
}

static unsigned long boot_from_uart(void)
{
	const struct bootrom_calls *romapi = romapi_get();

	if (romapi->xmodem_rcv((void*)CONFIG_LOAD_ADDR, 0x1000000) < 0)
		die();
	/* TODO: udelay is needed for the putc to work (why?) */
	romapi->udelay(100000);
	serial_puts("\n");

	return CONFIG_LOAD_ADDR;
}

#ifdef CONFIG_HT
static unsigned char compare_uboot_versions( unsigned int Signature1, unsigned int Version1, unsigned int Signature2, unsigned int Version2 )
{
    unsigned short v[2] ;

    if ( Signature1 != GS_PROD_SIGNATURE && Signature2 != GS_PROD_SIGNATURE )
        return 0;
    if ( Signature1 != GS_PROD_SIGNATURE )
        return 1;
    if ( Signature2 != GS_PROD_SIGNATURE )
        return 0;

    v[0] = ( Version1 >> 24) & 0xff;
    v[1] = ( Version2 >> 24) & 0xff;
    if ( v[0] > v[1] )
        return 0;
    else if ( v[1] > v[0] )
        return 1;

    v[0] = ( Version1 >> 16) & 0xff;
    v[1] = ( Version2 >> 16) & 0xff;
    if ( v[0] > v[1] )
        return 0;
    else if ( v[1] > v[0] )
        return 1;

    v[0] = ( Version1 >> 8) & 0xff;
    v[1] = ( Version2 >> 8) & 0xff;
    if ( v[0] > v[1] )
        return 0;
    else if ( v[1] > v[0] )
        return 1;

    v[0] = Version1 & 0xff;
    v[1] = Version2 & 0xff;
    if ( v[0] > v[1] )
        return 0;
    else if ( v[1] > v[0] )
        return 1;

    return 0;
}

static int calculate_uboot_checksum( unsigned int addr, unsigned int offset, unsigned int size, unsigned short header_checksum )
{
    int i;
    unsigned short check_sum = 0;
    unsigned short *ptr = ( unsigned short *)addr;
	const struct bootrom_calls *romapi = romapi_get();

	romapi->bootdev_read((u_char *)addr, offset, size );
    for (i = 0; i < size/sizeof( unsigned short); i++)
    {
        check_sum += gs_endian_conv16( ptr[i] );
    }

    /* verify the checksum */
    if (((header_checksum + check_sum) & 0xffff) != 0)
        return 1;
    return 0;    
}

static int gs_uboot_check(unsigned int *offset, int dual_uboot)
{
	unsigned int addr = CONFIG_LOAD_ADDR;
    ImageHeader header[2];
    unsigned char uboot_index = 0;
    int found = 0;

	const struct bootrom_calls *romapi = romapi_get();

    //read in headers
	romapi->bootdev_read((u_char *)&header[0], offset[0], sizeof(ImageHeader) );
    
    if ( dual_uboot ) 
    {
	    romapi->bootdev_read((u_char *)&header[1], offset[1], sizeof(ImageHeader) );

        if ( gs_endian_conv32(header[0].Signature) == GS_PROD_SIGNATURE && 
             gs_endian_conv32(header[1].Signature) == GS_PROD_SIGNATURE )
        {
            //compare provision counters
            if ( gs_endian_conv32(header[0].ProvisionCounter) == gs_endian_conv32(header[1].ProvisionCounter) )
            {
                //compare versions
                uboot_index = compare_uboot_versions( gs_endian_conv32(header[0].Signature), header[0].Version, gs_endian_conv32(header[1].Signature), header[1].Version );
            }
            else
                uboot_index = ( gs_endian_conv32(header[0].ProvisionCounter) < gs_endian_conv32(header[1].ProvisionCounter) )?1:0;

        }
        else if ( gs_endian_conv32(header[1].Signature) == GS_PROD_SIGNATURE )
        {
            uboot_index = 1;
        }
    }
    else 
    {
        uboot_index = 0;
    }
    
    //calculate checksum
    if ( gs_endian_conv32(header[uboot_index].Signature) == GS_PROD_SIGNATURE &&
         calculate_uboot_checksum( addr, offset[uboot_index] + HEADER_BLOCK_SIZE, gs_endian_conv32(header[uboot_index].SizeUsed), gs_endian_conv16(header[uboot_index].Checksum) ) == 0 )
    {
        found = 1;
    }
    else
    {
        if ( dual_uboot ) 
        {
            uboot_index = ( uboot_index == 0 )?1:0;

            //calculate checksum
            if( gs_endian_conv32(header[uboot_index].Signature) == GS_PROD_SIGNATURE &&
                calculate_uboot_checksum( addr, offset[uboot_index] + HEADER_BLOCK_SIZE, gs_endian_conv32(header[uboot_index].SizeUsed), gs_endian_conv16(header[uboot_index].Checksum) ) == 0 )
            {
                found = 1;
            }
            else 
            {
                uboot_index = 1;
            }
        }
    }

    if ( found )
    {
        serial_puts("Found a valid GS uboot image\n");
        if ( uboot_index )
            serial_puts("Using uboot2 partition image\n");
        else
            serial_puts("Using uboot partition image\n");

        // update uboot image offset
        offset[uboot_index] += HEADER_BLOCK_SIZE;
    }
    else
        serial_puts("No GS uboot image found\n");
    return uboot_index;
}
#endif

void bootastic(int bootsel)
{
	void (*loader)(void);
	unsigned long loader_addr = 0;
	long addr;
	int secureboot = 0;
	int fallback = 0;
#ifdef CONFIG_HT
    unsigned int offset[2];
    int dual_uboot;
    int index;
#endif
	const struct bootrom_calls *romapi = romapi_get();

#if MAINCLK_OUT == 1
	/* enable main_clk_out in syscfg GCR1 register */
	syscfg_enable_main_clk_out();
	/* disable bgpio31 */
	gpio_disable(BGPIO(31));
#endif

	/*
	 * debug: enable this to wait for jtag attach before continuing to
	 * boot. in the debugger, simply set the PC to the next instruction
	 * and then go/singlestep.
	 */
#if 0
	asm("1: b 1b");
#endif

	/* allow serial to flush remaining data */
	serial_exit();

	daif_init();

	set_dpu_dcdc_regulator_level(VCORE);
	set_3v_regulator_to_3_3v();
	set_vref_to_vcca_div_by_2();

	cmu_setup();

	serial_init(115200);
	serial_puts("Bootastic " VERSION "\n");

#if CONSOLE == 1
	/* press "enter" before power up to enter console */
	if (serial_chkc() == FEATURE_ACTIVATION_KEY)
		monitor(commands);
#elif XMODEM_HOOK == 1
	if (serial_chkc() == FEATURE_ACTIVATION_KEY) {
		serial_putc('X');
		bootsel = BOOTSEL_UART;
		/* wait 2s to give the user time to release the Enter key */
		romapi->udelay(2000000);
		serial_flush();
	}
#endif

#ifdef CONFIG_HT
    if (0 != ht_dram_init()) {
        serial_puts("Failed basic DRAM initialization test\n");
        die();      
    }
#else
	if (0 != dram_init()) {
		serial_puts("Failed basic DRAM initialization test\n");
		die();		
	}
#endif

	if (syscfg_get_chip_rev() > 1) {
		secureboot = romapi->is_dev_secured();
		if (IS_ERR(ERR_PTR(secureboot)))
			die();
	}

	/* load the bootloader to its destination */
	switch (bootsel) {
		case BOOTSEL_UART:
			loader_addr = boot_from_uart();
			break;
		case BOOTSEL_NAND:
		case BOOTSEL_QSPI:
			/* scan NAND and look for a valid u-boot image header */

#ifdef CONFIG_GXP
            addr = load_find_loader(0, 0x400000);
#elif defined( CONFIG_HT )
            offset[0] = 0x20000;
            offset[1] = 0xfc0000;
            dual_uboot = 1;
            
            index = gs_uboot_check(offset, dual_uboot);
			addr = load_find_loader(offset[index], offset[index]+0x40000);
#else
#error
#endif
			if (addr < 0) {
				/* fallback to xmodem */
				addr = boot_from_uart();
				fallback = 1;
			}

			loader_addr = (unsigned long)addr;
			break;
	}

	loader = load_loader(loader_addr, secureboot, fallback);
	if (IS_ERR(loader) && (bootsel != BOOTSEL_UART))
		/* fallback to xmodem */
		loader = load_loader(boot_from_uart(), secureboot, 1);
	if (IS_ERR(loader))
		die();

	serial_exit();

	/* we do not return from this call */
	loader();
}

void __entry _start(uint32_t rom_version, const struct bootrom_calls *rom_calls, int bootsel)
{
	romapi_init(rom_version, rom_calls);

	bootastic(bootsel);
}
