#ifndef __BOOTROM_H
#define __BOOTROM_H

struct pubkey_t {
	uint8_t n[256];			/* public modulus (0x24)*/
	uint32_t e;			/* public exponent (0x124)*/
} __attribute__ ((packed));

struct secureboot_header {
	struct pubkey_t pubkey;		/* public key (0x24)*/
	uint8_t pre_sha256[32];		/* preloader sha256 digest (0x128)*/
	uint8_t load_sha256[32];	/* bootloader sha256 digest (0x148)*/
	uint8_t signature[256];		/* header signature (0x168)*/
} __attribute__ ((packed));

struct image_header {
	uint32_t size;			/* size of this header + 4byte crc32 (0x0)*/
	uint32_t flags;			/* optional flags (0x4)*/
	uint32_t pre_offset;		/* preloader: starts at offset (0x8)*/
	uint32_t pre_size;		/* preloader: size of preloader (0xc)*/
	uint32_t pre_crc32;		/* preloader: crc32 checksum (0x10)*/
	uint32_t load_offset;		/* bootloader: start at offset (0x14)*/
	uint32_t load_size;		/* bootloader: size of bootloader (0x18)*/
	uint32_t load_addr;		/* bootloader: load to address (0x1c)*/
	uint32_t load_crc32;		/* bootloader: crc32 checksum (0x20)*/
	struct secureboot_header sb_hdr; 	/* secureboot header (0x24)*/
	uint32_t crc32;                 /* crc32 of all above fields */
} __attribute__ ((packed));

#define HDRFLAG_CRC32   0x00000001
#define HDRFLAG_SECURE	0x00000002

enum bootrom_errors {
	ERR_BOOTCONF = 1,
	ERR_UART_IMAGE,
	ERR_NAND_INIT,
	ERR_EMMC_INIT,
	ERR_EMMC_IO,
	ERR_EMMC_NOPART,
	ERR_EMMC_NOPARTACTIVE,
	ERR_PRE_IO,
	ERR_PRE_IMAGE,
	ERR_PRE_CRC32,
	ERR_UNEXPECTED,
	ERR_DIVBYZERO,
	ERR_EXCEPTION,
	ERR_CRYPTO_INVAL,
	ERR_CRYPTO_IO,
	ERR_SBOOT_NOSIG,
	ERR_SBOOT_INVPUBKEY,
	ERR_SBOOT_INVSIG,
	ERR_SBOOT_INVIMG,
	NUM_BOOTROM_ERRORS,
};

#define RAM_PRELOADER_SIZE 256*1024
#define RAM_MALLOC_SIZE    16*1024

#ifdef BOOTROM

#define __weak     __attribute__((weak))
#define __noinline __weak __attribute__((noinline))
#define __noreturn __attribute__((noreturn))

#define DECLARE_BOOTROM_ERROR \
	enum bootrom_errors bootrom_error __attribute__((section(".error")))

extern unsigned char ram_preloader[RAM_PRELOADER_SIZE];

#define DECLARE_RAM_PRELOADER \
	unsigned char ram_preloader[RAM_PRELOADER_SIZE] \
		__attribute__((section(".preloader")))

extern unsigned char ram_malloc[RAM_MALLOC_SIZE];

#define DECLARE_RAM_MALLOC \
	unsigned char ram_malloc[RAM_MALLOC_SIZE] \
		__attribute__((section(".malloc")))

void loop(void) __noreturn __noinline;
const char *strerror(enum bootrom_errors error);
const char *u32tohexstr(unsigned int val);
void perror(enum bootrom_errors error);
void die(enum bootrom_errors error) __noreturn;
void nand_init(void);
void next_bootloader(void *addr) __noreturn;

#endif /* BOOTROM */

#endif
