/*
 * (C) Copyright 2012
 * DSPG Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include "common.h"
#include "hardware.h"
#include "romapi.h"

#include "cmu.h"
#include "syscfg.h"
#include "ddrconf/ddrconf.h"

void reset_release(unsigned int block)
{
	unsigned int off = (block / 32) * sizeof(uint32_t);
	uint32_t tmp;

	block &= 0x1f;

	tmp = readl(DVF99_CMU_BASE + CMU_SWRST1 + off);
	tmp &= ~(1 << block);
	writel(tmp, DVF99_CMU_BASE + CMU_SWRST1 + off);
}

void clk_enable(unsigned int block)
{
	unsigned int off = (block / 32) * sizeof(uint32_t);
	uint32_t tmp;

	block &= 0x1f;

	tmp = readl(DVF99_CMU_BASE + CMU_SWCLKEN1 + off);
	tmp |= (1 << block);
	writel(tmp, DVF99_CMU_BASE + CMU_SWCLKEN1 + off);
}

static unsigned long
dvf99_calc_pllfreq(unsigned long input_freq, int pll_ctrl)
{
	unsigned long pll_out_freq;

	if (pll_ctrl & (1<<16)) {
		/* bypassed */
		pll_out_freq = input_freq;
	} else {
		uint32_t r, f, od;

		r = (pll_ctrl & 0x1f) + 1;
		f = ((pll_ctrl >> 5) & 0x7f) + 1;
		od = 1 << ((pll_ctrl >> 12) & 3);

		pll_out_freq = input_freq * f / r / od;
	}

	return pll_out_freq;
}

static uint32_t
dvf99_get_main_clk(void)
{
	uint32_t reg;
	uint32_t sysclk, mainclk, dectclk = 13824000ul;

	reg = readl(DVF99_CMU_BASE + CMU_RSTSR);

	/* 25mhz/50mhz main clock selection */
	if (reg & 0x4)
		mainclk = 25000000ul;
	else
		mainclk = 50000000ul;

	/* main clock can be overwritten by dect clk via main_clk_sel_strap */
	sysclk = (reg & 0x2) ? dectclk : mainclk;

	return sysclk;
}

static uint32_t
dvf99_get_pll3src(void){
	uint32_t pll3src = (readl(DVF99_CMU_BASE + CMU_CPUCLKCTL) >> 16) & 0x3;
	switch (pll3src) {
		case 0: return dvf99_get_main_clk();
		case 1: return 13824000ul;
		case 2: return 12000000ul;
	}
	return 0;
}

static uint32_t
dvf99_get_sysclk(void)
{
	uint32_t sysclksrc = (readl(DVF99_CMU_BASE + CMU_CLKSWCTL) >> 1) & 0x1;

	if (!sysclksrc) {
		uint32_t altsyssrc = (readl(DVF99_CMU_BASE + CMU_CPUCLKCTL) >> 14) & 0x3;
		switch (altsyssrc) {
			case 0: return dvf99_get_main_clk();
			case 1: return 13824000ul;
			case 2: return 12000000ul;
		}
	} else {
		uint32_t pll1 = readl(DVF99_CMU_BASE + CMU_PLL1);
		switch ((pll1 >> 19) & 0x3) {
			case 0: return dvf99_calc_pllfreq(dvf99_get_main_clk(), pll1);
			case 1: return dvf99_calc_pllfreq(13824000ul, pll1);
			case 2: return dvf99_calc_pllfreq(12000000ul, pll1);
		}
	}

	return 0;
}

static void
dram_freq(uint32_t main, uint32_t freq,
	  uint32_t *prediv, uint32_t *mult, uint32_t *postdiv, uint32_t *band)
{
	*mult = 0;
	*prediv = 1;
	*postdiv = 0;
	*band = 0;

	switch (freq) {
	case 50000000ul:
		switch (main) {
		case 13824000ul: /* 62.208MHz */
			*mult = 36;
			*postdiv = 3;	/* divide by 8 */
			break;
		case 25000000ul:
			*mult = 16;
			*postdiv = 3;	/* divide by 8 */
			break;
		case 50000000ul:
			*prediv = 2;
			*mult = 16;
			*postdiv = 3;	/* divide by 8 */
			break;
		}
		break;
	case 125000000ul:
		switch (main) {
		case 13824000ul: /* 124.416MHz */
			*mult = 36;
			*postdiv = 2;	/* divide by 4 */
			break;
		case 25000000ul:
			*mult = 20;
			*postdiv = 2;	/* divide by 4 */
			break;
		case 50000000ul:
			*prediv = 2;
			*mult = 20;
			*postdiv = 2;	/* divide by 4 */
			break;
		}
		break;
	case 250000000ul:
		switch (main) {
		case 13824000ul: /* 248.832MHz */
			*mult = 36;
			*postdiv = 1;
			break;
		case 25000000ul:
			*mult = 20;
			*postdiv = 1;
			break;
		case 50000000ul:
			*prediv = 5;
			*mult = 50;
			*postdiv = 1;
			break;
		}
		break;
	case 300000000ul:
		switch (main) {
		case 13824000ul: /* 297.216MHz */
			*mult = 43;
			*postdiv = 1;
			break;
		case 25000000ul:
			*mult = 12;
			break;
		case 50000000ul:
			*mult = 6;
			break;
		}
		break;
	case 325000000ul:
		switch (main) {
		case 13824000ul: /* 297.216MHz */
			*mult = 47;
			*postdiv = 1;
			break;
		case 25000000ul:
			*mult = 13;
			break;
		case 50000000ul:
			*prediv = 2;
			*mult = 13;
			break;
		}
		break;
	case 350000000ul:
		switch (main) {
		case 13824000ul: /* 352.512MHz */
			*mult = 51;
			*postdiv = 1;
			*band = 1;
			break;
		case 25000000ul:
			*mult = 14;
			break;
		case 50000000ul:
			*prediv = 5;
			*mult = 35;
			break;
		}
		break;
	case 375000000ul:
		switch (main) {
		case 13824000ul: /* 373.248MHz */
			*mult = 27;
			break;
		case 25000000ul:
			*mult = 15;
			break;
		case 50000000ul:
			*prediv = 2;
			*mult = 15;
			break;
		}
		break;
	}
}

#define ABSDIFF(x, y)	((x) > (y) ? (x) - (y) : (y) - (x))

static long
round_rate_pll(unsigned long in_rate, unsigned long rate, unsigned long *rprediv, unsigned long *rmult, unsigned long *rpostdiv, unsigned long *rband)
{
	unsigned long prediv, mult, postdiv;
	unsigned long match_prediv=0, match_mult=0, match_postdiv=0;
	unsigned long match_rate=0, match_diff=~0ul, match_vco=~0ul;

	/* loop through all settings finding best match */
	for (prediv=1; prediv<=32; prediv++) {
		unsigned long phaseclk = (in_rate >> 10) / prediv;

		/* keep in phase detector reference frequency: 10..50MHz */
		if (phaseclk < 10000000>>10 || phaseclk > 50000000>>10)
			continue;

		for (mult=6; mult<=100; mult++) {
			unsigned long vcoclk = phaseclk * mult;

			/* keep in VCO frequency range: 300..1000MHz */
			if (vcoclk < 300000000>>10 || vcoclk > 1000000000>>10)
				continue;

			for (postdiv=0; postdiv<=3; postdiv++) {
				unsigned long diff, out = vcoclk >> postdiv;
				unsigned long tmp;

				/* quick check if in target range */
				if (out-10000 > rate>>10 || out+10000 < rate>>10)
					continue;

				/* ok, do the precise calculation */
				tmp = in_rate / 1000 * mult;
				tmp = tmp / (unsigned long)prediv;
				out = (unsigned long)tmp >> postdiv;
				out *= 1000;

				/* better match? */
				diff = ABSDIFF(rate, out);
				if ((diff <  match_diff) ||
				    (diff == match_diff && vcoclk < match_vco)) {
					match_diff    = diff;
					match_vco     = vcoclk;
					match_rate    = out;
					match_prediv  = prediv;
					match_mult    = mult;
					match_postdiv = postdiv;
				}
			}
		}
	}

	if (!match_rate)
		return -1;

	*rprediv = match_prediv - 1;
	*rmult = match_mult - 1;
	*rpostdiv = match_postdiv;
	*rband = match_vco > 600000000>>10 ? 1: 0;

	return match_rate;
}


static void
pll_setup(void)
{
	uint32_t reg, mult = 0, prediv = 1, postdiv = 0, band = 0;
	uint32_t main, sysclk, pllsrc;

#if USB_FULLSPEED==1
	/* To support FS-USB on chip versions before ABC/BBC
	 * system clock will be sourced from 12MHz oscillator */
	if (syscfg_get_chip_rev() <= 1) {
		reg = readl(DVF99_CMU_BASE + CMU_CLKOSC);
		reg |= (1 << 2);
		writel(reg, DVF99_CMU_BASE + CMU_CLKOSC);

		/* wait until stable */
		do {
			reg = readl(DVF99_CMU_BASE + CMU_CLKOSC);
		} while (!(reg & 1 << 5));
	}
#endif

	/* Before PLL1 is set up, make sure the system is in a safe state */

	/* make sure hclk is <= 200MHz */
	reg = readl(DVF99_CMU_BASE + CMU_CPUCLKCTL);
	reg &= ~(3 << 4);
	reg |= (AHBCLK_DIV-1) << 4;   /* ahbclkdivrate: 3 or 4 */
	reg &= ~(3 << 16);
#if OSC13==1
	reg |=   1 << 16;         /* audsrc:        13.824MHz */
#endif
	reg &= ~(3 << 14);

#if USB_FULLSPEED==1
	/* To support FS-USB on chip versions before ABC/BBC
	 * system clock will be sourced from 12MHz oscillator */
	if (syscfg_get_chip_rev() <= 1)
		reg |= 2 << 14;
#endif
	writel(reg, DVF99_CMU_BASE + CMU_CPUCLKCTL);

	/* feed AP-CPU from altsys */
	reg = readl(DVF99_CMU_BASE + CMU_CLKSWCTL);
	reg &= ~(1 << 1);         /* swsysclksrc:   altsyssrc */
	writel(reg, DVF99_CMU_BASE + CMU_CLKSWCTL);

	main = dvf99_get_main_clk();
	sysclk = dvf99_get_sysclk();
	/*
	 * use sysclk as input for PLL1 because later we switch the actual
	 * muxer for the PLL1 to exactly the same value
	 */

#if USB_FULLSPEED==1
	/* Get the nearest clock rate that is a multiple of 48MHz for FS-USB
	 * compatibility for chips before ABC/BBC */
	/* sysclk is already switched to 12MHz OSC by now */
	if (syscfg_get_chip_rev() <= 1)
		round_rate_pll(sysclk, (((int)(CPU/48))*48 * 1000000), &prediv, &mult, &postdiv, &band);
	else
#endif
	round_rate_pll(sysclk, CPU * 1000000, &prediv, &mult, &postdiv, &band);

	/* feed CSS-CPU from pll1 (i.e., altsys) */
	reg = readl(DVF99_CMU_BASE + CMU_COMCLKSEL);
	reg &= ~(3 << 10);        /* swcomsrc:      clk_sysclk */
	writel(reg, DVF99_CMU_BASE + CMU_COMCLKSEL);

	/* configure PLL1 for ~450MHz */
	reg = ( prediv  <<  0) |  /* pre-div:       1 */
	      ( mult    <<  5) |  /* multiplicator: see above */
	      ( postdiv << 12) |  /* output-div:    1 */
	      ( band    << 14) |  /* band:          300 - 600 Mhz */
	      ( 1       << 15) |  /* power-down:    1 */
	      ( 0       << 16);   /* bypass:        0 */

#if USB_FULLSPEED==1
	/* To support FS-USB on chip versions before ABC/BBC
	 * system clock will be sourced from 12MHz oscillator */
	if (syscfg_get_chip_rev() <= 1)
		reg |= 2 << 19;   /* input:         clk_12m */
#endif
	writel(reg, DVF99_CMU_BASE + CMU_PLL1);

	/* wait at least 500 ns */
	delay(1);

	reg &= ~(1 << 15);        /* power-down:    0 */
	writel(reg, DVF99_CMU_BASE + CMU_PLL1);

	/* wait for lock */
	do {
		reg = readl(DVF99_CMU_BASE + CMU_PLL1);
	} while ((reg & (1 << 18)) == 0);

	/* feed AP-CPU from PLL1 */
	reg = readl(DVF99_CMU_BASE + CMU_CLKSWCTL);
	reg |=  (  1 <<  1);      /* swsysclksrc:   pll1 */
	writel(reg, DVF99_CMU_BASE + CMU_CLKSWCTL);

	/* feed CSS-CPU from PLL1 or PLL2 (if PLL1>500MHz) */
	reg = readl(DVF99_CMU_BASE + CMU_COMCLKSEL);
	reg &= ~(0xf <<  0);      /* comprediv:     0 */
#if CPU > 500
	reg &= ~(0x3 << 10);
	reg |=   0x2 << 10;       /* swcomsrc: clk_audclk */
#endif
	reg |=   1 << 14;         /* combmpsrcsel:  if_clk_if */
#if OSC13==1
	reg &= ~(1 << 14);        /* combmpsrcsel:  if_clk_pll */
#endif
	writel(reg, DVF99_CMU_BASE + CMU_COMCLKSEL);

	pllsrc = dvf99_get_pll3src();
	if ((pllsrc == 13824000ul) || (OSC13 == 1)) {
		/* set CSS clock rate near to 500MHz */
		round_rate_pll(13824000ul, 500000000ul, &prediv, &mult, &postdiv, &band);
	} else {
		/* Setup to come as close to 13.824MHz as possible:
		 * prediv = 2 or 4; f = 73; postdiv = 2; Fout = 456,25 MHz
		 * APU can run at 456,25MHz / 33 = 13.8258MHz */
		round_rate_pll(pllsrc, 456250000ul, &prediv, &mult, &postdiv, &band);
	}
	reg = ( prediv  <<  0) |  /* pre-div:       1 */
	      ( mult    <<  5) |  /* multiplicator: see above */
	      ( postdiv << 12) |  /* output-div:    1 */
	      ( band    << 14) |  /* band:          300 - 600 Mhz */
	      ( 1       << 15) |  /* power-down:    1 */
	      ( 0       << 16);   /* bypass:        0 */

	writel(reg, DVF99_CMU_BASE + CMU_PLL3);

	/* wait at least 500 ns */
	delay(1);

	reg &= ~(1 << 15);        /* power-down:    0 */
	writel(reg, DVF99_CMU_BASE + CMU_PLL3);

	/* wait for lock */
	do {
		reg = readl(DVF99_CMU_BASE + CMU_PLL3);
	} while ((reg & (1 << 18)) == 0);

	dram_freq(main, ddrconf_speed(), &prediv, &mult, &postdiv, &band);
	if (!mult)
		return;

	reg = ((prediv-1)<<  0) | /* pre-div:       */
	      ((mult-1)  <<  5) | /* multiplicator: */
	      ( postdiv  << 12) | /* output-div:    */
	      ( band     << 14) | /* band:          300 - 600 Mhz */
	      ( 1        << 15) | /* power-down:    1 */
	      ( 0        << 16);  /* bypass:        0 */
	writel(reg, DVF99_CMU_BASE + CMU_PLL2);

	/* wait at least 500 ns */
	delay(100);

	reg &= ~(1 << 15);        /* power-down:    0 */
	writel(reg, DVF99_CMU_BASE + CMU_PLL2);

	/* wait for lock */
	do {
		reg = readl(DVF99_CMU_BASE + CMU_PLL2);
	} while ((reg & (1 << 18)) == 0);
}

static void
per_setup(void)
{
	uint32_t reg;

	reg = readl(DVF99_CMU_BASE + CMU_CLKSWCTL);
	reg |=  ( 1 <<  0);      /* pll2 output to external pads (lcdc-pclk, clockout, etc.) */
	reg |=  ( 1 <<  2);      /* swaudclksrc:   pll3 output */
	if (ddrconf_speed() > 375000)
		reg |=  (1<<15); /* gpudivsrc:     sysclk */
	else
		reg &= ~(1<<15); /* gpudivsrc:     pll2   */
	reg |=  ( 1 << 16);      /* swifclksrc:    apudiv */
	reg |=  ( 1 << 17);      /* lcdpixdivsrc:  pll2   */
	reg &= ~( 1 << 18);      /* ciudivsrc:     sysclk */
	reg &= ~( 3 << 19);
	reg |=  ( 1 << 19);      /* tdm1divsrc:    audclk */
	reg &= ~( 3 << 21);
	reg |=  ( 1 << 21);      /* tdm2divsrc:    audclk */
	reg |=  ( 1 << 25);      /* ethrdivsrc:    pll2   */
	reg &= ~( 1 << 26);      /* qspidivsrc:    sysclk */
	reg &= ~( 1 << 27);      /* sddivsrc:      sysclk */
	writel(reg, DVF99_CMU_BASE + CMU_CLKSWCTL);

	/* FS-USB will be sourced from pll4 for ABC/BBC and later chips */
	if (syscfg_get_chip_rev() > 1) {
		reg = readl(DVF99_CMU_BASE + CMU_CLKDIV4);
		reg |= 3 << 28;  /* usb11divsrc:   pll4   */
		writel(reg, DVF99_CMU_BASE + CMU_CLKDIV4);
	}

	reg = readl(DVF99_CMU_BASE + CMU_CLKDIV5);
	reg &= ~(0xff << 24);
	reg |=   0x20 << 24;
	writel(reg, DVF99_CMU_BASE + CMU_CLKDIV5);

	/* don't stop cpu clock when entering WFI	*/
	/* this allows connecting with debugger when the cpu in WFI, */
	/* since on dvf99, corsight is clocked by cpu clock */
	reg = readl(DVF99_CMU_BASE + CMU_TEST);
	reg |= (1 << 11);	/* CSS_ARM_CLK_EN_BP */
	reg |= (1 << 12);	/* ARM_CLK_EN_BP */
	writel(reg, DVF99_CMU_BASE + CMU_TEST);
}

void
cmu_setup(void)
{
	const struct bootrom_calls *romapi = romapi_get();

	if (romapi->clkchg_pre) {
		romapi->clkchg_pre(CLKCHG_PCLK,   150000000);
		romapi->clkchg_pre(CLKCHG_SYSCLK, 450000000);
	}

	pll_setup();
	per_setup();

	/* release ETM from reset */
	reset_release(BLOCK_MCU_ETM);
	clk_enable(BLOCK_MCU_ETM);

	if (romapi->clkchg_post) {
		romapi->clkchg_post(CLKCHG_PCLK,   150000000);
		romapi->clkchg_post(CLKCHG_SYSCLK, 450000000);
	}
}

uint32_t
cmu_get_pclk(void)
{
	uint32_t clkcntrl = (readl(DVF99_CMU_BASE + CMU_CPUCLKCTL) >> 4) & 0x3;

	/* divide by ahbclkdivrate */
	if (clkcntrl == 0)
		return dvf99_get_sysclk() / 3;
	return dvf99_get_sysclk() / (clkcntrl+1);
}
