/*
 * (C) Copyright 2013
 * DSPG Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include "tools.h"
#include "io.h"
#include "md5.h"
#include "configs.h"

/*************************************************************************************
 *  Local Struct and Definition
 *************************************************************************************/
typedef struct
{
	char start; /* bit start */
	char end;   /* bit end */

	/* bit group description token, format list below.
	 * general description, description for value = 0, description for value = 1,...
	 */
	char * description;
} BitDescription;

typedef struct
{
	unsigned int address;
	char * name;
	char * description;
	const BitDescription * bitDescriptions;
} RegisterDescription;

/*************************************************************************************
 *  Local Variables
 *************************************************************************************/
const BitDescription reg_RSTSR[] =
{
	{1,1,  "MAINCLKSEL_STRAP CLK_MAIN_CLK IF_CLK_PLL"},
	{2,2,  "CLK_25M_50M_SEL_STRAP 50MHz 25MHz"},
	{3,3,  "WDRSTIND_watchdog_reseted not_reseted reseted"},
	{4,4,  "SWWRMRSTIND_software_warm_reseted not_reseted reseted"},
	{5,5,  "SWSYSRSTIND_software_full_reseted not_reseted reseted"},
	{6,7,  "CLK_OUT_STRAP MAIN_CLK HOST_CPU_CLK HOST_AHB_CLK CLKOUTDIV"},
	{8,8,  "UCFG_UART_opt1"},
	{9,11, "UCFG_boot_from resv resv JTAG PRODUCTION resv UART NAND QSPI"},
	{12,12, "UCFG_UART_opt2"},
	{0}, /* description=0 -> termination */
};

const BitDescription reg_PLL1CNTRL[] =
{
	{0,4,  "PLL1R_Pre_divider"},
	{5,11, "PLL1F_Feedback_divider"},
	{12,13,"PLL1O_Output_division_configuration"},
	{14,14,"PLL1BS_VCO_range 300-600MHz 600-1000MHz"},
	{15,15,"PLL1PD_power_status active power_down_mode"},
	{16,16,"PLL1BP_bypass_status not_bypassed bypassed"},
	{17,17,"PLL1ERR_software_setting_status no_error error"},
	{18,18,"PLL1LD_lock_status not_locked locked"},
	{19,20,"PLL1SRC_clock_source MAIN 13.824MHz 12MHz resv"},
	{0}, /* description=0 -> termination */
};

const BitDescription reg_PLL2CNTRL[] =
{
	{0,4,  "PLL2R_Pre_divider"},
	{5,11, "PLL2F_Feedback_divider"},
	{12,13,"PLL2O_Output_division_configuration"},
	{14,14,"PLL2BS_VCO_range 300-600MHz 600-1000MHz"},
	{15,15,"PLL2PD_power_status active power_down_mode"},
	{16,16,"PLL2BP_bypass_status not_bypassed bypassed"},
	{17,17,"PLL2ERR_software_setting_status no_error error"},
	{18,18,"PLL2LD_lock_status not_locked locked"},
	{0}, /* description=0 -> termination */
};

const BitDescription reg_CPUCLKCNTRL[] =
{
	{0,3,  "CPUDIV_CPU_clock_divider"},
	{5,4,  "AHBCLKDIV_AHB_bus_clock_divider 2 2 3 4"},
	{14,15,"ALTSYSSRC_alternative_system_clock_source MAIN 13.824MHz 12MHz resv"},
	{16,17,"AUDSRC_clock_source MAIN 13.824MHz 12MHz resv"},
	{0}, /* description=0 -> termination */
};

const RegisterDescription registerDescriptions[] =
{
	{ 0x5300048,"RSTSR","reset status register", reg_RSTSR },
	{ 0x5300000,"PLL1CNTRL","PLL1 control register", reg_PLL1CNTRL },
	{ 0x5300004,"PLL2CNTRL","PLL2 control register", reg_PLL2CNTRL },
	{ 0x5300014,"CPUCLKCNTRL","Main CPU clock control register", reg_CPUCLKCNTRL },
	{0}, /* description=0 -> termination */
};

/*************************************************************************************
 *  commands functions
 *************************************************************************************/
unsigned char cmdMemoryDisplay (unsigned char argc, char * argv[] )
{
	unsigned char status = -1;
	unsigned int i, address, size;
	unsigned int * data;

	if ( argc == 2 || argc == 3 ) {
		address  = hex2int(argv[1]);

		if ( argc == 3 )
			size     = dec2int(argv[2]);
		else
			size = 1;

		data     = (unsigned int *)address;

		for ( i = 0; i < size; i++ ) {
			if ( i == 0 )
				printf ( "%0.8x: ", data );
			else if ( i % 4 == 0 )
				printf ( "\n%0.8x: ", data );

			printf("%0.8x ", (unsigned int)*data++ );
		}
		printf ("\n");

		status = 0;
	} else {
		printf("usage : %s address [size] - display address in hex with size in 4 bytes \n", argv[0]);
	}

	return status;
}

extern MemoryConfig memoryConfigs[];
extern const int memoryConfigsCount;

unsigned char cmdMemoryListMemoryConfigs (unsigned char argc, char * argv[] )
{
	unsigned char status = -1;
	unsigned int i;

	if ( argc == 1 ) {
		printf("Config(s) available\n" );
		printf("=============================\n" );
		for ( i = 0; i < memoryConfigsCount; i ++ ) {
			printf("%s \n", memoryConfigs[i].memory_config);
		}
		status = 0;
	} else {
		printf("usage : %s - list memory configs available\n", argv[0]);
	}

	return status;
}

extern int dram_deinit(void);
extern int dram_init_configs(uint32_t * denali_setup_configs);

unsigned char cmdMemoryConfigMemory (unsigned char argc, char * argv[] )
{
	unsigned char status = -1;
	unsigned int i;

	if ( argc == 2 ) {
		for ( i = 0; i < memoryConfigsCount; i ++ ) {
			if ((strcmp(argv[1], memoryConfigs[i].memory_config)==0) ) {
				dram_deinit();
				dram_init_configs(memoryConfigs[i].denali_setup);
			}
		}
	} else {
		printf("usage : %s configuration - config ddr with configs selects\n", argv[0]);
	}

	return status;
}

unsigned char cmdMemoryListRegister (unsigned char argc, char * argv[] )
{
	unsigned char status = -1;
	unsigned int i;

	if ( argc == 1 ) {
		printf("Register(s) available\n" );
		printf("=============================\n" );
		for ( i = 0; registerDescriptions[i].description != 0; i ++ ) {
			printf("%s - %s\n", registerDescriptions[i].name, registerDescriptions[i].description);
		}
		status = 0;
	} else {
		printf("usage : %s - list register available for rreg\n", argv[0]);
	}

	return status;
}

unsigned char cmdMemoryReadRegister (unsigned char argc, char * argv[] )
{
	unsigned int * pAddress;
	unsigned int regValue;
	unsigned char status = -1;
	unsigned char tokenCount;
	unsigned char value;
	unsigned int i;
	unsigned char j,k;
	unsigned int start, end;
	char buffer[100];
	char * token[20];

	if ( argc == 2 ) {
		/* registerDescriptions[i].description == 0 -> list termination */
		for ( i = 0; registerDescriptions[i].description != 0; i ++ ) {
			if ((strcmp(argv[1],registerDescriptions[i].name)==0) ) {
				pAddress = (unsigned int *)registerDescriptions[i].address;
				regValue = *pAddress;
				printf("%s(0x%0.8x) = 0x%0.8x\n", registerDescriptions[i].name, pAddress,regValue );
				printf("===========================================\n");

				/* registerDescriptions[i].bitDescriptions[j].description == 0 -> list termination */
				for ( j = 0; registerDescriptions[i].bitDescriptions[j].description != 0; j++ ) {

					/* parse bits description */
					strncpy(buffer,registerDescriptions[i].bitDescriptions[j].description,100);
					tokenCount = makeargv(buffer,20,token);

					start = registerDescriptions[i].bitDescriptions[j].start;
					end = registerDescriptions[i].bitDescriptions[j].end;

					/* calculate value accordingly */
					value = 0;
					for ( k = start; k <= end; k++ ) {
						if ( ( 1 << k ) & regValue )
							value += (1<<(k-start));
					}
					/* print value first */
					printf("%d:%d,%s = 0x%x", start, end, token[0], value );

					/* have token description */
					if ((tokenCount>=3)&&((value+1)<tokenCount))
						printf(", %s", token[1+value] );

					printf("\n");
				}
				status = 0;
			}
		}
	} else {
		printf( "usage : %s [register_name] - read value of register_name\n", argv[0]);
	}

	return status;
}

unsigned char cmdMemoryWrite (unsigned char argc, char * argv[] )
{
	unsigned int address;
	unsigned int data;
	unsigned char status = -1;

	if ( argc == 3 ) {
		address  = hex2int(argv[1]);
		data     = hex2int(argv[2]);
		*(unsigned int *)address = data;
		printf("Write Address : 0x%x, Value=0x%x\n", address, data );
		status = 0;
	} else {
		printf( "usage : %s [unsigned int address] [unsigned int data]- write unsigned int to address\n", argv[0]);
	}

	return status;
}

#define MD5_SIZE 0x4000
#define MD5_CYCLE 100
unsigned char cmdMemoryMd5Test (unsigned char argc, char * argv[] )
{
	static int set;
	unsigned char * string;
	unsigned char md5_calc[16];
	unsigned char md5_calc_next[16];
	unsigned int address = 0, md5size = MD5_SIZE, index = MD5_CYCLE;
	int len, i, j, l, d, nextCharIndex = 0;
	char nextChar[5] = "0o.o";

	if ( argc == 2 || argc == 3 || argc == 4 ) {
		address = hex2int(argv[1]);
		if ( argc > 2 )
			md5size = dec2int(argv[2]);
		if ( argc > 3 )
			index = dec2int(argv[3]);

		string = (unsigned char *)address;

		printf("run md5 test: address=0x%0.8x size=%d cycle=%d\n", address, md5size, index);

		set = 0;
		for (i=0;i<md5size;i++) {
			string[i] = (char) i & 0xff;
		}

		for (j=0;j<index;j++) {
			for (i=0;i<md5size;i++) {
				if (set==0) {
					set = 1;
				} else {
					set = 0;
				}
				string[i] = (char) i & 0xff;
				string[i] = ~((char) string[i]);
			}

			for (i=0;i<md5size;i++) {
				if (set==1) {
					if ( string[i]!=(i&0xff) ) {
						printf("Write it the problem Iteration # %d\n",i);
						break;
					}
				} else if (set==0) {
					d = ~((char) string[i]);
					d = d & 0xff;
					if (d != (i&0xff)) {
						printf("Write it the problem Iteration # %d\n",i);
						break;
					}
				}
			}

			MD5calc(md5_calc, string, md5size);
			MD5calc(md5_calc_next, string, md5size);

			for (i=0;i<16;i++) {
				if (md5_calc[i] != md5_calc_next[i]) {
					printf("Different hash values detected Iteration # %d\n", j);
					for (l=0;l<md5size;l++) {
						d = ~((char) string[l]);
						d = d & 0xff;
						if (d != (l&0xff)) {

							printf("Here the error Address: 0x%x\n",l);
							len = l;
							for (j=len-0x20;j<len+0x20;j++)
								printf(" #%x\n",string[j]);
							break;
						}
					}
					break;
				}
			}

			if ( ++nextCharIndex == 4 ) nextCharIndex = 0;
			printf ("\b\b\b\b\b\b%c %0.2d %c", nextChar[nextCharIndex], (int)((j*100)/index), nextChar[nextCharIndex] );
		}
		printf (" done\n");
	} else {
		printf("usage : %s address [size] [cycles] - address in hex with size in byte to run cycles time test \n", argv[0]);
	}

	return 0;
}
