/*
 * (C) Copyright 2013
 * DSPG Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include "types.h"
#include "version.h"
#include "common.h"
#include "hardware.h"
#include "cryptohw.h"

#ifndef CONFIG_CRYPTOHW_MEMADDR
  #error "crypto HW mem address not defined!"
#endif

static void cryptohw_writel(unsigned long reg, unsigned long val)
{
	writel(val, DVF99_CRYPTO_BASE + reg);
}

int cryptohw_sha256(const unsigned char *message, unsigned int len,
		unsigned char *digest)
{
	/* hw descriptor should be located in DDR.
	 * look in the Makefile for the address */
	struct cryptohw_desc *hw_desc =
		(struct cryptohw_desc*)CONFIG_CRYPTOHW_MEMADDR;

	if (!len)
		return -1;

	cryptohw_memset(hw_desc, 0, sizeof(struct cryptohw_desc));

	hw_desc->desc_ctl = DESC_CTL_DESC_OWNER_HW | DESC_CTL_MODE_SHA2_256;

	hw_desc->data_in_phys = (uint32_t)message;
	hw_desc->data_in_len = len;
	hw_desc->data_out_phys = (uint32_t)digest;
	hw_desc->data_out_len = 32;
	hw_desc->link_ptr = 0;

	cryptohw_writel(REG_CHANNEL_HEADPTR, (unsigned long)hw_desc);
	cryptohw_writel(REG_HEADPTR_SET, 1);

	/* wait for hw to finish */
	while(hw_desc->desc_ctl & DESC_CTL_DESC_OWNER_HW)
	{
		delay(1);
	}

	return 0;
}


/************************************************************/
/* memset and memcmp functions below (taken from string.c)  */
/************************************************************/

/**
 * memset - Fill a region of memory with the given value
 * @s: Pointer to the start of the area.
 * @c: The byte to fill the area with
 * @count: The size of the area.
 *
 * Do not use memset() to access IO space, use memset_io() instead.
 */
void * cryptohw_memset(void * s,int c,size_t count)
{
	unsigned long *sl = (unsigned long *) s;
	unsigned long cl = 0;
	char *s8;
	int i;

	/* do it one word at a time (32 bits or 64 bits) while possible */
	if ( ((unsigned long)s & (sizeof(*sl) - 1)) == 0) {
		for (i = 0; i < sizeof(*sl); i++) {
			cl <<= 8;
			cl |= c & 0xff;
		}
		while (count >= sizeof(*sl)) {
			*sl++ = cl;
			count -= sizeof(*sl);
		}
	}
	/* fill 8 bits at a time */
	s8 = (char *)sl;
	while (count--)
		*s8++ = c;

	return s;
}

/**
 * memcmp - Compare two areas of memory
 * @cs: One area of memory
 * @ct: Another area of memory
 * @count: The size of the area.
 */
int cryptohw_memcmp(const void * cs,const void * ct,size_t count)
{
	const unsigned char *su1, *su2;
	int res = 0;

	for( su1 = cs, su2 = ct; 0 < count; ++su1, ++su2, count--)
		if ((res = *su1 - *su2) != 0)
			break;
	return res;
}



