/*
 * DSPG DVF9A core driver including access to the DAIF interface
 *
 * Copyright (c) 2012, DSP Group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include "common.h"
#include "hardware.h"
#include "cmu.h"
#include "syscfg.h"

#define DAIF_WAPU_TRAN_MASK		(1<<15)
#define DAIF_RAPU_NEW_VALUE		(1<<15)

#define DAIF_RAPU			41  // 0x00a4
#define DAIF_WAPU			64  // 0x0100
#define DAIF_MODE			117 // 0x01d4

#define DAIF_INDIRECT_START		128 // offset to Analog reg 0

#define DAIF_ANA_DATA_WR_LSB		65
#define DAIF_ANA_DATA_WR_MSB		66
#define DAIF_ANA_ADDR_WR		67
#define DAIF_ANA_ADDR_RD		68
#define DAIF_ANA_DATA_RD_LSB		69
#define DAIF_ANA_DATA_RD_MSB		70

#define DAIF_MODE_RAPU_PRIO		(1<<2)
#define DAIF_MODE_WAPU_PRIO		(1<<3)

#define DAIF_BASE			0x06300000

static int
daif_is_indirect_analog(uint16_t reg)
{
	if (reg >= DAIF_INDIRECT_START)
		return 1;
	return 0;
}

static uint16_t
daif_readw(uint16_t reg)
{
	return readw(DVF99_DAIF_BASE + (reg * 4));
}

static void
daif_writew(uint16_t val, uint16_t reg)
{
	writew(val, DVF99_DAIF_BASE + (reg * 4));
}

static void
daif_wait_wapu_tran(void)
{
	while (daif_readw(DAIF_WAPU) & DAIF_WAPU_TRAN_MASK)
		;
}

static uint16_t
daif_read_via_rapu(uint16_t reg)
{
	uint16_t val;

	/* waiting for resource */
	//while (!(daif_readw(DAIF_RAPU) & DAIF_RAPU_NEW_VALUE))
		;

	daif_writew(reg, DAIF_RAPU);

	while (!((val = daif_readw(DAIF_RAPU)) & DAIF_RAPU_NEW_VALUE))
		;

	/* Check if indicated register address (RAPU[14:8]) matches */
	if (((val & 0x7f00) >> 8) == reg)
		return val & 0xff;
	else
		return daif_read_via_rapu(reg); /* repeat request */
}

static void
daif_write_via_wapu(uint16_t reg, uint16_t value)
{
	daif_wait_wapu_tran();
	daif_writew(((reg & 0x7f) << 8) | (value & 0xff), DAIF_WAPU);
	daif_wait_wapu_tran();
}

uint16_t
daif_read(uint16_t reg)
{
	uint16_t ret = 0;

	if (daif_is_indirect_analog(reg)) {
		/* set required address */
		reg -= DAIF_INDIRECT_START;
		daif_write_via_wapu(DAIF_ANA_ADDR_RD, reg >> 1);
		if (daif_readw(DAIF_MODE) & DAIF_MODE_RAPU_PRIO)
			daif_wait_wapu_tran();

		/* read value */
		ret  = daif_read_via_rapu(DAIF_ANA_DATA_RD_MSB) << 8;
		ret |= daif_read_via_rapu(DAIF_ANA_DATA_RD_LSB) & 0xff;
	} else {
		ret = daif_read_via_rapu(reg);
	}

	return ret;
}

void
daif_write(uint16_t reg, uint16_t value)
{
	if (daif_is_indirect_analog(reg)) {
		/* write value msb, lsb */
		daif_write_via_wapu(DAIF_ANA_DATA_WR_MSB, value >> 8);
		daif_write_via_wapu(DAIF_ANA_DATA_WR_LSB, value & 0xff);

		/* set required address */
		reg -= DAIF_INDIRECT_START;
		daif_write_via_wapu(DAIF_ANA_ADDR_WR, reg >> 1);
	} else {
		daif_write_via_wapu(reg, value);
	}
}

void
daif_init(void)
{
	reset_release(RESET_PMU);
	clk_enable(CLOCK_PMU);
	clk_enable(CLOCK_APU);
}
