/*
 * (C) Copyright 2013
 * DSPG Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/* support printf function */
#include "stdarg.h"
#include "vsnprintf.h"

/* mapping getc and putc to terminal char input and output */
#include "serial.h"
#define getc serial_getc
#define putc serial_putc

/*************************************************************************************
 *  Global functions
 *************************************************************************************/
/*!
*****************************************************************************

 \fn int printf(const char* format, ...);

 <b>Description:</b>\n
 Print a string to the terminal

 \param[in] format	The format string to use

 \return :: length of the print string

******************************************************************************/
#define MAX_PRINTF_LENGTH 256
int printf(const char * format, ...)
{
	int i = 0;
	va_list ap;
	char buffer[MAX_PRINTF_LENGTH];

	va_start(ap, format);
	vsnprintf(buffer, MAX_PRINTF_LENGTH, format, ap);
	va_end(ap);

	for (i=0; i<MAX_PRINTF_LENGTH; i++ ) 
	{
		if ( buffer[i] == 0 )
			break;
		putc( buffer[i] );
	}

	return i;
}

/*!
*****************************************************************************

 \fn int read(char* input);

 <b>Description:</b>\n
 read string input from the terminal

 \param[in] input	input buffer to hold the string

 \return :: numeric value in unsigned long

******************************************************************************/
#define BUFFER_SIZE 256
static char buffer[BUFFER_SIZE] = {0};

int read(char * input)
{
	char key[2]={0}, finished=0, control=0;
	int len=0, i=0;

	/* terminate input first */
	*input = 0;
	while (!finished)
	{
		/* read the input char */
		key[1]=key[0];
		key[0]=getc();
		if (control)
		{
			/* arrow up key */
			if (key[0]=='A')
			{
				/* clear input char on terminal */
				for (i=0; i<len; i++)
				{
					putc('\b');
					putc(' ');
					putc('\b');
				}
				/* copy and print previous buffer on terminal */
				for (len=0; buffer[len] && len<(BUFFER_SIZE-1); len++)
					putc(input[len]=buffer[len]);
				input[len]=0;
			}
			/* clear control status */
			control = 0;
		}
		else if ( key[1]==27 && key[0]==91 )
		{
			/* this is a control key */
			control = 1;
		}
		else if ( key[0]=='\b' )
		{
			if (len>0)
			{
				/* erase one char on terminal */
				putc('\b');
				putc(' ');
				putc('\b');
				input[--len]=0;
			}
		}
		else if ( key[0]=='\r' || key[0]=='\n' )
		{
			/* print linefeed return on terminal */
			putc('\r');
			putc('\n');
			input[len]=0;
			for (len=0; input[len] && len<(BUFFER_SIZE-1); len++)
				buffer[len]=input[len];
			buffer[len]=0;
			/* stop receiving input */
			finished = 1;
		}
		else if ( key[0] >= ' ' && key[0] <= '~' )
		{
			/* print input key on terminal */
			putc(input[len++]=key[0]);
		}
	}
	return len;
}

