/*
 * (C) Copyright 2012
 * DSPG Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include "common.h"
#include "romapi.h"
#include "bootrom.h"
#include "serial.h"
#include "load.h"
#include "cryptohw.h"

static struct image_header header;

static void memcpy(void *d, void *s, unsigned int size)
{
	unsigned int i;

	for (i = 0; i < size; i++) {
		((uint8_t *)d)[i] = ((uint8_t *)s)[i];
	}
}

void *load_loader(unsigned long loader_addr, int secureboot, int fallback) {
	uint32_t crc;
	void *entry;
	int ret;
	const struct bootrom_calls *romapi = romapi_get();

	if (!fallback) {
		ret = romapi->bootdev_read(&header, loader_addr, sizeof(header));
		if (ret) {
			serial_puts("error reading header\n");
			return ERR_PTR(-1);
		}
	} else {
		memcpy(&header, (void *)loader_addr, sizeof(header));
	}

	crc = romapi->crc32(&header, sizeof(header) - sizeof(uint32_t));
	if (crc != header.crc32) {
		serial_puts("error in header\n");
		return ERR_PTR(-1);
	}

	/* if secure boot, verify header signature */
	if (secureboot) {
		serial_puts(" * verifying loader header... ");
		ret = romapi->sboot_check_hdr(&header);
		if (ret) {
			serial_puts("error in sboot_check_hdr\n");
			return ERR_PTR(-1);
		}
		serial_puts("ok\n");
	}

	entry = (void *)header.load_addr;

	if (!fallback) {
		ret = romapi->bootdev_read(entry, loader_addr + header.load_offset, header.load_size);
		if (ret) {
			serial_puts("error reading loader\n");
			return ERR_PTR(-1);
		}
	} else {
		memcpy(entry, (void *)(loader_addr + header.load_offset), header.load_size);
	}

	/* if secure boot verify loader sha256 */
	if (secureboot) {
		serial_puts(" * verifying loader image... ");
#ifdef USE_CRYPTOHW
		/* digest should be located on the DDR.
		 * look in the Makefile for the address */
		unsigned char *digest =
			(unsigned char*)(CONFIG_CRYPTOHW_MEMADDR + sizeof(struct cryptohw_desc));
		ret = cryptohw_sha256(entry, header.load_size, digest);
		if (ret) {
			serial_puts("hw error\n");
			return ERR_PTR(-1);
		}
		ret = cryptohw_memcmp(digest, header.sb_hdr.load_sha256, 32);
#else
		ret = romapi->sboot_check_img(entry, header.load_size,
					header.sb_hdr.load_sha256);
#endif
		if (ret) {
			serial_puts("error\n");
			return ERR_PTR(-1);
		}
		serial_puts("ok\n");
	}

	if (header.flags & HDRFLAG_CRC32) {
		crc = romapi->crc32(entry, header.load_size);
		if (crc != header.load_crc32) {
			serial_puts("error in loader (invalid CRC32)\n");
			return ERR_PTR(-1);
		}
	}

	return entry;
}

long load_find_loader(unsigned long start, unsigned long end)
{
	uint32_t crc;
	int ret;
	unsigned long addr = start;
#ifdef CONFIG_GXP
    unsigned long jump = 0x10000; // try locating the image every 64KB 
#else
	unsigned long jump = 0x40000; // try locating the image every 256KB
#endif
	const struct bootrom_calls *romapi = romapi_get();

	/* scan the boot device, look for a valid u-boot header */
	while (addr < end) {
		ret = romapi->bootdev_read(&header, addr, sizeof(header));
		if (ret) {
			serial_puts("error reading header\n");
			return ERR_IO;
		}

		crc = romapi->crc32(&header, sizeof(header) - sizeof(uint32_t));

		if (header.crc32 == crc &&
		    header.load_size > 0) {
			return addr;
		}

		addr += jump;
	}

	return ERR_NFOUND;
}
