/*
 * mkbootimg - Boot Image Builder
 *
 *  Copyright (C) 2010-2012 DSP Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdarg.h>
#include <getopt.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <errno.h>
#ifdef SECUREBOOT
#include <bfd.h>

#include <openssl/rsa.h>
#include <openssl/pem.h>
#include <openssl/sha.h>
#include <openssl/err.h>
#endif

#undef DEBDEBUG

struct pubkey_t {
	uint8_t n[256];
	uint32_t e;
} __attribute__ ((packed));

struct secure_header {
	struct pubkey_t pubkey;		//0x24
	uint8_t pre_sha256[32];		//0x128
	uint8_t load_sha256[32];	//0x148
	uint8_t signature[256];		//0x168
} __attribute__ ((packed));

struct image_header {
	uint32_t size;			//0x0
	uint32_t flags;			//0x4
	uint32_t pre_offset;	//0x8
	uint32_t pre_size;		//0xc
	uint32_t pre_crc32;		//0x10
	uint32_t load_offset;	//0x14
	uint32_t load_size;		//0x18
	uint32_t load_addr;		//0x1c
	uint32_t load_crc32;	//0x20
	struct secure_header sb_hdr; //0x24
	uint32_t crc32;			//0x268
} __attribute__ ((packed));

#define HDRFLAG_CRC32  0x00000001
#define HDRFLAG_SECURE 0x00000002

/*
 * CRC calculation from the Linux kernel, lib/crc32.c:
 */
#include "crc32table.h"
#define DO_CRC(x) crc = tab[ (crc ^ (x)) & 255 ] ^ (crc>>8)

/**
 * crc32_le() - Calculate bitwise little-endian Ethernet AUTODIN II CRC32
 * @crc: seed value for computation.  ~0 for Ethernet, sometimes 0 for
 *	other uses, or the previous crc32 value if computing incrementally.
 * @p: pointer to buffer over which CRC is run
 * @len: length of buffer @p
 */
static uint32_t crc32_le(uint32_t crc, uint32_t const *b, uint32_t len)
{
	const uint8_t *p = (uint8_t *)b;
	const unsigned long *tab = crc32table_le;

	/* Align it */
	if(((long)b)&3 && len){
		do {
			uint8_t *p = (uint8_t *)b;
			DO_CRC(*p++);
			b = (void *)p;
		} while ((--len) && ((long)b)&3 );
	}
	if(len >= 4){
		/* load data 32 bits wide, xor data 32 bits wide. */
		uint32_t save_len = len & 3;
	        len = len >> 2;
		--b; /* use pre increment below(*++b) for speed */
		do {
			crc ^= *++b;
			DO_CRC(0);
			DO_CRC(0);
			DO_CRC(0);
			DO_CRC(0);
		} while (--len);
		b++; /* point to next byte(s) */
		len = save_len;
	}
	/* And the last few bytes */
	if(len){
		do {
			uint8_t *p = (uint8_t *)b;
			DO_CRC(*p++);
			b = (void *)p;
		} while (--len);
	}

	return crc;
}

/* standard crc32 checksum */
static uint32_t calculate_crc32(void *buffer, uint32_t size)
{
	return crc32_le(~0u, (uint32_t const *)buffer, size);
}

/*
 * mkbootimg implementation
 */
#define align4(a) (((a) + 3) & ~3)

char *cmd;
char *fname_pre;
char *fname_load;
char *fname_key;
char *fname_uboot_elf;
char *fname_out = "bootloader.img";
uint32_t load_addr = 0x41000000;
uint32_t flags;

char *fname_key2;
int do_psign = 0;
int do_gsign = 0;
int do_gcode = 0;

enum options {
	OPTION_PRELOADER,
	OPTION_LOADER,
	OPTION_LOAD_ADDR,
	OPTION_CRC32,
	OPTION_SIGN,
	OPTION_PATCH_UBOOT_PUBKEY,
	OPTION_OUTPUT,
/* Options for testing only */
	OPTION_PSIGN,
	OPTION_GSIGN,
	OPTION_GCODE,
};

struct option long_opts[] = {
	{ "help",          0, 0, 'h' },
	{ "preloader",     1, 0, OPTION_PRELOADER },
	{ "loader",        1, 0, OPTION_LOADER },
	{ "loader-addr",   1, 0, OPTION_LOAD_ADDR },
	{ "crc32",         0, 0, OPTION_CRC32 },
#ifdef SECUREBOOT
	{ "sign",          1, 0, OPTION_SIGN },
	{ "patch-uboot",   1, 0, OPTION_PATCH_UBOOT_PUBKEY},
#endif
	{ "output",        1, 0, OPTION_OUTPUT },
#ifdef SECUREBOOT
/* Options for testing only */
	{ "psign",	   1, 0, OPTION_PSIGN},
	{ "gsign",	   0, 0, OPTION_GSIGN},
	{ "gcode",	   0, 0, OPTION_GCODE},
#endif
	{ NULL,            0, 0, 0 },
};

static void hexdump(uint8_t* buff, unsigned int len)
{
	int i;

	for (i=0; i<len; i++) {
		printf("%02X", buff[i]);
	}
	printf("\n");
}

static void die(const char *fmt, ...)
{
	va_list params;

	va_start(params, fmt);
	vfprintf(stderr, fmt, params);
	va_end(params);
	exit(EXIT_FAILURE);
}

static void die_file(const char *fname)
{
	die("'%s': %s\n", fname, strerror(errno));
}

#ifdef SECUREBOOT
static void bfd_fatal(const char *string)
{
	const char *errmsg = bfd_errmsg(bfd_get_error());

	die("%s: %s\n", string, errmsg);
}
#endif

static void usage(void)
{
	fprintf(stderr, "Usage: %s <opts>\n"
			"  -h, --help\n"
			"      --preloader=<file>\n"
			"      --loader=<file>\n"
			"      --loader-addr=<addr>\n"
			"      --crc32\n"
#ifdef SECUREBOOT
			"      --sign=<file.pem>\n"
			"      --patch-uboot=<u-boot elf>  patch u-boot image with a valid public\n"
			"                                 key digest for securedev command.\n"
			"                                 (must be used together with --sign)\n"
#endif
			"      --output=<file>\n"
#ifdef SECUREBOOT
			"\n"
			"  options for testing only:\n"
			"      --psign=<file.pem>         Sign with a different private key.\n"
			"      --gsign                    Put garbage in signature field.\n"
			"      --gcode                    Replace code with garbage after signning.\n"
#endif
	        , cmd);
	exit(EXIT_FAILURE);
}

static void parse_cmdline(int argc, char *argv[])
{
	int ret;
	int optnum;
	cmd = argv[0];

	while (1) {
		ret = getopt_long(argc, argv, "h", long_opts, &optnum);
		if (ret < 0)
			break;

		/*
		printf("%s(): got param '%s' with argument '%s'\n", __func__,
		       long_opts[optnum].name, optarg);
		      */

		switch (ret) {
		case OPTION_PRELOADER:
			fname_pre = strdup(optarg);
			break;
		case OPTION_LOADER:
			fname_load = strdup(optarg);
			break;
		case OPTION_LOAD_ADDR:
			load_addr = strtoul(optarg, NULL, 0);
			break;
		case OPTION_CRC32:
			flags |= HDRFLAG_CRC32;
			break;
#ifdef SECUREBOOT
		case OPTION_SIGN:
			flags |= HDRFLAG_SECURE;
			fname_key = strdup(optarg);
			break;
		case OPTION_PATCH_UBOOT_PUBKEY:
			fname_uboot_elf = strdup(optarg);
			break;
#endif
		case OPTION_OUTPUT:
			fname_out = strdup(optarg);
			break;
#ifdef SECUREBOOT
		case OPTION_PSIGN:
			fname_key2 = strdup(optarg);
			do_psign = 1;
			break;
		case OPTION_GSIGN:
			do_gsign = 1;
			break;
		case OPTION_GCODE:
			do_gcode = 1;
			break;
#endif
		case 'h':
		default:
			usage();
		}
	}

	/* abort if OPTION_PATCH_UBOOT_PUBKEY without OPTION_SIGN */
	if (fname_uboot_elf && (!fname_key)) {
		fprintf(stderr, "error: --patch-uboot must be used together with --sign\n");
		usage();
	}
}

off_t get_filesize(int fd)
{
	struct stat s;
	int ret;

	ret = fstat(fd, &s);
	if (ret < 0)
		die("%s(): stat: %s\n", __func__, strerror(errno));

	return s.st_size;
}

int copy_file(int fd_in, int fd_out)
{
	char buf[4096];
	int ret;

	do {
		ret = read(fd_in, buf, sizeof(buf));
		if (ret < 0)
			goto out;

		ret = write(fd_out, buf, ret);
	} while (ret > 0);

out:
	if (ret < 0)
		die("%s: %s\n", __func__, strerror(errno));
}

void file_align4(int fd)
{
	off_t pos = lseek(fd, 0, SEEK_CUR);
	int pad = pos % 4;
	char buf[3] = { 0xff, 0xff, 0xff };
	int ret = 0;

	if (pad)
		ret = write(fd, buf, pad);

	if (ret < 0)
		die("%s: error while padding\n", __func__);
}

static struct image_header h;

void cleanup(int status, void *arg)
{
	if (status == EXIT_SUCCESS)
		return;

	/* delete output file */
	unlink(fname_out);
}

int EMSA_encode_sha256(uint8_t *header_encode, uint8_t *header_dgst)
{
	int i;
	uint8_t T[51];
	uint8_t SHA256_ID[19] = {0x30, 0x31, 0x30, 0x0d, 0x06, 0x09, 0x60, 0x86,
		0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x01, 0x05, 0x00, 0x04, 0x20};

	if (!header_encode || !header_dgst)
		return -1;

	// copy digest algo id
	memcpy(T, SHA256_ID, 19);
	// copy sha256 digest of message
	memcpy(T+19, header_dgst, 32);

	*(header_encode++) = 0x00;
	*(header_encode++) = 0x01;
	for (i=0; i<256-sizeof(T)-3; i++) {
		*(header_encode++) = 0xFF;
	}
	*(header_encode++) = 0x00;
	memcpy(header_encode, T, sizeof(T));

	return 0;
}

#ifdef SECUREBOOT
int pem_rsa_init(RSA **rsa, char *keyfile)
{
	FILE *fp_key;

	/* essential for private key read and sha256 */
	OpenSSL_add_all_algorithms();
	ERR_load_crypto_strings();

	fp_key = fopen(keyfile, "r");
	if (fp_key == NULL) {
		fprintf(stderr, "error: cannot open file %s.\n", keyfile);
		return -1;
	}

	fprintf(stdout, "reading key from %s\n", keyfile);
	if ((*rsa = PEM_read_RSAPrivateKey(fp_key, NULL, 0, NULL)) == NULL) {
		fprintf(stderr, "error: cannot load private key from %s. (%s)\n",
				keyfile, ERR_error_string(ERR_get_error(), NULL));
		fclose(fp_key);
		return -1;
	}

	fclose(fp_key);

	/* verify public key */
	if (BN_num_bytes((*rsa)->n) != 256 || BN_num_bytes((*rsa)->e) > 4) {
		fprintf(stderr, "error: illegal public key\n");
		return -1;
	}

	return 0;
}

void pem_rsa_free(RSA **rsa)
{
	RSA_free(*rsa);
}

/*
 * uboot_pubkey_get_address
 *
 * Get the address of 'securedev_pubkey' symbol in u-boot elf file
 * to patch it later with the the public key digest value.
 */

int uboot_pubkey_get_address(char* uboot_elf_fn, unsigned long *addr)
{
	bfd *abfd;
	asymbol **sy;
	int storage;
	int symcount;
	int i;
	unsigned long ram_base = 0;
	unsigned long pubkey_addr = 0;

	bfd_init();
	bfd_set_default_target("i686-pc-linux-gnu");

	abfd = bfd_openr(uboot_elf_fn, NULL);
	if (!abfd)
		die("'%s': cannot bfd_openr()\n", uboot_elf_fn);

	if (bfd_check_format(abfd, bfd_object) == 0)
		die("'%s': not a bfd_object file\n", uboot_elf_fn);

	storage = bfd_get_symtab_upper_bound(abfd);
	if (storage < 0)
		bfd_fatal(bfd_get_filename(abfd));
	if (storage)
		sy = malloc(storage);

	symcount = bfd_canonicalize_symtab(abfd, sy);
	if (symcount < 0)
		bfd_fatal(bfd_get_filename(abfd));
	if (symcount == 0)
		die("'%s': no symbols", bfd_get_filename(abfd));

	for (i = 0; i < symcount; i++) {
		if (strcmp(bfd_asymbol_name(sy[i]), "securedev_pubkey") == 0)
		{
			pubkey_addr = bfd_asymbol_value(sy[i]);
		}
		else if (strcmp(bfd_asymbol_name(sy[i]), "__image_copy_start") == 0) {
			ram_base = bfd_asymbol_value(sy[i]);
		}
	}
	free(sy);
	bfd_close(abfd);

	if (!pubkey_addr) {
		fprintf(stderr, "error: cannot find symbol 'securedev_pubkey'!\n");
		return -1;
	}
	if (!ram_base) {
		fprintf(stderr, "error: cannot find symbol '__image_copy_start'!\n");
		return -1;
	}

	*addr = pubkey_addr - ram_base;
	return 0;
}

/*
 * uboot_pubkey_patch
 *
 * Patch u-boot pubkey (symbol 'securedev_pubkey')
 * with public key digest from key file
 */

int uboot_pubkey_patch(RSA *rsa, struct image_header *header, int fd_out)
{
	unsigned char *image;
	struct pubkey_t pubkey;
	uint32_t sha256_pubkey[8];
	unsigned long pubkey_address;
	uint32_t *pubkey_p;
	int i, ret;

	/* calculate sha256 of public key */
	BN_bn2bin(rsa->n, pubkey.n);
	pubkey.e = 0;
	BN_bn2bin(rsa->e, (unsigned char*)&(pubkey.e));

	if (SHA256((unsigned char*)&pubkey, 260, (unsigned char*)sha256_pubkey) == NULL) {
		fprintf(stderr, "error: failed calculate SHA256\n");
		return -1;
	}

	ret = uboot_pubkey_get_address(fname_uboot_elf, &pubkey_address);
	if (ret < 0) {
		fprintf(stderr, "error: cannot find u-boot pubkey address\n");
		return -1;
	}

	// map output image and patch u-boot pubkey)
	image = mmap(NULL, 10 << 20, PROT_WRITE, MAP_SHARED, fd_out, 0);
	if (image == NULL) {
		fprintf(stderr, "error: cannot map output file\n");
		return -1;
	}

	pubkey_p = (uint32_t *)((unsigned long)image + header->load_offset + pubkey_address);
	for (i=0; i<8; i++) {
		pubkey_p[i] = sha256_pubkey[i];
	}

	munmap(image, 10<<20);

	return 0;
}

/*
 * Secure Boot image sign
 *
 * Sign the image header with RSA private key contained in
 * the supplied file (in PEM format) according to PKCS1.5
 * EMSA padding scheme.
 */

int sign_image(RSA *rsa, struct image_header *header, int fd_out)
{
	uint8_t header_encode[256];
	uint8_t header_dgst[32];
	size_t size;
	unsigned char *image;

	/* write public key to header */
	BN_bn2bin(rsa->n, header->sb_hdr.pubkey.n);
	BN_bn2bin(rsa->e, (unsigned char*)&header->sb_hdr.pubkey.e);

	/* calculate and write SHA256 of preloader and loader */
	image = mmap(NULL, 10 << 20, PROT_READ, MAP_PRIVATE, fd_out, 0);
	if (image == NULL) {
		fprintf(stderr, "error: cannot map output file\n");
		return -1;
	}

	if (header->pre_size > 0)
		SHA256(image + header->pre_offset, header->pre_size,
			   header->sb_hdr.pre_sha256);

	if (header->load_size > 0)
		SHA256(image + header->load_offset, header->load_size,
			   header->sb_hdr.load_sha256);

	munmap(image, 10<<20);

	/* compute sha256 on image header*/
	size = (size_t)header->sb_hdr.signature - (size_t)header;
	SHA256((unsigned char *)header, size, header_dgst);

	/* encode(pad) the sha256 digest according to PKCS1-V1.5 EMSA scheme */
	if (EMSA_encode_sha256(header_encode, header_dgst) != 0) {
		fprintf(stderr, "error: error padding header digest\n");
		return -1;
	}

#ifdef DEBDEBUG
	printf("header encode:\n");
	hexdump(header_encode, 256);
#endif

	/* sign the PKCS EMSA padded sha256 digest of header */
	if (RSA_private_encrypt(256, header_encode, header->sb_hdr.signature, rsa, RSA_NO_PADDING) < 0) {
		fprintf(stderr, "error: cannot sign encoded message (%s)\n",
				ERR_error_string(ERR_get_error(), NULL));
		return -1;
	}

	/*********************/
	/** for testing only */
	/*********************/
	/* if psign is selected, override signature with different key */
	if (do_psign)
	{
		RSA *rsa2;
		FILE *fp_key2;

		fp_key2 = fopen(fname_key2, "r");
		if (fp_key2 == NULL) {
			fprintf(stderr, "error: cannot open file %s.\n", fname_key2);
			return -1;
		}

		fprintf(stdout, "reading key from %s\n", fname_key2);
		if ((rsa2 = PEM_read_RSAPrivateKey(fp_key2, NULL, 0, NULL)) == NULL) {
			fprintf(stderr, "error: cannot load private key from %s. (%s)\n",
					fname_key2, ERR_error_string(ERR_get_error(), NULL));
			fclose(fp_key2);
			return -1;
		}

		fclose(fp_key2);

		if (RSA_private_encrypt(256, header_encode, header->sb_hdr.signature, rsa2, RSA_NO_PADDING) < 0) {
			fprintf(stderr, "error: cannot sign encoded message (%s)\n",
					ERR_error_string(ERR_get_error(), NULL));
			return -1;
		}
	}

	/* if gsign is selected, override signature with garbage */
	if (do_gsign)
	{
		int fd = open("/dev/urandom", O_RDONLY);
		if (fd < 0) {
			fprintf(stderr, "error: cannot open /dev/urandom\n");
			return -1;
		}

		size = read(fd, header->sb_hdr.signature, 256);
		if (size != 256) {
			fprintf(stderr, "error: failed to write 256 garbage bytes as signature\n");
			return -1;
		}

		close(fd);
	}
	/*************************/
	/** END for testing only */
	/*************************/


#ifdef DEBDEBUG
	uint8_t test[256];
	printf("signature:\n");
	hexdump(header->sb_hdr.signature, 256);

	RSA_public_decrypt(256, header->sb_hdr.signature, test, rsa, RSA_NO_PADDING);
	printf("test:\n");
	hexdump(test, 256);
#endif

	return 0;
}
#endif

int main(int argc, char *argv[])
{
	int i, c;
	int fd_out, fd_in;
	unsigned int totalsize;
	int ret;
	char *image, *p;
#ifdef SECUREBOOT
	RSA *rsa = NULL;
#endif

	parse_cmdline(argc, argv);

	on_exit(cleanup, NULL);

	fd_out = open(fname_out, O_CREAT|O_TRUNC|O_RDWR, 0666);
	if (fd_out < 0)
		die("'%s': %s\n", strerror(errno));

	/* header is written last */
	ret = lseek(fd_out, sizeof(h), SEEK_SET);
	if (ret < 0)
		die_file(fname_out);

#ifdef SECUREBOOT
	/* init RSA structure from pem file (if needed) */
	if (fname_uboot_elf || fname_key) {
		ret = pem_rsa_init(&rsa, fname_key);
		if (ret < 0)
			die("pem_rsa_init failed!\n");
	}
#endif

	/*
	 * 1. write contents from external files
	 */

	file_align4(fd_out);

	if (fname_pre) {
		fd_in = open(fname_pre, O_RDONLY);
		if (fd_in < 0)
			die_file(fname_pre);

		h.pre_offset = lseek(fd_out, 0, SEEK_CUR);
		h.pre_size = get_filesize(fd_in);
		h.pre_size = align4(h.pre_size);

		copy_file(fd_in, fd_out);
		close(fd_in);
	}

	file_align4(fd_out);

	if (fname_load) {
		fd_in = open(fname_load, O_RDONLY);
		if (fd_in < 0)
			die_file(fname_load);

		h.load_offset = lseek(fd_out, 0, SEEK_CUR);
		h.load_size = get_filesize(fd_in);
		h.load_size = align4(h.load_size);
		h.load_addr = load_addr;

		copy_file(fd_in, fd_out);
		close(fd_in);
	}

	file_align4(fd_out);

#ifdef SECUREBOOT
	/*
	 * 2. patch u-boot image data with public key
	 */
	if (fname_uboot_elf) {
		ret = uboot_pubkey_patch(rsa, &h, fd_out);
		if (ret < 0)
			die("error: error patching u-boot pubkey\n");
	}
#endif

	/*
	 * 3. write header size and flags
	 */

	h.size = sizeof(h);
	h.flags = flags;

	ret = lseek(fd_out, 0, SEEK_SET);
	if (ret < 0)
		die_file(fname_out);

	/*
	 * 4. calculate crc32 of preloader and/or loader image
	 */
	if (h.flags & HDRFLAG_CRC32) {
		image = mmap(NULL, 10 << 20, PROT_READ, MAP_PRIVATE, fd_out, 0);
		if (image == MAP_FAILED)
			die_file(fname_out);

		if (fname_pre) {
			p = image + h.pre_offset;
			h.pre_crc32 = calculate_crc32(p, h.pre_size);
		}

		if (fname_load) {
			p = image + h.load_offset;
			h.load_crc32 = calculate_crc32(p, h.load_size);
		}
	}

#ifdef SECUREBOOT
	/*
	 * 5. sign header with RSA key
	 */
	if (h.flags & HDRFLAG_SECURE) {
		if (sign_image(rsa, &h, fd_out) != 0) {
			fprintf(stderr, "error: image cannot be signed!\n");
			h.flags &= ~HDRFLAG_SECURE;
		}
	}
#endif

	/*
	 * 6. calculate crc32 of header
	 */
	h.crc32 = calculate_crc32(&h, sizeof(h) - sizeof(h.crc32));

#ifdef SECUREBOOT
	if (rsa)
		pem_rsa_free(&rsa);

	/*********************/
	/** for testing only */
	/*********************/
	/* if gcode is selected override code with garbage */
	if (do_gcode) {
		int fd, size;
		char tmp_byte;

		ret = lseek(fd_out, h.pre_offset, SEEK_SET);
		if (ret < 0)
			die_file(fname_out);

		fd = open("/dev/urandom", O_RDONLY);
		if (fd < 0) {
			fprintf(stderr, "error: cannot open /dev/urandom\n");
			return -1;
		}

		size = h.pre_size;
		while (size) {
			if (read(fd, &tmp_byte, 1) <= 0)
				break;

			if (write(fd_out, &tmp_byte, 1) <= 0)
				break;

			size--;
		}
		if (size > 0) {
			fprintf(stderr, "error: failed to replace preloader code with garbage\n");
			return -1;
		}

		close(fd);
	}
	/*************************/
	/** END for testing only */
	/*************************/
#endif

	/*
	 * write the header
	 */
	ret = lseek(fd_out, 0, SEEK_SET);
	if (ret < 0)
		die_file(fname_out);

	ret = write(fd_out, &h, sizeof(h));
	if (ret < 0)
		die_file(fname_out);
	if (ret < sizeof(h))
		die("'%s': could not write full header\n", fname_out);

	return 0;
}
