/*
 * (C) Copyright 2012
 * DSPG Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include "cmu.h"
#include "common.h"
#include "gpio.h"
#include "hardware.h"
#include "romapi.h"
#include "syscfg.h"

#include "dram_settings.h"

#define MPMCControl              0x00000000
#define MPMCStatus               0x00000004
#define MPMCConfig               0x00000008
#define MPMCDynamicControl       0x00000020
#define MPMCDynamicRefresh       0x00000024
#define MPMCDynamicReadConfig    0x00000028
#define MPMCDynamictRP           0x00000030
#define MPMCDynamictRAS          0x00000034
#define MPMCDynamictSREX         0x00000038
#define MPMCDynamictWR           0x00000044
#define MPMCDynamictRC           0x00000048
#define MPMCDynamictRFC          0x0000004c
#define MPMCDynamictXSR          0x00000050
#define MPMCDynamictRRD          0x00000054
#define MPMCDynamictMRD          0x00000058
#define MPMCDynamictCDLR         0x0000005c
#define MPMCStaticExtendedWait   0x00000080
#define MPMCDynamicConfig0       0x00000100
#define MPMCDynamicRasCas0       0x00000104
#define MPMCDynamicConfig1       0x00000120
#define MPMCDynamicRasCas1       0x00000124
#define MPMCDynamicConfig2       0x00000140
#define MPMCDynamicRasCas2       0x00000144
#define MPMCDynamicConfig3       0x00000160
#define MPMCDynamicRasCas3       0x00000164
#define MPMCAHBTimeout0          0x00000408
#define MPMCAHBTimeout1          0x00000428
#define MPMCAHBTimeout2          0x00000448
#define MPMCAHBTimeout3          0x00000468
#define MPMCAHBTimeout4          0x00000488
#define MPMCAHBTimeout5          0x000004a8

static void mpmc_writel(uint32_t val, int offset)
{
	writel(val, DVF99_MPMC_BASE + offset);
}

int dram_init(void)
{
	int i;
	const struct bootrom_calls *romapi = romapi_get();

	reset_release(BLOCK_MPMC);
	clk_enable(BLOCK_MPMC);

	romapi->udelay(100);

	/* enable MEMBOOT_PAD_EN */
	syscfg_memboot_pad_enable();
	syscfg_sdram_slew_rate(3);

	/* disable GPIOs */
	gpio_disable(CGPIO(16)); /* MEMXA0 */
	gpio_disable(CGPIO(17)); /* MEMXA1 */
	gpio_disable(BGPIO(26)); /* MEMXA2 */
	gpio_disable(CGPIO(19)); /* MEMXA3 */
	gpio_disable(CGPIO(20)); /* MEMXA4 */
	gpio_disable(CGPIO(21)); /* MEMXA5 */
	gpio_disable(CGPIO(22)); /* MEMXA6 */
	gpio_disable(CGPIO(23)); /* MEMXA7 */
	gpio_disable(CGPIO(24)); /* MEMXA8 */
	gpio_disable(CGPIO(25)); /* MEMXA9 */
	gpio_disable(CGPIO(26)); /* MEMXA10 */
	gpio_disable(CGPIO(27)); /* MEMXA11 */
	gpio_disable(CGPIO(18)); /* MEMXA12 */
	gpio_disable(DGPIO(1));  /* MEMXA13 */
	gpio_disable(DGPIO(2));  /* MEMXA14 */
	gpio_disable(CGPIO(0));  /* MEMXD0 */
	gpio_disable(CGPIO(1));  /* MEMXD0 */
	gpio_disable(CGPIO(2));  /* MEMXD0 */
	gpio_disable(CGPIO(3));  /* MEMXD0 */
	gpio_disable(CGPIO(4));  /* MEMXD0 */
	gpio_disable(CGPIO(5));  /* MEMXD0 */
	gpio_disable(CGPIO(6));  /* MEMXD0 */
	gpio_disable(CGPIO(7));  /* MEMXD0 */
	gpio_disable(CGPIO(8));  /* MEMXD0 */
	gpio_disable(CGPIO(9));  /* MEMXD0 */
	gpio_disable(CGPIO(10)); /* MEMXD10 */
	gpio_disable(CGPIO(11)); /* MEMXD11 */
	gpio_disable(CGPIO(12)); /* MEMXD12 */
	gpio_disable(CGPIO(13)); /* MEMXD13 */
	gpio_disable(CGPIO(14)); /* MEMXD14 */
	gpio_disable(CGPIO(15)); /* MEMXD15 */
	gpio_disable(DGPIO(3));  /* MEMCAS */
	gpio_disable(CGPIO(31)); /* MEMCKE */
	gpio_disable(DGPIO(5));  /* MEMDQMBLS0 */
	gpio_disable(CGPIO(28)); /* MEMDQMBLS1 */
	gpio_disable(DGPIO(4));  /* MEMRAS */
	gpio_disable(CGPIO(30)); /* MEMWR */
	gpio_disable(DGPIO(8));  /* MEMCS2 */
	gpio_disable(DGPIO(0));  /* MEMCLK */

	/*
	 * Setup basic controller registers.
	 */
	mpmc_writel(0, MPMCConfig); /* Set little endian */
	mpmc_writel(0x11, MPMCDynamicReadConfig); /* command delayed strategy */

	/*
	 * Let DRAM settle after clock is applied: set SDRAM Initialisation
	 * Value (I) to NOP.
	 */
	mpmc_writel(0x0183, MPMCDynamicControl);
	romapi->udelay(200);

	/*
	 * Precharge all banks and place the device into the all banks idle
	 * state: set SDRAM Initialisation Value (I) to PALL. Then quickly
	 * perform a number of refresh cycles.
	 */
	mpmc_writel(0x103, MPMCDynamicControl);
	mpmc_writel(4, MPMCDynamicRefresh);
	romapi->udelay(1000);

	/*
	 * Program operational values.
	 */
	mpmc_writel(50, MPMCDynamicRefresh); /* TODO: calculate refresh rate */
	mpmc_writel(mpmc_ras_cas_setup, MPMCDynamicRasCas1);
	mpmc_writel(mpmc_config_setup, MPMCDynamicConfig1);
	for (i = 0; i < 3; i++)
		mpmc_writel(mpmc_timing_setup[i], MPMCDynamictRP + i*4);
	for (i = 0; i < 7; i++)
		mpmc_writel(mpmc_timing_setup[i+3], MPMCDynamictWR + i*4);

	/*
	 * Program mode register by first setting SDRAM Initialisation Value
	 * (I) to MODE and then doing a read transaction with the mode register
	 * values as address offset. The actual offset depends on the address
	 * mapping as configured in MPMCDynamicConfigX.
	 */
	mpmc_writel(0x83, MPMCDynamicControl);
	readl(0xC8000000UL + mpmc_mode_setup);
	romapi->udelay(100); /* TODO: really needed? */

	/*
	 * Done: set SDRAM Initialisation Value (I) to NORMAL.
	 */
	mpmc_writel(3, MPMCDynamicControl);

	return 0;
}

#ifdef CONFIG_HT
int ht_dram_init(void)
{
    // HT8xx has two sdram type: 16MB and 32MB
    // try 32MB sdram first
    mpmc_config_setup = 0x680;
	if (0 != dram_init())
        return -1;
    
    writel( 0x12345678, 0xc8000000 );
    writel( 0x5555aaaa, 0xc9000000 );
    if ( readl( 0xc8000000 ) == 0x12345678 &&
         readl( 0xc9000000 ) == 0x5555aaaa )
         return 0;

    // sdram is 16MB
    mpmc_config_setup = 0x480;
	if (0 != dram_init())
        return -1;

    return 0;
}
#endif


