/*
 * (C) Copyright 2012
 * DSPG Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include "common.h"
#include "hardware.h"
#include "serial.h"
#include "cmu.h"
#include "romapi.h"
#include "syscfg.h"
#include "gpio.h"

#define UART1		1
#define UART2_LOC0	2
#define UART2_LOC1	3

static int uart = USE_UART;
static unsigned long uart_base[] =
{ DVF99_UART1_BASE, DVF99_UART1_BASE, DVF99_UART2_BASE, DVF99_UART2_BASE };
static unsigned long uart_pad[] =
{ PAD_UART1, PAD_UART1, PAD_UART2, PAD_UART2 };
static unsigned long uart_tx_pin[] =
{ DGPIO(20), DGPIO(20), EGPIO(16), AGPIO(28) };
static unsigned long uart_rx_pin[] =
{ DGPIO(21), DGPIO(21), EGPIO(17), AGPIO(29) };

#define UART_CTL		0x00
#define UART_CFG		0x04
#define UART_INT_DIV		0x08
#define UART_FRAC_DIV		0x0c
#define UART_TX_FIFO_WM		0x10
#define UART_RX_FIFO_WM		0x14
#define UART_FIFO_ICR		0x20
#define UART_TX_FIFO_LVL	0x24
#define UART_RX_FIFO_LVL	0x28
#define UART_TX_DATA		0x2c
#define UART_RX_DATA		0x30
#define UART_STAT		0x34

#define UART_CFG_8N1		((1 << 11) | (0 << 2))
#define RX_EMPTY		(1<<8)

void serial_enable(void)
{
	writel(1, uart_base[uart] + UART_CTL);
}

void serial_disable(void)
{
	writel(0, uart_base[uart] + UART_CTL);
}

int serial_get_uart_conf(void)
{
	uint8_t ucfg;

#if USE_UART != 0
	return USE_UART;
#endif

	ucfg = (readl(DVF99_CMU_BASE + CMU_RSTSR) >> 8) & 0xff;

	if ((ucfg & 0x11) == 0x11)
		return UART1;
	else if ((ucfg & 0x11) == 0x01)
		return UART2_LOC0;
	else if ((ucfg & 0x11) == 0x10)
		return UART2_LOC1;

	return UART1;
}

void serial_init(unsigned int baudrate)
{
	unsigned int pclk = cmu_get_pclk();
	unsigned int integer, fraction;

	uart = serial_get_uart_conf();

	reset_release(BLOCK_UART1);
	clk_enable(BLOCK_UART1);

	if (!readl(uart_base[uart] + UART_CTL)) {
		pad_enable(uart_pad[uart]);
		gpio_disable(uart_tx_pin[uart]);
		gpio_disable(uart_rx_pin[uart]);
		if (uart == UART2_LOC1)
			pad_enable(PAD_UART2_LOC);
		writel(UART_CFG_8N1, uart_base[uart] + UART_CFG);
	} else {
		serial_disable();
	}

	integer = (pclk / baudrate) >> 4;
	fraction = (pclk / baudrate) & 0xf;

	writel(integer,  uart_base[uart] + UART_INT_DIV);
	writel(fraction, uart_base[uart] + UART_FRAC_DIV);

	serial_enable();
}

void serial_putc_hw(const char c)
{
	/* wait until there is space in TX FIFO */
	while (readl(uart_base[uart] + UART_TX_FIFO_LVL) == 0x80)
		;

	writel(c, uart_base[uart] + UART_TX_DATA);
}

void serial_putc(const char c)
{
	if (c == '\n')
		serial_putc_hw('\r');

	serial_putc_hw(c);
}

void serial_puts(const char *s)
{
	while (*s)
		serial_putc(*s++);
}

int serial_tstc(void)
{
	if (!(readl(uart_base[uart] + UART_STAT) & RX_EMPTY))
		return 1;

	return 0;
}

static int serial_getc_hw(void)
{
	unsigned int data;

	data = readl(uart_base[uart] + UART_RX_DATA);
	if (data & 0x600) /* parity or frame error */
		return -1;

	data &= 0xff; /* limit to 8 characters */

	return (int)data;
}

void serial_flush(void)
{
	while (serial_tstc())
		serial_getc_hw();
}

int serial_chkc(void)
{
	return serial_getc_hw();
}

int serial_getc(void)
{
	while (serial_tstc() == 0)
		;
	return serial_getc_hw();
}

void serial_exit(void)
{
	const struct bootrom_calls *romapi = romapi_get();
	uart = serial_get_uart_conf();

	/* UART running? */
	if (!readl(uart_base[uart] + UART_CTL))
		return;

	/* wait until the TX FIFO is emptied */
	while (readl(uart_base[uart] + UART_TX_FIFO_LVL) > 0)
		;

	/* the last character is still being shifted out */
	romapi->udelay(100);

	serial_disable();
}
