/*
 * (C) Copyright 2013
 * DSPG Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/*************************************************************************************
 *  Global functions
 *************************************************************************************/

/*!
*****************************************************************************

 \fn int strlen (const char* str);

 <b>Description:</b>\n
 Returns the length of the C string str.

 \param[in] s		The input string to length

 \return :: length in integer

******************************************************************************/
int strlen (const char *str)
{
        const char *s;
        for (s = str; *s; ++s)
                ;
        return (s - str);
}

/*!
*****************************************************************************

 \fn int strcmp (const char* s1, const char* s2);

 <b>Description:</b>\n
 Compares the C string s1 to the C string s2.

 \param[in] s1		C string to be compared.
 \param[in] s2		C string to be compared.

 \return :: A zero value indicates that both strings are equal.
	A value greater than zero indicates that the first character that 
        does not match has a greater value in str1 than in str2;
	And a value less than zero indicates the opposite.

******************************************************************************/
int strcmp (const char* s1, const char* s2)
{
	while(*s1 && (*s1==*s2))
		s1++,s2++;
	return *(const unsigned char*)s1-*(const unsigned char*)s2;
}

/*!
*****************************************************************************

 \fn char * strncpy(char* dst, const char* src, int n);

 <b>Description:</b>\n
 Copies the first num characters of source to destination. If the end of the 
 source C string (which is signaled by a null-character) is found before num
 characters have been copied, destination is padded with zeros until a total
 of num characters have been written to it.

 \param[in] dst		Pointer to the destination is to be copied.
 \param[in] src		C string to be copied.
 \param[in] n		maximum number of characters to be copied from source.

 \return :: destination is returned

******************************************************************************/
char * strncpy(char *dst, const char *src, int n)
{
	while (n-- && (*dst++ = *src++)!= '\0')
    		;
	return dst;
}

/*!
*****************************************************************************

 \fn char * strcpy(char* dst, const char* src);

 <b>Description:</b>\n
 Copies the C string pointed by source into the array pointed by destination, 
 including the terminating null character (and stopping at that point).

 \param[in] dst		Pointer to the destination is to be copied.
 \param[in] src		C string to be copied.

 \return :: destination is returned

******************************************************************************/
char * strcpy(char *dst, const char *src)
{
	while((*dst++ = *src++)!= '\0')
		;
	return dst;
}

/*!
*****************************************************************************

 \fn int dec2int ( char* s );

 <b>Description:</b>\n
 Parse decimal value into integer, accepting with or without -+ sign

 \param[in] s		The input string to parse

 \return :: integer value parsed

******************************************************************************/
int dec2int(char * s)
{
	int sum = 0, sign = 1;
	if ( *s == '-' )
	{
		s++;
		sign = -1;
	}
	else if ( *s == '+' ) 
		s++;

	while(*s)
	{
		if ( '0' <= *s && *s <= '9' )
			sum = sum * 10 + (*s-'0');
		else 
			break;
		s++;
	}
	return sum * sign;
}

/*!
*****************************************************************************

 \fn unsigned int hex2int ( char* s );

 <b>Description:</b>\n
 Parse hexdecimal value into unsigned integer, accepting with or without 0x

 \param[in] s		The input string to parse

 \return :: unsigned integer value parsed

******************************************************************************/
unsigned int hex2int(char * s)
{
	unsigned int sum = 0;

	if ( s[0]!=0 && s[1]!=0 && s[0]=='0' && s[1]=='x' )
		s+=2;

	while(*s)
	{
		if ( '0' <= *s && *s <= '9' )
			sum = sum * 16 + (*s-'0');
		else if ( 'A' <= *s && *s <= 'F' )
			sum = sum * 16 + (*s-'A') + 10;
		else if ( 'a' <= *s && *s <= 'f' )
			sum = sum * 16 + (*s-'a') + 10;
		else 
			break;
		s++;
	}
	return sum;
}

/*!
*****************************************************************************

 \fn int makeargv(char * s, unsigned char argvsz, char * argv[]);

 <b>Description:</b>\n
 Tokenize buffer S into argurments and arguement count

 \param[in] s		array of cliCommands
 \param[in] argvsz	number Of commands
 \param[in] argv	version string

 \return :: number of argurment tokenized

******************************************************************************/
int makeargv(char * s, unsigned char argvsz, char * argv[])
{
	int argc = 0;

	/* split into argv */
	while (argc < argvsz - 1) 
	{
		/* skip any white space */
		while ((*s == ' ') || (*s == '\t'))
			++s;
		/* end of s, no more args */
		if (*s == '\0')
			break;
		/* begin of argument string */
		argv[argc] = s;
		/* find end of string */
		while (*s && (*s != ' ') && (*s != '\t'))
			++s;
		argc++;
		/* end of s, no more args */
		if (*s == '\0')
			break;
		/* terminate current arg */
		*s++ = '\0';
	}
	argv[argc] = 0;

	return argc;
}
