/*
 * (C) Copyright 2012 DSP Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <asm/io.h>
#include <asm/arch/clock.h>
#include <asm/arch/hardware.h>
#include <asm/arch/serial.h>

#define CMU_PLL1      0x00
#define CMU_PLL2      0x04
#define CMU_PLL3      0x08
#define CMU_CPUCLKCTL 0x14
#define CMU_COMCLKSEL 0x18
#define CMU_CLKSWCTL  0x1C
#define CMU_CLKDIV1   0x20
#define CMU_CLKDIV2   0x24
#define CMU_CLKDIV3   0x28
#define CMU_CLKDIV4   0x2C
#define CMU_CLKDIV5   0x30
#define CMU_OSCCR     0x3C
#define CMU_RSTSR     0x48
#define CMU_SWCHWRST1 0x50
#define CMU_SWCHWRST2 0x54
#define CMU_SWCLKENR1 0x58
#define CMU_SWCLKENR2 0x5C

struct clk {
	char *name;
	int clockbit;
	int resetbit;
	unsigned long rate;
	int (*set_rate)(struct clk *clk, unsigned long rate);
};

static int etherdiv_set_rate(struct clk *clk, unsigned long rate);
static int apu_set_rate(struct clk *clk, unsigned long rate);
static int lcdc_set_rate(struct clk *clk, unsigned long rate);
static int sdmmc_set_rate(struct clk *clk, unsigned long rate);
static int qspi_set_rate(struct clk *clk, unsigned long rate);
static int spi2_set_rate(struct clk *clk, unsigned long rate);
static int usb11_set_rate(struct clk *clk, unsigned long rate);
static int tdm1_set_rate(struct clk *clk, unsigned long rate);

#define __stringify_1(x)	#x
#define __stringify(x)		__stringify_1(x)

#define CLK(_name, _clockbit, _resetbit)    \
	struct clk _name = {                \
		.name = __stringify(_name), \
		.clockbit = _clockbit,      \
		.resetbit = _resetbit,      \
	}

#define CLK_RATE(_name, _clockbit, _resetbit, _set_rate) \
	struct clk _name = {                \
		.name = __stringify(_name), \
		.clockbit = _clockbit,      \
		.resetbit = _resetbit,      \
		.rate = 0,                  \
		.set_rate = _set_rate,      \
	}

CLK(usb1_phy_por,          -1, 32 + 25);
CLK(usb1_mac_phyif,        -1, 32 + 23);
CLK(ccu0,                  -1, 32 + 20);
CLK(coresight,        32 + 20, 32 + 30);
CLK(cmu_int_etherdiv, 32 + 17, -1);
CLK_RATE(cmu_etherdiv,     32 + 16, -1, etherdiv_set_rate);
CLK(css_gdmac,        32 + 15, 32 + 15);
CLK(css_timer2,       32 + 14, 32 + 13);
CLK(css_timer1,       32 + 13, 32 + 13);
CLK(dram_ctl,         32 + 12, 32 + 12);
CLK(clkout,           32 + 11,      -1);
CLK(mpmc,             32 + 10, 32 + 10);
CLK_RATE(usb11_mac,   32 +  9, 32 +  9, usb11_set_rate);
CLK(slvmii,           32 +  8, 32 +  8);
CLK_RATE(spi2,        32 +  7, 32 +  7, spi2_set_rate);
CLK(spi1,             32 +  6, 32 +  6);
CLK(i2c2,             32 +  5, 32 +  5);
CLK(i2c1,             32 +  4, 32 +  4);
CLK(timer4,           32 +  3, 32 +  1);
CLK(timer3,           32 +  2, 32 +  1);
CLK(timer2,           32 +  1, 32 +  0);
CLK(timer1,           32 +  0, 32 +  0);

CLK(css_etm,               31,      31);
CLK(css_bmp,               30,      30);
CLK(css_adpcm,             29,      29);
CLK(ref_clk,               28,      -1);
CLK(mcu_etm,               27,      -1);
CLK_RATE(apu,              26,      26, apu_set_rate);
CLK(rfapu,                 25, 32 +  3);
CLK(gpu,                   24,      24);
CLK(css,                   23,      23);
CLK(sec,                   22,      22);
CLK_RATE(lcdc,             21,      21, lcdc_set_rate);
CLK(pmu,                   20, 32 +  2);
CLK(ciu,                   19,      19);
CLK(dclass,                18,      18);
CLK(usb_phy,               17,      -1);
CLK(usb_mac,               16, 32 + 22);
CLK(acp,                   -1,      15);
CLK(nfc,                   14,      14);
CLK_RATE(sdmmc,            13,      13, sdmmc_set_rate);
CLK(isdn,                  12,      12);
CLK_RATE(qspi,             11,      11, qspi_set_rate);
CLK(tdm2,                  10,      10);
CLK_RATE(tdm1,                   9,       9, tdm1_set_rate);
CLK(drt_com,                8,       8);
CLK(uart2,                  7,       7);
CLK(uart1,                  6,       6);
CLK(rmii_if,                5,      -1);
CLK(secure,                 4,       4);
CLK(ethmac,                 3,       3);
CLK(rtc,                    2,       2);
CLK(css_arm926,             1,       1);
CLK(gdmac,                  0,       0);

static struct clk *clocks[] = {
	&usb1_phy_por,
	&usb1_mac_phyif,
	&ccu0,
	&coresight,
	&cmu_int_etherdiv,
	&cmu_etherdiv,
	&css_gdmac,
	&css_timer2,
	&css_timer1,
	&dram_ctl,
	&clkout,
	&mpmc,
	&usb11_mac,
	&slvmii,
	&spi2,
	&spi1,
	&i2c2,
	&i2c1,
	&timer4,
	&timer3,
	&timer2,
	&timer1,
	&css_etm,
	&css_bmp,
	&css_adpcm,
	&ref_clk,
	&mcu_etm,
	&apu,
	&rfapu,
	&gpu,
	&css,
	&sec,
	&lcdc,
	&pmu,
	&ciu,
	&dclass,
	&usb_phy,
	&usb_mac,
	&acp,
	&nfc,
	&sdmmc,
	&isdn,
	&qspi,
	&tdm2,
	&tdm1,
	&drt_com,
	&uart2,
	&uart1,
	&rmii_if,
	&secure,
	&ethmac,
	&rtc,
	&css_arm926,
	&gdmac,
};

static unsigned long
dvf99_calc_pllfreq(unsigned long input_freq, int pll_ctrl)
{
	unsigned long pll_out_freq;

	if (pll_ctrl & (1<<16)) {
		/* bypassed */
		pll_out_freq = input_freq;
	} else {
		uint32_t r, f, od;

		r = (pll_ctrl & 0x1f) + 1;
		f = ((pll_ctrl >> 5) & 0x7f) + 1;
		od = 1 << ((pll_ctrl >> 12) & 3);

		pll_out_freq = input_freq * f / r / od;
	}

	return pll_out_freq;
}

unsigned long
dvf99_get_main_clk(void)
{
	uint32_t reg;
	unsigned int sysclk, mainclk, dectclk = 13824000;

	reg = readl(DVF99_CMU_BASE + CMU_RSTSR);

	/* 25mhz/50mhz main clock selection */
	if (reg & 0x4)
		mainclk = 25000000;
	else
		mainclk = 50000000;

	/* main clock can be overwritten by dect clk via main_clk_sel_strap */
	sysclk = (reg & 0x2) ? dectclk : mainclk;

	return sysclk;
}

static unsigned long
dvf99_get_sysclk(void)
{
	uint32_t sysclksrc = (readl(DVF99_CMU_BASE + CMU_CLKSWCTL) >> 1) & 0x1;

	if (!sysclksrc) {
		uint32_t altsyssrc = (readl(DVF99_CMU_BASE + CMU_CPUCLKCTL) >> 14) & 0x3;
		switch (altsyssrc) {
			case 0: return dvf99_get_main_clk();
			case 1: return 13824000ul;
			case 2: return 12000000ul;
		}
	} else {
		uint32_t pll1 = readl(DVF99_CMU_BASE + CMU_PLL1);
		switch ((pll1 >> 19) & 0x3) {
			case 0: return dvf99_calc_pllfreq(dvf99_get_main_clk(), pll1);
			case 1: return dvf99_calc_pllfreq(13824000ul, pll1);
			case 2: return dvf99_calc_pllfreq(12000000ul, pll1);
		}
	}

	return 0;
}

static unsigned long
dvf99_get_pll2(void)
{
	uint32_t pll2 = readl(DVF99_CMU_BASE + CMU_PLL2);

	return dvf99_calc_pllfreq(dvf99_get_main_clk(), pll2);
}

unsigned long
dvf99_get_audclk(void)
{
	uint32_t swaudclksrc = (readl(DVF99_CMU_BASE + CMU_CLKSWCTL) >> 2) & 0x1;
	uint32_t audsrc = (readl(DVF99_CMU_BASE + CMU_CPUCLKCTL) >> 16) & 0x3;
	uint32_t pll3 = readl(DVF99_CMU_BASE + CMU_PLL3);
	uint32_t src_clk = 0;

	switch (audsrc) {
		case 0:
			src_clk = dvf99_get_main_clk();
			break;
		case 1:
			src_clk = 13824000ul;
			break;
		case 2:
			src_clk = 12000000ul;
			break;
	}

	if (!swaudclksrc)
		return src_clk;

	return dvf99_calc_pllfreq(src_clk, pll3);
}

unsigned long
dvf99_get_pll4(void)
{
	return 48000000ul;
}

static unsigned int
clk_divider(unsigned long parent_rate, unsigned long f, int bits)
{
	int div = ((parent_rate + f/2) / f) - 1;

	if (div < 0)
		div = 0;
	else if (div >= (1 << bits))
		div = (1 << bits) - 1;

	return div;
}

static int
etherdiv_set_rate(struct clk *clk, unsigned long rate)
{
	unsigned long reg;
	unsigned long ethrdivsrc;
	unsigned long parent_rate;
	unsigned int div;

	if (!rate)
		return -1;

#ifndef CONFIG_RMII_CLK_OFF_CHIP
    {
        unsigned long val = readl( DVF99_CMU_BASE + CMU_CLKSWCTL );

        val |= ( 1 << 25 );
        writel( val, DVF99_CMU_BASE + CMU_CLKSWCTL );
    }
#endif

	ethrdivsrc = readl(DVF99_CMU_BASE + CMU_CLKSWCTL) & (1 << 25);
	if (ethrdivsrc) /* ethrdivsrc: pll2 */
    {
		parent_rate = dvf99_get_pll2();
        printf("ethrdivsrc pll2 - rate %d\n", parent_rate);
    }
	else           /* ethrdivsrc: sysclk */
    {
		parent_rate = dvf99_get_sysclk();
        printf("ethrdivsrc sysclk - rate %d\n", parent_rate);
    }

	/* we must set ethrdiv rate to be EXACTLY rate */
	div = clk_divider(parent_rate,   rate, 4);
	if (parent_rate / (div+1) != rate)
		return -1;

	reg = readl(DVF99_CMU_BASE + CMU_CLKDIV2);
	reg &= ~(0xf << 20); /* ethrdiv */
	reg |= div << 20;
	writel(reg, DVF99_CMU_BASE + CMU_CLKDIV2);

	return 0;
}

static int
apu_set_rate(struct clk *clk, unsigned long rate)
{
	unsigned long reg;

	if (!rate)
		return -1;

	reg = readl(DVF99_CMU_BASE + CMU_CLKDIV5);
	reg &= ~(0xff << 24); /* apudiv */
	reg |= (clk_divider(dvf99_get_audclk(), rate, 8) << 24);
	writel(reg, DVF99_CMU_BASE + CMU_CLKDIV5);

	return 0;
}

static int
lcdc_set_rate(struct clk *clk, unsigned long rate)
{
	unsigned long reg, lcddivsrc;

	if (!rate)
		return -1;

	reg = readl(DVF99_CMU_BASE + CMU_CLKDIV1);
	reg &= ~(0xff << 8); /* lcdpixdiv */

	lcddivsrc = readl(DVF99_CMU_BASE + CMU_CLKSWCTL) & (1 << 17);
	if (lcddivsrc) /* lcddivsrc: pll2 */
		reg |= (clk_divider(dvf99_get_pll2(),   rate, 8) << 8);
	else           /* lcddivsrc: sysclk */
		reg |= (clk_divider(dvf99_get_sysclk(), rate, 8) << 8);

	writel(reg, DVF99_CMU_BASE + CMU_CLKDIV1);

	return 0;
}

static int
sdmmc_set_rate(struct clk *clk, unsigned long rate)
{
	unsigned long reg, sddivsrc;

	if (!rate)
		return -1;

	reg = readl(DVF99_CMU_BASE + CMU_CLKDIV1);
	reg &= ~(0xff << 24); /* sddiv */

	sddivsrc = readl(DVF99_CMU_BASE + CMU_CLKSWCTL) & (1 << 27);
	if (sddivsrc) /* sddivsrc: pll2 */
		reg |= (clk_divider(dvf99_get_pll2(),   rate, 8) << 24);
	else          /* sddivsrc: sysclk */
		reg |= (clk_divider(dvf99_get_sysclk(), rate, 8) << 24);

	writel(reg, DVF99_CMU_BASE + CMU_CLKDIV1);

	return 0;
}

static int
qspi_set_rate(struct clk *clk, unsigned long rate)
{
	unsigned long reg, qspidivsrc;

	if (!rate)
		return -1;

	reg = readl(DVF99_CMU_BASE + CMU_CLKDIV3);
	reg &= ~(0xff << 8); /* qspidiv */

	qspidivsrc = readl(DVF99_CMU_BASE + CMU_CLKSWCTL) & (1 << 26);
	if (qspidivsrc) /* qspidivsrc: pll2 */
		reg |= (clk_divider(dvf99_get_pll2(),   rate, 8) << 8);
	else            /* qspidivsrc: sysclk */
		reg |= (clk_divider(dvf99_get_sysclk(), rate, 8) << 8);

	writel(reg, DVF99_CMU_BASE + CMU_CLKDIV3);

	return 0;
}

static int
spi2_set_rate(struct clk *clk, unsigned long rate)
{
	unsigned long reg;

	if (!rate)
		return -1;

	reg = readl(DVF99_CMU_BASE + CMU_CLKDIV3);
	reg &= ~(0xff << 24); /* spi2div */
	reg |= (clk_divider(dvf99_get_sysclk(), rate, 8) << 24);
	writel(reg, DVF99_CMU_BASE + CMU_CLKDIV3);

	return 0;
}

static int
usb11_set_rate(struct clk *clk, unsigned long rate)
{
	unsigned long reg, usb11divsrc;

	if (!rate)
		return -1;

	reg = readl(DVF99_CMU_BASE + CMU_CLKDIV4);
	reg &= ~(0xf << 24); /* usb11div */

	usb11divsrc = (readl(DVF99_CMU_BASE + CMU_CLKDIV4) >> 28) & 0x3;
	switch (usb11divsrc) {
	case 0:
		reg |= (clk_divider(dvf99_get_sysclk(), rate, 4) << 24);
		break;
	case 1:
		reg |= (clk_divider(dvf99_get_pll2(), rate, 4) << 24);
		break;
	case 2:
		reg |= (clk_divider(dvf99_get_audclk(), rate, 4) << 24);
		break;
	case 3:
		reg |= (clk_divider(dvf99_get_pll4(), rate, 4) << 24);
		break;
	}
	writel(reg, DVF99_CMU_BASE + CMU_CLKDIV4);

	return 0;
}

static int
tdm1_set_rate(struct clk *clk, unsigned long rate)
{
	unsigned long reg, tdm1divsrc;

	if (!rate)
		return -1;

	reg = readl(DVF99_CMU_BASE + CMU_CLKDIV4);
	reg &= ~(0xfff); 

	tdm1divsrc = (readl(DVF99_CMU_BASE + CMU_CLKSWCTL) >> 19) & 0x3;
	switch (tdm1divsrc) {
	case 0:
		reg |= clk_divider(dvf99_get_sysclk(), rate, 12);
		break;
	case 1:
		reg |= clk_divider(dvf99_get_audclk(), rate, 12);
		break;
	case 2:
		reg |= clk_divider(dvf99_get_pll2(), rate, 12);
		break;
	}
	writel(reg, DVF99_CMU_BASE + CMU_CLKDIV4);

	return 0;
}

unsigned long
dvf99_get_hclk(void)
{
	uint32_t clkcntrl = (readl(DVF99_CMU_BASE + CMU_CPUCLKCTL) >> 4) & 0x3;

	/* divide by ahbclkdivrate */
	if (clkcntrl == 0)
		return dvf99_get_sysclk() / 3;
	return dvf99_get_sysclk() / (clkcntrl+1);
}

unsigned long
dvf99_get_cpufreq(void)
{
	uint32_t cpudiv = readl(DVF99_CMU_BASE + CMU_CPUCLKCTL) & 0xf;

	return dvf99_get_sysclk() / (cpudiv + 1);
}

unsigned long
dvf99_get_cssfreq(void)
{
	uint32_t comclksel = readl(DVF99_CMU_BASE + CMU_COMCLKSEL);
	uint32_t swcomsrc = (comclksel >> 10) & 0x3;
	uint32_t comprediv = (comclksel & 0xf) + 1;

	switch (swcomsrc) {
	case 0: return dvf99_get_sysclk() / comprediv;
	case 1: return dvf99_get_pll2() / comprediv;
	case 2: return dvf99_get_audclk() / comprediv;
	case 3: return dvf99_get_main_clk() / comprediv;
	}

	return 0;
}

unsigned long
dvf99_get_dramfreq(void)
{
	return dvf99_get_pll2();
}

void
clk_init(void)
{
	uint32_t reg, slow = 0;
	uint32_t main = dvf99_get_main_clk();

	/* The "slow" clock drives the RTC, which is used as a "high
	 * resolution" timer in the kernel. Therefore we set it to 1MHz. */
	switch (main) {
	case 13824000ul:
		slow = 13; /* 1.06MHz */
		break;
	case 25000000ul:
		slow = 25; /* 1MHz */
		break;
	case 50000000ul:
		slow = 50; /* 1MHz */
		break;
	}

	/* adjust SLOWDIV to ~1MHz */
	reg = readl(DVF99_CMU_BASE + CMU_CLKSWCTL);
	reg &= ~(0x7ff << 4);
	reg |=  ((slow - 1) << 4);
	writel(reg, DVF99_CMU_BASE + CMU_CLKSWCTL);
}

static struct clk *
clk_get(const char *id)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(clocks); i++) {
		if (!strcmp(id, clocks[i]->name))
			return clocks[i];
	}

	return NULL;
}

static void
clock_set_onoff(struct clk *clk, int on)
{
	int tmp, bit = clk->clockbit, reg = CMU_SWCLKENR2;

	if (bit == -1)
		return;
	else if (bit < 32)
		reg = CMU_SWCLKENR1;
	else
		bit -= 32;

	tmp = readl(DVF99_CMU_BASE + reg);
	if (on)
		tmp |=  (1 << bit);
	else
		tmp &= ~(1 << bit);
	writel(tmp, DVF99_CMU_BASE + reg);
}

int
clk_set_rate(char *id, unsigned long rate)
{
	struct clk *clock = clk_get(id);
	int ret = -1;

	if (clock && clock->set_rate)
		ret = clock->set_rate(clock, rate);

	if (!ret)
		clock->rate = rate;

	return ret;
}

void
clk_enable(char *id)
{
	struct clk *clock = clk_get(id);

	if (clock)
		clock_set_onoff(clock, 1);
}

void
clk_disable(char *id)
{
	struct clk *clock = clk_get(id);

	if (clock)
		clock_set_onoff(clock, 0);
}

static void
reset_set_onoff(struct clk *clk, int on)
{
	int tmp, bit = clk->resetbit, reg = CMU_SWCHWRST2;

	if (bit == -1)
		return;
	else if (bit < 32)
		reg = CMU_SWCHWRST1;
	else
		bit -= 32;

	tmp = readl(DVF99_CMU_BASE + reg);
	if (on)
		tmp |=  (1 << bit);
	else
		tmp &= ~(1 << bit);
	writel(tmp, DVF99_CMU_BASE + reg);
}

void
reset_set(char *id)
{
	struct clk *clock = clk_get(id);

	if (clock)
		reset_set_onoff(clock, 1);
}

void
reset_release(char *id)
{
	struct clk *clock = clk_get(id);

	if (clock)
		reset_set_onoff(clock, 0);
}


static void cmu_update_field(u8 reg, u8 shift, u8 size, u32 val)
{
	uint32_t regval;
	uint32_t mask = ((1 << size)-1) << shift;

	regval = readl(DVF99_CMU_BASE + reg);
	regval &= ~mask;
	regval |= (val << shift) & mask;
	writel(regval, DVF99_CMU_BASE + reg);
}

static int check_eth_config(enum eth_mode mode, enum eth_speed speed)
{
	switch (mode) {
		case ETH_MII:
		case ETH_SLVMII:
		case ETH_RMII:
		case ETH_SLVRMII:
			if (speed == ETH_SPEED_10 || speed == ETH_SPEED_100)
				return 0;
			break;
		case ETH_RGMII:
			if (speed == ETH_SPEED_1000)
				return 0;
			break;
		default:
			break;
	};

	return -1;
}

struct eth_mode_settings {
	uint8_t tx_clk_sel;
	uint8_t rx_clk_sel;
	uint8_t tx_pad_src_sel;
	uint8_t tx_pad_oen;
	uint8_t rx_pad_oen;
	uint8_t int_ethrdiv_en;
	int ethint_rate[NUM_ETH_SPEEDS];
};

/* Table 1 from CMU specification */
struct eth_mode_settings eth_clk_settings[NUM_ETH_MODES] = {
	/* MII */		{0, 0, 0, 0, 1, 0,	{2500000,	25000000,	-1} },
	/* SLAVE_MII */		{1, 0, 0, 0, 1, 0,	{2500000,	25000000,	-1} },
	/* RMII */		{1, 1, 0, 0, 1, 1,	{2500000,	25000000,	-1} },
	/* SLAVE RMII */	{1, 1, 0, 0, 1, 1,	{2500000,	25000000,	-1} },
	/* RGMII */		{1, 0, 3, 0, 1, 1,	{-1,		-1,		125000000} },
};

unsigned long ethrdiv_rates[] =
{
	/* MII */	25000000,
	/* SLVMII */	25000000,
	/* RMII */	50000000,
	/* SLVRMII */	50000000,
	/* RGMII */	125000000,
};

int clk_update_etherdiv(enum eth_mode mode, enum eth_speed speed)
{
    unsigned long rate = ethrdiv_rates[mode];
    struct eth_mode_settings *settings;

    settings = &eth_clk_settings[mode];

    if (settings->int_ethrdiv_en) {
        uint32_t div = rate / settings->ethint_rate[speed];

        if (div * settings->ethint_rate[speed] != rate) {
            printf("error: can't set etherintdiv clock rate to %d\n", settings->ethint_rate[speed]);
            return -1;
        }
        cmu_update_field(CMU_CLKDIV2,  0, 6, div - 1);
    }

    return 0;
}

int clk_eth_config(enum eth_mode mode, enum eth_speed speed)
{
	int ret;
	unsigned long rate = ethrdiv_rates[mode];
	struct eth_mode_settings *settings;

	if (check_eth_config(mode, speed))
		return -1;

	settings = &eth_clk_settings[mode];

#ifdef CONFIG_RMII_CLK_OFF_CHIP
	cmu_update_field(CMU_CLKDIV1, 23, 1, 1);
	clk_disable("cmu_etherdiv");
	settings->tx_pad_oen = 1;
#else
	ret = clk_set_rate("cmu_etherdiv", rate);
	if (ret) {
		cmu_update_field(CMU_CLKSWCTL, 25, 1, 0); /* ethrdivsrc: sysclk */
		ret = clk_set_rate("cmu_etherdiv", rate);
	}
	if (ret) {
		printf("error: can't set etherdiv clock rate to %lu\n", rate);
		return ret;
	}

	cmu_update_field(CMU_CLKDIV1, 23, 1, 0);
	clk_enable("cmu_etherdiv");
#endif

	cmu_update_field(CMU_CLKSWCTL, 23, 1, settings->tx_clk_sel);
	cmu_update_field(CMU_CLKSWCTL, 24, 1, settings->rx_clk_sel);
	cmu_update_field(CMU_CLKSWCTL, 30, 2, settings->tx_pad_src_sel);
	cmu_update_field(CMU_CLKSWCTL, 29, 1, settings->tx_pad_oen);
	cmu_update_field(CMU_CLKSWCTL, 28, 1, settings->rx_pad_oen);

	if (settings->int_ethrdiv_en) {
		uint32_t div = rate / settings->ethint_rate[speed];

		if (div * settings->ethint_rate[speed] != rate) {
			printf("error: can't set etherintdiv clock rate to %d\n", settings->ethint_rate[speed]);
			return -1;
		}

		clk_enable("cmu_int_etherdiv");

		cmu_update_field(CMU_CLKDIV2,  0, 6, div - 1);
	}

	return 0;
}

#define USBOTG1_GCR1 0x5c
#define USBOTG1_GCR1_SUSPEND_MODE	(1 << 19)
#define USBOTG1_GCR1_SUSPEND		(1 << 20)

static void syscfg_set_bits(u32 reg, u32 bits)
{
	u32 val = readl(DVF99_SYSCFG_BASE + reg);
	val |= bits;
	writel(val, DVF99_SYSCFG_BASE + reg);
}

void clk_enable_usb1(void)
{
	u32 reg;

	/* enable 12MHz OSC */
	reg = readl(DVF99_CMU_BASE + CMU_OSCCR);
	reg |= (1 << 2);
	writel(reg, DVF99_CMU_BASE + CMU_OSCCR);

	/* wait until stable */
	do {
		reg = readl(DVF99_CMU_BASE + CMU_OSCCR);
	} while (!(reg & 1 << 5));

	/* Special power-up sequence needed */
	clk_enable("usb_phy");
	udelay(10*2);
	syscfg_set_bits(USBOTG1_GCR1, USBOTG1_GCR1_SUSPEND);
	udelay(3100*2);
	reset_release("usb1_phy_por");
	clk_enable("usb_mac");
	udelay(10*2);
	clk_disable("usb_mac");
	reset_release("usb1_mac_phyif");
	clk_enable("usb_mac");
	udelay(10*2);
	clk_disable("usb_mac");
	reset_release("usb_mac");
	clk_enable("usb_mac");
	syscfg_set_bits(USBOTG1_GCR1, USBOTG1_GCR1_SUSPEND_MODE);
	udelay(10*2);
}

#define ABSDIFF(x, y)	((x) > (y) ? (x) - (y) : (y) - (x))

static long
round_rate_pll(unsigned long in_rate, unsigned long rate, unsigned long *rprediv, unsigned long *rmult, unsigned long *rpostdiv, unsigned long *rband)
{
	unsigned long prediv, mult, postdiv;
	unsigned long match_prediv = 0, match_mult = 0, match_postdiv = 0;
	unsigned long match_rate = 0, match_diff = ~0ul, match_vco = ~0ul;

	/* loop through all settings finding best match */
	for (prediv = 1; prediv <= 32; prediv++) {
		unsigned long phaseclk = (in_rate >> 10) / prediv;

		/* keep in phase detector reference frequency: 10..50MHz */
		if (phaseclk < 10000000>>10 || phaseclk > 50000000>>10)
			continue;

		for (mult = 6; mult <= 100; mult++) {
			unsigned long vcoclk = phaseclk * mult;

			/* keep in VCO frequency range: 300..1000MHz */
			if (vcoclk < 300000000>>10 || vcoclk > 1000000000>>10)
				continue;

			for (postdiv = 0; postdiv <= 3; postdiv++) {
				unsigned long diff, out = vcoclk >> postdiv;
				unsigned long tmp;

				/* quick check if in target range */
				if (out-10000 > rate>>10 || out+10000 < rate>>10)
					continue;

				/* ok, do the precise calculation */
				tmp = in_rate / 1000 * mult;
				tmp = tmp / (unsigned long)prediv;
				out = (unsigned long)tmp >> postdiv;
				out *= 1000;

				/* better match? */
				diff = ABSDIFF(rate, out);
				if ((diff <  match_diff) ||
				    (diff == match_diff && vcoclk < match_vco)) {
					match_diff    = diff;
					match_vco     = vcoclk;
					match_rate    = out;
					match_prediv  = prediv;
					match_mult    = mult;
					match_postdiv = postdiv;
				}
			}
		}
	}

	if (!match_rate)
		return -1;

	*rprediv = match_prediv - 1;
	*rmult = match_mult - 1;
	*rpostdiv = match_postdiv;
	*rband = match_vco > 600000000>>10 ? 1: 0;

	return match_rate;
}

int clk_reconfigure(int cpu_speed)
{
	unsigned long reg, mult = 0, prediv = 1, postdiv = 0, band = 0, sysclk;
	int i;

	if (dvf99_get_cpufreq() == cpu_speed)
		return 0;

	reg = readl(DVF99_CMU_BASE + CMU_PLL1) & (3 << 19);
	if ((reg == (2 << 19)) && ((cpu_speed % 48000000) != 0)) {
		printf("Must use multiple of 48MHz\n");
		return -1;
	}

	serial_exit();

	/* feed AP-CPU from altsys */
	reg = readl(DVF99_CMU_BASE + CMU_CLKSWCTL);
	reg &= ~(1 << 1);         /* swsysclksrc:   altsyssrc */
	writel(reg, DVF99_CMU_BASE + CMU_CLKSWCTL);

	/* feed CSS-CPU from pll1 (i.e., altsys) */
	reg = readl(DVF99_CMU_BASE + CMU_COMCLKSEL);
	reg &= ~(3 << 10);        /* swcomsrc:      clk_sysclk */
	writel(reg, DVF99_CMU_BASE + CMU_COMCLKSEL);

	sysclk = dvf99_get_sysclk();
	cpu_speed = round_rate_pll(sysclk, cpu_speed, &prediv, &mult, &postdiv, &band);

	/* make sure hclk is <= 200MHz */
	reg = readl(DVF99_CMU_BASE + CMU_CPUCLKCTL);
	reg &= ~(3 << 4);
	if (cpu_speed/2 < 200000000UL)
		reg |= 1 << 4;
	else if (cpu_speed/3 < 200000000UL)
		reg |= 2 << 4;
	else
		reg |= 3 << 4;
	writel(reg, DVF99_CMU_BASE + CMU_CPUCLKCTL);

	reg = readl(DVF99_CMU_BASE + CMU_PLL1);
	reg &= ~(0x3 << 19);
	/* re-configure PLL1 */
	reg |= ( prediv  <<  0) |  /* pre-div:       1 */
	      ( mult    <<  5) |  /* multiplicator: see above */
	      ( postdiv << 12) |  /* output-div:    1 */
	      ( band    << 14) |  /* band:          300 - 600 Mhz */
	      ( 1       << 15) |  /* power-down:    1 */
	      ( 0       << 16);   /* bypass:        0 */

	writel(reg, DVF99_CMU_BASE + CMU_PLL1);

	/* wait at least 500 ns */
	udelay(1);

	reg &= ~(1 << 15);        /* power-down:    0 */
	writel(reg, DVF99_CMU_BASE + CMU_PLL1);

	/* wait for lock */
	do {
		reg = readl(DVF99_CMU_BASE + CMU_PLL1);
	} while ((reg & (1 << 18)) == 0);

	/* feed AP-CPU from PLL1 */
	reg = readl(DVF99_CMU_BASE + CMU_CLKSWCTL);
	reg |=  (  1 <<  1);      /* swsysclksrc:   pll1 */
	writel(reg, DVF99_CMU_BASE + CMU_CLKSWCTL);

	/* feed CSS-CPU from PLL1 or PLL2 (if PLL1>500MHz) */
	reg = readl(DVF99_CMU_BASE + CMU_COMCLKSEL);
	reg &= ~(0xf <<  0);      /* comprediv:     0 */
	if (cpu_speed > 500000000) {
		reg &= ~(0x3 << 10);
		reg |=   0x2 << 10; /* swcomsrc:    clk_audclk */
	}
	writel(reg, DVF99_CMU_BASE + CMU_COMCLKSEL);

	serial_init();

	for (i = 0; i < ARRAY_SIZE(clocks); i++) {
		if (clocks[i]->set_rate && (clocks[i]->rate != 0))
			clocks[i]->set_rate(clocks[i], clocks[i]->rate);
	}

	return 0;
}
