/*
 *  DSPG UART support
 *
 *  Speed and options are hardcoded to 115200 8N1.
 *
 *  (C) Copyright 2006-2007, 2010, 2012, DSP Group
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <config.h>
#include <common.h>
#include <command.h>
#include <serial.h>
#include <environment.h>
#include <asm/io.h>
#include <asm/arch/gpio.h>
#include <asm/arch/hardware.h>
#include <asm/arch/clock.h>
#include <asm/arch/pads.h>
#include <asm/arch/keypad.h>
#include <asm/arch/serial.h>

#define UART_FIFO_SIZE		(128)

#define UART_CTL		0x00
#define UART_CFG		0x04
#define UART_INT_DIV		0x08
#define UART_FRAC_DIV		0x0C
#define UART_TX_FIFO_WM		0x10
#define UART_RX_FIFO_WM		0x14
#define UART_FIFO_INTSTAT	0x18
#define UART_FIFO_INTEN		0x1C
#define UART_FIFO_INT_CLR	0x20
#define UART_TX_FIFO_LVL	0x24
#define UART_RX_FIFO_LVL	0x28
#define UART_TX_DAT		0x2C
#define UART_RX_DAT		0x30
#define UART_STAT		0x34

#define RX_EMPTY		(1<<8)
#define TX_FULL			(1<<1)

#define BAUDRATE 115200

DECLARE_GLOBAL_DATA_PTR;

static void dvf99_serial_putc(int index, char c);
static int  dvf99_serial_getc(int index);
static int  dvf99_serial_tstc(int index);

#ifndef CONFIG_SERIAL_UNLOCKED
static int gxp_console = 1;
#endif

static int __attribute__((section (".data"))) uart = 0;
static int dvf99_uart_base[] = { DVF99_UART1_BASE, DVF99_UART2_BASE, DVF99_UART2_BASE };

static void dvf99_serial_putc(int index, char c)
{
	while (readl(dvf99_uart_base[index] + UART_STAT) & TX_FULL)
		;

	writel(c, dvf99_uart_base[index] + UART_TX_DAT);
}

static int dvf99_serial_tstc(int index)
{
	if (!(readl(dvf99_uart_base[index] + UART_STAT) & RX_EMPTY))
		return 1;
#ifdef CONFIG_USB_TTY
	extern void usbtty_poll(void);
	usbtty_poll();
#endif
	dvf99_keypad_debug();

	return 0;
}

static int dvf99_serial_getc(int index)
{
	unsigned int data;

	while (dvf99_serial_tstc(index) == 0)
		;

	data = readl(dvf99_uart_base[index] + UART_RX_DAT);
	if (data & 0x600) /* parity or frame error */
		return -1;

	data &= 0xff; /* limit to 8 characters */

	return (int)data;
}

#ifdef CONFIG_SERIAL_MULTI
static void serial_putc_dev(int index, char c)
{
	if (c == '\n')
		dvf99_serial_putc(index, '\r');
	dvf99_serial_putc(index, c);
}
#else
void serial_putc(char c)
{
	if (c == '\n')
		dvf99_serial_putc(uart, '\r');
	dvf99_serial_putc(uart, c);
}
#endif

#ifdef CONFIG_SERIAL_MULTI
static int serial_tstc_dev(int index)
{
#ifndef CONFIG_SERIAL_UNLOCKED
    if (gxp_console)
        return 0;
#endif
	return dvf99_serial_tstc(index);
}
#else
int serial_tstc(void)
{
#ifndef CONFIG_SERIAL_UNLOCKED
    if (gxp_console)
        return 0;
#endif
	return dvf99_serial_tstc(uart);
}
#endif

#ifdef CONFIG_SERIAL_MULTI
static int serial_getc_dev(int index)
{
#ifndef CONFIG_SERIAL_UNLOCKED
    if (gxp_console)
        return -1;
#endif
	return dvf99_serial_getc(index);
}
#else
int serial_getc(void)
{
#ifndef CONFIG_SERIAL_UNLOCKED
    if (gxp_console)
        return -1;
#endif
	return dvf99_serial_getc(uart);
}
#endif

static int serial_init_dev(int index)
{
	unsigned int intgr, frac, reg;
	unsigned int dvf99_curr_clk_rate;

	reg = readl(dvf99_uart_base[index] + UART_CTL);
	if (reg & 1) {
		/* wait until the TX FIFO is emptied */
		while (readl(dvf99_uart_base[uart] + UART_TX_FIFO_LVL) > 0)
			;

		/* the last character is still being shifted out */
		udelay(100);
	}

	dvf99_curr_clk_rate = dvf99_get_hclk();

	intgr = (dvf99_curr_clk_rate / (CONFIG_BAUDRATE << 4));
	frac = (dvf99_curr_clk_rate / CONFIG_BAUDRATE) & 0xf;

	writel(0, dvf99_uart_base[index] + UART_CTL); /* disable */
	/* data length: 8 bits; enable break detection logic */
	writel(1<<11, dvf99_uart_base[index] + UART_CFG);
	writel(intgr, dvf99_uart_base[index] + UART_INT_DIV);
	writel(frac, dvf99_uart_base[index] + UART_FRAC_DIV);
	writel((1<<12)|(1<<9)|(1<<8)|1, dvf99_uart_base[index] + UART_FIFO_INT_CLR);
	writel(UART_FIFO_SIZE/2, dvf99_uart_base[index] + UART_TX_FIFO_WM);
	writel(UART_FIFO_SIZE/2, dvf99_uart_base[index] + UART_RX_FIFO_WM);
	writel(1, dvf99_uart_base[index] + UART_CTL); /* enable */

	return 0;
}

static int serial_uninit_dev(int index)
{
	unsigned int reg;

	reg = readl(dvf99_uart_base[index] + UART_CTL);
	if (reg & 1) {
		/* wait until the TX FIFO is emptied */
		while (readl(dvf99_uart_base[uart] + UART_TX_FIFO_LVL) > 0)
			;

		/* the last character is still being shifted out */
		udelay(100);

		writel(0, dvf99_uart_base[index] + UART_CTL); /* disable */
	}

	return 0;
}

int get_uart_config(void)
{
#if CONFIG_USE_UART == 0
	int ucfg = (readl(DVF99_CMU_BASE + DVF99_CMU_RSTSR) >> 8) & 0xff;

	if ((ucfg & 0x11) == 0x11)
		return 0;
	else if ((ucfg & 0x11) == 0x01)
		return 1;
	else if ((ucfg & 0x11) == 0x10)
		return 2;
	return 0;
#else
	return CONFIG_USE_UART - 1;
#endif
}

void serial_exit(void)
{
	uart = get_uart_config();

	(void)serial_uninit_dev(uart);
}

#ifndef CONFIG_SERIAL_MULTI
int serial_init(void)
{
	uart = get_uart_config();

	return serial_init_dev(uart);
}
#endif

#ifdef CONFIG_SERIAL_MULTI
void serial_puts_dev(int index, const char *s)
{
#ifndef CONFIG_SERIAL_UNLOCKED
    if (gxp_console)
        return;
#endif
	while (*s)
		serial_putc_dev(index, *s++);
}
#else
void serial_puts(const char *s)
{
#ifndef CONFIG_SERIAL_UNLOCKED
    if (gxp_console)
        return;
#endif
	while (*s)
		serial_putc(*s++);
}
#endif

#ifdef CONFIG_SERIAL_MULTI
static void serial_setbrg_dev(void) { }
#else
void serial_setbrg(void) { }
#endif

void serial_unlock(void)
{
#ifndef CONFIG_SERIAL_UNLOCKED
    gxp_console = 0;
#endif
}

#ifdef CONFIG_SERIAL_MULTI

/* Multi serial device functions */
#define DECLARE_DVF_SERIAL_FUNCTIONS(port) \
	void dvf_uart##port##_init(void) \
	{ \
		serial_init_dev(port); \
	} \
	void dvf_uart##port##_uninit(void) \
	{ \
		serial_uninit_dev(port); \
	} \
	int dvf_uart##port##_getc(void) \
	{ \
		return serial_getc_dev(port); \
	} \
	int dvf_uart##port##_tstc(void) \
	{ \
		return serial_tstc_dev(port); \
	} \
	void dvf_uart##port##_putc(const char c) \
	{ \
		serial_putc_dev(port, c); \
	} \
	void dvf_uart##port##_puts(const char *s) \
	{ \
		serial_puts_dev(port, s); \
	}

#define INIT_DVF_SERIAL_STRUCTURE(port, name) { \
	name, \
	dvf_uart##port##_init, \
	dvf_uart##port##_uninit, \
	serial_setbrg_dev, \
	dvf_uart##port##_getc, \
	dvf_uart##port##_tstc, \
	dvf_uart##port##_putc, \
	dvf_uart##port##_puts, \
}

DECLARE_DVF_SERIAL_FUNCTIONS(0);
struct serial_device dvf_serial0_device =
INIT_DVF_SERIAL_STRUCTURE(0, "dvf_uart0");
DECLARE_DVF_SERIAL_FUNCTIONS(1);
struct serial_device dvf_serial1_device =
INIT_DVF_SERIAL_STRUCTURE(1, "dvf_uart1");

struct serial_device *default_serial_console(void)
{
	if (uart == 0)
		return &dvf_serial0_device;
	else
		return &dvf_serial1_device;
}

#endif
