/*
 * (C) Copyright 2012 DSP Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <command.h>
#include <asm/io.h>
#include <asm/arch/clock.h>
#include <asm/arch/hardware.h>

#define TMR_LOAD		0x00
#define TMR_VALUE		0x04
#define TMR_CONTROL		0x08

/* Control */
#define TMR_CTRL_ONESHOT	(1 << 0)
#define TMR_CTRL_32BIT		(1 << 1)
#define TMR_CTRL_DIV1		(0 << 2)
#define TMR_CTRL_DIV16		(1 << 2)
#define TMR_CTRL_DIV256		(2 << 2)
#define TMR_CTRL_IE		(1 << 5)
#define TMR_CTRL_PERIODIC	(1 << 6)
#define TMR_CTRL_ENABLE		(1 << 7)

DECLARE_GLOBAL_DATA_PTR;

static void *base = (void *)DVF99_TIMER0_1_BASE;

/*
 * We're using the gd pointer to store the timer state because the timer is
 * initialized before relocation and all global variables are re-initialized by
 * the relocation. The following gd variables are used by the timer:
 *
 * timer_rate_hz: timer rate in hz
 * lastinc:       last timer counter value (note: timers are decrementing!)
 * tbu:           ticks per jiffie unit
 * tbl:           last jiffie
 */

void reset_timer_masked(void)
{
	/* capture current decrementer value time */
	gd->lastinc = readl(base + TMR_VALUE);
	gd->tbl = 0; /* start "advancing" last time stamp from 0 */
}

void reset_timer(void)
{
	reset_timer_masked();
}

unsigned long get_timer_masked(void)
{
	unsigned long elapsed, now;

	now = readl(base + TMR_VALUE);
	elapsed = (gd->lastinc - now) / gd->tbu;

	gd->tbl += elapsed;
	gd->lastinc -= elapsed * gd->tbu;

	return gd->tbl;
}

unsigned long get_timer(unsigned long base)
{
	return get_timer_masked() - base;
}

/* delay x useconds AND preserve advance timestamp value */
void __udelay(unsigned long usec)
{
	/* Fine grained or coarse/long delay? */
	if (usec >= 10000) {
		/* do it with CONFIG_SYS_HZ resolution */
		unsigned long tmo, endtime;
		signed long diff;

		tmo = usec / 1000;	/* start to normalize for usec to ticks per sec */
		tmo *= CONFIG_SYS_HZ;	/* find number of "ticks" to wait to achieve target */
		tmo /= 1000;		/* finish normalize. */

		endtime = get_timer_masked() + tmo;
		do {
			unsigned long now = get_timer_masked();
			diff = endtime - now;
		} while (diff >= 0);
	} else {
		/* do it with hardware resolution */
		long tmo = usec * (gd->timer_rate_hz / 1000) / 1000;
		unsigned long now, last;

		last = readl(base + TMR_VALUE);
		while (tmo > 0) {
			now = readl(base + TMR_VALUE);
			tmo -= last - now;
			last = now;
		}
	}
}

unsigned long long get_ticks(void)
{
	return get_timer(0);
}

unsigned long get_tbclk(void)
{
	return CONFIG_SYS_HZ;
}

int timer_init(void)
{
	reset_release("timer1");
	clk_enable("timer1");

	/* disable timer just in case */
	writel(TMR_CTRL_DIV16 | TMR_CTRL_32BIT, base + TMR_CONTROL);

	writel(0xffffffff, base + TMR_LOAD);
	writel(readl(base + TMR_CONTROL) | TMR_CTRL_ENABLE, base + TMR_CONTROL);

	gd->timer_rate_hz = dvf99_get_hclk() / 16;
	gd->tbu = gd->timer_rate_hz / CONFIG_SYS_HZ;

	/* init the timestamp and lastdec value */
	reset_timer_masked();

	return 0;
}

int do_dvf99_test_timers(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	int count = 10;

	printf("Counting with mdelay: %2d ", count);
	while (count > 0) {
		mdelay(1000);
		printf("\b\b\b%2d ", --count);
	}
	printf("\n");

	count = 10;
	printf("Counting with udelay: %2d ", count);
	while (count > 0) {
		udelay(1000000);
		printf("\b\b\b%2d ", --count);
	}
	printf("\n");

	return 0;
}

U_BOOT_CMD(test_timers, 1, 0, do_dvf99_test_timers,
	"run timers test (countdown for 10 seconds)",
	""
);
