/*
 * (C) Copyright 2014, DSP Group
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <config.h>

#if defined(CONFIG_SOFTWARE_UPGRADE_UBOOT) || defined(CONFIG_SOFTWARE_UPGRADE_ROOTFS)
#include <jffs2/load_kernel.h>
#include <linux/list.h>
#include <asm/io.h>
#include <asm/arch/hardware.h>
#include <asm/arch/clock.h>
#include <asm/arch/upgrade.h>
#include <asm/arch/serial.h>

#define UPDATE_STATUS_INVALID		0
#define UPDATE_STATUS_AVAILABLE		1

#ifdef CONFIG_CMD_NAND
#include <nand.h>
extern nand_info_t nand_info[CONFIG_SYS_MAX_NAND_DEVICE];
#endif /* CONFIG_CMD_NAND */

typedef void (*new_uboot)(void);
new_uboot *new_uboot_image;

extern struct list_head devices;

static struct part_info *
get_partition(const char *id)
{
	struct mtd_device *mtd_dev;
	struct part_info *part;
	struct list_head *dentry;
	struct list_head *pentry;

	list_for_each(dentry, &devices) {
		mtd_dev = list_entry(dentry, struct mtd_device, link);
		list_for_each(pentry, &mtd_dev->parts) {
			part = list_entry(pentry, struct part_info, link);
			if (strcmp(part->name, id) == 0)
				return part;
		}
	}

	return NULL;
}

#ifdef CONFIG_SOFTWARE_UPGRADE_ROOTFS
static void
fixup_bootargs(const char *rootfs)
{
	char buf[256], *start, *end;
	char *cmdline = getenv("linux_bootargs");
	int offset, paramlen;

	if (!cmdline || !rootfs)
		return;

	paramlen = strlen(rootfs);
	if (rootfs[paramlen - 1] == '\n')
		paramlen--;

	if (((start = strstr(cmdline, "ubi.mtd=")) != NULL) ||
	    ((start = strstr(cmdline, "root=mtd:")) != NULL)) {
		end = strchr(start, ' ');
		offset = start - cmdline; // + 5;
		strncpy(buf, cmdline, offset);
		strncpy(buf + offset, rootfs, paramlen);
		offset += paramlen;
		if (end)
			strcpy(buf + offset, end);
		else
			buf[offset] = '\0';
		setenv("linux_bootargs", buf);
	}
}
#endif /* CONFIG_SOFTWARE_UPGRADE_ROOTFS */

/* FIXME: add watchdog here */

static void
load_and_start_updated_uboot(const char *partition_name)
{
	unsigned long offset = 0;
	struct part_info *pinfo = NULL;
#ifdef CONFIG_CMD_NAND
	size_t cnt = 0;
#endif

	pinfo = get_partition(partition_name);
	if (pinfo == NULL) {
		printf("unknown partition: %s\n", partition_name);
		printf("not updating U-Boot\n");
		return;
	}

#ifdef CONFIG_CMD_NAND
	cnt = pinfo->size;
	offset = pinfo->offset;
	nand_read_skip_bad(&nand_info[0], offset,
	                   &cnt, (u_char *)CONFIG_SYS_TEXT_BASE);
	new_uboot_image = (new_uboot *)CONFIG_SYS_TEXT_BASE;
#else
	/* TODO: QSPI */
	offset = CONFIG_SYS_FLASH_BASE + pinfo->offset;
	new_uboot_image = (new_uboot *)offset;
#endif /* CONFIG_CMD_NAND */
	printf("Loading updated u-boot image @0x%x\n",
	       (unsigned int)offset);
	serial_exit();
	((new_uboot)(new_uboot_image))();
}

void
dvf99_software_upgrade(void)
{
	char *update_status_string, *partition_string;
	int update_status_uboot;
#ifdef CONFIG_SOFTWARE_UPGRADE_ROOTFS
	int update_status_rootfs;
#endif /* CONFIG_SOFTWARE_UPGRADE_ROOTFS */
	char tmp[100];

	if (mtdparts_init())
		printf("mtdparts failed\n");

#ifdef CONFIG_SOFTWARE_UPGRADE_UBOOT
#ifdef CONFIG_SOFTWARE_UPGRADE_VERSION
	if (((update_status_string = getenv("uboot_version")) == NULL) ||
	   (strcmp(update_status_string, CONFIG_SOFTWARE_UPGRADE_VERSION))) {
		setenv("uboot_version", CONFIG_SOFTWARE_UPGRADE_VERSION);
		saveenv();
	}
#endif /* CONFIG_SOFTWARE_UPGRADE_VERSION */

	update_status_uboot = UPDATE_STATUS_INVALID;
	if ((update_status_string = getenv("uboot_update")) != NULL)
		update_status_uboot = simple_strtol(update_status_string, NULL, 10);

	if (update_status_uboot == UPDATE_STATUS_AVAILABLE) {
		sprintf(tmp, "%d", UPDATE_STATUS_INVALID);
		setenv("uboot_update", tmp);
		saveenv();

		if ((partition_string = getenv("uboot_partition_new")) != NULL)
			load_and_start_updated_uboot((const char *)partition_string);
	} else {
		if ((partition_string = getenv("uboot_partition_safe")) != NULL)
			load_and_start_updated_uboot((const char *)partition_string);
	}
#endif /* CONFIG_SOFTWARE_UPGRADE_UBOOT */

#ifdef CONFIG_SOFTWARE_UPGRADE_ROOTFS
	update_status_rootfs = UPDATE_STATUS_INVALID;
	if ((update_status_string = getenv("rootfs_update")) != NULL)
		update_status_rootfs = simple_strtol(update_status_string, NULL, 10);

	if (update_status_rootfs == UPDATE_STATUS_AVAILABLE) {
		sprintf(tmp, "%d", UPDATE_STATUS_INVALID);
		setenv("rootfs_update", tmp);
		saveenv();
	}

	if (update_status_rootfs == UPDATE_STATUS_AVAILABLE) {
		if ((partition_string = getenv("rootfs_partition_new"))
		    != NULL) {
			printf("Update available for rootfs\n");
			fixup_bootargs(partition_string);
		}
	}
#endif /* CONFIG_SOFTWARE_UPGRADE_ROOTFS */
}
#endif
