/*
 * Copyright (C) 2009 Texas Instruments Incorporated - http://www.ti.com/
 *
 * Driver for SPI controller on DVF99. Based on atmel_spi.c
 * by Atmel Corporation
 *
 * Copyright (C) 2007 Atmel Corporation
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */
#include <common.h>
#include <spi.h>
#include <malloc.h>
#include <asm/io.h>
#include <asm/hardware.h>

struct dspg_spi_slave {
	struct spi_slave slave;
	unsigned long regs_base;
	unsigned int freq;
	unsigned int mode;
};

static inline struct dspg_spi_slave *to_dspg_spi(struct spi_slave *slave)
{
	return container_of(slave, struct dspg_spi_slave, slave);
}

/* DMW96 SPI controller (supporting only master mode) */
#define DMW_SPI_CFG_REG		0x00
#define DMW_SPI_RATE_CNTL	0x04
#define DMW_SPI_INT_EN		0x08
#define DMW_SPI_INT_ST		0x0C
#define DMW_SPI_INT_CLR		0x10
#define DMW_SPI_CAUSE_REG	0x14
#define DMW_SPI_TX_DAT		0x18
#define DMW_SPI_RX_DAT		0x1C
#define DMW_SPI_DEL_VAL		0x20

#define BIT(x)			(1 << (x))

/* SPI_CFG_REG */
#define CFG_CS0_DIS		BIT(1)
#define CFG_CS1_DIS		BIT(3)
#define CFG_DATA_ORD		BIT(5)
#define CFG_SWCS		BIT(6)
#define CFG_DIRECT_CS		BIT(7)
#define CFG_SPI_EN		BIT(31)

/* DCX SPI controller (supporting master and slave mode) */
#define DCX_SPI_GLOBAL		0x00
#define DCX_SPI_CON		0x04
#define DCX_SPI_FRM		0x08
#define DCX_SPI_IER		0x0c
#define DCX_SPI_STAT		0x10
#define DCX_SPI_DAT		0x14

/* SPI_STAT register */
#define STAT_BUSY		(1<<6)
#define STAT_RBE		(1<<3)
#define STAT_TBF		(1<<2)
#define STAT_TBE		(1<<0)

void spi_init()
{
	/* Nothing to do */
}

struct spi_slave *
spi_setup_slave(unsigned int bus, unsigned int cs,
		unsigned int max_hz, unsigned int mode)
{
	struct dspg_spi_slave *ds;

	if (!spi_cs_is_valid(bus, cs))
		return NULL;

	ds = malloc(sizeof(*ds));
	if (!ds)
		return NULL;

	ds->slave.bus = bus;
	ds->slave.cs = cs;

	if (bus == 0)
		ds->regs_base = DVF99_SPI1_BASE;
	else
		ds->regs_base = DVF99_SPI2_BASE;

	ds->freq = max_hz;
	ds->mode = mode;

	return &ds->slave;
}

void spi_free_slave(struct spi_slave *slave)
{
	struct dspg_spi_slave *ds = to_dspg_spi(slave);

	free(ds);
}

int spi_claim_bus(struct spi_slave *slave)
{
	struct dspg_spi_slave *ds = to_dspg_spi(slave);
	unsigned int cfg;

	if (slave->bus == 0) {
		/* Set rate - currently fixed... */
		writel(0x08, ds->regs_base + DMW_SPI_RATE_CNTL); /* FIXME */

		/* no interrupts */
		writel(0x00, ds->regs_base + DMW_SPI_INT_EN);
		writel(0x26, ds->regs_base + DMW_SPI_INT_CLR);

		/* SWCS = 0, so selected CS is asserted */
		cfg = CFG_DIRECT_CS | CFG_DATA_ORD | CFG_SPI_EN;

		if (slave->cs == 0)
			cfg |= CFG_CS1_DIS;
		else
			cfg |= CFG_CS0_DIS;

		/* data_ord: 0 (bit 5), clkpol: 0 (bit 0) */

		/* Chipselect controlled by software, data order: MSB first */
		writel(cfg, ds->regs_base + DMW_SPI_CFG_REG);
	} else {
		/* Reset */
		writew(0x0006, ds->regs_base + DCX_SPI_GLOBAL);
		udelay(10);

		/* Set rate==1, mode==0 or 3, master, 8-bit */
		if (ds->mode == 0)
			writew(0x0e81, ds->regs_base + DCX_SPI_CON);
		else
			writew(0x0f81, ds->regs_base + DCX_SPI_CON);

		/* no interrupts */
		writew(0x0000, ds->regs_base + DCX_SPI_IER);
	}

	return 0;
}

void spi_release_bus(struct spi_slave *slave)
{
	struct dspg_spi_slave *ds = to_dspg_spi(slave);
	unsigned int cfg;

	if (slave->bus == 0) {
		/* Disable CS (might be not needed since disabling of the SPI will do the same anyway... */
		cfg = readl(ds->regs_base + DMW_SPI_CFG_REG);
		cfg |= CFG_SWCS;
		writel(cfg, ds->regs_base + DMW_SPI_CFG_REG);

		/* Disable the SPI hardware */
		writel(0, ds->regs_base + DMW_SPI_CFG_REG);
	}
}

static inline unsigned int
dmw_spi_write_read_byte(struct dspg_spi_slave *ds, unsigned int data)
{
	/* send out data */
	writel(data, ds->regs_base + DMW_SPI_TX_DAT);

	/* wait for the data to clock in/out */
	while (!(readl(ds->regs_base + DMW_SPI_INT_ST) & (1<<10))) /* SPI_DONE */
		;

	/* read the byte we just got and return it */
	/* because we always read the byte, we don't have to use RX_IGNORE or RX_FIFO_FLUSH */
	return readl(ds->regs_base + DMW_SPI_RX_DAT);
}

static inline unsigned char
dcx_spi_read_byte(struct dspg_spi_slave *ds)
{
	while (readw(ds->regs_base + DCX_SPI_STAT) & STAT_RBE)
		;

	return readb(ds->regs_base + DCX_SPI_DAT);
}

static inline void
dcx_spi_write_byte(struct dspg_spi_slave *ds, unsigned char data)
{
	while (readw(ds->regs_base + DCX_SPI_STAT) & STAT_TBF)
		;

	writeb(data, ds->regs_base + DCX_SPI_DAT);
}

int spi_xfer(struct spi_slave *slave, unsigned int bitlen,
	     const void *dout, void *din, unsigned long flags)
{
	struct dspg_spi_slave *ds = to_dspg_spi(slave);
	unsigned int len, i, tmp;
	unsigned char *buf;

	/* No support for full-duplex, and only support full bytes */
	if ((dout && din) || (bitlen % 8))
		return -1;

	if (slave->bus == 1) {
		if (flags & SPI_XFER_BEGIN)
			spi_cs_activate(slave);

		/* FIFO pointer reset */
		writew(0x0004, ds->regs_base + DCX_SPI_GLOBAL);

		tmp = readw(ds->regs_base + DCX_SPI_CON);
		tmp &= ~0x2000; /* disable SHIFT_OFF */
		writew(tmp, ds->regs_base + DCX_SPI_CON);
	}

	len = bitlen / 8;

	if (dout) { /* Write */
		buf = (unsigned char*)dout;

		if (slave->bus == 0) {
			for (i = 0; i < len; i++)
				(void)dmw_spi_write_read_byte(ds, buf[i]);
		} else {
			for (i = 0; i < len; i++)
				dcx_spi_write_byte(ds, buf[i]);
		}
	} else if (din) { /* Read */
		buf = (unsigned char*)din;

		if (slave->bus == 0) {
			for (i = 0; i < len; i++)
				buf[i] = dmw_spi_write_read_byte(ds, 0x00);
		} else {
			/* dummy read required for receiving */
			if (din)
				(void)dcx_spi_read_byte(ds);

			for (i = 0; i < len-1; i++)
				buf[i] = dcx_spi_read_byte(ds);

			tmp = readw(ds->regs_base + DCX_SPI_CON);
			tmp |= 0x2000; /* enable SHIFT_OFF */
			writew(tmp, ds->regs_base + DCX_SPI_CON);

			buf[len-1] = dcx_spi_read_byte(ds);
		}
	}

	if (slave->bus == 1) {
		if (flags & SPI_XFER_END) {
			//while (readw(DVF99_SPI2_BASE + DCX_SPI_STAT) & STAT_BUSY)
				;

			/* Wait until the transfer is completely done */
			spi_cs_deactivate(slave);
		}
	}

	return 0;
}

int spi_cs_is_valid(unsigned int bus, unsigned int cs)
{
	return ((bus == 0) || (bus == 1)) && ((cs == 0) || (cs == 1));
}
