/*
 *  linux/arch/arm/mach-dvf/arch.c
 *
 *  Copyright (C) 2010-2012 DSPG Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <linux/delay.h>
#include <linux/export.h>
#include <linux/irq.h>
#include <linux/irqdomain.h>
#include <linux/of.h>
#include <linux/of_address.h>
#include <linux/of_irq.h>
#include <linux/of_platform.h>
#include <linux/memblock.h>
#include <linux/regulator/machine.h>
#include <linux/dma-mapping.h>

#include <asm/mach/arch.h>
#include <asm/hardware/plicu.h>
#include <asm/mach/map.h>
#include <asm/memblock.h>
#include <asm/system.h>

#include <mach/hardware.h>
#include <generated/mach-types.h>

#include <mach/clock.h>

static bool dvf_nodebug = false;
static void __iomem *denali_sync;

static __init int gxp_console_setup(char* str)
{
    get_option(&str, &gxp_console);
    return 0;
}
early_param("gxp_console", gxp_console_setup);


int dvf_get_usb11_gcr(void)
{
	return readl(IO_ADDRESS(DVF_SYSCFG_BASE) + DVF_SYSCFG_USB11_GCR1);
}
EXPORT_SYMBOL(dvf_get_usb11_gcr);

void dvf_set_usb11_gcr(int value)
{
	writel(value, IO_ADDRESS(DVF_SYSCFG_BASE) + DVF_SYSCFG_USB11_GCR1);
}
EXPORT_SYMBOL(dvf_set_usb11_gcr);

int dvf_get_chip_rev(void)
{
	return readl(IO_ADDRESS(DVF_SYSCFG_BASE) + DVF_SYSCFG_CHIP_REV);
}
EXPORT_SYMBOL(dvf_get_chip_rev);

void dvf_denali_sync(void)
{
	/* Perform read-after-write to make sure CPU port on Denali controller
	 * is flushed. */
	if (denali_sync) {
		writel_relaxed(0, denali_sync);
		barrier();
		readl_relaxed(denali_sync);
	}
}
EXPORT_SYMBOL(dvf_denali_sync);

static int __init dvf_denali_sync_init(void)
{
	dma_addr_t paddr;

	denali_sync = dma_alloc_coherent(NULL, 4, &paddr, GFP_KERNEL);
	if (!denali_sync)
		panic("dvf_denali_sync_init failed!");

	return 0;
}
arch_initcall(dvf_denali_sync_init);

bool dvf_debug_disabled(void)
{
	return dvf_nodebug;
}

static int __init nodebug_setup(char *__unused)
{
	dvf_nodebug = true;

	return 1;
}

__setup("nodebug", nodebug_setup);

static struct map_desc dvf_io_desc[] __initdata = {
	{
		.virtual	= IO_ADDRESS(DVF_CMU_BASE),
		.pfn		= __phys_to_pfn(DVF_CMU_BASE),
		.length		= SZ_4K,
		.type		= MT_DEVICE
	}, {
		.virtual	= IO_ADDRESS(DVF_SYSCFG_BASE),
		.pfn		= __phys_to_pfn(DVF_SYSCFG_BASE),
		.length		= SZ_4K,
		.type		= MT_DEVICE
	},
};

static const char *dvf_dt_match[] = {
	"dspg,dvf99",
	NULL
};

static const struct of_device_id plicu_of_match[] __initconst = {
	{ .compatible = "dspg,plicu", .data = plicu_init, },
	{ /* Sentinel */ }
};

static void __init dvf_map_io(void)
{
	init_consistent_dma_size(14 << 20);
	iotable_init(dvf_io_desc, ARRAY_SIZE(dvf_io_desc));
}

static void __init dvf_early_init(void)
{
	if (dvf_clk_init())
		panic("Clock init failed!");
}

static void __init dvf_init_irq(void)
{
	of_irq_init(plicu_of_match);
}

static const struct of_dev_auxdata dvf99_auxdata_lookup[] __initconst = {
	OF_DEV_AUXDATA("dspg,tdm", 0x5f00000, "dvf-tdm.0", NULL),
	OF_DEV_AUXDATA("dspg,tdm", 0x6000000, "dvf-tdm.1", NULL),
	{},
};

static void __init dvf_init_machine(void)
{
	regulator_has_full_constraints();
	of_platform_populate(NULL, of_default_bus_match_table, dvf99_auxdata_lookup, NULL);
}

static void dvf_reset(char mode, const char *cmd)
{
	unsigned int value;

	/* Clear WDRSTIND and SWRSTIND before SW resetting */
	value = readl(IO_ADDRESS(DVF_CMU_BASE) + DVF_CMU_RSTSR);
	value |= (1<<5) | (1<<4) | (1<<3);
	writel(value, IO_ADDRESS(DVF_CMU_BASE) + DVF_CMU_RSTSR);

    while(1) {
        writel(0x7496398B, IO_ADDRESS(DVF_CMU_BASE) + DVF_CMU_SWSYSRST);
        writel(0xB5C9C674, IO_ADDRESS(DVF_CMU_BASE) + DVF_CMU_SWSYSRST);
    }
}

extern struct sys_timer dvf_timer;

static phys_addr_t dvf_reservemem_base;
static phys_addr_t dvf_reservemem_next;
static size_t dvf_reservemem_size;

static size_t __init dvf_reservemem_remain(void)
{
	size_t used = dvf_reservemem_next - dvf_reservemem_base;
	return dvf_reservemem_size - used;
}

static int
dvf_reservemem_get(phys_addr_t *pa, size_t size)
{
	size = roundup(size, SZ_1M);

	if (dvf_reservemem_remain() < size)
		return -ENOMEM;

	*pa = dvf_reservemem_next;
	dvf_reservemem_next += size;
	return 0;
}

dma_addr_t
dvf_dev_reserve_dmamem(struct device *dev, size_t size, int exclusive)
{
	int flags = DMA_MEMORY_MAP;
	phys_addr_t pa;
	int ret;

	if (exclusive)
		flags |= DMA_MEMORY_EXCLUSIVE;

	ret = dvf_reservemem_get(&pa, size);
	if (ret)
		return -ENOMEM;

	ret = dma_declare_coherent_memory(dev, pa, pa, size, flags);
	if(!(ret & DMA_MEMORY_MAP))
		return -ENOMEM;

	return pa;
}

static void __init dvf_reserve(void)
{
#ifndef CONFIG_DVF_RESERVEMEM_CSS
    dvf_reservemem_size = 0;
#else
	dvf_reservemem_size = CONFIG_DVF_RESERVEMEM_CSS * SZ_1M;

	dvf_reservemem_base = dvf_reservemem_next =
		memblock_alloc(dvf_reservemem_size, SZ_1M);
	memblock_free(dvf_reservemem_base, dvf_reservemem_size);
	memblock_remove(dvf_reservemem_base, dvf_reservemem_size);
#endif
}

MACHINE_START(DVF99, "DSPG DVF99")
	.map_io		= dvf_map_io,
	.nr_irqs	= NR_IRQS_LEGACY,
	.init_early	= dvf_early_init,
	.init_machine	= dvf_init_machine,
	.init_irq	= dvf_init_irq,
	.handle_irq	= plicu_handle_irq,
	.timer		= &dvf_timer,
	.reserve	= dvf_reserve,
	.dt_compat	= dvf_dt_match,
	.restart	= dvf_reset,
MACHINE_END

