/*
 *  linux/arch/arm/mach-dvf/clock.c
 *
 *  Copyright (C) 2010 DSP Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/errno.h>
#include <linux/err.h>
#include <linux/string.h>
#include <linux/clk.h>
#include <linux/clkdev.h>
#include <linux/mutex.h>
#include <linux/proc_fs.h>
#include <linux/io.h>
#include <linux/delay.h>
#include <linux/device.h>
#include <linux/seq_file.h>
#include <asm/div64.h>
#include <asm/vfp.h>

#include <mach/platform.h>
#include <mach/hardware.h>
#include <mach/reset.h>

#include <mach/clock.h>

atomic_t dvf_active_bm_cnt = ATOMIC_INIT(0);
static long dvf_active_bm_mask = 0;

#define BM_USB11	0
#define BM_USB20	1
#define BM_CSS		2

static LIST_HEAD(clocks_list);

#define cmu_read(_reg)		readl(IO_ADDRESS(DVF_CMU_BASE) + (_reg))
#define cmu_write(_val, _reg)	writel((_val), IO_ADDRESS(DVF_CMU_BASE) + (_reg))

#define WAIT_PLL_1		(1<<0)
#define WAIT_PLL_2		(1<<1)
#define WAIT_PLL_3		(1<<2)
#define WAIT_OSC_12M	(1<<3)

static struct clk *get_parent_muxed(struct clk *clk);
static int set_parent_muxed(struct clk *clk, struct clk *parent);

static unsigned long get_rate_osc_main(struct clk *clk);
static unsigned long get_rate_parent(struct clk *clk);
static unsigned long get_rate_fixed(struct clk *clk);
static int set_rate_fixed(struct clk *clk, unsigned long rate);
static int get_on_off_always_on(struct clk *clk);
static int get_on_off_gated(struct clk *clk);
static int set_on_off_gated(struct clk *clk, int on);

static int get_on_off_osc_12m(struct clk *clk);
static int set_on_off_osc_12m(struct clk *clk, int on);
static int get_on_off_osc_dect(struct clk *clk);

static unsigned long get_rate_pll(struct clk *clk);
static int set_rate_pll(struct clk *clk, unsigned long rate);
static long round_rate_pll(struct clk *clk, unsigned long rate);
static int get_on_off_pllx(struct clk *clk);
static int set_on_off_pllx(struct clk *clk, int on);

static unsigned long get_rate_parent_div_const(struct clk *clk);

static unsigned long get_rate_ahb(struct clk *clk);
static int set_rate_ahb(struct clk *clk, unsigned long rate);
static long round_rate_ahb(struct clk *clk, unsigned long rate);

static unsigned long get_rate_div(struct clk *clk);
static int set_rate_div(struct clk *clk, unsigned long rate);
static long round_rate_div(struct clk *clk, unsigned long rate);

static int get_on_off_divsclk(struct clk *clk);
static int set_on_off_divsclk(struct clk *clk, int on);

static unsigned long get_rate_bmp(struct clk *clk);
static int set_rate_bmp(struct clk *clk, unsigned long rate);
static long round_rate_bmp(struct clk *clk, unsigned long rate);

static struct clk osc_main = {
	.name = "osc_main",
	.get_rate = get_rate_osc_main,
	.get_on_off = get_on_off_always_on,
};

static struct clk osc_12m = {
	.name = "osc_12m",
	.rate = 12000000,
	.get_rate = get_rate_fixed,
	.get_on_off = get_on_off_osc_12m,
	.set_on_off = set_on_off_osc_12m,
};

static struct clk osc_dect = {
	.name = "osc_dect",
	.rate = 13824000,
	.get_rate = get_rate_fixed,
	.get_on_off = get_on_off_osc_dect,
	/* can't be switched on/off */
};

struct clk* parents_mainclk[] = {&osc_main, &osc_dect};

static struct clk mainclk = {
	.name = "mainclk",
	.mux.reg = DVF_CMU_RSTSR,
	.mux.shift = 1,
	.mux.size = 1,
	.mux.clocks = parents_mainclk,
	.get_parent = get_parent_muxed,
	/* can't set parent */
	.get_rate = get_rate_parent,
	.get_on_off = get_on_off_always_on,
};

struct clk* parents_root[] = {&mainclk, &osc_dect, &osc_12m, &mainclk};

static struct clk audsrcclk = {
	.name = "audsrcclk",
	.mux.reg = DVF_CMU_CPUCLKCNTRL,
	.mux.shift = 16,
	.mux.size = 2,
	.mux.clocks = parents_root,
	.get_parent = get_parent_muxed,
	.set_parent = set_parent_muxed,
	.get_rate = get_rate_parent,
	.get_on_off = get_on_off_always_on,
};

static struct clk pll1 = {
	.name = "pll1",
	.priv.pll.reg = DVF_CMU_PLL1CONTROL,
	.priv.pll.waitflag = WAIT_PLL_1,
	.mux.reg = DVF_CMU_PLL1CONTROL,
	.mux.shift = 19,
	.mux.size = 2,
	.mux.clocks = parents_root,
	.get_parent = get_parent_muxed,
	.set_parent = set_parent_muxed,
	.get_rate = get_rate_pll,
	.set_rate = set_rate_pll,
	.round_rate = round_rate_pll,
	.get_on_off = get_on_off_pllx,
	.set_on_off = set_on_off_pllx,
};

static struct clk pll2 = {
	.name = "pll2",
	.priv.pll.reg = DVF_CMU_PLL2CONTROL,
	.priv.pll.waitflag = WAIT_PLL_2,
	.parent = &mainclk,
	.get_rate = get_rate_pll,
	.set_rate = set_rate_pll,
	.round_rate = round_rate_pll,
	.get_on_off = get_on_off_pllx,
	.set_on_off = set_on_off_pllx,
};

static struct clk pll3 = {
	.name = "pll3",
	.priv.pll.reg = DVF_CMU_PLL3CONTROL,
	.priv.pll.waitflag = WAIT_PLL_3,
	.parent = &audsrcclk,
	.get_rate = get_rate_pll,
	.set_rate = set_rate_pll,
	.round_rate = round_rate_pll,
	.get_on_off = get_on_off_pllx,
	.set_on_off = set_on_off_pllx,
};

struct clk* parents_audclk[] = {&audsrcclk, &pll3};

static struct clk audclk = {
	.name = "audclk",
	.mux.reg = DVF_CMU_CLKSWCNTRL,
	.mux.shift = 2,
	.mux.size = 1,
	.mux.clocks = parents_audclk,
	.get_parent = get_parent_muxed,
	.set_parent = set_parent_muxed,
	.get_rate = get_rate_parent,
	.get_on_off = get_on_off_always_on,
};

static struct clk audclkdiv16 = {
	.name = "audclkdiv16",
	.priv.gate.div = 16,
	.get_rate = get_rate_parent_div_const,
	.parent = &audclk,
	.get_on_off = get_on_off_always_on,
};

static struct clk altsysclk = {
	.name = "altsysclk",
	.mux.reg = DVF_CMU_CPUCLKCNTRL,
	.mux.shift = 14,
	.mux.size = 2,
	.mux.clocks = parents_root,
	.get_parent = get_parent_muxed,
	.set_parent = set_parent_muxed,
	.get_rate = get_rate_parent,
	.get_on_off = get_on_off_always_on,
};

struct clk* parents_sysclk[] = {&altsysclk, &pll1};

static struct clk sysclk = {
	.name = "sysclk",
	.mux.reg = DVF_CMU_CLKSWCNTRL,
	.mux.shift = 1,
	.mux.size = 1,
	.mux.clocks = parents_sysclk,
	.get_parent = get_parent_muxed,
	.set_parent = set_parent_muxed,
	.get_rate = get_rate_parent,
	.get_on_off = get_on_off_always_on,
};

#define CLK_DIV_ON(_name, _parent, _reg, _shift, _size, _flags)	\
	static struct clk _name = {					\
		.name = __stringify(_name),				\
		.flags = (_flags),						\
		.parent = &_parent,						\
		.priv.gate.reg = (_reg),				\
		.priv.gate.shift = (_shift),			\
		.priv.gate.size = (_size),				\
		.get_rate = get_rate_div,				\
		.set_rate = set_rate_div,				\
		.round_rate = round_rate_div,			\
		.get_on_off = get_on_off_always_on,		\
	}

/*          name    	parent		reg                		shift 	size	flags */
CLK_DIV_ON(	hostcpu,	sysclk,		DVF_CMU_CPUCLKCNTRL,	0,		4,		FLAG_ENABLE );

static struct clk slowclk = {
	.name = "slowclk",
	.parent = &mainclk,
	.priv.gate.reg = DVF_CMU_CLKSWCNTRL,
	.priv.gate.shift = 4,
	.priv.gate.size = 11,
	.get_rate = get_rate_div,
	.set_rate = set_rate_div,
	.round_rate = round_rate_div,
	.get_on_off = get_on_off_divsclk,
	.set_on_off = set_on_off_divsclk,
};

static struct clk rtc = {
	.name = "rtc",
	.parent = &slowclk,
	.priv.gate.bit = 2,
	.get_rate = get_rate_parent,
	.get_on_off = get_on_off_gated,
	.set_on_off = set_on_off_gated,
};

static struct clk ahb = {
	.name = "ahb",
	.parent = &hostcpu,
	.get_rate = get_rate_ahb,
	.set_rate = set_rate_ahb,
	.round_rate = round_rate_ahb,
	.get_on_off = get_on_off_always_on,
};

#define CLK_DIV_PER(_name, _parent, _bit, _reg, _shift, _size, _flags)	\
	static struct clk _name = {					\
		.name = __stringify(_name),				\
		.flags = (_flags),						\
		.parent = &_parent,						\
		.priv.gate.bit = (_bit),				\
		.priv.gate.reg = (_reg),				\
		.priv.gate.shift = (_shift),			\
		.priv.gate.size = (_size),				\
		.get_rate = get_rate_div,				\
		.set_rate = set_rate_div,				\
		.round_rate = round_rate_div,			\
		.get_on_off = get_on_off_gated,			\
		.set_on_off = set_on_off_gated,			\
	}

/*          name	parent		bit	reg						shift	size	flags */
CLK_DIV_PER(etm,	hostcpu,	27,	DVF_CMU_CLKDIVCNTRL1,	16,		4,		FLAG_BM | FLAG_IO | FLAG_ENABLE);
CLK_DIV_PER(dclass,	audclk,		18,	DVF_CMU_CLKDIVCNTRL3,	16,		10,		FLAG_BM | FLAG_IO);
CLK_DIV_PER(isdn,	audclk,		12,	DVF_CMU_CLKDIVCNTRL2,	8,		8,		FLAG_NONE);	//TODO: check flags
CLK_DIV_PER(spi2,	sysclk,		39,	DVF_CMU_CLKDIVCNTRL3,	24,		8,		FLAG_IO);

static struct clk dram = {
	.name = "dram",
	.get_rate = get_rate_parent,
	.parent = &pll2,
	.get_on_off = get_on_off_always_on,
};

static struct clk dramctrl = {
	.name = "dramctrl",
	.flags = FLAG_ENABLE,
	.priv.gate.div = 2,
	.priv.gate.bit = 44,
	.get_rate = get_rate_parent_div_const,
	.parent = &dram,
	.get_on_off = get_on_off_gated,
	.set_on_off = set_on_off_gated,
};

static struct clk pll4 = {
        .name = "pll4",
        .parent = &osc_12m,
	.get_on_off = get_on_off_always_on,
        .rate = 480000000,
        .get_rate = get_rate_fixed,
};

#define CLK_DIV_MUX(_name, _bit, _reg, _shift, _size, _muxreg, _muxshift, _muxsize, _parents, _flags)	\
	static struct clk _name = {				\
		.name = __stringify(_name),			\
		.flags = (_flags),					\
		.priv.gate.bit = (_bit),			\
		.priv.gate.reg = (_reg),			\
		.priv.gate.shift = (_shift),		\
		.priv.gate.size = (_size),			\
		.mux.reg = (_muxreg),				\
		.mux.shift = (_muxshift),			\
		.mux.size = (_muxsize),				\
		.mux.clocks = (_parents),			\
		.get_rate = get_rate_div,			\
		.set_rate = set_rate_div,			\
		.round_rate = round_rate_div,		\
		.get_on_off = get_on_off_gated,		\
		.set_on_off = set_on_off_gated,		\
		.get_parent = get_parent_muxed,		\
		.set_parent = set_parent_muxed,		\
	}

struct clk* parents_2fast[] = {&sysclk, &pll2};
struct clk* parents_4fast[] = {&sysclk, &audclk, &pll2, &sysclk};
struct clk* parents_usb11[] = {&sysclk, &pll2, &audclk, &pll4};
struct clk* parents_css[] = {&sysclk, &pll2, &audclk, &mainclk};
struct clk* parents_gpu[] = {&pll2, &sysclk};
struct clk* parents_apuclk[] = {&audclkdiv16, &audclk};

/*          name		bit	reg						shift	size	muxreg	*/
/*			muxshift	muxsize	parents				flags					*/
CLK_DIV_MUX(lcdc,		21,	DVF_CMU_CLKDIVCNTRL1,	8,		8,		DVF_CMU_CLKSWCNTRL,
			17,			1,	parents_2fast,			FLAG_BM | FLAG_IO | FLAG_DONT_DISABLE);
CLK_DIV_MUX(sdmmc,		13,	DVF_CMU_CLKDIVCNTRL1,	24,		8,		DVF_CMU_CLKSWCNTRL,
			27,			1,	parents_2fast,			FLAG_BM | FLAG_IO);
CLK_DIV_MUX(qspic,		11,	DVF_CMU_CLKDIVCNTRL3,	8,		8,		DVF_CMU_CLKSWCNTRL,
			26,			1,	parents_2fast,			FLAG_BM | FLAG_IO);
CLK_DIV_MUX(ciu,		19,	DVF_CMU_CLKDIVCNTRL1,	0,		8,		DVF_CMU_CLKSWCNTRL,
			18,			1,	parents_2fast,			FLAG_BM | FLAG_IO); /* camxck */
CLK_DIV_MUX(refout,		28,	DVF_CMU_CLKDIVCNTRL3,	0,		8,		DVF_CMU_CLKDIVCNTRL2,
			6,			2,	parents_4fast,			FLAG_NONE);
CLK_DIV_MUX(usb11,		41,	DVF_CMU_CLKDIVCNTRL4,	24,		4,		DVF_CMU_CLKDIVCNTRL4,
			28,			2,	parents_usb11,			FLAG_DONT_DISABLE);
CLK_DIV_MUX(tdm1,		9,	DVF_CMU_CLKDIVCNTRL4,	0,		12,		DVF_CMU_CLKSWCNTRL,
			19,			2,	parents_4fast,			FLAG_IO);
CLK_DIV_MUX(tdm2,		10,	DVF_CMU_CLKDIVCNTRL4,	12,		12,		DVF_CMU_CLKSWCNTRL,
			21,			2,	parents_4fast,			FLAG_IO);
CLK_DIV_MUX(css,		23, DVF_CMU_COMCLKSEL,		0,		4,		DVF_CMU_COMCLKSEL,
			10,			2,	parents_css,			FLAG_NONE);
CLK_DIV_MUX(ethr,		48, DVF_CMU_CLKDIVCNTRL2,	20,		8,		DVF_CMU_CLKSWCNTRL,
			25,			1,	parents_2fast,			FLAG_DONT_DISABLE);
CLK_DIV_MUX(gpu,		24,	DVF_CMU_CLKDIVCNTRL2,	16,		4,		DVF_CMU_CLKSWCNTRL,
			15,			1,	parents_gpu,			FLAG_BM);
CLK_DIV_MUX(apuclk,		26,	DVF_CMU_CLKDIVCNTRL5,	24,		8,		DVF_CMU_CLKSWCNTRL,
			16,			1,	parents_apuclk,			FLAG_NONE);

/* ethr with phase shift of one clock cycle of PLL2 out */
static struct clk ethr_ph1 = {
	.name = "ethr_ph1",
	.parent = &ethr,
	.get_rate = get_rate_parent,
	.get_on_off = get_on_off_always_on,
};

/* ethr with phase shift of half clock cycle of PLL2 out */
static struct clk ethr_ph2 = {
	.name = "ethr_ph2",
	.parent = &ethr,
	.get_rate = get_rate_parent,
	.get_on_off = get_on_off_always_on,
};

/* &ethrin is filled at dvf_clk_init */
struct clk* parents_ethmax_txpad[] = {&ethr, NULL /*&ethrin*/, &ethr_ph1, &ethr_ph2 };

static struct clk ethmac_txpad = {
	.name = "ethmac_txpad",
	.mux.reg = DVF_CMU_CLKSWCNTRL,
	.mux.shift = 30,
	.mux.size = 2,
	.mux.clocks = parents_ethmax_txpad,
	.get_parent = get_parent_muxed,
	.set_parent = set_parent_muxed,
	.get_rate = get_rate_parent,
	.get_on_off = get_on_off_always_on,
};

/* rate of eth_external_tx should be set if this clock is used! */
static struct clk eth_external_tx = {
	.name = "eth_external_tx",
	.set_rate = set_rate_fixed,
	.get_rate = get_rate_fixed,
	.get_on_off = get_on_off_always_on,
};

struct clk* parents_eth_txpad[] = {&ethmac_txpad, &eth_external_tx};

static struct clk eth_txpad = {
	.name = "eth_txpad",
	.mux.reg = DVF_CMU_CLKSWCNTRL,
	.mux.shift = 29,
	.mux.size = 1,
	.mux.clocks = parents_eth_txpad,
	.get_parent = get_parent_muxed,
	.set_parent = set_parent_muxed,
	.get_rate = get_rate_parent,
	.get_on_off = get_on_off_always_on,
};

struct clk* parents_eth_tx[] = {&ethr, &eth_txpad};

static struct clk eth_rmii = {
	.name = "eth_rmii",
	.flags = FLAG_DONT_DISABLE,
	.priv.gate.bit = 5,
	.mux.reg = DVF_CMU_CLKDIVCNTRL1,
	.mux.shift = 23,
	.mux.size = 1,
	.mux.clocks = parents_eth_tx,
	.get_parent = get_parent_muxed,
	.set_parent = set_parent_muxed,
	.get_rate = get_rate_parent,
	.get_on_off = get_on_off_gated,
	.set_on_off = set_on_off_gated,
};

CLK_DIV_MUX(ethrin,		49,	DVF_CMU_CLKDIVCNTRL2,	0,		6,		DVF_CMU_CLKDIVCNTRL1,
			23,			1,	parents_eth_tx,			FLAG_DONT_DISABLE);

/* rate of eth_external_rx should be set if this clock is used! */
static struct clk eth_external_rx = {
	.name = "eth_external_rx",
	.set_rate = set_rate_fixed,
	.get_rate = get_rate_fixed,
	.get_on_off = get_on_off_always_on,
};

struct clk* parents_eth_rxpad[] = {&ethrin, &eth_external_rx};

static struct clk eth_rxpad = {
	.name = "eth_rxpad",
	.mux.reg = DVF_CMU_CLKSWCNTRL,
	.mux.shift = 28,
	.mux.size = 1,
	.mux.clocks = parents_eth_rxpad,
	.get_parent = get_parent_muxed,
	.set_parent = set_parent_muxed,
	.get_rate = get_rate_parent,
	.get_on_off = get_on_off_always_on,
};

struct clk* parents_ethmac_rx[] = {&eth_rxpad, &ethrin};

static struct clk ethmac_rx = {
	.name = "ethmac_rx",
	.flags = FLAG_IO | FLAG_BM,
	.priv.gate.bit = 3,	/* same bit as ethmac_tx */
	.mux.reg = DVF_CMU_CLKSWCNTRL,
	.mux.shift = 24,
	.mux.size = 1,
	.mux.clocks = parents_ethmac_rx,
	.get_parent = get_parent_muxed,
	.set_parent = set_parent_muxed,
	.get_rate = get_rate_parent,
	.get_on_off = get_on_off_gated,
	.set_on_off = set_on_off_gated,
};

struct clk* parents_ethmac_tx[] = {&eth_txpad, &ethrin};

static struct clk ethmac_tx = {
	.name = "ethmac_tx",
	.flags = FLAG_IO | FLAG_BM,
	.priv.gate.bit = 3,	/* same bit as ethmac_rx */
	.mux.reg = DVF_CMU_CLKSWCNTRL,
	.mux.shift = 23,
	.mux.size = 1,
	.mux.clocks = parents_ethmac_tx,
	.get_parent = get_parent_muxed,
	.set_parent = set_parent_muxed,
	.get_rate = get_rate_parent,
	.get_on_off = get_on_off_gated,
	.set_on_off = set_on_off_gated,
};

static struct clk gpudiv2 = {
	.name = "gpudiv2",
	.priv.gate.div = 2,
	.get_rate = get_rate_parent_div_const,
	.parent = &gpu,
	.get_on_off = get_on_off_always_on,
};

static struct clk css_arm = {
	.name = "css_arm",
	.priv.gate.bit = 1,
	.parent = &css,
	.get_on_off = get_on_off_gated,
	.set_on_off = set_on_off_gated,
	.get_rate = get_rate_parent,
};

static struct clk css_etm = {
	.name = "css_etm",
	.priv.gate.bit = 31,
	.parent = &css,
	.get_on_off = get_on_off_gated,
	.set_on_off = set_on_off_gated,
	.get_rate = get_rate_parent,
};

static struct clk css_ahb = {
	.name = "css_ahb",
	.priv.gate.div = 2,
	.get_rate = get_rate_parent_div_const,
	.parent = &css,
	.get_on_off = get_on_off_always_on,
};

/*          name		parent		bit	reg					shift	size	flags */
CLK_DIV_PER(css_adpcm,	css_ahb,	29,	DVF_CMU_COMCLKSEL,	4,		5,		FLAG_NONE);

struct clk* parents_bmp[] = {&osc_dect, &apuclk};

static struct clk bmp = {
	.name = "bmp",
	.priv.gate.bit = 30,
	.mux.reg = DVF_CMU_COMCLKSEL,
	.mux.shift = 14,
	.mux.size = 1,
	.mux.clocks = parents_bmp,
	.get_parent = get_parent_muxed,
	.set_parent = set_parent_muxed,
	.get_rate = get_rate_bmp,
	.set_rate = set_rate_bmp,
	.round_rate = round_rate_bmp,
	.get_on_off = get_on_off_gated,
	.set_on_off = set_on_off_gated,
};

static struct clk css_drt = {
	.name = "css_drt",
	.priv.gate.bit = 8,
	.parent = &bmp,
	.get_on_off = get_on_off_gated,
	.set_on_off = set_on_off_gated,
	.get_rate = get_rate_parent,
};

#define CLK_CSS_AHB(_name, _bit, _flags)	\
	static struct clk _name = {		\
		.name = __stringify(_name),	\
		.flags = (_flags),		\
		.parent = &css_ahb,		\
		.priv.gate.bit = (_bit),	\
		.get_rate = get_rate_parent,	\
		.get_on_off = get_on_off_gated,	\
		.set_on_off = set_on_off_gated,	\
	}

CLK_CSS_AHB(rfapu,	25,	0);
CLK_CSS_AHB(css_timer1,	45,	0);
CLK_CSS_AHB(css_timer2,	46,	0);
CLK_CSS_AHB(css_gdmac,	47,	0);

static struct clk usb20_phy = {
	.name = "usb20_phy",
	.parent = &osc_12m,
	.priv.gate.bit = 17,
	.get_on_off = get_on_off_gated,
	.set_on_off = set_on_off_gated,
	.get_rate = get_rate_parent,
};

static struct clk usb20_mac = {
	.name = "usb20_mac",
	.parent = &usb20_phy,
	.priv.gate.bit = 16,
	.get_on_off = get_on_off_gated,
	.set_on_off = set_on_off_gated,
	.rate = 480000000,
	.get_rate = get_rate_fixed,
};

#define CLK_AHB(_name, _bit, _flags)	\
	static struct clk _name = {			\
		.name = __stringify(_name),		\
		.flags = (_flags),				\
		.parent = &ahb,					\
		.priv.gate.bit = (_bit),		\
		.get_rate = get_rate_parent,	\
		.get_on_off = get_on_off_gated,	\
		.set_on_off = set_on_off_gated,	\
	}

CLK_AHB(gdmac,      0, FLAG_BM);
CLK_AHB(secure,     4, FLAG_BM);
CLK_AHB(uart1,      6, 0);
CLK_AHB(uart2,      7, 0);
CLK_AHB(nfc,       14, FLAG_IO | FLAG_BM);
CLK_AHB(pmu,       20, 0);
CLK_AHB(sec,       22, 0);
CLK_AHB(timer1,    32, 0);
CLK_AHB(timer2,    33, 0);
CLK_AHB(timer3,    34, 0);
CLK_AHB(timer4,    35, 0);
CLK_AHB(i2c1,      36, FLAG_IO);
CLK_AHB(i2c2,      37, FLAG_IO);
CLK_AHB(spi1,      38, FLAG_IO);
CLK_AHB(mpmc,      42, FLAG_IO | FLAG_DONT_DISABLE);
CLK_AHB(coresight, 52, FLAG_ENABLE);

struct clk* parents_clkoutdiv[] = {
	&sysclk,	&tdm1,		&tdm2,		&lcdc,		&ethr,		&ethrin,	&css_arm,	&css_adpcm,
	&bmp,		&pll2,		&usb20_mac,	&osc_12m,	&osc_main,	&osc_dect,	&mainclk,	&audclk,
	&slowclk,	&apuclk,	&dclass,	&isdn,		&gpu,		&gpudiv2,	&css_arm,	&qspic,
	&refout,	&ciu,		&sdmmc,		&spi2,		&usb11,		&sysclk,	&sysclk,	&sysclk
};

CLK_DIV_MUX(clkoutdiv,	43, DVF_CMU_CLKDIVCNTRL5,	12,		8,		DVF_CMU_TEST,
			0,			5,	parents_clkoutdiv,		FLAG_NONE);

struct clk* parents_clkout[] = {&mainclk, &hostcpu, &ahb, &clkoutdiv};

static struct clk clkout = {
	.name = "clkout",
	.mux.reg = DVF_CMU_RSTSR,
	.mux.shift = 6,
	.mux.size = 2,
	.mux.clocks = parents_clkout,
	.get_parent = get_parent_muxed,
	/* can't set parent - strap pin */
	.get_on_off = get_on_off_always_on,
	.get_rate = get_rate_parent,
};

static struct clk *onchip_clks[] = {
	&osc_main,
	&osc_12m,
	&osc_dect,
	&mainclk,
	&audsrcclk,
	&pll1,
	&pll2,
	&pll3,
	&pll4,
	&audclk,
	&audclkdiv16,
	&altsysclk,
	&sysclk,
	&hostcpu,
	&spi2,
	&apuclk,
	&slowclk,
	&rtc,
	&ahb,
	&etm,
	&dclass,
	&isdn,
	&gpu,
	&gpudiv2,
	&dram,
	&dramctrl,
	&lcdc,
	&sdmmc,
	&qspic,
	&ciu,
	&refout,
	&usb11,
	&tdm1,
	&tdm2,
	&css,
	&clkoutdiv,
	&clkout,
	&css_arm,
	&css_etm,
	&css_ahb,
	&css_adpcm,
	&bmp,
	&css_drt,
	&rfapu,
	&css_timer1,
	&css_timer2,
	&css_gdmac,
	&usb20_phy,
	&usb20_mac,
	&gdmac,
	&secure,
	&uart1,
	&uart2,
	&nfc,
	&pmu,
	&sec,
	&timer1,
	&timer2,
	&timer3,
	&timer4,
	&i2c1,
	&i2c2,
	&spi1,
	&mpmc,
	&coresight,
	&ethr,
	&ethr_ph1,
	&ethr_ph2,
	&ethmac_txpad,
	&eth_external_tx,
	&eth_txpad,
	&eth_rmii,
	&ethrin,
	&eth_external_rx,
	&eth_rxpad,
	&ethmac_rx,
	&ethmac_tx,
};

/*****************************************************************************
 * Device clock mappings
 ****************************************************************************/

#define CLK_MAP_DEV(_devname, _clk)	\
	{								\
		.dev_id = (_devname),		\
		.clk = (_clk),				\
	}

#define CLK_MAP_DEV_CON(_devname, _conname, _clk)	\
	{												\
		.dev_id = (_devname),						\
		.con_id = (_conname),						\
		.clk = (_clk),								\
	}

static struct clk_lookup clk_mappings[] = {
	CLK_MAP_DEV("5400000.nand",		&nfc),
	CLK_MAP_DEV("5500000.qspic",		&qspic),
	CLK_MAP_DEV("5600000.rtc",	&rtc),
	CLK_MAP_DEV("rtc",		&rtc),
	CLK_MAP_DEV("timer",		&timer1),
	CLK_MAP_DEV("5800020.timer",	&timer2),
	CLK_MAP_DEV("5900000.timer",	&timer3),
	CLK_MAP_DEV("5900020.timer",	&timer4),
	CLK_MAP_DEV("5a00000.serial",	&uart1),
	CLK_MAP_DEV("5b00000.serial",	&uart2),
	CLK_MAP_DEV("5d00000.spi",		&spi1),
	CLK_MAP_DEV("5e00000.spi",		&spi2),
	CLK_MAP_DEV("6100000.i2c",		&i2c2),
	CLK_MAP_DEV("6200000.i2c",		&i2c1),
	CLK_MAP_DEV("6b00000.display",		&lcdc),
	CLK_MAP_DEV("6c00000.secure",	&secure),
	CLK_MAP_DEV("6e00000.ethernet",	&ethmac_tx),	/* can be also ethmac_rx - just for enable/disable */

	CLK_MAP_DEV("900000.mmc",		&sdmmc),
	CLK_MAP_DEV("a00000.sec",		&sec),
	CLK_MAP_DEV("1000000.mpmc",		&mpmc),

	CLK_MAP_DEV("dmw96dma",			&gdmac),
	CLK_MAP_DEV("dvf99ciu.0",		&ciu),
	CLK_MAP_DEV("dvf-tdm.0",		&tdm1),
	CLK_MAP_DEV("dvf-tdm.1",		&tdm2),
	CLK_MAP_DEV("600000.gpu",		&gpu),

	CLK_MAP_DEV_CON("6300000.dvf9a", "pmu",		&pmu),
	CLK_MAP_DEV_CON("6300000.dvf9a", "apu",		&apuclk),

	CLK_MAP_DEV_CON("300000.usb",	"mac",		&usb20_mac),
	CLK_MAP_DEV_CON("300000.usb",	"phy",		&usb20_phy),
	CLK_MAP_DEV_CON("400000.usb",	"mac",		&usb11),

	CLK_MAP_DEV_CON("8200000.css",	"arm",		&css_arm),
	CLK_MAP_DEV_CON("8200000.css",	"css",		&css),
	CLK_MAP_DEV_CON("8200000.css",	"etm",		&css_etm),
	CLK_MAP_DEV_CON("8200000.css",	"timer1",	&css_timer1),
	CLK_MAP_DEV_CON("8200000.css",	"timer2",	&css_timer2),
	CLK_MAP_DEV_CON("8200000.css",	"gdmac",	&css_gdmac),
	CLK_MAP_DEV_CON("8200000.css",	"adpcm",	&css_adpcm),
	CLK_MAP_DEV_CON("8200000.css",	"bmp",		&bmp),
	CLK_MAP_DEV_CON("8200000.css",	"uart0",	&uart1),
	CLK_MAP_DEV_CON("8200000.css",	"uart1",	&uart2),
	CLK_MAP_DEV_CON("8200000.css",	"rtc",		&rtc),
	CLK_MAP_DEV_CON("8200000.css",	"clkout",	&clkout),
	CLK_MAP_DEV_CON("8200000.css",	"dclass",	&dclass),
	CLK_MAP_DEV_CON("8200000.css",	"drt",		&css_drt),
	CLK_MAP_DEV_CON("8200000.css",	"rfapu",	&rfapu),
	CLK_MAP_DEV_CON("8200000.css",	"tdm1",		&tdm1),
	CLK_MAP_DEV_CON("8200000.css",	"tdm2",		&tdm2),
	CLK_MAP_DEV_CON("8200000.css",	"apu",		&apuclk),

	CLK_MAP_DEV_CON("cpufreq-dvf",	"cpu",		&hostcpu),
	CLK_MAP_DEV_CON("pm.0",		"dram",		&dram),
	CLK_MAP_DEV_CON("pm.0",		"sysclk",	&sysclk),
	CLK_MAP_DEV_CON("pm.0",		"altsysclk",	&altsysclk),
};

/*****************************************************************************
 * Clock specific callbacks
 ****************************************************************************/

static struct clk *
get_parent_muxed(struct clk *clk)
{
	u32 reg = cmu_read(clk->mux.reg);
	u32 val = (reg >> clk->mux.shift) & ((1 << clk->mux.size)-1);
	return clk->mux.clocks[val];
}

static int
set_parent_muxed(struct clk *clk, struct clk *parent)
{
	u32 i;
	u32 num_clocks = 1 << clk->mux.size;

	for (i=0 ; i<num_clocks ; i++)
		if (clk->mux.clocks[i] == parent) {
			u32 val = cmu_read(clk->mux.reg) & ~(((1 << clk->mux.size)-1) << clk->mux.shift);
			val |= i << clk->mux.shift;
			cmu_write(val, clk->mux.reg);
			return 0;
		}

	return -EINVAL;
}

static unsigned long
get_rate_osc_main(struct clk *clk)
{
	u32 rstsr = cmu_read(DVF_CMU_RSTSR);

	if (rstsr & (1 << 2))
		return 25000000;
	else
		return 50000000;
}

static unsigned long
get_rate_parent(struct clk *clk)
{
	return clk->parent->get_rate(clk->parent);
}

static unsigned long
get_rate_fixed(struct clk *clk)
{
	return clk->rate;
}

static int
set_rate_fixed(struct clk *clk, unsigned long rate)
{
	clk->rate = rate;
	return 0;
}

static int
get_on_off_always_on(struct clk *clk)
{
	return 1;
}

static int
get_on_off_gated(struct clk *clk)
{
	u32 clken = 0, bit = clk->priv.gate.bit;

	if (bit < 32) {
		clken = cmu_read(DVF_CMU_SWCLKENR1);
	} else if (bit < 64) {
		clken = cmu_read(DVF_CMU_SWCLKENR2);
		bit -= 32;
	}

	return !!(clken & (1 << bit));
}

static int
set_on_off_gated(struct clk *clk, int on)
{
	u32 clken, bit = clk->priv.gate.bit;
	u32 reg = DVF_CMU_SWCLKENR1;

	if (bit >= 32) {
		reg = DVF_CMU_SWCLKENR2;
		bit -= 32;
	}

	do {
		clken = cmu_read(reg);
		if (on)
			clken |=  (1 << bit);
		else
			clken &= ~(1 << bit);
		cmu_write(clken, reg);
	} while (cmu_read(reg) != clken);

	return 0;
}


static int
get_on_off_osc_12m(struct clk *clk)
{
	return !!(cmu_read(DVF_CMU_CLKOSCCR) & (1 << 2));
}

static int
set_on_off_osc_12m(struct clk *clk, int on)
{
	u32 reg;
	int ret = 0;

	/* Note: the HW will not shut-down the oscillator if it is in used by one of it sources */
	/* Our code should not get into that situation... */
	reg = cmu_read(DVF_CMU_CLKOSCCR);

	if (on) {
		reg |= (1 << 2);
		ret = WAIT_OSC_12M;
	} else
		reg &= ~(1 << 2);

	cmu_write(reg, DVF_CMU_CLKOSCCR);

	return ret;
}

static int
get_on_off_osc_dect(struct clk *clk)
{
	/* Note: according to the spec - 0 is enabled! */
	return !(cmu_read(DVF_CMU_CLKOSCCR) & (1 << 0));
}

static unsigned long
get_rate_pll(struct clk *clk)
{
	unsigned long in, prediv, mult, postdiv;
	unsigned long long tmp;
	u32 reg = cmu_read(clk->priv.pll.reg);

	in = clk->parent->get_rate(clk->parent);
	prediv  = ((reg >>  0) & 0x1f) + 1;
	mult    = ((reg >>  5) & 0x7f) + 1;
	postdiv = (reg >> 12) & 0x03;

	tmp = (unsigned long long)in * mult;
	do_div(tmp, prediv);

	return (unsigned long)tmp >> postdiv;
}

#define ABSDIFF(x, y)	((x) > (y) ? (x) - (y) : (y) - (x))

static long
round_rate_pll(struct clk *clk, unsigned long rate)
{
	unsigned long prediv, mult, postdiv;
	unsigned long match_prediv=0, match_mult=0, match_postdiv=0;
	unsigned long match_rate=0, match_diff=~0ul, match_vco=~0ul;

	unsigned long in = clk->parent->get_rate(clk->parent);

	/* loop through all settings finding best match */
	for (prediv=1; prediv<=32; prediv++) {
		unsigned long phaseclk = (in >> 10) / prediv;

		/* keep in phase detector reference frequency: 10..50MHz */
		if (phaseclk < 10000000>>10 || phaseclk > 50000000>>10)
			continue;

		for (mult=6; mult<=100; mult++) {
			unsigned long vcoclk = phaseclk * mult;

			/* keep in VCO frequency range: 300..1000MHz */
			if (vcoclk < 300000000>>10 || vcoclk > 1000000000>>10)
				continue;

			for (postdiv=0; postdiv<=3; postdiv++) {
				unsigned long diff, out = vcoclk >> postdiv;
				unsigned long long tmp;

				/* quick check if in target range */
				if (out-10000 > rate>>10 || out+10000 < rate>>10)
					continue;

				/* ok, do the precise calculation */
				tmp = (unsigned long long)in * mult;
				do_div(tmp, prediv);
				out = (unsigned long)tmp >> postdiv;

				/* better match? */
				diff = ABSDIFF(rate, out);
				if ((diff <  match_diff) ||
				    (diff == match_diff && vcoclk < match_vco)) {
					match_diff    = diff;
					match_vco     = vcoclk;
					match_rate    = out;
					match_prediv  = prediv;
					match_mult    = mult;
					match_postdiv = postdiv;
				}
			}
		}
	}

	if (!match_rate)
		return -EINVAL;

	/* cache setting */
	clk->priv.pll.round_rate     = match_rate;
	clk->priv.pll.round_setting  = (match_prediv  - 1) <<  0;
	clk->priv.pll.round_setting |= (match_mult    - 1) <<  5;
	clk->priv.pll.round_setting |= (match_postdiv    ) << 12;
	clk->priv.pll.round_setting |= match_vco > 600000000>>10 ? 1 << 14 : 0;

	return match_rate;
}

static int
set_rate_pll(struct clk *clk, unsigned long rate)
{
	u32 reg;

	/* See if rate is supported */
	if (rate != clk->priv.pll.round_rate) {
		long ret;
		if ((ret = round_rate_pll(clk, rate)) <= 0)
			return ret;
		if (ret != rate)
			return -EINVAL;
	}

	/* Make sure the PLL is not running */
	if (clk->get_on_off(clk))
		return -EBUSY;

	/* Apply settings */
	reg = cmu_read(clk->priv.pll.reg);
	reg &= ~0x7ffful;
	reg |= clk->priv.pll.round_setting;
	cmu_write(reg, clk->priv.pll.reg);

	ndelay(500);

	return 0;
}

static int
get_on_off_pllx(struct clk *clk)
{
	return !(cmu_read(clk->priv.pll.reg) & 0x8000);
}

static int
set_on_off_pllx(struct clk *clk, int on)
{
	u32 reg = cmu_read(clk->priv.pll.reg);
	if (on) {
		if ((reg & 0x8000) == 0)
			return 0;

		reg &= ~0x18000;
		cmu_write(reg, clk->priv.pll.reg);
		return clk->priv.pll.waitflag;
	} else {
		reg |= 0x18000;
		cmu_write(reg, clk->priv.pll.reg);
		return 0;
	}
}

static unsigned long
get_rate_parent_div_const(struct clk *clk)
{
	return clk->parent->get_rate(clk->parent) / clk->priv.gate.div;
}

static unsigned long get_rate_ahb(struct clk *clk)
{
	unsigned long in = clk->parent->get_rate(clk->parent);
	u32 reg = cmu_read(DVF_CMU_CPUCLKCNTRL);
	u32 div = (reg >> 4) & 0x3;

	if (div == 0)
		div = 2;

	return in / (div + 1);
}

static int set_rate_ahb(struct clk *clk, unsigned long rate)
{
	unsigned long in = clk->parent->get_rate(clk->parent);
	unsigned long div, reg;

	if (!rate)
		return -EINVAL;

	div = in / rate;

	/* AHB divider can divide by 2/3/4 */
	if (div < 2 || div > 4)
		return -EINVAL;
	if (in / div != rate)
		return -EINVAL;

	div--;

	reg = cmu_read(DVF_CMU_CPUCLKCNTRL) & ~(0x3 << 4);
	reg |= div << 4;
	cmu_write(reg, DVF_CMU_CPUCLKCNTRL);

	return 0;
}

static long round_rate_ahb(struct clk *clk, unsigned long rate)
{
	unsigned long in = clk->parent->get_rate(clk->parent);
	unsigned long div;

	if (!rate)
		return -EINVAL;

	/* AHB divider can divide by 2/3/4 */
	div = (in + rate/2) / rate;
	if (div < 2)
		div = 2;
	if (div > 4)
		div = 4;

	return in / div;
}

static unsigned long
get_rate_div(struct clk *clk)
{
	unsigned long in = clk->parent->get_rate(clk->parent);
	u32 reg = cmu_read(clk->priv.gate.reg);
	u32 div = (reg >> clk->priv.gate.shift) & ((1 << clk->priv.gate.size)-1);
	return in / (div + 1);
}

static int
set_rate_div(struct clk *clk, unsigned long rate)
{
	unsigned long in = clk->parent->get_rate(clk->parent);
	unsigned long div, reg;

	if (!rate)
		return -EINVAL;

	div = in / rate;
	if (div == 0 || div > (1 << clk->priv.gate.size))
		return -EINVAL;
	if (in / div != rate)
		return -EINVAL;

	div--;

	reg = cmu_read(clk->priv.gate.reg);
	reg &= ~(((1ul << clk->priv.gate.size) - 1ul) << clk->priv.gate.shift);
	reg |= div << clk->priv.gate.shift;
	cmu_write(reg, clk->priv.gate.reg);

	return 0;
}

static long
round_rate_div(struct clk *clk, unsigned long rate)
{
	unsigned long in = clk->parent->get_rate(clk->parent);
	unsigned long div;

	if (!rate)
		return -EINVAL;

	div = (in + rate/2) / rate;
	if (div == 0)
		div = 1;
	if (div > (1 << clk->priv.gate.size))
		div = 1 << clk->priv.gate.size;

	return in / div;
}

static int
get_on_off_divsclk(struct clk *clk)
{
	return !!(cmu_read(DVF_CMU_CLKSWCNTRL) & (1 << 3));
}

static int
set_on_off_divsclk(struct clk *clk, int on)
{
	u32 reg = cmu_read(DVF_CMU_CLKSWCNTRL);

	if (on)
		reg |= 1 << 3;
	else
		reg &= ~(1 << 3);

	cmu_write(reg, DVF_CMU_CLKSWCNTRL);
	return 0;
}

static unsigned long get_rate_bmp(struct clk *clk)
{
	unsigned long in = clk->parent->get_rate(clk->parent);
	u32 reg = cmu_read(DVF_CMU_COMCLKSEL);

	return (reg & (1 << 9)) ? (in / 12) : in;
}

static int set_rate_bmp(struct clk *clk, unsigned long rate)
{
	unsigned long in = clk->parent->get_rate(clk->parent);
	u32 reg = cmu_read(DVF_CMU_COMCLKSEL);

	if (rate == in)
		reg &= ~(1 << 9);
	else if (rate == (in /12))
		reg |= 1 << 9;
	else
		return -EINVAL;

	cmu_write(reg, DVF_CMU_COMCLKSEL);

	return 0;
}

static long round_rate_bmp(struct clk *clk, unsigned long rate)
{
	unsigned long in = clk->parent->get_rate(clk->parent);

	if (rate < in)
		return in/12;
	else
		return in;
}

/*****************************************************************************
 * Internal helper functions
 ****************************************************************************/

static void
clk_inc_io(struct clk *clk)
{
	clk->iocnt++;
	if (clk->iocnt == 1 && clk->parent)
		clk_inc_io(clk->parent);
}

static int
clk_enable_locked(struct clk *clk)
{
	int ret = 0;

	clk->refcnt++;

	if (clk->refcnt == 1) {
		if (clk->parent)
			ret |= clk_enable_locked(clk->parent);

		if (!clk->get_on_off(clk) && clk->set_on_off)
			ret |= clk->set_on_off(clk, 1);

		if (clk->flags & FLAG_BM)
			atomic_inc(&dvf_active_bm_cnt);
		if (clk->flags & FLAG_IO)
			clk_inc_io(clk);
	}

	return ret;
}

static void
clk_dec_io(struct clk *clk)
{
	clk->iocnt--;
	if (clk->iocnt == 0 && clk->parent)
		clk_dec_io(clk->parent);
}

static void
clk_disable_locked(struct clk *clk)
{
	clk->refcnt--;
	BUG_ON(clk->refcnt < 0);

	if (clk->refcnt == 0) {
		if (clk->get_on_off(clk) && clk->set_on_off)
			clk->set_on_off(clk, 0);

		if (clk->parent)
			clk_disable_locked(clk->parent);

		if (clk->flags & FLAG_BM)
			atomic_dec(&dvf_active_bm_cnt);
		if (clk->flags & FLAG_IO)
			clk_dec_io(clk);
	}
}

/*
 * Re-parenting is a bit more involved. It must be divided in two parts because
 * the new parent may have to be enabled before actually switching to it. This
 * requires enabling IRQs and FIQs to keep the latency impact to the system to
 * a minimum.
 */

static int
clk_reparent_locked_prepare(struct clk *clk, struct clk *new_parent)
{
	struct clk *old_parent = clk->parent;
	int ret = 0;

	if (clk->flags & FLAG_REPARENTING)
		return -EBUSY;

	if (old_parent == new_parent)
		goto done;

	if (!clk->set_parent)
		return -EINVAL;

	if (clk->refcnt) {
		ret = clk_enable_locked(new_parent);
		/* make sure we can't be switched off */
		if (ret >= 0)
			ret |= clk_enable_locked(clk);
	}

done:
	if (ret >= 0)
		clk->flags |= FLAG_REPARENTING;
	return ret;
}

static int
clk_reparent_locked_switch(struct clk *clk, struct clk *new_parent)
{
	struct clk *old_parent = clk->parent;
	int ret = 0;

	if (old_parent == new_parent)
		goto done;

	ret = clk->set_parent(clk, new_parent);
	if (ret < 0) {
		/* Revert clk_reparent_locked_prepare() actions */
		if (clk->refcnt) {
			clk_disable_locked(clk);
			clk_disable_locked(new_parent);
		}
		goto done;
	}

	list_move(&clk->child_node, &new_parent->childs);
	clk->parent = new_parent;

	if (clk->refcnt) {
		/* adjust IO refcount if necessary */
		if (clk->flags & FLAG_IO) {
			clk_inc_io(new_parent);
			clk_dec_io(old_parent);
		}

		clk_disable_locked(clk);
		clk_disable_locked(old_parent);
	}

done:
	clk->flags &= ~FLAG_REPARENTING;
	return ret;
}

static int
clk_wait(int flags)
{
	if (flags <= 0)
		return flags;

repeat:
	if (flags & WAIT_PLL_1 && !(cmu_read(DVF_CMU_PLL1CONTROL) & (1 << 18)))
		goto repeat;
	if (flags & WAIT_PLL_2 && !(cmu_read(DVF_CMU_PLL2CONTROL) & (1 << 18)))
		goto repeat;
	if (flags & WAIT_PLL_3 && !(cmu_read(DVF_CMU_PLL3CONTROL) & (1 << 18)))
		goto repeat;
	if (flags & WAIT_OSC_12M && !(cmu_read(DVF_CMU_CLKOSCCR) & (1 << 5)))
		goto repeat;

	return 0;
}

/*****************************************************************************
 * Public interface
 ****************************************************************************/

int
clk_enable(struct clk *clk)
{
	unsigned long flags;
	int ret = 0;

	if (!clk)
		return -ENODEV;

	raw_local_irq_save(flags);
	local_fiq_disable();

	ret = clk_enable_locked(clk);

	raw_local_irq_restore(flags);

	return clk_wait(ret);
}
EXPORT_SYMBOL(clk_enable);

void clk_disable(struct clk *clk)
{
	unsigned long flags;

	if (!clk)
		return;

	raw_local_irq_save(flags);
	local_fiq_disable();

	clk_disable_locked(clk);

	raw_local_irq_restore(flags);
}
EXPORT_SYMBOL(clk_disable);

long
clk_round_rate(struct clk *clk, unsigned long rate)
{
	unsigned long flags;

	if (!clk)
		return -ENODEV;

	if (clk->round_rate) {
		raw_local_irq_save(flags);
		local_fiq_disable();

		rate = clk->round_rate(clk, rate);

		raw_local_irq_restore(flags);
	}

	return rate;
}
EXPORT_SYMBOL(clk_round_rate);

unsigned long
clk_get_rate(struct clk *clk)
{
	unsigned long rate, flags;

	if (!clk)
		return -ENODEV;

	raw_local_irq_save(flags);
	local_fiq_disable();

	rate = clk->get_rate(clk);

	raw_local_irq_restore(flags);

	return rate;
}
EXPORT_SYMBOL(clk_get_rate);

int clk_set_rate(struct clk *clk, unsigned long rate)
{
	unsigned long flags;
	int ret = -EINVAL;

	if (!clk)
		return -ENODEV;

	if (clk->set_rate) {
		raw_local_irq_save(flags);
		local_fiq_disable();

		ret = clk->set_rate(clk, rate);

		raw_local_irq_restore(flags);
	}

	return clk_wait(ret);
}
EXPORT_SYMBOL(clk_set_rate);

struct clk *clk_get_parent(struct clk *clk)
{
	if (!clk)
		return ERR_PTR(-ENODEV);

	return clk->parent;
}
EXPORT_SYMBOL(clk_get_parent);

int clk_set_parent(struct clk *clk, struct clk *parent)
{
	unsigned long flags;
	int ret;

	if (!clk)
		return -ENODEV;

	raw_local_irq_save(flags);
	local_fiq_disable();
	ret = clk_reparent_locked_prepare(clk, parent);
	raw_local_irq_restore(flags);

	if (ret < 0)
		return ret;
	clk_wait(ret);

	raw_local_irq_save(flags);
	local_fiq_disable();
	ret = clk_reparent_locked_switch(clk, parent);
	raw_local_irq_restore(flags);

	return ret;
}
EXPORT_SYMBOL(clk_set_parent);


/* /proc/clocks information */

static void
clocks_show_clock(struct seq_file *f, struct clk *clk, int indent)
{
	for (; indent>0; indent--)
		seq_printf(f, "\t");

	if (clk->get_on_off(clk)) {
		unsigned long rate = clk_get_rate(clk);
		unsigned long int mhz = rate / 1000000;
		unsigned long int khz = (rate - (mhz * 1000000)) / 1000;

		seq_printf(f, "%s\t refcnt:%d iocnt:%d", clk->name,
			   clk->refcnt, clk->iocnt);
		if (mhz && khz)
			seq_printf(f, " rate:%lu.%03luMHz", mhz, khz);
		else if (mhz && !khz)
			seq_printf(f, " rate:%luMHz", mhz);
		else
			seq_printf(f, " rate:%lukHz", khz);

		if (clk->flags & FLAG_BM)
			seq_printf(f, " +BM");
		if (clk->flags & FLAG_IO)
			seq_printf(f, " +IO");
	} else {
		seq_printf(f, "(%s)\t", clk->name);
		seq_printf(f, " refcnt:%d iocnt:%d", clk->refcnt,
			   clk->iocnt);
	}

	if (clk->getcnt)
		seq_printf(f, " users:%d", clk->getcnt);

	seq_printf(f, "\n");
}

static void
clocks_show_list(struct seq_file *f, struct list_head *list, int indent)
{
	struct clk *clk;

	list_for_each_entry(clk, list, child_node) {
		clocks_show_clock(f, clk, indent);
		clocks_show_list(f, &clk->childs, indent+1);
	}
}

static int clocks_show(struct seq_file *f, void *unused)
{
	int i;

	seq_printf(f, "Active bus masters: %d", atomic_read(&dvf_active_bm_cnt));
	if (atomic_read(&dvf_active_bm_cnt)) {
		seq_printf(f, " (");

		for (i = 0; i < ARRAY_SIZE(onchip_clks); i++) {
			struct clk *clk = onchip_clks[i];
			if (clk->flags & FLAG_BM && clk->refcnt)
				seq_printf(f, " %s", clk->name);
		}

		if (dvf_active_bm_mask & (1 << BM_USB11))
			seq_printf(f, " usb1.1");
		if (dvf_active_bm_mask & (1 << BM_USB20))
			seq_printf(f, " usb2.0");
		if (dvf_active_bm_mask & (1 << BM_CSS))
			seq_printf(f, " css");

		seq_printf(f, ")");
	}

	seq_printf(f, "\nActive IO clocks on PLL1: %d\n\n", pll1.iocnt);

	clocks_show_list(f, &clocks_list, 0);

	return 0;
}

static int clocks_open(struct inode *inode, struct file *file)
{
	return single_open(file, clocks_show, NULL);
}

static const struct file_operations clock_fops = {
	.owner = THIS_MODULE,
	.open = clocks_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = single_release,
};

static void clk_init_node(struct clk *clk)
{
	struct clk *child;

	/* First init childs */
	list_for_each_entry(child, &clk->childs, child_node)
		clk_init_node(child);

	/* Enable if requested */
	if (clk->flags & FLAG_ENABLE) {
		/* enable statically */
		clk_enable_locked(clk);
	} else if (!clk->refcnt && clk->get_on_off(clk)) {
		if (clk->flags & FLAG_DONT_DISABLE) {
			/* propagate flag upwards */
			struct clk *parent = clk->parent;
			while (parent) {
				parent->flags |= FLAG_DONT_DISABLE;
				parent = parent->parent;
			}
		} else if (clk->set_on_off)
			clk->set_on_off(clk, 0);
	}
}

static void
reset_set_onoff(int id, int on)
{
	int tmp, bit = id, reg = DVF_CMU_SWCHWRST2;
	unsigned long flags;

	if ((id < 0) || (id > 63))
		return;
	else if (bit < 32)
		reg = DVF_CMU_SWCHWRST1;
	else
		bit -= 32;

	raw_local_irq_save(flags);
	local_fiq_disable();

	do {
		tmp = cmu_read(reg);
		if (on)
			tmp |=  (1 << bit);
		else
			tmp &= ~(1 << bit);
		cmu_write(tmp, reg);
	} while (cmu_read(reg) != tmp);

	raw_local_irq_restore(flags);
}

void
reset_set(int id)
{
	reset_set_onoff(id, 1);
}
EXPORT_SYMBOL(reset_set);

void
reset_release(int id)
{
	reset_set_onoff(id, 0);
}
EXPORT_SYMBOL(reset_release);

int
dvf_get_wdt_reset_ind(void)
{
	int ret;

	ret = !!(cmu_read(DVF_CMU_RSTSR) & (1<<3));

	return ret;
}

void dvf_clk_print_active(void)
{
	int i;

	printk("Active IO/BM clocks:\n");

	for (i = 0; i < ARRAY_SIZE(onchip_clks); i++) {
		struct clk *clk = onchip_clks[i];

		if ((clk->flags & (FLAG_IO|FLAG_BM)) && clk->refcnt > 0)
			printk("  %s%s%s\n",
			       clk->name,
			       clk->flags & FLAG_IO ? " +IO" : "",
			       clk->flags & FLAG_BM ? " +BM" : "");
	}
}

void dvf_usb11_busy(void)
{
	if (!test_and_set_bit(BM_USB11, &dvf_active_bm_mask))
		atomic_inc(&dvf_active_bm_cnt);
}

void dvf_usb11_idle(void)
{
	if (test_and_clear_bit(BM_USB11, &dvf_active_bm_mask))
		atomic_dec(&dvf_active_bm_cnt);
}

void dvf_usb20_busy(void)
{
	if (!test_and_set_bit(BM_USB20, &dvf_active_bm_mask))
		atomic_inc(&dvf_active_bm_cnt);
}

void dvf_usb20_idle(void)
{
	if (test_and_clear_bit(BM_USB20, &dvf_active_bm_mask))
		atomic_dec(&dvf_active_bm_cnt);
}

void dvf_css_busy(void)
{
	if (!test_and_set_bit(BM_CSS, &dvf_active_bm_mask))
		atomic_inc(&dvf_active_bm_cnt);
}

void dvf_css_idle(void)
{
	if (test_and_clear_bit(BM_CSS, &dvf_active_bm_mask))
		atomic_dec(&dvf_active_bm_cnt);
}

static void cmu_update_field(u8 reg, u8 shift, u8 size, u32 val)
{
	unsigned long flags;
	u32 regval;
	u32 mask = ((1 << size)-1) << shift;

	raw_local_irq_save(flags);
	local_fiq_disable();

	regval = cmu_read(reg);
	regval &= ~mask;
	regval |= (val << shift) & mask;
	cmu_write(regval, reg);

	raw_local_irq_restore(flags);
}

static int check_eth_config(enum eth_mode mode, enum eth_speed speed)
{
	switch (mode) {
		case ETH_MII:
		case ETH_SLVMII:
		case ETH_RMII:
		case ETH_SLVRMII:
			if (speed == ETH_SPEED_10 || speed == ETH_SPEED_100)
				return 0;
			break;
		case ETH_RGMII:
			if (speed == ETH_SPEED_1000)
				return 0;
			break;
		default:
			break;
	};

	return -1;
}

struct eth_mode_settings {
	u8 tx_clk_sel;
	u8 rx_clk_sel;
	u8 tx_pad_src_sel;
	u8 tx_pad_oen;
	u8 rx_pad_oen;
	u8 int_ethrdiv_en;
	int ethint_rate[NUM_ETH_SPEEDS];
};

/* Table 1 from CMU specification */
struct eth_mode_settings eth_clk_settings[NUM_ETH_MODES] = {
	/* MII */			{0, 0, 0, 0, 1, 0,	{2500000,	25000000,	-1} },
	/* SLAVE_MII */		{1, 0, 0, 0, 1, 0,	{2500000,	25000000,	-1} },
	/* RMII */			{1, 1, 0, 0, 1, 1,	{2500000,	25000000,	-1} },
	/* SLAVE RMII */	{1, 1, 0, 0, 1, 1,	{2500000,	25000000,	-1} },
	/* RGMII */			{1, 0, 3, 0, 1, 1,	{-1,		-1,			125000000} },
};

unsigned long ethrdiv_rates[] =
{
	/* MII */		25000000,
	/* SLVMII */	25000000,
	/* RMII */		50000000,
	/* SLVRMII */	50000000,
	/* RGMII */		125000000,
};

int clk_update_etherdiv(enum eth_mode mode, enum eth_speed speed)
{
	struct eth_mode_settings *settings;
	unsigned long rate = ethrdiv_rates[mode];

	settings = &eth_clk_settings[mode];
	if (settings->int_ethrdiv_en) {
		u32 div = rate / settings->ethint_rate[speed];

		if (div * settings->ethint_rate[speed] != rate)
			return -EINVAL;

		cmu_update_field(DVF_CMU_CLKDIVCNTRL2,  0, 6, div - 1);
	}

	return 0;
}

int dvf_eth_config(enum eth_mode mode, enum eth_speed speed)
{
	int ret;
	struct eth_mode_settings *settings;
	unsigned long rate = ethrdiv_rates[mode];

	if (check_eth_config(mode, speed))
		return -EINVAL;

	ret = clk_set_rate(&ethr, rate);

	if (ret)
		return ret;
	
	settings = &eth_clk_settings[mode];

	cmu_update_field(DVF_CMU_CLKSWCNTRL, 23, 1, settings->tx_clk_sel);
	cmu_update_field(DVF_CMU_CLKSWCNTRL, 24, 1, settings->rx_clk_sel);
	cmu_update_field(DVF_CMU_CLKSWCNTRL, 30, 2, settings->tx_pad_src_sel);
	cmu_update_field(DVF_CMU_CLKSWCNTRL, 29, 1, settings->tx_pad_oen);
	cmu_update_field(DVF_CMU_CLKSWCNTRL, 28, 1, settings->rx_pad_oen);
	cmu_update_field(DVF_CMU_SWCLKENR2,  17, 1, settings->int_ethrdiv_en);

	if (settings->int_ethrdiv_en) {
		u32 div = rate / settings->ethint_rate[speed];

		if (div * settings->ethint_rate[speed] != rate)
			return -EINVAL;

		cmu_update_field(DVF_CMU_CLKDIVCNTRL2,  0, 6, div - 1);
		clk_enable(&ethrin);
	}

	return 0;
}

int __init dvf_clk_init(void)
{
	int i;
	struct clk *clk;
	unsigned long tmp;

	/* Fill &ethrin here and not at decleration because of sircular dependencies... */
	parents_ethmax_txpad[1] = &ethrin;

	/* On chips prior to ABC/BBC the FS-USB cannot be clocked from PLL4 */
	if (dvf_get_chip_rev() <= 1)
		parents_usb11[3] = &sysclk;

	/* Reset SD/MMC block */
	tmp = cmu_read(DVF_CMU_SWCHWRST1);
	tmp |= (1<<13);
	cmu_write(tmp, DVF_CMU_SWCHWRST1);
	udelay(1000);

	/* Release reset of all blocks (except for USB20 & CSS) */
	tmp = cmu_read(DVF_CMU_SWCHWRST1);
	tmp &= ~(0x0f7bffdd);
	cmu_write(tmp, DVF_CMU_SWCHWRST1);
	tmp = cmu_read(DVF_CMU_SWCHWRST2);
	tmp &= ~(0x003f0ff7);
	cmu_write(tmp, DVF_CMU_SWCHWRST2);

	/* Configure clkoutdiv parent to osc_dect without division */
	tmp = cmu_read(DVF_CMU_TEST);
	tmp &= ~0x1f;
	tmp |= 13; /* CLK_13M_OSC */
	cmu_write(tmp, DVF_CMU_TEST);
	tmp = cmu_read(DVF_CMU_CLKDIVCNTRL5);
	tmp &= ~(0xff << 12);
	cmu_write(tmp, DVF_CMU_CLKDIVCNTRL5);

	/* Round 1: set up parents */
	for (i = 0; i < ARRAY_SIZE(onchip_clks); i++) {
		clk = onchip_clks[i];

		INIT_LIST_HEAD(&clk->childs);
		if (clk->get_parent)
			clk->parent = clk->get_parent(clk);
	}

	/* Round 2: set up child lists */
	for (i = 0; i < ARRAY_SIZE(onchip_clks); i++) {
		clk = onchip_clks[i];

		if (clk->parent)
			list_add(&clk->child_node, &clk->parent->childs);
		else
			list_add(&clk->child_node, &clocks_list);
	}

	/* Round 3: enable the clocks needed during early bootup, disable unneeded ones */
	list_for_each_entry(clk, &clocks_list, child_node)
		clk_init_node(clk);

	/* Add clock mappings for devices */
	for (i = 0; i < ARRAY_SIZE(clk_mappings); i++)
		clkdev_add(&clk_mappings[i]);

#if 0
	/* cpu freq is not possible since CPU is running from the same clock as peripherals (sysclk) */
	for (i = 0; i < ARRAY_SIZE(clk_mappings_cpufreq); i++)
		clkdev_add(&clk_mappings_cpufreq[i]);
#endif

	return 0;
}

static int __init clk_init_late(void)
{
	/* Create proc entry */
	proc_create_data("clocks", S_IRUGO, NULL, &clock_fops, NULL);

	/* Disable CoreSight if debugging disabled */
	if (dvf_debug_disabled()) {
		unsigned long tmp;

		clk_disable(&etm);
		clk_disable(&coresight);

		tmp = cmu_read(DVF_CMU_TEST);
		tmp &= ~(1 << 11); /* CSS_ARM_CLK_EN_BP */
		tmp &= ~(1 << 12); /* ARM_CLK_EN_BP */
		cmu_write(tmp, DVF_CMU_TEST);
	}

	return 0;
}

arch_initcall(clk_init_late);

