/****************************************************************************
 *
 * FILENAME:        $RCSfile: factory.c,v $
 *
 * LAST REVISION:   $Revision: 1.6 $
 * LAST MODIFIED:   $Date: 2014/03/05 15:07:38 $
 *
 * DESCRIPTION:     factory data support
 *
 * vi: set ts=4:
 *
 * Copyright (c) 2013 by Grandstream Networks, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***************************************************************************/

//=======================
//  Includes
//=======================
#include <linux/types.h>
#include <linux/vmalloc.h>
#include <linux/mtd/mtd.h>
#include <mach/factory.h>
#include <provision.h>

//=======================
//  Defines
//=======================
//must be updated for every new model inthe table
#define HW_MODEL_TABLE_SIZE     (sizeof(hw_ids) / (3 * sizeof(uint16_t)) )

#ifndef UINT16
#define UINT16 unsigned short int
#endif

//=======================
//  Globals
//=======================

//=======================
//  Typedefs
//=======================

//=======================
//  Locals
//=======================

    /*
     * Private Functions
     */

    /*
     * Private Data
     */
static hw_info_t hw_info;
static sec_info_t sec_info;
static char factory_para_buf[GS_FACTORY_CFG_SIZE];
static GS_FACTORYPARA * factory_para = NULL;
static int feature_table_loaded = 0;
static int factory_para_dirty = 0;

//=======================
//  Functions
//=======================

//===========================================================================
static void
load_factory_para(
    void
)
//===========================================================================
{
    int len;
    struct mtd_info *mtd;

    mtd = get_mtd_device_nm( "factory" );
    if ( mtd <= 0 )
    {
        printk("load_factory_para: unable to open mtd device\n");
        return;
    }

    mtd_read( mtd, 0, GS_FACTORY_CFG_SIZE, &len, factory_para_buf );
    if ( len != GS_FACTORY_CFG_SIZE )
    {
        printk("load_factory_para: mtd_read read size != GS_FACTORY_CFG_SIZE\n");
    }
    

    put_mtd_device( mtd );
}

//===========================================================================
uint16_t
fact_para_checksum(
    GS_FACTORYPARA * factPara
)
//===========================================================================
{
    GS_FACTORYPARA tmpData;
    uint8_t *p = (uint8_t*)&tmpData;
    uint16_t sum = 0;
    uint16_t checksum = 0;
    unsigned int length = 0;
    int i;

    length = ( factPara->length[1] << 16 ) | factPara->length[0];
    if ( length > GS_FACTORY_CFG_SIZE )
    {
        printk("fact_para_checksum: length exceeds GS_FACTORY_CFG_SIZE\n");
        return 0xDEAD;
    }
    p = ( unsigned char * )vmalloc( length );
	if (p == NULL) {
        printk("fact_para_checksum: vmalloc failed\n");
        return 0xDEAD;
    }
    memcpy( p, factPara, length );
    ((GS_FACTORYPARA*)p)->checksum = 0;

    for ( i = 0; i < length; i+= 2 )
    {
         sum += ((p[i] << 8) | p[i+1]);
    }
    checksum = 0xffff - sum + 0x01;

    vfree( p );

    return ((checksum >> 8) | (checksum << 8));
}


//===========================================================================
GS_FACTORYPARA *
fact_para_get(
    void
)
//===========================================================================
{
    if ( factory_para == NULL || factory_para_dirty )
    {
        factory_para = (GS_FACTORYPARA *)factory_para_buf;
        load_factory_para();
        factory_para_dirty = 0;
    }
    return factory_para;
}

//===========================================================================
void
free_memory(
    void
)
//===========================================================================
{
    if ( hw_info.hw_logo_data )
        vfree( hw_info.hw_logo_data );

    if ( sec_info.certificate.data )
        vfree( sec_info.certificate.data );

    if ( sec_info.private_key.data )
        vfree( sec_info.private_key.data );

    if ( sec_info.sn.data )
        vfree( sec_info.sn.data );

    if ( sec_info.hash.data )
        vfree( sec_info.hash.data );

    if ( sec_info.password.data )
        vfree( sec_info.password.data );

    if ( sec_info.xml_key.data )
        vfree( sec_info.xml_key.data );

    if ( sec_info.checksum.data )
        vfree( sec_info.checksum.data );
}

//===========================================================================
void
set_factory_para_dirty(
    void
)
//===========================================================================
{
    factory_para_dirty = 1;
    feature_table_loaded = 0;
}

//===========================================================================
sec_info_t*
sec_info_get(
    void
)
//===========================================================================
{
    hw_get_hw_feature();
    return &sec_info;
}

//===========================================================================
hw_info_t *
hw_get_hw_feature(
    void
)
// Description: Get the hardware logic features from HW feature table in flash factory
//              parameter sector
//===========================================================================
{
    factory_param_sector_t *fact_para;
    hw_feature_table_header_t *hwTableHdr;
    hw_feature_attrib_t *featurePtr;
    uint16_t tableLength, parsedLength, index, featureLen;
    static int has_security_table = 0;

    if ( feature_table_loaded )
    {
        return &hw_info;
    }

    if ( factory_para_dirty )
        free_memory();
    
    fact_para = (factory_param_sector_t*)fact_para_get();
    memset( &hw_info, 0, sizeof ( hw_info ) );
    memset( &sec_info, 0, sizeof ( sec_info ) );

    hw_info.poe_class = 0xFF; /* Set so we can test for existence in flash */
    hw_info.num_linekeys = 0xFF;	// Initial NULL value for models that do not have it in flash.
    hw_info.lcd_backlight = 0xFF; /* NULL value set so we can test for existence in flash */

	// make sure there is at least one HW feature attribute in the flash
   if( !(fact_para->factory_para.length[0] >= sizeof(factory_param_sector_t)) )
   {
       printk("Warning: No factory table\n");
       return &hw_info;
   }

    // make sure the flash sector contains a complete HW feature table
    tableLength = fact_para->hw_feature_table_header.hw_feature_table_len;
    tableLength += sizeof(GS_FACTORYPARA);

    if ( fact_para->factory_para.length[0] >= tableLength )
    {
        has_security_table = 1;
    }

    // locate the HW feature table in the sector, in case there are other tables in it
    hwTableHdr = &fact_para->hw_feature_table_header;
    while (HW_FEATURE_TABLE_VERSION != hwTableHdr->hw_feature_table_version)
    {
        hwTableHdr = (hw_feature_table_header_t*)( (uint16_t*)hwTableHdr + hwTableHdr->hw_feature_table_len);
        if ((uint16_t*)hwTableHdr-(uint16_t*)fact_para >= fact_para->factory_para.length[0])
        {  // did not find the HW feature table in the flash sector
            hwTableHdr = NULL;
            break;
        }
    }
    if ( !(NULL!= hwTableHdr) )
    {
        printk("Warning: Can't find feature table\n");
        return &hw_info;
    }

    tableLength = sizeof(uint16_t) * hwTableHdr->hw_feature_table_len;
    featurePtr = &fact_para->hw_feature;
    parsedLength = 6;

    for (index=0; index<hwTableHdr->hw_feature_attrib_count; index++)
    {
        switch (featurePtr->hw_feature_type)
        {
            case HW_FEATURE_VENDOR:
                memcpy( &hw_info.vendor, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(UINT16) );
                break;
            case HW_FEATURE_MODELSTR:
                memcpy( &hw_info.modelstr, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(UINT16) );
                break;
            case HW_FEATURE_MODELALIAS:
                memcpy( &hw_info.modelalias, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(UINT16) );
                break;
            case HW_FEATURE_OEM_ID:
                hw_info.oem_id = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_VENDOR_FULLNAME:
                memcpy( &hw_info.vendor_fullname, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(UINT16) );
                break;
            case HW_FEATURE_FWKEY:
                memcpy( &hw_info.fwkey, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof(UINT16) );
                break;
            case HW_FEATURE_ETH:
                hw_info.eth = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_HANDSET:
                hw_info.handset = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_NUM_ACCTS:
                hw_info.num_accts = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_EXT:
                hw_info.ext = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_NUM_MPK:
                hw_info.mpk = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_CPU:
                memcpy( &hw_info.cpu, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( UINT16 ) );
                break;
            case HW_FEATURE_DECT:
                hw_info.dect = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_DAP1912:
                hw_info.dap1912 = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_RTL8363_SMI:
                hw_info.rtl8363_smi = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_DECT_HS:
                hw_info.dect_hs = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_DECT_REGION:
                memcpy( &hw_info.dect_region, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( UINT16 ) );
                break;
            case HW_FEATURE_POE_CLASS:
                hw_info.poe_class = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_FLASH_SIZE:
                memcpy( &hw_info.flash_size, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( UINT16 ) );
                break;
            case HW_FEATURE_NUM_SOFTKEYS:
                hw_info.num_softkeys = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_MODEL_BASE:
                memcpy( &hw_info.gs_model, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( UINT16 ) );
                break;
            case HW_FEATURE_BLUETOOTH:
                hw_info.bluetooth = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_BLUETOOTH_CHIP:
                memcpy( &hw_info.bluetooth_chip, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( UINT16 ) );
                break;
            case HW_FEATURE_NO_USB:
                hw_info.usb_unsupported = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_SI3217X_DEV:
                hw_info.si3217x_dev = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_NUV_DEV:
                hw_info.nuv_dev = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_NUV_NEON:
                hw_info.nuv_neon = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_NUM_LINEKEYS:
                hw_info.num_linekeys = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_HEADSET_MIC_DETECT:
                hw_info.headset_mic_detect = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_MANUFACT_ID:
                 hw_info.hw_manufact_id = featurePtr->hw_feature_value;
                 break;
            case HW_FEATURE_MANUFACT_OPT_1:
                 hw_info.hw_manufact_opt_1 = featurePtr->hw_feature_value;
                 break;
            case HW_FEATURE_MANUFACT_OPT_2:
                 hw_info.hw_manufact_opt_2 = featurePtr->hw_feature_value;
                 break;
            case HW_FEATURE_SLIC_VENDOR:
                 hw_info.slic_vendor = featurePtr->hw_feature_value;
                 break;
            case HW_FEATURE_SLIC_DCDC_TYPE:
                 hw_info.slic_dcdc_type = featurePtr->hw_feature_value;
                 break;
            case HW_FEATURE_SLIC_DCDC_VIN:
                 hw_info.slic_dcdc_vin = featurePtr->hw_feature_value;
                 break;
            case HW_FEATURE_SLIC_HW_PROT:
                 hw_info.slic_hw_prot = featurePtr->hw_feature_value;
                 break;
            case HW_FEATURE_NUM_SLIC_DEV:
                 hw_info.num_slic_dev = featurePtr->hw_feature_value;
                 break;
            case HW_FEATURE_NUM_CH_PER_SLIC:
                 hw_info.num_ch_per_slic = featurePtr->hw_feature_value;
                 break;
            case HW_FEATURE_NUM_DAA_DEV:
                 hw_info.num_daa_dev = featurePtr->hw_feature_value;
                 break;
            case HW_FEATURE_NUM_CH_PER_DAA:
                 hw_info.num_ch_per_daa = featurePtr->hw_feature_value;
                 break;
            case HW_FEATURE_NEON_SUPPORT:
                 hw_info.neon_support = featurePtr->hw_feature_value;
                 break;
            case HW_FEATURE_MAX_REN_PER_CH:
                 hw_info.max_ren_per_ch = featurePtr->hw_feature_value;
                 break;
            case HW_FEATURE_LCD_BACKLIGHT:
                 hw_info.lcd_backlight = featurePtr->hw_feature_value;
                 break;
            case HW_FEATURE_DCLASS_FIX:
                hw_info.dclass_fix = featurePtr->hw_feature_value;
                break;
            case HW_FEATURE_REQUIRE_PEM:
                 hw_info.require_pem = featurePtr->hw_feature_value;
                 break;
            case HW_FEATURE_REQUIRE_SN:
                 hw_info.require_sn = featurePtr->hw_feature_value;
                 break;
            case HW_FEATURE_REQUIRE_ADMIN_PWD:
                 hw_info.require_admin_pwd = featurePtr->hw_feature_value;
                 break;
            case HW_FEATURE_REQUIRE_XML_KEY:
                 hw_info.require_xml_key = featurePtr->hw_feature_value;
                 break;
            case HW_FEATURE_REQUIRE_HASH:
                 hw_info.require_hash = featurePtr->hw_feature_value;
                 break;
            case HW_FEATURE_ETH_CLOCK_SOURCE:
                 hw_info.eth_clock_source = featurePtr->hw_feature_value;
                 break;
            default:
                break;
        }

        // move to the next attribute
        featureLen = sizeof(hw_feature_attrib_t) + sizeof(uint16_t)*( featurePtr->hw_feature_len - 1 );
        parsedLength += featureLen;
        featurePtr = (hw_feature_attrib_t*)((char*)featurePtr + featureLen);
        if (parsedLength >= tableLength)
        {
            break;
        }
    }

    feature_table_loaded = 1;
    
    // locate the security table
    hwTableHdr = &fact_para->hw_feature_table_header;
    while (HW_SECURITY_TABLE_VERSION != hwTableHdr->hw_feature_table_version)
    {
        hwTableHdr = (hw_feature_table_header_t*)( (uint16_t*)hwTableHdr + hwTableHdr->hw_feature_table_len);
        if ((uint16_t*)hwTableHdr-(uint16_t*)fact_para >= fact_para->factory_para.length[0] || hwTableHdr->hw_feature_table_version == 0)
        {  // did not find the HW feature table in the flash sector
            hwTableHdr = NULL;
            break;
        }
    }
    if ( !(NULL!= hwTableHdr) )
    {
        //dump_stack();
        printk("Warning: Can't find security table\n");
        return &hw_info;
    }

    tableLength = sizeof(uint16_t) * hwTableHdr->hw_feature_table_len;
    featurePtr = (hw_feature_attrib_t*)( (uint8_t*)hwTableHdr + sizeof ( hw_feature_table_header_t ) );
    parsedLength = 6;

    //printk("Load Security Table\n");
    //printk("tableLength: %d\n", tableLength);
    //printk("attribCount: %d\n", hwTableHdr->hw_feature_attrib_count);
    
    for (index=0; index<hwTableHdr->hw_feature_attrib_count; index++)
    {
        //printk("Found Feature: %d\n", featurePtr->hw_feature_type);
        switch (featurePtr->hw_feature_type)
        {
            case SEC_FEATURE_SN:
                sec_info.sn.data = vmalloc( featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                memcpy( sec_info.sn.data, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                sec_info.sn.len = featurePtr->hw_feature_len * sizeof ( uint16_t  );
                break;
            case SEC_FEATURE_PRIVATE_KEY:
                sec_info.private_key.data = vmalloc( featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                memcpy( sec_info.private_key.data, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                sec_info.private_key.len = featurePtr->hw_feature_len * sizeof ( uint16_t );
                //printk("Feature Size: %d\n", sec_info.private_key.len);
                break;
            case SEC_FEATURE_CERTIFICATE:
                sec_info.certificate.data = vmalloc( featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                memcpy( sec_info.certificate.data, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                sec_info.certificate.len = featurePtr->hw_feature_len * sizeof ( uint16_t );
                //printk("Feature Size: %d\n", sec_info.certificate.len );
                break;
            case SEC_FEATURE_CHECKSUM:
                sec_info.checksum.data = vmalloc( featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                memcpy( sec_info.checksum.data, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                sec_info.checksum.len = featurePtr->hw_feature_len * sizeof ( uint16_t );
                //printk("Feature Size: %d\n", sec_info.checksum.len );
                break;
            case SEC_FEATURE_VONAGE_HASH:
                sec_info.hash.data = vmalloc( featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                memcpy( sec_info.hash.data, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                sec_info.hash.len = featurePtr->hw_feature_len * sizeof ( uint16_t );
                break;
            case SEC_FEATURE_VONAGE_PASSWORD:
                sec_info.password.data = vmalloc( featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                memcpy( sec_info.password.data, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                sec_info.password.len = featurePtr->hw_feature_len * sizeof ( uint16_t );
                break;
            case SEC_FEATURE_VONAGE_XML_KEY:
                sec_info.xml_key.data = vmalloc( featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                memcpy( sec_info.xml_key.data, &featurePtr->hw_feature_value, featurePtr->hw_feature_len * sizeof ( uint16_t ) );
                sec_info.xml_key.len = featurePtr->hw_feature_len * sizeof ( uint16_t );
                break;
        }
        
        // move to the next attribute
        featureLen = sizeof(hw_feature_attrib_t) + sizeof(uint16_t)*( featurePtr->hw_feature_len - 1 );
        parsedLength += featureLen;
        featurePtr = (hw_feature_attrib_t*)((char*)featurePtr + featureLen);
        if (parsedLength >= tableLength)
        {
            break;
        }
    }

    //printk("Done\n");

    return &hw_info;
}

/* EOF */
