/****************************************************************************
 *
 * FILENAME:            $RCSfile: factory_proc.c,v $
 *
 * LAST REVISION:       $Revision: 1.14 $
 * LAST MODIFIED:   $Date: 2014/03/05 15:10:41 $
 *
 * DESCRIPTION:         sipp55 application entry point
 *
 * vi: set ts=4:
 *
 * Copyright (c) 2009 by Grandstream Networks, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***************************************************************************/

//=======================
//      Includes
//=======================
#include <linux/types.h>
#include <linux/proc_fs.h>
#include <linux/uaccess.h>
#include <linux/module.h>
#include <linux/mtd/mtd.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/mfd/dvf9a.h>
#include <linux/gpio.h>
#include <linux/of.h>
#include <linux/of_gpio.h>
#include <linux/seq_file.h>
#include <video/dspgfb.h>
#include <mach/factory.h>
#include <provision_funcs.h>

//=======================
//     Globals
//=======================

//=======================
//      Defines
//=======================
#define DISABLE_SECURITY_PROCFS_WRITE

#define DECLARE_READ_PROC( var, type )\
static int \
var##_read_proc(\
    char * page,\
    char ** start,\
    off_t off,\
    int count,\
    int * eof,\
    int * data\
)\
{\
    return type##_read_proc( page, start, off, count, eof, data, &var );\
}

#define DECLARE_WRITE_PROC( var, type, accessor )\
static int \
var##_write_proc(\
    struct file * file,\
    const char * buffer,\
    unsigned long count,\
    void * data\
)\
{\
    return type##_write_proc( file, buffer, count, data, &var, accessor );\
}


//=======================
//      Typedefs
//=======================
struct gs_procfs_entry
{
    char * name;
    read_proc_t * read_proc;
    write_proc_t * write_proc;
    void ( *init )( void );
    int value;
};

//=======================
//      Locals
//=======================

        /*
         * Private Functions
         */
static int erase_helper( struct mtd_info* mtd, loff_t offset, size_t len );
static void erase_callback( struct erase_info* info );
#ifdef CONFIG_GXP
extern void gs_lcd_set_contrast( int percent );
void lcd_brightness_set( int value );
static void lcd_brightness_init(void);
#endif
        /*
         * Private Data
         */
static struct {
    int enabled;
    unsigned int value;
} oem_override;

#ifdef CONFIG_GXP
static unsigned int lcd_on_off;
static unsigned int lcd_brightness;
static unsigned int lcd_contrast;
#endif
#ifdef CONFIG_HT
static unsigned int reset_gpio;
#endif
static char syslog_id[64];

//=======================
//      Functions
//=======================

//=============================================================================
static int
erase_helper(
    struct mtd_info* mtd,
    loff_t offset,
    size_t len
)
//=============================================================================
{
    struct erase_info info;
    DECLARE_COMPLETION_ONSTACK(complete);
    int ret;

    memset( &info, 0, sizeof ( info ) );
    info.mtd = mtd;
    info.addr = offset;
    info.len = len;
    info.callback = erase_callback;
    info.priv = (long)&complete;

    ret = mtd_erase( mtd, &info );
    if ( ret )
        return -EIO;

    wait_for_completion( &complete );
    if ( info.state != MTD_ERASE_DONE )
        return -EIO;

    return 0;
}

//=============================================================================
static void
erase_callback(
    struct erase_info* info
)
//=============================================================================
{
    complete( (struct completion *)info->priv );
}

#ifdef CONFIG_GXP
//=============================================================================
static int
integer_read_proc(
    char * page,
    char ** start,
    off_t off,
    int count,
    int * eof,
    int * data,
    int * integer
)
//=============================================================================
{
    char * p = page;
    int len = 0;

    p += sprintf( p, "%d\n", *integer );

    len = p - page;
    if ( len <= off + count )
    {
        *eof = 1;
    }

    *start = page + off;
    len -= off;

    if ( len > count )
    {
        len = count;
    }
    if ( len < 0 )
    {
        len = 0;
    }

    return len;
}

//=============================================================================
static int
integer_write_proc(
    struct file * file,
    const char * buffer,
    unsigned long count,
    void * data,
    int * integer,
    void (*set)( int val )
)
//=============================================================================
{
    char state_string[ 12 ] = {'\0'};
    char * endp;
    unsigned long new_value = 0;

    if ( ( count > sizeof ( state_string ) - 1 ) )
    {
        return -EINVAL;
    }

    if ( copy_from_user( state_string, buffer, count ) )
    {
        return -EFAULT;
    }

    state_string[ count ] = '\0';

    new_value = simple_strtoul( state_string, &endp, 10 /* Base */ );
	*integer = new_value;
    set( *integer );

    return count;
}

//=============================================================================
static void 
lcd_on_off_init( void )
//=============================================================================
{
	lcd_on_off = 1; // lcd is enabled by default
}

//=============================================================================
void 
lcd_on_off_set( int value )
//=============================================================================
{
	printk("%s: value = %d\n", __func__, value);
	if (value == 0) {
        {
            dvf9a_pwm_stop();
        }
	}
	else if (value == 1) {
        dvf9a_pwm_start();
#if defined( CONFIG_FB_DSPG )
        dspgfb_set_brightness(lcd_brightness);
#else
        dvf9a_pwm_set_duty(lcd_brightness);
#endif
	}
#if defined( CONFIG_FB_DSPG )
    dspgfb_set_lcd_on_off(value);
#endif
}

DECLARE_WRITE_PROC( lcd_on_off, integer, lcd_on_off_set );
DECLARE_READ_PROC( lcd_on_off, integer);

//=============================================================================
static void 
lcd_brightness_init(void)
//=============================================================================
{
    lcd_brightness = 70;
}

//=============================================================================
void 
lcd_brightness_set( int value )
// LCD brightness value range: [0:100]
//=============================================================================
{
	//printk("%s: value = %d\n", __func__, value);
#if defined( CONFIG_FB_DSPG )
    dspgfb_set_brightness(value);
#else
    dvf9a_pwm_set_duty(lcd_brightness);
#endif
}

DECLARE_WRITE_PROC( lcd_brightness, integer, lcd_brightness_set );
DECLARE_READ_PROC( lcd_brightness, integer );

//=============================================================================
static void
lcd_contrast_init(void)
//=============================================================================
{
	lcd_contrast = 33;
}
//=============================================================================
static void lcd_contrast_set( int value )
//=============================================================================
{
    if ( value < 0 || value > 100 )
    {
        value = 33;
    }

    lcd_contrast = value;
    gs_lcd_set_contrast( lcd_contrast );
}

DECLARE_WRITE_PROC( lcd_contrast, integer, lcd_contrast_set );
DECLARE_READ_PROC( lcd_contrast, integer );
#endif

#ifdef CONFIG_HT
//=============================================================================
static void 
reset_button_init(void)
//=============================================================================
{
}

//=============================================================================
static int
reset_button_read_proc(
    char * page,
    char ** start,
    off_t off,
    int count,
    int * eof,
    int * data
)
//=============================================================================
{
    char * p = page;
    int len = 0;
    int gpio_val = !gpio_get_value( reset_gpio );

    p += sprintf( p, "%x\n", gpio_val );

    len = p - page;
    if ( len <= off + count )
    {
        *eof = 1;
    }

    *start = page + off;
    len -= off;

    if ( len > count )
    {
        len = count;
    }
    if ( len < 0 )
    {
        len = 0;
    }

    return len;
}

//=============================================================================
static int
reset_button_write_proc(
    struct file * file,
    const char * buffer,
    unsigned long count,
    void * data
)
//=============================================================================
{
    return -EINVAL;
}
#endif

//=============================================================================
static int
gs_oem_override(
    struct file * file,
    const char * buffer,
    unsigned long count,
    void * data
)
//=============================================================================
{
    int i;

    for ( i = 0; i < count - 1; i++ )
    {
        if ( !( buffer[i] >= '0' && buffer[i] <= '9' ) )
        {
            printk("%s: value must be a positive integer\n", __func__);
            return -EINVAL;
        }
    }

    oem_override.value = (unsigned int)simple_strtoul( buffer, NULL, 10 );
    oem_override.enabled = 1;

    return count;
}


/*
 * ProcFS Entries
 */
struct gs_procfs_entry gs_procfs_entry[] =
{
#define GS_PROCFS_ENTRY( name ) { #name, (read_proc_t*)name##_read_proc, (write_proc_t*)name##_write_proc, name##_init }
#ifdef CONFIG_GXP
    GS_PROCFS_ENTRY( lcd_on_off ),		/* 1 = on; 0 = off */
    GS_PROCFS_ENTRY( lcd_brightness ),
    GS_PROCFS_ENTRY( lcd_contrast ),
#endif
#ifdef CONFIG_HT
    GS_PROCFS_ENTRY( reset_button ),
#endif
#undef GS_PROCFS_ENTRY
};
#define NUM_PROCFS_ENTRIES ( sizeof ( gs_procfs_entry ) / sizeof ( struct gs_procfs_entry ) )


static int devinfo_proc_read( char * page, char **start, off_t off, int count, int * eof, void * data );
struct gs_procfs_entry gs_devinfo_entry[] =
{
#define GS_DEVINFO_ENTRY( name ) { #name, devinfo_proc_read, NULL, NULL }
#define iPN 0
        GS_DEVINFO_ENTRY( PN ),
#define idev_alias 1
        GS_DEVINFO_ENTRY( dev_alias ),
#define idev_data 2
        GS_DEVINFO_ENTRY( dev_data ),
#define idev_id 3
        GS_DEVINFO_ENTRY( dev_id ),
#define idev_mac 4
        GS_DEVINFO_ENTRY( dev_mac ),
#define idev_rev 5
        GS_DEVINFO_ENTRY( dev_rev ),
#define irequired_bits 6
        GS_DEVINFO_ENTRY( required_bits ),
#define imodel_key 7
        GS_DEVINFO_ENTRY( model_key ),
#define idev_prefix 8
            GS_DEVINFO_ENTRY( dev_prefix )
#undef GS_DEVINFO_ENTRY
};
#define NUM_DEVINFO_ENTRIES ( sizeof ( gs_devinfo_entry ) / sizeof ( struct gs_procfs_entry ) )

static int hw_feature_proc_read( char * page, char **start, off_t off, int count, int * eof, void * data );
struct gs_procfs_entry hw_feature_entry[] =
{
#define HW_FEATURE_ENTRY( name, value ) { #name, NULL, NULL, NULL, value }
    HW_FEATURE_ENTRY( vendor, HW_FEATURE_VENDOR ),
        HW_FEATURE_ENTRY( model_str, HW_FEATURE_MODELSTR ),
        HW_FEATURE_ENTRY( model_alias, HW_FEATURE_MODELALIAS ),
        HW_FEATURE_ENTRY( default_fwkey, HW_FEATURE_FWKEY ),
        HW_FEATURE_ENTRY( oem_id, HW_FEATURE_OEM_ID ),
        HW_FEATURE_ENTRY( vendor_fullname, HW_FEATURE_VENDOR_FULLNAME ),
        HW_FEATURE_ENTRY( eth, HW_FEATURE_ETH ),
        HW_FEATURE_ENTRY( handset, HW_FEATURE_HANDSET ),
        HW_FEATURE_ENTRY( nuv_dev, HW_FEATURE_NUV_DEV ),
        HW_FEATURE_ENTRY( si3217x_dev, HW_FEATURE_SI3217X_DEV ),
        HW_FEATURE_ENTRY( nuv_neon, HW_FEATURE_NUV_NEON ),
        HW_FEATURE_ENTRY( num_accts, HW_FEATURE_NUM_ACCTS ),
        HW_FEATURE_ENTRY( ext, HW_FEATURE_EXT ),
        HW_FEATURE_ENTRY( num_mpk, HW_FEATURE_NUM_MPK ),
        HW_FEATURE_ENTRY( num_softkeys, HW_FEATURE_NUM_SOFTKEYS ),
        HW_FEATURE_ENTRY( gs_model, HW_FEATURE_MODEL_BASE ),
        HW_FEATURE_ENTRY( cpu, HW_FEATURE_CPU ),
        HW_FEATURE_ENTRY( flash, HW_FEATURE_FLASH_SIZE ),
        HW_FEATURE_ENTRY( dect, HW_FEATURE_DECT ),
        HW_FEATURE_ENTRY( dect_hs, HW_FEATURE_DECT_HS ),
        HW_FEATURE_ENTRY( bluetooth, HW_FEATURE_BLUETOOTH ),
        HW_FEATURE_ENTRY( poe_class, HW_FEATURE_POE_CLASS ),
        HW_FEATURE_ENTRY( dect_region, HW_FEATURE_DECT_REGION ),
        HW_FEATURE_ENTRY( rtl8363_smi, HW_FEATURE_RTL8363_SMI ),
        HW_FEATURE_ENTRY( dap1912, HW_FEATURE_DAP1912 ),
        HW_FEATURE_ENTRY( num_linekeys, HW_FEATURE_NUM_LINEKEYS ),
        HW_FEATURE_ENTRY( headset_mic_detect, HW_FEATURE_HEADSET_MIC_DETECT ),
        HW_FEATURE_ENTRY( hw_manufact_id, HW_FEATURE_MANUFACT_ID ),
        HW_FEATURE_ENTRY( hw_manufact_opt_1, HW_FEATURE_MANUFACT_OPT_1 ),
        HW_FEATURE_ENTRY( hw_manufact_opt_2, HW_FEATURE_MANUFACT_OPT_2 ),
        HW_FEATURE_ENTRY( slic_vendor, HW_FEATURE_SLIC_VENDOR ),
        HW_FEATURE_ENTRY( slic_dcdc_type, HW_FEATURE_SLIC_DCDC_TYPE ),
        HW_FEATURE_ENTRY( slic_dcdc_vin, HW_FEATURE_SLIC_DCDC_VIN ),
        HW_FEATURE_ENTRY( slic_hw_prot, HW_FEATURE_SLIC_HW_PROT ),
        HW_FEATURE_ENTRY( num_slic_dev, HW_FEATURE_NUM_SLIC_DEV ),
        HW_FEATURE_ENTRY( num_ch_per_slic, HW_FEATURE_NUM_CH_PER_SLIC ),
        HW_FEATURE_ENTRY( num_daa_dev, HW_FEATURE_NUM_DAA_DEV ),
        HW_FEATURE_ENTRY( num_ch_per_daa, HW_FEATURE_NUM_CH_PER_DAA ),
        HW_FEATURE_ENTRY( neon_support, HW_FEATURE_NEON_SUPPORT ),
        HW_FEATURE_ENTRY( max_ren_per_ch, HW_FEATURE_MAX_REN_PER_CH ),
        HW_FEATURE_ENTRY( lcd_backlight, HW_FEATURE_LCD_BACKLIGHT ),
        HW_FEATURE_ENTRY( dclass_fix, HW_FEATURE_DCLASS_FIX ),
        HW_FEATURE_ENTRY( require_pem, HW_FEATURE_REQUIRE_PEM ),
        HW_FEATURE_ENTRY( require_sn, HW_FEATURE_REQUIRE_SN ),
        HW_FEATURE_ENTRY( require_admin_pwd, HW_FEATURE_REQUIRE_ADMIN_PWD ),
        HW_FEATURE_ENTRY( require_xml_key, HW_FEATURE_REQUIRE_XML_KEY ),
        HW_FEATURE_ENTRY( require_hash, HW_FEATURE_REQUIRE_HASH ),
        HW_FEATURE_ENTRY( eth_clock_source, HW_FEATURE_ETH_CLOCK_SOURCE ),
};
#define NUM_HW_FEATURE_ENTRIES ( sizeof ( hw_feature_entry ) / sizeof ( struct gs_procfs_entry ) )

static int sec_feature_proc_read( char *page, char **start, off_t off, int count, int *eof, void *data );
struct gs_procfs_entry sec_feature_entry[] =
{
#define SEC_FEATURE_ENTRY( name, value ) { #name, NULL, NULL, NULL, value }
    SEC_FEATURE_ENTRY( private_key, SEC_FEATURE_PRIVATE_KEY ),
    SEC_FEATURE_ENTRY( certificate, SEC_FEATURE_CERTIFICATE ),
    SEC_FEATURE_ENTRY( checksum, SEC_FEATURE_CHECKSUM ),
    SEC_FEATURE_ENTRY( sn, SEC_FEATURE_SN ),
    SEC_FEATURE_ENTRY( hash, SEC_FEATURE_VONAGE_HASH ),
    SEC_FEATURE_ENTRY( password, SEC_FEATURE_VONAGE_PASSWORD ),
    SEC_FEATURE_ENTRY( xml_key, SEC_FEATURE_VONAGE_XML_KEY ),
};
#define NUM_SEC_FEATURE_ENTRIES ( sizeof ( sec_feature_entry ) / sizeof ( struct gs_procfs_entry ) )

//=============================================================================
hw_info_t *
hw_feature_get(
    void
)
//=============================================================================
{
    static hw_info_t * hw_info = NULL;

    if ( hw_info == NULL )
    {
        GS_FACTORYPARA * factory_para;

        factory_para = fact_para_get();
        if ( factory_para->checksum != fact_para_checksum( factory_para ) )
        {
            printk("INVALID FACTORY CHECKSUM\n");
        }

        hw_info = hw_get_hw_feature();
    }

    return hw_info;
}

//=============================================================================
static int hw_feature_proc_read(
    char * page,
    char ** start,
    off_t off,
    int count,
    int * eof,
    void * data
)
//=============================================================================
{
    int len = 0;
    char str[128];
    hw_info_t * hw_info = hw_feature_get();

    switch ( (int)data )
    {
        case HW_FEATURE_VENDOR:
            strcpy( str, hw_info->vendor );
            break;
        case HW_FEATURE_VENDOR_FULLNAME:
            strcpy( str, hw_info->vendor_fullname );
            break;
        case HW_FEATURE_MODELSTR:
            strcpy( str, hw_info->modelstr );
            break;
        case HW_FEATURE_MODELALIAS:
            strcpy( str, hw_info->modelalias );
            break;
        case HW_FEATURE_MODEL_BASE:
            strcpy( str, hw_info->gs_model );
            break;
        case HW_FEATURE_OEM_ID:
            sprintf( str, "%x", hw_info->oem_id );
            break;
        case HW_FEATURE_NUV_DEV:
            sprintf( str, "%x", hw_info->nuv_dev );
            break;
        case HW_FEATURE_SI3217X_DEV:
            sprintf( str, "%x", hw_info->si3217x_dev );
            break;
        case HW_FEATURE_NUV_NEON:
            sprintf( str, "%x", hw_info->nuv_neon );
            break;
        case HW_FEATURE_FWKEY:
            if ( hw_info->fwkey[0] == 0 )
            {
                strcpy( str, "nokey" );
            }
            else
            {
                strcpy( str, hw_info->fwkey );
            }
            break;
        case HW_FEATURE_ETH:
            switch ( hw_info->eth ) {
                case ETH_TYPE_KSZ8873:
                    sprintf( str, "KSZ8873" );
                    break;
                case ETH_TYPE_KSZ8863:
                    sprintf( str, "KSZ8863" );
                    break;
                case ETH_TYPE_ICPLUS:
                    sprintf( str, "IC+" );
                    break;
                case ETH_TYPE_KSZ8051:
                    sprintf( str, "KSZ8051" );
                    break;
                case ETH_TYPE_KSZ8031:
                    sprintf( str, "KSZ8031" );
                    break;
                case ETH_TYPE_RTL8363SB:
                    sprintf( str, "RTL8363SB" );
                    break;
                case ETH_TYPE_RTL8304:
                    sprintf( str, "RTL8304" );
                    break;
                default:
                    sprintf( str, "Unknown" );
                    break;
            }
            break;
        case HW_FEATURE_HANDSET:
            switch ( hw_info->handset ) {
            case HANDSET_WIDEBAND:
                sprintf( str, "wideband" );
                break;
            case HANDSET_NARROWBAND:
                sprintf( str, "narrowband" );
                break;
            }
            break;
        case HW_FEATURE_NUM_ACCTS:
            sprintf( str, "%d", hw_info->num_accts );
            break;
        case HW_FEATURE_EXT:
            sprintf( str, "%d", hw_info->ext );
            break;
        case HW_FEATURE_BLUETOOTH:
            sprintf( str, "%d", hw_info->bluetooth);
            break;
        case HW_FEATURE_NUM_MPK:
            sprintf( str, "%d", hw_info->mpk );
            break;
        case HW_FEATURE_NUM_SOFTKEYS:
            sprintf( str, "%d", hw_info->num_softkeys );
            break;
        case HW_FEATURE_NO_USB:
            sprintf( str, "%d", hw_info->usb_unsupported );
            break;

        case HW_FEATURE_CPU:
            if ( hw_info->cpu[0] == 0 )
            {
                sprintf( str, "9918" );
            }
            else
                strcpy( str, hw_info->cpu );
            break;
        case HW_FEATURE_FLASH_SIZE:
            if ( hw_info->flash_size[0] == 0 )
            {
                sprintf( str, "16" );
            }
            else
                strcpy( str, hw_info->flash_size );
            break;

        case HW_FEATURE_DECT:
            sprintf( str, "%d", hw_info->dect );
            break;
        case HW_FEATURE_DECT_HS:
            sprintf( str, "%d", hw_info->dect_hs );
            break;
        case HW_FEATURE_DECT_REGION:
            strcpy( str, hw_info->dect_region );
            break;
        case HW_FEATURE_POE_CLASS:
            sprintf( str, "%d", hw_info->poe_class );
            break;
        case HW_FEATURE_RTL8363_SMI:
            sprintf( str, "%d", hw_info->rtl8363_smi );
            break;
        case HW_FEATURE_DAP1912:
            sprintf( str, "%d", hw_info->dap1912 );
            break;
        case HW_FEATURE_NUM_LINEKEYS:
            sprintf( str, "%d", hw_info->num_linekeys );
            break;
        case HW_FEATURE_HEADSET_MIC_DETECT:
            sprintf( str, "%d", hw_info->headset_mic_detect );
            break;

        case HW_FEATURE_MANUFACT_ID:
             sprintf( str, "%d", hw_info->hw_manufact_id );
             break;
        case HW_FEATURE_MANUFACT_OPT_1:
             sprintf( str, "%d", hw_info->hw_manufact_opt_1 );
             break;
        case HW_FEATURE_MANUFACT_OPT_2:
             sprintf( str, "%d", hw_info->hw_manufact_opt_2 );
             break;
        case HW_FEATURE_SLIC_VENDOR:
             sprintf( str, "%d", hw_info->slic_vendor );
             break;
        case HW_FEATURE_SLIC_DCDC_TYPE:
             sprintf( str, "%d", hw_info->slic_dcdc_type );
             break;
        case HW_FEATURE_SLIC_DCDC_VIN:
             sprintf( str, "%d", hw_info->slic_dcdc_vin );
             break;
        case HW_FEATURE_SLIC_HW_PROT:
             sprintf( str, "%d", hw_info->slic_hw_prot );
             break;
        case HW_FEATURE_NUM_SLIC_DEV:
             sprintf( str, "%d", hw_info->num_slic_dev );
             break;
        case HW_FEATURE_NUM_CH_PER_SLIC:
             sprintf( str, "%d", hw_info->num_ch_per_slic );
             break;
        case HW_FEATURE_NUM_DAA_DEV:
             sprintf( str, "%d", hw_info->num_daa_dev );
             break;
        case HW_FEATURE_NUM_CH_PER_DAA:
             sprintf( str, "%d", hw_info->num_ch_per_daa );
             break;
        case HW_FEATURE_NEON_SUPPORT:
             sprintf( str, "%d", hw_info->neon_support );
             break;
        case HW_FEATURE_MAX_REN_PER_CH:
             sprintf( str, "%d", hw_info->max_ren_per_ch );
             break;
        case HW_FEATURE_LCD_BACKLIGHT:
             // If lcd_backlight is not set in Flash, assume that the backlight is there. 
             sprintf( str, "%d", (hw_info->lcd_backlight == 0xFF) ? 1 : hw_info->lcd_backlight );
             break;
        case HW_FEATURE_DCLASS_FIX:
             sprintf( str, "%d", hw_info->dclass_fix );
             break;
         case HW_FEATURE_REQUIRE_PEM:
              sprintf( str, "%d", hw_info->require_pem );
              break;
         case HW_FEATURE_REQUIRE_SN:
               sprintf( str, "%d", hw_info->require_sn );
               break;
         case HW_FEATURE_REQUIRE_ADMIN_PWD:
              sprintf( str, "%d", hw_info->require_admin_pwd );
              break;
         case HW_FEATURE_REQUIRE_XML_KEY:
              sprintf( str, "%d", hw_info->require_xml_key );
              break;
         case HW_FEATURE_REQUIRE_HASH:
              sprintf( str, "%d", hw_info->require_hash );
              break;
         case HW_FEATURE_ETH_CLOCK_SOURCE:
              sprintf( str, "%d", hw_info->eth_clock_source );
              break;
    }

    len = strlen( str );
    memcpy( page, str, len );

    if ( len <= off + count )
        *eof = 1;
    *start = page + off;
    len -= off;
    if ( len > count )
        len = count;
    if ( len < 0 )
        len = 0;

    return len;
}

//=============================================================================
static int sec_feature_proc_read(
    char * page,
    char ** start,
    off_t off,
    int count,
    int * eof,
    void * data
)
//=============================================================================
{
    int len = 0;
    sec_info_t * sec_info = sec_info_get();

    switch ( (int)data )
    {
        case SEC_FEATURE_SN:
            if ( sec_info->sn.len == 0 )
            {
                printk("serial number not set\n");
                return -EINVAL;
            }
            len = strlen( sec_info->sn.data );
            if( len > sec_info->sn.len )
                printk( "sn length is greater than the stored length\n" );
            memcpy( page, sec_info->sn.data, len );
            break;
        case SEC_FEATURE_PRIVATE_KEY:
            if ( sec_info->private_key.len == 0 )
            {
                printk("private key not set\n");
                return -EINVAL;
            }
            len = strlen( sec_info->private_key.data );
            if( len > sec_info->private_key.len )
                printk( "private key length is greater than the stored length\n" );
            memcpy( page, sec_info->private_key.data, len );
            break;
        case SEC_FEATURE_CERTIFICATE:
            if ( sec_info->certificate.len == 0 )
            {
                printk("certificate not set\n");
                return -EINVAL;
            }
            len = strlen( sec_info->certificate.data );
            if( len > sec_info->certificate.len )
                printk( "certificate length is greater than the stored length\n" );
            memcpy( page, sec_info->certificate.data, len );
            break;
        case SEC_FEATURE_CHECKSUM:
            if ( sec_info->checksum.len == 0 )
            {
                printk("checksum not set\n");
                return -EINVAL;
            }
            len = strlen( sec_info->checksum.data );
            if( len > sec_info->checksum.len )
                printk( "checksum length is greater than the stored length\n" );
            memcpy( page, sec_info->checksum.data, len );
            break;
        case SEC_FEATURE_VONAGE_HASH:
            if ( sec_info->hash.len == 0 )
            {
                printk("hash not set\n");
                return -EINVAL;
            }
            len = strlen( sec_info->hash.data );
            if( len > sec_info->hash.len )
                printk( "hash length is greater than the stored length\n" );
            memcpy( page, sec_info->hash.data, len );
            break;
        case SEC_FEATURE_VONAGE_PASSWORD:
            if ( sec_info->password.len == 0)
            {
                printk("password not set\n");
                return -EINVAL;
            }
            len = strlen( sec_info->password.data );
            if( len > sec_info->password.len )
                printk( "password length is greater than the stored length\n" );
            memcpy( page, sec_info->password.data, len );
            break;
        case SEC_FEATURE_VONAGE_XML_KEY:
            if ( sec_info->xml_key.len == 0)
            {
                printk("xml_key not set\n");
                return -EINVAL;
            }
            len = strlen( sec_info->xml_key.data );
            if( len > sec_info->xml_key.len )
                printk( "xml key length is greater than the stored length\n" );
            memcpy( page, sec_info->xml_key.data, len );
            break;
        default:
            return -EINVAL;
    }

    if ( len <= off + count )
        *eof = 1;
    *start = page + off;
    len -= off;
    if ( len > count )
        len = count;
    if ( len < 0 )
        len = 0;

    return len;

    return len;
}

#ifndef DISABLE_SECURITY_PROCFS_WRITE
//=============================================================================
static int
sec_feature_proc_write(
        struct file     *file,
        const char      *buffer,
        unsigned long   count,
        void            *data
)
//=============================================================================
{
    factory_param_sector_t* factory_para;
    GS_FACTORYPARA * fact_para;
    struct mtd_info* mtd;
    int len;
    uint8_t* factory_buf;
    hw_feature_table_header_t *hw_table;
    hw_feature_attrib_t *hw_feature;
    hw_feature_attrib_t *existing_features[SEC_FEATURE_MAX];
    int existing_feature_count;
    int i;
    int sec_table_size, sec_table_entries;
    char *sec_table;
    int factpara_len = sizeof ( GS_FACTORYPARA );

    factory_para = (factory_param_sector_t*)fact_para_get();
    fact_para = fact_para_get();
    if ( fact_para->checksum != fact_para_checksum( fact_para ) )
    {
        printk("devinfo_proc_write: checksum fail\n");
        return -EINVAL;
    }

    //printk("sec_feature_proc_write: got %lu bytes\n", count);

    mtd = get_mtd_device_nm( "factory" );
    if ( mtd <= 0 )
    {
        printk("devinfo_proc_write: can't get mtd device\n");
        return -EFAULT;
    }

    factory_buf = vmalloc( GS_FACTORY_CFG_SIZE_FLASH );
    if ( factory_buf == NULL )
    {
        printk("devinfo_proc_write: can't allocate buffer\n");
        put_mtd_device( mtd );
        return -EFAULT;
    }

    /* See if there is a security table already */
    hw_table = &factory_para->hw_feature_table_header;
    while ( hw_table->hw_feature_table_version != HW_SECURITY_TABLE_VERSION )
    {
        //printk("factory_para: %08x hw_table: %08x version: %08x\n", factory_para, hw_table, hw_table->hw_feature_table_version);
        factpara_len += hw_table->hw_feature_table_len * sizeof (uint16_t);
        if ( hw_table->hw_feature_table_version == 0 )
        {
            /* create security table */
            hw_table->hw_feature_table_version = HW_SECURITY_TABLE_VERSION;
            hw_table->hw_feature_attrib_count = 0;

            break;
        }
        hw_table = (hw_feature_table_header_t*)( (uint16_t*)hw_table + hw_table->hw_feature_table_len );
    }

    /* locate any existing entries */
    sec_table_size = sizeof ( hw_feature_table_header_t );
    sec_table_entries = 0;
    existing_feature_count = 0;
    if ( hw_table->hw_feature_attrib_count > 0 )
    {
        //printk("found existing security table size %d entries %d\n", hw_table->hw_feature_table_len, hw_table->hw_feature_attrib_count);
        memset( existing_features, 0, sizeof ( existing_features ) );
        hw_feature = (uint8_t*)hw_table + sizeof ( hw_feature_table_header_t );
        for ( i = 0; i < hw_table->hw_feature_attrib_count; i++ )
        {
            if ( (int)data != hw_feature->hw_feature_type )
            {
                //printk("found existing feature type %d len %d\n", hw_feature->hw_feature_type, hw_feature->hw_feature_len);
                existing_features[existing_feature_count++] = hw_feature;
                sec_table_size += ( sizeof(uint16_t) * ( hw_feature->hw_feature_len - 1 ) ) + sizeof ( hw_feature_attrib_t );
                sec_table_entries++;
            }
            hw_feature = (char*)hw_feature + sizeof(uint16_t) * ( hw_feature->hw_feature_len - 1 ) + sizeof ( hw_feature_attrib_t );
        }
    }
    sec_table_size += count + 1 + sizeof ( hw_feature_attrib_t ); /*Add 1 for the last '\0' in the string*/
    sec_table_entries++;

    if ( sec_table_size & 1 )
        sec_table_size++;

    //printk("security table size: %d\n", sec_table_size);
    //printk("security table entries: %d\n", sec_table_entries);

    sec_table = vmalloc( sec_table_size );
    memset( sec_table, 0, sec_table_size );

    ((hw_feature_table_header_t*)sec_table)->hw_feature_table_version = HW_SECURITY_TABLE_VERSION;
    ((hw_feature_table_header_t*)sec_table)->hw_feature_table_len = sec_table_size / sizeof ( uint16_t );
    ((hw_feature_table_header_t*)sec_table)->hw_feature_attrib_count = sec_table_entries;

    hw_feature = (hw_feature_attrib_t*)( sec_table + sizeof ( hw_feature_table_header_t ) );
    hw_feature->hw_feature_type = (int)data;
    hw_feature->hw_feature_len = ( count + 1 )/ sizeof ( uint16_t );
    if ( ( count + 1 ) & 1 )
        hw_feature->hw_feature_len++;
    
    if ( copy_from_user( &hw_feature->hw_feature_value, buffer, count ) )
    {
        printk("security feature update: copy error\n");
        return -EFAULT;
    }

    //printk("New Feature ( type: %d, len: %d )\n", hw_feature->hw_feature_type, hw_feature->hw_feature_len);
    
    for ( i = 0; i < existing_feature_count; i++ )
    {
        hw_feature = (hw_feature_attrib_t*)( (char*)hw_feature + sizeof(uint16_t) * (hw_feature->hw_feature_len - 1 ) + sizeof ( hw_feature_attrib_t ) );
        memcpy( hw_feature, existing_features[i], ( existing_features[i]->hw_feature_len - 1 ) * sizeof ( uint16_t ) + sizeof ( hw_feature_attrib_t ) );
    }

    memcpy( hw_table, sec_table, sec_table_size );

    fact_para->length[0] = factpara_len + sec_table_size; 
    fact_para->checksum = fact_para_checksum( fact_para );
  
    //printk("length: %d\n", fact_para->length[0]);
    //printk("cksum: %08x\n", fact_para->checksum);

    mtd_read( mtd, 0, GS_FACTORY_CFG_SIZE_FLASH, &len, factory_buf );
    if ( len != GS_FACTORY_CFG_SIZE_FLASH )
    {
        printk("devinfo_proc_write: error reading factory part\n");
        put_mtd_device( mtd );
        vfree( factory_buf );
        return -EFAULT;
    }

    memcpy( factory_buf + GS_FACTORY_CFG_OFFSET, factory_para, GS_FACTORY_CFG_SIZE_FLASH );
    if ( erase_helper( mtd, 0, GS_FACTORY_CFG_SIZE_FLASH ) < 0 )
    {
        printk("devinfo_proc_write: error erasing factory part\n");
        put_mtd_device( mtd );
        vfree( factory_buf );
        return -EFAULT;
    }

    mtd_write( mtd, 0, GS_FACTORY_CFG_SIZE_FLASH, &len, factory_buf );
    if ( len != GS_FACTORY_CFG_SIZE_FLASH )
    {
        printk("devinfo_proc_write: error writing factory part\n");
        put_mtd_device( mtd );
        vfree( factory_buf );
        return -EFAULT;
    }

    set_factory_para_dirty();
    vfree( factory_buf );
    vfree( sec_table );
    put_mtd_device( mtd );

    return count;
}
#endif

//=============================================================================
static int devinfo_proc_read(
    char * page,
    char ** start,
    off_t off,
    int count,
    int * eof,
    void * data
)
//=============================================================================
{
    int len = 0;
    int checksum_valid = 1;
    char * nickstr = HW_UNKNOWN;
    char * default_mac = "00:0B:82:FF:FF:FF";
    char * pn;
    char mac_str[64];
    unsigned short int required_bits[4];
    GS_FACTORYPARA * factory_para;
    hw_info_t * hw_info;

    factory_para = fact_para_get();

    if ( factory_para->checksum != fact_para_checksum( factory_para ) )
    {
        printk("INVALID CHECKSUM\n");
        checksum_valid = 0;
    }

    hw_info = hw_feature_get();

    switch ( (int) data )
    {
        case imodel_key:
            sprintf( mac_str, "%04x", factory_para->hw_model );
            len = strlen( mac_str );
            memcpy( page, mac_str, len );
            break;
        case idev_prefix:
            printk("WARNING -- DEPRECATED PROC ENTRY PREFIX ACCESSED\n");
            strcpy( mac_str, hw_info->modelstr );
            len = strlen( mac_str );
            memcpy( page, mac_str, len );
            break;
        case irequired_bits:
            if ( checksum_valid )
            {
                required_bits[0] = factory_para->RequiredBits[0];
                required_bits[1] = factory_para->RequiredBits[1];
                required_bits[2] = factory_para->RequiredBits[2];
                required_bits[3] = factory_para->RequiredBits[3];
                /* This gets byte swapped in flash by hexpnxbin, we want all instances
                 * of required bits to look the same for convenience */
                required_bits[0] = ( ( required_bits[0] & 0x00FF ) << 8 ) | ( ( required_bits[0] & 0xFF00 ) >> 8 );
                required_bits[1] = ( ( required_bits[1] & 0x00FF ) << 8 ) | ( ( required_bits[1] & 0xFF00 ) >> 8 );
                required_bits[2] = ( ( required_bits[2] & 0x00FF ) << 8 ) | ( ( required_bits[2] & 0xFF00 ) >> 8 );
                required_bits[3] = ( ( required_bits[3] & 0x00FF ) << 8 ) | ( ( required_bits[3] & 0xFF00 ) >> 8 );
            }
            else
            {
                memset( required_bits, 0, sizeof ( required_bits ) );
            }
            sprintf( mac_str, "%04x%04x%04x%04x",
                required_bits[0],
                required_bits[1],
                required_bits[2],
                required_bits[3] );
            len = strlen( mac_str );
            memcpy( page, mac_str, len );
            break;
        case iPN:
            if ( checksum_valid )
            {
                pn = (char *)factory_para->PartNumber;
                len = strlen( pn );
                memcpy( page, pn, len );
            }
            else
            {
                len = strlen( HW_UNKNOWN );
                memcpy( page, HW_UNKNOWN, len );
            }
            break;

        case idev_alias:
            nickstr = hw_info->modelalias;
            len = strlen( nickstr );
            memcpy( page, nickstr, len );
            break;

        case idev_id:
            nickstr = hw_info->modelstr;
            len = strlen( nickstr );
            memcpy( page, nickstr, len );
            break;

        case idev_data:
            len = ( ( factory_para->length[1] << 16 ) | factory_para->length[0] );
            if ( len > sizeof ( GS_FACTORYPARA ) )
                len = sizeof ( GS_FACTORYPARA ); /* How did we pass checksum if this is true? */
            if ( checksum_valid )
                memcpy( page, factory_para, sizeof ( GS_FACTORYPARA ) );
            else
                memset( page, 0, sizeof ( GS_FACTORYPARA ) );
            break;

        case idev_mac:
            if ( checksum_valid )
            {
                sprintf( mac_str, "%02x:%02x:%02x:%02x:%02x:%02x",
                    factory_para->mac[0] & 0xFF, (factory_para->mac[0] >> 8) & 0xFF,
                    factory_para->mac[1] & 0xFF, (factory_para->mac[1] >> 8) & 0xFF,
                    factory_para->mac[2] & 0xFF, (factory_para->mac[2] >> 8) & 0xFF );
                len = strlen( mac_str );
                memcpy( page, mac_str, len );
            }
            else
            {
                len = strlen( default_mac );
                memcpy( page, default_mac, len );
            }
            break;
        case idev_rev:
            if ( checksum_valid )
            {
                pn = (char *)factory_para->PartNumber;
                sprintf(mac_str, "V%c.%c%c", pn[8], pn[9], pn[10]);
                len = strlen( mac_str );
                memcpy( page, mac_str, len );
            }
            else
            {
                len = strlen( HW_UNKNOWN );
                memcpy( page, HW_UNKNOWN, len );
            }
            break;
    }


    if ( len <= off + count )
        *eof = 1;
    *start = page + off;
    len -= off;
    if ( len > count )
        len = count;
    if ( len < 0 )
        len = 0;

    return len;
}

//=============================================================================
static int
devinfo_proc_write (
        struct file     *file,
        const char      *buffer,
        unsigned long   count,
        void            *data
)
//=============================================================================
{
    GS_FACTORYPARA * factory_para;
    unsigned short mac[3];
    struct mtd_info* mtd;
    int len;
    uint8_t* factory_buf;

    if ( count != sizeof ( mac ) )
    {
        printk("devinfo_proc_write: invalid size\n");
        return -EINVAL;
    }

    if ( copy_from_user( mac, buffer, 6 ) )
    {
        printk("devinfo_proc_write: copy error\n");
        return -EFAULT;
    }

    factory_para = fact_para_get();
    if ( factory_para->checksum != fact_para_checksum( factory_para ) )
    {
        printk("devinfo_proc_write: checksum fail\n");
        return -EINVAL;
    }

    if ( !( factory_para->mac[0] == 0x0b00 &&
            factory_para->mac[1] == 0x0082 &&
            factory_para->mac[2] == 0x0000 ) )
    {
        printk("devinfo_proc_write: mac already set\n");
        return -EFAULT;
    }

    mtd = get_mtd_device_nm( "factory" );
    if ( mtd <= 0 )
    {
        printk("devinfo_proc_write: can't get mtd device\n");
        return -EFAULT;
    }

    factory_buf = vmalloc( GS_FACTORY_CFG_SIZE_FLASH );
    if ( factory_buf == NULL )
    {
        printk("devinfo_proc_write: can't allocate buffer\n");
        put_mtd_device( mtd );
        return -EFAULT;
    }

    printk("devinfo_proc_write: setting MAC %04x:%04x:%04x\n", mac[0], mac[1], mac[2]);

    factory_para->mac[0] = mac[0];
    factory_para->mac[1] = mac[1];
    factory_para->mac[2] = mac[2];

    factory_para->checksum = fact_para_checksum( factory_para );
    
    mtd_read( mtd, 0, GS_FACTORY_CFG_SIZE_FLASH, &len, factory_buf );
    if ( len != GS_FACTORY_CFG_SIZE_FLASH )
    {
        printk("devinfo_proc_write: error reading factory part\n");
        put_mtd_device( mtd );
        vfree( factory_buf );
        return -EFAULT;
    }

    memcpy( factory_buf + GS_FACTORY_CFG_OFFSET, factory_para, GS_FACTORY_CFG_SIZE_FLASH );
    if ( erase_helper( mtd, 0, GS_FACTORY_CFG_SIZE_FLASH ) < 0 )
    {
        printk("devinfo_proc_write: error erasing factory part\n");
        put_mtd_device( mtd );
        vfree( factory_buf );
        return -EFAULT;
    }

    mtd_write( mtd, 0, GS_FACTORY_CFG_SIZE_FLASH, &len, factory_buf );
    if ( len != GS_FACTORY_CFG_SIZE_FLASH )
    {
        printk("devinfo_proc_write: error writing factory part\n");
        put_mtd_device( mtd );
        vfree( factory_buf );
        return -EFAULT;
    }

    set_factory_para_dirty();
    vfree( factory_buf );
    put_mtd_device( mtd );

    return count;
}

//=============================================================================
static int
cfg_reload_write(
        struct file     *file,
        const char      *buffer,
        unsigned long   count,
        void            *data
)
//=============================================================================
{
    // use this to trigger factory reload after writing security table by factory tool 
    set_factory_para_dirty();
    return count;
}

// For security table item certificate and private key,
// their length may exceed one page, use seq_file

//=============================================================================
static void *
private_key_seq_start(
        struct seq_file *s,
        loff_t *pos
)
//=============================================================================
{
    sec_info_t * sec_info = sec_info_get();

    if ( sec_info == NULL || 
         sec_info->private_key.len == 0 ||
         *pos > 0 )
        return NULL;
    else
    	return (void *)( sec_info->private_key.data );

}

//=============================================================================
static void *
private_key_seq_next(
        struct seq_file *s, 
        void *v, 
        loff_t *pos
)
//=============================================================================
{
	(*pos)++;
	return private_key_seq_start(s, pos);
}

//=============================================================================
static void 
private_key_seq_stop(
        struct seq_file *s, 
        void *v
)
//=============================================================================
{
	return;
}

//=============================================================================
static int private_key_seq_show(
        struct seq_file *s, 
        void *v
)
//=============================================================================
{
	seq_printf(s, "%s", (char *) v);

	return 0;
}

static struct seq_operations private_key_seq_ops={
	.start = private_key_seq_start,
	.next = private_key_seq_next,
	.stop = private_key_seq_stop,
	.show = private_key_seq_show
};

//=============================================================================
static int 
private_key_proc_open(
        struct inode *inode, 
        struct file *file
)
//=============================================================================
{
	return seq_open(file, &private_key_seq_ops);
}

static struct file_operations private_key_proc_ops={
	.owner = THIS_MODULE,
	.open = private_key_proc_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = seq_release
};

//=============================================================================
static void *
certificate_seq_start(
        struct seq_file *s,
        loff_t *pos
)
//=============================================================================
{
    sec_info_t * sec_info = sec_info_get();

    if ( sec_info == NULL || 
         sec_info->certificate.len == 0 ||
         *pos > 0 )
        return NULL;
    else
    	return (void *)( sec_info->certificate.data );

}

//=============================================================================
static void *
certificate_seq_next(
        struct seq_file *s, 
        void *v, 
        loff_t *pos
)
//=============================================================================
{
	(*pos)++;
	return certificate_seq_start(s, pos);
}

//=============================================================================
static void 
certificate_seq_stop(
        struct seq_file *s, 
        void *v
)
//=============================================================================
{
	return;
}

//=============================================================================
static int certificate_seq_show(
        struct seq_file *s, 
        void *v
)
//=============================================================================
{
	seq_printf(s, "%s", (char *) v);
	return 0;
}

static struct seq_operations certificate_seq_ops={
	.start = certificate_seq_start,
	.next = certificate_seq_next,
	.stop = certificate_seq_stop,
	.show = certificate_seq_show
};

//=============================================================================
static int 
certificate_proc_open(
        struct inode *inode, 
        struct file *file
)
//=============================================================================
{
	return seq_open(file, &certificate_seq_ops);
}

static struct file_operations certificate_proc_ops={
	.owner = THIS_MODULE,
	.open = certificate_proc_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = seq_release
};

static int __init gs_procfs_init( void )
{
    int i;
    struct proc_dir_entry * gs_procfs_root;
    struct proc_dir_entry * gs_devinfo_root;
    struct proc_dir_entry * gs_hw_feature_root;
    struct proc_dir_entry *gs_sec_feature_root;
    struct proc_dir_entry * procfs_entry;

    memset( &oem_override, 0, sizeof ( oem_override ) );

    gs_procfs_root = proc_mkdir( "gxp", NULL );
    
    procfs_entry = create_proc_entry( "oem_override", S_IFREG | S_IRUGO, gs_procfs_root );
    procfs_entry->write_proc = gs_oem_override;

    procfs_entry = create_proc_entry( "cfg_reload", S_IFREG | S_IRUGO, gs_procfs_root );
    procfs_entry->write_proc = cfg_reload_write;

    for ( i = 0; i < NUM_PROCFS_ENTRIES; i++ )
    {
        procfs_entry = create_proc_entry( gs_procfs_entry[ i ].name, S_IFREG | S_IRUGO, gs_procfs_root );
        if ( !procfs_entry )
        {
            printk("GS ProcFS creation failure.\n");
            return -1;
        }

        procfs_entry->read_proc = gs_procfs_entry[ i ].read_proc;
        procfs_entry->write_proc = *(gs_procfs_entry[ i ].write_proc);

        gs_procfs_entry[ i ].init();
    }

    gs_devinfo_root = proc_mkdir( "dev_info", gs_procfs_root );
    for ( i = 0; i < NUM_DEVINFO_ENTRIES; i++ )
    {
        procfs_entry = create_proc_entry( gs_devinfo_entry[ i ].name, S_IFREG | S_IRUGO, gs_devinfo_root );
        if ( !procfs_entry )
        {
            printk("GS ProcFS devinfo creation failure\n");
            return -1;
        }

        procfs_entry->read_proc = devinfo_proc_read;
        if ( i == idev_data )
            procfs_entry->write_proc = devinfo_proc_write;
        procfs_entry->data = (int*)i;
    }

    gs_hw_feature_root = proc_mkdir( "hw_features", gs_devinfo_root );

    for ( i = 0; i < NUM_HW_FEATURE_ENTRIES; i++ )
    {
        procfs_entry = create_proc_entry( hw_feature_entry[ i ].name, S_IFREG | S_IRUGO, gs_hw_feature_root );
        if ( !procfs_entry )
        {
            printk("GS ProcFS hw_feature creation failure\n");
            return -1;
        }

        procfs_entry->read_proc = hw_feature_proc_read;
        procfs_entry->data = (int*)( hw_feature_entry[ i ].value );
    }

    gs_sec_feature_root = proc_mkdir( "security", gs_devinfo_root );

    for ( i = 0; i < NUM_SEC_FEATURE_ENTRIES; i++ )
    {
        procfs_entry = create_proc_entry( sec_feature_entry[ i ].name, S_IFREG | S_IRUGO, gs_sec_feature_root );
        if ( !procfs_entry )
        {
            printk("gs procfs sec_feature creation failure\n");
            return -1;
        }

        if ( sec_feature_entry[ i ].value == SEC_FEATURE_PRIVATE_KEY )
        {
            procfs_entry->proc_fops = &private_key_proc_ops;
        }
        else if ( sec_feature_entry[ i ].value == SEC_FEATURE_CERTIFICATE )
        {
            procfs_entry->proc_fops = &certificate_proc_ops;
        }
        else
        {
            procfs_entry->read_proc = sec_feature_proc_read;
#ifndef DISABLE_SECURITY_PROCFS_WRITE
            procfs_entry->write_proc = sec_feature_proc_write;
#endif
            procfs_entry->data = (int*)( sec_feature_entry[ i ].value );
        }
    }

    return 0;
}

static void __exit gs_procfs_exit( void )
{

}

char* get_syslog_id(void)
{
    static int init = 0;

    if (!init) {
        GS_FACTORYPARA * factory_para;
        hw_info_t * hw_info;

        hw_info = hw_get_hw_feature();
        factory_para = fact_para_get();
        memset( syslog_id, 0, sizeof(syslog_id) );
        sprintf( syslog_id, "%s_KERNEL [%02x:%02x:%02x:%02x:%02x:%02x]", 
                    hw_info->modelalias,
                    factory_para->mac[0] & 0xFF, (factory_para->mac[0] >> 8) & 0xFF,
                    factory_para->mac[1] & 0xFF, (factory_para->mac[1] >> 8) & 0xFF,
                    factory_para->mac[2] & 0xFF, (factory_para->mac[2] >> 8) & 0xFF );
    }

    return syslog_id;
}

EXPORT_SYMBOL(get_syslog_id);
EXPORT_SYMBOL(hw_feature_get);
#ifdef CONFIG_GXP
EXPORT_SYMBOL(lcd_on_off_set);
EXPORT_SYMBOL(lcd_brightness_set);
#endif

subsys_initcall(gs_procfs_init);
module_exit(gs_procfs_exit);

#ifdef CONFIG_HT
static int __devinit reset_button_probe(struct platform_device *pdev) {
    struct device_node *np = pdev->dev.of_node;

    reset_gpio = of_get_named_gpio(np, "reset_button", 0);
    printk("reset_button: %s\n", gpio_is_valid(reset_gpio) ? "valid" : "invalid");
    if (gpio_is_valid(reset_gpio)) {
        gpio_request(reset_gpio, "grandstream-reset");
        gpio_direction_input(reset_gpio);
    }

    return 0;
}

static int __devexit reset_button_remove(struct platform_device *pdev) {
    gpio_free(reset_gpio);
    return 0;
}

static const struct of_device_id reset_button_of_match[] __devinitconst = {
	{ .compatible = "grandstream,reset", },
	{},
};
MODULE_DEVICE_TABLE(of, reset_button_of_match);

static struct platform_driver reset_button_driver = {
	.probe		= reset_button_probe,
	.remove		= __devexit_p(reset_button_remove),
	.driver		= {
		.name		= "reset-button-driver",
		.owner		= THIS_MODULE,
		.of_match_table = reset_button_of_match,
	},
};

static int __init gs_reset_button_init(void)
{
	return platform_driver_register(&reset_button_driver);
}

static void __exit gs_reset_button_exit(void)
{
	platform_driver_unregister(&reset_button_driver);
}

module_init(gs_reset_button_init);
module_exit(gs_reset_button_exit);
#endif

