/*
 *  linux/arch/arm/mach-dvf/fixups.c
 *
 *  Copyright (C) 2012 DSPG Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/of.h>
#include <linux/phy.h>
#include <linux/phy_fixed.h>
#include <linux/dmw96dma.h>
#include <linux/platform_device.h>
#include <linux/irqdomain.h>

#include <mach/irqs.h>
#include <mach/hardware.h>

static int __init dvf_dma_fixup_add_res(struct platform_device *pdev, int ch1,
					int ch2)
{
	struct resource *res;
	int num = pdev->num_resources;

	res = krealloc(pdev->resource, (num + 2) * sizeof(*res), GFP_KERNEL);
	if (!res)
		return -ENOMEM;

	res[num].start = ch1;
	res[num].end = ch1;
	res[num].flags = IORESOURCE_DMA;
	res[num + 1].start = ch2;
	res[num + 1].end = ch2;
	res[num + 1].flags = IORESOURCE_DMA;

	pdev->num_resources = num + 2;
	pdev->resource = res;

	return 0;
}

static int __init dvf_dma_fixup_device(struct device *dev, void *data)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct resource *res;
	int ret = 0;

	/* get base address */
	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (res == NULL)
		return 0;

	switch (res->start) {
	case DVF_TDM1_BASE:
		ret = dvf_dma_fixup_add_res(pdev, 8, 9);
		break;

	case DVF_TDM2_BASE:
		ret = dvf_dma_fixup_add_res(pdev, 10, 11);
		break;

	case DVF_SPI1_BASE:
		ret = dvf_dma_fixup_add_res(pdev, 12, 13);
		break;
	}

	return ret;
}

static int __init dvf_dma_fixup(void)
{
	return bus_for_each_dev(&platform_bus_type, NULL, NULL,
			dvf_dma_fixup_device);
}
arch_initcall(dvf_dma_fixup);


static struct dmw96dma_ch_config gdmac_configs[] = {
	/* ch08: TDM1 RX FIFO */
	[8] = {
		.direction       = 0, /* A -> B */
		.hw_port         = 0, /* Peripheral on Port A */
		.transfer_length = 2*16 -1, /* 16 audio frames (stereo samples) */
		.width           = DMW96DMA_WIDTH_32,
		.mode            = DMW96DMA_MODE_HW_SINGLE,
		.base            = DVF_TDM1_BASE + 0x3c,
	},
	/* ch09: TDM1 TX FIFO */
	[9] = {
		.direction       = 1, /* B -> A */
		.hw_port         = 0, /* Peripheral on Port A */
		.transfer_length = 2*16 -1, /* 16 audio frames (stereo samples) */
		.width           = DMW96DMA_WIDTH_32,
		.mode            = DMW96DMA_MODE_HW_SINGLE,
		.base            = DVF_TDM1_BASE + 0x38,
	},
	/* ch10: TDM2 RX FIFO */
	[10] = {
		.direction       = 0, /* A -> B */
		.hw_port         = 0, /* Peripheral on Port A */
		.transfer_length = 2*16 -1, /* 16 audio frames (stereo samples) */
		.width           = DMW96DMA_WIDTH_32,
		.mode            = DMW96DMA_MODE_HW_SINGLE,
		.base            = DVF_TDM2_BASE + 0x3c,
	},
	/* ch11: TDM2 TX FIFO */
	[11] = {
		.direction       = 1, /* B -> A */
		.hw_port         = 0, /* Peripheral on Port A */
		.transfer_length = 2*16 -1, /* 16 audio frames (stereo samples) */
		.width           = DMW96DMA_WIDTH_32,
		.mode            = DMW96DMA_MODE_HW_SINGLE,
		.base            = DVF_TDM2_BASE + 0x38,
	},
	/* ch12: SPI1 RX FIFO */
	[12] = {
		.direction       = 0, /* A -> B */
		.hw_port         = 0, /* Peripheral on Port A */
		.transfer_length = 4 -1, /* 16 bytes or 4 words */
		.width           = DMW96DMA_WIDTH_8,
		.mode            = DMW96DMA_MODE_HW_SINGLE,
		.base            = DVF_SPI1_BASE + 0x1C,
	},
	/* ch13: SPI1 TX FIFO */
	[13] = {
		.direction       = 1, /* B -> A */
		.hw_port         = 0, /* Peripheral on Port A */
		.transfer_length = 4 -1, /* n+1: 16 bytes or 4 words */
		.width           = DMW96DMA_WIDTH_8,
		.mode            = DMW96DMA_MODE_HW_SINGLE,
		.base            = DVF_SPI1_BASE + 0x18,
	},
};

/*
 * DMA channel priorities. Valid values are from 0 to 31 where 0 has the
 * highest priority. Every priority must be used only once, including the ones
 * given below for rotate and memcpy!
 */
u8 gdmac_priorities[ARRAY_SIZE(gdmac_configs)] = {
	[8]  =  0, /* ch08: TDM1 RX FIFO */
	[9]  =  1, /* ch09: TDM1 TX FIFO */
	[10] =  2, /* ch10: TDM2 RX FIFO */
	[11] =  3, /* ch11: TDM2 TX FIFO */
	[2]  =  4, /* ch02: UART1 RX FIFO */
	[4]  =  5, /* ch04: UART2 RX FIFO */
	[3]  =  6, /* ch03: UART1 TX FIFO */
	[5]  =  7, /* ch05: UART2 TX FIFO */
	[0]  =  8, /* ch00: QSPI */
	[12] =  9, /* ch12: SPI1 RX FIFO */
	[13] = 10, /* ch13: SPI1 TX FIFO */

	[1]  = -1, /* ch01: reserved */
	[6]  = -1, /* ch06: reserved */
	[7]  = -1, /* ch07: reserved */
};

static struct dmw96dma_platform_data gdmac_pdata = {
	.hw_version	 = 1, /* DVF99 style, no rotation, 16ch */
	.timeout         = 0x0FFF,
	.desc_port       = 1,
	.num_channels    = ARRAY_SIZE(gdmac_configs),
	.ch_configs      = gdmac_configs,
	.ch_priorities   = gdmac_priorities,
	.memcpy_priority = 14, /* uses priorities 14 and 15! */
	.memcpy_base     = DVF_SRAM_GDMAC_BASE,
};

static struct resource gdmac_resources[] = {
	{
		/* DVF_IID_GDMAC_DMA_DONE */
		.flags	= IORESOURCE_IRQ,
	},
	{
		/* DVF_IID_GDMAC_ERROR */
		.flags	= IORESOURCE_IRQ,
	},
	{
		.start	= DVF_GDMAC_BASE,
		.end	= DVF_GDMAC_BASE + 0x200 - 1,
		.flags	= IORESOURCE_MEM,
	},
};

static struct platform_device gdmac_device = {
	.name		= "dmw96dma",
	.id		= -1,
	.dev		= {
		.coherent_dma_mask = ~0,
		.platform_data     = &gdmac_pdata,
	},
	.num_resources	= ARRAY_SIZE(gdmac_resources),
	.resource	= gdmac_resources,
};

static int __init of_add_fixed_phys(void)
{
	int ret;
	struct device_node *np;
	u32 *fixed_link;
	struct fixed_phy_status status = {};

	for_each_node_by_name(np, "ethernet") {
		fixed_link  = (u32 *)of_get_property(np, "fixed-link", NULL);
		if (!fixed_link)
			continue;

		status.link = 1;
		status.duplex = be32_to_cpu(fixed_link[1]);
		status.speed = be32_to_cpu(fixed_link[2]);
		status.pause = be32_to_cpu(fixed_link[3]);
		status.asym_pause = be32_to_cpu(fixed_link[4]);

		ret = fixed_phy_add(PHY_POLL, be32_to_cpu(fixed_link[0]), &status);
		if (ret) {
			of_node_put(np);
			return ret;
		}
	}

	/* map GDMAC IRQs */
	gdmac_resources[0].start = irq_create_mapping(NULL, DVF_IID_GDMAC_DMA_DONE);
	gdmac_resources[0].end = gdmac_resources[0].start;
	gdmac_resources[1].start = irq_create_mapping(NULL, DVF_IID_GDMAC_ERROR);
	gdmac_resources[1].end = gdmac_resources[1].start;
	if (!gdmac_resources[0].start || !gdmac_resources[1].start) {
		printk(KERN_ERR "could not map GDMAC IRQs\n");
		return -ENODEV;
	}

	return platform_device_register(&gdmac_device);
}
arch_initcall(of_add_fixed_phys);
