/*
 *  linux/arch/arm/mach-dvf/pm-sleep.S
 *
 *  Copyright (C) 2013 DSP Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

#include <linux/linkage.h>
#include <mach/hardware.h>
#include <mach/platform.h>
#include "pm.h"

#define lowpower_cmd_reg	(21*4)
#define lowpower_cmd_msk	(0xff << 24)
#define lowpower_cmd_off	24
#define lowpower_state_reg	(22*4)

#define controller_busy_reg	(41*4)
#define CONTROLLER_BUSY		(0x1 << 16)

#define param_user_def_reg0	(106*4)
#define comp_disable_drv	0x1
#define comp_disable_rtt	0x2

#define DVF_CMU_PLL_PD		(1 << 15)
#define DVF_CMU_PLL_BP		(1 << 16)
#define DVF_CMU_PLL_LD		(1 << 18)

.globl dvf_sleep_selfrefresh_end
	.text
/*
 * dvf_sleep_selfrefresh(u32 mode, u32 denali_base, u32 cmu_base)
 *
 * Puts the RAM into self refresh.
 *
 * The whole function runs out of internal memory with locked TLB entries
 * because no memory access is possible after the RAM is in self refresh mode.
 *
 * Must be called with IRQs (and, if used, FIQs) disabled!
 *
 * r0: mode
 * r1: DENALI_BASE
 * r2: CMU_BASE
 * r3, r4, r5: scratch registers
 */
ENTRY(dvf_sleep_selfrefresh)
	push	{r4-r6}

	mov	r4, #0
	mrc	p15, 0, r3, c1, c0, 0		@ Read control register
	mcr	p15, 0, r4, c7, c10, 4		@ Drain write buffer
	bic	r4, r3, #(1 << 12)
	mcr	p15, 0, r4, c1, c0, 0		@ Disable I cache

	/*
	 * Put the DDR controller in low power mode. 
	 */
1:
	ldr	r4, [r1, #controller_busy_reg]
	tst	r4, #CONTROLLER_BUSY
	bne	1b

	ldr	r4, [r1, #param_user_def_reg0]	@ disable RTT and DRV
	orr	r4, r4,  #(comp_disable_drv | comp_disable_rtt)
	str	r4, [r1, #param_user_def_reg0]

	tst	r0, #SLEEP_DDR_LP_MODE
	ldrne	r4, [r1, #lowpower_cmd_reg]
	bicne	r4, r4,  #lowpower_cmd_msk
	orrne	r4, r4, r0, lsl #lowpower_cmd_off
	orrne	r4, r4,  #DDR_MODE_ENTER	@ enter low-power DDR mode
	strne	r4, [r1, #lowpower_cmd_reg]

1:
	ldr	r4, [r1, #lowpower_state_reg]
	tst	r4, #DDR_MODE_LOCKED
	beq	1b

	tst	r0, #SLEEP_DDR_PLL_OFF
	ldrne	r4, [r2, #DVF_CMU_PLL2CONTROL]	@ power down and bypass DDR-PLL
	orrne	r5, r4,  #(DVF_CMU_PLL_PD | DVF_CMU_PLL_BP)
	strne	r5, [r2, #DVF_CMU_PLL2CONTROL]
	ldrne	r6, [r2, #DVF_CMU_SWCLKENR2]
	bicne	r5, r6, #(1<<12)
	strne	r5, [r2, #DVF_CMU_SWCLKENR2]

	/*
	 * Zzzzzzz
	 */

	mcr	p15, 0, r5, c7, c0, 4		@ Wait for interrupt
	mcr	p15, 0, r3, c1, c0, 0		@ Restore ICache enable

	strne	r6, [r2, #DVF_CMU_SWCLKENR2]
	strne	r4, [r2, #DVF_CMU_PLL2CONTROL]	@ restore DDR-PLL
1:
	ldr	r4, [r2, #DVF_CMU_PLL2CONTROL]
	tst	r4, #DVF_CMU_PLL_LD
	beq	1b

	tst	r0, #SLEEP_DDR_LP_MODE
	ldrne	r4, [r1, #lowpower_cmd_reg]
	bicne	r4, r4,  #lowpower_cmd_msk
	orrne	r4, r4,  #DDR_MODE_LEAVE	@ leave low-power DDR mode
	strne	r4, [r1, #lowpower_cmd_reg]

1:
	ldr	r4, [r1, #lowpower_state_reg]
	tst	r4, #DDR_MODE_LOCKED
	beq	1b

	ldr	r4, [r1, #param_user_def_reg0]	@ enable RTT and DRV
	bic	r4, r4,  #(comp_disable_drv | comp_disable_rtt)
	str	r4, [r1, #param_user_def_reg0]

	/*
	 * Fill one cache line (16 instructions) to avoid fetching instructions
	 * when the DRAM is still off.
	 */
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop

	pop	{r4-r6}
	mov	pc, lr				@ done
dvf_sleep_selfrefresh_end:
ENDPROC(dvf_sleep_selfrefresh)
