/*
 *  linux/arch/arm/mach-dvf/pm.c
 *
 *  Copyright (C) 2013 DSP Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

#include <linux/clk.h>
#include <linux/clkdev.h>
#include <linux/err.h>
#include <linux/of.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/io.h>
#include <linux/pm.h>
#include <linux/cpuidle.h>
#include <linux/suspend.h>
#include <mach/hardware.h>
#include <asm/system_misc.h>
#include <asm/hardware/plicu.h>
#include "pm.h"

#define STATE_COUNT	3

#define cmu_readl(_reg)		readl(cmu_base + (_reg))
#define cmu_writel(_val, _reg)	writel(_val, cmu_base + (_reg))

static void *denali_base, *intmem_base, *cmu_base;
static struct clk *altsysclk, *sysclk;
static struct cpuidle_device dvf_idle_dev;
static int ddr_type = 2;

static void (*sleep_selfrefresh)(int mode, unsigned long denali_base, unsigned long cmu_base);

static int
dvf_sleep(int mode)
{
	struct clk *parent = NULL;
	int active_io, active_bm;
	unsigned long swclkenr1 = 0;

	if (ddr_type != 2)
		mode &= ~(SLEEP_DDR_LP_MODE | SLEEP_DDR_PLL_OFF);

	/* Stay here until an IRQ is pending */
	while (!plicu_pending()) {
		active_io = sysclk->iocnt > 0;
		active_bm = atomic_read(&dvf_active_bm_cnt) > 0;

		if ((mode & SLEEP_CPU_PLL_OFF) && !active_io && !active_bm) {
			/* Switch sysclk to slow clock if possible */
			parent = clk_get_parent(sysclk);
			clk_set_parent(sysclk, altsysclk);

			/* disable UARTs */
			swclkenr1 = cmu_readl(DVF_CMU_SWCLKENR1);
			cmu_writel(swclkenr1 & ~(0x3 << 6), DVF_CMU_SWCLKENR1);
		}

		sleep_selfrefresh(mode, (unsigned long)denali_base, (unsigned long)cmu_base);

		if ((mode & SLEEP_CPU_PLL_OFF) && !active_io && !active_bm) {
			/* enable UARTs */
			cmu_writel(swclkenr1, DVF_CMU_SWCLKENR1);

			/* Switch back to fast clock */
			clk_set_parent(sysclk, parent);
		}
	}

	return 0;
}

static int
dvf_idle_enter(struct cpuidle_device *dev, struct cpuidle_driver *drv, int index)
{
	ktime_t kt_preidle, kt_postidle;
	s64 idle_time;
	struct cpuidle_state_usage *state_usage;
	unsigned long flags;

	/* Fall back to safe idle mode if DRAM access is required */
	if (atomic_read(&dvf_active_bm_cnt) > 0)
		index = 0;

	state_usage = &dev->states_usage[index];

	flags = (unsigned long)cpuidle_get_statedata(state_usage);

	/* Used to keep track of the total time in idle */
	kt_preidle = ktime_get_real();

	dvf_sleep(flags);

	/* calculate sleep time */
	kt_postidle = ktime_get_real();

	idle_time = ktime_to_us(ktime_sub(kt_postidle, kt_preidle));

	/* Update device last_residency */
	dev->last_residency = (int)idle_time;

	local_irq_enable(); /* idle handler must re-enable IRQs on exit */

	return index;
}

static unsigned long state_flags[STATE_COUNT] = {
	WFI,
	DDR_MODE_SELF_REFRESH,
	DDR_MODE_SELF_REFRESH_GATING /* | SLEEP_CPU_PLL_OFF */
};

static struct cpuidle_driver dvf_idle_driver = {
	.name  = "dvf_idle",
	.owner = THIS_MODULE,
	.states = {
		{
			.name = "C0",
			.desc = "wfi",
			.flags = CPUIDLE_FLAG_TIME_VALID,
			.exit_latency = 0, /* in US */
			.power_usage = 200, /* in mW */
			.target_residency = 0, /* in US */
			.enter = dvf_idle_enter,
		},
		{
			.name = "C1",
			.desc = "self-refresh",
			.flags = CPUIDLE_FLAG_TIME_VALID,
			.exit_latency = 10, /* in US */
			.power_usage = 150, /* in mW */
			.target_residency = 20, /* in US */
			.enter = dvf_idle_enter,
		},
		{
			.name = "C2",
			.desc = "self-refresh, PLLs off",
			.flags = CPUIDLE_FLAG_TIME_VALID,
			.exit_latency = 200, /* in US */
			.power_usage = 50, /* in mW */
			.target_residency = 400, /* in US */
			.enter = dvf_idle_enter,
		},
	},
	.state_count = STATE_COUNT,
};

static int dvf_pm_prepare(void)
{
	disable_hlt();

	return 0;
}

static int dvf_pm_prepare_late(void)
{
	if (sysclk->iocnt > 0 || atomic_read(&dvf_active_bm_cnt)) {
		printk(KERN_WARNING "%d active IO devices on SYSCLK and"
		       " %d active bus masters remaining!\n",
		       sysclk->iocnt, atomic_read(&dvf_active_bm_cnt));
		dvf_clk_print_active();
	}

	return 0;
}

static int dvf_pm_enter(suspend_state_t state)
{
	int ret = 0;

	switch (state) {
	case PM_SUSPEND_STANDBY:
	case PM_SUSPEND_MEM:
		ret = dvf_sleep(DDR_MODE_SELF_REFRESH_GATING | SLEEP_CPU_PLL_OFF);
		break;
	default:
		ret = -EINVAL;
	}

	return ret;
}

static void dvf_pm_finish(void)
{
	enable_hlt();
}

static struct platform_suspend_ops dvf_pm_ops = {
	.prepare	= dvf_pm_prepare,
	.prepare_late	= dvf_pm_prepare_late,
	.enter		= dvf_pm_enter,
	.finish		= dvf_pm_finish,
	.valid		= suspend_valid_only_mem,
};

static int __init dvf_pm_probe(struct platform_device *pdev)
{
	int i, ret = 0;
	struct cpuidle_state_usage *state_usage;
	unsigned long tmp = 0;

	denali_base = ioremap_nocache(DVF_DDRC_BASE, SZ_4K);
	if (!denali_base) {
		dev_err(&pdev->dev, "could not map Denali controller!\n");
		return -ENOMEM;
	}

	/* check if Denali is active and configured for DDR2 */
	if ((virt_to_phys(dvf_pm_probe) >= (DVF_DRAM1_BASE + DVF_DRAM_SIZE)) ||
	    ((readl(denali_base) & 0xf00) != 0x400))
		ddr_type = 3;

	intmem_base = ioremap_nocache(DVF_SRAM_BASE, 8*SZ_4K);
	if (!intmem_base) {
		dev_err(&pdev->dev, "could not map internal memory!\n");
		ret = -ENOMEM;
		goto free_denali;
	}

	memcpy(intmem_base,   dvf_sleep_selfrefresh,
	       (unsigned long)dvf_sleep_selfrefresh_end -
	       (unsigned long)dvf_sleep_selfrefresh);
	sleep_selfrefresh = intmem_base;

	cmu_base = ioremap_nocache(DVF_CMU_BASE, 8*SZ_4K);
	if (!cmu_base) {
		dev_err(&pdev->dev, "could not map CMU memory!\n");
		ret = -ENOMEM;
		goto free_intmem;
	}

	sysclk = clk_get(&pdev->dev, "sysclk");
	if (IS_ERR(sysclk)) {
		dev_err(&pdev->dev, "sysclk missing!\n");
		ret = PTR_ERR(sysclk);
		goto free_cmu;
	}

	altsysclk = clk_get(&pdev->dev, "altsysclk");
	if (IS_ERR(altsysclk)) {
		dev_err(&pdev->dev, "altsysclk missing!\n");
		ret = PTR_ERR(altsysclk);
		goto put_sysclk_clk;
	}

	/* load addresses used by assembly code into TLB lockdown area */
	asm volatile (
	"mcr	p15, 0, %0, c8, c7, 1\n"	/* invalidate single TLB entry to ensure */
	"mcr	p15, 0, %1, c8, c7, 1\n"	/* that the address is not already in the */
	"mcr	p15, 0, %2, c8, c7, 1\n"	/* TLB */
	"mrc	p15, 0, %3, c10, c0, 0\n"	/* read the lockdown register */
	"orr	%3, %3, #1\n"			/* set preserve bit */
	"mcr	p15, 0, %3, c10, c0, 0\n"	/* write the lockdown register */
	"ldr	%3, [%0]\n"			/* issue TLB miss on desired address */
	"ldr	%3, [%1]\n"			/* issue TLB miss on desired address */
	"ldr	%3, [%2]\n"			/* issue TLB miss on desired address */
	"mrc	p15, 0, %3, c10, c0, 0\n"	/* read the lockdown register */
	"bic	%3, %3, #1\n"			/* clear preserve bit */
	"mcr	p15, 0, %3, c10, c0, 0\n"	/* write the lockdown register */
	: : "r" (intmem_base), "r" (denali_base), "r" (cmu_base), "r" (tmp));

	suspend_set_ops(&dvf_pm_ops);

	for (i = 0; i < STATE_COUNT; i++) {
		state_usage = &dvf_idle_dev.states_usage[i];

		cpuidle_set_statedata(state_usage, (void *)state_flags[i]);
	}
	dvf_idle_dev.state_count = STATE_COUNT;

	ret = cpuidle_register_driver(&dvf_idle_driver);
	if (!ret)
		ret = cpuidle_register_device(&dvf_idle_dev);

	if (ret)
		dev_warn(&pdev->dev, "idle driver not registered: %d\n", ret);

	return 0;

put_sysclk_clk:
	clk_put(sysclk);
free_cmu:
	iounmap(cmu_base);
free_intmem:
	iounmap(intmem_base);
free_denali:
	iounmap(denali_base);

	return ret;
}

static const struct of_device_id dvf_pm_ids[] = {
	{ .compatible = "dspg,pm", },
	{ /* Sentinel */ },
};

static struct platform_driver dvf_pm_plat_driver = {
	.driver = {
		.name	 = "dvf-pm",
		.of_match_table	= dvf_pm_ids,
		.owner	 = THIS_MODULE,
	},
};

static int __init dvf_pm_init(void)
{
	return platform_driver_probe(&dvf_pm_plat_driver, dvf_pm_probe);
}
late_initcall(dvf_pm_init);
