/****************************************************************************
 *
 * FILENAME:        $RCSfile: provision.c
 *
 * LAST REVISION:   $Revision: 1.0 
 * LAST MODIFIED:   $Date: 2014/09/15
 *
 * DESCRIPTION:     provision support
 *
 * vi: set ts=4:
 *
 * Copyright (c) 2013 by Grandstream Networks, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***************************************************************************/

//=======================
//  Includes
//=======================
#include <linux/types.h>
#include <linux/init.h>
#include <linux/vmalloc.h>
#include <linux/proc_fs.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/partitions.h>
#include <mach/factory.h>
#include <provision.h>

//=======================
//  Defines
//=======================
//must be updated for every new model inthe table

//=======================
//  Globals
//=======================

//=======================
//  Typedefs
//=======================

// reference from drivers/mtd/mtdpart.c
struct mtd_part {
	struct mtd_info mtd;
	struct mtd_info *master;
	uint64_t offset;
	struct list_head list;
};
#define PART(x)  ((struct mtd_part *)(x))

//=======================
//  Locals
//=======================

    /*
     * Private Functions
     */

    /*
     * Private Data
     */
static char * provision_part_name[] = 
{
#if defined( CONFIG_HT )
    "uboot",
    "uboot2",
    "core",
    "core2",
    "base",
    "base2",
    "prog",
    "prog2",
    "lang",
    "lang2",
    "oem"
#elif defined( CONFIG_GXP )
    "boot",
    "core",
    "base",
    "prog",
    "oem",
    "lcl"
#else
#error NO PRODUCT DEFINED
#endif
};

#define NUM_PROV_ENTRIES   sizeof( provision_part_name ) / sizeof( char * )

#if defined( CONFIG_HT )
static int provision_parts_group = 0;
#endif

//=======================
//  Functions
//=======================

#if defined( CONFIG_HT )
//=============================================================================
static int __init prov_parts_group_setup(
    char *str
)
//=============================================================================
{
    if ( str[0] == '1' )
        provision_parts_group = 0;
    else
        provision_parts_group = 1;

    return 1;
}
__setup("bootpartition=", prov_parts_group_setup);
#endif

//=============================================================================
int
gxp_internal_mtd_header(
	int part
)
//=============================================================================
{
    struct mtd_info *mtd;
    struct mtd_part *partition;

    if ( part >= NUM_PROV_ENTRIES )
    {
        return -ENODEV;
    }
    
    mtd = get_mtd_device_nm( provision_part_name[part] );
    if ( mtd <= 0 )
    {
        return -ENODEV;
    }

    partition = PART(mtd);
#if defined( CONFIG_HT )
    return (int)partition->offset;
#elif defined( CONFIG_GXP )
    // GXP boot image using tail header 
    if ( strcmp( provision_part_name[part], "boot" ) == 0 )
        return (int)(partition->offset + mtd->size - HEADER_BLOCK_SIZE);
    else
        return (int)partition->offset;
#else
#error NO PRODUCT DEFINED
#endif
}


//=============================================================================
int
gxp_internal_mtd_read(
        u_int32_t addr,
        u_int32_t len,
        char *buf_ptr
)
//=============================================================================
{
    size_t retlen;
    int ret;

    if ( gxp_mtd == NULL )
    {
        printk("GXP MTD NOT CONFIGURED\n");
        return 1;
    }

    ret = (*(gxp_mtd->_read))(gxp_mtd, addr, len, &retlen, (unsigned char *)buf_ptr);
    if (ret || retlen != len)
    {
        printk ("mtd read failed.\n");
        return 1;
    }
    return 0;
}

//=============================================================================
void prov_update_partitions(
    struct mtd_partition *parts
)
//=============================================================================
{
    int i;

    parts->hdr_offset = 0;
    for ( i = 0; i < NUM_PROV_ENTRIES; i ++ )
    {
        if ( strcmp( parts->name, provision_part_name[i] ) == 0 )
        {
            parts->hdr_offset = HEADER_BLOCK_SIZE;
            break;
        }
    }
}

#if defined( CONFIG_HT )
//=============================================================================
static int
prov_part_grp_read_proc(
    char * page,
    char ** start,
    off_t off,
    int count,
    int * eof,
    void * data
)
//=============================================================================
{
    int len = 1;

    if ( provision_parts_group )
        page[0] = '1';
    else
        page[0] = '0';

    if ( len <= off + count )
        *eof = 1;
    *start = page + off;
    len -= off;
    if ( len > count )
        len = count;
    if ( len < 0 )
        len = 0;

    return (len);
}
#endif

//=============================================================================
static int
prov_read_proc(
    char * page,
    char ** start,
    off_t off,
    int count,
    int * eof,
    void * data
)
//=============================================================================
{
    int partition = (int)data;
    int len = 0;
    unsigned int addr;
    int ret =  0;
    pImageHeader p = (pImageHeader)page;

    len = sizeof ( ImageHeader );
    addr = gxp_internal_mtd_header( partition );
    if ( addr < 0 )
    {
        printk("Invalid Parititon %d\n", partition );
        memset( p, 0, len );
    }
    else 
    {
        ret = gxp_internal_mtd_read( addr, len, (char*)p );

        if ( p->Signature != GS_PROD_SIGNATURE || ret != 0 )
        {
            printk("Invalid Signature for Parititon %s\n", provision_part_name[partition] );
            memset( p, 0, len );
        }
    }

    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

//=============================================================================
static int
prov_write_proc(
    struct file * file,
    const char * buffer,
    unsigned long count,
    void * data
)
//=============================================================================
{
    return -EINVAL;
}

//=============================================================================
static int __init gs_procfs_init(
    void
)
//=============================================================================
{
    int i;
    struct proc_dir_entry * gs_provfs_root;
    struct proc_dir_entry * procfs_entry;

    gs_provfs_root = proc_mkdir( "provision", NULL );

#if defined( CONFIG_HT )
    procfs_entry = create_proc_entry( "provision_group", S_IFREG | S_IRUGO, gs_provfs_root );
    if ( !procfs_entry )
    {
        printk("GS ProcFS prov creation failure\n");
        return -1;
    }
    procfs_entry->read_proc = prov_part_grp_read_proc;
    procfs_entry->write_proc = NULL;
    procfs_entry->data = (int*)( 0 );
#endif

    gs_provfs_root = proc_mkdir( "partition", gs_provfs_root );
    for ( i = 0; i < NUM_PROV_ENTRIES; i++ )
    {
        procfs_entry = create_proc_entry( provision_part_name[ i ], S_IFREG | S_IRUGO, gs_provfs_root );
        if ( !procfs_entry )
        {
            printk("GS ProcFS prov creation failure\n");
            return -1;
        }

        procfs_entry->read_proc = prov_read_proc;
        procfs_entry->write_proc = prov_write_proc;
        procfs_entry->data = (int*)( i ); /* Partition Number */
    }

    return 0;
}

//=============================================================================
static void __exit gs_procfs_exit(
    void
)
//=============================================================================
{

}

subsys_initcall(gs_procfs_init);
module_exit(gs_procfs_exit);


/* EOF */
