/*
 * si3217x.c  --  Silicon Labs si3217x ALSA SoC Audio Codec driver
 *
 * Copyright (C) 2011 DSPG (DSP Group)
 *
 * Based on wm8978.c:
 *  Copyright (C) 2009-2010 Guennadi Liakhovetski <g.liakhovetski@gmx.de>
 *  Copyright (C) 2007 Carlos Munoz <carlos@kenati.com>
 *  Copyright 2006-2009 Wolfson Microelectronics PLC.
 *  Based on wm8974 and wm8990 by Liam Girdwood <lrg@slimlogic.co.uk>
 *
 * Adapted by Murali TD. Mohan <Murali.Mohan@dspg.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

//#define V_DEBUG
//#define DEBUG

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/pm.h>
#include <linux/spi/spi.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>
#include <sound/initval.h>
#include <sound/tlv.h>
#include <asm/div64.h>
#include <linux/gpio.h>
#include <linux/ioctl.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <linux/err.h>
#include <linux/list.h>
#include <linux/errno.h>
#include <linux/mutex.h>
#include <linux/of_gpio.h>

#include <linux/spi/spi.h>
#include <linux/spi/spidev.h>
#include <sound/si32178.h>
#include <asm/uaccess.h>

void si3217x_reset(unsigned int gpio)
{
	gpio_direction_output(gpio, 1);
	udelay(100);
	gpio_set_value(gpio, 0);
	udelay(1000);
	gpio_set_value(gpio, 1);
	udelay(1000);
}

/* Must keep this private pointer in the local scope of the file so 
   get_si3217x_priv() can correctly return it to the bsp driver of the slic  
*/
static struct si3217x_priv * ptr_si3217x[NUM_SPI_BANKS] = { 0, 0};

static int si3217x_spi_probe(struct spi_device *spi)
{
	struct si3217x_priv * si3217x = NULL;
	struct device_node *np;
	int reset_gpio, ret;
	unsigned int id;

	np = spi->dev.of_node;
	if (np == NULL)
		dev_info(&spi->dev, "of node is null\n");

	/* Get the reset line associated wtih the devices on this SPI bank */
	if ((reset_gpio = of_get_named_gpio(np, "rst-gpio", 0)) < 0) {
		dev_err(&spi->dev, "Reset value not specified in devtree\n");
		return -ENODEV;
	}

	/* Read the SPI id and verify it before moving on */
	ret = of_property_read_u32(np, "id", &id);
	if (ret && ret != -EINVAL) {
		dev_err(&spi->dev, "Error parsing id value in devtree\n");
		return -ENODEV;
	}
	if ( id >= NUM_SPI_BANKS ) {
		printk("Driver support only %d SPI banks. Must increase NUM_SPI_BANKS macro\n", NUM_SPI_BANKS);
		return -ENODEV;
	}

	si3217x = kzalloc(sizeof(struct si3217x_priv), GFP_KERNEL);
	if (si3217x == NULL)
		return -ENOMEM;

	si3217x->si3217x_reset = reset_gpio;
	si3217x->si3217x_id = id;

	spi_set_drvdata(spi, si3217x);
	printk("spi_dev[%d] = 0x%8p\n", id, spi);
	ptr_si3217x[id] = si3217x;
	printk("ptr_si3217x[%d] = 0x%8p\n", id, ptr_si3217x[id]);
	/*
	* setup spi parameters; this makes sure that parameters we request
	* are acceptable by the spi driver
	*/
	spi->mode = SPI_MODE_3; /* clk active high */
	spi->bits_per_word = 8;

	ret = spi_setup(spi);
	if (ret < 0) {
		dev_err(&spi->dev, "spi_setup() failed\n");
		goto err_si3217x;
	}

	si3217x->spi = spi;

	ret = gpio_request(si3217x->si3217x_reset, "si3217x reset");
	if (ret < 0) {
		dev_err(&spi->dev, "failed requesting reset gpio\n");
		goto err_si3217x;
	}

	/* Resetting the SPI devices will be done by the low level drivers  */
    //	si3217x_reset(si3217x->si3217x_reset);

	dev_set_name(&spi->dev, "%s.%u", "si3217x-daa", si3217x->si3217x_id);
	dev_info(&spi->dev, "successfully registered\n");

	return ret;

err_si3217x:
	kfree(si3217x);
	return ret;
}

/* Helper function to allow low level BSP driver finish all the setups */
struct si3217x_priv * get_si3217x_priv(int spi_dev_id)
{
	printk("calling get_si3217x_priv reutrn spi_dev[%d] = 0x%8p\n", spi_dev_id, ptr_si3217x[spi_dev_id]);

	return ptr_si3217x[spi_dev_id];
}
EXPORT_SYMBOL(get_si3217x_priv);

static int __devexit si3217x_spi_remove(struct spi_device *spi)
{
	struct si3217x_priv *si3217x = spi_get_drvdata(spi);

	kfree(si3217x);
	return 0;
}

static const struct of_device_id dspg_si32178_dt_match[] = {
	{ .compatible = "dspg,dvf_si32178_spi", },
	{}
};

static struct spi_driver dvf_si32178_spi_driver = {
	.driver = {
		.name	= "dvf_si32178_spi",
		.bus	= &spi_bus_type,
		.owner	= THIS_MODULE,
		.of_match_table = dspg_si32178_dt_match,
	},
	.probe	= si3217x_spi_probe,
	.remove	= __exit_p(si3217x_spi_remove),
};

static int __init si3217x_module_init(void)
{
	return spi_register_driver(&dvf_si32178_spi_driver);
}

static __devexit void si3217x_module_exit(void)
{
	spi_unregister_driver(&dvf_si32178_spi_driver);
}

module_init(si3217x_module_init);
module_exit(si3217x_module_exit);

MODULE_DESCRIPTION("ASoC si3217x D.A.A. codec driver");
MODULE_AUTHOR("DSP Group");
MODULE_LICENSE("GPL");
