/****************************************************************************
 *
 * FILENAME:        $RCSfile: gxp_ext_host.c,v $
 *
 * LAST REVISION:   $Revision: 1.2 $
 * LAST MODIFIED:   $Date: 2013/08/01 19:19:04 $
 *
 * DESCRIPTION:     EXT Support
 *
 * vi: set ts=4:
 *
 * Copyright (c) 2005 by Grandstream Networks, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***************************************************************************/

//=======================
//  Includes
//=======================
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>
#include <linux/kfifo.h>
#include <linux/kthread.h>
#include <linux/cdev.h>
#include <linux/dma-mapping.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/irq.h>
#include <linux/vmalloc.h>
#include <linux/workqueue.h>
#include <linux/cdev.h>
#include <linux/wait.h>
#include <linux/i2c.h>
#include <linux/fs.h>
#include <linux/list.h>
#include <linux/crc32.h>
#include <linux/of_gpio.h>

#include <mach/gpio.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include "gxp_ext.h"

//=======================
//  Defines
//=======================
/* Note: power of two break something in the lower level that cause all transmits to timeout */
#define I2C_MTU_HOST 2000
#define I2C_MTU_EXT 250

// Time period for pulsar task
#define PULSAR_PERIOD 1

// Time period in milliseconds to poll I2C message receiving.
#define I2C_RECEIVE_POLL_PERIOD	10

// Buffer pool size
#define I2C_QUEUE_SIZE 50

/* BOARD_DISCONNECTED and BOARD_CONNECTED are hard states that indicate we *know* the
 * current status of the board.  BOARD_FLUX is a soft state that indicates the board
 * *may* be powered, but we're not sure.  In most cases, it indicates that the board
 * is powered up, however in cases where a board in the middle of the chain is in the
 * process of rebooting, downstream boards may be disconnected with no notification.
 */
#define BOARD_DISCONNECTED 		0
#define BOARD_FLUX           1
#define BOARD_CONNECTED         2
#define BOARD_PROBE             4

#define EXT_CHECK_PERIOD 1
#define MAX_I2C_TX_BUFFERS 4
#define ADVANCE_BUFFER_INDEX(index)  { index = (index + 1) % MAX_I2C_TX_BUFFERS; }
//#define RETRY_READ_MAX_COUNT 300
#define RETRY_READ_MAX_COUNT 10
#define RETRY_WRITE_MAX_COUNT 10
//#define EXT_IRQ_ENABLE

#define EXT_FIRMWARE_DIR "/system/ext_fw"

#define EXT_DBG
#ifdef EXT_DBG
	#define LOGD printk
#else
	#define LOGD(format, arg...)
#endif

#define MAX_ABSENSE 5

#define gpio_to_irq __gpio_to_irq

//=======================
//  Typedefs
//=======================
struct ext_con_platform_data {
	int i2c_bus;
	int ext_sel;
	int ext_int;
	int irq;
};

struct ext_queue {
    struct kfifo fifo;
    spinlock_t lock;
};

enum file_transfer_status {
    FILE_TRANSFER_IDLE = 0,
    FILE_TRANSFER_IN_PROGRESS,
    FILE_TRANSFER_COMPLETE,
    FILE_TRANSFER_ABORT
};
#define MAX_FILENAME 255

struct file_transfer_chunk {
    struct list_head list;
    uint32_t offset;
    long timestamp;
};

struct file_transfer_client {
    char filename[MAX_FILENAME];
    atomic_t status;
    uint8_t* buffer;
    uint32_t filesize;
    uint32_t offset_start;
    uint32_t offset_end;
    struct {
        struct file_transfer_chunk pending;
        struct file_transfer_chunk in_flight;
        struct file_transfer_chunk completed;
    } chunks;
    struct ext_queue recv_queue;
};

struct file_transfer_server {
    atomic_t status;
    uint8_t* buffer;
    uint32_t filesize;
    struct ext_queue recv_queue;
    wait_queue_head_t wq;
};

struct board_status {
	atomic_t connection;		// Connection status of EXT device: 0 (disconnected), 1 (connected)
    struct ext_queue send_queue;
    struct ext_queue recv_queue;
    struct ext_queue dbus_queue;
    wait_queue_head_t wq;
    struct file_transfer_server file_transfer_server;
    struct file_transfer_client file_transfer_client;
    int absense;
};

enum pulsar_state {
    PULSAR_STATE_IDLE,
    PULSAR_STATE_DEBOUNCE,
    PULSAR_STATE_ACTIVE,
};

typedef void (*packet_handler_f)(struct ext_packet*, int);

//=======================
//  Locals
//=======================

    /*
     * Private Functions
     */
static void poll_timer_expired(
	unsigned long arg );
static void pulsar_timer_expired(
    unsigned long arg );
static void poll_startup_finished(
    unsigned long arg );
static void ext_receive_packet(
    struct ext_packet * pkt,
    int board );
static void do_pulsar(
    struct work_struct* work );

    /*
     * Private Data
     */
DECLARE_WAIT_QUEUE_HEAD( ext_wq );
DECLARE_WAIT_QUEUE_HEAD( ext_buffer_wq );
DECLARE_WAIT_QUEUE_HEAD( ext_connection_wq );
DECLARE_WORK( pulsar_work, do_pulsar );
static atomic_t connection_state_changed;
static struct board_status ext_boards[NUM_BOARDS];
static struct timer_list poll_timer;
static int system_up_ready = 0;
static dev_t dev_no;
static struct cdev cdev;
static struct class* ext_class = NULL;
static dev_t dev_no_status;
static struct cdev cdev_status;
static struct class* ext_status_class = NULL;
static struct {
    int ext_int;
    int ext_sel;
} ext_gpio;
static int i2c_bus;
static void* buffer_pool;
static struct ext_queue buffers;
static packet_handler_f packet_handler[EXT_PACKET_MAX];
static struct {
    struct timer_list timer;
    int state;
} pulsar;
unsigned int ext_sel_irq;
struct irq_chip* ext_sel_irq_chip;
struct timer_list poll_startup;

//=======================
//  Functions
//=======================

//=============================================================================
static void
checksum_add(
    struct ext_packet* packet
)
//=============================================================================
{
    packet->crc = 0;
    packet->pad = 0xAA;
    packet->crc = crc32( 0xbeef, (unsigned char*)packet, packet->len + sizeof ( *packet ) );
}

//=============================================================================
static int
checksum_check(
    struct ext_packet* packet
)
//=============================================================================
{
    u32 crc;

    if ( packet->len > I2C_MTU_HOST )
        return -1;

    crc = packet->crc;
    packet->crc = 0;
    if ( crc != crc32( 0xbeef, (unsigned char*)packet, packet->len + sizeof ( *packet ) ) )
    {
        return -1;
    }
    return 0;
}

//=============================================================================
static int
gxp_i2c_send(
    u8 devaddr, 
    u8 *buf, 
    int count
)
/* Write to I2C device */
//=============================================================================
{

	struct i2c_adapter *adap;
	int err=0;
	struct i2c_msg msg[1];

	adap = i2c_get_adapter(i2c_bus-1);
	if (!adap)
		return -ENODEV;

	//printk("%s\n", __func__);

	msg->addr = devaddr;	/* I2C address of chip */
	msg->flags = 0;
	msg->len = count;
	msg->buf = buf;
	err = i2c_transfer(adap, msg, 1);
	i2c_put_adapter(adap);
	if (err >= 0)
		return 0;

	return err;

}

//=============================================================================
static int 
gxp_i2c_recv(
    u8 devaddr, 
    u8 *buf, 
    int count
)
/* Read from I2C device */
//=============================================================================
{

	struct i2c_adapter *adap;
	int err=0;
	struct i2c_msg msg[1];

	adap = i2c_get_adapter(i2c_bus-1);
	if (!adap)
		return -ENODEV;

	msg->addr = devaddr;	/* I2C address */
	msg->flags = I2C_M_RD;
	msg->len = count;
	msg->buf = buf;
	err = i2c_transfer(adap, msg, 1);
	i2c_put_adapter(adap);

	if (err >= 0)
		return 0;
	return err;

}

//=============================================================================
static int 
gxp_ext_write(
	char addr,
	unsigned int len,
	u8 *buf
)
//=============================================================================
{
	//printk("%s: addr=%d len=%d\n", __func__, addr, len);
    wmb();
	return gxp_i2c_send(addr, buf, len);

};

//=============================================================================
static int 
gxp_ext_read(
	char addr,
	u8 *buf,
	unsigned int len
)
//=============================================================================
{
    int ret;

	//printk("%s: addr=%d len=%d\n", __func__, addr, len);

	ret = gxp_i2c_recv(addr, buf, len);
    rmb();

    return ret;
};

//=============================================================================
static int
ext_server_task(
    void* data
)
//=============================================================================
{
    int board = (int)data;
    int sig;
    uint8_t* buffer;
    struct ext_packet* packet;
    int status;

    LOGD("ext_server_task started\n");

    /* drain any stray messages */
    while ( kfifo_out_spinlocked( 
                &ext_boards[board].file_transfer_server.recv_queue.fifo, 
                &buffer, 
                sizeof ( buffer ), 
                &buffers.lock ) != 0 )
    {
        kfifo_in_spinlocked( &buffers.fifo, &buffer, sizeof ( buffer ), &buffers.lock );
    }

    /* respond to client */
    while ( kfifo_out_spinlocked( &buffers.fifo, &buffer, sizeof ( buffer ), &buffers.lock ) == 0 )
    {
        LOGD("gxp_ext_host: server waiting on buffers...\n");
        wait_event_interruptible( ext_buffer_wq, kfifo_len( &buffers.fifo ) != 0 );
    }
    packet = ( struct ext_packet* )buffer;

    LOGD("gxp_ext_host: response for file request, filesize is %d\n", ext_boards[board].file_transfer_server.filesize);

    packet->type = EXT_PACKET_FILE_RESPONSE;
    packet->len = 0;
    packet->msg.filesize = ext_boards[board].file_transfer_server.filesize;

    kfifo_in_spinlocked( &ext_boards[board].send_queue.fifo, &buffer, sizeof ( buffer ), &ext_boards[board].send_queue.lock );
    wake_up_interruptible_all( &ext_wq );

    /* process chunk requests */
    while ( 1 )
    {
        sig = wait_event_interruptible_timeout(
                ext_boards[board].file_transfer_server.wq,
                kfifo_len( &ext_boards[board].file_transfer_server.recv_queue.fifo ) != 0,
                HZ * 30 );

        if ( sig < 0 )
            continue;

        if ( sig == 0 )
        {
            LOGD("gxp_ext_host: file transfer for board %d timed out\n", board);
            if ( ext_boards[board].file_transfer_server.buffer )
            {
                vfree( ext_boards[board].file_transfer_server.buffer );
                ext_boards[board].file_transfer_server.buffer = NULL;
            }
            else
            {
                LOGD("buffer is null....?\n");
            }
            atomic_set( &ext_boards[board].file_transfer_server.status, FILE_TRANSFER_IDLE );
            return 0;
        }

        if ( kfifo_out_spinlocked( 
                    &ext_boards[board].file_transfer_server.recv_queue.fifo,
                    &buffer,
                    sizeof ( buffer ),
                    &ext_boards[board].file_transfer_server.recv_queue.lock ) == 0 )
        {
            continue;
        }
        packet = ( struct ext_packet* )buffer;
        
        if ( packet->type == EXT_PACKET_FILE_RELEASE )
        {
            LOGD("gxp_ext_host: receive file release message for board %d\n", board);
            atomic_set( &ext_boards[board].file_transfer_server.status, FILE_TRANSFER_IDLE);
            kfifo_in_spinlocked( &buffers.fifo, &buffer, sizeof ( buffer ), &buffers.lock );
            if ( ext_boards[board].file_transfer_server.buffer )
                vfree( ext_boards[board].file_transfer_server.buffer );
            wake_up_interruptible_all( &ext_buffer_wq );
            return 0;
        }
        else if ( packet->type == EXT_PACKET_CHUNK_REQUEST )
        {
            uint32_t offset = packet->msg.chunk.offset;

            packet->type = EXT_PACKET_CHUNK_RESPONSE;
            if ( offset + ( I2C_MTU_HOST - sizeof(*packet) ) > ext_boards[board].file_transfer_server.filesize )
            {
                packet->len = ext_boards[board].file_transfer_server.filesize - offset;
            }
            else
            {
                packet->len = ( I2C_MTU_HOST - sizeof(*packet) );
            }

            memcpy( buffer + sizeof (*packet), ext_boards[board].file_transfer_server.buffer + offset, packet->len );

            kfifo_in_spinlocked(
                    &ext_boards[board].send_queue.fifo,
                    &buffer,
                    sizeof ( buffer ),
                    &ext_boards[board].send_queue.lock );
            wake_up_interruptible_all( &ext_wq );
        }
        else
        {
            LOGD("gxp_ext_host: received incorrect packet type %d during transfer for board %d\n", packet->type, board);
            kfifo_in_spinlocked( &buffers.fifo, &buffer, sizeof ( buffer ), &buffers.lock );
            wake_up_interruptible_all( &ext_buffer_wq );
        }
    
        status = atomic_read( &ext_boards[board].file_transfer_server.status );
        if ( status == FILE_TRANSFER_ABORT )
        {
            if ( ext_boards[board].file_transfer_server.buffer )
            {
                vfree( ext_boards[board].file_transfer_server.buffer );
                ext_boards[board].file_transfer_server.buffer = NULL;
            }
            atomic_set( &ext_boards[board].file_transfer_server.status, FILE_TRANSFER_IDLE );
            return 0;
        }
    }

    return 0;
}

//=============================================================================
static int
ext_task(
    void* data
)
//=============================================================================
{
    int ret;
    int sig;
    int i;
    uint8_t* buffer;

    while ( 1 )
    {
        sig = wait_event_interruptible(
                ext_wq,
                kfifo_len( &ext_boards[0].send_queue.fifo ) != 0 ||
                kfifo_len( &ext_boards[1].send_queue.fifo ) != 0 ||
                kfifo_len( &ext_boards[2].send_queue.fifo ) != 0 ||
                kfifo_len( &ext_boards[3].send_queue.fifo ) != 0 ||
                gpio_get_value( ext_gpio.ext_int ) == 1 );

        if ( sig < 0 )
            continue;

        /* poll for read requests */
        /* Read data from individual EXT devices */
        for ( i = 0; i < NUM_BOARDS; i++ )
        {
            uint8_t* buffer;

            if ( gpio_get_value( ext_gpio.ext_int ) == 0 )
            {
                /* no more data to read */
                break;
            }

            if ( atomic_read( &ext_boards[i].connection ) == BOARD_DISCONNECTED )
            {
                continue;
            }

            if ( kfifo_out_spinlocked( &buffers.fifo, &buffer, sizeof ( buffer ), &buffers.lock ) == 0 )
            {
                /* Try to free some buffers */
                //LOGD("gxp_ext_host: no free buffers to service read...skip to write\n");
                goto do_xmit;
            }

            /* Note: I2C read does not reliably error out when an EXT is held in power on reset....
             * clear the buffer before the read so we don't process garbage when this happens */
            memset( buffer, 0, I2C_MTU_EXT );
            ret = gxp_ext_read( 
                    I2C_DEVICE_ADDR_START + i,
                    buffer, 
                    I2C_MTU_EXT );

            if ( ret != 0 && atomic_read( &ext_boards[i].connection ) == BOARD_CONNECTED ) {
                LOGD("gxp_ext_host: read failure from board %d\n", i);
                ext_boards[i].absense++;
                if ( ext_boards[i].absense > MAX_ABSENSE )
                {
                    ext_boards[i].absense = 0;
                    /* Set board powered...board may be rebooting, unless we hear
                     * otherwise from it's parent lets continue to poll it */
                    atomic_set( &ext_boards[i].connection, BOARD_FLUX );
                    atomic_set( &connection_state_changed, 1 );
                    wake_up_interruptible( &ext_connection_wq );
                }
                kfifo_in_spinlocked( &buffers.fifo, &buffer, sizeof ( buffer ), &buffers.lock );
                wake_up_interruptible_all( &ext_buffer_wq );
                continue;
            }

            // Reading packet is successful.
            ext_boards[i].absense = 0;
            ext_receive_packet( (struct ext_packet*)buffer, i );
        }

        /* service write requests */
do_xmit:
        for ( i = 0; i < NUM_BOARDS; i++ )
        {
            if ( kfifo_len( &ext_boards[i].send_queue.fifo ) != 0 )
            {
                int retry_count = 0;

                if ( kfifo_out_spinlocked( 
                        &ext_boards[i].send_queue.fifo, 
                        &buffer, 
                        sizeof ( buffer ), 
                        &ext_boards[i].send_queue.lock ) == 0 )
                {
                    continue;
                }

                checksum_add( ( struct ext_packet* )buffer );

                while ( atomic_read( &ext_boards[i].connection ) != BOARD_DISCONNECTED )
                {
                    ret = gxp_ext_write(
                            I2C_DEVICE_ADDR_START + i,
                            ((struct ext_packet*)buffer)->len + sizeof ( struct ext_packet ),
                            buffer );

                    if ( ret == 0 )
                    {
                        break;
                    }

                    if ( retry_count <  RETRY_WRITE_MAX_COUNT )
                    {
                        retry_count++;

                        set_current_state( TASK_UNINTERRUPTIBLE );
                        schedule_timeout( HZ/1000 );
                    }
                    else
                    {
                        LOGD("gxp_ext_host: i2c transmit to addr %d fails with error %d\n",
                               I2C_DEVICE_ADDR_START + i, ret);
                        break;
                    }
                }

                kfifo_in_spinlocked( &buffers.fifo, &buffer, sizeof ( buffer ), &buffers.lock );
                wake_up_interruptible_all( &ext_buffer_wq );
            }
        }
    }

    return 0;
}

//=============================================================================
void
ext_packet_drop(
    struct ext_packet* pkt,
    int board
)
//=============================================================================
{
	if ( pkt->type != 0 && pkt->len != 0 )
	{
    	LOGD("gxp_ext_host: dropping packet of type %d from board %d\n", pkt->type, board);
	}
    kfifo_in_spinlocked( &buffers.fifo, &pkt, sizeof ( pkt ), &buffers.lock );
    wake_up_interruptible_all( &ext_buffer_wq );
}

//=============================================================================
void
ext_packet_dbus(
    struct ext_packet* pkt,
    int board
)
//=============================================================================
{
    kfifo_in_spinlocked( 
        &ext_boards[board].dbus_queue.fifo,
        &pkt,
        sizeof ( pkt ),
        &ext_boards[board].dbus_queue.lock );
    wake_up_interruptible_all( &ext_boards[board].wq );
}

//=============================================================================
void
ext_packet_file(
    struct ext_packet* pkt,
    int board
)
//=============================================================================
{
    if ( pkt->type == EXT_PACKET_FILE_REQUEST )
    {
        struct file* f;
        mm_segment_t fs;
        char fullpath[512];
        uint32_t filesize;

        if ( atomic_read( &ext_boards[board].file_transfer_server.status ) != FILE_TRANSFER_IDLE )
        {
            LOGD("gxp_ext_host: ignoring request for file, transfer in progress\n");
            return;
        }
        atomic_set( &ext_boards[board].file_transfer_server.status, FILE_TRANSFER_IN_PROGRESS );

		sprintf( fullpath, EXT_FIRMWARE_DIR "/%s", (char*)pkt + sizeof(*pkt) );
        f = filp_open( fullpath, O_RDONLY, 0 );
        LOGD("gxp_ext_host: servicing request for file %s from board %d\n", fullpath, board);
        if ( IS_ERR(f) )
        {
            LOGD("gxp_ext_host: unable to open %s\n", fullpath);

            pkt->len = 0;
            pkt->type = EXT_PACKET_FILE_RESPONSE;
            pkt->msg.filesize = 0;

            kfifo_in_spinlocked( 
                &ext_boards[board].send_queue.fifo, 
                &pkt, 
                sizeof ( pkt ), 
                &ext_boards[board].send_queue.lock );
            wake_up_interruptible_all( &ext_wq );
            atomic_set( &ext_boards[board].file_transfer_server.status, FILE_TRANSFER_IDLE );

            return;
        }

        fs = get_fs();
        set_fs( get_ds() );
        filesize = f->f_op->llseek( f, 0, SEEK_END );
        f->f_op->llseek( f, 0, SEEK_SET );
        
        ext_boards[board].file_transfer_server.filesize = filesize;
        ext_boards[board].file_transfer_server.buffer = vmalloc( filesize );
        //printk("vmalloc'd %08x for board %d\n", \
		//	   ext_boards[board].file_transfer_server.buffer, board);
        if ( ext_boards[board].file_transfer_server.buffer == NULL )
        {
            LOGD("gxp_ext_host: unable to allocate file buffer\n");
            set_fs( fs );
            filp_close( f, NULL );
            
            pkt->len = 0;
            pkt->type = EXT_PACKET_FILE_RESPONSE;
            pkt->msg.filesize = 0;

            kfifo_in_spinlocked( 
                &ext_boards[board].send_queue.fifo, 
                &pkt, 
                sizeof ( pkt ), 
                &ext_boards[board].send_queue.lock );
            wake_up_interruptible_all( &ext_wq );
            atomic_set( &ext_boards[board].file_transfer_server.status, FILE_TRANSFER_IDLE );

            return;
        }
        kfifo_in_spinlocked( &buffers.fifo, &pkt, sizeof ( pkt ), &buffers.lock );
        wake_up_interruptible_all( &ext_buffer_wq );


        f->f_op->read( f, ext_boards[board].file_transfer_server.buffer, filesize, &f->f_pos );
        set_fs( fs );
        filp_close( f, NULL );

        kthread_run( ext_server_task, (void*)board, "ext_transfer_task_%d", board );
        return;
    }
    else if ( pkt->type == EXT_PACKET_CHUNK_REQUEST ||
              pkt->type == EXT_PACKET_FILE_RELEASE )
    {
        kfifo_in_spinlocked(
                &ext_boards[board].file_transfer_server.recv_queue.fifo,
                &pkt,
                sizeof ( pkt ),
                &ext_boards[board].file_transfer_server.recv_queue.lock );
        wake_up_interruptible_all( &ext_boards[board].file_transfer_server.wq );
        return;
    }
    else /* ext->host transfer not yet supported */
        ext_packet_drop( pkt, board );
}

//=============================================================================
static void
ext_packet_link(
    struct ext_packet* pkt,
    int board
)
//=============================================================================
{
    //printk("ext_packet_link: board = %d, pkt->msg.link = %d\n", board, pkt->msg.link);

    if ( pkt->msg.link & BOARD_FLUX )
    {
        LOGD("gxp_ext_host: board %d powered \n", board + 1);
        /* Don't override a CONNECTED state if a board in the chain is soft-booted */
        local_irq_disable();
        if ( atomic_read( &ext_boards[board+1].connection ) == BOARD_PROBE ||
             atomic_read( &ext_boards[board+1].connection ) == BOARD_DISCONNECTED )
        {
            ext_boards[board+1].absense = 0;
            atomic_set( &ext_boards[board + 1].connection, BOARD_FLUX );
        }
        local_irq_enable();
    }
    if ( pkt->msg.link & BOARD_CONNECTED )
    {
        LOGD("gxp_ext_host: board %d connected\n", board );
        atomic_set( &ext_boards[board].connection, BOARD_CONNECTED );
    }
    if ( pkt->msg.link == BOARD_DISCONNECTED )
    {
        int i;

        /* link down is detected by previous board in the chain */
        for ( i = board + 1; i < NUM_BOARDS; i++ )
        {
            LOGD("gxp_ext_host: board %d disconnected\n", i);
            atomic_set( &ext_boards[i].connection, BOARD_DISCONNECTED );
        }
    }
    atomic_set( &connection_state_changed, 1 );
    wake_up_interruptible( &ext_connection_wq );
 
    kfifo_in_spinlocked( &buffers.fifo, &pkt, sizeof ( pkt ), &buffers.lock );
    wake_up_interruptible_all( &ext_buffer_wq );
}

//=============================================================================
static void
ext_receive_packet(
    struct ext_packet * pkt,
    int board
)
//=============================================================================
{
    if ( pkt->type > 0 && pkt->type < EXT_PACKET_MAX ) 
    {
		//printk("%s: pkt->type=%d  pkt->len=%d\n", __func__, pkt->type, pkt->len);
        if ( checksum_check( pkt ) < 0 )
        {
            printk("gxp_ext_host: packet failed checksum board %d type %d len %d sum %08x\n",\
				   board, pkt->type, pkt->len, pkt->crc);
            ext_packet_drop( pkt, board );
        }
        else
            packet_handler[pkt->type]( pkt, board );
    }
    else
        ext_packet_drop( pkt, board );
}

//=============================================================================
static irqreturn_t
ext_sel_irq_fxn(
    int irq,
    void* priv
)
//=============================================================================
{
    int i;

    for ( i = 0; i < NUM_BOARDS; i++ )
        atomic_set( &ext_boards[i].connection, BOARD_DISCONNECTED );

    atomic_set( &connection_state_changed, 1 );
    wake_up_interruptible( &ext_connection_wq );

    return IRQ_HANDLED;
}

#ifdef EXT_IRQ_ENABLE
//=============================================================================
static irqreturn_t
ext_irq(
	int irq,
	void * priv
)
// Description: Attached EXT requests host read back
//=============================================================================
{
    int i;

    wake_up_interruptible_all( &ext_wq );

	return IRQ_HANDLED;
}
#else
//=============================================================================
static void
poll_timer_expired(
	unsigned long arg
)
//=============================================================================
{
    poll_timer.expires = jiffies + (I2C_RECEIVE_POLL_PERIOD * HZ) / 1000;
    add_timer( &poll_timer );

    if ( gpio_get_value( ext_gpio.ext_int ) == 1 )
    {
        wake_up_interruptible_all( &ext_wq );
    }
}
#endif

//=============================================================================
static void
pulsar_timer_expired(
    unsigned long arg
)
//=============================================================================
{
    pulsar.timer.expires = jiffies + HZ * PULSAR_PERIOD;
    add_timer( &pulsar.timer);

    schedule_work( &pulsar_work );
}

//=============================================================================
static void
do_pulsar(
    struct work_struct* work
)
//=============================================================================
{
    int i;
   
    /* Send a single pulse to board 1 to notify it of it's address, then
     * check if the board disconnected while interrupts were disasbled.  Each
     * board will pass along the number of detected pulses + 1 to notify
     * the next board of its address.
     */

    if ( gpio_get_value( ext_gpio.ext_sel ) == 1 )
    {
        switch( pulsar.state )
        {
            case PULSAR_STATE_IDLE:
                pulsar.state = PULSAR_STATE_DEBOUNCE;
                break;
            case PULSAR_STATE_DEBOUNCE:
                pulsar.state = PULSAR_STATE_ACTIVE;
            case PULSAR_STATE_ACTIVE:
                if ( atomic_read( &ext_boards[0].connection ) != BOARD_CONNECTED )
                {
                    ext_sel_irq_chip->irq_mask( irq_get_irq_data( ext_sel_irq ) );
                    gpio_direction_output( ext_gpio.ext_sel, 0 );
                    set_current_state( TASK_UNINTERRUPTIBLE );
                    schedule_timeout( HZ / 10 );
                    gpio_direction_output( ext_gpio.ext_sel, 1 );
                    /* Let the capacitor hanging off ext_sel charge before polling for
                     * disconnection */
                    set_current_state( TASK_UNINTERRUPTIBLE );
                    schedule_timeout( HZ / 10 );
                    gpio_direction_input( ext_gpio.ext_sel );
                    ext_sel_irq_chip->irq_unmask( irq_get_irq_data( ext_sel_irq ) );

                    if ( gpio_get_value( ext_gpio.ext_sel ) == 0 )
                    {
                        for ( i = 0; i < NUM_BOARDS; i++ )
                        {
                            atomic_set( &ext_boards[i].connection, BOARD_DISCONNECTED );
                        }
                        atomic_set( &connection_state_changed, 1 );
                        wake_up_interruptible( &ext_connection_wq );
                    }
                    else if ( atomic_read( &ext_boards[0].connection ) == BOARD_DISCONNECTED )
                    {
                        ext_boards[0].absense = 0;
                        atomic_set( &ext_boards[0].connection, BOARD_FLUX );
                        atomic_set( &connection_state_changed, 1 );
                        wake_up_interruptible( &ext_connection_wq );
                    }
                }
                break;
        }
    }
    else
    {
        switch( pulsar.state )
        {
            case PULSAR_STATE_IDLE:
                break;
            case PULSAR_STATE_DEBOUNCE:
                pulsar.state = PULSAR_STATE_IDLE;
                break;
            case PULSAR_STATE_ACTIVE:
                pulsar.state = PULSAR_STATE_IDLE;
                for ( i = 0; i < NUM_BOARDS; i++ )
                {
                    atomic_set( &ext_boards[i].connection, BOARD_DISCONNECTED );
                }
                atomic_set( &connection_state_changed, 1 );
                wake_up_interruptible( &ext_connection_wq );
                break;
        }
    }

}

//=============================================================================
static int 
poll_all_boards(
    void* data
)
//=============================================================================
{
    uint8_t* buffer;
    struct ext_packet* packet;
    int i;
    int sig;

    printk("gxp_ext_host: poll all boards\n");
    for ( i = 0; i < NUM_BOARDS; i++ )
    {
        atomic_cmpxchg( &ext_boards[i].connection, BOARD_DISCONNECTED, BOARD_PROBE );

        while ( kfifo_out_spinlocked( &buffers.fifo, &buffer, sizeof ( buffer ), &buffers.lock ) == 0 )
        {
            LOGD("gxp_ext_host: poll waiting on buffers...\n");
            wait_event_interruptible( ext_buffer_wq, kfifo_len( &buffers.fifo ) != 0 );
        }
        packet = ( struct ext_packet* )buffer;

        packet->type = EXT_PACKET_LINK;
        packet->len = 0;
       
        printk("probe board %d\n", i ); 
        kfifo_in_spinlocked(
            &ext_boards[i].send_queue.fifo, 
            &buffer, 
            sizeof ( buffer ), 
            &ext_boards[i].send_queue.lock );
        wake_up_interruptible_all( &ext_wq );
    }

    poll_startup.expires = jiffies + HZ * 4;
    add_timer( &poll_startup );

    return 0;
}

//=============================================================================
static void
poll_startup_finished(
    unsigned long arg
)
//=============================================================================
{
    int i;

    printk("gxp_ext_host: poll startup finished\n");
    for ( i = 0; i < NUM_BOARDS; i++ )
    {
        if ( atomic_cmpxchg( &ext_boards[i].connection,  BOARD_PROBE, BOARD_DISCONNECTED ) ==
                BOARD_DISCONNECTED )
        {
            printk("gxp_ext_host: board %d set to disconnected by startup poll\n", i );
        }
    }

    pulsar.timer.expires = jiffies + HZ * PULSAR_PERIOD;
    add_timer( &pulsar.timer );
}

//=============================================================================
static unsigned int
dev_ext_status_poll(
    struct file* filp,
    poll_table* wait
)
//=============================================================================
{
    unsigned int mask = 0;
    
    if ( atomic_read( &connection_state_changed ) ) {
        atomic_set( &connection_state_changed, 0 );
        mask = POLLIN | POLLRDNORM;
    }

    poll_wait( filp, &ext_connection_wq, wait );

    return mask;
}

//=============================================================================
static long
dev_ext_status_ioctl(
    struct file* filp,
    unsigned int cmd,
    unsigned long arg
)
//=============================================================================
{
#define EXT_IOC_MAGIC 'X'
#define EXT_IOC_STATUS _IOR( EXT_IOC_MAGIC, 0, int[NUM_BOARDS] )
    int ext_status[NUM_BOARDS];
    int i;

    switch ( cmd )
    {
        case EXT_IOC_STATUS:
            for ( i = 0; i < NUM_BOARDS; i++ )
            {
                ext_status[i] = atomic_read( &ext_boards[i].connection );
            }

            if ( copy_to_user( ( int __user * )arg, &ext_status[0], sizeof ( ext_status ) ) )
                return -EFAULT;

            return 0;
        default:
            return -EINVAL;
    }

    return -EINVAL;
}

static struct file_operations dev_ext_status_fops = {
    owner: THIS_MODULE,
    unlocked_ioctl: dev_ext_status_ioctl,
    poll: dev_ext_status_poll
};

//=============================================================================
static unsigned int
dev_ext_poll(
    struct file* filp,
    poll_table* wait
)
//=============================================================================
{
    unsigned int mask = 0;
    int board = iminor( filp->f_dentry->d_inode );

    if ( kfifo_len( &ext_boards[board].dbus_queue.fifo  ) != 0 )
    {
        mask = ( POLLIN | POLLRDNORM );
    }

    poll_wait( filp, &ext_boards[board].wq, wait );

    return mask;
}

//=============================================================================
static int
dev_ext_read(
    struct file* filp,
    char __user* buf,
    size_t count,
    loff_t* ppos
)
//=============================================================================
{
    int board = iminor( filp->f_dentry->d_inode );
    uint8_t* buffer;
    ssize_t ret;

	//printk("%s: count=%d\n", __func__, count);

    if ( kfifo_len( &ext_boards[board].dbus_queue.fifo ) == 0 )
        return 0;

    if ( kfifo_out_spinlocked( 
            &ext_boards[board].dbus_queue.fifo,
            &buffer,
            sizeof ( buffer ),
            &ext_boards[board].dbus_queue.lock ) == 0 )
    {
        return 0;
    }

    if ( count < ((struct ext_packet*)buffer)->len )
    {
        LOGD("gxp_ext_host: dev_ext_read: user buffer too small, board is %d, packet len is %d\n", board, ((struct ext_packet*)buffer)->len);
        kfifo_in_spinlocked(
            &ext_boards[board].dbus_queue.fifo,
            &buffer,
            sizeof ( buffer ),
            &ext_boards[board].dbus_queue.lock );
        return -EAGAIN;
    }

    if ( copy_to_user( buf, buffer + sizeof ( struct ext_packet ), ((struct ext_packet*)buffer)->len ) )
    {
        kfifo_in_spinlocked(
            &ext_boards[board].dbus_queue.fifo,
            &buffer,
            sizeof ( buffer ),
            &ext_boards[board].dbus_queue.lock );
        return -EFAULT;
    }
    ret = ((struct ext_packet*)buffer)->len;

    kfifo_in_spinlocked( &buffers.fifo, &buffer, sizeof ( buffer ), &buffers.lock );
    wake_up_interruptible_all( &ext_buffer_wq );

    return ret;
}

//=============================================================================
static int
dev_ext_write(
    struct file* filp,
    const char* buf,
    size_t count,
    loff_t* ppos
)
//=============================================================================
{
    int board = iminor( filp->f_dentry->d_inode );
    uint8_t* buffer;
    
	//printk("%s: count=%d\n", __func__, count);

	if ( count + sizeof ( struct ext_packet ) > I2C_MTU_HOST )
    {
        LOGD("gxp_ext_host: dev_ext_write: message too big\n");
        return -EFAULT;
    }

    while ( kfifo_out_spinlocked( &buffers.fifo, &buffer, sizeof ( buffer ), &buffers.lock ) == 0 )
    {
        //LOGD("gxp_ext_host: dev_ext_write: waiting on buffer...\n");
        if ( wait_event_interruptible( ext_buffer_wq, kfifo_len( &buffers.fifo ) != 0 ) < 0 )
        {
            return -EINTR;
        }
    }

    ((struct ext_packet*)buffer)->len = count;
    ((struct ext_packet*)buffer)->type = EXT_PACKET_DBUS;

    if ( copy_from_user( buffer + sizeof ( struct ext_packet ), buf, count ) )
    {
        kfifo_in_spinlocked( &buffers.fifo, &buffer, sizeof ( buffer ), &buffers.lock );
        return -EFAULT;
    }

    kfifo_in_spinlocked(
        &ext_boards[board].send_queue.fifo,
        &buffer,
        sizeof ( buffer ),
        &ext_boards[board].send_queue.lock );

    wake_up_interruptible_all( &ext_wq );

    return count;
}

//=============================================================================
static int
dev_ext_mmap(
    struct file* filp,
    struct vm_area_struct* vma
)
//=============================================================================
{
    int board = iminor( filp->f_dentry->d_inode );
    unsigned long start = vma->vm_start;
    unsigned long size = vma->vm_end - vma->vm_start;
    unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;
    unsigned long page, pos;

    if ( offset + size > MAX_FILE_SIZE )
    {
        return -EINVAL;
    }

    if ( atomic_read( &ext_boards[board].file_transfer_server.status ) == FILE_TRANSFER_IN_PROGRESS )
    {
        if ( ext_boards[board].file_transfer_server.buffer == NULL )
            return -EINVAL;

        pos = ( unsigned long )ext_boards[board].file_transfer_server.buffer + offset;
    }
    else if ( atomic_read( &ext_boards[board].file_transfer_client.status ) == FILE_TRANSFER_IN_PROGRESS )
    {
        if ( ext_boards[board].file_transfer_client.buffer == NULL )
            return -EINVAL;

        pos = ( unsigned long )ext_boards[board].file_transfer_client.buffer + offset;
    }
    else
        return -EINVAL;

    while ( size > 0 )
    {
        page = vmalloc_to_pfn( (void*)pos );
        if ( remap_pfn_range(
                    vma,
                    start,
                    page,
                    PAGE_SIZE,
                    pgprot_noncached( vma->vm_page_prot ) ) )
        {
            return -EAGAIN;
        }

        start += PAGE_SIZE;
        pos += PAGE_SIZE;

        if ( size > PAGE_SIZE )
            size -= PAGE_SIZE;
        else
            size = 0;
    }

    return 0;
}

//=============================================================================
static long
dev_ext_ioctl(
    struct file* filp,
    unsigned int cmd,
    unsigned long arg
)
//=============================================================================
{
#if 0
    int board = iminor( filp->f_dentry->d_inode );
    int filesize;
    int stat;

    switch ( cmd )
    {
        default:
            return -EINVAL;
    }
#endif

    return -EINVAL;
}

static struct file_operations dev_ext_fops = {
    owner: THIS_MODULE,
    unlocked_ioctl: dev_ext_ioctl,
    mmap: dev_ext_mmap,
    read: dev_ext_read,
    write: dev_ext_write,
    poll: dev_ext_poll
};

/*ext proc ops*/
//=============================================================================
static int
ext_proc_read(
    char *page, 
    char **start, 
    off_t off,
    int count, 
    int *eof, 
    void *data
)
//=============================================================================
{
	int len = sizeof(system_up_ready);

	sprintf(page, "%d", system_up_ready);

	if (len <= off + count)
		*eof = 1;
	*start = page + off;
	len -= off;
	if (len > count)
		len = count;
	if (len <0)
		len = 0;

	return len;
}

//=============================================================================
static int 
ext_proc_write(
    struct file *file, 
    const char *buffer,
    unsigned long count, 
    void *data
)
//=============================================================================
{
	char str;
	int value = 0;

	if (count > sizeof(int))
	{
		return (-EINVAL);
	}

	if (copy_from_user(&str, buffer, count))
	{
		return (-EFAULT);
	}

	sscanf(&str, "%d", &value);

	if (value)
		system_up_ready = 1;
	else
		system_up_ready = 0;

	return count;
}


/*
 * Platform Support
 */

//=============================================================================
static int __init
ext_host_probe(
	struct platform_device *pdev
)
//=============================================================================
{
    struct device *dev = &pdev->dev;
    struct ext_con_platform_data *pdata = dev->platform_data;
	int ret, i;
	struct proc_dir_entry *ent = NULL;
	uint32_t val;

	LOGD("%s\n", __func__);
	
	// Get ext_host device parameters.
	if (pdev->dev.of_node) {
		LOGD("Getting ext_host HW parameters from device tree.\n");
		ret = of_property_read_u32(pdev->dev.of_node, "i2c-bus", &val);
		if (ret < 0) {
			LOGD("Error: Invalid or missing 'i2c-bus' property!\n");
			return ret;
		}
		i2c_bus = val;
		ext_gpio.ext_sel = of_get_named_gpio(pdev->dev.of_node, "ext-sel-gpio", 0);
		ext_gpio.ext_int = of_get_named_gpio(pdev->dev.of_node, "ext-int-gpio", 0);
	} 
	else if (pdata){
		LOGD("Getting ext_host HW parameters from device driver.\n");
		i2c_bus = pdata->i2c_bus;
	    ext_gpio.ext_int = pdata->ext_int;
	    ext_gpio.ext_sel = pdata->ext_sel;
	}
	else {
		LOGD("Error: ext_host HW parameters not found!\n");
		return -ENODEV;
	}

    if ( alloc_chrdev_region( &dev_no, 0, NUM_BOARDS, "ext" ) )
    {
        LOGD("gxp_ext_host: unable to allocate chrdev region\n");
        return -ENODEV;
    }

    cdev_init( &cdev, &dev_ext_fops );
    if ( cdev_add( &cdev, dev_no, NUM_BOARDS ) )
    {
        LOGD("gxp_ext_host: unable to init cdev\n");
        unregister_chrdev_region( dev_no, NUM_BOARDS );
        return -ENODEV;
    }

    ext_class = class_create( THIS_MODULE, "ext" );
    if ( !ext_class )
    {
        unregister_chrdev_region( dev_no, NUM_BOARDS );
        cdev_del( &cdev );
        LOGD("gxp_ext_host: unable to create class\n");
        return -ENODEV;
    }

    for ( i = 0; i < NUM_BOARDS; i++ )
    {
        device_create( ext_class, NULL, MKDEV( MAJOR( dev_no ), i ), NULL, "ext%d", i );
    }

    alloc_chrdev_region( &dev_no_status, 0, 1, "ext_status" );
    cdev_init( &cdev_status, &dev_ext_status_fops );
    cdev_add( &cdev_status, dev_no_status, 1 );
    ext_status_class = class_create( THIS_MODULE, "ext_status" );
    device_create( ext_status_class, NULL, MKDEV( MAJOR( dev_no_status ), 0 ), NULL, "ext_status" );

    atomic_set( &connection_state_changed, 0 );
    memset( ext_boards, 0, sizeof ( ext_boards ) );
    for ( i = 0; i < NUM_BOARDS; i++  )
    {
		atomic_set( &ext_boards[i].connection, BOARD_DISCONNECTED );
        if ( kfifo_alloc( 
                &ext_boards[i].send_queue.fifo, 
                I2C_QUEUE_SIZE * sizeof ( uint8_t* ),
                GFP_KERNEL ) )
        {
            LOGD("gxp_ext_host: unable to allocate fifo\n");
            return -ENODEV;
        }
        kfifo_reset( &ext_boards[i].send_queue.fifo );
        if ( kfifo_alloc(
                &ext_boards[i].recv_queue.fifo,
                I2C_QUEUE_SIZE * sizeof ( uint8_t* ),
                GFP_KERNEL ) )
        {
            LOGD("gxp_ext_host: unable to allocate fifo\n");
            return -ENODEV;
        }
        kfifo_reset( &ext_boards[i].recv_queue.fifo );
        if ( kfifo_alloc(
                &ext_boards[i].dbus_queue.fifo,
                I2C_QUEUE_SIZE * sizeof ( uint8_t* ),
                GFP_KERNEL ) )
        {
            LOGD("gxp_ext_host: unable to allocate fifo\n");
            return -ENODEV;
        }
        kfifo_reset( &ext_boards[i].dbus_queue.fifo );
        init_waitqueue_head( &ext_boards[i].wq );

        atomic_set( &ext_boards[i].file_transfer_server.status, FILE_TRANSFER_IDLE );
        ext_boards[i].file_transfer_server.buffer = NULL;
        if ( kfifo_alloc(
                    &ext_boards[i].file_transfer_server.recv_queue.fifo,
                    I2C_QUEUE_SIZE * sizeof ( uint8_t* ),
                    GFP_KERNEL ) )
        {
            LOGD("gxp_ext_host: unable to allocate fifo\n");
            return -ENODEV;
        }
        kfifo_reset( &ext_boards[i].file_transfer_server.recv_queue.fifo );
        init_waitqueue_head( &ext_boards[i].file_transfer_server.wq );
    }

    if ( kfifo_alloc(
            &buffers.fifo,
            I2C_QUEUE_SIZE * sizeof ( uint8_t* ),
            GFP_KERNEL ) )
    {
        LOGD("gxp_ext_host: unable to allocate fifo\n");
        return -ENODEV;
    }
    kfifo_reset( &buffers.fifo );
    
    buffer_pool = vmalloc( I2C_QUEUE_SIZE * I2C_MTU_HOST );
    for ( i = 0; i < I2C_QUEUE_SIZE; i++ )
    {
        uint8_t* buffer = (uint8_t*)buffer_pool + ( i * I2C_MTU_HOST );

        if ( buffer == NULL )
        {
            LOGD("gxp_ext_host: unable to allocate buffers\n");
            return -EFAULT;
        }

        kfifo_in_spinlocked( &buffers.fifo, &buffer, sizeof ( buffer ), &buffers.lock );
    }

    packet_handler[EXT_PACKET_NOP] = ext_packet_drop;
    packet_handler[EXT_PACKET_DISCOVERY] = ext_packet_drop;
    packet_handler[EXT_PACKET_LINK] = ext_packet_link;
    packet_handler[EXT_PACKET_NET] = ext_packet_drop;
    packet_handler[EXT_PACKET_DBUS] = ext_packet_dbus;
    packet_handler[EXT_PACKET_FILE_REQUEST] = ext_packet_file;
    packet_handler[EXT_PACKET_FILE_RESPONSE] = ext_packet_file;
    packet_handler[EXT_PACKET_FILE_RELEASE] = ext_packet_file;
    packet_handler[EXT_PACKET_CHUNK_REQUEST] = ext_packet_file;
    packet_handler[EXT_PACKET_CHUNK_RESPONSE] = ext_packet_file;

#ifndef EXT_IRQ_ENABLE
	// Init timer for I2C-receive polling.
	init_timer( &poll_timer );
	poll_timer.data = 0;
	/* Make sure ext is power up before i2c receive data */
	poll_timer.function = poll_timer_expired;
    poll_timer.expires = jiffies + HZ / 2;
    add_timer( &poll_timer );
#else
	/* Register IRQ Handlers */
	pdata->irq = gpio_to_irq(ext_gpio.ext_int);
	if (gpio_request(ext_gpio.ext_int, "ext_int") < 0)
	{
		LOGD("Can't get ext int gpio!\n");
		return -ENODEV;
	}

	ret = request_irq(pdata->irq, ext_irq, IRQF_TRIGGER_RISING|IRQF_TRIGGER_FALLING, "ext-irq", 0 );
	if ( ret < 0 )
	{
		LOGD("EXT unable to request irq\n");
		return -ENODEV;
	}
#endif

    init_timer( &pulsar.timer );
    pulsar.timer.data = 0;
    pulsar.timer.function = pulsar_timer_expired;

    init_timer( &poll_startup );
    poll_startup.function = poll_startup_finished;

    kthread_run( ext_task, 0, "ext_task" );

    ret = gpio_request( ext_gpio.ext_sel, "ext-host" );
	if (ret != 0)
	{
        LOGD("gxp_ext_host: gpio_request fails\n");
		return  -EBUSY;
	}
	gpio_direction_input( ext_gpio.ext_sel );
    ext_sel_irq = gpio_to_irq( ext_gpio.ext_sel );
    ext_sel_irq_chip = irq_get_chip( ext_sel_irq );

    ret = request_irq( ext_sel_irq, ext_sel_irq_fxn, IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING, "ext-sel-irq", 0 );
    if ( ret < 0 )
    {
        LOGD("gxp_ext_host: unable to request irq\n");
        return -ENODEV;
    }

    if ( gpio_get_value( ext_gpio.ext_sel ) == 1 )
    {
		//printk("%s: poll_all_boards\n", __func__);
        /* poll all boards for their own status if one is already connected */
        kthread_run( poll_all_boards, 0, "ext_polling" );
    }
    else
    {
        pulsar.timer.expires = jiffies + HZ * PULSAR_PERIOD;
        add_timer( &pulsar.timer );
    }

	/*Add proc interface to notify system up is ready */
	ent = create_proc_entry("ext_ready", S_IFREG|S_IRUGO, NULL);
	if (ent)
	{
		ent->read_proc = ext_proc_read;
		ent->write_proc = ext_proc_write;
		ent->data = NULL;
	}
	else
	{
		LOGD("Create ext proc error!\n");
    }
	LOGD("GXP EXT host platform device registered\n");

	return 0;
}

//=============================================================================
static int __devexit
ext_host_remove(
	struct platform_device *pdev
)
{
	return 0;
}

//=============================================================================

static struct ext_con_platform_data ext_con_info = {
	// These parameters are not used if they are available from kernel device-tree.
	.i2c_bus = 1,
	.ext_sel = GPIO_PORTE(6),
	.ext_int = GPIO_PORTD(24),
};

static struct platform_device ext_host_device = {
	.name = "ext_connect",
	.dev = {
		.platform_data = &ext_con_info,
	}
};

static struct of_device_id ext_host_of_ids[] = {
    { .compatible = "ext_host" },
    { },
};

MODULE_DEVICE_TABLE(of, ext_host_of_ids);

static struct platform_driver __refdata ext_host_driver = {
    .driver = {
        .name = "ext_host",
		.owner = THIS_MODULE,
        .of_match_table = of_match_ptr(ext_host_of_ids),
    },
    .probe = ext_host_probe,
    .remove = __devexit_p(ext_host_remove),
};


//=============================================================================
static int __init
ext_host_init(
	void
)
//=============================================================================
{
	LOGD("%s\n", __func__);

	platform_device_register(&ext_host_device);
    return platform_driver_probe(&ext_host_driver, ext_host_probe);
}

//=============================================================================
static void __exit
ext_host_exit(
	void
)
//=============================================================================
{
    platform_driver_unregister(&ext_host_driver);
}

module_init(ext_host_init);
module_exit(ext_host_exit);
