/*
 * Copyright 2013 Grandstream Networks, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

#include <linux/err.h>
#include <linux/mutex.h>
#include <linux/module.h>
#define DEBUG 1
#include <linux/device.h>
#include <linux/platform_device.h>
#include <linux/gpio.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/of.h>
#include <linux/of_gpio.h>
#include <linux/hrtimer.h>
#include <linux/mfd/dvf9a.h>
#include <linux/workqueue.h>

#define EHS_EVENT_DCIN_POLLING

extern unsigned short daif_read_dcin(int dcin_number);

enum led_state {
    LED_STATE_OFF,
    LED_STATE_RED,
    LED_STATE_GREEN,
    LED_STATE_ORANGE
};


enum leds {
    LED_R_G_CTRL = 0, // this gpio port doesn't control anything
    LED_L1_R,
    LED_L2_R,
    LED_L1_G,
    LED_L2_G,
    LED_MSG_R,
    LED_PROG1,
    LED_PROG2,
    LED_PROG3,
    LED_PROG4,
    LED_PROG5,
    LED_PROG6,
    LED_PROG7,
    LED_PROG8,
    LED_MAX
};

char led_names[][16] = {
     "LED_R_G_CTRL",
     "LED_L1_R",
     "LED_L2_R",
     "LED_L1_G",
     "LED_L2_G",
     "LED_MSG_R",
    "LED_PROG1",
    "LED_PROG2",
    "LED_PROG3",
    "LED_PROG4",
    "LED_PROG5",
    "LED_PROG6",
    "LED_PROG7",
    "LED_PROG8"
};

enum HRT_TASK {
    HRT_TASK_LED = 0,
    HRT_TASK_DCIN,
    NUM_HRT_TASK 
};

struct hrt_data {
    void* hrt_data[NUM_HRT_TASK];
    void (*hrt_tasks[NUM_HRT_TASK])(void*);
    
    ktime_t delay;
    struct hrtimer timer;
    struct device* dev;
};

struct led_data {
    int gpio[LED_MAX];
    uint32_t valid_mask;
    enum led_state r_g_control;
    int blinker;
};
static uint32_t l1_mask;
static uint32_t l1_blink_mask;
static uint32_t l2_mask;
static uint32_t l2_blink_mask;
static uint32_t msg_mask;
static uint32_t msg_blink_mask;
static uint32_t red_mask;
static uint32_t green_mask;
static uint32_t blink_mask;

static struct led_data drvdata_t;

#ifdef EHS_EVENT_DCIN_POLLING
struct dcin_data {
    int configured;
    int dcin_number;
    int dcin_ticks;
    struct device* dev;
};

typedef struct {
    struct work_struct work;
    int event;
    struct dcin_data* drvdata;
} dcin_event_work_t;

static struct workqueue_struct* dcin_event_wq;
static struct dcin_data dcindata;

static void* dcin_state_idle(struct dcin_data* drvdata, int dcin_pin);
static void* dcin_state_debounce(struct dcin_data* drvdata, int dcin_pin);
static void* dcin_state_counting(struct dcin_data* drvdata, int dcin_pin);
static void dcin_send_event(struct work_struct * work);


static void dcin_send_event(struct work_struct * work)
{
    char dcin_state_string[12];
    char* envp[] = { dcin_state_string, NULL };
    dcin_event_work_t* dcin_work = (dcin_event_work_t *) work;

    snprintf(dcin_state_string, sizeof(dcin_state_string), "DCIN=%d", dcin_work->event);
    kobject_uevent_env(&(dcin_work->drvdata->dev->kobj), KOBJ_CHANGE, envp);
    
    kfree((void *) work);
    return;
}

static void* dcin_state_idle(struct dcin_data* drvdata, int dcin_pin)
{
    switch(dcin_pin) {
        case 1:
        default:
            drvdata->dcin_ticks = 0;
            return dcin_state_idle;

        case 0:
            drvdata->dcin_ticks++;
            return dcin_state_debounce;
    }
}

static void* dcin_state_debounce(struct dcin_data* drvdata, int dcin_pin)
{
    switch(dcin_pin) {
        case 1:
        default:
            drvdata->dcin_ticks = 0;
            return dcin_state_idle;

        case 0:
            drvdata->dcin_ticks++;
            return dcin_state_counting;
    }
}

static void* dcin_state_counting(struct dcin_data* drvdata, int dcin_pin)
{
    dcin_event_work_t* work;
    
    switch(dcin_pin) {
        case 1:
        default:
            printk("DCIN: Event ( %d )\n", drvdata->dcin_ticks);
            work = (dcin_event_work_t *) kmalloc(sizeof(dcin_event_work_t), GFP_KERNEL);
            if (work) {
              INIT_WORK((struct work_struct *)work, dcin_send_event);
              work->event = drvdata->dcin_ticks;
              work->drvdata = drvdata;
              queue_work(dcin_event_wq, (struct work_struct *) work);
            }
            drvdata->dcin_ticks = 0;
            return dcin_state_idle;
        
        case 0:
            drvdata->dcin_ticks++;
            return dcin_state_counting;
    }
}

static void dcin_task(void* priv)
{
    static void* (*dcin_state_fxn)(struct dcin_data*, int) = dcin_state_idle;
    struct dcin_data* drvdata = (struct dcin_data*)priv;

    if (!drvdata->configured)
        return;

    dcin_state_fxn = dcin_state_fxn(drvdata, daif_read_dcin(drvdata->dcin_number));
}
#endif

static enum hrtimer_restart hrt_task(struct hrtimer *timer)
{
    struct hrt_data *drvdata = container_of(timer, struct hrt_data, timer);
    int i;

    hrtimer_forward_now(&drvdata->timer, drvdata->delay);

    for (i=0; i < NUM_HRT_TASK; i++) {
        if (drvdata->hrt_tasks[i]) {
            drvdata->hrt_tasks[i](drvdata->hrt_data[i]);
        }
    }

    return HRTIMER_RESTART;
}

static void set_l1_green(void)
{
    struct led_data* drvdata = &drvdata_t;

    gpio_set_value(drvdata->gpio[LED_L1_R], 0);
    gpio_set_value(drvdata->gpio[LED_L1_G], 1);
}

static void set_l1_red(void)
{
    struct led_data* drvdata = &drvdata_t;

    gpio_set_value(drvdata->gpio[LED_L1_R], 1);
    gpio_set_value(drvdata->gpio[LED_L1_G], 0);
}

static void set_l1_off(void)
{
    struct led_data* drvdata = &drvdata_t;

    gpio_set_value(drvdata->gpio[LED_L1_R], 0);
    gpio_set_value(drvdata->gpio[LED_L1_G], 0);
}

static void set_l2_green(void)
{
    struct led_data* drvdata = &drvdata_t;
    
    gpio_set_value(drvdata->gpio[LED_L2_R], 0);
    gpio_set_value(drvdata->gpio[LED_L2_G], 1);
}

static void set_l2_red(void)
{
    struct led_data* drvdata = &drvdata_t;
    
    gpio_set_value(drvdata->gpio[LED_L2_R], 1);
    gpio_set_value(drvdata->gpio[LED_L2_G], 0);
}

static void set_l2_off(void)
{
    struct led_data* drvdata = &drvdata_t;
    
    gpio_set_value(drvdata->gpio[LED_L2_R], 0);
    gpio_set_value(drvdata->gpio[LED_L2_G], 0);
}

static void set_msg_red(void)
{
    struct led_data* drvdata = &drvdata_t;

    gpio_set_value(drvdata->gpio[LED_MSG_R], 0);
}

static void set_msg_off(void)
{
    struct led_data* drvdata = &drvdata_t;
           
    gpio_set_value(drvdata->gpio[LED_MSG_R], 1);
}


static void led_task(void* priv)
{
    struct led_data* drvdata = (struct led_data*)priv;
    int i;
    uint32_t *active_mask;
    int r_g;

    drvdata =  &drvdata_t;
	/* Static Line Control */
    switch (l1_mask)
    {
        case 0: 
            set_l1_off();
            break;
        case 1:
            set_l1_red();
            break;
        case 2:
            set_l1_green();
            break;
        default:
            break;
    }
    
    if ( l1_blink_mask && drvdata->blinker<50)
    {
        set_l1_off();
    }

     switch (l2_mask)
    {
        case 0: 
            set_l2_off();
            break;
        case 1:
            set_l2_red();
            break;
        case 2:
            set_l2_green();
            break;
        default:
            break;
    }
    
    if ( l2_blink_mask && drvdata->blinker<50)
    {
        set_l2_off();
    }

    switch (msg_mask)
    {
        case 0: 
            set_msg_off();
            break;
        case 1:
            set_msg_red();
            break;
        default:
            break;
    }
    
    if ( msg_blink_mask && drvdata->blinker<50)
    {
        set_msg_off();
    }
  
  	/* Strobe Control */
	if (drvdata->r_g_control == LED_STATE_RED) {
        active_mask = &green_mask;
        r_g = 0;
        drvdata->r_g_control = LED_STATE_GREEN;
    }
    else {
        active_mask = &red_mask;
        r_g = 1;
        drvdata->r_g_control = LED_STATE_RED;
    }
        
	// Turn off all LEDs.
    for (i=1; i<LED_MAX; i++) {
        if (drvdata->valid_mask & (1<<i))
            gpio_set_value(drvdata->gpio[i], 1);
    }
	
	// Set red/green color of all LEDs.
    if (gpio_is_valid(drvdata->gpio[LED_R_G_CTRL]))
        gpio_set_value(drvdata->gpio[LED_R_G_CTRL], r_g);

	// Turn on all LEDs.
	for (i=1;i<LED_MAX;i++ ) {
        if (*active_mask & (1<<i) && drvdata->valid_mask & (1<<i)) {
			if ((blink_mask & (1<<i)) && drvdata->blinker>=50)
            	gpio_set_value(drvdata->gpio[i], 0);
			else if (!(blink_mask & (1<<i)))
				gpio_set_value(drvdata->gpio[i], 0);
        }
    }
  
	drvdata->blinker++;
	if ( drvdata->blinker >= 100 )
		drvdata->blinker = 0;
}

static struct ctl_table_header *sysctl_header;

/* 
    0 -- off
    1 -- red
    2 -- green
*/

static ctl_table led_table[] = {
    {
        .procname = "set_l1_color",
        .data = &l1_mask,
        .maxlen = sizeof(uint32_t),
        .mode = 0644,
        .proc_handler = proc_dointvec,
    },
    {
        .procname = "set_l1_blink",
        .data = &l1_blink_mask,
        .maxlen = sizeof(uint32_t),
        .mode = 0644,
        .proc_handler = proc_dointvec,
    },
    {
        .procname = "set_l2_color",
        .data = &l2_mask,
        .maxlen = sizeof(uint32_t),
        .mode = 0644,
        .proc_handler = proc_dointvec,
    },
    {
        .procname = "set_l2_blink",
        .data = &l2_blink_mask,
        .maxlen = sizeof(uint32_t),
        .mode = 0644,
        .proc_handler = proc_dointvec,
    },
    {
        .procname = "set_msg_color",
        .data = &msg_mask,
        .maxlen = sizeof(uint32_t),
        .mode = 0644,
        .proc_handler = proc_dointvec,
    },
    {
        .procname = "set_msg_blink",
        .data = &msg_blink_mask,
        .maxlen = sizeof(uint32_t),
        .mode = 0644,
        .proc_handler = proc_dointvec,
    },
    {  
        .procname = "red",
        .data = &red_mask,
        .maxlen = sizeof(uint32_t),
        .mode = 0644,
        .proc_handler = proc_dointvec,
    },
    {  
        .procname = "green",
        .data = &green_mask,
        .maxlen = sizeof(uint32_t),
        .mode = 0644,
        .proc_handler = proc_dointvec,
    },
    {   
        .procname = "blink",
        .data = &blink_mask,
        .maxlen = sizeof(uint32_t),
        .mode = 0644,
        .proc_handler = proc_dointvec,
    },


    { }
};

#ifdef EHS_EVENT_DCIN_POLLING
static ctl_table dcin_table[] = {
    {
        .procname = "configured",
        .data = NULL,
        .maxlen = sizeof(uint32_t),
        .mode = 0644,
        .proc_handler = proc_dointvec,
    },
    { }
};
#endif

static ctl_table led_root[] = {
    {
        .procname = "led",
        .mode = 0555,
        .child = led_table,
    },
#ifdef EHS_EVENT_DCIN_POLLING
    {
        .procname = "dcin",
        .mode = 0555,
        .child = dcin_table,
    },
#endif
    { }
};

static ctl_table dev_root[] = {
    {
        .procname = "dev",
        .mode = 0555,
        .child = led_root,
    },
    { }
};

static void* __devinit led_probe(struct platform_device *pdev) {
    struct led_data *drvdata;
    struct device_node *np = pdev->dev.of_node;
    int i;

    printk("<----%s\n", __func__);
    drvdata = &drvdata_t;

#if 0
	drvdata = kzalloc(sizeof(struct led_data), GFP_KERNEL);
	if (drvdata == NULL) {
		dev_err(&pdev->dev, "Failed to allocate device data for LED\n");
	    return NULL;
    }
#endif

        
#if 0 // for debug 
       drvdata->gpio[0] = of_get_named_gpio(np, "LED_R_G_CTRL", 0);
       printk("get_named_gpio = %d\n", drvdata->gpio[0]);

       drvdata->gpio[1] = of_get_named_gpio(np, "LED_L1_R", 0);
       printk("get_named_gpio = %d\n", drvdata->gpio[1]);


       drvdata->gpio[2] = of_get_named_gpio(np, "LED_L2_R", 0);
       printk("get_named_gpio = %d\n", drvdata->gpio[2]);
        

       drvdata->gpio[3] = of_get_named_gpio(np, "LED_L1_G", 0);
       printk("get_named_gpio = %d\n", drvdata->gpio[3]);
           gpio_set_value(drvdata->gpio[1], 0);

       drvdata->gpio[4] = of_get_named_gpio(np, "LED_L2_G", 0);
       printk("get_named_gpio = %d\n", drvdata->gpio[4]);

       drvdata->gpio[5] = of_get_named_gpio(np, "LED_MSG_R", 0);
       printk("get_named_gpio = %d\n", drvdata->gpio[5]);
#endif


    for ( i=0; i<LED_MAX; i++ ) 
    {
        drvdata->gpio[i] = of_get_named_gpio(np, led_names[i], 0);
        dev_dbg(&pdev->dev, "%s: %s\n", led_names[i], gpio_is_valid(drvdata->gpio[i]) ? "valid" : "invalid");
        printk("%s: %s\n", led_names[i], gpio_is_valid(drvdata->gpio[i]) ? "valid" : "invalid");
        if (gpio_is_valid(drvdata->gpio[i])) {
            gpio_request(drvdata->gpio[i], "grandstream-led");
            gpio_direction_output(drvdata->gpio[i], 1);
            
            /* Only add strobed LEDs to valid_mask */
            if ( !( i == LED_L1_R || i == LED_L2_R || i == LED_L1_G || i == LED_L2_G || i == LED_MSG_R ) )
                drvdata->valid_mask |= (1<<i);
        }
    }
    printk("valid_mask = %08x LED_MAX=%d\n", drvdata->valid_mask, LED_MAX);

    drvdata->r_g_control = LED_STATE_RED;

    printk("---->%s\n", __func__);
    return drvdata;
}

#ifdef EHS_EVENT_DCIN_POLLING
static void* __devinit dcin_probe(struct platform_device *pdev)
{
    struct dcin_data* drvdata;

#if 0
    drvdata = kzalloc(sizeof(struct dcin_data), GFP_KERNEL);
    if (drvdata == NULL) {
        dev_err(&pdev->dev, "Failed to allocated device data for DCIN\n");
        return NULL;
    }
#endif

    drvdata = &dcindata;

    drvdata->dev = &pdev->dev;
    drvdata->dcin_number = 2;   // DCIN2 for EHS_DET signal.
    dcin_table[0].data = &drvdata->configured;

    dcin_event_wq = create_workqueue("dcin_event_queue");

    return drvdata;
}
#endif

static int __devinit hrt_probe(struct platform_device *pdev)
{
	struct hrt_data *drvdata;

    printk("hrt_probe\n");

    drvdata = kzalloc(sizeof(struct hrt_data), GFP_KERNEL);
    if (drvdata == NULL) {
        dev_err(&pdev->dev, "Failed to allocate device data\n");
        return -ENOMEM;
    }
    
    drvdata->dev = &pdev->dev;
    if ((drvdata->hrt_data[HRT_TASK_LED] = led_probe(pdev))) {
        drvdata->hrt_tasks[HRT_TASK_LED] = led_task;
    }
#ifdef EHS_EVENT_DCIN_POLLING
    if ((drvdata->hrt_data[HRT_TASK_DCIN] = dcin_probe(pdev))) {
        drvdata->hrt_tasks[HRT_TASK_DCIN] = dcin_task;
    }
#endif
    
    hrtimer_init(&drvdata->timer, CLOCK_MONOTONIC, HRTIMER_MODE_REL);
    drvdata->timer.function = hrt_task;
    drvdata->delay = ktime_set(0, 5000000); /* 5ms */
	
    platform_set_drvdata(pdev, drvdata);

    sysctl_header = register_sysctl_table(dev_root);

    hrtimer_start(&drvdata->timer, drvdata->delay, HRTIMER_MODE_REL);

	dev_dbg(&pdev->dev, "HRT driver setup complete\n");

	return 0;
}

static void __devexit led_remove(struct led_data* drvdata) {
    int i;

    for (i=0;i<LED_MAX;i++) {
        if (gpio_is_valid(drvdata->gpio[i]))
            gpio_free(drvdata->gpio[i]);
    }

    kfree(drvdata);
}

#ifdef EHS_EVENT_DCIN_POLLING
static void __devexit dcin_remove(struct dcin_data* drvdata) {
    kfree(drvdata);
    flush_workqueue(dcin_event_wq);
    destroy_workqueue(dcin_event_wq);
}
#endif

static int __devexit hrt_remove(struct platform_device *pdev)
{
	struct hrt_data *drvdata = platform_get_drvdata(pdev);

    if (drvdata->hrt_data[HRT_TASK_LED])
        led_remove(drvdata->hrt_data[HRT_TASK_LED]);
#ifdef EHS_EVENT_DCIN_POLLING
    if (drvdata->hrt_data[HRT_TASK_DCIN])
        dcin_remove(drvdata->hrt_data[HRT_TASK_DCIN]);
#endif

    kfree(drvdata);

	return 0;
}

static const struct of_device_id led_of_match[] __devinitconst = {
	{ .compatible = "grandstream,led", },
	{},
};
MODULE_DEVICE_TABLE(of, led_of_match);

static struct platform_driver hrt_driver = {
	.probe		= hrt_probe,
	.remove		= __devexit_p(hrt_remove),
	.driver		= {
		.name		= "led-driver",
		.owner		= THIS_MODULE,
		.of_match_table = led_of_match, /* for historical reasons, this will remain as 'led' in the device tree */
	},
};

static int __init hrt_init(void)
{
	return platform_driver_register(&hrt_driver);
}
module_init(hrt_init);

static void __exit hrt_exit(void)
{
	platform_driver_unregister(&hrt_driver);
}
module_exit(hrt_exit);

MODULE_AUTHOR("Steven Frick (sfrick@grandstream.com)");
MODULE_DESCRIPTION("HRT Driver");
MODULE_LICENSE("GPL");
MODULE_ALIAS("platform:led-driver");
