/*
 *  Keyboard driver for DSPG keypad controller
 *
 *  Copyright (c) 2012 DSPG Technologies GmbH, Andreas Weissel
 *
 *  Based on corgikbd.c, which is based on xtkbd.c/locomkbd.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/delay.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/rwsem.h>
#include <linux/input.h>
#include <linux/input/dspg-keypad.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/of_device.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include <linux/gpio.h>
#include <linux/gpio-dspg.h>

#define DSPG_KBD_CFG 0x00
#define DSPG_KBD_ROW 0x04
#define DSPG_KBD_COL 0x08

#define DSPG_KEYPAD_ACTIVATE_DELAY  20 /* us */
#define DSPG_KEYPAD_DEBOUNCE_DELAY  50 /* ms */

unsigned row_pinmap[MAX_ROW] =
{
    GPIO_PORTE(18),
    GPIO_PORTE(19),
    GPIO_PORTE(20),
    GPIO_PORTE(21),
    GPIO_PORTE(22),
    GPIO_PORTE(29),
    GPIO_PORTE(30),
    GPIO_PORTE(31),
    GPIO_PORTA(28),
    GPIO_PORTA(29)
};
static inline void
dspg_keypad_activate_col(struct dspg_keypad *dspg_keypad, int col)
{
	/* STROBE col -> High, not col -> HiZ */
	if ((1 << col) & dspg_keypad->config->cols)
		writel(1 << col, dspg_keypad->regs + DSPG_KBD_COL);
}

static inline void
dspg_keypad_activate_all(struct dspg_keypad *dspg_keypad)
{
	/* STROBE ALL -> High, against the column bitmask */
	writel(dspg_keypad->config->cols, dspg_keypad->regs + DSPG_KBD_COL);
}

static int
dspg_keypad_key_pressed(struct dspg_keypad *dspg_keypad)
{
	return (readl(dspg_keypad->regs + DSPG_KBD_ROW) &
		dspg_keypad->config->rows);
}

static void
dspg_keypad_setup_sense_lines_and_strobe_high(struct dspg_keypad *dspg_keypad)
{
	unsigned long val;

	val = dspg_keypad->config->cols | ( dspg_keypad->config->rows << 8 );
	writel(val, dspg_keypad->regs + DSPG_KBD_CFG);
}

/*
 * The dw keyboard generates interrupts as long as a key is pressed.
 * When a key is pressed, we disable the interrupt and enable a timer to
 * periodically can the keyboard to detect when the key is released.
 */

/* Scan the hardware keyboard and push any changes up through the input layer */
static void
dspg_keypad_scankeyboard(unsigned long data)
{
	struct dspg_keypad *dspg_keypad = (struct dspg_keypad *)data;
	unsigned int max_rows, max_cols, col, row, row_mask, scancode;
	unsigned int pressed, num_pressed = 0;
	struct list_head *entry;
	int ret;

	max_cols = fls(dspg_keypad->config->cols);
	max_rows = fls(dspg_keypad->config->rows);

	for (col = 0; col < max_cols; col++) {
		if ( ( 1 << col ) & dspg_keypad->config->cols ) {
			/*
			 * Discharge the output driver capacitance
			 * in the keyboard matrix. (Yes it is significant..)
			 */
			dspg_keypad_activate_col(dspg_keypad, col);
			udelay(DSPG_KEYPAD_ACTIVATE_DELAY);

			row_mask = dspg_keypad_key_pressed(dspg_keypad);


			for (row = 0; row < max_rows; row++) {
				if ( ( 1 << row ) & dspg_keypad->config->rows ) {
					scancode = MATRIX_SCAN_CODE(row, col, ROW_SHIFT);
					pressed = !!(row_mask & (1 << row));
					ret = 0;

					down_read(&dspg_keypad->handler_lock);
					list_for_each(entry, &dspg_keypad->handler_list) {
						struct dspg_keypad_handler *handler;

						handler = list_entry(entry, struct dspg_keypad_handler,
								     next_handler);
						ret |= handler->handler(handler, scancode,
									pressed);
					}
					up_read(&dspg_keypad->handler_lock);

					if (ret == 0) {
						input_report_key(dspg_keypad->input,
								 dspg_keypad->keycodes[scancode],
								 pressed);
					}
					if (pressed)
						num_pressed++;
				}
			}
		}
	}

	input_sync(dspg_keypad->input);

	dspg_keypad_activate_all(dspg_keypad);

	/* if any keys are pressed, enable the timer */
	if (num_pressed)
		mod_timer(&dspg_keypad->timer, jiffies +
		          msecs_to_jiffies(DSPG_KEYPAD_DEBOUNCE_DELAY));
	else {
		/* we finished the polling phase, so register for the interrupt */
		udelay(DSPG_KEYPAD_ACTIVATE_DELAY);
		dspg_keypad->irq_disabled = 0;
		enable_irq(dspg_keypad->irq);
	}
}

static irqreturn_t
dspg_keypad_interrupt(int irq, void *dev_id)
{
	struct dspg_keypad *dspg_keypad = (struct dspg_keypad *)dev_id;

	/* Scan the keyboard once. Further scans will be triggered by the
	 * timer. Keep the interrupt disabled in the meantime because the
	 * scanning will trigger it again. */
	disable_irq_nosync(dspg_keypad->irq);
	dspg_keypad->irq_disabled = 1;
	mod_timer(&dspg_keypad->timer, jiffies);

	return IRQ_HANDLED;
}

void
dspg_keypad_register(struct dspg_keypad *dspg_keypad,
		     struct dspg_keypad_handler *handler)
{
	down_write(&dspg_keypad->handler_lock);
	list_add_tail(&handler->next_handler, &dspg_keypad->handler_list);
	up_write(&dspg_keypad->handler_lock);
}
EXPORT_SYMBOL(dspg_keypad_register);

void
dspg_keypad_unregister(struct dspg_keypad *dspg_keypad,
		       struct dspg_keypad_handler *handler)
{
	down_write(&dspg_keypad->handler_lock);
	list_del(&handler->next_handler);
	up_write(&dspg_keypad->handler_lock);
}
EXPORT_SYMBOL(dspg_keypad_unregister);

#ifdef CONFIG_PM
static int dspg_keypad_suspend(struct platform_device *dev, pm_message_t state)
{
	struct dspg_keypad *dspg_keypad = platform_get_drvdata(dev);

	if (device_may_wakeup(&dev->dev))
		enable_irq_wake(dspg_keypad->irq);

	/*
	 * Prevent a race with the interrupt / scanning function.
	 */
	disable_irq(dspg_keypad->irq);
	del_timer_sync(&dspg_keypad->timer);
	if (dspg_keypad->irq_disabled) {
		enable_irq(dspg_keypad->irq);
		dspg_keypad->irq_disabled = 0;
	}
	enable_irq(dspg_keypad->irq);

	return 0;
}
static int dspg_keypad_resume(struct platform_device *dev)
{
	struct dspg_keypad *dspg_keypad = platform_get_drvdata(dev);

	if (device_may_wakeup(&dev->dev))
		disable_irq_wake(dspg_keypad->irq);

	return 0;
}
#else
#define dspg_keypad_suspend NULL
#define dspg_keypad_resume  NULL
#endif

struct input_dev *event_dev;

//=============================================================================
int
gs_inject_key_event(
	unsigned int event,
	unsigned int status
)
//=============================================================================
{
	if (event_dev == NULL )
		return -1;

	printk("gs_inject_key_event: event=%d; status=%d\n", event, status);
	local_irq_disable();
	input_report_key( event_dev, event, 1 ); // simulate key press
	input_sync( event_dev );
	input_report_key( event_dev, event, 0 ); // simulate key release
	input_sync( event_dev );
	local_irq_enable();

	return 0;
}
EXPORT_SYMBOL(gs_inject_key_event);

static struct matrix_keymap_data keymap_data;

#ifdef CONFIG_OF
static struct dspg_keypad_config * __init
dspg_keypad_parse_dt(struct platform_device *pdev)
{
	struct dspg_keypad_config *pdata;
	struct device_node *np = pdev->dev.of_node;
	uint32_t *keymap; 
	int i;
	u32 prop;

	if (!np)
		return NULL;

	pdata = kzalloc(sizeof(*pdata), GFP_KERNEL);
	if (!pdata)
		return NULL;

	if (of_property_read_u32(np, "dspg,rows", &prop)) {
		dev_err(&pdev->dev, "missing dspg,rows\n");
		goto err;
	}
	pdata->rows = ( prop << 1 ) | 1; // increase 1 row for fake key event

	if (of_property_read_u32(np, "dspg,cols", &prop)) {
		dev_err(&pdev->dev, "missing dspg,cols\n");
		goto err;
	}
	pdata->cols = prop;

	pdata->keymap_data = matrix_keyboard_of_fill_keymap(np, NULL);
	if (!pdata->keymap_data) {
		dev_err(&pdev->dev, "missing linux,keymap\n");
		goto err;
	}

	keymap = kzalloc((pdata->keymap_data->keymap_size + fls(pdata->cols))  * sizeof( uint32_t ), GFP_KERNEL);
	if (!keymap)
		return NULL;

	for ( i = 0 ; i < pdata->keymap_data->keymap_size; i++ )
	{
		keymap[i] = pdata->keymap_data->keymap[i];
	}

	keymap[i] = ( (fls(pdata->rows) - 1) << 24) |  KEY_EVENT_NET;	
	printk("keymap[%d] = %x\n", i, keymap[i]);	


	keymap_data.keymap_size = pdata->keymap_data->keymap_size +  1;
	keymap_data.keymap = keymap;

	matrix_keyboard_of_free_keymap(pdata->keymap_data);

	pdata->keymap_data = &keymap_data;

	return pdata;

err:
	kfree(pdata);
	return NULL;
}
#else
static inline struct dspg_keypad_config * __init
dspg_keypad_parse_dt(struct platform_device *pdev)
{
	return NULL;
}
#endif

static int __init
dspg_keypad_probe(struct platform_device *pdev)
{
	struct device_node *np = pdev->dev.of_node;
	struct dspg_keypad_config *pdata;
	struct dspg_keypad *dspg_keypad;
	struct input_dev *input_dev;
	struct resource *res;
	int ret;
    int i;

	dspg_keypad = kzalloc(sizeof(struct dspg_keypad), GFP_KERNEL);
	if (!dspg_keypad)
		return -ENOMEM;

	dspg_keypad->input = event_dev = input_dev = input_allocate_device();
	if (!input_dev) {
		ret = -ENOMEM;
		goto out_input_alloc_fail;
	}

	platform_set_drvdata(pdev, dspg_keypad);
	if (pdev->dev.of_node)
		pdata = dspg_keypad_parse_dt(pdev);
	else
		pdata = pdev->dev.platform_data;
	if (!pdata) {
		dev_err(&pdev->dev, "no platform data defined\n");
		ret = -EINVAL;
		goto out_no_pdata;
	}
	dspg_keypad->config = pdata;

    //disable internal pull-up on row pins
#if 0
    /* SCF - dvf99 needed this, it seems dvf9918 does not */
    for ( i = 0; i < MAX_ROW; i++ )
    {
        if ( dspg_keypad->config->rows & ( 1 << i ) )
        {
            dspg_gpio_set_pull_enable( row_pinmap[i], 0 );
        }
    }
#endif
    
	dspg_keypad->irq = platform_get_irq(pdev, 0);
	if (dspg_keypad->irq < 0) {
		dev_err(&pdev->dev, "no IRQ resource defined\n");
		ret = -ENODEV;
		goto out_no_irq;
	}

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (res == NULL) {
		dev_err(&pdev->dev, "no memory resource defined\n");
		ret = -ENODEV;
		goto out_no_regs;
	}

	dspg_keypad->mem = request_mem_region(res->start, resource_size(res),
	                            pdev->name);
	if (!dspg_keypad->mem) {
		dev_err(&pdev->dev, "cannot get register range\n");
		ret = EBUSY;
		goto out_mem_busy;
	}

	dspg_keypad->regs = ioremap(res->start, resource_size(res));
	if (dspg_keypad->regs == NULL) {
		dev_err(&pdev->dev, "failed to ioremap() registers\n");
		ret = -ENODEV;
		goto out_ioremap_failed;
	}

	/* Init Keyboard rescan timer */
	init_timer(&dspg_keypad->timer);
	dspg_keypad->timer.function = dspg_keypad_scankeyboard;
	dspg_keypad->timer.data = (unsigned long)dspg_keypad;

	input_dev->name = "DSPG keypad";
	input_dev->phys = pdev->name;
	input_dev->id.bustype = BUS_HOST;
	input_dev->id.vendor = 0x0001;
	input_dev->id.product = 0x0001;
	input_dev->id.version = 0x0100;
	input_dev->dev.parent = &pdev->dev;
	input_set_drvdata(input_dev, dspg_keypad);

	__set_bit(EV_SYN, input_dev->evbit);
	__set_bit(EV_KEY, input_dev->evbit);
	input_dev->keycode = dspg_keypad->keycodes;
	input_dev->keycodesize = sizeof(dspg_keypad->keycodes[0]);
	input_dev->keycodemax = fls(pdata->rows) << ROW_SHIFT;

	matrix_keypad_build_keymap(pdata->keymap_data, ROW_SHIFT,
			input_dev->keycode, input_dev->keybit);

	ret = input_register_device(dspg_keypad->input);
	if (ret < 0) {
		dev_err(&pdev->dev, "failed to register input device\n");
		goto out_register_failed;
	}

	/* Set Strobe lines as outputs - set high */
	dspg_keypad_setup_sense_lines_and_strobe_high(dspg_keypad);
	dspg_keypad_activate_all(dspg_keypad);

	/* Setup sense interrupts - RisingEdge Detect, sense lines as inputs */
	ret = request_irq(dspg_keypad->irq, dspg_keypad_interrupt, 0,
	                  "dspg_keypad", dspg_keypad);
	if (ret < 0) {
		dev_err(&pdev->dev, "cannot get irq\n");
		goto out_irq_request_failed;
	}

	init_rwsem(&dspg_keypad->handler_lock);
	INIT_LIST_HEAD(&dspg_keypad->handler_list);

	ret = of_platform_populate(np, NULL, NULL, &pdev->dev);
	if (ret < 0) {
		dev_err(&pdev->dev, "error populating subdevices\n");
		goto out_of_populate_failed;
	}

	/* Keypad is wakeup capable by default */
	device_init_wakeup(&pdev->dev, 1);

	return 0;

out_of_populate_failed:
	free_irq(dspg_keypad->irq, dspg_keypad);
out_irq_request_failed:
	input_unregister_device(dspg_keypad->input);
out_register_failed:
	iounmap(dspg_keypad->regs);
out_ioremap_failed:
	release_mem_region(dspg_keypad->mem->start, resource_size(dspg_keypad->mem));
out_mem_busy:
out_no_regs:
out_no_irq:
	if (!pdev->dev.platform_data) {
		matrix_keyboard_of_free_keymap(pdata->keymap_data);
		kfree(pdata);
	}
out_no_pdata:
	input_free_device(dspg_keypad->input);
out_input_alloc_fail:
	kfree(dspg_keypad);

	return ret;
}

static int __devexit
dspg_keypad_remove(struct platform_device *pdev)
{
	struct dspg_keypad *dspg_keypad = platform_get_drvdata(pdev);

	free_irq(dspg_keypad->irq, dspg_keypad);
	del_timer_sync(&dspg_keypad->timer);
	input_unregister_device(dspg_keypad->input);
	iounmap(dspg_keypad->regs);
	if (!pdev->dev.platform_data) {
		matrix_keyboard_of_free_keymap(dspg_keypad->config->keymap_data);
		kfree(dspg_keypad->config);
	}
	input_free_device(dspg_keypad->input);
	kfree(dspg_keypad);

	return 0;
}

static const struct of_device_id dspg_keypad_of_match[] = {
	{ .compatible = "dspg,keypad", },
	{ },
};
MODULE_DEVICE_TABLE(of, dspg_keypad_of_match);

static struct platform_driver dspg_keypad_driver = {
	.remove		= __devexit_p(dspg_keypad_remove),
	.suspend	= dspg_keypad_suspend,
	.resume		= dspg_keypad_resume,
	.driver		= {
		.name	= "dspg-keypad",
		.owner  = THIS_MODULE,
		.of_match_table = dspg_keypad_of_match,
	},
};

static int __init dspg_keypad_init(void)
{
	return platform_driver_probe(&dspg_keypad_driver, dspg_keypad_probe);
}

static void __exit dspg_keypad_exit(void)
{
	platform_driver_unregister(&dspg_keypad_driver);
}

module_init(dspg_keypad_init);
module_exit(dspg_keypad_exit);

MODULE_AUTHOR("Andreas Weissel <andreas.weissel@dspg.com>");
MODULE_DESCRIPTION("DSPG Keypad Driver");
MODULE_LICENSE("GPL");
