/*
 * DSPG DVF9A core driver including access to the DAIF interface
 *
 * Copyright (c) 2012, DSP Group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/err.h>
#include <linux/io.h>
#include <linux/irq.h>
#include <linux/irqdomain.h>
#include <linux/interrupt.h>
#include <linux/clk.h>
#include <linux/slab.h>
#include <linux/platform_device.h>
#include <linux/mfd/core.h>
#include <linux/mfd/dvf9a.h>
#include <linux/of.h>
#include <linux/of_device.h>
#include <linux/miscdevice.h>
#include <linux/fcntl.h>
#include <linux/fs.h>
#include <linux/io.h>
#include <linux/uaccess.h>
#include <linux/bitops.h>
#include <linux/delay.h>
#include <mach/platform.h>

#include <asm/unaligned.h>

struct dvf9a;

struct anaint {
	u16 hwirq;
	struct dvf9a *dvf9a;
	struct irq_domain *irq_domain;
};

struct dvf9a {
	void __iomem *regs;
	struct clk *pmu_clk;
	struct clk *apu_clk;
	spinlock_t lock;
	int irq;
	int owner;
	void (*handle)(struct dvf9a *);

	struct device *dev;
	struct resource *res;
	struct irq_domain *anaint_domain;
	struct anaint anaints[DVF9A_NUM_IRQS];
	int user_irq[DVF9A_NUM_IRQS];

	struct miscdevice miscdev;

	int irq_enable_cur; /* currently active value */
	int daif_indirect_end;
	int num_irqs;
	int version;
	int filterless;
};

struct dvf9a *the_dvf9a;
int dcin_enabled;

#define DVF9A				"dvf9a"

#define DAIF_WAPU_TRAN_MASK		(1<<15)
#define DAIF_RAPU_NEW_VALUE		(1<<15)

#define DAIF_DIRECT_MIRRORED_START	0
#define DAIF_DIRECT_MIRRORED_END	64
#define DAIF_DIRECT_RAPU_WAPU_START	65
#define DAIF_DIRECT_RAPU_WAPU_END	114

#define DAIF_RAPU			41  // 0x00a4
#define DAIF_WAPU			64  // 0x0100
#define DAIF_MODE			117 // 0x01d4
#define DAIF_PEND			118 // 0x01d8
#define DAIF_PENDB0			119 // 0x01dc
#define DAIF_PENDB1			120 // 0x01e0
#define DAIF_PENDB2			121 // 0x01e4
#define DAIF_PENDB3			122 // 0x01e8

#define DAIF_ANA_DATA_WR_LSB		65
#define DAIF_ANA_DATA_WR_MSB		66
#define DAIF_ANA_ADDR_WR		67
#define DAIF_ANA_ADDR_RD		68
#define DAIF_ANA_DATA_RD_LSB		69
#define DAIF_ANA_DATA_RD_MSB		70

#define DAIF_MODE_RAPU_PRIO		(1<<2)
#define DAIF_MODE_WAPU_PRIO		(1<<3)

/*
 * Reserved registers: 9, 12, 13, 27, 41, 45, 50, 52..64, 80, 90, 105,
 * 115..127.
 */
#define M(off) (1UL << ((off) & 31UL))
#define MM(bits, off) (((1UL << (bits)) - 1UL) << ((off & 31UL)))

static DECLARE_BITMAP(reserved_regs, DAIF_INDIRECT_END) = {
	/*  0..31  */ [0] = M(9) | MM(2, 12) | M(27),
	/* 32..63  */ [1] = M(41) | M(45) | M(50) | MM(12, 52),
	/* 64..95  */ [2] = M(64) | M(80) | M(90),
	/* 96..127 */ [3] = M(105) | MM(13, 115),
};

static int irq_to_ana_stat_en[DVF9A_NUM_IRQS] = {
	0, 1, 2, 3, 4, -1, 6, -1, -1, -1, -1, 7, -1
};

void
dvf9a_handle_ownership(struct dvf9a *dvf9a, void (*handle)(struct dvf9a *))
{
	dvf9a->handle = handle;
}
EXPORT_SYMBOL(dvf9a_handle_ownership);

void
dvf9a_acquire(struct dvf9a *dvf9a)
{
	if (dvf9a->owner)
		return;

	if (dvf9a->handle)
		dvf9a->handle(dvf9a);

	dvf9a->owner = 1;
}
EXPORT_SYMBOL(dvf9a_acquire);

void
dvf9a_set_owner(struct dvf9a *dvf9a, int owner)
{
	dvf9a->owner = !!owner;
}
EXPORT_SYMBOL(dvf9a_set_owner);


static int
daif_is_direct_mirrored(unsigned short reg)
{
	if ((reg <= DAIF_DIRECT_MIRRORED_END) ||
	   ((reg >= DAIF_MODE) && (reg <= DAIF_PENDB3)))
		return 1;
	return 0;
}

static int
daif_is_direct_rapu_wapu(unsigned short reg)
{
	if (reg - DAIF_DIRECT_RAPU_WAPU_START <= DAIF_DIRECT_RAPU_WAPU_END -
	          DAIF_DIRECT_RAPU_WAPU_START)
		return 1;
	return 0;
}

static int
daif_is_indirect_analog(unsigned short reg)
{
	if (reg >= DAIF_INDIRECT_START)
		return 1;
	return 0;
}

static unsigned short
daif_readw(struct dvf9a *dvf9a, unsigned short reg)
{
	return readw(dvf9a->regs + (reg * 4));
}

static void
daif_writew(struct dvf9a *dvf9a, unsigned short val, unsigned short reg)
{
	writew(val, dvf9a->regs + (reg * 4));
}

static void
daif_wait_wapu_tran(struct dvf9a *dvf9a)
{
	while (daif_readw(dvf9a, DAIF_PEND))
		;
}

static unsigned short
daif_read_via_rapu(struct dvf9a *dvf9a, unsigned short reg)
{
	unsigned short val;

	daif_writew(dvf9a, reg, DAIF_RAPU);

	while (!((val = daif_readw(dvf9a, DAIF_RAPU)) & DAIF_RAPU_NEW_VALUE))
		;

	/* Check if indicated register address (RAPU[14:8]) matches */
	if (((val & 0x7f00) >> 8) == reg)
		return val & 0xff;
	else
		return daif_read_via_rapu(dvf9a, reg); /* repeat request */
}

static void
daif_write_via_wapu(struct dvf9a *dvf9a, unsigned short reg, unsigned short value)
{
	daif_writew(dvf9a, ((reg & 0x7f) << 8) | (value & 0xff), DAIF_WAPU);
	daif_wait_wapu_tran(dvf9a);
}

static void
daif_direct_mirrored_init(struct dvf9a *dvf9a)
{
	unsigned short val, reg;

	/* enable "prev_replica" */
	daif_writew(dvf9a, 2, DAIF_MODE);

	for (reg = 0; reg < DAIF_DIRECT_MIRRORED_END; reg++) {
		if (reg == DAIF_RAPU)
			continue;
		val = daif_read_via_rapu(dvf9a, reg);
		udelay(10);
		daif_writew(dvf9a, val & 0xff, reg);
	}

	/* disable "prev_replica" */
	daif_writew(dvf9a, 0, DAIF_MODE);
}

unsigned short
daif_read_dcin(int dcin_number)
{
    unsigned short reg;

    if ( !dcin_enabled )
        return 1;

    reg = daif_read(the_dvf9a, DAIF_ICU_ANA_RAW);
    reg = ( reg >> dcin_number ) & 0x01;

    return reg;
}
EXPORT_SYMBOL(daif_read_dcin);

unsigned short
daif_read(struct dvf9a *dvf9a, unsigned short reg)
{
	unsigned short ret = 0;

    if (!irqs_disabled())
        spin_lock_bh(&dvf9a->lock);

	clk_enable(dvf9a->pmu_clk);
	clk_enable(dvf9a->apu_clk);

	dvf9a_acquire(dvf9a);

	if (daif_is_direct_mirrored(reg)) {
		ret = daif_readw(dvf9a, reg); /* immediate access */
	} else if (daif_is_direct_rapu_wapu(reg)) {
		ret = daif_read_via_rapu(dvf9a, reg);
	} else if (daif_is_indirect_analog(reg)) {
		/* set required address */
		reg -= DAIF_INDIRECT_START;
		daif_write_via_wapu(dvf9a, DAIF_ANA_ADDR_RD, reg >> 1);
		if (daif_readw(dvf9a, DAIF_MODE) & DAIF_MODE_RAPU_PRIO)
			daif_wait_wapu_tran(dvf9a);

		/* read value */
		ret  = daif_read_via_rapu(dvf9a, DAIF_ANA_DATA_RD_MSB) << 8;
		udelay(10);
		ret |= daif_read_via_rapu(dvf9a, DAIF_ANA_DATA_RD_LSB) & 0xff;
	}

    if (!irqs_disabled())
        spin_unlock_bh(&dvf9a->lock);

	clk_disable(dvf9a->apu_clk);
	clk_disable(dvf9a->pmu_clk);

	return ret;
}
EXPORT_SYMBOL(daif_read);

void
daif_write(struct dvf9a *dvf9a, unsigned short reg, unsigned short value)
{
    if (!irqs_disabled())
        spin_lock_bh(&dvf9a->lock);

	clk_enable(dvf9a->pmu_clk);
	clk_enable(dvf9a->apu_clk);

	dvf9a_acquire(dvf9a);

	if (dvf9a->filterless && (reg == DAIF_DCLS_ANA_CTRL)) {
		value &= ~0x40; /* disable compatibility mode */
		if (value & 0x01)
			value |= 0x1e; /* enable HI_INP and SHORT protection */
	}

	if (daif_is_direct_mirrored(reg)) {
		unsigned short pend_reg = reg >> 4; /* register number 0..3 */
		unsigned short pend_mask = reg - (pend_reg << 4);   /* 0..15 */
		pend_mask = 1 << pend_mask; /* bit mask */

		/* immediate access */
		while (daif_readw(dvf9a, DAIF_PENDB0 + pend_reg) & pend_mask)
			;
		daif_writew(dvf9a, value, reg);
		while (daif_readw(dvf9a, DAIF_PENDB0 + pend_reg) & pend_mask)
			;
	} else if (daif_is_direct_rapu_wapu(reg)) {
		daif_write_via_wapu(dvf9a, reg, value);
	} else if (daif_is_indirect_analog(reg)) {
		/* write value msb, lsb */
		daif_write_via_wapu(dvf9a, DAIF_ANA_DATA_WR_MSB, value >> 8);
		daif_write_via_wapu(dvf9a, DAIF_ANA_DATA_WR_LSB, value & 0xff);

		/* set required address */
		reg -= DAIF_INDIRECT_START;
		daif_write_via_wapu(dvf9a, DAIF_ANA_ADDR_WR, reg >> 1);
	}

	clk_disable(dvf9a->apu_clk);
	clk_disable(dvf9a->pmu_clk);

    if (!irqs_disabled())
        spin_unlock_bh(&dvf9a->lock);
}
EXPORT_SYMBOL(daif_write);

int
daif_translate_irq(struct dvf9a *dvf9a, int hw_irq)
{
	return irq_create_mapping(dvf9a->anaint_domain, hw_irq);
}
EXPORT_SYMBOL(daif_translate_irq);

static void
dvf9a_irq_mask(struct irq_data *data)
{
	struct anaint *anaint = irq_data_get_irq_chip_data(data);
	struct dvf9a *dvf9a = anaint->dvf9a;
	unsigned short val;

	dvf9a->irq_enable_cur &= ~(1 << anaint->hwirq);

	daif_write(dvf9a, DAIF_ICU_ANA_EN1, dvf9a->irq_enable_cur & 0x7f);
	if (irq_to_ana_stat_en[anaint->hwirq] != -1) {
		val = daif_read(dvf9a, DAIF_ICU_ANA_STAT_EN1);
		val &= ~(1 << irq_to_ana_stat_en[anaint->hwirq]);
		daif_write(dvf9a, DAIF_ICU_ANA_STAT_EN1, val);
	}
	daif_write(dvf9a, DAIF_ICU_ANA_EN2, dvf9a->irq_enable_cur >> 7);

	clk_disable(dvf9a->apu_clk);
	clk_disable(dvf9a->pmu_clk);
}

static void
dvf9a_irq_unmask(struct irq_data *data)
{
	struct anaint *anaint = irq_data_get_irq_chip_data(data);
	struct dvf9a *dvf9a = anaint->dvf9a;
	unsigned short val;

	clk_enable(dvf9a->pmu_clk);
	clk_enable(dvf9a->apu_clk);

	dvf9a->irq_enable_cur |= 1 << anaint->hwirq;

	daif_write(dvf9a, DAIF_ICU_ANA_EN1, dvf9a->irq_enable_cur & 0x7f);
	if (irq_to_ana_stat_en[anaint->hwirq] != -1) {
		val = daif_read(dvf9a, DAIF_ICU_ANA_STAT_EN1);
		val |= 1 << irq_to_ana_stat_en[anaint->hwirq];
		daif_write(dvf9a, DAIF_ICU_ANA_STAT_EN1, val);
	}
	daif_write(dvf9a, DAIF_ICU_ANA_EN2, dvf9a->irq_enable_cur >> 7);
}

static struct irq_chip dvf9a_irq_chip = {
	.name = DVF9A,
	.irq_mask = dvf9a_irq_mask,
	.irq_disable = dvf9a_irq_mask,
	.irq_unmask = dvf9a_irq_unmask,
};

static irqreturn_t
dvf9a_irq_thread(int irq, void *priv)
{
	struct dvf9a *dvf9a = priv;
	int i, pending, pend_misc = 0;

	pending = daif_read(dvf9a, DAIF_ICU_ANA_SRC1);
	if (!pending)
		/* if ANA_MISC_INT is not set, assume dclass short or high-input */
		pending |= 0x80;

	if (pending & 0x80) {
		pend_misc = daif_read(dvf9a, DAIF_ICU_ANA_STAT2);
		daif_write(dvf9a, DAIF_ICU_ANA_STAT2, pend_misc);
	}
	pending &= 0x7f;

	if (pending)
		daif_write(dvf9a, DAIF_ICU_ANA_STAT1, pending);

	pending |= pend_misc << 7;

	if (!pending)
		return IRQ_NONE;

	for (i = 0; pending != 0; i++, pending >>= 1) {
		if (!(pending & 1))
			continue;

		/* Check the mask again. Another handler/thread might have
		 * changed it in between. */
		if (dvf9a->irq_enable_cur & (1 << i))
			handle_nested_irq(irq_linear_revmap(dvf9a->anaint_domain, i));
	}

	return IRQ_HANDLED;
}

static int
dvf9a_irq_map(struct irq_domain *dom, unsigned int irq,
	     irq_hw_number_t hw_irq)
{
	struct dvf9a *dvf9a = dom->host_data;

	irq_set_chip_data(irq, &(dvf9a->anaints[hw_irq]));
	irq_set_chip(irq, &dvf9a_irq_chip);
	irq_set_nested_thread(irq, 1);
	set_irq_flags(irq, IRQF_VALID);

	return 0;
}

static struct irq_domain_ops dvf9a_irq_ops = {
	.map	= dvf9a_irq_map,
	.xlate	= irq_domain_xlate_onecell,
};

#define DAIF_BYTES DAIF_INDIRECT_END

static int
dvf9a_rw_validate(loff_t pos, size_t *len)
{
	if (pos >= the_dvf9a->daif_indirect_end)
		return 0;	/* Past EOF */

	if (*len > the_dvf9a->daif_indirect_end - pos)
		*len = the_dvf9a->daif_indirect_end - pos;

	/* make sure start is aligned */
	if (pos >= DAIF_INDIRECT_START && (pos & 1))
		return -EINVAL;

	/* make sure end is aligned */
	if ((pos + *len) >= DAIF_INDIRECT_START && ((pos + *len) & 1))
		return -EINVAL;

	return 1;
}

static loff_t
dvf9a_llseek(struct file *file, loff_t offset, int origin)
{
	switch (origin) {
	case 0:
		/* nothing to do */
		break;
	case 1:
		offset += file->f_pos;
		break;
	case 2:
		offset += the_dvf9a->daif_indirect_end;
		break;
	default:
		return -EINVAL;
	}

	/* be a bit more strict than POSIX because seek'ing beyond the end
	 * doesn't make sense in our case */
	if (offset < 0 || offset >= the_dvf9a->daif_indirect_end)
		return -EINVAL;

	/* ensure correct word alignment on indirect registers */
	if (offset >= DAIF_INDIRECT_START && (offset & 1))
		return -EINVAL;

	file->f_pos = offset;
	return offset;
}

static ssize_t
dvf9a_read(struct file *file, char __user *buf, size_t count, loff_t *ppos)
{
	unsigned char contents[DAIF_BYTES];
	unsigned i = *ppos;
	unsigned char *tmp;
	ssize_t ret;

	if ((ret = dvf9a_rw_validate(*ppos, &count)) <= 0)
		return ret;

	tmp = contents;
	while (count) {
		unsigned short val;

		if (test_bit(i, reserved_regs))
			val = 0;
		else
			val = daif_read(the_dvf9a, i);

		if (i >= DAIF_INDIRECT_START) {
			put_unaligned(val, (u16*)tmp);
			tmp += 2;
			i += 2;
			count -= 2;
		} else {
			*tmp = val;
			tmp++;
			i++;
			count--;
		}
	}

	if (copy_to_user(buf, contents, tmp - contents))
		return -EFAULT;

	*ppos = i;

	return tmp - contents;
}

static ssize_t
dvf9a_write(struct file *file, const char __user *buf, size_t count, loff_t *ppos)
{
	unsigned char contents[DAIF_BYTES];
	unsigned i = *ppos;
	unsigned char *tmp;
	ssize_t ret;

	if ((ret = dvf9a_rw_validate(*ppos, &count)) <= 0)
		return ret;

	if (copy_from_user(contents, buf, count))
		return -EFAULT;

	tmp = contents;
	while (count) {
		if (i >= DAIF_INDIRECT_START) {
			if (!test_bit(i, reserved_regs))
				daif_write(the_dvf9a, i, get_unaligned((u16*)tmp));
			i += 2;
			tmp += 2;
			count -= 2;
		} else {
			if (!test_bit(i, reserved_regs))
				daif_write(the_dvf9a, i, *tmp);
			i++;
			tmp++;
			count--;
		}
	}

	*ppos = i;

	return tmp - contents;
}

static irqreturn_t
dvf9a_user_irq_thread(int irq, void *priv)
{
	int hw_irq;
	struct dvf9a *dvf9a = priv;
	char irq_num[8], hw_num[10];
	char *envp[] = { irq_num, hw_num, NULL };

	for (hw_irq = 0; hw_irq < dvf9a->num_irqs; hw_irq++)
		if (dvf9a->user_irq[hw_irq] == irq)
			break;

	snprintf(irq_num, 8, "IRQ=%d", irq);
	snprintf(hw_num, 10, "HWIRQ=%d", hw_irq);
	kobject_uevent_env(&dvf9a->dev->kobj, KOBJ_CHANGE, envp);

	return IRQ_HANDLED;
}

static long
dvf9a_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	struct dvf9a *dvf9a = the_dvf9a;
	int irq;
    int dcin;
	int ret = 0;

	/* Check type and command number */
	if (_IOC_TYPE(cmd) != DVF9A_IOC_MAGIC)
		return -ENOIOCTLCMD;

	/* Check access direction */
	if (_IOC_DIR(cmd) & _IOC_READ)
		ret = !access_ok(VERIFY_WRITE,
				(void __user *)arg, _IOC_SIZE(cmd));
	if (ret == 0 && _IOC_DIR(cmd) & _IOC_WRITE)
		ret = !access_ok(VERIFY_READ,
				(void __user *)arg, _IOC_SIZE(cmd));
	if (ret)
		return -EFAULT;

	switch(cmd) {
	case DVF9A_IOC_REQ_IRQ:
		if (__get_user(irq, (int __user *)arg))
			return -EFAULT;
		if ((irq < 0) || (irq >= dvf9a->num_irqs) || dvf9a->user_irq[irq])
			ret = -EBUSY;

		dvf9a->user_irq[irq] = daif_translate_irq(dvf9a, irq);
		if (!ret)
			ret = request_threaded_irq(dvf9a->user_irq[irq],
						   NULL, dvf9a_user_irq_thread,
						   0, "dvf9a-user", dvf9a);
		if (ret)
			dvf9a->user_irq[irq] = 0;

		if (__put_user(ret, (int __user *)arg))
			return -EFAULT;

		return 0;
	case DVF9A_IOC_REL_IRQ:
		if (__get_user(irq, (int __user *)arg))
			return -EFAULT;
		if ((irq < 0) || (irq >= dvf9a->num_irqs) || !dvf9a->user_irq[irq])
			ret = -EINVAL;

		if (!ret)
			free_irq(daif_translate_irq(dvf9a, irq), dvf9a);
		if (!ret)
			dvf9a->user_irq[irq] = 0;

		return 0;
    case DVF9A_IOC_DCIN:
        if (__get_user(dcin, (int __user *)arg))
            return -EFAULT;
        if ((dcin < 0) || (dcin > 1))
            return -EINVAL;
        dcin_enabled = dcin;
        return 0;
	}
        

	return -ENOIOCTLCMD;
}

static const struct file_operations dvf9a_fops = {
	.owner		= THIS_MODULE,
	.llseek		= dvf9a_llseek,
	.read		= dvf9a_read,
	.write		= dvf9a_write,
	.unlocked_ioctl	= dvf9a_ioctl,
};

static int
dvf9a_probe(struct platform_device *pdev)
{
printk("<<<-- %s\n", __func__); 
	struct dvf9a *dvf9a;
	struct device_node *np = pdev->dev.of_node;
	struct resource *res;
	int i, ret = 0;
	unsigned long rate;

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res)
		return -EINVAL;

	dvf9a = kzalloc(sizeof *dvf9a, GFP_KERNEL);
	if (!dvf9a)
		return -ENOMEM;
	dvf9a->dev = &pdev->dev;

    dcin_enabled = 0;
	the_dvf9a = dvf9a;

	platform_set_drvdata(pdev, dvf9a);
	dvf9a->res = res;

	spin_lock_init(&dvf9a->lock);

	dvf9a->pmu_clk = clk_get(&pdev->dev, "pmu");
	if (IS_ERR(dvf9a->pmu_clk)) {
		dev_err(&pdev->dev, "cannot get clock\n");
		ret = PTR_ERR(dvf9a->pmu_clk);
		goto err_request_res;
	}
	clk_enable(dvf9a->pmu_clk);

	dvf9a->apu_clk = clk_get(&pdev->dev, "apu");
	if (IS_ERR(dvf9a->apu_clk)) {
		dev_err(&pdev->dev, "cannot get clock\n");
		ret = PTR_ERR(dvf9a->apu_clk);
		goto err_pmu_clk;
	}
	rate = clk_round_rate(dvf9a->apu_clk, 13824000);
	clk_set_rate(dvf9a->apu_clk, rate);
	clk_enable(dvf9a->apu_clk);

	dvf9a->irq = platform_get_irq(pdev, 0);
	if (dvf9a->irq < 0) {
		dev_err(&pdev->dev, "need an irq resource\n");
		ret = dvf9a->irq;
		goto err_apu_clk;
	}

	if (!request_mem_region(res->start, res->end - res->start, DVF9A)) {
		dev_err(&pdev->dev, "failed to request memory region\n");
		ret = -EBUSY;
		goto err_apu_clk;
	}

	dvf9a->regs = ioremap(res->start, res->end - res->start + 1);
	if (!dvf9a->regs) {
		dev_err(&pdev->dev, "failed to map registers\n");
		ret = -ENOMEM;
		goto err_ioremap;
	}

	daif_direct_mirrored_init(dvf9a);

	dvf9a->daif_indirect_end = DAIF_INDIRECT_END;
	dvf9a->num_irqs = DVF9A_NUM_IRQS;
	dvf9a->version = daif_read(dvf9a, DAIF_HW_REVISION);
	if (dvf9a->version < 2) {
		dvf9a->daif_indirect_end = DAIF_AFEICTRL2+2;
		dvf9a->num_irqs = 11;
	}
	dvf9a->filterless = (dvf_get_chip_rev() > 1) &&
			    of_property_read_bool(np, "filterless-mode");
	if (dvf9a->filterless) {
		/* New overlap mode 5 (ENN), 5 (ENP).
		 * Limit current to 740mA (0x4) (spare for diff PVT) */
		daif_write(dvf9a, DAIF_DCLASS_ANA_CFG, 0x5da4);

		/* Set HI_INP_DELAY_DET to 0xf and DCLS_SHRT_DET_THRS to 0x0 */
		daif_write(dvf9a, DAIF_DCLS_SWCH_CFG, 0x0f30);

		/* Set HI_INP_ON_THRS to 0x1 */
		daif_write(dvf9a, DAIF_DCLS_HI_INP_THRS, 0x0427);
	}

	/* Setup IRQs */
	daif_write(dvf9a, DAIF_ICU_INT_POL,      0x0000);
	daif_write(dvf9a, DAIF_ICU_ANA_EN1,      0x0000);
	daif_write(dvf9a, DAIF_ICU_ANA_EN2,      0x0000);
	daif_write(dvf9a, DAIF_ICU_ANA_STAT_EN1, 0x0000);
	daif_write(dvf9a, DAIF_ICU_ANA_STAT1,    0x00ff);
	daif_write(dvf9a, DAIF_ICU_ANA_STAT2,    0x00ff);

	/* PWR interrupts are not (yet) supported */
	daif_write(dvf9a, DAIF_ICU_PWR_EN,       0x0000);
	daif_write(dvf9a, DAIF_ICU_PWR_STAT_EN,  0x0000);
	daif_write(dvf9a, DAIF_ICU_PWR_STAT,     0x00ff);

	clk_disable(dvf9a->apu_clk);
	clk_disable(dvf9a->pmu_clk);

	dvf9a->anaint_domain = irq_domain_add_linear(np, dvf9a->num_irqs,
						    &dvf9a_irq_ops, dvf9a);
	if (!dvf9a->anaint_domain) {
		dev_err(&pdev->dev, "cannot register IRQ domain!\n");
		ret = -ENOMEM;
		goto err_irq_domain;
	}

	for (i = 0; i < dvf9a->num_irqs; i++) {
		dvf9a->anaints[i].dvf9a = dvf9a;
		dvf9a->anaints[i].irq_domain = dvf9a->anaint_domain;
		dvf9a->anaints[i].hwirq = i;
	}

	ret = request_threaded_irq(dvf9a->irq, NULL, dvf9a_irq_thread,
				   IRQF_TRIGGER_HIGH | IRQF_ONESHOT,
				   DVF9A, dvf9a);
	if (ret < 0) {
		dev_err(&pdev->dev, "request irq %d failed\n", dvf9a->irq);
		goto err_irq_domain;
	}

	ret = of_platform_populate(np, NULL, NULL, &pdev->dev);
	if (ret < 0) {
		dev_err(&pdev->dev, "error populating subdevices\n");
		goto err_of_populate;
	}

	dvf9a->miscdev.minor = MISC_DYNAMIC_MINOR;
	dvf9a->miscdev.name = DVF9A;
	dvf9a->miscdev.fops = &dvf9a_fops;
	dvf9a->miscdev.parent = &pdev->dev;
	ret = misc_register(&dvf9a->miscdev);
	if (ret) {
		dev_err(&pdev->dev, "can't misc_register\n");
		goto err_of_populate;
	}

	dev_info(&pdev->dev, "successfully registered (DVF9A version %d)\n",
		 dvf9a->version);

printk("--> %s\n", __func__); 
	return 0;

err_of_populate:
	free_irq(dvf9a->irq, dvf9a);
err_irq_domain:
	iounmap(dvf9a->regs);
err_ioremap:
	release_resource(dvf9a->res);
err_apu_clk:
	clk_disable(dvf9a->apu_clk);
	clk_put(dvf9a->apu_clk);
err_pmu_clk:
	clk_disable(dvf9a->pmu_clk);
	clk_put(dvf9a->pmu_clk);
err_request_res:
	kfree(dvf9a);

	return ret;
}

static int __exit
dvf9a_remove(struct platform_device *dev)
{
	struct dvf9a *dvf9a = platform_get_drvdata(dev);

	misc_deregister(&dvf9a->miscdev);
	free_irq(dvf9a->irq, dvf9a);
	iounmap(dvf9a->regs);
	release_resource(dvf9a->res);
	clk_disable(dvf9a->pmu_clk);
	clk_put(dvf9a->pmu_clk);
	clk_disable(dvf9a->apu_clk);
	clk_put(dvf9a->apu_clk);
	kfree(dvf9a);

	return 0;
}

static struct of_device_id dvf9a_of_ids[] = {
	{ .compatible = "dspg,dvf9a" },
	{ },
};

static struct platform_driver dvf9a_platform_driver = {
	.driver = {
		.name	= DVF9A,
		.owner	= THIS_MODULE,
		.of_match_table = dvf9a_of_ids,
	},
	.probe		= dvf9a_probe,
	.remove		= __exit_p(dvf9a_remove),
};

static int __init
dvf9a_init(void)
{
	int retval = 0;

	retval = platform_driver_register(&dvf9a_platform_driver);
	return retval;
}

static void __exit
dvf9a_exit(void)
{
	platform_driver_unregister(&dvf9a_platform_driver);
}

module_init(dvf9a_init);
module_exit(dvf9a_exit);

MODULE_DESCRIPTION("DSPG DVF9A driver");
MODULE_LICENSE("GPL v2");
MODULE_AUTHOR("DSP Group");
MODULE_ALIAS("platform:dvf9a-core");
