/*
 * DVF9a PWM Driver
 *
 * (C) 2013 Grandstream Networks, Inc.
 * Author: Steven Frick
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/device.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/of.h>
#include <linux/of_device.h>
#include <linux/slab.h>
#include <linux/mfd/dvf9a.h>

/* Rsense = 5 ohm */
#define LCSENSE 0x80
/* Rref = 1000 ohm */
#define LCREF 0x04


struct dvf9a_pwm {
    struct dvf9a *dvf9a;
    int pwm;
    int khz;
};

static struct dvf9a_pwm *pwm_drv;

void dvf9a_pwm_set_duty(unsigned int duty)
{
    unsigned int value;

    if (pwm_drv == NULL) {
		return;
	}

#if defined( CONFIG_MFD_DSPG_DVF9A_PWM_BACKLIGHT_PWM )
    if (duty > 100)
        duty = 100;
    
    duty = 100 - duty;
    value = ((duty * 155)/100) + 100;	//Map duty value to range [100:255]
    daif_write(pwm_drv->dvf9a, DAIF_PWM1_DUTY, value); /* PWM1_DUTY */
#elif defined( CONFIG_MFD_DSPG_DVF9A_PWM_BACKLIGHT_CURRENT_CONTROL )
    /* map duty to current scale - max value of 9 -> 18mA */
    value = ( 9 * duty * 1000 / ( 100 * 1000 ) );
    value |= ( LCSENSE );
    daif_write(pwm_drv->dvf9a, DAIF_LC1CFG1, value);
#endif
}
EXPORT_SYMBOL(dvf9a_pwm_set_duty);

void dvf9a_pwm_start()
{
	if (pwm_drv == NULL) {
		return;
    }

    printk("%s\n", __func__);

#if defined( CONFIG_MFD_DSPG_DVF9A_PWM_BACKLIGHT_PWM )
    // Enabling and disabling of PWM feedbacks before and after enabling of PWM is needed
    // to fix black-screen error on GXP2130.
    daif_write(pwm_drv->dvf9a, DAIF_PWM1_CFG1, 0x0F); /* PWMFB_ENx: Enable PWM feedbacks */
    
    // Set PWM-out Hi-Z when disabled; Enable PWM; Set PWM Modulo 256.
    daif_write(pwm_drv->dvf9a, DAIF_PWM1_CFG2, (1 << 6) | (1 << 5) | (7 << 2)); 

    daif_write(pwm_drv->dvf9a, DAIF_PWM1_CFG1, 0x00); /* PWMFB_ENx: Disable PWM feedbacks */
#elif defined( CONFIG_MFD_DSPG_DVF9A_PWM_BACKLIGHT_CURRENT_CONTROL )

#endif
}
EXPORT_SYMBOL(dvf9a_pwm_start);

void dvf9a_pwm_stop()
{
	if (pwm_drv == NULL) {
		return;
	}

    printk("%s\n", __func__);

    daif_write(pwm_drv->dvf9a, DAIF_PWM1_CFG2, 0); /* PWM1_CFG2: Disable PWM */
    daif_write(pwm_drv->dvf9a, DAIF_PWM1_CFG2, (1 << 6)); /* Set PWM-out Hi-Z */
}
EXPORT_SYMBOL(dvf9a_pwm_stop);

static int __devinit dvf9a_pwm_probe(struct platform_device *pdev)
{
    struct dvf9a_pwm *pwm;
    struct device_node *np;
    uint32_t val;
    int ret;

    printk("%s\n", __func__);

    pwm = kzalloc(sizeof(*pwm), GFP_KERNEL);
    if (!pwm)
        return -ENOMEM;

    pwm->dvf9a = dev_get_drvdata(pdev->dev.parent);
	pwm_drv = pwm;
	
    np = pdev->dev.of_node;

    ret = of_property_read_u32(np, "dspg,id", &val);
    if (ret < 0 || val < 0 || val > 1) {
        dev_err(&pdev->dev, "invalid or missing 'dspg,id' property!\n");
        return ret;
    }
    pwm->pwm = val;

    ret = of_property_read_u32(np, "dspg,khz", &val);
    if (ret < 0 || val < 1 || val > 25) {
        dev_err(&pdev->dev, "invalid or missing 'dspg,khz' property!\n");
        return ret;
    }
    pwm->khz = val;

    platform_set_drvdata(pdev, pwm);

    /* Turn on bandgap */
    /* Note: Use DAIF, HW bug is present when accessing via memory mapped address */
    daif_write(pwm->dvf9a, DAIF_AUXEN, 0x01);

#if defined( CONFIG_MFD_DSPG_DVF9A_PWM_BACKLIGHT_PWM )
    // Configure PWM1.
    daif_write(pwm->dvf9a, DAIF_LC1CFG2, 3); /* LCMODE (LED1/PWM1): Set PWM mode */
    // PWM1 needs to be disabled first before writing settings.
    daif_write(pwm_drv->dvf9a, DAIF_PWM1_CFG2, 0); /* PWM1_CFG2: Disable PWM */
    daif_write(pwm_drv->dvf9a, DAIF_PWM1_CFG3, 2); /* Prescaler */
    daif_write(pwm_drv->dvf9a, DAIF_PWM1_CFG1, 0x00); /* PWMFB_ENx: Disable PWM feedbacks */
#elif defined( CONFIG_MFD_DSPG_DVF9A_PWM_BACKLIGHT_CURRENT_CONTROL )
    // Configure Current Control

    daif_write(pwm->dvf9a, DAIF_LC1CFG1, LCSENSE ); /* Idac = 0 */
    daif_write(pwm->dvf9a, DAIF_LC1CFG2, 0x01 | LCREF ); /* Set current control mode */
    daif_write(pwm->dvf9a, DAIF_LED_CTRL, 0x00 ); /* Pin is high impedence when off */
#endif
    
    dvf9a_pwm_start();      // Start LCD backlight
    dvf9a_pwm_set_duty(70); // Set LCD brightness to 70 during bootup
    
    return 0;
}

static int __devexit dvf9a_pwm_remove(struct platform_device *pdev)
{
    struct dvf9a_pwm *pwm = platform_get_drvdata(pdev);

    kfree(pwm);

    return 0;
}

static struct of_device_id dvf9a_pwm_of_ids[] = {
    { .compatible = "dspg,dvf9a-pwm" },
    { },
};

static struct platform_driver dvf9a_pwm_driver = {
    .driver = {
        .name = "dvf9a-pwm",
        .of_match_table = dvf9a_pwm_of_ids,
    },
    .probe = dvf9a_pwm_probe,
    .remove = __devexit_p(dvf9a_pwm_remove),
};

module_platform_driver(dvf9a_pwm_driver);

MODULE_AUTHOR("Steven Frick (sfrick@grandstream.com)");
MODULE_DESCRIPTION("DVF99 PWM Driver");
MODULE_LICENSE("GPL");
MODULE_ALIAS("platform:dvf9a-pwm");

