/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright © 2010 Atheros Communications, Inc.,  All Rights Reserved.
 *
 * Modified by Grandstream for dvf99
 *
 */
//=================
// Includes
//=================
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/phy.h>
#include "athrs17_phy.h"
#include "stmmac.h"

//=================
// Defines
//=================
/* for PHY related debugging messages */
#undef S17_PHY_DEBUG
#ifdef S17_PHY_DEBUG
#define DPRINTF_PHY(_fmt,...) do {         \
                printk(MODULE_NAME":"_fmt, __VA_ARGS__);      \
} while (0)
#define DPRINTF_PHY2(_fmt,...)
#else
#define DPRINTF_PHY(_fmt,...)
#define DPRINTF_PHY2(_fmt,...)
#endif

#define S17_WAN_PORT    5
#define S17_WAN_VID 2
#define S17_LAN_VID 1

#define DRV_LOG(DBG_SW, X0, X1, X2, X3, X4, X5, X6)
#define DRV_MSG(x,a,b,c,d,e,f)
#define DRV_PRINT(DBG_SW,X)

#define S17_LAN_PORT_VLAN          S17_LAN_VID
#define S17_WAN_PORT_VLAN         S17_WAN_VID

#define ENET_UNIT_GE1 1
#define ENET_UNIT_GE0 0 /* Connected to the switch */

#define TRUE    1
#define FALSE   0

#define S17_PHY0_ADDR   0x0
#define S17_PHY1_ADDR   0x1
#define S17_PHY2_ADDR   0x2
#define S17_PHY3_ADDR   0x3
#define S17_PHY4_ADDR   0x4
#define S17_IND_PHY 4

#define MODULE_NAME "ATHRS17"

#define S17_PHY_MAX 5

/* Range of valid PHY IDs is [MIN..MAX] */
#define S17_ID_MIN 0
#define S17_ID_MAX (S17_PHY_MAX-1)

/* Convenience macros to access myPhyInfo */
#define S17_IS_ENET_PORT(phyUnit) (athrPhyInfo[phyUnit].isEnetPort)
#define S17_IS_PHY_ALIVE(phyUnit) (athrPhyInfo[phyUnit].isPhyAlive)
#define S17_ETHUNIT(phyUnit) (athrPhyInfo[phyUnit].ethUnit)
#define S17_PHYBASE(phyUnit) (athrPhyInfo[phyUnit].phyBase)
#define S17_PHYADDR(phyUnit) (athrPhyInfo[phyUnit].phyAddr)


//=================
// Typedefs
//=================
/* PHY selections and access functions */
typedef enum
{
    PHY_SRCPORT_INFO,
    PHY_PORTINFO_SIZE,
} PHY_CAP_TYPE;

typedef enum
{
    PHY_SRCPORT_NONE,
    PHY_SRCPORT_VLANTAG,
    PHY_SRCPORT_TRAILER,
} PHY_SRCPORT_TYPE;

/*
 * Track per-PHY port information.
 */
typedef struct
{
    BOOL   isEnetPort;       /* normal enet port */
    BOOL   isPhyAlive;       /* last known state of link */
    int    ethUnit;          /* MAC associated with this phy port */
    uint32_t phyBase;
    uint32_t phyAddr;          /* PHY registers associated with this phy port */
    uint32_t VLANTableSetting; /* Value to be written to VLAN table */
#ifdef S17_8023AZ_FEATURE
    int azFeature;
#endif
} athrPhyInfo_t;

//=================
// Locals
//=================

 /*
  * Private Data
  */
static struct net_device* netdev;
/*
 * Per-PHY information, indexed by PHY unit number
 * MAC port 0 - CPU port 0x100
 * All ports are connected to GE0 of the SoC MAC
 * LAN/WAN seperation by VLAN tags (port 0-3 with VLANID 1; port 4 with VLANID 2
 */
static athrPhyInfo_t athrPhyInfo[] =
{
    {
        TRUE,   /* phy port 0 -- MAC port 1 0x200 */
        FALSE,
        ENET_UNIT_GE1,
        0,
        S17_PHY0_ADDR,
        S17_LAN_PORT_VLAN
    },

    {
        TRUE,   /* phy port 1 -- MAC port 2 0x300 */
        FALSE,
        ENET_UNIT_GE1,
        0,
        S17_PHY1_ADDR,
        S17_LAN_PORT_VLAN
    },

    {
        TRUE,   /* phy port 2 -- MAC port 3 0x400 */
        FALSE,
        ENET_UNIT_GE1,
        0,
        S17_PHY2_ADDR,
        S17_LAN_PORT_VLAN
    },

    {
        TRUE,   /* phy port 3 -- MAC port 4 0x500 */
        FALSE,
        ENET_UNIT_GE1,
        0,
        S17_PHY3_ADDR,
        S17_LAN_PORT_VLAN
    },

    {
        TRUE,   /* phy port 4 -- WAN port or MAC port 5 0x600 */
        FALSE,
        ENET_UNIT_GE0,
        0,
        S17_PHY4_ADDR,
        S17_WAN_PORT_VLAN   /* set as WAN port */
    },

    {
        FALSE,  /* phy port 5 -- CPU port (no RJ45 connector) */
        TRUE,
        ENET_UNIT_GE0,
        0,
        0x00,
        S17_LAN_PORT_VLAN    /* Send to all ports */
    },
};
static const athr_switch_intf_t default_rgmii_only =
{
    .mac0_type      = RGMII,
    .mac6_type      = NOT_CONNECT,
    .mac5_type      = DONT_CARE,
    .wan_port       = S17_WAN_PORT,
    .vlan_id        = 0,
    .wan2_port      = 0xff,
    .use_vlan_table = 0,
};

static const athr_switch_intf_t default_rgmii_sgmii =
{
    .mac0_type      = SGMII,
    .mac6_type      = RGMII,
    .mac5_type      = DONT_CARE,
    .wan_port       = S17_WAN_PORT,
    .vlan_id        = 1,
    .wan2_port      = 0xff,
    .use_vlan_table = 0,
};

static const athr_switch_intf_t default_sgmii_only =
{
    .mac0_type      = SGMII,
    .mac6_type      = NOT_CONNECT,
    .mac5_type      = DONT_CARE,
    .wan_port       = S17_WAN_PORT,
    .vlan_id        = 1,
    .wan2_port      = 0xff,
    .use_vlan_table = 0,
};

static const athr_switch_intf_t default_rgmii_rgmii =
{
    .mac0_type      = RGMII,
    .mac6_type      = RGMII,
    .mac5_type      = DONT_CARE,
    .wan_port       = S17_WAN_PORT,
    .vlan_id        = 1,
    .wan2_port      = 0xff,
    .use_vlan_table = 0,
};

 /*
  * Private Functions
  */
static uint16_t phy_reg_read(int dev_id, uint32_t phy_addr, uint8_t reg);
static void phy_reg_write(int dev_id, uint32_t phy_addr, uint8_t reg, uint16_t data);
static void athrs17_reg_rmw(unsigned int reg_addr, unsigned int reg_val);

//=================
// Functions
//=================
//=============================================================================
void
athrs17_phy_init(
    struct net_device *ndev
)
//=============================================================================
{   
    netdev = ndev;
}

//=============================================================================
static uint16_t
phy_reg_read(
    int dev_id,
    uint32_t phy_addr,
    uint8_t reg
)
//=============================================================================
{   
    uint16_t data = 0;
    struct stmmac_priv *priv = netdev_priv(netdev);
    struct mii_bus *miibus = priv->mii;

    data = miibus->read(miibus, (unsigned char)phy_addr, reg);
    return data;
}

//=============================================================================
static void
phy_reg_write(
    int dev_id,
    uint32_t phy_addr,
    uint8_t reg,
    uint16_t data
)
//=============================================================================
{   
    struct stmmac_priv *priv = netdev_priv(netdev);
    struct mii_bus *miibus = priv->mii;

    miibus->write(miibus, (unsigned char)phy_addr, reg, data);
}

//=============================================================================
void
athrs17_pc_port_enable(
    void
)
//=============================================================================
{
    athrs17_reg_write(S17_P2STATUS_REG, S17_PORT_STATUS_DEFAULT);
    printk("PC port enabled.\n");
}

//=============================================================================
void
athrs17_pc_port_disable(
    void
)
//=============================================================================
{
    athrs17_reg_write(S17_P2STATUS_REG, 0);
    printk("PC port disabled.\n");
}

//=============================================================================
void
athrs17_pc_port_mirror_enable(
    void
)
//=============================================================================
{
    unsigned int value;

    // Set mirror port number 2.
    value = athrs17_reg_read(S17_GLOFW_CTRL0_REG);
    value &= ~(0x0F << S17_MIRROR_PORT_SHIFT);
    athrs17_reg_write( S17_GLOFW_CTRL0_REG, value | (2 << S17_MIRROR_PORT_SHIFT) );

    // Enable mirror for incoming packets to port 0.
    value = athrs17_reg_read(S17_P0LOOKUP_CTRL_REG);
    athrs17_reg_write( S17_P0LOOKUP_CTRL_REG, value | (1 << 25) );

    // Enable mirror for outgoing packets from port 0.
    value = athrs17_reg_read(S17_PORT0_HOL_CTRL1);
    athrs17_reg_write( S17_PORT0_HOL_CTRL1, value | (1 << 16) );

    printk("PC port mirror enabled.\n");
}

//=============================================================================
void 
athrs17_pc_port_mirror_disable(
    void
)
//=============================================================================
{
    unsigned int value;

    // Disable mirror for incoming packets to port 0.
    value = athrs17_reg_read(S17_P0LOOKUP_CTRL_REG);
    athrs17_reg_write( S17_P0LOOKUP_CTRL_REG, value & ~(1 << 25) );

    // Disable mirror for outgoing packets from port 0.
    value = athrs17_reg_read(S17_PORT0_HOL_CTRL1);
    athrs17_reg_write( S17_PORT0_HOL_CTRL1, value & ~(1 << 16) );

    printk("PC port mirror disabled.\n");
}

//=============================================================================
static int
athrs17_port_status(
    int port
)
// Description: Check status of the LAN port
//=============================================================================
{
    static int last_status = 0;
    uint16_t phyHwStatus, phyHwControl;
    uint32_t phyBase;
    uint32_t phyAddr;

    phyBase = S17_PHYBASE( port );
    phyAddr = S17_PHYADDR( port );

    if ( last_status )
    {
        /* link was previously up */
        phyHwStatus = phy_reg_read( phyBase, phyAddr, S17_PHY_SPEC_STATUS );

        /* See if we've lost link */
        if ( !( phyHwStatus & S17_STATUS_LINK_PASS ) )
        {
            /* Flush the port from ARL entry */
            athrs17_reg_rmw( S17_ATU_FUNC_REG, ( port << 8 ) );

            last_status = 0;
        }
    }
    else
    {
        /* last known link status was DEAD */
        /* Check for reset complete */
        phyHwStatus = phy_reg_read( phyBase, phyAddr, S17_PHY_STATUS );
        if ( !S17_RESET_DONE( phyHwStatus ) )
        {
            return 0;
        }

        phyHwControl = phy_reg_read( phyBase, phyAddr, S17_PHY_CONTROL );
        /* Check for AutoNegotiation complete */
        if ( ( !( phyHwControl & S17_CTRL_AUTONEGOTIATION_ENABLE ) ) || S17_AUTONEG_DONE( phyHwStatus ) )
        {
            phyHwStatus = phy_reg_read( phyBase, phyAddr, S17_PHY_SPEC_STATUS );

            if ( phyHwStatus & S17_STATUS_LINK_PASS )
            {
                last_status = 1;
            }
        }
    }

    return last_status;
}

//=============================================================================
int
athrs17_lan_status(
    void
)
//=============================================================================
{
    return athrs17_port_status( 2 );
}

//=============================================================================
int
athrs17_pc_status(
    void
)
//=============================================================================
{
    return athrs17_port_status( 1 );
}

//=============================================================================
unsigned int
athrs17_reg_read(
    unsigned int reg_addr
)
//=============================================================================
{
    uint32_t reg_word_addr;
    uint32_t phy_addr, tmp_val, reg_val;
    uint16_t phy_val;
    uint8_t phy_reg;

    /* change reg_addr to 16-bit word address, 32-bit aligned */
    reg_word_addr = (reg_addr & 0xfffffffc) >> 1;

    /* configure register high address */
    phy_addr = 0x18;
    phy_reg = 0x0;
    phy_val = (uint16_t) ((reg_word_addr >> 8) & 0x3ff);  /* bit16-8 of reg address */
    phy_reg_write(0, phy_addr, phy_reg, phy_val);

    /* For some registers such as MIBs, since it is read/clear, we should */
    /* read the lower 16-bit register then the higher one */

    /* read register in lower address */
    phy_addr = 0x10 | ((reg_word_addr >> 5) & 0x7); /* bit7-5 of reg address */
    phy_reg = (uint8_t) (reg_word_addr & 0x1f);   /* bit4-0 of reg address */
    reg_val = (uint32_t) phy_reg_read(0, phy_addr, phy_reg);

    /* read register in higher address */
    reg_word_addr++;
    phy_addr = 0x10 | ((reg_word_addr >> 5) & 0x7); /* bit7-5 of reg address */
    phy_reg = (uint8_t) (reg_word_addr & 0x1f);   /* bit4-0 of reg address */
    tmp_val = (uint32_t) phy_reg_read(0, phy_addr, phy_reg);
    reg_val |= (tmp_val << 16);

    return reg_val;
}

//=============================================================================
void
athrs17_reg_write(
    unsigned int reg_addr, 
    unsigned int reg_val
)
//=============================================================================
{
    uint32_t reg_word_addr;
    uint32_t phy_addr;
    uint16_t phy_val;
    uint8_t phy_reg;

    /* change reg_addr to 16-bit word address, 32-bit aligned */
    reg_word_addr = (reg_addr & 0xfffffffc) >> 1;

    /* configure register high address */
    phy_addr = 0x18;
    phy_reg = 0x0;
    phy_val = (uint16_t) ((reg_word_addr >> 8) & 0x3ff);  /* bit16-8 of reg address */
    phy_reg_write(0, phy_addr, phy_reg, phy_val);

    /* For S17 registers such as ARL and VLAN, since they include BUSY bit */
    /* in higher address, we should write the lower 16-bit register then the */
    /* higher one */

    /* write register in lower address */
    phy_addr = 0x10 | ((reg_word_addr >> 5) & 0x7); /* bit7-5 of reg address */
    phy_reg = (uint8_t) (reg_word_addr & 0x1f);   /* bit4-0 of reg address */
    phy_val = (uint16_t) (reg_val & 0xffff);
    phy_reg_write(0, phy_addr, phy_reg, phy_val);

    /* read register in higher address */
    reg_word_addr++;
    phy_addr = 0x10 | ((reg_word_addr >> 5) & 0x7); /* bit7-5 of reg address */
    phy_reg = (uint8_t) (reg_word_addr & 0x1f);   /* bit4-0 of reg address */
    phy_val = (uint16_t) ((reg_val >> 16) & 0xffff);
    phy_reg_write(0, phy_addr, phy_reg, phy_val);
}

//=============================================================================
static void
athrs17_reg_rmw(
    unsigned int reg_addr, 
    unsigned int reg_val
)
//=============================================================================
{
    reg_val |= athrs17_reg_read(reg_addr);
    athrs17_reg_write(reg_addr,reg_val);
}

/* EOF */

