/*
 * drivers/rtc/rtc-dspg.c - driver for the RTC block of the DP52 chip
 *
 * Copyright (C) 2011 DSPG Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/delay.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/rtc.h>
#include <linux/slab.h>
#include <linux/time.h>
#include <linux/io.h>
#include <linux/of.h>
#include <linux/clk.h>

static char dspg_rtc_name[] = "dspg-rtc";

#define DSPG_RTC_CFG		0x00
#define DSPG_RTC_ALARMVAL	0x08
#define DSPG_RTC_STATUS		0x0C
#define DSPG_RTC_INTCLR		0x10
#define DSPG_RTC_TIME		0x14

#define RTC_CFG_ALARMEN		(3 << 0)
#define RTC_STATUS_WRITEVALID	(1 << 1)
#define RTC_STATUS_ALARM	(1 << 0)
#define RTC_INTCLR_ALARM	(1 << 0)

struct dspg_rtc {
	struct rtc_device *dev;
	void *base;
	struct clk *clk;
	unsigned long rate;
	int irq;

	/*
	 * Those two variables build the foundation of this driver. Since the
	 * rtc only features a counter which cannot be reset and which value is
	 * unknown at system boot, we need a fixed point in time and a
	 * corresponding rtc counter sample to be able to calculate the current
	 * time. This pair is set via the settime call.
	 *
	 * rtc_reference: the rtc counter sample (one tick is one second)
	 * rtc_time:      the time at the rtc counter sample in seconds (mktime)
	 */
	unsigned long rtc_reference;
	unsigned long rtc_time;
};
struct dspg_rtc *dspg_rtc;

static unsigned long
dspg_rtc_read_sample(struct dspg_rtc *rtc)
{
	return readl(rtc->base + DSPG_RTC_TIME) / rtc->rate;
}

static irqreturn_t
dspg_rtc_irq(int irq, void *priv)
{
	struct dspg_rtc *rtc = priv;

	/* clear interrupt */
	writel(RTC_INTCLR_ALARM, rtc->base + DSPG_RTC_INTCLR);

	/* update status to rtc layer */
	rtc_update_irq(rtc->dev, 1, RTC_AF);

	return IRQ_HANDLED;
}

static int
dspg_rtc_readtime(struct device *dev, struct rtc_time *tm)
{
	struct dspg_rtc *rtc = platform_get_drvdata(to_platform_device(dev));
	unsigned long sample;

	/*
	 * get the sample first, this will also retrieve the valid bit
	 * used below
	 */
	sample = dspg_rtc_read_sample(rtc);

	rtc_time_to_tm(rtc->rtc_time + sample - rtc->rtc_reference, tm);

	return 0;
}

static int
dspg_rtc_settime(struct device *dev, struct rtc_time *tm)
{
	struct dspg_rtc *rtc = platform_get_drvdata(to_platform_device(dev));

	rtc->rtc_reference = dspg_rtc_read_sample(rtc);
	rtc->rtc_time = mktime(tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday,
	                       tm->tm_hour, tm->tm_min, tm->tm_sec);

	return 0;
}

static int
dspg_rtc_readalarm(struct device *dev, struct rtc_wkalrm *wkalrm)
{
	struct dspg_rtc *rtc = platform_get_drvdata(to_platform_device(dev));
	unsigned long alarm;
	struct rtc_time *tm = &wkalrm->time;

	alarm  = readl(rtc->base + DSPG_RTC_ALARMVAL);

	wkalrm->enabled = !!(readl(rtc->base + DSPG_RTC_CFG) & RTC_CFG_ALARMEN);

	alarm = alarm - rtc->rtc_reference + rtc->rtc_time;

	rtc_time_to_tm(alarm, tm);

	return 0;
}

static int
dspg_rtc_setalarm(struct device *dev, struct rtc_wkalrm *wkalrm)
{
	struct dspg_rtc *rtc = platform_get_drvdata(to_platform_device(dev));
	unsigned long alarm;
	struct rtc_time *tm = &wkalrm->time;
	unsigned long cfg;

	alarm = mktime(tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday,
	               tm->tm_hour, tm->tm_min, tm->tm_sec);
	alarm = alarm - rtc->rtc_time + rtc->rtc_reference;

	while (!(readl(rtc->base + DSPG_RTC_STATUS) & RTC_STATUS_WRITEVALID))
		;

	writel(alarm * rtc->rate, rtc->base + DSPG_RTC_ALARMVAL);
	cfg = readl(rtc->base + DSPG_RTC_CFG);
	writel(cfg | RTC_CFG_ALARMEN, rtc->base + DSPG_RTC_CFG);

	return 0;
}

static int
dspg_rtc_proc(struct device *dev, struct seq_file *seq)
{
	return 0;
}

static const struct rtc_class_ops dspg_rtc_ops = {
	.read_time	= dspg_rtc_readtime,
	.set_time	= dspg_rtc_settime,
	.read_alarm	= dspg_rtc_readalarm,
	.set_alarm	= dspg_rtc_setalarm,
	.proc		= dspg_rtc_proc,
};

static int __devinit
dspg_rtc_probe(struct platform_device *pdev)
{
	struct dspg_rtc *rtc;
	struct resource *res;
	int ret;
	unsigned long cfg;

	rtc = kzalloc(sizeof(*rtc), GFP_KERNEL);
	if (!rtc)
		return -ENOMEM;

	dspg_rtc = rtc;

	rtc->clk = clk_get(&pdev->dev, NULL);
	if (IS_ERR(rtc->clk)) {
		dev_err(&pdev->dev, "cannot get clock\n");
		ret = PTR_ERR(rtc->clk);
		goto err_free;
	}
	clk_enable(rtc->clk);
	rtc->rate = clk_get_rate(rtc->clk);

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res) {
		ret = -ENXIO;
		dev_err(&pdev->dev, "no memory resource\n");
		goto err_clk;
	}

	if (!devm_request_mem_region(&pdev->dev, res->start,
				     resource_size(res), pdev->name)) {
		ret = -EBUSY;
		dev_err(&pdev->dev, "cannot request memory region\n");
		goto err_clk;
	}

	rtc->base = ioremap(res->start, resource_size(res));
	if (!rtc->base) {
		ret = -ENXIO;
		dev_err(&pdev->dev, "cannot map memory region\n");
		goto err_clk;
	}

	if ((rtc->irq = platform_get_irq(pdev, 0)) < 0) {
		ret = rtc->irq;
		dev_err(&pdev->dev, "no irq\n");
		goto err_iounmap;
	}

	platform_set_drvdata(pdev, rtc);

	device_init_wakeup(&pdev->dev, 1);

	rtc->dev = rtc_device_register(dspg_rtc_name, &pdev->dev,
				       &dspg_rtc_ops, THIS_MODULE);
	if (IS_ERR(rtc->dev)) {
		ret = PTR_ERR(rtc->dev);
		dev_err(&pdev->dev, "cannot allocate RTC\n");
		goto err_iounmap;
	}

	/* reset alarm */
	cfg = readl(rtc->base + DSPG_RTC_CFG);
	writel(cfg & ~RTC_CFG_ALARMEN, rtc->base + DSPG_RTC_CFG);

	rtc->irq = platform_get_irq(pdev, 0);
	if (!rtc->irq) {
		ret = -ENXIO;
		dev_err(&pdev->dev, "cannot get IRQ\n");
		goto err_unregister;
	}

	ret = request_threaded_irq(rtc->irq, NULL, dspg_rtc_irq, 0,
				   dspg_rtc_name, rtc);
	if (ret) {
		dev_err(&pdev->dev, "failed to request IRQ\n");
		goto err_unregister;
	}

	dev_info(&pdev->dev, "initialized\n");

	return 0;

err_unregister:
	rtc_device_unregister(rtc->dev);
err_iounmap:
	iounmap(rtc->base);
err_clk:
	clk_disable(rtc->clk);
	clk_put(rtc->clk);
err_free:
	kfree(rtc);
	return ret;
}

static int __exit
dspg_rtc_remove(struct platform_device *pdev)
{
	struct dspg_rtc *rtc = platform_get_drvdata(pdev);

	free_irq(rtc->irq, rtc);
	rtc_device_unregister(rtc->dev);
	iounmap(rtc->base);
	clk_disable(rtc->clk);
	clk_put(rtc->clk);
	kfree(rtc);

	return 0;
}

#ifdef CONFIG_PM
static int
dspg_rtc_suspend(struct platform_device *pdev, pm_message_t state)
{
	struct dspg_rtc *rtc = platform_get_drvdata(pdev);

	if (device_may_wakeup(&pdev->dev))
		enable_irq_wake(rtc->irq);

	return 0;
}

static int
dspg_rtc_resume(struct platform_device *pdev)
{
	struct dspg_rtc *rtc = platform_get_drvdata(pdev);

	if (device_may_wakeup(&pdev->dev))
		disable_irq_wake(rtc->irq);

	return 0;
}
#else
#define dspg_rtc_suspend NULL
#define dspg_rtc_resume NULL
#endif

static const struct of_device_id dspg_rtc_of_match[] = {
	{.compatible = "dspg,rtc", },
	{ },
};

static struct platform_driver dspg_rtc_driver = {
	.remove  = __exit_p(dspg_rtc_remove),
	.probe	 = dspg_rtc_probe,
	.suspend = dspg_rtc_suspend,
	.resume  = dspg_rtc_resume,
	.driver  = {
		.name = dspg_rtc_name,
		.owner = THIS_MODULE,
		.of_match_table = dspg_rtc_of_match,
	},
};

static int __init dspg_rtc_init(void)
{
	return platform_driver_register(&dspg_rtc_driver);
}

static void __exit dspg_rtc_exit(void)
{
	platform_driver_unregister(&dspg_rtc_driver);
}

module_init(dspg_rtc_init);
module_exit(dspg_rtc_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("DSPG Technologies GmbH");
MODULE_DESCRIPTION("Driver for the RTC block of the DSPG-DVF99");
