/*
 * spi-dspg-dmw.c - SPI master driver for the DMW96 style controller
 *
 * Copyright (C) 2012 DSPG Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <linux/clk.h>
#include <linux/delay.h>
#include <linux/err.h>
#include <linux/errno.h>
#include <linux/gpio.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/module.h>
#include <linux/of_gpio.h>
#include <linux/platform_device.h>
#include <linux/spi/spi.h>
#include <linux/spi/spi_bitbang.h>
#include <linux/spi/spi_dspg_dmw.h>
#include <linux/spinlock.h>
#include <linux/workqueue.h>
#ifdef CONFIG_DMW96_GDMAC
#include <linux/dmw96dma.h>
#endif
#include <linux/mm.h>

#define DRIVER_NAME "spi-dspg-dmw"

#define DMW_SPI_CFG_REG		0x00
#define DMW_SPI_RATE_CNTL	0x04
#define DMW_SPI_INT_EN		0x08
#define DMW_SPI_INT_STAT	0x0C
#define DMW_SPI_INT_CLR		0x10
#define DMW_SPI_INT_CAUSE	0x14
#define DMW_SPI_TX_DAT		0x18
#define DMW_SPI_RX_DAT		0x1C
#define DMW_SPI_DEL_VAL		0x20
#define DMW_SPI_RX_BUFF		0x3C
#define DMW_SPI_DBG		0x40

#define DMW_SPI_FIFO_SIZE	16

#define DMW_SPI_NUM_CHIPSELECT	2

#define INT_RX_FIFO_WM		(1<<0)
#define INT_RX_TO		(1<<1)
#define INT_RX_FIFO_UR		(1<<2)
#define INT_TX_FIFO_WM		(1<<8)
#define INT_TX_FIFO_OV		(1<<9)
#define INT_ALL_INTS		(INT_RX_FIFO_WM | INT_RX_TO | \
				 INT_RX_FIFO_UR | INT_TX_FIFO_WM | \
				 INT_TX_FIFO_OV)
#define INT_NO_INTS		0

#define INT_FULL_FIFO_WM	(15<<12)

#define SPI_IDLE		(1<<15)
#define SPI_DONE		(1<<10)

#define CFG_EN			(1<<31)
#define CFG_FIFOS_FLUSH		((1<<17) | (1<<18))
#define CFG_TX_DMA_EN		(1<<16)
#define CFG_RX_DMA_EN		(1<<15)
#define CFG_RX_IGNORE		(1<<8)
#define CFG_DIRECT_CS		(1<<7)
#define CFG_SWCS		(1<<6)
#define CFG_CS1_HOLD		(1<<4)
#define CFG_CS1_DIS		(1<<3)
#define CFG_CS0_HOLD		(1<<2)
#define CFG_CS0_DIS		(1<<1)

struct dspg_dmw_spi {
	/* bitbang has to be first */
	struct spi_bitbang	 bitbang;
	struct completion	 done;

	void __iomem		*regs;
	int			 irq;
	int			 remaining_bytes;

	/* data buffers */
	const unsigned char	*tx_ptr;
	unsigned char		*rx_ptr;

	int			chip_select;
	unsigned		cs_gpios[DMW_SPI_NUM_CHIPSELECT];

	struct clk		*clk;
	struct resource		*ioarea;
	struct spi_master	*master;

	struct device		*dev;

#ifdef CONFIG_DMW96_GDMAC
	/* DMA */
	struct dmw96dma_list	*dma_rx_list;
	struct dmw96dma_list	*dma_tx_list;
	unsigned int             dma_rx_channel;
	unsigned int             dma_tx_channel;
#endif
};

static unsigned int dma = 1;
module_param(dma, uint, 0);
MODULE_PARM_DESC(dma, "Support for DMA mode (default: 1)");

static void dspg_dmw_spi_debug(struct dspg_dmw_spi *hw);

#ifdef CONFIG_DMW96_GDMAC
static void
dspg_dmw_spi_dma_tx_callback(int status, void *context)
{
	struct dspg_dmw_spi *hw = (struct dspg_dmw_spi *)context;

	if (!status)
		hw->remaining_bytes = 0;

	if (!hw->dma_rx_list)
		complete(&hw->done);
}

static void
dspg_dmw_spi_dma_rx_callback(int status, void *context)
{
	struct dspg_dmw_spi *hw = (struct dspg_dmw_spi *)context;

	if (!status)
		hw->remaining_bytes = 0;

	complete(&hw->done);
}
#endif

static inline struct dspg_dmw_spi *
to_hw(struct spi_device *sdev)
{
	return spi_master_get_devdata(sdev->master);
}

static void
dspg_dmw_spi_set_cs(struct dspg_dmw_spi *hw, unsigned int chip_select, int val, int pol)
{
	unsigned int ctrl;
	unsigned int bit = CFG_CS0_HOLD;

	clk_enable(hw->clk);

	ctrl = readl(hw->regs + DMW_SPI_CFG_REG);

	if (hw->chip_select != 0)
		bit = CFG_CS1_HOLD;

	ctrl |= CFG_DIRECT_CS;
	if ((pol && (val == BITBANG_CS_ACTIVE)) ||
	   (!pol && (val == BITBANG_CS_INACTIVE))) {
		ctrl |=  CFG_SWCS;
		//ctrl &= ~bit; /* don't negate chip select at end of transaction */
	} else {
		ctrl &= ~CFG_SWCS;
		//ctrl |=  bit; /* negate chip select at end of transaction */
	}

	writel(ctrl, hw->regs + DMW_SPI_CFG_REG);

	clk_disable(hw->clk);
}

static void
dspg_dmw_spi_chipsel(struct spi_device *spi, int is_on)
{
	struct dspg_dmw_spi *hw = to_hw(spi);
	unsigned int cspol = spi->mode & SPI_CS_HIGH ? 1 : 0;
	unsigned int chip_select = spi->chip_select;
	unsigned cs = hw->cs_gpios[chip_select];

	if (gpio_is_valid(cs))
		gpio_direction_output(cs, is_on ? cspol : !cspol);
	else
		dspg_dmw_spi_set_cs(hw, chip_select, is_on, cspol);
}

static int
dspg_dmw_spi_setupxfer(struct spi_device *spi, struct spi_transfer *t)
{
	struct dspg_dmw_spi *hw = to_hw(spi);
	unsigned int bpw;
	unsigned long clk_rate = clk_get_rate(hw->clk);
	unsigned int hz;
	unsigned int div;
	unsigned int div_code;
	unsigned int ctrl = 0;
	unsigned msb_first = ~(spi->mode & SPI_LSB_FIRST);
	unsigned int clk_polarity =
			 spi->mode & SPI_CPOL ? 0 : 1; /* 0 if active high */

	bpw = t ? t->bits_per_word : 0;
	if (!bpw)
		bpw = spi->bits_per_word;
	hz  = t ? t->speed_hz : 0;
	if (!hz)
		hz = spi->max_speed_hz;

	dev_dbg(&spi->dev, "hz = %u, clockrate = %lu, clk_polarity = %s\n",
		hz, clk_rate, clk_polarity ? "active high" : "active low");

	if (!hz) {
		dev_err(&spi->dev, "invalid speed value (%dHz)\n", hz);
		return -EINVAL;
	}

	if (hz > spi->max_speed_hz) {
		dev_err(&spi->dev, "speed values > %dHz not supported\n",
			spi->max_speed_hz);
		return -EINVAL;
	}

	if (bpw != 8) {
		dev_err(&spi->dev, "invalid bits-per-word (%d)\n", bpw);
		return -EINVAL;
	}

	div = clk_rate / hz;
	if (div < 2)
		div = 2;

	div_code = 0;
	while ((div > 2) && (div_code < 10)) {
		div_code++;
		div = (div+1)/2; /* round up */
	}

	dev_dbg(&spi->dev, "setting pre-scaler to %d, code %u " \
		"(hz %u, clk %lu)\n", div, div_code, hz, clk_rate);

	clk_enable(hw->clk);

	writel(div_code, hw->regs + DMW_SPI_RATE_CNTL);

	writel(0x100, hw->regs + DMW_SPI_DEL_VAL);

	hw->chip_select = spi->chip_select;

	/* important: keep cs at the same configuration, since setup_transfer might be called */
	/* between 2 consecutive transfers on the same device (for example: spi_write_then_read) */
	ctrl = readl(hw->regs + DMW_SPI_CFG_REG) & CFG_SWCS;
	ctrl |= CFG_FIFOS_FLUSH |
	       ((msb_first & 0x1) << 5) | (clk_polarity & 0x1) | CFG_DIRECT_CS;
	if (hw->chip_select == 0)
		ctrl |= CFG_CS1_DIS;
	else
		ctrl |= CFG_CS0_DIS;
	writel(ctrl, hw->regs + DMW_SPI_CFG_REG);

	ctrl &= ~CFG_FIFOS_FLUSH;
	writel(ctrl, hw->regs + DMW_SPI_CFG_REG);

	spin_lock(&hw->bitbang.lock);
	if (!hw->bitbang.busy) {
		hw->bitbang.chipselect(spi, BITBANG_CS_INACTIVE);
		/* need to ndelay for 0.5 clocktick ? */
	}

	spin_unlock(&hw->bitbang.lock);

	clk_disable(hw->clk);

	return 0;
}

static int
dspg_dmw_spi_setup(struct spi_device *spi)
{
	int ret;

	if (!spi->bits_per_word)
		spi->bits_per_word = 8;

	ret = dspg_dmw_spi_setupxfer(spi, NULL);
	if (ret < 0) {
		dev_err(&spi->dev, "setupxfer returned %d\n", ret);
		return ret;
	}

	dev_dbg(&spi->dev, "%s: mode %d, %u bpw, %d hz\n",
		__func__, spi->mode, spi->bits_per_word,
		spi->max_speed_hz);

	return 0;
}

static void
dspg_dmw_spi_enable(struct dspg_dmw_spi *hw)
{
	unsigned long val;

	while (!(readl(hw->regs + DMW_SPI_INT_STAT) & SPI_DONE))
		;

	val = readl(hw->regs + DMW_SPI_CFG_REG);
	val |= CFG_EN;
	writel(val, hw->regs + DMW_SPI_CFG_REG);
}

#ifdef CONFIG_DMW96_GDMAC
static void
dspg_dmw_spi_dma(struct dspg_dmw_spi *hw, int flags)
{
	unsigned long val;

	val = readl(hw->regs + DMW_SPI_CFG_REG);
	if (val & CFG_EN) {
		while (!(readl(hw->regs + DMW_SPI_INT_STAT) & SPI_DONE))
			;

		val &= ~CFG_EN;
		writel(val, hw->regs + DMW_SPI_CFG_REG);
	}

	val &= ~(CFG_TX_DMA_EN | CFG_RX_DMA_EN);
	val |= flags;
	if ((flags & (CFG_RX_DMA_EN|CFG_TX_DMA_EN)) == CFG_TX_DMA_EN)
		val |=  CFG_RX_IGNORE;
	else
		val &= ~CFG_RX_IGNORE;
	if (flags)
		val |=  CFG_FIFOS_FLUSH;
	writel(val, hw->regs + DMW_SPI_CFG_REG);

	val &= ~CFG_FIFOS_FLUSH;
	if (flags)
		val |= CFG_EN;
	writel(val, hw->regs + DMW_SPI_CFG_REG);
}

static void
dspg_dmw_spi_watermarks(struct dspg_dmw_spi *hw, int tx, int rx)
{
	writel((((tx-1) & 0xF) << 12) | (((rx-1) & 0xF) << 4),
	       hw->regs + DMW_SPI_INT_EN);
}
#endif

static void
dspg_dmw_spi_disable_and_clear_ints(struct dspg_dmw_spi *hw)
{
	writel(INT_NO_INTS,  hw->regs + DMW_SPI_INT_EN);
	writel(INT_ALL_INTS, hw->regs + DMW_SPI_INT_CLR);
}

static void
dspg_dmw_spi_enable_rx_wm_int(struct dspg_dmw_spi *hw, int count)
{
	writel(INT_RX_FIFO_WM | (((count-1) & 0xf)<<4),
	       hw->regs + DMW_SPI_INT_EN);
}

static int
dspg_dmw_spi_rx_not_empty(struct dspg_dmw_spi *hw)
{
	return !!(readl(hw->regs + DMW_SPI_INT_STAT) & INT_RX_FIFO_WM);
}

static int
dspg_dmw_spi_tx_not_full(struct dspg_dmw_spi *hw)
{
	return !!(readl(hw->regs + DMW_SPI_INT_STAT) & INT_TX_FIFO_WM);
}

static int
dspg_dmw_spi_fill_tx_fifo(struct dspg_dmw_spi *hw)
{
	int written = 0;
	unsigned char dat = 0;

	while (dspg_dmw_spi_tx_not_full(hw) && (hw->remaining_bytes > 0) &&
	       (written < DMW_SPI_FIFO_SIZE)) {
		if (hw->tx_ptr)
			dat = *hw->tx_ptr++;
		writel(dat, hw->regs + DMW_SPI_TX_DAT);
		hw->remaining_bytes--;
		written++;
	}

	return written;
}

static int
dspg_dmw_spi_txrx(struct spi_device *spi, struct spi_transfer *t)
{
	struct dspg_dmw_spi *hw = to_hw(spi);
#ifdef CONFIG_DMW96_GDMAC
	int use_dma = 0;
#endif

	hw->chip_select = spi->chip_select;
	hw->tx_ptr = t->tx_buf;
	hw->rx_ptr = t->rx_buf;
	hw->remaining_bytes = t->len;
	INIT_COMPLETION(hw->done);

	dev_dbg(&spi->dev, "transfer %d bytes (%s %s)\n", t->len,
	        t->tx_buf ? "tx" : "  ", t->rx_buf ? "rx" : "  ");

	clk_enable(hw->clk);
	dspg_dmw_spi_disable_and_clear_ints(hw);

#ifdef CONFIG_DMW96_GDMAC
	use_dma = dma &&
		!(hw->remaining_bytes % DMW_SPI_FIFO_SIZE) &&
		(!hw->rx_ptr || virt_addr_valid(hw->rx_ptr)) &&
		(!hw->tx_ptr || virt_addr_valid(hw->tx_ptr));
	if (use_dma) {
		if (hw->rx_ptr) {
			hw->dma_rx_list = dmw96dma_alloc_io_list(0, hw->dma_rx_channel,
			                                         dspg_dmw_spi_dma_rx_callback,
			                                         (void *)hw);
			if (!hw->dma_rx_list) {
				dev_err(&spi->dev, "Out of memory for DMA\n");
				clk_disable(hw->clk);
				return t->len;
			}

			dev_dbg(&spi->dev, "setting up RX DMA transfer for %d bytes\n",
			        t->len);

			dmw96dma_add_io_transfer_buf(hw->dma_rx_list,
			                             (void *)hw->rx_ptr,
			                             t->len, NULL,
			                             (void *)hw);
		}

		hw->dma_tx_list = dmw96dma_alloc_io_list(0, hw->dma_tx_channel,
		                                         dspg_dmw_spi_dma_tx_callback,
		                                         (void *)hw);
		if (!hw->dma_tx_list) {
			dev_err(&spi->dev, "Out of memory for DMA\n");
			if (hw->dma_rx_list) {
				dmw96dma_free_list(hw->dma_rx_list);
				hw->dma_rx_list = NULL;
			}
			clk_disable(hw->clk);
			return t->len;
		}

		dev_dbg(&spi->dev, "setting up TX DMA transfer for %d bytes\n",
		        t->len);

		dmw96dma_add_io_transfer_buf(hw->dma_tx_list,
		                             hw->tx_ptr ? (void *)hw->tx_ptr :
		                                          (void *)hw->rx_ptr,
		                             t->len, NULL,
		                             (void *)hw);

		dspg_dmw_spi_watermarks(hw, DMW_SPI_FIFO_SIZE, DMW_SPI_FIFO_SIZE);

		if (hw->rx_ptr)
			dmw96dma_submit(hw->dma_rx_list);
		dmw96dma_submit(hw->dma_tx_list);

		if (hw->rx_ptr)
			dspg_dmw_spi_dma(hw, CFG_RX_DMA_EN | CFG_TX_DMA_EN);
		else
			dspg_dmw_spi_dma(hw, CFG_TX_DMA_EN);
	} else
#endif
	{
		int count;

		dspg_dmw_spi_enable(hw);

		count = dspg_dmw_spi_fill_tx_fifo(hw);

		dspg_dmw_spi_enable_rx_wm_int(hw, count);
	}

	wait_for_completion(&hw->done);

#ifdef CONFIG_DMW96_GDMAC
	if (use_dma) {
		dspg_dmw_spi_dma(hw, 0);

		if (hw->dma_rx_list) {
			dmw96dma_free_list(hw->dma_rx_list);
			hw->dma_rx_list = NULL;
		}

		if (hw->dma_tx_list) {
			dmw96dma_free_list(hw->dma_tx_list);
			hw->dma_tx_list = NULL;
		}
	}
#endif

	dspg_dmw_spi_disable_and_clear_ints(hw);
	clk_disable(hw->clk);

	return t->len - hw->remaining_bytes;
}

static void
dspg_dmw_spi_debug(struct dspg_dmw_spi *hw)
{
	int val;

	printk(KERN_INFO "dspg_dmw_spi: CFG_REG   = 0x%x\n", readl(hw->regs + DMW_SPI_CFG_REG));
	printk(KERN_INFO "dspg_dmw_spi: RATE_CNTL = 0x%x\n", readl(hw->regs + DMW_SPI_RATE_CNTL));
	printk(KERN_INFO "dspg_dmw_spi: DEL_VAL   = 0x%x\n", readl(hw->regs + DMW_SPI_DEL_VAL));
	printk(KERN_INFO "dspg_dmw_spi: INT_EN =    0x%x\n", readl(hw->regs + DMW_SPI_INT_EN));
	printk(KERN_INFO "dspg_dmw_spi: INT_STAT =  0x%x\n", readl(hw->regs + DMW_SPI_INT_STAT));
	printk(KERN_INFO "dspg_dmw_spi: INT_CLR =   0x%x\n", readl(hw->regs + DMW_SPI_INT_CLR));
	printk(KERN_INFO "dspg_dmw_spi: INT_CAUSE = 0x%x\n", readl(hw->regs + DMW_SPI_INT_CAUSE));

	val = readl(hw->regs + DMW_SPI_DBG);

	printk(KERN_INFO "dspg_dmw_spi: TX_BUSY:       %d\n", (val & (0x1<<31))>>31);
	printk(KERN_INFO "dspg_dmw_spi: SPI_FSM_CUR:   %d\n", (val & (0x3<<18))>>18);
	printk(KERN_INFO "dspg_dmw_spi: SPI_FSM_NXT:   %d\n", (val & (0x3<<16))>>16);
	printk(KERN_INFO "dspg_dmw_spi: TX_FIFO_LVL:   %d\n", (val & (0x1f<<9))>>9);
	printk(KERN_INFO "dspg_dmw_spi: TX_FIFO_EMPTY: %d\n", (val & (0x1<<8))>>8);
	printk(KERN_INFO "dspg_dmw_spi: TX_FIFO_FULL:  %d\n", (val & (0x1<<7))>>7);
	printk(KERN_INFO "dspg_dmw_spi: RX_FIFO_LVL:   %d\n", (val & (0x1f<<2))>>2);
	printk(KERN_INFO "dspg_dmw_spi: RX_FIFO_EMPTY: %d\n", (val & (0x1<<1))>>1);
	printk(KERN_INFO "dspg_dmw_spi: RX_FIFO_FULL:  %d\n", val & 0x1);
}

irqreturn_t
dspg_dmw_spi_irq(int irq, void *dev)
{
	struct dspg_dmw_spi *hw = dev;
	int cause;
	unsigned char dat;

	cause = readl(hw->regs + DMW_SPI_INT_CAUSE);
	dspg_dmw_spi_disable_and_clear_ints(hw);

	if (cause & INT_RX_FIFO_WM) { /* A transmit has completed */
		while (dspg_dmw_spi_rx_not_empty(hw)) {
			/* Read out all the data from the Rx FIFO */
			dat = readl(hw->regs + DMW_SPI_RX_DAT);
			if (hw->rx_ptr)
				*hw->rx_ptr++ = dat;
		}

		if (hw->remaining_bytes > 0) {
			int count = dspg_dmw_spi_fill_tx_fifo(hw);

			dspg_dmw_spi_enable_rx_wm_int(hw, count);
		} else {
			/* No more data to send.
			 * Indicate the transfer is completed.
			 */

			complete(&hw->done);
		}
	} else {
		printk(KERN_ERR "dspg_dmw_spi: unhandled or ghost interrupt (0x%x)\n", cause);
		dspg_dmw_spi_debug(hw);
	}

	return IRQ_HANDLED;
}

static int
dspg_dmw_spi_probe(struct platform_device *pdev)
{
	struct dspg_dmw_spi *hw;
	struct spi_master *master;
	struct resource *res;
	int err = 0;

	master = spi_alloc_master(&pdev->dev, sizeof(struct dspg_dmw_spi));
	if (master == NULL) {
		dev_err(&pdev->dev, "No memory for spi_master\n");
		err = -ENOMEM;
		goto err_nomem;
	}

	master->bus_num = -1;
	master->num_chipselect = DMW_SPI_NUM_CHIPSELECT;
	master->dev.of_node = pdev->dev.of_node;

	hw = spi_master_get_devdata(master);
	memset(hw, 0, sizeof(struct dspg_dmw_spi));

	hw->master = spi_master_get(master);
	hw->dev = &pdev->dev;
	hw->cs_gpios[0] = -1;
	hw->cs_gpios[1] = -1;

	if (pdev->dev.of_node) {
		struct device_node *np = pdev->dev.of_node;

		hw->cs_gpios[0] = of_get_named_gpio(np, "cs-gpios", 0);
		hw->cs_gpios[1] = of_get_named_gpio(np, "cs-gpios", 1);
	} else if (pdev->dev.platform_data) {
		struct spi_dspg_dmw_info *pdata = pdev->dev.platform_data;

		master->bus_num = pdata->bus_num;
		hw->cs_gpios[0] = pdata->cs0_gpio;
		hw->cs_gpios[1] = pdata->cs1_gpio;
	}

	if (gpio_is_valid(hw->cs_gpios[0])) {
		err = gpio_request(hw->cs_gpios[0], "spi-dspg-dmw-cs0");
		if (err) {
			dev_err(&pdev->dev, "cannot get CS0 GPIO\n");
			goto err_req_gpio0;
		}
	}

	if (gpio_is_valid(hw->cs_gpios[1])) {
		err = gpio_request(hw->cs_gpios[1], "spi-dspg-dmw-cs1");
		if (err) {
			dev_err(&pdev->dev, "cannot get CS1 GPIO\n");
			goto err_req_gpio1;
		}
	}

	platform_set_drvdata(pdev, hw);
	init_completion(&hw->done);

	/* setup the state for the bitbang driver */
	hw->bitbang.master         = hw->master;
	hw->bitbang.setup_transfer = dspg_dmw_spi_setupxfer;
	hw->bitbang.chipselect     = dspg_dmw_spi_chipsel;
	hw->bitbang.txrx_bufs      = dspg_dmw_spi_txrx;
	hw->bitbang.master->setup  = dspg_dmw_spi_setup;

	dev_dbg(hw->dev, "bitbang at %p\n", &hw->bitbang);

	/* find and map our resources */
#ifdef CONFIG_DMW96_GDMAC
	res = platform_get_resource(pdev, IORESOURCE_DMA, 0);
	if (res == NULL) {
		dev_err(&pdev->dev, "Cannot get RX DMA channel\n");
		err = -ENOENT;
		goto err_no_iores;
	} else
		hw->dma_rx_channel = res->start;

	res = platform_get_resource(pdev, IORESOURCE_DMA, 1);
	if (res == NULL) {
		dev_err(&pdev->dev, "Cannot get TX DMA channel\n");
		err = -ENOENT;
		goto err_no_iores;
	} else
		hw->dma_tx_channel = res->start;
#endif

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (res == NULL) {
		dev_err(&pdev->dev, "Cannot get IORESOURCE_MEM\n");
		err = -ENOENT;
		goto err_no_iores;
	}

	hw->ioarea = request_mem_region(res->start, (res->end - res->start)+1,
	                                pdev->name);

	if (hw->ioarea == NULL) {
		dev_err(&pdev->dev, "Cannot reserve region\n");
		err = -ENXIO;
		goto err_no_iores;
	}

	hw->regs = ioremap_nocache(res->start, (res->end - res->start)+1);
	if (hw->regs == NULL) {
		dev_err(&pdev->dev, "Cannot map IO\n");
		err = -ENXIO;
		goto err_no_iomap;
	}

	hw->irq = platform_get_irq(pdev, 0);
	if (hw->irq < 0) {
		dev_err(&pdev->dev, "No IRQ specified\n");
		err = -ENOENT;
		goto err_no_iomap;
	}

	hw->clk = clk_get(&pdev->dev, NULL);
	if (IS_ERR(hw->clk)) {
		dev_err(&pdev->dev, "No clock for device\n");
		err = PTR_ERR(hw->clk);
		goto err_no_clk;
	}

	clk_enable(hw->clk);

	dspg_dmw_spi_disable_and_clear_ints(hw);

	err = request_irq(hw->irq, dspg_dmw_spi_irq, 0, DRIVER_NAME, hw);
	if (err) {
		dev_err(&pdev->dev, "Cannot claim IRQ\n");
		goto err_no_irq;
	}

	/* register our spi controller */
	err = spi_bitbang_start(&hw->bitbang);
	if (err) {
		dev_err(&pdev->dev, "Failed to register SPI master\n");
		goto err_register;
	}

	clk_disable(hw->clk);

	return 0;

 err_register:
	free_irq(hw->irq, hw);

 err_no_irq:
	clk_disable(hw->clk);
	clk_put(hw->clk);

 err_no_clk:
	iounmap(hw->regs);

 err_no_iomap:
	release_resource(hw->ioarea);
	kfree(hw->ioarea);

 err_no_iores:
	if (gpio_is_valid(hw->cs_gpios[1]))
		gpio_free(hw->cs_gpios[1]);

 err_req_gpio1:
	if (gpio_is_valid(hw->cs_gpios[0]))
		gpio_free(hw->cs_gpios[0]);

 err_req_gpio0:
	spi_master_put(hw->master);

 err_nomem:
	return err;
}

static int
dspg_dmw_spi_remove(struct platform_device *dev)
{
	struct dspg_dmw_spi *hw = platform_get_drvdata(dev);
	unsigned long val;

	platform_set_drvdata(dev, NULL);

	spi_unregister_master(hw->master);

	val = readl(hw->regs + DMW_SPI_CFG_REG);
	val &= ~(CFG_EN);
	writel(val, hw->regs + DMW_SPI_CFG_REG);

	clk_disable(hw->clk);
	clk_put(hw->clk);

	free_irq(hw->irq, hw);
	iounmap(hw->regs);

	release_resource(hw->ioarea);
	kfree(hw->ioarea);

	if (gpio_is_valid(hw->cs_gpios[1]))
		gpio_free(hw->cs_gpios[1]);

	if (gpio_is_valid(hw->cs_gpios[0]))
		gpio_free(hw->cs_gpios[0]);

	spi_master_put(hw->master);

	return 0;
}

static const struct of_device_id spi_dspg_dmw_dt_ids[] = {
	{ .compatible = "dspg,spi-dmw", },
	{ /* sentinel */ }
};

static struct platform_driver dspg_dmw_spidrv = {
	.probe		= dspg_dmw_spi_probe,
	.remove		= dspg_dmw_spi_remove,

	.driver		= {
		.name	= DRIVER_NAME,
		.owner	= THIS_MODULE,
		.of_match_table = spi_dspg_dmw_dt_ids,
	},
};

static int __init
dspg_dmw_spi_init(void)
{
	return platform_driver_register(&dspg_dmw_spidrv);
}

static void __exit
dspg_dmw_spi_exit(void)
{
	platform_driver_unregister(&dspg_dmw_spidrv);
}

subsys_initcall(dspg_dmw_spi_init);
module_exit(dspg_dmw_spi_exit);

MODULE_DESCRIPTION("DSPG DMW style SPI Driver");
MODULE_AUTHOR("Andreas Weissel <andreas.weissel@dpsg.com>");
MODULE_LICENSE("GPL");
