/*
 * DSPG crypto coma driver
 *
 * Copyright (c) 2012, DSP Group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/mutex.h>
#include <linux/err.h>
#include <linux/clk.h>
#include <linux/slab.h>
#include <linux/platform_device.h>
#include <linux/debugfs.h>
#include <linux/uaccess.h>
#include <linux/of.h>
#include <linux/dma-mapping.h>
#include <linux/completion.h>
#include <linux/coma/coma.h>
#include <linux/coma/cfifo.h>
#include "cmsg-crypto.h"
#include "coma-crypto.h"

#define COMA_CRYPTO "coma-crypto"

struct phys_memory {
	void *data;
	unsigned int size;
	dma_addr_t phys;
	unsigned int rsize;
};

struct crypto_coma {
	struct platform_device *pdev;
	int service_id;
	struct dentry *pdentry;
	struct mutex lock;
	struct phys_memory iv;
	struct phys_memory key;
	struct phys_memory data;
	struct phys_memory finished;
	union cmsg_crypto_params reply;
	struct completion css_done;
	struct completion crypt_done;
};
struct crypto_coma *crypto_coma;

static int
crypto_coma_create_message(enum cmsg_crypto_types type,
			  union cmsg_crypto_params *params)
{
	return coma_cmsg_send(crypto_coma->service_id, (int)type,
			      (void *)params, sizeof(*params),
			      NULL, 0);
}

static void
crypto_coma_process_message(void *arg, struct cmsg *cmsg)
{
	volatile int *done = crypto_coma->finished.data;
	union cmsg_crypto_params *params =
				(union cmsg_crypto_params *)cmsg_params(cmsg);

	switch (cmsg->type) {
	case CMSG_CRYPTO_OPEN_REPLY:
	case CMSG_CRYPTO_CLOSE_REPLY:
		memcpy(&crypto_coma->reply, params, sizeof(*params));
		complete(&crypto_coma->css_done);
		break;
	case CMSG_CRYPTO_DONE:
		if (*done)
			complete(&crypto_coma->crypt_done);
		break;
	default:
		dev_err(&crypto_coma->pdev->dev, "invalid message type\n");
		return;
	}
}

#ifdef CONFIG_DEBUG_FS
static int crypto_debug_common_open(struct crypto_coma *p, struct file *f,
				    struct phys_memory *pm)
{
	f->private_data = pm;
	return 0;
}

static int crypto_debug_common_release(struct inode *inode, struct file *f)
{
	return 0;
}

static int crypto_debug_iv_open(struct inode *inode, struct file *f)
{
	struct crypto_coma *p = inode->i_private;

	return crypto_debug_common_open(p, f, &p->iv);
}

static int crypto_debug_key_open(struct inode *inode, struct file *f)
{
	struct crypto_coma *p = inode->i_private;

	return crypto_debug_common_open(p, f, &p->key);
}

static int crypto_debug_data_open(struct inode *inode, struct file *f)
{
	struct crypto_coma *p = inode->i_private;

	return crypto_debug_common_open(p, f, &p->data);
}

struct crypto_session {
	int session;
	struct crypto_coma *p;
};

static int crypto_debug_start_open(struct inode *inode, struct file *f)
{
	struct crypto_coma *p = inode->i_private;
	struct crypto_session *s;
	union cmsg_crypto_params params;
	int ret = -EACCES;

	mutex_lock(&p->lock);

	s = kmalloc(sizeof(struct crypto_session), GFP_KERNEL);
	if (!s) {
		ret = -ENOMEM;
		goto out;
	}

	params.crypto_open.algorithm = ALGORITHM_AES_GCM;
	ret = crypto_coma_create_message(CMSG_CRYPTO_OPEN, &params);
	if (ret != 0)
		goto out_fail;
	wait_for_completion(&crypto_coma->css_done);

	ret = crypto_coma->reply.crypto_open_reply.status;
	if (ret < 0)
		goto out_fail;

	s->session = ret;
	s->p = p;
	f->private_data = s;
	goto out;
out_fail:
	kfree(s);
out:
	mutex_unlock(&p->lock);
	return ret;
}

static int crypto_debug_start_release(struct inode *inode, struct file *f)
{
	struct crypto_coma *p = inode->i_private;
	struct crypto_session *s = f->private_data;
	union cmsg_crypto_params params;
	int ret = 0;

	mutex_lock(&p->lock);

	params.crypto_close.session = s->session;
	ret = crypto_coma_create_message(CMSG_CRYPTO_CLOSE, &params);
	if (ret != 0)
		goto out;
	wait_for_completion(&crypto_coma->css_done);

	ret = crypto_coma->reply.crypto_close_reply.status;
	if (ret < 0)
		goto out;
	kfree(s);
out:
	mutex_unlock(&p->lock);
	return ret;
}

static int crypto_debug_start_write(char action, struct crypto_session *s, struct crypto_coma *p, int sync)
{
	union cmsg_crypto_params params;
	int ret = 0;
	volatile int *done = p->finished.data;
	struct timespec before, after;
	unsigned long nsec;
	unsigned long sec;

	mutex_lock(&p->lock);

	getnstimeofday(&before);

	params.crypto_start.session = s->session;
	params.crypto_start.action = (action == 'e'? ACTION_ENCRYPT: ACTION_DECRYPT);
	params.crypto_start.flags = (sync ? 0: CRYPTO_FLAG_SIGNAL);
	params.crypto_start.finished = (void *)p->finished.phys;
	params.crypto_start.iv_len = p->iv.rsize;
	params.crypto_start.iv = (void *)p->iv.phys;
	params.crypto_start.key_len = p->key.rsize;
	params.crypto_start.key = (void *)p->key.phys;
	params.crypto_start.data_len = p->data.rsize;
	params.crypto_start.data = (void *)p->data.phys;
	ret = crypto_coma_create_message(CMSG_CRYPTO_START, &params);
	if (ret != 0) {
		dev_err(&p->pdev->dev, "cannot start: %d\n", ret);
		goto out;
	}

	if (sync)
		while (*done == 0);
	else
		wait_for_completion(&p->crypt_done);

	getnstimeofday(&after);

	/* reset done flag */
	ret = *done;
	*done = 0;

	sec = after.tv_sec - before.tv_sec;
	if (!sec) {
		if (after.tv_nsec > before.tv_nsec)
			nsec = after.tv_nsec - before.tv_nsec;
		else
			nsec = after.tv_nsec + 1000000000ul - before.tv_nsec;
	} else {
		sec--;
		if (before.tv_nsec != after.tv_nsec)
			nsec = 1000000000ul - before.tv_nsec + after.tv_nsec;
		else
			nsec = 0;
	}

	dev_info(&p->pdev->dev, "%s-%s %s %lu.%lu - %lu.%lu (elapsed: %lu.%lu)\n",
		 sync ? "sync": "async",
		 action == 'e'? "encryption": "decryption",
		 ret == 1? "done": "failed",
		 before.tv_sec,
		 before.tv_nsec,
		 after.tv_sec,
		 after.tv_nsec,
		 sec,
		 nsec);
out:
	mutex_unlock(&p->lock);
	return ret;
}

static ssize_t crypto_debug_start_write_sync(struct file *f, const char __user *usr, size_t size, loff_t *off)
{
	size_t read = size;
	char action;
	int ret = 0;
	struct crypto_session *s = f->private_data;
	struct crypto_coma *p = s->p;

	if (*off >= 1)
		return 0;
	if (*off + size > 1)
		read = 1 - *off;

	if (copy_from_user((void *)&action + *off, usr, read))
		return -EFAULT;

	*off += size;

	ret = crypto_debug_start_write(action, s, p, 1);

	return size;
}

static ssize_t crypto_debug_start_write_async(struct file *f, const char __user *usr, size_t size, loff_t *off)
{
	size_t read = size;
	char action;
	int ret = 0;
	struct crypto_session *s = f->private_data;
	struct crypto_coma *p = s->p;

	if (*off >= 1)
		return 0;
	if (*off + size > 1)
		read = 1 - *off;

	if (copy_from_user((void *)&action + *off, usr, read))
		return -EFAULT;

	*off += size;

	ret = crypto_debug_start_write(action, s, p, 0);

	return size;
}

static ssize_t crypto_debug_write(struct file *f, const char __user *usr, size_t size, loff_t *off)
{
	struct phys_memory *pm = f->private_data;

	if (*off >= pm->size) {
		return 0;
	}
	if (*off + size > pm->size)
		size = pm->size - *off;

	if (copy_from_user(pm->data + *off, usr, size)) {
		return -EFAULT;
	}

	*off += size;
	pm->rsize = *off;

	return size;
}

static int crypto_alloc_dma_mem(struct crypto_coma *p)
{
	int ret = 0;

	p->iv.size = 512;
	p->iv.data = dma_alloc_coherent(&p->pdev->dev,
					p->iv.size,
					&p->iv.phys,
					GFP_KERNEL);
	if (!p->iv.data) {
		dev_err(&p->pdev->dev, "no memory left for iv buffer\n");
		ret = -ENOMEM;
		goto out;
	}
	p->key.size = 512;
	p->key.data = dma_alloc_coherent(&p->pdev->dev,
					 p->key.size,
					 &p->key.phys,
					 GFP_KERNEL);
	if (!p->key.data) {
		dev_err(&p->pdev->dev, "no memory left for key buffer\n");
		ret = -ENOMEM;
		goto out_free_iv;
	}
	p->data.size = 64 * 1024;
	p->data.data = dma_alloc_coherent(&p->pdev->dev,
					  p->data.size,
					  &p->data.phys,
					  GFP_KERNEL);
	if (!p->data.data) {
		dev_err(&p->pdev->dev, "no memory left for data buffer\n");
		ret = -ENOMEM;
		goto out_free_key;
	}
	p->finished.size = sizeof(int);
	p->finished.data = dma_alloc_coherent(&p->pdev->dev,
					      p->finished.size,
					      &p->finished.phys,
					      GFP_KERNEL);
	if (!p->data.data) {
		dev_err(&p->pdev->dev, "no memory left for data buffer\n");
		ret = -ENOMEM;
		goto out_free_data;
	}

	goto out;
out_free_data:
	dma_free_coherent(&p->pdev->dev, p->data.size, p->data.data, p->data.phys);
out_free_key:
	dma_free_coherent(&p->pdev->dev, p->key.size, p->key.data, p->key.phys);
out_free_iv:
	dma_free_coherent(&p->pdev->dev, p->iv.size, p->iv.data, p->iv.phys);
out:
	return ret;
}

static void crypto_free_dma_mem(struct crypto_coma *p)
{
	dma_free_coherent(&p->pdev->dev, p->iv.size, p->iv.data, p->iv.phys);
	dma_free_coherent(&p->pdev->dev, p->key.size, p->key.data, p->key.phys);
	dma_free_coherent(&p->pdev->dev, p->data.size, p->data.data, p->data.phys);
	dma_free_coherent(&p->pdev->dev, p->finished.size, p->finished.data, p->finished.phys);
}

static struct file_operations debug_iv_fops = {
	.open		= crypto_debug_iv_open,
	.write		= crypto_debug_write,
	.release	= crypto_debug_common_release,
};

static struct file_operations debug_key_fops = {
	.open		= crypto_debug_key_open,
	.write		= crypto_debug_write,
	.release	= crypto_debug_common_release,
};

static struct file_operations debug_data_fops = {
	.open		= crypto_debug_data_open,
	.write		= crypto_debug_write,
	.release	= crypto_debug_common_release,
};

static struct file_operations debug_start_sync_fops = {
	.open		= crypto_debug_start_open,
	.write		= crypto_debug_start_write_sync,
	.release	= crypto_debug_start_release,
};

static struct file_operations debug_start_async_fops = {
	.open		= crypto_debug_start_open,
	.write		= crypto_debug_start_write_async,
	.release	= crypto_debug_start_release,
};

static void crypto_create_debugfs_files(struct crypto_coma *p)
{
	int ret;
	struct dentry *pdentry;

	ret = crypto_alloc_dma_mem(p);
	if (ret != 0)
		return;

	pdentry = debugfs_create_dir("css-crypto", NULL);
	p->pdentry = pdentry;

	debugfs_create_file("iv", S_IWUSR, pdentry, p, &debug_iv_fops);
	debugfs_create_file("key", S_IWUSR, pdentry, p, &debug_key_fops);
	debugfs_create_file("data", S_IWUSR, pdentry, p, &debug_data_fops);
	debugfs_create_file("start_sync", S_IWUSR, pdentry, p, &debug_start_sync_fops);
	debugfs_create_file("start_async", S_IWUSR, pdentry, p, &debug_start_async_fops);
}

static void crypto_remove_debugfs_files(struct crypto_coma *p)
{
	if (p->pdentry)
		debugfs_remove_recursive(p->pdentry);
	crypto_free_dma_mem(p);
}
#else
#define crypto_create_debugfs_files(p)
#define crypto_remove_debugfs_files(p)
#endif

static void
crypto_coma_remove_service(void *arg)
{

	crypto_coma->service_id = -1;
	crypto_remove_debugfs_files(crypto_coma);
}

int crypto_coma_service_init(void)
{
	int ret;

	if (!crypto_coma)
		return -ENODEV;

	ret = coma_register("crypto", 10240, NULL, crypto_coma_process_message,
			    crypto_coma_remove_service, NULL);
	if (ret < 0) {
		dev_err(&crypto_coma->pdev->dev, "coma_register failed: %d\n", ret);
		return ret;
	}
	crypto_coma->service_id = ret;

	crypto_create_debugfs_files(crypto_coma);

	return 0;
}
EXPORT_SYMBOL(crypto_coma_service_init);

static int
crypto_coma_probe(struct platform_device *pdev)
{
	crypto_coma = kzalloc(sizeof *crypto_coma, GFP_KERNEL);
	if (!crypto_coma) {
		dev_err(&pdev->dev, "Out of memory\n");
		return -ENOMEM;
	}
	crypto_coma->pdev = pdev;
	crypto_coma->service_id = -1;

	mutex_init(&crypto_coma->lock);
	init_completion(&crypto_coma->css_done);
	init_completion(&crypto_coma->crypt_done);

	platform_set_drvdata(pdev, crypto_coma);

	dev_info(&pdev->dev, "successfully registered\n");

	return 0;
}

static int __exit
crypto_coma_remove(struct platform_device *dev)
{
	/* FIXME: racy wrt. crypto_coma_remove_service callback */
	if (crypto_coma->service_id >= 0) {
		coma_deregister(crypto_coma->service_id);
		crypto_coma_remove_service(NULL);
	}
	mutex_destroy(&crypto_coma->lock);
	kfree(crypto_coma);
	crypto_coma = NULL;

	return 0;
}

static struct of_device_id crypto_coma_of_ids[] = {
	{ .compatible = "dspg,crypto-coma" },
	{ },
};

static struct platform_driver crypto_coma_platform_driver = {
	.driver = {
		.name	= COMA_CRYPTO,
		.owner	= THIS_MODULE,
		.of_match_table = crypto_coma_of_ids,
	},
	.probe		= crypto_coma_probe,
	.remove		= __exit_p(crypto_coma_remove),
};

static int __init crypto_coma_init(void)
{
	return platform_driver_register(&crypto_coma_platform_driver);
}

static void __exit crypto_coma_exit(void)
{
	platform_driver_unregister(&crypto_coma_platform_driver);
}

module_init(crypto_coma_init);
module_exit(crypto_coma_exit);

MODULE_DESCRIPTION("DSPG COMA crypto driver");
MODULE_LICENSE("GPL v2");
MODULE_AUTHOR("DSP Group");
MODULE_ALIAS("platform:crypto-coma");
