/*
 * DSPG shared memory coma driver
 *
 * Copyright (c) 2013, DSP Group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/err.h>
#include <linux/io.h>
#include <linux/ioctl.h>
#include <linux/slab.h>
#include <linux/cdev.h>
#include <linux/file.h>
#include <linux/fs.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/completion.h>
#include <linux/debugfs.h>
#include <linux/coma/coma.h>
#include <linux/coma/coma-sharedmem.h>
#include <asm/uaccess.h>
#include <mach/platform.h>
#include "cmsg-sharedmem.h"

#define INIT_TIMEOUT   500 /* ms */
#define COMA_SHAREDMEM "coma-sharedmem"

struct sharedmem_coma {
	struct platform_device *pdev;
	int service_id;
	struct cdev cdev;
	struct device *dev;
	struct device *css_dev;
	dev_t chrdev;
	struct class *class;
	struct mutex lock;
	struct completion done;
	int status;
	int major;
	dma_addr_t base;
#ifdef CONFIG_DEBUG_FS
	struct dentry *pdentry;
	struct debugfs_blob_wrapper mem_blob;
#endif
};
struct sharedmem_coma *sharedmem_coma;

static int
sharedmem_coma_create_message(enum cmsg_sharedmem_types type,
			      union cmsg_sharedmem_params *params)
{
	return coma_cmsg_send(sharedmem_coma->service_id, (int)type,
			      (void *)params, sizeof(*params), NULL, 0);
}

#ifdef CONFIG_DEBUG_FS
static void
sharedmem_coma_debugfs_create_files(struct sharedmem_coma *sharedmem_coma)
{
	struct dentry *pdentry;

	pdentry = debugfs_create_dir("sharedmem", NULL);
	sharedmem_coma->pdentry = pdentry;

	sharedmem_coma->mem_blob.data = ioremap(sharedmem_coma->base,
						CONFIG_DVF_RESERVEMEM_CSS*SZ_1M);
	sharedmem_coma->mem_blob.size = CONFIG_DVF_RESERVEMEM_CSS*SZ_1M;

	if (sharedmem_coma->mem_blob.data)
		debugfs_create_blob("memory", S_IRUSR, pdentry,
				    &sharedmem_coma->mem_blob);
}

static void
sharedmem_coma_debugfs_remove_files(struct sharedmem_coma *sharedmem_coma)
{
	if (sharedmem_coma->mem_blob.data)
		iounmap(sharedmem_coma->mem_blob.data);
	if (sharedmem_coma->pdentry)
		debugfs_remove_recursive(sharedmem_coma->pdentry);
}
#else
#define sharedmem_coma_debugfs_create_files(p)
#define sharedmem_coma_debugfs_remove_files(p)
#endif

static int
sharedmem_coma_initialize(size_t size)
{
	union cmsg_sharedmem_params params;
	int ret = 0;
	unsigned long timedout;

	if (sharedmem_coma->service_id < 0) {
		dev_err(&sharedmem_coma->pdev->dev,
			"sharedmem service not yet initialized\n");
		return -1;
	}

	params.init.virt_base = (unsigned long)sharedmem_coma->base;
	params.init.phys_base = (unsigned long)sharedmem_coma->base;
	params.init.size = (unsigned int)size;
	ret = sharedmem_coma_create_message(CMSG_SHAREDMEM_INIT, &params);
	if (ret != 0) {
		dev_err(&sharedmem_coma->pdev->dev,
			"coma communication failed: %d\n", ret);
		return ret;
	}

	timedout = wait_for_completion_timeout(&sharedmem_coma->done,  HZ); //msecs_to_jiffies(INIT_TIMEOUT));

	if (sharedmem_coma->status) {
		dev_err(&sharedmem_coma->pdev->dev,
			"initialization (MMU mapping) in CSS failed: %d\n",
			sharedmem_coma->status);
		ret = -1;
	}

	if (timedout <= 0) {
		dev_err(&sharedmem_coma->pdev->dev,
			"initialization (MMU mapping) in CSS timed out\n");
		ret = -1;
	}

	sharedmem_coma_debugfs_create_files(sharedmem_coma);

	return ret;
}

void *
sharedmem_coma_alloc(size_t size, dma_addr_t *handle)
{
	void *addr;

	if (sharedmem_coma->service_id < 0) {
		dev_err(&sharedmem_coma->pdev->dev,
			"sharedmem service not yet initialized\n");
		return NULL;
	}

	mutex_lock(&sharedmem_coma->lock);

	addr = dma_alloc_coherent(sharedmem_coma->css_dev, size, handle,
				  GFP_KERNEL);

	mutex_unlock(&sharedmem_coma->lock);

	return addr;
}
EXPORT_SYMBOL(sharedmem_coma_alloc);

int
sharedmem_coma_free(size_t size, void *cpu_addr, dma_addr_t handle)
{
	if (sharedmem_coma->service_id < 0)
		return -EFAULT;

	mutex_lock(&sharedmem_coma->lock);

	dma_free_coherent(sharedmem_coma->css_dev, size, cpu_addr, handle);

	mutex_unlock(&sharedmem_coma->lock);

	return 0;
}
EXPORT_SYMBOL(sharedmem_coma_free);

int
sharedmem_is_shared(void *addr)
{
	if (sharedmem_coma->base &&
	    ((dma_addr_t)addr >= sharedmem_coma->base) &&
	    ((dma_addr_t)addr <  sharedmem_coma->base +
	    (CONFIG_DVF_RESERVEMEM_CSS*SZ_1M)))
		return 1;
	return 0;
}
EXPORT_SYMBOL(sharedmem_is_shared);

static void
sharedmem_coma_process_message(void *arg, struct cmsg *cmsg)
{
	int ret = 0;
	union cmsg_sharedmem_params *params =
			(union cmsg_sharedmem_params *)cmsg_params(cmsg);

	switch (cmsg->type) {
	case CMSG_SHAREDMEM_INIT_REPLY:
		sharedmem_coma->status = params->init_reply.status;
		complete(&sharedmem_coma->done);
		break;
	default:
		dev_err(&sharedmem_coma->pdev->dev, "invalid message type\n");
		return;
	}

	if (ret)
		dev_err(&sharedmem_coma->pdev->dev, "coma_cmsg_send failed\n");
}

static void
sharedmem_coma_remove_service(void *arg)
{
	mutex_lock(&sharedmem_coma->lock);
	sharedmem_coma->service_id = -1;
	sharedmem_coma_debugfs_remove_files(sharedmem_coma);
	mutex_unlock(&sharedmem_coma->lock);
}

int sharedmem_coma_service_init(struct device *dev, dma_addr_t dma_base, size_t size)
{
	int ret;

	if (!sharedmem_coma)
		return -ENODEV;

	mutex_lock(&sharedmem_coma->lock);

	sharedmem_coma->css_dev = dev;
	sharedmem_coma->base = dma_base;

	if (sharedmem_coma->service_id >= 0) {
		ret = -EBUSY;
		goto out;
	}

	ret = coma_register("sharedmem", 1024, NULL,
			    sharedmem_coma_process_message,
			    sharedmem_coma_remove_service, NULL);
	if (ret < 0) {
		dev_err(&sharedmem_coma->pdev->dev, "coma_register failed\n");
		goto out;
	}
	sharedmem_coma->service_id = ret;

	if (sharedmem_coma_initialize(size) < 0) {
		ret = -1;
		goto out;
	}

	ret = 0;
out:
	mutex_unlock(&sharedmem_coma->lock);
	return ret;
}
EXPORT_SYMBOL(sharedmem_coma_service_init);

static const struct vm_operations_struct mmap_mem_ops = {
};

static int
sharedmem_coma_mmap(struct file *file, struct vm_area_struct *vma)
{
	size_t size = vma->vm_end - vma->vm_start;

	if (!sharedmem_coma->base) {
		dev_err(&sharedmem_coma->pdev->dev, "shared memory area not available\n");
		return -ENODEV;
	}

	if ((vma->vm_start < sharedmem_coma->base) ||
	    (vma->vm_end   > sharedmem_coma->base +
	    (CONFIG_DVF_RESERVEMEM_CSS*SZ_1M))) {
		dev_err(&sharedmem_coma->pdev->dev, "invalid memory range\n");
		return -EIO;
	}

	vma->vm_pgoff = vma->vm_start >> PAGE_SHIFT;
	vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
	vma->vm_ops = &mmap_mem_ops;

	/* Remap-pfn-range will mark the range VM_IO and VM_RESERVED */
	if (remap_pfn_range(vma,
			    vma->vm_start,
			    vma->vm_pgoff,
			    size,
			    vma->vm_page_prot)) {
		dev_err(&sharedmem_coma->pdev->dev, "remap failed\n");
		return -EAGAIN;
	}

	return 0;
}

static long
sharedmem_coma_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	int size;
	void __user *argp = (void __user *)arg;
	struct sharedmem_coma_handle handle;

	if (_IOC_TYPE(cmd) != SHAREDMEM_IOC_MAGIC)
		return -EINVAL;

	if (_IOC_NR(cmd) > SHAREDMEM_IOC_MAXNR)
		return -EINVAL;

	size = _IOC_SIZE(cmd);

	if ((_IOC_DIR(cmd) & _IOC_READ) && !access_ok(VERIFY_WRITE, argp, size))
		return -EFAULT;

	if ((_IOC_DIR(cmd) & _IOC_WRITE) && !access_ok(VERIFY_READ, argp, size))
		return -EFAULT;

	switch (cmd) {
	case SHAREDMEM_IOCALLOC:
	{
		void *addr;

		if (copy_from_user(&handle, argp, sizeof(handle)))
			return -EFAULT;
		addr = sharedmem_coma_alloc(handle.size,
					    (dma_addr_t *)&handle.pa);
		handle.addr = (unsigned long)addr;
		return copy_to_user(argp, &handle, sizeof(handle));
	}

	case SHAREDMEM_IOCFREE:
	{
		if (copy_from_user(&handle, argp, sizeof(handle)))
			return -EFAULT;
		return sharedmem_coma_free(handle.size,
					   (void *)handle.addr,
					   handle.pa);
	}

	default:
		return -EINVAL;
	}
}

static int
sharedmem_coma_open(struct inode *inode, struct file *file)
{
	return 0;
}

static int
sharedmem_coma_release(struct inode *inode, struct file *file)
{
	return 0;
}

static struct file_operations sharedmem_coma_fops = {
	.owner		= THIS_MODULE,
	.unlocked_ioctl	= sharedmem_coma_ioctl,
	.open		= sharedmem_coma_open,
	.release	= sharedmem_coma_release,
	.mmap		= sharedmem_coma_mmap,
};

static int
sharedmem_coma_probe(struct platform_device *pdev)
{
	int ret;

	sharedmem_coma = kzalloc(sizeof *sharedmem_coma, GFP_KERNEL);
	if (!sharedmem_coma) {
		dev_err(&pdev->dev, "Out of memory\n");
		return -ENOMEM;
	}
	sharedmem_coma->pdev = pdev;
	sharedmem_coma->service_id = -1;

	ret = alloc_chrdev_region(&sharedmem_coma->chrdev, 0, 1,
				  COMA_SHAREDMEM);
	if (ret < 0) {
		dev_err(&pdev->dev,
			"error %d adding character device region\n", ret);
		return ret;
	}

	cdev_init(&sharedmem_coma->cdev, &sharedmem_coma_fops);
	sharedmem_coma->cdev.owner = THIS_MODULE;
	ret = cdev_add(&sharedmem_coma->cdev, sharedmem_coma->chrdev, 1);
	if (ret) {
		dev_err(&pdev->dev, "error %d adding character device", ret);
		return ret;
	}

	/* Create a sysfs class */
	sharedmem_coma->class = class_create(THIS_MODULE, "sharedmem");
	if (IS_ERR(sharedmem_coma->class)) {
		ret = PTR_ERR(sharedmem_coma->class);
		dev_err(&pdev->dev, "error %d creating the voice class", ret);
		return ret;
	}

	sharedmem_coma->dev = device_create(sharedmem_coma->class,
					    &platform_bus,
					    MKDEV(MAJOR(sharedmem_coma->chrdev),
					    0), NULL, "sharedmem");
	if (IS_ERR(sharedmem_coma->dev)) {
		ret = PTR_ERR(sharedmem_coma->dev);
		dev_err(&pdev->dev, "could not create device");
		return ret;
	}

	mutex_init(&sharedmem_coma->lock);
	init_completion(&sharedmem_coma->done);

	platform_set_drvdata(pdev, sharedmem_coma);

	dev_info(&pdev->dev, "successfully registered\n");

	return 0;
}

static int __exit
sharedmem_coma_remove(struct platform_device *dev)
{
	/* FIXME: racy wrt. sharedmem_coma_remove_service callback */
	if (sharedmem_coma->service_id >= 0) {
		coma_deregister(sharedmem_coma->service_id);
		sharedmem_coma_remove_service(NULL);
	}
	kfree(sharedmem_coma);
	sharedmem_coma = NULL;

	return 0;
}

static struct of_device_id sharedmem_coma_of_ids[] = {
	{ .compatible = "dspg,sharedmem-coma" },
	{ },
};

static struct platform_driver sharedmem_coma_platform_driver = {
	.driver = {
		.name	= COMA_SHAREDMEM,
		.owner	= THIS_MODULE,
		.of_match_table = sharedmem_coma_of_ids,
	},
	.probe		= sharedmem_coma_probe,
	.remove		= __exit_p(sharedmem_coma_remove),
};

static int __init sharedmem_coma_init(void)
{
	return platform_driver_register(&sharedmem_coma_platform_driver);
}

static void __exit sharedmem_coma_exit(void)
{
	platform_driver_unregister(&sharedmem_coma_platform_driver);
}

module_init(sharedmem_coma_init);
module_exit(sharedmem_coma_exit);

MODULE_DESCRIPTION("DSPG SHAREDMEM-COMA driver");
MODULE_LICENSE("GPL v2");
MODULE_AUTHOR("DSP Group");
MODULE_ALIAS("platform:sharedmem-coma");
