/*
 *  drivers/cordless/dev-voice.c - voice character device
 *
 *  This driver registers one character device for each available voice channel.
 *  It allows only one reader and one writer at the same time.
 *
 *  Copyright (C) 2007 NXP Semiconductors
 *  Copyright (C) 2008, 2009 DSPG Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* TODO:
 *  - if minor numbers dont start at zero, this code will fail
 *  - read/write several rtp packets at once
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/err.h>
#include <linux/cdev.h>
#include <linux/uio.h>
#include <linux/poll.h>
#include <linux/fs.h>
#include <linux/wait.h>
#include <linux/mutex.h>
#include <linux/kthread.h>
#include <linux/net.h>
#include <linux/file.h>
#include <linux/completion.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/inetdevice.h>
#include <linux/netdevice.h>
#include <linux/platform_device.h>
#include <linux/coma/coma.h>
#include <linux/coma/cfifo.h>
#include <asm/uaccess.h>
#include <asm/types.h>

#include "voice.h"
#include "cmsg-voice.h"
#include "coma-voice.h"

MODULE_AUTHOR("DSP Group Inc.");
MODULE_LICENSE("GPL");

#define SERVICE_NAME  "voice"
#define DEFAULT_CODEC RTP_SESSION_PT_G722
#define FIFO_SIZE     (4096*5)

static dev_t voice_chrdev;
static struct cdev voice_cdev;
static struct class *voice_class;
static struct device *voice_dev[CONFIG_CORDLESS_NUM_VOICELINES];
static int voice_reg;
static int service_id;
const unsigned int voice_numlines = CONFIG_CORDLESS_NUM_VOICELINES;

static DECLARE_COMPLETION(voice_reply_compl);
static DEFINE_MUTEX(voice_mutex);
static DEFINE_MUTEX(lock);

static struct cmsg_voice_params last_cmsg_voice_params;
static struct task_struct *kmode_thread = NULL;

enum voice_chan_status {
	CHANNEL_INVALID = 0,
	CHANNEL_INITIALIZED,
	CHANNEL_STARTED,
};

/**
 * struct voice_chan - all the info needed to run one audio channel
 *
 * @enc_fifo:  pointer to cfifo used for the encoder
 * @dec_fifo:  pointer to cfifo used for the decoder
 * @reader:    pointer to the reading process
 * @writer:    pointer to the writing process
 *
 */
struct voice_chan {
	struct cfifo *enc_fifo;
	struct cfifo *dec_fifo;
	struct file *reader;
	struct file *writer;
	wait_queue_head_t enc_wq;
	enum voice_chan_status status;
	int id;
	struct mutex codec_mutex;
	int flushed;
	struct socket *sock;
	struct sockaddr_storage remote_addr;
	struct socket *sock_rtcp;
	struct sockaddr_storage remote_addr_rtcp;
	struct msghdr txmsg;
	struct msghdr rxmsg;
	/* dtmf handling */
	rtp_dtmf_event dtmf_event;
	struct mutex dtmf_event_mutex;
	wait_queue_head_t dtmf_wq;
	enum rtcp_mode rtcpmode;
	int isSymmetricRTP;
	rtp_session_config config;
	struct timeval last_timestamp;
	unsigned long last_rtptime;
	struct mutex sock_read_write_mutex;
};

static struct voice_chan voice_channels[CONFIG_CORDLESS_NUM_VOICELINES];

static int
voice_registered(void)
{
	return (voice_reg != 0);
}

static void
voice_generic_reply(struct cmsg_voice_params *params)
{
	last_cmsg_voice_params = *params;
	complete(&voice_reply_compl);
}

static int
coma_create_voice_message(enum cmsg_voice_types type,
			  struct cmsg_voice_params *params,
			  void *payload, unsigned int payload_size)
{
	return coma_cmsg_send(service_id, (int)type,
			      (void *)params, sizeof(*params),
			      payload, payload_size);
}

/*
 * voice_chan specific function
 */
static void
voice_chan_reset(struct voice_chan *chan)
{
	mutex_lock(&voice_mutex);
	cfifo_reset(chan->enc_fifo);
	cfifo_reset(chan->dec_fifo);
	chan->reader = 0;
	chan->writer = 0;
	chan->status = CHANNEL_INVALID;
	chan->sock = NULL;
	chan->isSymmetricRTP = 0;
	mutex_unlock(&voice_mutex);
}

static int
voice_request_get_chan(int id)
{
	int ret;
	struct cmsg_voice_params params;

	if (!voice_registered())
		return -EFAULT;

	params.id = id;
	ret = coma_create_voice_message(CMSG_VOICE_REQUEST_GET_CHAN, &params,
					NULL, 0);

	return ret;
}

static int
voice_get_chan(struct voice_chan *chan)
{
	int ret;

	mutex_lock(&voice_mutex);

	ret = voice_request_get_chan(chan->id);
	if (ret < 0) {
		mutex_unlock(&voice_mutex);
		return -EFAULT;
	}

	wait_for_completion(&voice_reply_compl);

	ret = last_cmsg_voice_params.result;
	mutex_unlock(&voice_mutex);

	return ret;
}

static int
voice_request_set_chan_fifos(int id, struct cfifo *enc, struct cfifo *dec)
{
	int ret;
	struct cmsg_voice_params params;

	if (!voice_registered())
		return -EFAULT;

	params.id = id;
	params.config.set_chan_fifos.enc = (unsigned int)cfifo_phys(enc);
	params.config.set_chan_fifos.dec = (unsigned int)cfifo_phys(dec);

	ret = coma_create_voice_message(CMSG_VOICE_REQUEST_SET_CHAN_FIFOS,
					&params, NULL, 0);

	return ret;
}

static int
voice_set_fifos(struct voice_chan *chan)
{
	int ret;

	mutex_lock(&voice_mutex);

	ret = voice_request_set_chan_fifos(chan->id, chan->enc_fifo,
					   chan->dec_fifo);
	if (ret < 0) {
		mutex_unlock(&voice_mutex);
		return -EFAULT;
	}

	wait_for_completion(&voice_reply_compl);

	ret = last_cmsg_voice_params.result;
	mutex_unlock(&voice_mutex);

	return ret;
}

static int
voice_request_start_chan(int id, rtp_session_config *config)
{
	int ret;
	struct cmsg_voice_params params;

	if (!voice_registered())
		return -EFAULT;

	params.id = id;
	memcpy(&params.config.start_chan,config,sizeof(rtp_session_config));

	ret = coma_create_voice_message(CMSG_VOICE_REQUEST_START_CHAN,
					&params, NULL, 0);

	return ret;
}

static int
voice_start_chan(struct voice_chan *chan)
{
	int ret;

	mutex_lock(&voice_mutex);

	ret = voice_request_start_chan(chan->id, &chan->config);
	if (ret < 0) {
		mutex_unlock(&voice_mutex);
		return -EFAULT;
	}

	wait_for_completion(&voice_reply_compl);

	ret = last_cmsg_voice_params.result;
	mutex_unlock(&voice_mutex);

	return ret;
}

static int
voice_request_stop_chan(int id)
{
	int ret;
	struct cmsg_voice_params params;

	if (!voice_registered())
		return -EFAULT;

	params.id = id;

	ret = coma_create_voice_message(CMSG_VOICE_REQUEST_STOP_CHAN, &params,
					NULL, 0);

	return ret;
}

static int
voice_stop_chan(struct voice_chan *chan)
{
	int ret;

	mutex_lock(&voice_mutex);

	ret = voice_request_stop_chan(chan->id);
	if (ret < 0) {
		mutex_unlock(&voice_mutex);
		return -EFAULT;
	}

	wait_for_completion(&voice_reply_compl);

	ret = last_cmsg_voice_params.result;
	mutex_unlock(&voice_mutex);

	return ret;
}

static int
voice_request_start_rtcp(int id, struct voice_kernelmode_rtcp *km_rtcp)
{
	int ret;

	struct cmsg_voice_params params;

	if (!voice_registered())
		return -EFAULT;

	params.id = id;
	memcpy(&params.config.start_rtcp, &km_rtcp->rtcpCfg,
	       sizeof(rtcp_session_config));
	ret = coma_create_voice_message(CMSG_VOICE_REQUEST_START_RTCP,
					&params, NULL, 0);

	return ret;
}

static int
voice_request_update_chan(int id, rtp_session_config *mediaParams)
{
	struct cmsg_voice_params params;

	if (!voice_registered())
		return -EFAULT;

	params.id = id;

	/* update configuration */
	memcpy(&params.config.update_chan, mediaParams,
	       sizeof(rtp_session_config));

	return coma_create_voice_message(CMSG_VOICE_REQUEST_UPDATE_CHAN,
					 &params, NULL, 0);
}

static int
voice_start_rtcp(struct voice_chan *chan, struct voice_kernelmode_rtcp *km_rtcp)
{
	int ret;

	mutex_lock(&voice_mutex);

	ret = voice_request_start_rtcp(chan->id, km_rtcp);
	if (ret < 0) {
		mutex_unlock(&voice_mutex);
		return -EFAULT;
	}

	wait_for_completion(&voice_reply_compl);

	ret = last_cmsg_voice_params.result;
	mutex_unlock(&voice_mutex);

	return ret;
}

static int
voice_request_stop_rtcp(int id)
{
	int ret;
	struct cmsg_voice_params params;

	if (!voice_registered())
		return -EFAULT;

	params.id = id;

	ret = coma_create_voice_message(CMSG_VOICE_REQUEST_STOP_RTCP, &params,
					NULL, 0);

	return ret;
}

static int
voice_stop_rtcp(struct voice_chan *chan)
{
	int ret;

	mutex_lock(&voice_mutex);

	ret = voice_request_stop_rtcp(chan->id);
	if (ret < 0) {
		mutex_unlock(&voice_mutex);
		return -EFAULT;
	}

	wait_for_completion(&voice_reply_compl);

	ret = last_cmsg_voice_params.result;
	mutex_unlock(&voice_mutex);

	return ret;
}

static int
voice_request_free_chan(int id)
{
	int ret;
	struct cmsg_voice_params params;

	if (!voice_registered())
		return -EFAULT;

	params.id = id;

	ret = coma_create_voice_message(CMSG_VOICE_REQUEST_FREE_CHAN, &params,
					NULL, 0);

	return ret;
}

static int
voice_free_chan(struct voice_chan *chan)
{
	int ret;

	mutex_lock(&voice_mutex);

	ret = voice_request_free_chan(chan->id);
	if (ret < 0) {
		mutex_unlock(&voice_mutex);
		return -EFAULT;
	}

	wait_for_completion(&voice_reply_compl);

	ret = last_cmsg_voice_params.result;
	mutex_unlock(&voice_mutex);

	return ret;
}

static int
voice_chan_init(struct voice_chan *chan)
{
	int ret = 0, i;

	init_waitqueue_head(&chan->enc_wq);

	for (i = 0; i < MAX_NUM_OF_CODEC; i++) {
		chan->config.codec.rx_list[i].rx_pt = 0xff;
		memset(chan->config.codec.rx_list[i].CodecStr, 0,
		       sizeof(chan->config.codec.rx_list[i].CodecStr));
	}
	chan->config.codec.rx_list[0].rx_pt = DEFAULT_CODEC;
	chan->config.codec.rx_pt_event = 0xff;
	chan->config.codec.tx_pt = DEFAULT_CODEC;
	chan->config.codec.tx_pt_event = 0xff;
	chan->config.codec.duration = DEFAULT_DURATION;
	chan->config.codec.opts = 0;
	mutex_init(&chan->codec_mutex);
	mutex_init(&chan->dtmf_event_mutex);
	init_waitqueue_head(&chan->dtmf_wq);

	ret = voice_get_chan(chan);
	if (ret != 0)
		return ret;

	ret = voice_set_fifos(chan);
	if (ret != 0)
		goto err_set_fifos;

	chan->status = CHANNEL_INITIALIZED;

	return 0;

err_set_fifos:
	voice_free_chan(chan);

	return ret;
}


static inline unsigned long timeval_to_rtp(const struct timeval *tv)
{
	return (tv->tv_sec * 8000) + (tv->tv_usec / 125);
}

/*
 * kernelmode
 */
static char kthread_buf[1492];

static void
voice_kmode_process_send(struct voice_chan *chan, struct msghdr *msg)
{
	struct kvec vec;
	unsigned char *enc_buf;
	int to_send = 0, sent = 0;
	struct rtp_packet_header *packet_header = NULL;

	/* kernel mode is on */
	while ((to_send = cfifo_get(chan->enc_fifo, (void **)&enc_buf)) > 0) {
		/* send data */
		packet_header = (struct rtp_packet_header *)enc_buf;
		to_send = to_send - sizeof(struct rtp_packet_header);
		vec.iov_base = enc_buf + sizeof(struct rtp_packet_header);
		vec.iov_len = to_send;
		/* send to appropriate destination according to RTP/RTCP flag */
		msg->msg_namelen = sizeof(chan->remote_addr);
		
		/* If RTCP-MUX flag is enabled, Send RTCP Packet on the same RTP port*/
                if (chan->config.rtcp_mux){
                        msg->msg_name = &chan->remote_addr;
                        sent = kernel_sendmsg(chan->sock, msg, &vec, 1, to_send);
                } else {
			if (packet_header->packetType == CFIFO_RTP_PACKET) {
				msg->msg_name = &chan->remote_addr;
				sent = kernel_sendmsg(chan->sock, msg, &vec, 1, to_send);
			} else if (packet_header->packetType == CFIFO_RTCP_PACKET) {
				msg->msg_name = &chan->remote_addr_rtcp;
				sent = kernel_sendmsg(chan->sock_rtcp, msg, &vec, 1, to_send);
			}
		}
		if (sent < 0)
			pr_err("%s: sock_sendmsg() failed with error %d "
			       "(channel %d)\n", SERVICE_NAME, chan->id, sent);
		else if (sent != to_send)
			pr_err("%s: failed to send %d bytes (channel %d)\n",
			        SERVICE_NAME, chan->id, to_send - sent);

		cfifo_processed(chan->enc_fifo);
	}
}

static int
voice_kmode_process_recv(struct voice_chan *chan, struct msghdr *msg)
{
	struct kvec vec;
	unsigned char *dec_buf;
	int recvd = 0;
	struct timeval timestamp;
	unsigned long timediff;
	struct rtp_packet_header *packet_header = NULL;
	unsigned char ctrl_buf[100];
	struct cmsghdr *cmsg;
	int ttl = 0;
	RTP_HDR *pRtpHdr = NULL;	

	if (chan->config.codec.opts & RTP_ENABLE_SYMMETRIC_RSP) {
		msg->msg_name = &chan->remote_addr;
		msg->msg_namelen = sizeof(chan->remote_addr);
	}
	msg->msg_control = ctrl_buf;
	msg->msg_controllen = sizeof(ctrl_buf);
	memset(&ctrl_buf, 0, sizeof(ctrl_buf));

	while (1) {
		vec.iov_base = kthread_buf;
		vec.iov_len  = sizeof(kthread_buf);
		recvd = kernel_recvmsg(chan->sock, msg, &vec, 1,
				       sizeof(kthread_buf), MSG_DONTWAIT);
		if (recvd <= 0)
			break;
		
		do_gettimeofday(&timestamp);
		timediff = timeval_to_rtp(&timestamp) -
			   timeval_to_rtp(&chan->last_timestamp);
		chan->last_timestamp.tv_sec = timestamp.tv_sec;
		chan->last_timestamp.tv_usec = timestamp.tv_usec;

		dec_buf = cfifo_request(chan->dec_fifo, recvd +
					sizeof(struct rtp_packet_header));
		if (IS_ERR(dec_buf))
			return NF_DROP; /* drop the packet */

		cmsg = (struct cmsghdr *)ctrl_buf;
		msg->msg_control = ctrl_buf;
		msg->msg_controllen = sizeof(ctrl_buf);
		for (cmsg = CMSG_FIRSTHDR(msg); cmsg; cmsg = CMSG_NXTHDR(msg, cmsg)) {
			if (!CMSG_OK(msg, cmsg))
				break;

			if (cmsg->cmsg_level != SOL_IP)
				continue;

			if (cmsg->cmsg_type != IP_TTL)
				continue;

			ttl = *(int *)CMSG_DATA(cmsg);
		}

		packet_header = (struct rtp_packet_header *)dec_buf;
		packet_header->ttl = ttl;

		if (chan->config.rtcp_mux) {
			pRtpHdr = (RTP_HDR *)kthread_buf;
			if (pRtpHdr->bMarkerPayload == RTCP_TYPE_SR 
				|| pRtpHdr->bMarkerPayload == RTCP_TYPE_RR
					|| pRtpHdr->bMarkerPayload == RTCP_TYPE_XR)
				packet_header->packetType = CFIFO_RTCP_PACKET;
		} else {
			packet_header->packetType = CFIFO_RTP_PACKET;
		}

		chan->last_rtptime += timediff;
		packet_header->receiptTime = chan->last_rtptime;
		memcpy(dec_buf + sizeof(struct rtp_packet_header),
		       kthread_buf, recvd);
		cfifo_commit(chan->dec_fifo);
	}

	/* Try to Recv RTCP packets if any */
	if (!chan->sock_rtcp)
		return 0;

	if (chan->config.codec.opts & RTP_ENABLE_SYMMETRIC_RSP) {
		msg->msg_name = &chan->remote_addr_rtcp;
		msg->msg_namelen = sizeof(chan->remote_addr_rtcp);
	}

	while (1) {
		vec.iov_base = kthread_buf;
		vec.iov_len  = sizeof(kthread_buf);
		recvd = kernel_recvmsg(chan->sock_rtcp, msg, &vec, 1,
				       sizeof(kthread_buf), MSG_DONTWAIT);
		if (recvd <= 0)
			break;

		/* Add extra space to distinguish RTP and RTCP */
		dec_buf = cfifo_request(chan->dec_fifo, recvd +
					sizeof(struct rtp_packet_header));
		if (IS_ERR(dec_buf))
			return NF_DROP; /* drop the packet */

		cmsg = (struct cmsghdr *)ctrl_buf;
		msg->msg_control = ctrl_buf;
		msg->msg_controllen = sizeof(ctrl_buf);
		for (cmsg = CMSG_FIRSTHDR(msg); cmsg; cmsg = CMSG_NXTHDR(msg, cmsg)) {
			if (!CMSG_OK(msg, cmsg))
				break;

			if (cmsg->cmsg_level != SOL_IP)
				continue;

			if (cmsg->cmsg_type != IP_TTL)
				continue;

			ttl = *(int *)CMSG_DATA(cmsg);
		}

		packet_header = (struct rtp_packet_header *)dec_buf;
		packet_header->packetType = CFIFO_RTCP_PACKET;
		packet_header->ttl = ttl;
		packet_header->receiptTime = 0;
		memcpy(dec_buf + sizeof(struct rtp_packet_header),
		       kthread_buf, recvd);
		cfifo_commit(chan->dec_fifo);
	}

	return 0;
}

static void
voice_kmode_process_recv_loop(struct voice_chan *chan, struct msghdr *msg)
{
	struct kvec vec;
	int recvd = 0;
	int sent = 0;

	if (chan->config.codec.opts & RTP_ENABLE_SYMMETRIC_RSP) {
		msg->msg_name = &chan->remote_addr;
		msg->msg_namelen = sizeof(chan->remote_addr);
	}

	while (1) {
		vec.iov_base = kthread_buf;
		vec.iov_len  = sizeof(kthread_buf);
		recvd = kernel_recvmsg(chan->sock, msg, &vec, 1,
		                       sizeof(kthread_buf), MSG_DONTWAIT);
		if (recvd <= 0)
			break;

		if (recvd > 0) {
			vec.iov_base = kthread_buf;
			vec.iov_len  = sizeof(kthread_buf);

			msg->msg_namelen = sizeof(chan->remote_addr);
			msg->msg_name = &chan->remote_addr;

			sent = kernel_sendmsg(chan->sock, msg, &vec, 1, recvd);

			if (sent < 0)
				pr_err("%s: sock_sendmsg() failed with error %d "
				       "(channel %d)\n", SERVICE_NAME, chan->id, sent);
			else if (sent != recvd)
				pr_err("%s: failed to send %d bytes (channel %d)\n",
				       SERVICE_NAME, chan->id, recvd - sent);
		}
	}
}

static int
voice_kmode_thread(void *data)
{
	struct voice_chan *chan;
	struct task_struct *tsk = current;
	struct sched_param param = {.sched_priority = 1 };
	int i;

	sched_setscheduler(tsk, SCHED_FIFO, &param);

	while (!kthread_should_stop()) {
		for (i = 0; i < voice_numlines ; i++) {
			chan = &voice_channels[i];
			mutex_lock(&chan->sock_read_write_mutex);
			if (chan->status != CHANNEL_STARTED) {
				mutex_unlock(&chan->sock_read_write_mutex);
				continue;
			}

			if ((chan->config.media_loop_level == RTP_LOOP_IP_LEVEL)) {
				voice_kmode_process_recv_loop(chan, &chan->rxmsg);
				mutex_unlock(&chan->sock_read_write_mutex);
				continue;
			}

			if (chan->sock) {
				voice_kmode_process_send(chan, &chan->txmsg);
				if (voice_kmode_process_recv(chan, &chan->rxmsg) < 0)
					pr_warning("%s: FIFO full - dropping incoming RTP packet\n", SERVICE_NAME);
			} else {
				if (!cfifo_empty(chan->enc_fifo))
					wake_up(&chan->enc_wq);
			}
			mutex_unlock(&chan->sock_read_write_mutex);
		}
		usleep_range(4000, 4500);
	}

	return 0;
}

static int
voice_kmode_start(struct voice_chan *chan, struct voice_kernelmode *km)
{
	int ret = 0;
	int ttl = 1;

	/* check if kernelmode is already active */
	if (chan->sock) {
		pr_err("%s: start: kmode active\n", SERVICE_NAME);
		return -1;
	}

	/* set up kmode structure */
	chan->remote_addr = km->remote_addr;
	chan->sock = sockfd_lookup(km->sock_fd, &ret);
	chan->sock_rtcp  = 0;
	memset(&chan->txmsg, 0, sizeof(chan->txmsg));
	memset(&chan->rxmsg, 0, sizeof(chan->rxmsg));

	if (!chan->sock) {
		pr_err("%s: start: socket invalid\n", SERVICE_NAME);
		return -1;
	}
	kernel_setsockopt(chan->sock, IPPROTO_IP, IP_RECVTTL, (char *)&ttl, 1);

	cfifo_reset(chan->enc_fifo);

	return ret;
}

static void
voice_kmode_stop(struct voice_chan *chan)
{
	if (chan->sock) {
		sockfd_put(chan->sock);
		chan->sock = NULL;
	}

	if (chan->sock_rtcp) {
		sockfd_put(chan->sock_rtcp);
		chan->sock_rtcp = NULL;
	}
}

static int
voice_kmode_start_rtcp(struct voice_chan *chan,
                       struct voice_kernelmode_rtcp *km_rtcp)
{
	int ret = 0;

	if (km_rtcp->rtcpCfg.opts & RTCP_USER_MODE) {
		/* notify cordless */
		voice_start_rtcp(chan, km_rtcp);
		return 0;
	}
	chan->remote_addr_rtcp = km_rtcp->remote_addr;
	chan->sock_rtcp = sockfd_lookup(km_rtcp->sock_fd, &ret);
	
	/* notify cordless */
	voice_start_rtcp(chan, km_rtcp);

	return 0;
}

static void
voice_kmode_stop_rtcp(struct voice_chan *chan)
{
	if (chan->sock_rtcp) {
		sockfd_put(chan->sock_rtcp);
		chan->sock_rtcp = NULL;
	}
	/* notify cordless */
	voice_stop_rtcp(chan);
}

static int
voice_kmode_update_chan(struct voice_chan *chan,
			rtp_session_config *mediaParams)
{
	int ret;

	/* notify cordless */
	mutex_lock(&voice_mutex);

	ret = voice_request_update_chan(chan->id, mediaParams);
	if (ret < 0) {
		mutex_unlock(&voice_mutex);
		return -EFAULT;
	}

	wait_for_completion(&voice_reply_compl);
	ret = last_cmsg_voice_params.result;

	mutex_unlock(&voice_mutex);

	return ret;
}

/*
 * dtmf handling
 */
static void
voice_receive_dtmf(int id, RTP_EVENT *dtmf_event)
{
	struct voice_chan *chan = &voice_channels[id];

	mutex_lock(&chan->dtmf_event_mutex);
	memcpy(&chan->dtmf_event, dtmf_event, sizeof(rtp_dtmf_event));
	mutex_unlock(&chan->dtmf_event_mutex);
	wake_up(&chan->dtmf_wq);
}

static int
voice_request_send_dtmf(int id, rtp_dtmf_event *dtmf_event)
{
	int ret;
	struct cmsg_voice_params params;

	if (!voice_registered())
		return -EFAULT;

	params.id = id;
	memcpy(&params.config.send_dtmf, dtmf_event, sizeof(rtp_dtmf_event));
	ret = coma_create_voice_message(CMSG_VOICE_REQUEST_SEND_DTMF, &params,
					NULL, 0);

	return ret;
}

static int
voice_request_send_evt(int id, rtp_generic_event *evt)
{
	int ret;
	struct cmsg_voice_params params;

	if (!voice_registered())
		return -EFAULT;

	params.id = id;
	memcpy(&params.config.send_evt, evt, sizeof(rtp_generic_event));
	ret = coma_create_voice_message(CMSG_VOICE_REQUEST_SEND_EVT, &params,
					NULL, 0);

	return ret;
}

static int
voice_send_dtmf(struct voice_chan *chan, rtp_dtmf_event *new)
{
	int ret;

	mutex_lock(&voice_mutex);

	ret = voice_request_send_dtmf(chan->id, new);
	if (ret < 0) {
		mutex_unlock(&voice_mutex);
		return -EFAULT;
	}

	wait_for_completion(&voice_reply_compl);

	ret = last_cmsg_voice_params.result;
	mutex_unlock(&voice_mutex);

	return ret;
}

static int
voice_send_generic_evt(struct voice_chan *chan, rtp_generic_event *new)
{
	int ret;
	mutex_lock(&voice_mutex);

	ret = voice_request_send_evt(chan->id, new);
	if (ret < 0) {
		mutex_unlock(&voice_mutex);
		return -EFAULT;
	}

	wait_for_completion(&voice_reply_compl);

	ret = last_cmsg_voice_params.result;
	mutex_unlock(&voice_mutex);

	return ret;
}

/*
 * coma handling
 */
static void
voice_process_message(void *arg, struct cmsg *cmsg)
{
	int ret = 0;
	struct cmsg_voice_params *params =
				(struct cmsg_voice_params *)cmsg_params(cmsg);

	switch (cmsg->type) {
	case CMSG_VOICE_REPLY_GET_CHAN:
	case CMSG_VOICE_REPLY_SET_CHAN_FIFOS:
	case CMSG_VOICE_REPLY_START_CHAN:
	case CMSG_VOICE_REPLY_STOP_CHAN:
	case CMSG_VOICE_REPLY_FREE_CHAN:
	case CMSG_VOICE_REPLY_START_RTCP:
	case CMSG_VOICE_REPLY_STOP_RTCP:
	case CMSG_VOICE_REPLY_SEND_DTMF:
	case CMSG_VOICE_REPLY_UPDATE_CHAN:
	case CMSG_VOICE_REPLY_SEND_EVT:
		voice_generic_reply(params);
		break;

	case CMSG_VOICE_RECEIVE_DTMF:
		voice_receive_dtmf(params->id,
		                   &params->config.receive_dtmf);
		break;
	default:
		/* Handle RTP and RTCP Packets*/
		ret = -1;
		break;
	}

	return;
}

/*
 * character device functions
 */
static long
voice_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	int size;
	int ret = 0;
	void __user *argp = (void __user *)arg;
	unsigned int minor = iminor(file->f_dentry->d_inode);
	struct voice_chan *chan = &voice_channels[minor];

	if (_IOC_TYPE(cmd) != VOICE_IOC_MAGIC)
		return -EINVAL;

	if (_IOC_NR(cmd) > VOICE_IOC_MAXNR)
		return -EINVAL;

	size = _IOC_SIZE(cmd);

	if ((_IOC_DIR(cmd) & _IOC_READ) && !access_ok(VERIFY_WRITE, argp, size))
		return -EFAULT;

	if ((_IOC_DIR(cmd) & _IOC_WRITE) && !access_ok(VERIFY_READ, argp, size))
		return -EFAULT;

	switch (cmd) {
	case VOICE_IOCSETCODEC:
	{
		mutex_lock(&chan->codec_mutex);
		if (copy_from_user(&chan->config, argp, sizeof(chan->config))) {
			mutex_unlock(&chan->codec_mutex);
			return -EFAULT;
		}
		mutex_unlock(&chan->codec_mutex);

		/*
		 * TODO: switch codec in stack, this will also do
		 *       the sanity checking for us (result in reply)
		 */

		/* update our local info if codec switch went ok */
		if (chan->status == CHANNEL_INVALID)
			return -EFAULT;

		if (chan->status == CHANNEL_STARTED)
			voice_stop_chan(chan);

		/* If Symmetric Pkt count is configured as zero set to default value */
		if (chan->config.SymmRTPTxPktCnt == 0)
			chan->config.SymmRTPTxPktCnt = SYMMETRIC_TX_RTP_CNT;
		ret = voice_start_chan(chan);
		if (ret < 0)
			return -EFAULT;

		chan->last_timestamp.tv_sec = 0;
		chan->last_timestamp.tv_usec = 0;
		chan->last_rtptime = 0;

		chan->status = CHANNEL_STARTED;
		return 0;
	}

	case VOICE_IOCSENDDTMF:
	{
		rtp_dtmf_event dtmf;

		if (copy_from_user(&dtmf, argp, sizeof(dtmf)))
			return -EFAULT;

		ret = voice_send_dtmf(chan, &dtmf);
		if (ret != 0)
			return -EFAULT;
		return 0;
	}

	case VOICE_IOCGENERIC_EVENT:
	{
		rtp_generic_event event;

		if (copy_from_user(&event, argp, sizeof(event)))
			return -EFAULT;

		ret = voice_send_generic_evt(chan, &event);
		if (ret != 0)
			return -EFAULT;

		return 0;
	}

	case VOICE_IOCGETDTMF:
	{
		return 0;
	}

	case VOICE_IOCFLUSH:
	{
		cfifo_reset(chan->enc_fifo);
		return 0;
	}

	case VOICE_IOCKERNELMODE:
	{
		struct voice_kernelmode km;
		mutex_lock(&chan->codec_mutex);

		/*
		 * if argp is set, start kernelmode, if its NULL,
		 * stop kernelmode
		 */
		if (argp) {
			if (copy_from_user(&km, argp, sizeof(km)))
				return -EFAULT;

			mutex_lock(&chan->sock_read_write_mutex);
			ret = voice_kmode_start(chan, &km);
			mutex_unlock(&chan->sock_read_write_mutex);
		} else {
			mutex_lock(&chan->sock_read_write_mutex);
			voice_kmode_stop(chan);
			mutex_unlock(&chan->sock_read_write_mutex);
		}
		mutex_unlock(&chan->codec_mutex);

		return ret;
	}

	case VOICE_IOCRTCP:
	{
		struct voice_kernelmode_rtcp km_rtcp;

		/* if argp is set, start rtcp, if its NULL, stop rtcp */
		if (argp) {
			if (copy_from_user(&km_rtcp, argp, sizeof(km_rtcp)))
				return -EFAULT;
			mutex_lock(&chan->sock_read_write_mutex);
			ret = voice_kmode_start_rtcp(chan, &km_rtcp);
			mutex_unlock(&chan->sock_read_write_mutex);
		} else {
			mutex_lock(&chan->sock_read_write_mutex);
			voice_kmode_stop_rtcp(chan);
			mutex_unlock(&chan->sock_read_write_mutex);
	}

		return 0;
	}

	case VOICE_IOCUPDATE_CHAN:
	{
		if (copy_from_user(&chan->config, argp, sizeof(chan->config)))
			return -EFAULT;

		ret = voice_kmode_update_chan(chan, &chan->config);
		if (ret < 0)
			return -EFAULT;

		return 0;
	}

	default: /* redundant, as cmd was checked against MAXNR */
		return -EINVAL;
	}
}

static int
voice_open(struct inode *inode, struct file *file)
{
	int ret = 0;
	unsigned int flags = file->f_flags & O_ACCMODE;
	unsigned int minor = iminor(inode);
	struct voice_chan *chan = &voice_channels[minor];

	if (!voice_registered())
		return -1;

	/* only allow one reader */
	if (flags == O_RDONLY || flags == O_RDWR) {
		if (chan->reader) {
			return -EBUSY;
		} else {
			chan->reader = file;
			file->private_data = chan;
		}
	}

	/* only allow one writer */
	if (flags == O_WRONLY || flags == O_RDWR) {
		if (chan->writer) {
			return -EBUSY;
		} else {
			chan->writer = file;
			file->private_data = chan;
		}
	}

	if (chan->status == CHANNEL_INVALID) {
		chan->id = minor;
		ret = voice_chan_init(chan);
	}

	return ret;
}

static int
voice_release(struct inode *inode, struct file *file)
{
	int ret = 0;
	unsigned int minor = iminor(inode);
	struct voice_chan *chan = &voice_channels[minor];

	if (!voice_registered())
		return -1;

	mutex_lock(&chan->sock_read_write_mutex);
	if (chan->reader == file)
		chan->reader = NULL;

	if (chan->writer == file)
		chan->writer = NULL;

	if (!chan->reader && !chan->writer) {
		voice_stop_chan(chan);
		if (chan->sock)
			voice_kmode_stop(chan);
		voice_free_chan(chan);
		voice_chan_reset(chan);
	}
	mutex_unlock(&chan->sock_read_write_mutex);

	return ret;
}

static ssize_t
voice_read(struct file *file, char __user *buf, size_t count_want,
           loff_t *f_pos)
{
	struct voice_chan *chan = (struct voice_chan *)file->private_data;
	unsigned char *enc_buf;
	size_t not_copied;
	ssize_t to_copy;
	ssize_t ret;

	if (!voice_registered())
		return -1;

	if (chan->sock)
		return -EBUSY;

	if (wait_event_interruptible(chan->enc_wq,
				     ((ret = cfifo_get(chan->enc_fifo,
				     (void **)&enc_buf)) > 0)))
		return -ERESTARTSYS;

	/* copy one rtp packet at a time */
	to_copy = ret;
	/* Need to check if we are returning more than what has been asked ??? */
	if(count_want < to_copy){
		pr_err("%s: coma-voice: Got more data of '%d' than requested, copying only requested '%d' \n" , SERVICE_NAME , to_copy , count_want);
		to_copy = count_want;
	}
	not_copied = copy_to_user(buf, enc_buf, to_copy);
	cfifo_processed(chan->enc_fifo);

	return (to_copy - not_copied);
}

static ssize_t
voice_write(struct file *file, const char __user *buf, size_t count,
            loff_t *f_pos)
{
	int ret = 0;
	struct voice_chan *chan = (struct voice_chan *)file->private_data;
	char *fifo_buf;

	if (!voice_registered())
		return -1;

	if (chan->sock)
		return -EBUSY;
	fifo_buf = cfifo_request(chan->dec_fifo, count);
	if (IS_ERR(fifo_buf))
		return -ENOMEM;

	ret = copy_from_user(fifo_buf, buf, count);
	if (ret != 0)
		return -1;

	cfifo_commit(chan->dec_fifo);

	return count;
}

static unsigned int
voice_poll(struct file *file, poll_table * event_list)
{
	struct voice_chan *chan = (struct voice_chan *)file->private_data;
	unsigned int mask = POLLOUT | POLLWRNORM; /* we always allow writing */

	if (!cfifo_empty(chan->enc_fifo))
		mask |= POLLIN | POLLRDNORM;

	poll_wait(file, &chan->enc_wq, event_list);

	return mask;
}

static struct file_operations voice_fops = {
	.owner   = THIS_MODULE,
	.unlocked_ioctl   = voice_ioctl,
	.open    = voice_open,
	.release = voice_release,
	.read    = voice_read,
	.write   = voice_write,
	.poll    = voice_poll,
};

static void
voice_release_all(void)
{
	int i;
	struct voice_chan *chan;

	if (voice_registered()) {
		for (i = 0; i < CONFIG_CORDLESS_NUM_VOICELINES; i++) {
			chan = &voice_channels[i];

			mutex_lock(&chan->sock_read_write_mutex);
			if (chan->sock)
				voice_kmode_stop(chan);

			if (chan->status == CHANNEL_STARTED) {
				voice_stop_chan(chan);
				voice_free_chan(chan);
			}
			voice_chan_reset(chan);
			mutex_unlock(&chan->sock_read_write_mutex);
		}
	}
}

static void
voice_remove(void *arg)
{
	int i;

	mutex_lock(&lock);

	if (kmode_thread) {
		kthread_stop(kmode_thread);
		kmode_thread = NULL;
	}

	if (voice_reg) {
		voice_reg = 0;

		voice_release_all();
		for (i = 0; i < CONFIG_CORDLESS_NUM_VOICELINES; i++) {
			device_unregister(voice_dev[i]);
		}
		class_destroy(voice_class);
		cdev_del(&voice_cdev);
		unregister_chrdev_region(voice_chrdev, voice_numlines);
	}

	for (i = 0; i < CONFIG_CORDLESS_NUM_VOICELINES; i++) {
		if (voice_channels[i].enc_fifo) {
			cfifo_unref(voice_channels[i].enc_fifo);
			voice_channels[i].enc_fifo = NULL;
		}
		if (voice_channels[i].dec_fifo) {
			cfifo_unref(voice_channels[i].dec_fifo);
			voice_channels[i].dec_fifo = NULL;
		}
	}
	mutex_unlock(&lock);
}

static int
voice_suspend(void)
{
	int i;

	for (i = 0; i < CONFIG_CORDLESS_NUM_VOICELINES; i++) {
		if (voice_channels[i].status != CHANNEL_INVALID)
			return -EINVAL;
	}

	return 0;
}

int
voice_init(struct device *dev)
{
	int i, ret = 0;

	mutex_lock(&lock);

	ret = alloc_chrdev_region(&voice_chrdev, 0, voice_numlines, SERVICE_NAME);
	if (ret) {
		pr_err("%s: coma-voice: alloc of chrdev returned %d\n",
		       SERVICE_NAME, ret);
		goto out;
	}

	cdev_init(&voice_cdev, &voice_fops);

	voice_cdev.owner = THIS_MODULE;
	voice_cdev.ops   = &voice_fops;

	ret = cdev_add(&voice_cdev, voice_chrdev, voice_numlines);
	if (ret) {
		pr_err("%s: error %d adding character device", SERVICE_NAME, ret);
		goto err_unreg_chrdev_reg;
	}

	/* Create a sysfs class */
	voice_class = class_create(THIS_MODULE, "voice");
	if (IS_ERR(voice_class)) {
		ret = PTR_ERR(voice_class);
		pr_err("%s: error %d creating the voice class", SERVICE_NAME, ret);
		goto err_del_cdev;
	}

	for (i = 0; i < CONFIG_CORDLESS_NUM_VOICELINES; i++) {
		voice_channels[i].enc_fifo = cfifo_alloc(NULL, FIFO_SIZE, NULL, 0);
		voice_channels[i].dec_fifo = cfifo_alloc(NULL, FIFO_SIZE, NULL, 0);
		if (!voice_channels[i].enc_fifo || !voice_channels[i].dec_fifo) {
			ret = -ENOMEM;
			pr_err("%s: out of memory\n", SERVICE_NAME);
			goto err_cfifo_free;
		}
		mutex_init(&voice_channels[i].sock_read_write_mutex);
		voice_dev[i] = device_create(voice_class, &platform_bus,
					     MKDEV(MAJOR(voice_chrdev), i),
					     NULL, "voice%d", i);
		if (IS_ERR(voice_dev[i])) {
			pr_err("%s: could not create device\n", SERVICE_NAME);
			goto err_cfifo_free;
		}
	}

	ret = coma_register(SERVICE_NAME, FIFO_SIZE, NULL,
			    voice_process_message, voice_remove, voice_suspend);
	if (ret < 0) {
		pr_err("%s: registration failed: %d\n", SERVICE_NAME, ret);
		goto err_cfifo_free;
	}

	service_id = ret;
	voice_reg = 1;

	/* create sender thread */
	kmode_thread = kthread_run(voice_kmode_thread, NULL, "kvoice");
	if (IS_ERR(kmode_thread)) {
		kmode_thread = NULL;
		pr_err("%s: could not start kernel thread\n", SERVICE_NAME);
		goto err_cfifo_free;
	}

	pr_info("%s: character device initialized (major=%d)\n",
	       SERVICE_NAME, MAJOR(voice_chrdev));

	ret = 0;
	goto out;

err_cfifo_free:
	for (i = 0; i < CONFIG_CORDLESS_NUM_VOICELINES; i++) {
		if (voice_channels[i].enc_fifo) {
			cfifo_unref(voice_channels[i].enc_fifo);
			voice_channels[i].enc_fifo = NULL;
		}
		if (voice_channels[i].dec_fifo) {
			cfifo_unref(voice_channels[i].dec_fifo);
			voice_channels[i].dec_fifo = NULL;
		}
		if (voice_dev[i]) {
			device_unregister(voice_dev[i]);
			voice_dev[i] = NULL;
		}
	}
	class_destroy(voice_class);

err_del_cdev:
	cdev_del(&voice_cdev);

err_unreg_chrdev_reg:
	unregister_chrdev_region(voice_chrdev, voice_numlines);
out:
	mutex_unlock(&lock);
	return ret;
}

EXPORT_SYMBOL(voice_init);
