/*
 * drivers/staging/dspg/coma/coma.c - the cordless manager
 *
 * The cordless manager is responsible for maintaining the main cfifo between
 * Linux and cordless which is used for loading and unloading cordless and for
 * registration of other services (e.g., debug output, VoIP/RTP, etc.)
 *
 * Copyright (C) 2007 NXP Semiconductors
 * Copyright (C) 2008 - 2012 DSP Group Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <asm/uaccess.h>
#include <linux/coma/cfifo.h>
#include <linux/coma/coma.h>
#include <linux/device.h>
#include <linux/err.h>
#include <linux/file.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/mutex.h>
#include <linux/slab.h>
#include <linux/workqueue.h>
#include <linux/coma/coma-alsa.h>
#include <linux/coma/coma-sharedmem.h>

#include "coma-signaling.h"
#include "coma-datahandler.h"
#include "coma-tdm.h"
#include "coma-voice.h"
#include "coma-dvf9a.h"
#include "coma-crypto.h"

#include "cmsg-coma.h"

MODULE_AUTHOR("DSP Group Inc.");
MODULE_LICENSE("GPL");

#define MAX_SERVICES CONFIG_DSPG_COMA_MAX_SERVICES

struct coma_service {
	const char *name;
	void *arg;
	struct cfifo __rcu *l2c;
	struct cfifo __rcu *c2l;
	atomic_t received_bytes;
	atomic_t sent_bytes;
	coma_process_cb_t process_message;
	coma_remove_cb_t remove;
	coma_suspend_cb_t suspend;
	struct work_struct work;
};

static int setup_result;
static DECLARE_COMPLETION(setup_reply);
static DECLARE_COMPLETION(remove_reply);
static DEFINE_MUTEX(coma_mutex);
static unsigned long active_services[BITS_TO_LONGS(MAX_SERVICES)];
static struct coma_service services[MAX_SERVICES];
static struct workqueue_struct *receive_wq;
static struct device *css_dev;
static dma_addr_t dma_base;
static size_t dma_size;

/*
 * Statistics
 */

static ssize_t show_services(struct device *dev, struct device_attribute *attr, char *buf)
{
	unsigned int i = 1;
	ssize_t len = 0;

	mutex_lock(&coma_mutex);
	for_each_set_bit_from(i, active_services, MAX_SERVICES) {
		len += snprintf(buf + len, PAGE_SIZE - len, "%d", i);
		if (services[i].name)
			len += snprintf(buf + len, PAGE_SIZE - len,
				" %s", services[i].name);

		len += snprintf(buf + len, PAGE_SIZE - len, "\n");
	}
	mutex_unlock(&coma_mutex);

	return len;
}

static ssize_t show_cfifo(struct device *dev, struct device_attribute *attr, char *buf)
{
	unsigned int i = 1;
	ssize_t len = 0;

	mutex_lock(&coma_mutex);
	for_each_set_bit_from(i, active_services, MAX_SERVICES) {
		struct cfifo *l2c, *c2l;

		l2c = rcu_dereference_protected(services[i].l2c,
			mutex_is_locked(&coma_mutex));
		c2l = rcu_dereference_protected(services[i].c2l,
			mutex_is_locked(&coma_mutex));

		len += snprintf(buf + len, PAGE_SIZE - len, "%d %u %s %u %s\n", i,
			cfifo_size(l2c), cfifo_empty(l2c) ? "empty" : "filled",
			cfifo_size(c2l), cfifo_empty(c2l) ? "empty" : "filled");
	}
	mutex_unlock(&coma_mutex);

	return len;
}

static ssize_t show_stats(struct device *dev, struct device_attribute *attr, char *buf)
{
	int i = 1;
	ssize_t len = 0;

	mutex_lock(&coma_mutex);
	for_each_set_bit_from(i, active_services, MAX_SERVICES) {
		len += snprintf(buf + len, PAGE_SIZE - len,
			"%d %d %d\n", i,
			atomic_read(&services[i].received_bytes),
			atomic_read(&services[i].sent_bytes));
	}
	mutex_unlock(&coma_mutex);

	return len;
}

static ssize_t reset_stats(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
	unsigned long id;

	id = simple_strtoul(buf, NULL, 0);
	if (id == 0 || id >= MAX_SERVICES)
		return -EINVAL;

	atomic_set(&services[id].received_bytes, 0);
	atomic_set(&services[id].sent_bytes, 0);

	return count;
}

static struct device_attribute coma_attrs[] = {
	__ATTR(services,	S_IRUGO,		show_services,	0),
	__ATTR(cfifo,		S_IRUGO,		show_cfifo,	0),
	__ATTR(statistics,	S_IRUGO | S_IWUSR,	show_stats,	reset_stats),
};

static int coma_sysfs_init(struct device *dev)
{
	int i, ret = 0;

	for (i = 0; i < ARRAY_SIZE(coma_attrs); i++) {
		ret = device_create_file(dev, &coma_attrs[i]);
		if (ret)
			goto err;
	}

	return 0;

err:
	while (--i >= 0)
		device_remove_file(dev, &coma_attrs[i]);
	printk(KERN_ERR "coma: failed creating sysfs file %d\n", ret);
	return ret;
}

static void coma_sysfs_release(struct device *dev)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(coma_attrs); i++)
		device_remove_file(dev, &coma_attrs[i]);
}

/*
 * CMsg handling
 */

struct cmsg *coma_cmsg_alloc(unsigned int service, int type,
			     unsigned int params_size,
			     unsigned int payload_size)
{
	struct cfifo *cfifo;
	struct cmsg *p = ERR_PTR(-ENODEV);

	rcu_read_lock();
	cfifo = rcu_dereference(services[service].l2c);
	if (!cfifo)
		goto out_err;

	p = cfifo_request(cfifo, sizeof(*p) + params_size + payload_size);
	if (IS_ERR(p))
		goto out_err;

	p->type = type;
	p->params_size = params_size;
	p->payload_size = payload_size;

	/* not quite accurate to do it here but who cares... */
	atomic_add(params_size + payload_size, &services[service].sent_bytes);

	/* keep on holding RCU read lock */
	return p;

out_err:
	rcu_read_unlock();
	return p;
}
EXPORT_SYMBOL(coma_cmsg_alloc);

int coma_cmsg_commit(unsigned int service)
{
	struct cfifo *cfifo;
	int ret = -ENODEV;

	/* the RCU read lock is still held from coma_cmsg_alloc() */

	cfifo = rcu_dereference(services[service].l2c);
	if (!cfifo)
		goto out;

	ret = cfifo_commit(cfifo);
	if (ret == 0)
		coma_signal(service);

out:
	rcu_read_unlock();
	return ret;
}
EXPORT_SYMBOL(coma_cmsg_commit);

void coma_cmsg_abort(unsigned int service)
{
	/*
	 * Nothing to do in the CFIFO. The caller can invoke coma_cmsg_alloc()
	 * again for the next message and it will happily reuse the just
	 * allocated space.
	 */

	rcu_read_unlock();
}
EXPORT_SYMBOL(coma_cmsg_abort);

int coma_cmsg_send(unsigned int service, int type,
		   void *params, unsigned int params_size,
		   void *payload, unsigned int payload_size)
{
	struct cmsg *cmsg;

	cmsg = coma_cmsg_alloc(service, type, params_size, payload_size);
	if (IS_ERR(cmsg))
		return PTR_ERR(cmsg);

	if (params)
		memcpy(cmsg_params(cmsg), params, params_size);

	if (payload)
		memcpy(cmsg_payload(cmsg), payload, payload_size);

	return coma_cmsg_commit(service);
}
EXPORT_SYMBOL(coma_cmsg_send);

/*
 * message handling
 */

static void coma_receive_work(struct work_struct *work)
{
	struct coma_service *service = container_of(work, struct coma_service, work);
	coma_process_cb_t cb = service->process_message;
	struct cfifo *cfifo;
	struct cmsg *cmsg;
	int len;
	void *arg;

	if (!cb)
		return;

	rcu_read_lock();
	cfifo = rcu_dereference(service->c2l);
	if (!cfifo) {
		rcu_read_unlock();
		return;
	}
	cfifo_ref(cfifo); /* make sure CFIFO doesn't disappear */
	rcu_read_unlock();

	arg = service->arg;
	while ((len = cfifo_get(cfifo, (void **)&cmsg)) > 0) {
		cb(arg, cmsg);
		cfifo_processed(cfifo);

		atomic_add(cmsg->params_size + cmsg->payload_size,
			&service->received_bytes);
	}

	cfifo_unref(cfifo);
}

void coma_receive(unsigned int i)
{
	if (test_bit(i, active_services))
		queue_work(receive_wq, &services[i].work);
}

/*
 * service (de)registration
 */

static int
coma_create_coma_message(enum cmsg_coma_types type,
                         union cmsg_coma_params *params)
{
	return coma_cmsg_send(0, type, params, sizeof(*params), NULL, 0);
}

int coma_register(const char *name, unsigned int size, void *arg,
		  coma_process_cb_t process_message, coma_remove_cb_t remove,
		  coma_suspend_cb_t suspend)
{
	struct cfifo *l2c, *c2l;
	union cmsg_coma_params params;
	int id, ret;

	if (strlen(name) >= 16)
		return -EINVAL;

	mutex_lock(&coma_mutex);

	if (!test_bit(0, active_services)) {
		ret = -ENODEV;
		goto done;
	}

	/* service already registered? */
	for_each_set_bit(id, active_services, MAX_SERVICES) {
		if (strcmp(services[id].name, name) == 0) {
			ret = -EBUSY;
			goto done;
		}
	}

	/* find free service number, starting from 1 (0 == COMA service) */
	id = find_next_zero_bit(active_services, MAX_SERVICES, 1);
	if (id >= MAX_SERVICES) {
		ret = -EMFILE;
		goto done;
	}

	/* allocate FIFOs */
	l2c = cfifo_alloc(css_dev, size, NULL, 0);
	if (!l2c) {
		ret = -ENOMEM;
		goto done;
	}
	c2l = cfifo_alloc(css_dev, size, NULL, 0);
	if (!c2l) {
		cfifo_unref(l2c);
		ret = -ENOMEM;
		goto done;
	}

	/*
	 * Already mark the service as active because once we send the
	 * registration message we might immediately get messages from the
	 * cordless domain.
	 */
	services[id].name = kstrdup(name, GFP_KERNEL);
	services[id].arg = arg;
	rcu_assign_pointer(services[id].l2c, l2c);
	rcu_assign_pointer(services[id].c2l, c2l);
	services[id].process_message = process_message;
	services[id].remove = remove;
	services[id].suspend = suspend;
	atomic_set(&services[id].received_bytes, 0);
	atomic_set(&services[id].sent_bytes, 0);
	INIT_WORK(&services[id].work, coma_receive_work);
	wmb();

	/* register on cordless domain */
	params.request_register.id = id;
	strcpy(params.request_register.name, name);
	params.request_register.l2c = cfifo_phys(l2c);
	params.request_register.c2l = cfifo_phys(c2l);
	ret = coma_create_coma_message(CMSG_COMA_REQUEST_REGISTER, &params);
	if (ret)
		goto err;

	/* wait for reply from cordless domain */
	ret = wait_for_completion_timeout(&setup_reply, HZ);
	if (ret <= 0) {
		if (ret == 0)
			ret = -ETIMEDOUT;
		goto err;
	}

	if (setup_result != 0) {
		ret = -ECONNREFUSED;
		goto err;
	}

	set_bit(id, active_services);

	ret = id;
	goto done;

err:
	rcu_assign_pointer(services[id].l2c, NULL);
	rcu_assign_pointer(services[id].c2l, NULL);
	services[id].process_message = NULL;
	services[id].remove = NULL;
	services[id].suspend = NULL;
	kfree(services[id].name);
	services[id].name = NULL;

	/*
	 * No need to synchronize with RCU here: nobody knows about the CFIFOs
	 * yet.
	 */
	cfifo_unref(c2l);
	cfifo_unref(l2c);
done:
	mutex_unlock(&coma_mutex);
	return ret;
}
EXPORT_SYMBOL(coma_register);

int coma_deregister(unsigned int id)
{
	struct cfifo *l2c, *c2l;
	union cmsg_coma_params params;
	int ret = 0;

	if (id >= MAX_SERVICES)
		return -EINVAL;

	mutex_lock(&coma_mutex);

	if (!test_bit(id, active_services))
		goto done;

	l2c = rcu_dereference_protected(services[id].l2c,
		mutex_is_locked(&coma_mutex));
	c2l = rcu_dereference_protected(services[id].c2l,
		mutex_is_locked(&coma_mutex));

	/*
	 * Only clear FIFO pointers first. Some other thread might still be
	 * writing into a CFIFO, so we have to wait for RCU until it is safe to
	 * free them.
	 */
	clear_bit(id, active_services);;
	wmb();
	rcu_assign_pointer(services[id].l2c, NULL);
	rcu_assign_pointer(services[id].c2l, NULL);
	services[id].process_message = NULL;
	services[id].remove = NULL;
	services[id].suspend = NULL;
	kfree(services[id].name);
	services[id].name = NULL;

	/* Unregister from cordless domain. Wait for reply if message could be
	 * sent. */
	params.request_deregister.id = id;
	ret = coma_create_coma_message(CMSG_COMA_REQUEST_DEREGISTER, &params);
	if (ret == 0)
		wait_for_completion_timeout(&remove_reply, HZ);

	flush_work(&services[id].work);

	/* synchronize with RCU and then safely free the CFIFos */
	synchronize_rcu();
	cfifo_unref(c2l);
	cfifo_unref(l2c);

done:
	mutex_unlock(&coma_mutex);
	return ret;
}
EXPORT_SYMBOL(coma_deregister);

static void coma_process_message(void *arg, struct cmsg *cmsg)
{
	union cmsg_coma_params *params = cmsg_params(cmsg);

	switch(cmsg->type) {
	case CMSG_COMA_REPLY_REGISTER:
		setup_result = params->reply_register.result;
		complete(&setup_reply);
		break;

	case CMSG_COMA_REPLY_DEREGISTER:
		complete(&remove_reply);
		break;
	}
}

/*
 * start-up and release
 */

void init_kernel_coma_services(struct work_struct *work)
{
	tdm_service_init();
	datahandler_init();
	voice_init(css_dev);
	dvf9a_coma_service_init();
	crypto_coma_service_init();
	alsa_service_init();
	sharedmem_coma_service_init(css_dev, dma_base, dma_size);
}

static DECLARE_DELAYED_WORK(init_work, init_kernel_coma_services);

int coma_suspend(void)
{
	int i;
	int ret = 0;

	mutex_lock(&coma_mutex);
	for_each_set_bit(i, active_services, MAX_SERVICES) {
		if (services[i].suspend)
			ret = services[i].suspend();
		if (ret)
			break;
	}

	mutex_unlock(&coma_mutex);
	return ret;
}
EXPORT_SYMBOL(coma_suspend);

int coma_setup(struct device *dev, dma_addr_t *l2c_phys, dma_addr_t *c2l_phys,
	       dma_addr_t base, size_t size)
{
	int ret;
	struct cfifo *l2c, *c2l;

	/* guard against concurrent coma_register() calls */
	mutex_lock(&coma_mutex);

	dma_base = base;
	dma_size = size;

	css_dev = dev;

	receive_wq = alloc_workqueue("coma", WQ_NON_REENTRANT, 0);
	if (!receive_wq) {
		ret = -ENOMEM;
		goto out_unlock;
	}

	l2c = cfifo_alloc(dev, 1024, NULL, 0);
	if (!l2c) {
		ret = -ENOMEM;
		goto out_free_wq;
	}

	c2l = cfifo_alloc(dev, 1024, NULL, 0);
	if (!c2l) {
		ret = -ENOMEM;
		goto out_free_l2c;
	}

	memset(services, 0, sizeof(services));
	services[0].name = kstrdup("coma", GFP_KERNEL);
	rcu_assign_pointer(services[0].l2c, l2c);
	rcu_assign_pointer(services[0].c2l, c2l);
	services[0].process_message = coma_process_message;
	INIT_WORK(&services[0].work, coma_receive_work);
	wmb();
	set_bit(0, active_services);

	ret = coma_signaling_init();
	if (ret)
		goto out_free_c2l;

	ret = coma_sysfs_init(dev);
	if (ret)
		goto out_sig_exit;

	*l2c_phys = cfifo_phys(l2c);
	*c2l_phys = cfifo_phys(c2l);

	mutex_unlock(&coma_mutex);

	/*
	 * Defer the initialization of in-kernel services a bit. Initializing
	 * them here wouldn't work because the cordless domain is not running
	 * yet.
	 */
	queue_delayed_work(system_long_wq, &init_work, 1);

	return 0;

out_sig_exit:
	coma_signaling_exit();
out_free_c2l:
	clear_bit(0, active_services);
	kfree(services[0].name);
	memset(services, 0, sizeof(services));
	cfifo_unref(c2l);
out_free_l2c:
	cfifo_unref(l2c);
out_free_wq:
	flush_workqueue(receive_wq);
	destroy_workqueue(receive_wq);
out_unlock:
	mutex_unlock(&coma_mutex);
	return ret;
}
EXPORT_SYMBOL(coma_setup);

void coma_release(struct device *dev)
{
	int i;

	/* must not grab coma_mutex because init_work might also grab it */
	cancel_delayed_work_sync(&init_work);

	/* tear down all services forcefully */
	mutex_lock(&coma_mutex);

	coma_sysfs_release(dev);
	coma_signaling_exit();
	flush_workqueue(receive_wq);
	destroy_workqueue(receive_wq);

	for_each_set_bit(i, active_services, MAX_SERVICES) {
		struct cfifo *l2c, *c2l;

		l2c = rcu_dereference_protected(services[i].l2c,
			mutex_is_locked(&coma_mutex));
		c2l = rcu_dereference_protected(services[i].c2l,
			mutex_is_locked(&coma_mutex));

		/*
		 * Only clear FIFO pointers first. Some other thread might
		 * still be writing into a CFIFO, so we have to wait for RCU
		 * until it is safe to free them.
		 */
		clear_bit(i, active_services);
		rcu_assign_pointer(services[i].l2c, NULL);
		rcu_assign_pointer(services[i].c2l, NULL);
		kfree(services[i].name);

		/* synchronize with RCU and then safely free the CFIFos */
		synchronize_rcu();
		cfifo_unref(c2l);
		cfifo_unref(l2c);

		if (services[i].remove)
			services[i].remove(services[i].arg);
	}
	memset(services, 0, sizeof(services));
	mutex_unlock(&coma_mutex);
}
EXPORT_SYMBOL(coma_release);

