/*
 * linux/drivers/staging/dspg/dspg-crypto.c - DSPG cryptographic security module
 *
 *  Copyright (C) 2013 DSP Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

#include <linux/ccu.h>
#include <linux/clk.h>
#include <linux/err.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include <linux/spinlock.h>
#include <linux/dmapool.h>
#include <linux/dma-mapping.h>
#include <linux/sched.h>
#include <linux/of.h>
#include <linux/of_device.h>
#include <linux/mfd/core.h>
#include <linux/mfd/dspg-crypto.h>
#include <mach/platform.h>

#define DSPG_CRYPTO "dspg-crypto"

#define for_each_bit(i, bits) \
	for (i=0; (bits)!=0; i++, (bits)>>=1) if ((bits) & 1)

/* Register map */
enum REG_CFG {
	REG_CFG_CFG				= 0x0000,
	REG_CFG_QoS				= 0x0004,
};

/* True Random Number Generator (TRNG) registers */
enum REG_TRNG {
	REG_TRNG_CFG				= 0x0008,
	REG_TRNG_KICK				= 0x000C,
	REG_TRNG_STATUS				= 0x0010,
	REG_TRNG_CLR_STATUS			= 0x0014,
};

/* Interrupt status and control registers */
enum REG_IRQ {
	REG_IRQ_DESC_OWNER_CHANGED_STATUS	= 0x0100,
	REG_IRQ_DESC_OWNER_CHANGED_CAUSE	= 0x0104,
	REG_IRQ_DESC_OWNER_CHANGED_STATUS_CLR	= 0x0108,
	REG_IRQ_HEAD_PTR_WENT_NULL_STATUS	= 0x010C,
	REG_IRQ_HEAD_PTR_WENT_NULL_ENABLE	= 0x0110,
	REG_IRQ_HEAD_PTR_WENT_NULL_CAUSE	= 0x0114,
	REG_IRQ_HEAD_PTR_WENT_NULL_STATUS_CLR	= 0x0118,

	REG_IRQ_CRYPTO_AHB_EXCEPTION_STATUS	= 0x011C,
	REG_IRQ_CRYPTO_HW_EXCEPTION_STATUS	= 0x0120,
	REG_IRQ_CRYPTO_DESC_EXCEPTION_STATUS	= 0x0124,
	REG_IRQ_CRYPTO_EXCEPTION_ENABLE		= 0x0128,
	REG_IRQ_CRYPTO_EXCEPTION_CAUSE		= 0x0140,
	REG_IRQ_CRYPTO_EXCEPTION_STATUS_CLR	= 0x0144,
};

enum irq_exception {
	IRQ_EXCEPTION_AHB,
	IRQ_EXCEPTION_HW,
	IRQ_EXCEPTION_DESC,
};

/*
 * Channel Headpointer registers
 * All headpointers point to the first descriptor in the corresponding
 * linked list and are initialized by software.  There are:
 * - 24 high priority (VoIP) channels
 * - 4 medium priority (streaming data) channels.
 * - 4 medium priority (data) channels.
 */
enum  REG_HEADPTR {
	REG_CHANNEL_HEADPTR			= 0x0200,
	REG_HEADPTR_ENABLED			= 0x0280,
	REG_HEADPTR_SET				= 0x0284,
	REG_HEADPTR_RESET			= 0x0288,
};

#define NUM_CHANNELS			32
#define MIN_CHANNEL			 0
#define MAX_CHANNEL			(NUM_CHANNELS-1)

/* Descriptor control fields */
#define DESC_CTL_DESC_OWNER_HW	0x80000000 /* Ownership: 0 = SW; 1 = HW */
#define DESC_CTL_INT_REQ	0x40000000
#define DESC_CTL_LOCK_NEXT	0x20000000
#define DESC_CTL_SAVE_STATE	0x18000000
#define DESC_CTL_GCM_FAIL	0x04000000

/* Channel HW descriptor */
struct hw_crypto_desc {
	u32 desc_ctl;		/* Descriptor control */
	u32 link_ptr;		/* 32-bit pointer to next descriptor */
	u32 data_in_phys;	/* 32-bit pointer to input data */
	u32 data_out_phys;	/* 32-bit pointer to output data */
	u32 data_in_len;	/* 31-bit length of input data */
	u32 data_out_len;	/* 31-bit length of output data */
	u32 iv[8];		/* Initialization vector */
	u32 state_index;
	u32 data_in_total_len;	/* Full packet data length if last descriptor
				 * of multiple-descriptor packet of HASH/HMAC */
};

/* TRNG operation */
#define TRNG_STATUS_BUSY	0x00000001
#define TRNG_STATUS_READY	0x00000002
#define TRNG_MAGIC_SEQ		0xaa55bb11
#define TRNG_STATUS_READY_CLR	0x00000001

#define TRNG_CONFIG_HW_TUNE	0x00000000

/* Keystore mappping */
#define KS_KEY_128_START	0x1000 /* indices 0 - 31  */
#define KS_KEY_256_START	0x1200 /* indices 32 - 79 */
#define TRNG_CONFIG_KEY_INDEX	0x0000004F /* use the last 256 bit keystore entry */
#define SECRET_KEY_INDEX	0x00000050

struct dspg_crypto_dev {
	struct platform_device *pdev;
	void __iomem *regs;
	struct clk *clk;

	struct dma_pool *desc_dma_pool;
	spinlock_t lock;
	u32 irq;

	int chip_version;

	struct crypto_op *pending_ops[NUM_CHANNELS];
};

static struct dspg_crypto_dev *_dspg_crypto_dev = NULL;
static int key_store[NUM_CHANNELS] = { 0 };

static unsigned long
crypto_readl(unsigned long reg)
{
	return readl(_dspg_crypto_dev->regs + reg);
}

static void
crypto_writel(unsigned long reg, unsigned long val)
{
	writel(val, _dspg_crypto_dev->regs + reg);
}

static void
dspg_crypto_int_enable(unsigned long ch, int enable)
{
	u32 temp = crypto_readl(REG_IRQ_HEAD_PTR_WENT_NULL_ENABLE);

	if (enable)
		crypto_writel(REG_IRQ_HEAD_PTR_WENT_NULL_ENABLE, temp |   1 << ch);
	else
		crypto_writel(REG_IRQ_HEAD_PTR_WENT_NULL_ENABLE, temp & ~(1 << ch));

	temp = crypto_readl(REG_IRQ_CRYPTO_EXCEPTION_ENABLE);
	if (enable)
		crypto_writel(REG_IRQ_CRYPTO_EXCEPTION_ENABLE, temp |   1 << ch);
	else
		crypto_writel(REG_IRQ_CRYPTO_EXCEPTION_ENABLE, temp & ~(1 << ch));
}

static void
dspg_crypto_free_key(int channel, u8 key_idx)
{
	key_store[channel] = 0;
}

static void
dspg_crypto_op_done(u32 ch, struct dspg_crypto_dev *dev)
{
	struct device *ddev = &dev->pdev->dev;
	struct crypto_op *op = dev->pending_ops[ch];

	dma_unmap_single(ddev, op->hw_desc->data_in_phys,
			 op->hw_desc->data_in_len, DMA_TO_DEVICE);
	dma_unmap_single(ddev, op->hw_desc->data_out_phys,
			 op->hw_desc->data_out_len, DMA_FROM_DEVICE);
	dma_pool_free(_dspg_crypto_dev->desc_dma_pool, op->hw_desc,
		      op->hw_desc_phys);

	dspg_crypto_int_enable(ch, 0);

	dspg_crypto_free_key(ch, 0); //*key_idx);

	op->done(0, op->context);
}

static irqreturn_t
dspg_crypto_isr(int irq, void *dev_id)
{
	struct dspg_crypto_dev *dev = (struct dspg_crypto_dev*)dev_id;
	struct device *ddev = &dev->pdev->dev;

	u32 channel;
	u32 completed_desc = 0;
	u32 irq_head_ptr_went_null, irq_desc_owner_changed;
	u32 irq_exception_cause;
	u32 irq_ahb_exception_status, irq_hw_exception_status, irq_desc_exception_status;

	irq_ahb_exception_status = crypto_readl(REG_IRQ_CRYPTO_AHB_EXCEPTION_STATUS);
	irq_hw_exception_status = crypto_readl(REG_IRQ_CRYPTO_HW_EXCEPTION_STATUS);
	irq_desc_exception_status = crypto_readl(REG_IRQ_CRYPTO_DESC_EXCEPTION_STATUS);

	/* clear interrupts */
	irq_exception_cause = crypto_readl(REG_IRQ_CRYPTO_EXCEPTION_CAUSE);
	crypto_writel(REG_IRQ_CRYPTO_EXCEPTION_STATUS_CLR, irq_exception_cause);

	irq_desc_owner_changed = crypto_readl(REG_IRQ_DESC_OWNER_CHANGED_STATUS);
	crypto_writel(REG_IRQ_DESC_OWNER_CHANGED_STATUS_CLR, irq_desc_owner_changed);

	irq_head_ptr_went_null = crypto_readl(REG_IRQ_HEAD_PTR_WENT_NULL_STATUS);
	crypto_writel(REG_IRQ_HEAD_PTR_WENT_NULL_STATUS_CLR, irq_head_ptr_went_null);

	if (irq_exception_cause) {
		for_each_bit(channel, irq_exception_cause) {
			dev_err(ddev, "EXCEPTION irq on channel %d\n", channel);

			if (irq_ahb_exception_status & (1<< channel))
				dev_err(ddev, "AHB EXCEPTION irq on channel %d\n", channel);
			if (irq_hw_exception_status & (1<< channel))
				dev_err(ddev, "HW EXCEPTION irq on channel %d\n", channel);
			if (irq_desc_exception_status & (1<< channel))
				dev_err(ddev, "Descriptor EXCEPTION irq on channel %d\n", channel);

			/* @todo: invoke dspg_crypto_op_done and report failure */
		}
	}

	/*
	 * ATTENTION: The following call will guarantee that all memory
	 * transactions indicated by the device irq flags have actually hit the
	 * memory. Just process the flags which have been sampled. When reading
	 * the status again any new IRQ flags MUST NOT be processed until
	 * ccu_barrier() is called again!
	 */
	ccu_barrier();

	completed_desc = irq_exception_cause | irq_head_ptr_went_null;
	for_each_bit(channel, completed_desc)
		dspg_crypto_op_done(channel, dev);

	clk_disable(_dspg_crypto_dev->clk);

	return IRQ_HANDLED;
}

void
dspg_crypto_generate_trng(void)
{
	clk_enable(_dspg_crypto_dev->clk);

	while ((crypto_readl(REG_TRNG_STATUS) & TRNG_STATUS_BUSY) == TRNG_STATUS_BUSY)
		schedule();

	crypto_writel(REG_TRNG_CFG, TRNG_CONFIG_HW_TUNE | TRNG_CONFIG_KEY_INDEX);

	crypto_writel(REG_TRNG_KICK, TRNG_MAGIC_SEQ);
	while ((crypto_readl(REG_TRNG_STATUS) & TRNG_STATUS_READY) != TRNG_STATUS_READY)
		schedule();

	crypto_writel(REG_TRNG_CLR_STATUS, TRNG_STATUS_READY_CLR);

	clk_disable(_dspg_crypto_dev->clk);
}
EXPORT_SYMBOL(dspg_crypto_generate_trng);

int
dspg_crypto_get_trng(unsigned int *buf)
{
	int i;

	clk_enable(_dspg_crypto_dev->clk);

	for (i = 0; i < NUM_OF_TRNG; i++)
		buf[i] = crypto_readl(KS_KEY_256_START + 0x20 * (TRNG_CONFIG_KEY_INDEX-32)+i);

	clk_disable(_dspg_crypto_dev->clk);

	return NUM_OF_TRNG;
}
EXPORT_SYMBOL(dspg_crypto_get_trng);

/* @todo - handle allocation of SECRET_KEY */
static int
dspg_crypto_alloc_key(int channel, u8 key_size, u8 *key_idx, u8 *key)
{
	int i;
	u32 word, keybase;

	if (key_size <= 16) {
		*key_idx = channel;
		keybase = KS_KEY_128_START;
	} else if (key_size <= 32) {
		*key_idx = channel + 32;
		keybase = KS_KEY_256_START;
	} else {
		return -1;
	}

	if (!key_store[channel]) {
		/* load the key to the key-store */
		for (i = 0; i < key_size; i += 4) {
			word = *((u32*)(key+i));
			crypto_writel(keybase + i, word);
		}
		key_store[channel] = 1;
	}

	return 0;
}

static struct hw_crypto_desc *
dspg_crypto_alloc_hw_desc(struct crypto_op *op, u8 mode, int channel,
			  u8 key_size, u8 *key_idx, u8 *key)
{
	struct hw_crypto_desc *hw_desc;

	if (dspg_crypto_alloc_key(channel, key_size, key_idx, key))
		return NULL;

	hw_desc = dma_pool_alloc(_dspg_crypto_dev->desc_dma_pool, GFP_KERNEL, &op->hw_desc_phys);
	if (!hw_desc) {
		dspg_crypto_free_key(channel, *key_idx);
		return NULL;
	}

	op->hw_desc = hw_desc;
	key_size = (key_size == 0) ? 0 : key_size - 1;
	hw_desc->desc_ctl = DESC_CTL_DESC_OWNER_HW	|
			    DESC_CTL_INT_REQ		|
			    (*key_idx << 16)		|
			    (key_size << 8)		|
			mode;

	return hw_desc;
}

static int
dspg_crypto_verify_buffers(void *in, size_t in_len, void *out, size_t out_len)
{
	/*
	 * Output DMA is performed using 32-bit access, therefore 'out' must be 32-bit
	 * aligned start address. If 'out_len' is not integer multiple of 32 bits, then
	 * last write will overwrite up to 3 bytes beyond end of data to next 32-bit
	 * aligned address.
	 */
	/* if (((u32)out%4) || (out_len%4))
		goto err_align; */

	/*
	 * Byte-aligned input are allowed.  However, if encrypt/decrypt in-place is
	 * desired, 'in' has the same size & alignment restrictions as 'out'
	 */
	if ( (in==out) && (((u32)in%4) || (in_len%4)) )
		goto err_align;

	/*
	 * If selected mode is ECB or CBC block cipher (mode is set to DES/3DES ECB or CBC,
	 * AES ECB or CBC modes), then software is responsible for performing the Cipher
	 * padding to ensure input data length is multiple of selected Cipher block size.
	 * A Crypto Exception interrupt will be generated if software fails to perform this
	 * task.  Software must determine what cipher padding scheme is required and use
	 * this to pad input.
	 */
	 /* @todo - implement this logic */
	return 0;

err_align:
	dev_err(&_dspg_crypto_dev->pdev->dev, "Buffer alignment error\n");
	return -EINVAL;
}

/* @todo: implement allocation logic */
static int
dspg_crypto_allocate_channel(int channel)
{
	/* @todo - handle case of channel==-1 (driver chooses channel) */
	if ((channel < MIN_CHANNEL) || (channel > MAX_CHANNEL))
		return -EINVAL;

	/* @todo - manage real allocation of channel resource
	return EADDRINUSE; */
	return 0;
}

/* @todo: implement free logic */
static int
dspg_crypto_free_channel(int channel)
{
	return 0;
}

/* @todo - verify key len is legal */
static int
dspg_crypto_verify_mode(u8 mode, size_t out_len)
{
	if (_dspg_crypto_dev->chip_version <= 1) {
		if ((mode == DESC_CTL_MODE_AES_GCM_ENCRYPT) ||
		    (mode == DESC_CTL_MODE_AES_GCM_DECRYPT))
			return -1;
	}

	switch (mode) {
		case DESC_CTL_MODE_MD5:
		case DESC_CTL_MODE_MD5_HMAC:
			if (out_len!=16)
				return -1;
			break;

		case DESC_CTL_MODE_SHA1_160:
		case DESC_CTL_MODE_SHA1_160_HMAC:
			if (out_len!=20)
				return -1;
			break;

		case DESC_CTL_MODE_SHA2_224:
		case DESC_CTL_MODE_SHA2_224_HMAC:
			if (out_len!=28)
				return -1;
			break;

		case DESC_CTL_MODE_SHA2_256:
		case DESC_CTL_MODE_SHA2_256_HMAC:
			if (out_len!=32)
				return -1;
			break;

		case DESC_CTL_MODE_AES_CTR_ENCRYPT:
		case DESC_CTL_MODE_AES_CBC_ENCRYPT:
			/* if ((out_len!=16) && (out_len!=32) && (out_len!=64))
				return -1; */
			break;

		default:
			return 0;
	}

	return 0;
}

struct crypto_op *
dspg_crypto_alloc_operation(void *in, size_t in_len, void *out, size_t out_len,
			    u8 mode, int channel, u8 *key, size_t key_len,
			    void *iv, void *a, size_t a_len, size_t c_len)
{
	struct hw_crypto_desc *hw_desc;
	struct crypto_op *op;
	struct device *dev = &_dspg_crypto_dev->pdev->dev;
	u8 key_idx;

	if (dspg_crypto_verify_mode(mode, out_len))
		return NULL;

	if (dspg_crypto_verify_buffers(in, in_len + a_len, out, out_len))
		return NULL;

	if (dspg_crypto_allocate_channel(channel))
		return NULL;

	op = kzalloc(sizeof(struct crypto_op), GFP_KERNEL);
	
	if (!op) goto err_op_alloc;

	op->channel = channel;
	op->data_in_virt = in;
	op->data_out_virt = out;

	clk_enable(_dspg_crypto_dev->clk);

	hw_desc = dspg_crypto_alloc_hw_desc(op, mode, channel, key_len, &key_idx, key);
	if (!hw_desc) goto err_desc_alloc;

	_dspg_crypto_dev->pending_ops[channel] = op;

	if (mode == DESC_CTL_MODE_AES_GCM_DECRYPT)
		hw_desc->data_in_len = a_len - 12 + c_len + 16;
	else if (mode == DESC_CTL_MODE_AES_GCM_ENCRYPT)
		hw_desc->data_in_len = a_len - 12 + c_len;
	else
		hw_desc->data_in_len = in_len + a_len;
	hw_desc->data_in_phys = dma_map_single(dev, op->data_in_virt,
					       hw_desc->data_in_len,
					       DMA_TO_DEVICE);

	if (mode == DESC_CTL_MODE_AES_GCM_DECRYPT)
		hw_desc->data_out_len = c_len;
	else if (mode == DESC_CTL_MODE_AES_GCM_ENCRYPT)
		hw_desc->data_out_len = c_len + 16;
	else
		hw_desc->data_out_len = out_len;
	hw_desc->data_out_phys = dma_map_single(dev, op->data_out_virt,
						hw_desc->data_out_len,
						DMA_FROM_DEVICE);

	if (iv != NULL)
		memcpy(hw_desc->iv, iv, 16);
	if ((mode == DESC_CTL_MODE_AES_GCM_ENCRYPT) ||
	    (mode == DESC_CTL_MODE_AES_GCM_DECRYPT)) {
		hw_desc->iv[3] = 0;
		memcpy(&hw_desc->iv[4], a, 12);
		hw_desc->iv[7] = (a_len << 16) | c_len;
	}

	return op;

err_desc_alloc:
	dev_err(dev, "Failed to allocate HW descriptor memory\n");
	kfree(op);
	
err_op_alloc:
	dev_err(dev, "Failed to allocate operator memory\n");
	dspg_crypto_free_channel(channel);
	return NULL;
}
EXPORT_SYMBOL(dspg_crypto_alloc_operation);

/* @todo: implement logic */
void
dspg_crypto_free_operation(struct crypto_op *op)
{
	kfree(op);
}
EXPORT_SYMBOL(dspg_crypto_free_operation);

int
dspg_crypto_submit_operation(struct crypto_op *op)
{
	u32 temp = 0;
	u32 ch = op->channel;

	/* assign the desc to the channel's headpointer */
	crypto_writel(REG_CHANNEL_HEADPTR + (ch * 4), op->hw_desc_phys);

	dspg_crypto_int_enable(ch, 1);

	/* enable the channel */
	temp = crypto_readl(REG_HEADPTR_SET);
	crypto_writel(REG_HEADPTR_SET, temp | 1 << ch);

	return 0;
}
EXPORT_SYMBOL(dspg_crypto_submit_operation);

static int __init
dspg_crypto_probe(struct platform_device *pdev)
{
	struct dspg_crypto_dev *dev;
	struct device_node *np = pdev->dev.of_node;
	struct resource *res;
	int irq;
	int err = 0;
	u32 cfg;

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res)
		return -ENODEV;

	dev = kzalloc(sizeof(struct dspg_crypto_dev), GFP_KERNEL);
	if (!dev) {
		dev_err(&pdev->dev, "out of memory\n");
		return -ENOMEM;
	}

	_dspg_crypto_dev = dev;
	dev->pdev = pdev;
	platform_set_drvdata(pdev, dev);

	dev->chip_version = dvf_get_chip_rev();

	irq = platform_get_irq(pdev, 0);
	if (irq < 0) {
		dev_err(&pdev->dev, "out of memory\n");
		kfree(dev);
		return irq;
	}
	dev->irq = irq;

	if (!request_mem_region(res->start, res->end - res->start, DSPG_CRYPTO)) {
		dev_err(&pdev->dev, "failed to request memory region\n");
		err = -EBUSY;
		goto err_kfree;
	}

	dev->regs = ioremap_nocache(res->start, res->end - res->start);
	if (!dev->regs) {
		dev_err(&pdev->dev, "unable to map registers\n");
		err = -ENOMEM;
		goto err_release_reg;
	}

	dev->clk = clk_get(&pdev->dev, DSPG_CRYPTO);
	if (IS_ERR(dev->clk)) {
		dev_err(&pdev->dev, "could not get clock\n");
		err = PTR_ERR(dev->clk);
		goto err_clk;
	}

	spin_lock_init(&dev->lock);

	err = request_irq(dev->irq, dspg_crypto_isr, 0, DSPG_CRYPTO, dev);
	if (err)
		goto err_irq;

	dev->desc_dma_pool = dma_pool_create(DSPG_CRYPTO, &pdev->dev,
		sizeof(struct hw_crypto_desc), NUM_CHANNELS, 0);
	if (!dev->desc_dma_pool) {
		err = -ENOMEM;
		goto err_dma;
	}

	err = of_platform_populate(np, NULL, NULL, &pdev->dev);
	if (err < 0) {
		dev_err(&pdev->dev, "error populating subdevices\n");
		goto err_of_populate;
	}

	clk_enable(dev->clk);
	cfg = crypto_readl(REG_CFG_CFG);
	clk_disable(dev->clk);

	dev_info(&pdev->dev, "DSPG cryptographic security module (revision %d/%d)\n",
		 (cfg >> 20) & 0xf, (cfg >> 16) & 0xf);

	return 0;

err_of_populate:
	dma_pool_destroy(dev->desc_dma_pool);
err_dma:
	free_irq(irq, dev);
err_irq:
	clk_disable(dev->clk);
	clk_put(dev->clk);
err_clk:
	iounmap(dev->regs);
	dev->regs = NULL;
err_release_reg:
	release_mem_region(res->start, res->end - res->start);
err_kfree:
	kfree(_dspg_crypto_dev);
	_dspg_crypto_dev = NULL;
	return err;
}

static int
__exit dspg_crypto_remove(struct platform_device *pdev)
{
	struct dspg_crypto_dev *dev = platform_get_drvdata(pdev);

	dma_pool_destroy(dev->desc_dma_pool);
	free_irq(dev->irq, dev);
	clk_disable(dev->clk);
	clk_put(dev->clk);
	iounmap(dev->regs);
	kfree(dev);

	return 0;
}

static const struct of_device_id dspg_crypto_of_ids[] = {
	{ .compatible = "dspg,crypto" },
	{ },
};

MODULE_DEVICE_TABLE(of, dspg_crypto_of_ids);

static struct platform_driver dspg_crypto_driver = {
	.driver = {
		.name = DSPG_CRYPTO,
		.owner = THIS_MODULE,
		.of_match_table = dspg_crypto_of_ids,
	},
	.remove = __exit_p(dspg_crypto_remove),
};

static int __init dspg_crypto_init(void)
{
	return platform_driver_probe(&dspg_crypto_driver, dspg_crypto_probe);
}
module_init(dspg_crypto_init);

static void __exit dspg_crypto_exit(void)
{
	platform_driver_unregister(&dspg_crypto_driver);
}
module_exit(dspg_crypto_exit);

MODULE_AUTHOR("DSP Group Inc.");
MODULE_DESCRIPTION("DSPG Cryptographic security module driver");
MODULE_LICENSE("GPL");
