/*
 * Synopsys DesignWare Multimedia Card Interface driver
 *  (Based on NXP driver for lpc 31xx)
 *
 * Copyright (C) 2009 NXP Semiconductors
 * Copyright (C) 2009, 2010 Imagination Technologies Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <linux/bitops.h>
#include <linux/blkdev.h>
#include <linux/ccu.h>
#include <linux/clk.h>
#include <linux/debugfs.h>
#include <linux/delay.h>
#include <linux/device.h>
#include <linux/dma-mapping.h>
#include <linux/err.h>
#include <linux/gpio.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/irq.h>
#include <linux/mmc/dw_mmc_dspg.h>
#include <linux/mmc/host.h>
#include <linux/mmc/mmc.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/of_device.h>
#include <linux/of_gpio.h>
#include <linux/platform_device.h>
#include <linux/regulator/consumer.h>
#include <linux/scatterlist.h>
#include <linux/seq_file.h>
#include <linux/slab.h>
#include <linux/stat.h>
#include <linux/workqueue.h>

#include "dw_mmc_dspg.h"

/* Common flag combinations */
#define DW_MCI_DATA_ERROR_FLAGS	(SDMMC_INT_DTO | SDMMC_INT_DCRC | \
				 SDMMC_INT_SBE | SDMMC_INT_EBE)
#define DW_MCI_CMD_ERROR_FLAGS	(SDMMC_INT_RTO | SDMMC_INT_RCRC | \
				 SDMMC_INT_RESP_ERR)
#define DW_MCI_ERROR_FLAGS	(DW_MCI_DATA_ERROR_FLAGS | SDMMC_INT_HLE)
#define DW_MCI_SEND_STATUS	1
#define DW_MCI_RECV_STATUS	2
#define DW_MCI_DMA_THRESHOLD	16

#ifdef CONFIG_MMC_DW_DSPG_IDMAC
struct idmac_desc {
	u32		des0;	/* Control Descriptor */
#define IDMAC_DES0_DIC	BIT(1)
#define IDMAC_DES0_LD	BIT(2)
#define IDMAC_DES0_FD	BIT(3)
#define IDMAC_DES0_CH	BIT(4)
#define IDMAC_DES0_ER	BIT(5)
#define IDMAC_DES0_CES	BIT(30)
#define IDMAC_DES0_OWN	BIT(31)

	u32		des1;	/* Buffer sizes */
#define IDMAC_SET_BUFFER1_SIZE(d, s) \
	((d)->des1 = ((d)->des1 & 0x03ffe000) | ((s) & 0x1fff))

	u32		des2;	/* buffer 1 physical address */

	u32		des3;	/* buffer 2 physical address */
};
#endif /* CONFIG_MMC_DW_DSPG_IDMAC */

/**
 * struct dw_mci_slot - MMC slot state
 * @mmc: The mmc_host representing this slot.
 * @host: The MMC controller this slot is using.
 * @ctype: Card type for this slot.
 * @mrq: mmc_request currently being processed or waiting to be
 *	processed, or NULL when the slot is idle.
 * @queue_node: List node for placing this node in the @queue list of
 *	&struct dw_mci.
 * @clock: Clock rate configured by set_ios(). Protected by host->lock.
 * @div: Clock divider for card clock
 * @cycle_ns: Nano seconds per card clock cycle
 * @flags: Random state bits associated with the slot.
 * @id: Number of this slot.
 * @last_detect_state: Most recently observed card detect state.
 */
struct dw_mci_slot {
	struct mmc_host			*mmc;
	struct dw_mci			*host;
	struct dw_mci_board_slot	*pdata;

	struct regulator		*vmmc;	/* Power regulator */
	unsigned char			power_mode;
	int				cd_irq;

	u32				ctype;

	struct mmc_request		*mrq;
	struct list_head		queue_node;

	unsigned int			clock;
	unsigned int			div;
	unsigned long			cycle_ns;
	unsigned long			flags;
#define DW_MMC_CARD_PRESENT		0
#define DW_MMC_CARD_NEED_INIT		1
#define DW_MMC_CARD_POWERED		2
	int				id;
	int				last_detect_state;
};

static struct workqueue_struct *dw_mci_card_workqueue;

static void dw_mci_req_busy(struct dw_mci *host)
{
	del_timer_sync(&host->clk_off_timer);

	if (likely(host->clk_enabled))
		return;

	clk_enable(host->clk);
	host->clk_enabled = 1;
}

static void dw_mci_req_idle(struct dw_mci *host)
{
	mod_timer(&host->clk_off_timer, jiffies + HZ);
}

static void dw_mci_clk_off(unsigned long data)
{
	struct dw_mci *host = (struct dw_mci *)data;

	spin_lock_bh(&host->lock);
	if (!host->clk_enabled)
		return;
	clk_disable(host->clk);
	host->clk_enabled = 0;
	spin_unlock_bh(&host->lock);
}

#if defined(CONFIG_DEBUG_FS)
static int dw_mci_req_show(struct seq_file *s, void *v)
{
	struct dw_mci_slot *slot = s->private;
	struct mmc_request *mrq;
	struct mmc_command *cmd;
	struct mmc_command *stop;
	struct mmc_data	*data;

	/* Make sure we get a consistent snapshot */
	spin_lock_bh(&slot->host->lock);
	mrq = slot->mrq;

	if (mrq) {
		cmd = mrq->cmd;
		data = mrq->data;
		stop = mrq->stop;

		if (cmd)
			seq_printf(s,
				   "CMD%u(0x%x) flg %x rsp %x %x %x %x err %d\n",
				   cmd->opcode, cmd->arg, cmd->flags,
				   cmd->resp[0], cmd->resp[1], cmd->resp[2],
				   cmd->resp[2], cmd->error);
		if (data)
			seq_printf(s, "DATA %u / %u * %u flg %x err %d\n",
				   data->bytes_xfered, data->blocks,
				   data->blksz, data->flags, data->error);
		if (stop)
			seq_printf(s,
				   "CMD%u(0x%x) flg %x rsp %x %x %x %x err %d\n",
				   stop->opcode, stop->arg, stop->flags,
				   stop->resp[0], stop->resp[1], stop->resp[2],
				   stop->resp[2], stop->error);
	}

	spin_unlock_bh(&slot->host->lock);

	return 0;
}

static int dw_mci_req_open(struct inode *inode, struct file *file)
{
	return single_open(file, dw_mci_req_show, inode->i_private);
}

static const struct file_operations dw_mci_req_fops = {
	.owner		= THIS_MODULE,
	.open		= dw_mci_req_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= single_release,
};

static int dw_mci_regs_show(struct seq_file *s, void *v)
{
	struct dw_mci *host = s->private;

	clk_enable(host->clk);

	seq_printf(s, "STATUS:\t0x%08x\n",  mci_readl(host, STATUS));
	seq_printf(s, "RINTSTS:\t0x%08x\n", mci_readl(host, RINTSTS));
	seq_printf(s, "CMD:\t0x%08x\n",     mci_readl(host, CMD));
	seq_printf(s, "CTRL:\t0x%08x\n",    mci_readl(host, CTRL));
	seq_printf(s, "INTMASK:\t0x%08x\n", mci_readl(host, INTMASK));
	seq_printf(s, "CLKENA:\t0x%08x\n",  mci_readl(host, CLKENA));

	clk_disable(host->clk);

	return 0;
}

static int dw_mci_regs_open(struct inode *inode, struct file *file)
{
	return single_open(file, dw_mci_regs_show, inode->i_private);
}

static const struct file_operations dw_mci_regs_fops = {
	.owner		= THIS_MODULE,
	.open		= dw_mci_regs_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= single_release,
};

static void dw_mci_init_debugfs(struct dw_mci_slot *slot)
{
	struct mmc_host	*mmc = slot->mmc;
	struct dw_mci *host = slot->host;
	struct dentry *root;
	struct dentry *node;

	root = mmc->debugfs_root;
	if (!root)
		return;

	node = debugfs_create_file("regs", S_IRUSR, root, host,
				   &dw_mci_regs_fops);
	if (!node)
		goto err;

	node = debugfs_create_file("req", S_IRUSR, root, slot,
				   &dw_mci_req_fops);
	if (!node)
		goto err;

	node = debugfs_create_u32("state", S_IRUSR, root, (u32 *)&host->state);
	if (!node)
		goto err;

	node = debugfs_create_x32("pending_events", S_IRUSR, root,
				  (u32 *)&host->pending_events);
	if (!node)
		goto err;

	node = debugfs_create_x32("completed_events", S_IRUSR, root,
				  (u32 *)&host->completed_events);
	if (!node)
		goto err;

	return;

err:
	dev_err(&mmc->class_dev, "failed to initialize debugfs for slot\n");
}
#endif /* defined(CONFIG_DEBUG_FS) */

static bool mci_wait_reset(struct dw_mci *host, unsigned int resets)
{
	unsigned long timeout = jiffies + msecs_to_jiffies(500);
	unsigned int ctrl;

	ctrl = mci_readl(host, CTRL) | resets;
	mci_writel(host, CTRL, ctrl);

	/* wait till resets clear */
	do {
		ctrl = mci_readl(host, CTRL);
		if (!(ctrl & resets))
			return true;
	} while (time_before(jiffies, timeout));

	dev_err(&host->pdev->dev, "Timeout resetting block (ctrl %#x)\n", ctrl);

	return false;
}

static void dw_mci_set_timeout(struct dw_mci *host)
{
	unsigned long timeout;

	timeout = host->mrq->data->timeout_ns / host->cur_slot->cycle_ns;
	timeout += host->mrq->data->timeout_clks;
	if (timeout > 0x0FFFFFFFul)
		timeout = 0x0FFFFFFFul;

	mci_writel(host, TMOUT, (timeout << 8) | 0x40);
}

static u32 dw_mci_prepare_command(struct mmc_host *mmc, struct mmc_command *cmd)
{
	struct mmc_data	*data;
	u32 cmdr;
	cmd->error = -EINPROGRESS;

	cmdr = cmd->opcode;

	if (cmdr == MMC_STOP_TRANSMISSION)
		cmdr |= SDMMC_CMD_STOP;
	else
		cmdr |= SDMMC_CMD_PRV_DAT_WAIT;

	if (cmd->flags & MMC_RSP_PRESENT) {
		/* We expect a response, so set this bit */
		cmdr |= SDMMC_CMD_RESP_EXP;
		if (cmd->flags & MMC_RSP_136)
			cmdr |= SDMMC_CMD_RESP_LONG;
	}

	if (cmd->flags & MMC_RSP_CRC)
		cmdr |= SDMMC_CMD_RESP_CRC;

	cmdr |= SDMMC_CMD_USE_HOLD_REG;

	data = cmd->data;
	if (data) {
		cmdr |= SDMMC_CMD_DAT_EXP;
		if (data->flags & MMC_DATA_STREAM)
			cmdr |= SDMMC_CMD_STRM_MODE;
		if (data->flags & MMC_DATA_WRITE)
			cmdr |= SDMMC_CMD_DAT_WR;
	}

	return cmdr;
}

static bool mci_send_cmd(struct dw_mci_slot *slot, u32 cmd, u32 arg)
{
	struct dw_mci *host = slot->host;
	unsigned long timeout = jiffies + msecs_to_jiffies(500);
	unsigned int cmd_status = 0;

	mci_writel(host, CMDARG, arg);
	wmb();
	mci_writel(host, CMD, SDMMC_CMD_START | SDMMC_CMD_CARD_NUM(slot->id) | cmd);

	do {
		cmd_status = mci_readl(host, CMD);
		if (!(cmd_status & SDMMC_CMD_START))
			return true;
	} while (time_before(jiffies, timeout));

	dev_err(&host->pdev->dev,
		"Timeout sending command (slot %d, cmd %#x arg %#x):\n",
		slot->id, cmd, arg);
	dev_err(&host->pdev->dev,
		"    CMD %#x CLKENA %#x CTYPE %#x STATUS %#x\n",
		cmd_status, mci_readl(host, CLKENA), mci_readl(host, CTYPE),
		mci_readl(host, STATUS));
	dev_err(&host->pdev->dev, "    cmd_status %#x data_status %#x\n",
		host->cmd_status, host->data_status);
	dev_err(&host->pdev->dev,
		"    pending_events %#lx completed_events %#lx state %d\n",
		host->pending_events, host->completed_events, host->state);

	mci_wait_reset(host, SDMMC_CTRL_RESET | SDMMC_CTRL_FIFO_RESET |
			     SDMMC_CTRL_DMA_RESET);

	return false;
}

static void dw_mci_start_command(struct dw_mci *host,
				 struct mmc_command *cmd, u32 cmd_flags)
{
	host->cmd = cmd;
	host->cmd_status = 0;

	dev_vdbg(&host->pdev->dev,
		 "start command: ARGR=0x%08x CMDR=0x%08x\n",
		 cmd->arg, cmd_flags);

	mci_send_cmd(host->cur_slot, cmd_flags, cmd->arg);
}

static void send_stop_cmd(struct dw_mci *host, struct mmc_data *data)
{
	dw_mci_start_command(host, data->stop, host->stop_cmdr);
	set_bit(EVENT_SENT_STOP, &host->pending_events);
}

/*
 * The BIU data state machine jams if a start bit error happens. The only way
 * to recover from this situation is to send a stop command, even if the upper
 * layer didn't provide one.
 */
static void send_fake_stop(struct dw_mci *host)
{
	u32 cmd = SDMMC_CMD_STOP | SDMMC_CMD_RESP_EXP | 12;

	host->sent_fake_stop = 1;
	mci_send_cmd(host->cur_slot, cmd, 0x00);
	set_bit(EVENT_SENT_STOP, &host->pending_events);
}

/* DMA interface functions */
static void dw_mci_stop_dma(struct dw_mci *host)
{
	if (host->using_dma) {
		host->dma_ops->stop(host);
		host->using_dma = 0;
	}
}

#ifdef CONFIG_MMC_DW_DSPG_IDMAC
static void dw_mci_idmac_stop_dma(struct dw_mci *host)
{
	struct mmc_data *data = host->data;
	u32 temp;

	/* Stop the IDMAC running */
	temp = mci_readl(host, BMOD);
	temp &= ~(SDMMC_IDMAC_ENABLE | SDMMC_IDMAC_FB);
	mci_writel(host, BMOD, temp);

	wmb();

	/* Disable the IDMAC interface */
	temp = mci_readl(host, CTRL);
	temp &= ~SDMMC_CTRL_USE_IDMAC;
	mci_writel(host, CTRL, temp);

	if (data)
		dma_unmap_sg(&host->pdev->dev, data->sg, data->sg_len,
			     ((data->flags & MMC_DATA_WRITE)
			      ? DMA_TO_DEVICE : DMA_FROM_DEVICE));
}

static void dw_mci_translate_sglist(struct dw_mci *host, struct mmc_data *data,
				    unsigned int sg_len)
{
	int i;
	struct idmac_desc *desc = host->sg_cpu;

	for (i = 0; i < sg_len; i++, desc++) {
		unsigned int length = sg_dma_len(&data->sg[i]);
		u32 mem_addr = sg_dma_address(&data->sg[i]);

		/* Set the OWN bit and disable interrupts for this descriptor */
		desc->des0 = IDMAC_DES0_OWN | IDMAC_DES0_DIC | IDMAC_DES0_CH;

		/* Buffer length */
		IDMAC_SET_BUFFER1_SIZE(desc, length);

		/* Physical address to DMA to/from */
		desc->des2 = mem_addr;
	}

	/* Set first descriptor */
	desc = host->sg_cpu;
	desc->des0 |= IDMAC_DES0_FD;

	/* Set last descriptor */
	desc = host->sg_cpu + (i - 1) * sizeof(struct idmac_desc);
	desc->des0 &= ~(IDMAC_DES0_CH | IDMAC_DES0_DIC);
	desc->des0 |= IDMAC_DES0_LD;

	wmb();
}

static void dw_mci_idmac_start_dma(struct dw_mci *host, unsigned int sg_len)
{
	u32 temp;

	dw_mci_translate_sglist(host, host->data, sg_len);

	/* Reset DMA controller */
	mci_writel(host, BMOD, SDMMC_IDMAC_SWRESET);
	while (mci_readl(host, BMOD) & SDMMC_IDMAC_SWRESET)
		;

	/* Mask out interrupts - get Tx & Rx complete only */
	mci_writel(host, IDINTEN, SDMMC_IDMAC_INT_NI | SDMMC_IDMAC_INT_RI |
		   SDMMC_IDMAC_INT_TI);

	/* Set the descriptor base address */
	mci_writel(host, DBADDR, host->sg_dma);

	/* Select IDMAC interface */
	temp = mci_readl(host, CTRL);
	temp |= SDMMC_CTRL_USE_IDMAC;
	mci_writel(host, CTRL, temp);

	wmb();

	/* Enable the IDMAC */
	temp = mci_readl(host, BMOD);
	temp |= SDMMC_IDMAC_ENABLE | SDMMC_IDMAC_FB;
	mci_writel(host, BMOD, temp);
}

static int dw_mci_idmac_init(struct dw_mci *host)
{
	struct idmac_desc *p;
	int i;

	/* Number of descriptors in the ring buffer */
	host->ring_size = PAGE_SIZE / sizeof(struct idmac_desc);

	/* Forward link the descriptor list */
	for (i = 0, p = host->sg_cpu; i < host->ring_size - 1; i++, p++)
		p->des3 = host->sg_dma + (sizeof(struct idmac_desc) * (i + 1));

	/* Set the last descriptor as the end-of-ring descriptor */
	p->des3 = host->sg_dma;
	p->des0 = IDMAC_DES0_ER;

	return 0;
}

static struct dw_mci_dma_ops dw_mci_idmac_ops = {
	.init = dw_mci_idmac_init,
	.start = dw_mci_idmac_start_dma,
	.stop = dw_mci_idmac_stop_dma,
};
#endif /* CONFIG_MMC_DW_DSPG_IDMAC */

static int dw_mci_submit_data_dma(struct dw_mci *host, struct mmc_data *data)
{
	struct scatterlist *sg;
	unsigned int i, direction, sg_len;

	host->using_dma = 0;

	/* If we don't have a channel, we can't do DMA */
	if (!host->use_dma)
		return -ENODEV;

	/*
	 * We don't do DMA on "complex" transfers, i.e. with
	 * non-word-aligned buffers or lengths. Also, we don't bother
	 * with all the DMA setup overhead for short transfers.
	 */
	if (data->blocks * data->blksz < DW_MCI_DMA_THRESHOLD)
		return -EINVAL;
	if (data->blksz & 3)
		return -EINVAL;

	for_each_sg(data->sg, sg, data->sg_len, i) {
		if (sg->offset & 3 || sg->length & 3)
			return -EINVAL;
	}

	host->using_dma = 1;

	if (data->flags & MMC_DATA_READ)
		direction = DMA_FROM_DEVICE;
	else
		direction = DMA_TO_DEVICE;

	sg_len = dma_map_sg(&host->pdev->dev, data->sg, data->sg_len,
			    direction);

	dev_vdbg(&host->pdev->dev,
		 "sd sg_cpu: %#lx sg_dma: %#lx sg_len: %d\n",
		 (unsigned long)host->sg_cpu, (unsigned long)host->sg_dma,
		 sg_len);

	host->dma_ops->start(host, sg_len);

	return 0;
}

static void dw_mci_submit_data(struct dw_mci *host, struct mmc_data *data)
{
	u32 temp;

	data->error = -EINPROGRESS;

	WARN_ON(host->data);
	host->sg = NULL;
	host->data = data;

	if (data->flags & MMC_DATA_READ)
		host->dir_status = DW_MCI_RECV_STATUS;
	else
		host->dir_status = DW_MCI_SEND_STATUS;

	if (dw_mci_submit_data_dma(host, data)) {
		host->sg = data->sg;
		host->pio_offset = 0;
		host->part_buf_start = 0;
		host->part_buf_count = 0;

		mci_writel(host, RINTSTS, SDMMC_INT_TXDR | SDMMC_INT_RXDR);
		wmb();
		temp = mci_readl(host, INTMASK);
		temp |= SDMMC_INT_TXDR | SDMMC_INT_RXDR;
		mci_writel(host, INTMASK, temp);
	}
}

static bool dw_mci_setup_bus(struct dw_mci_slot *slot)
{
	struct dw_mci *host = slot->host;

	if (slot->clock != host->current_speed || host->cur_slot != slot) {
		/*
		 * The clock may only be changed if the current card is not busy!
		 */
		if (SDMMC_DATA_BUSY(mci_readl(host, STATUS)))
			return false;

		/* disable clock */
		mci_writel(host, CLKENA, 0);
		mci_writel(host, CLKSRC, 0);

		/* inform CIU */
		if (!mci_send_cmd(slot,
				  SDMMC_CMD_UPD_CLK | SDMMC_CMD_PRV_DAT_WAIT, 0))
			return false;

		/* set clock to desired speed */
		mci_writel(host, CLKDIV, slot->div);

		/* inform CIU */
		if (!mci_send_cmd(slot,
				  SDMMC_CMD_UPD_CLK | SDMMC_CMD_PRV_DAT_WAIT, 0))
			return false;

		/* enable clock */
		mci_writel(host, CLKENA, (SDMMC_CLKEN_ENABLE |
			   SDMMC_CLKEN_LOW_PWR) << slot->id);

		/* inform CIU */
		if (!mci_send_cmd(slot,
				  SDMMC_CMD_UPD_CLK | SDMMC_CMD_PRV_DAT_WAIT, 0))
			return false;

		host->current_speed = slot->clock;
	}

	/* Set the current slot bus width */
	mci_writel(host, CTYPE, (slot->ctype << slot->id));

	return true;
}

static void dw_mci_start_request(struct dw_mci *host)
{
	struct mmc_request *mrq = NULL;
	struct mmc_command *cmd;
	struct mmc_data	*data;
	struct dw_mci_slot *slot = NULL;
	LIST_HEAD(stalled_queue);
	u32 cmdflags;

	/*
	 * In case of a retry delay the poll_timer is already running. Once the
	 * timer expires will will get back here with the flag cleared.
	 */
	if (host->retry_delay) {
		dev_vdbg(&host->pdev->dev, "%s: retry bail out\n", __func__);
		return;
	}

	dw_mci_req_busy(host);

	while (!list_empty(&host->queue)) {
		slot = list_entry(host->queue.next, struct dw_mci_slot,
				  queue_node);
		list_del(&slot->queue_node);

		/* Slot specific timing and width adjustment */
		if (dw_mci_setup_bus(slot)) {
			dev_vdbg(&host->pdev->dev, "%s: %s is next\n", __func__,
				 mmc_hostname(slot->mmc));
			mrq = slot->mrq;
			if (host->pdata->select_slot)
				host->pdata->select_slot(slot->id);
			break;
		} else {
			dev_vdbg(&host->pdev->dev, "%s: %s stalled\n", __func__,
				 mmc_hostname(slot->mmc));
			list_add_tail(&slot->queue_node, &stalled_queue);
			mod_timer(&host->poll_timer,
				  jiffies + msecs_to_jiffies(10));
		}
	}

	/* re-insert stalled requests */
	list_splice(&stalled_queue, &host->queue);
	if (!mrq)
		return;

	host->state = STATE_SENDING_CMD;
	host->cur_slot = slot;
	host->mrq = mrq;

	host->pending_events = 0;
	host->completed_events = 0;
	host->data_status = 0;

	data = mrq->data;
	if (data) {
		/*
		 * Reset the FIFO to remove any data still lingering around.
		 * Its also mandatory to reset the DMA interface. Otherwise we
		 * may get unexplainable FIFO overruns if we mix PIO and DMA
		 * transfers.
		 */
		mci_wait_reset(host, SDMMC_CTRL_FIFO_RESET | SDMMC_CTRL_DMA_RESET);
		dw_mci_set_timeout(host);
		mci_writel(host, BYTCNT, data->blksz*data->blocks);
		mci_writel(host, BLKSIZ, data->blksz);
	}

	cmd = mrq->cmd;
	cmdflags = dw_mci_prepare_command(slot->mmc, cmd);

	/* this is the first command, send the initialization clock */
	if (test_and_clear_bit(DW_MMC_CARD_NEED_INIT, &slot->flags))
		cmdflags |= SDMMC_CMD_INIT;

	if (data) {
		dw_mci_submit_data(host, data);
		wmb();
	}

	dw_mci_start_command(host, cmd, cmdflags);

	if (mrq->stop)
		host->stop_cmdr = dw_mci_prepare_command(slot->mmc, mrq->stop);
}

/* must be called with host->lock held */
static void dw_mci_queue_request(struct dw_mci *host, struct dw_mci_slot *slot,
				 struct mmc_request *mrq)
{
	dev_vdbg(&slot->mmc->class_dev, "queue request: state=%d\n",
		 host->state);

	slot->mrq = mrq;

	list_add_tail(&slot->queue_node, &host->queue);
	if (host->state == STATE_IDLE)
		dw_mci_start_request(host);
}

static void dw_mci_request(struct mmc_host *mmc, struct mmc_request *mrq)
{
	struct dw_mci_slot *slot = mmc_priv(mmc);
	struct dw_mci *host = slot->host;

	WARN_ON(slot->mrq);

	/*
	 * The check for card presence and queueing of the request must be
	 * atomic, otherwise the card could be removed in between and the
	 * request wouldn't fail until another card was inserted.
	 */
	spin_lock_bh(&host->lock);

	if (!test_bit(DW_MMC_CARD_PRESENT, &slot->flags) ||
	    !test_bit(DW_MMC_CARD_POWERED, &slot->flags)) {
		spin_unlock_bh(&host->lock);
		mrq->cmd->error = -ENOMEDIUM;
		mmc_request_done(mmc, mrq);
		return;
	}

	dw_mci_queue_request(host, slot, mrq);

	spin_unlock_bh(&host->lock);
}

static void dw_mci_poll_queue(unsigned long data)
{
	struct dw_mci *host = (struct dw_mci *)data;

	/*
	 * Check if the controller is idle and there is still a request in the
	 * queue. This may happen if the request could not be processed because
	 * another card was busy or if we have to wait before a retry.
	 */
	spin_lock(&host->lock);

	dev_vdbg(&host->pdev->dev, "%s\n", __func__);
	host->retry_delay = 0;
	if (host->state == STATE_IDLE && !list_empty(&host->queue))
		dw_mci_start_request(host);

	spin_unlock(&host->lock);
}

static void dw_mci_set_power(struct dw_mci_slot *slot,
			     unsigned char power_mode,
			     unsigned int vdd)
{
	switch (power_mode) {
	case MMC_POWER_UP:
		set_bit(DW_MMC_CARD_NEED_INIT, &slot->flags);
		set_bit(DW_MMC_CARD_POWERED, &slot->flags);
		if (slot->vmmc)
			mmc_regulator_set_ocr(slot->mmc, slot->vmmc, vdd);
		break;

	case MMC_POWER_OFF:
		clear_bit(DW_MMC_CARD_POWERED, &slot->flags);
		if (slot->vmmc)
			mmc_regulator_set_ocr(slot->mmc, slot->vmmc, 0);
		break;
	default:
		break;
	}
}

static void dw_mci_set_ios(struct mmc_host *mmc, struct mmc_ios *ios)
{
	struct dw_mci_slot *slot = mmc_priv(mmc);
	u32 regs;

	/* set default 1 bit mode */
	slot->ctype = SDMMC_CTYPE_1BIT;

	switch (ios->bus_width) {
	case MMC_BUS_WIDTH_1:
		slot->ctype = SDMMC_CTYPE_1BIT;
		break;
	case MMC_BUS_WIDTH_4:
		slot->ctype = SDMMC_CTYPE_4BIT;
		break;
	case MMC_BUS_WIDTH_8:
		slot->ctype = SDMMC_CTYPE_8BIT;
		break;
	}

	regs = mci_readl(slot->host, UHS_REG);

	/* DDR mode set */
	if (ios->timing == MMC_TIMING_UHS_DDR50)
		regs |= (0x1 << slot->id) << 16;
	else
		regs &= ~(0x1 << slot->id) << 16;

	mci_writel(slot->host, UHS_REG, regs);

	if (ios->clock && ios->clock != slot->clock) {
		unsigned int slot_hz;
		struct dw_mci *host = slot->host;

		if (host->bus_hz % ios->clock)
			/*
			 * move the + 1 after the divide to prevent
			 * over-clocking the card.
			 */
			slot->div = ((host->bus_hz / ios->clock) >> 1) + 1;
		else
			slot->div = (host->bus_hz  / ios->clock) >> 1;

		slot_hz = slot->div ? ((host->bus_hz / slot->div) >> 1)
				    : host->bus_hz;
		slot->cycle_ns = 1000000000 / slot_hz;
		slot->clock = ios->clock;

		dev_info(&slot->mmc->class_dev,
			 "Bus speed (slot %d) = %dHz (slot req %dHz, actual %dHZ"
			 " div = %d)\n", slot->id, host->bus_hz, slot->clock,
			 slot_hz, slot->div);
	}

	if (slot->power_mode != ios->power_mode) {
		dw_mci_set_power(slot, ios->power_mode, ios->vdd);
		slot->power_mode = ios->power_mode;
	}
}

static int dw_mci_get_ro(struct mmc_host *mmc)
{
	int read_only;
	struct dw_mci_slot *slot = mmc_priv(mmc);

	/* Use GPIO, else try on board write protect */
	if (mmc->caps & MMC_CAP_NONREMOVABLE)
		read_only = 0;
	else if (gpio_is_valid(slot->pdata->wp_gpio))
		read_only = gpio_get_value(slot->pdata->wp_gpio);
	else
		read_only =
			mci_readl(slot->host, WRTPRT) & (1 << slot->id) ? 1 : 0;

	if (slot->pdata->wp_inverted)
		read_only = !read_only;

	dev_dbg(&mmc->class_dev, "card is %s\n",
		read_only ? "read-only" : "read-write");

	return read_only;
}

static int dw_mci_get_cd(struct mmc_host *mmc)
{
	int present;
	struct dw_mci_slot *slot = mmc_priv(mmc);
	struct dw_mci_board *brd = slot->host->pdata;

	/* Use GPIO, else try onboard card detect */
	if (mmc->caps & MMC_CAP_NONREMOVABLE)
		present = 1;
	else if (brd->quirks & DW_MCI_QUIRK_BROKEN_CARD_DETECTION)
		present = 1;
	else if (gpio_is_valid(slot->pdata->cd_gpio))
		present = !gpio_get_value(slot->pdata->cd_gpio);
	else
		present = (mci_readl(slot->host, CDETECT) & (1 << slot->id))
			== 0 ? 1 : 0;

	if (slot->pdata->cd_inverted)
		present = !present;

	if (present)
		dev_dbg(&mmc->class_dev, "card is present\n");
	else
		dev_dbg(&mmc->class_dev, "card is not present\n");

	return present;
}

static void dw_mci_enable_sdio_irq(struct mmc_host *mmc, int enb)
{
	struct dw_mci_slot *slot = mmc_priv(mmc);
	struct dw_mci *host = slot->host;
	u32 int_mask;

	spin_lock_bh(&host->lock);

	clk_enable(host->clk);

	/* Enable/disable Slot Specific SDIO interrupt */
	int_mask = mci_readl(host, INTMASK);
	if (enb && !host->sdio_irq_enabled) {
		clk_enable(host->clk);
		mci_writel(host, INTMASK,
			   (int_mask | SDMMC_INT_SDIO(slot->id)));
		host->sdio_irq_enabled = 1;
	} else if (!enb && host->sdio_irq_enabled) {
		mci_writel(host, INTMASK,
			   (int_mask & ~(SDMMC_INT_SDIO(slot->id))));
		clk_disable(host->clk);
		host->sdio_irq_enabled = 0;
	}

	clk_disable(host->clk);

	spin_unlock_bh(&host->lock);
}

static const struct mmc_host_ops dw_mci_ops = {
	.request		= dw_mci_request,
	.set_ios		= dw_mci_set_ios,
	.get_ro			= dw_mci_get_ro,
	.get_cd			= dw_mci_get_cd,
	.enable_sdio_irq	= dw_mci_enable_sdio_irq,
};

static void dw_mci_request_end(struct dw_mci *host, struct mmc_request *mrq)
	__releases(&host->lock)
	__acquires(&host->lock)
{
	struct mmc_host	*prev_mmc = host->cur_slot->mmc;
	u32 temp;

	WARN_ON(host->cmd || host->data);

	dw_mci_stop_dma(host);

	/* Always disable RX/TX/HTO IRQs at the end. */
	temp = mci_readl(host, INTMASK);
	temp &= ~(SDMMC_INT_RXDR | SDMMC_INT_TXDR | SDMMC_INT_HTO);
	mci_writel(host, INTMASK, temp);

	host->cur_slot->mrq = NULL;
	host->mrq = NULL;
	host->state = STATE_IDLE;
	if (!list_empty(&host->queue)) {
		dw_mci_start_request(host);
	} else {
		dev_vdbg(&host->pdev->dev, "list empty\n");
		dw_mci_req_idle(host);
	}

	spin_unlock(&host->lock);
	mmc_request_done(prev_mmc, mrq);
	spin_lock(&host->lock);
}

static void dw_mci_command_complete(struct dw_mci *host, struct mmc_command *cmd)
{
	u32 status = host->cmd_status;

	/* Read the response from the card (up to 16 bytes) */
	if (cmd->flags & MMC_RSP_PRESENT) {
		if (cmd->flags & MMC_RSP_136) {
			cmd->resp[3] = mci_readl(host, RESP0);
			cmd->resp[2] = mci_readl(host, RESP1);
			cmd->resp[1] = mci_readl(host, RESP2);
			cmd->resp[0] = mci_readl(host, RESP3);
		} else {
			cmd->resp[0] = mci_readl(host, RESP0);
			cmd->resp[1] = 0;
			cmd->resp[2] = 0;
			cmd->resp[3] = 0;
		}
	}

	if (status & SDMMC_INT_RTO)
		cmd->error = -ETIMEDOUT;
	else if ((cmd->flags & MMC_RSP_CRC) && (status & SDMMC_INT_RCRC))
		cmd->error = -EILSEQ;
	else if (status & SDMMC_INT_RESP_ERR)
		cmd->error = -EIO;
	else
		cmd->error = 0;

	if (cmd->error) {
		/* newer ip versions need a delay between retries */
		if (host->quirks & DW_MCI_QUIRK_RETRY_DELAY) {
			host->retry_delay = 1;
			mod_timer(&host->poll_timer,
				  jiffies + msecs_to_jiffies(20));
		}
	}
}

static int dw_mci_data_cmd_failed(struct dw_mci *host, struct mmc_command *cmd)
{
	u32 temp, status = host->cmd_status;

	/*
	 * On a response timeout error the DMA and data state machines are
	 * typiclly idle. But there is a slight window of opportunity that the
	 * data path still started receiving, namely if a start bit was
	 * detected before the response timeout happended. This case is handled
	 * below.
	 */
	if (status & SDMMC_INT_RTO && !SDMMC_MC_BUSY(mci_readl(host, STATUS))) {
		host->data->error = -ETIMEDOUT;
		if (host->data->stop)
			host->data->stop->error = -ETIMEDOUT;
		host->data = NULL;

		dev_dbg(&host->pdev->dev, "response timeout (status=%08x)\n",
			status);

		dw_mci_request_end(host, host->mrq);
		return STATE_IDLE;
	}

	/*
	 * The CIU data state machine is armed in this state. Either we get
	 * busy if the card answers or we run into a data timeout. The internal
	 * DMA has already bailed out, except for pure SDMMC_INT_RESP_ERRs. We
	 * stop the transfer and switch to PIO to dump any remaining data!
	 */
	if ((status & DW_MCI_CMD_ERROR_FLAGS) != SDMMC_INT_RESP_ERR) {
		if (host->data->stop)
			send_stop_cmd(host, host->data);

		/* stop DMA and mark data as bad */
		dw_mci_stop_dma(host);
		host->data_status = SDMMC_INT_DCRC;

		/*
		 * Enable RX/TX and HTO interrupts. Somehow the PIO interrupts
		 * are sometimes not triggering but the HTO can save us if it
		 * happens
		 */
		temp = mci_readl(host, INTMASK);
		temp |= SDMMC_INT_TXDR | SDMMC_INT_RXDR | SDMMC_INT_HTO;
		mci_writel(host, INTMASK, temp);

		dev_dbg(&host->pdev->dev, "response error (status=0x%08x)"
			"; dump FIFO with PIO\n", status);
		return STATE_DATA_BUSY;
	} else {
		dev_dbg(&host->pdev->dev, "plain response error\n");
		return STATE_SENDING_DATA;
	}
}

static void dw_mci_tasklet_func(unsigned long priv)
{
	struct dw_mci *host = (struct dw_mci *)priv;
	struct mmc_data	*data;
	struct mmc_command *cmd;
	enum dw_mci_state state;
	enum dw_mci_state prev_state;
	u32 status;

	spin_lock(&host->lock);

	state = host->state;
	data = host->data;

	do {
		prev_state = state;

		switch (state) {
		case STATE_IDLE:
			break;

		case STATE_SENDING_CMD:
			if (!test_and_clear_bit(EVENT_CMD_COMPLETE,
						&host->pending_events))
				break;

			cmd = host->cmd;
			host->cmd = NULL;
			set_bit(EVENT_CMD_COMPLETE, &host->completed_events);
			dw_mci_command_complete(host, host->mrq->cmd);

			if (!cmd->data) {
				dw_mci_request_end(host, host->mrq);
				goto unlock;
			} else if (cmd->error) {
				state = dw_mci_data_cmd_failed(host, cmd);
				if (state == STATE_IDLE)
					goto unlock;
				break;
			}

			prev_state = state = STATE_SENDING_DATA;
			/* fall through */

		case STATE_SENDING_DATA:
			if (test_and_clear_bit(EVENT_DATA_ERROR,
					       &host->pending_events)) {
				if (data->stop)
					send_stop_cmd(host, data);
				else if (host->data_status & SDMMC_INT_SBE &&
					 host->quirks & DW_MCI_QUIRK_SBE_NEED_STOP)
					send_fake_stop(host);

				if (host->data_status & SDMMC_INT_DTO)
					set_bit(EVENT_DATA_COMPLETE, &host->pending_events);

				set_bit(EVENT_DATA_ERROR, &host->completed_events);
				state = STATE_DATA_BUSY;
				break;
			}

			if (!test_and_clear_bit(EVENT_XFER_COMPLETE,
						&host->pending_events))
				break;

			set_bit(EVENT_XFER_COMPLETE, &host->completed_events);
			prev_state = state = STATE_DATA_BUSY;
			/* fall through */

		case STATE_DATA_BUSY:
			if (test_and_clear_bit(EVENT_DATA_ERROR,
					       &host->pending_events)) {
				if (host->data_status & SDMMC_INT_SBE &&
				    host->quirks & DW_MCI_QUIRK_SBE_NEED_STOP &&
				    !data->stop && !host->sent_fake_stop)
					send_fake_stop(host);

				set_bit(EVENT_DATA_ERROR, &host->completed_events);
			}

			if (!test_and_clear_bit(EVENT_DATA_COMPLETE,
						&host->pending_events))
				break;

			host->data = NULL;
			set_bit(EVENT_DATA_COMPLETE, &host->completed_events);
			status = host->data_status;
			host->data_status = 0;

			if (status & DW_MCI_DATA_ERROR_FLAGS) {
				if (status & SDMMC_INT_DTO) {
					data->error = -ETIMEDOUT;
				} else if (status & SDMMC_INT_DCRC) {
					data->error = -EILSEQ;
				} else if (status & SDMMC_INT_EBE &&
					   host->dir_status ==
							DW_MCI_SEND_STATUS) {
					/*
					 * No data CRC status was returned.
					 * The number of bytes transferred will
					 * be exaggerated in PIO mode.
					 */
					data->bytes_xfered = 0;
					data->error = -ETIMEDOUT;
				} else {
					dev_err(&host->pdev->dev,
						"data FIFO error "
						"(status=%08x)\n",
						status);
					data->error = -EIO;
				}
			} else {
				data->bytes_xfered = data->blocks * data->blksz;
				data->error = 0;
			}

			if (!data->stop && !host->sent_fake_stop) {
				dw_mci_request_end(host, host->mrq);
				goto unlock;
			}

			prev_state = state = STATE_SENDING_STOP;
			if (!test_and_clear_bit(EVENT_SENT_STOP, &host->pending_events))
				send_stop_cmd(host, data);
			set_bit(EVENT_SENT_STOP, &host->completed_events);
			/* fall through */

		case STATE_SENDING_STOP:
			if (!test_and_clear_bit(EVENT_CMD_COMPLETE,
						&host->pending_events))
				break;

			host->cmd = NULL;
			if (!host->sent_fake_stop)
				dw_mci_command_complete(host, host->mrq->stop);
			else
				host->sent_fake_stop = 0;
			dw_mci_request_end(host, host->mrq);
			goto unlock;
		}
	} while (state != prev_state);

	host->state = state;
unlock:
	spin_unlock(&host->lock);
}

/* push final bytes to part_buf, only use during push */
static void dw_mci_set_part_bytes(struct dw_mci *host, void *buf, int cnt)
{
	memcpy((void *)&host->part_buf, buf, cnt);
	host->part_buf_count = cnt;
}

/* append bytes to part_buf, only use during push */
static int dw_mci_push_part_bytes(struct dw_mci *host, void *buf, int cnt)
{
	cnt = min(cnt, (1 << host->data_shift) - host->part_buf_count);
	memcpy((void *)&host->part_buf + host->part_buf_count, buf, cnt);
	host->part_buf_count += cnt;
	return cnt;
}

/* pull first bytes from part_buf, only use during pull */
static int dw_mci_pull_part_bytes(struct dw_mci *host, void *buf, int cnt)
{
	cnt = min(cnt, (int)host->part_buf_count);
	if (cnt) {
		memcpy(buf, (void *)&host->part_buf + host->part_buf_start,
		       cnt);
		host->part_buf_count -= cnt;
		host->part_buf_start += cnt;
	}
	return cnt;
}

/* pull final bytes from the part_buf, assuming it's just been filled */
static void dw_mci_pull_final_bytes(struct dw_mci *host, void *buf, int cnt)
{
	memcpy(buf, &host->part_buf, cnt);
	host->part_buf_start = cnt;
	host->part_buf_count = (1 << host->data_shift) - cnt;
}

static void dw_mci_push_data16(struct dw_mci *host, void *buf, int cnt)
{
	/* try and push anything in the part_buf */
	if (unlikely(host->part_buf_count)) {
		int len = dw_mci_push_part_bytes(host, buf, cnt);
		buf += len;
		cnt -= len;
		if (!sg_next(host->sg) || host->part_buf_count == 2) {
			mci_writew(host, DATA, host->part_buf16);
			host->part_buf_count = 0;
		}
	}
#ifndef CONFIG_HAVE_EFFICIENT_UNALIGNED_ACCESS
	if (unlikely((unsigned long)buf & 0x1)) {
		while (cnt >= 2) {
			u16 aligned_buf[64];
			int len = min(cnt & -2, (int)sizeof(aligned_buf));
			int items = len >> 1;
			int i;
			/* memcpy from input buffer into aligned buffer */
			memcpy(aligned_buf, buf, len);
			buf += len;
			cnt -= len;
			/* push data from aligned buffer into fifo */
			for (i = 0; i < items; ++i)
				mci_writew(host, DATA, aligned_buf[i]);
		}
	} else
#endif
	{
		u16 *pdata = buf;
		for (; cnt >= 2; cnt -= 2)
			mci_writew(host, DATA, *pdata++);
		buf = pdata;
	}
	/* put anything remaining in the part_buf */
	if (cnt) {
		dw_mci_set_part_bytes(host, buf, cnt);
		if (!sg_next(host->sg))
			mci_writew(host, DATA, host->part_buf16);
	}
}

static void dw_mci_pull_data16(struct dw_mci *host, void *buf, int cnt)
{
#ifndef CONFIG_HAVE_EFFICIENT_UNALIGNED_ACCESS
	if (unlikely((unsigned long)buf & 0x1)) {
		while (cnt >= 2) {
			/* pull data from fifo into aligned buffer */
			u16 aligned_buf[64];
			int len = min(cnt & -2, (int)sizeof(aligned_buf));
			int items = len >> 1;
			int i;
			for (i = 0; i < items; ++i)
				aligned_buf[i] = mci_readw(host, DATA);
			/* memcpy from aligned buffer into output buffer */
			memcpy(buf, aligned_buf, len);
			buf += len;
			cnt -= len;
		}
	} else
#endif
	{
		u16 *pdata = buf;
		for (; cnt >= 2; cnt -= 2)
			*pdata++ = mci_readw(host, DATA);
		buf = pdata;
	}
	if (cnt) {
		host->part_buf16 = mci_readw(host, DATA);
		dw_mci_pull_final_bytes(host, buf, cnt);
	}
}

static void dw_mci_push_data32(struct dw_mci *host, void *buf, int cnt)
{
	/* try and push anything in the part_buf */
	if (unlikely(host->part_buf_count)) {
		int len = dw_mci_push_part_bytes(host, buf, cnt);
		buf += len;
		cnt -= len;
		if (!sg_next(host->sg) || host->part_buf_count == 4) {
			mci_writel(host, DATA, host->part_buf32);
			host->part_buf_count = 0;
		}
	}
#ifndef CONFIG_HAVE_EFFICIENT_UNALIGNED_ACCESS
	if (unlikely((unsigned long)buf & 0x3)) {
		while (cnt >= 4) {
			u32 aligned_buf[32];
			int len = min(cnt & -4, (int)sizeof(aligned_buf));
			int items = len >> 2;
			int i;
			/* memcpy from input buffer into aligned buffer */
			memcpy(aligned_buf, buf, len);
			buf += len;
			cnt -= len;
			/* push data from aligned buffer into fifo */
			for (i = 0; i < items; ++i)
				mci_writel(host, DATA, aligned_buf[i]);
		}
	} else
#endif
	{
		u32 *pdata = buf;
		for (; cnt >= 4; cnt -= 4)
			mci_writel(host, DATA, *pdata++);
		buf = pdata;
	}
	/* put anything remaining in the part_buf */
	if (cnt) {
		dw_mci_set_part_bytes(host, buf, cnt);
		if (!sg_next(host->sg))
			mci_writel(host, DATA, host->part_buf32);
	}
}

static void dw_mci_pull_data32(struct dw_mci *host, void *buf, int cnt)
{
#ifndef CONFIG_HAVE_EFFICIENT_UNALIGNED_ACCESS
	if (unlikely((unsigned long)buf & 0x3)) {
		while (cnt >= 4) {
			/* pull data from fifo into aligned buffer */
			u32 aligned_buf[32];
			int len = min(cnt & -4, (int)sizeof(aligned_buf));
			int items = len >> 2;
			int i;
			for (i = 0; i < items; ++i)
				aligned_buf[i] = mci_readl(host, DATA);
			/* memcpy from aligned buffer into output buffer */
			memcpy(buf, aligned_buf, len);
			buf += len;
			cnt -= len;
		}
	} else
#endif
	{
		u32 *pdata = buf;
		for (; cnt >= 4; cnt -= 4)
			*pdata++ = mci_readl(host, DATA);
		buf = pdata;
	}
	if (cnt) {
		host->part_buf32 = mci_readl(host, DATA);
		dw_mci_pull_final_bytes(host, buf, cnt);
	}
}

static void dw_mci_push_data64(struct dw_mci *host, void *buf, int cnt)
{
	/* try and push anything in the part_buf */
	if (unlikely(host->part_buf_count)) {
		int len = dw_mci_push_part_bytes(host, buf, cnt);
		buf += len;
		cnt -= len;
		if (!sg_next(host->sg) || host->part_buf_count == 8) {
			mci_writew(host, DATA, host->part_buf);
			host->part_buf_count = 0;
		}
	}
#ifndef CONFIG_HAVE_EFFICIENT_UNALIGNED_ACCESS
	if (unlikely((unsigned long)buf & 0x7)) {
		while (cnt >= 8) {
			u64 aligned_buf[16];
			int len = min(cnt & -8, (int)sizeof(aligned_buf));
			int items = len >> 3;
			int i;
			/* memcpy from input buffer into aligned buffer */
			memcpy(aligned_buf, buf, len);
			buf += len;
			cnt -= len;
			/* push data from aligned buffer into fifo */
			for (i = 0; i < items; ++i)
				mci_writeq(host, DATA, aligned_buf[i]);
		}
	} else
#endif
	{
		u64 *pdata = buf;
		for (; cnt >= 8; cnt -= 8)
			mci_writeq(host, DATA, *pdata++);
		buf = pdata;
	}
	/* put anything remaining in the part_buf */
	if (cnt) {
		dw_mci_set_part_bytes(host, buf, cnt);
		if (!sg_next(host->sg))
			mci_writeq(host, DATA, host->part_buf);
	}
}

static void dw_mci_pull_data64(struct dw_mci *host, void *buf, int cnt)
{
#ifndef CONFIG_HAVE_EFFICIENT_UNALIGNED_ACCESS
	if (unlikely((unsigned long)buf & 0x7)) {
		while (cnt >= 8) {
			/* pull data from fifo into aligned buffer */
			u64 aligned_buf[16];
			int len = min(cnt & -8, (int)sizeof(aligned_buf));
			int items = len >> 3;
			int i;
			for (i = 0; i < items; ++i)
				aligned_buf[i] = mci_readq(host, DATA);
			/* memcpy from aligned buffer into output buffer */
			memcpy(buf, aligned_buf, len);
			buf += len;
			cnt -= len;
		}
	} else
#endif
	{
		u64 *pdata = buf;
		for (; cnt >= 8; cnt -= 8)
			*pdata++ = mci_readq(host, DATA);
		buf = pdata;
	}
	if (cnt) {
		host->part_buf = mci_readq(host, DATA);
		dw_mci_pull_final_bytes(host, buf, cnt);
	}
}

static void dw_mci_pull_data(struct dw_mci *host, void *buf, int cnt)
{
	int len;

	/* get remaining partial bytes */
	len = dw_mci_pull_part_bytes(host, buf, cnt);
	if (unlikely(len == cnt))
		return;
	buf += len;
	cnt -= len;

	/* get the rest of the data */
	host->pull_data(host, buf, cnt);
}

static void dw_mci_read_data_pio(struct dw_mci *host)
{
	struct scatterlist *sg = host->sg;
	void *buf = sg_virt(sg);
	unsigned int offset = host->pio_offset;
	struct mmc_data	*data = host->data;
	int shift = host->data_shift;
	u32 status;
	unsigned int nbytes = 0, len;

	do {
		len = host->part_buf_count +
			(SDMMC_GET_FCNT(mci_readl(host, STATUS)) << shift);
		if (offset + len <= sg->length) {
			dw_mci_pull_data(host, (void *)(buf + offset), len);

			offset += len;
			nbytes += len;

			if (offset == sg->length) {
				flush_dcache_page(sg_page(sg));
				host->sg = sg = sg_next(sg);
				if (!sg)
					goto done;

				offset = 0;
				buf = sg_virt(sg);
			}
		} else {
			unsigned int remaining = sg->length - offset;
			dw_mci_pull_data(host, (void *)(buf + offset),
					 remaining);
			nbytes += remaining;

			flush_dcache_page(sg_page(sg));
			host->sg = sg = sg_next(sg);
			if (!sg)
				goto done;

			offset = len - remaining;
			buf = sg_virt(sg);
			dw_mci_pull_data(host, buf, offset);
			nbytes += offset;
		}

		status = mci_readl(host, MINTSTS);
		mci_writel(host, RINTSTS, SDMMC_INT_RXDR);
	} while (status & SDMMC_INT_RXDR); /*if the RXDR is ready read again*/
	host->pio_offset = offset;
	if (data)
		data->bytes_xfered += nbytes;
	return;

done:
	if (data) {
		data->bytes_xfered += nbytes;
		smp_wmb();
		set_bit(EVENT_XFER_COMPLETE, &host->pending_events);
	}
}

static void dw_mci_read_data_dummy(struct dw_mci *host)
{
	u32 status;

	do {
		unsigned int len = SDMMC_GET_FCNT(mci_readl(host, STATUS));

		switch (host->data_shift) {
		case 1:
			while (len--) {
				mci_readw(host, DATA);
				mci_readw(host, DATA);
			}
			break;
		case 2:
			while (len--)
				mci_readl(host, DATA);
			break;
		case 3:
			while (len--)
				mci_readq(host, DATA);
			break;
		}
		mci_writel(host, RINTSTS, SDMMC_INT_RXDR);
		status = mci_readl(host, MINTSTS);
	} while (status & SDMMC_INT_RXDR);
}

static void dw_mci_write_data_pio(struct dw_mci *host)
{
	struct scatterlist *sg = host->sg;
	void *buf = sg_virt(sg);
	unsigned int offset = host->pio_offset;
	struct mmc_data	*data = host->data;
	int shift = host->data_shift;
	u32 status;
	unsigned int nbytes = 0, len;

	do {
		len = ((host->fifo_depth -
			SDMMC_GET_FCNT(mci_readl(host, STATUS))) << shift)
			- host->part_buf_count;
		if (offset + len <= sg->length) {
			host->push_data(host, (void *)(buf + offset), len);

			offset += len;
			nbytes += len;
			if (offset == sg->length) {
				host->sg = sg = sg_next(sg);
				if (!sg)
					goto done;

				offset = 0;
				buf = sg_virt(sg);
			}
		} else {
			unsigned int remaining = sg->length - offset;

			host->push_data(host, (void *)(buf + offset),
					remaining);
			nbytes += remaining;

			host->sg = sg = sg_next(sg);
			if (!sg)
				goto done;

			offset = len - remaining;
			buf = sg_virt(sg);
			host->push_data(host, (void *)buf, offset);
			nbytes += offset;
		}

		status = mci_readl(host, MINTSTS);
		mci_writel(host, RINTSTS, SDMMC_INT_TXDR);
	} while (status & SDMMC_INT_TXDR); /* if TXDR write again */
	host->pio_offset = offset;
	if (data)
		data->bytes_xfered += nbytes;

	return;

done:
	if (data) {
		data->bytes_xfered += nbytes;
		smp_wmb();
		set_bit(EVENT_XFER_COMPLETE, &host->pending_events);
		tasklet_schedule(&host->tasklet);
	}
}

static void dw_mci_write_data_dummy(struct dw_mci *host)
{
	u32 status;

	do {
		unsigned int len = host->fifo_depth -
			SDMMC_GET_FCNT(mci_readl(host, STATUS));

		switch (host->data_shift) {
		case 1:
			while (len--) {
				mci_writew(host, DATA, 0);
				mci_writew(host, DATA, 0);
			}
			break;
		case 2:
			while (len--)
				mci_writel(host, DATA, 0);
			break;
		case 3:
			while (len--)
				mci_writeq(host, DATA, 0);
			break;
		}
		mci_writel(host, RINTSTS, SDMMC_INT_TXDR);
		status = mci_readl(host, MINTSTS);
	} while (status & SDMMC_INT_TXDR);
}

static irqreturn_t dw_mci_cdetect_interrupt(int irq, void *dev_id)
{
	struct dw_mci *host = dev_id;

	queue_work(dw_mci_card_workqueue, &host->card_work);
	return IRQ_HANDLED;
}

static irqreturn_t dw_mci_interrupt(int irq, void *dev_id)
{
	struct dw_mci *host = dev_id;
	u32 status, pending;
	unsigned int pass_count = 0;
	int i;

	do {
		status = mci_readl(host, RINTSTS);
		pending = mci_readl(host, MINTSTS); /* read-only mask reg */

		/*
		 * DTO fix - version 2.10a and below, and only if internal DMA
		 * is configured.
		 */
		if (host->quirks & DW_MCI_QUIRK_IDMAC_DTO) {
			if (!pending &&
			    ((mci_readl(host, STATUS) >> 17) & 0x1fff))
				pending |= SDMMC_INT_DATA_OVER;
		}

		if (!pending)
			break;

		if (pending & DW_MCI_DATA_ERROR_FLAGS) {
			/* if there is an error report DATA_ERROR */
			mci_writel(host, RINTSTS, pending & DW_MCI_DATA_ERROR_FLAGS);
			host->data_status |= status;
			smp_wmb();
			set_bit(EVENT_DATA_ERROR, &host->pending_events);
			tasklet_schedule(&host->tasklet);
		}

		if (pending & SDMMC_INT_DATA_OVER) {
			mci_writel(host, RINTSTS, SDMMC_INT_DATA_OVER);
			host->data_status |= status;
			smp_wmb();
			if (host->dir_status == DW_MCI_RECV_STATUS) {
				if (host->sg != NULL)
					dw_mci_read_data_pio(host);
			}

			set_bit(EVENT_DATA_COMPLETE, &host->pending_events);
			tasklet_schedule(&host->tasklet);
		}

		if (pending & SDMMC_INT_RXDR) {
			mci_writel(host, RINTSTS, SDMMC_INT_RXDR);
			if (host->dir_status == DW_MCI_RECV_STATUS && host->sg)
				dw_mci_read_data_pio(host);
			else
				dw_mci_read_data_dummy(host);
		}

		if (pending & SDMMC_INT_TXDR) {
			mci_writel(host, RINTSTS, SDMMC_INT_TXDR);
			if (host->dir_status == DW_MCI_SEND_STATUS && host->sg)
				dw_mci_write_data_pio(host);
			else
				dw_mci_write_data_dummy(host);
		}

		if (pending & SDMMC_INT_HTO) {
			dev_vdbg(&host->pdev->dev,
				 "HTO occured! Extra dummy PIO access.\n");
			if (host->dir_status == DW_MCI_RECV_STATUS)
				dw_mci_read_data_dummy(host);
			else
				dw_mci_write_data_dummy(host);
			mci_writel(host, RINTSTS, SDMMC_INT_HTO);
		}

		if (pending & SDMMC_INT_CMD_DONE) {
			mci_writel(host, RINTSTS, DW_MCI_CMD_ERROR_FLAGS |
				   SDMMC_INT_CMD_DONE);
			host->cmd_status = status;
			smp_wmb();
			set_bit(EVENT_CMD_COMPLETE, &host->pending_events);
			tasklet_schedule(&host->tasklet);
		}

		if (pending & SDMMC_INT_CD) {
			mci_writel(host, RINTSTS, SDMMC_INT_CD);
			queue_work(dw_mci_card_workqueue, &host->card_work);
		}

		/* Handle SDIO Interrupts */
		for (i = 0; i < host->num_slots; i++) {
			struct dw_mci_slot *slot = host->slot[i];
			if (pending & SDMMC_INT_SDIO(i)) {
				mci_writel(host, RINTSTS, SDMMC_INT_SDIO(i));
				mmc_signal_sdio_irq(slot->mmc);
			}
		}

	} while (pass_count++ < 5);

#ifdef CONFIG_MMC_DW_DSPG_IDMAC
	/* Handle DMA interrupts */
	pending = mci_readl(host, IDSTS);
	if (pending & (SDMMC_IDMAC_INT_TI | SDMMC_IDMAC_INT_RI)) {
		mci_writel(host, IDSTS, SDMMC_IDMAC_INT_TI | SDMMC_IDMAC_INT_RI);
		mci_writel(host, IDSTS, SDMMC_IDMAC_INT_NI);

		ccu_barrier();

		dev_vdbg(&host->pdev->dev, "IDMAC complete\n");

		set_bit(EVENT_XFER_COMPLETE, &host->pending_events);
		tasklet_schedule(&host->tasklet);
	}
#endif

	return IRQ_HANDLED;
}

static void dw_mci_work_routine_card(struct work_struct *work)
{
	struct dw_mci *host = container_of(work, struct dw_mci, card_work);
	int i;

	clk_enable(host->clk);

	for (i = 0; i < host->num_slots; i++) {
		struct dw_mci_slot *slot = host->slot[i];
		struct mmc_host *mmc;
		struct mmc_request *mrq;
		int present;

		if (!slot)
			continue;

		mmc = slot->mmc;
		present = dw_mci_get_cd(mmc);
		while (present != slot->last_detect_state) {
			dev_dbg(&slot->mmc->class_dev, "card %s\n",
				present ? "inserted" : "removed");

			spin_lock_bh(&host->lock);

			/* Card change detected */
			slot->last_detect_state = present;

			/* Mark card as present if applicable */
			if (present != 0)
				set_bit(DW_MMC_CARD_PRESENT, &slot->flags);
			else
				clear_bit(DW_MMC_CARD_PRESENT, &slot->flags);


			/* Clean up queue if present */
			mrq = slot->mrq;
			if (mrq && mrq != host->mrq) {
				list_del(&slot->queue_node);
				mrq->cmd->error = -ENOMEDIUM;
				if (mrq->data)
					mrq->data->error = -ENOMEDIUM;
				if (mrq->stop)
					mrq->stop->error = -ENOMEDIUM;
				slot->mrq = NULL;

				spin_unlock(&host->lock);
				mmc_request_done(slot->mmc, mrq);
				spin_lock(&host->lock);
			}

			spin_unlock_bh(&host->lock);

			present = dw_mci_get_cd(mmc);
		}

		mmc_detect_change(slot->mmc,
			msecs_to_jiffies(host->pdata->detect_delay_ms));
	}

	clk_disable(host->clk);
}

static int __init dw_mci_init_slot(struct dw_mci *host, unsigned int i,
				   struct dw_mci_board_slot *pdata)
{
	struct mmc_host *mmc;
	struct dw_mci_slot *slot;
	int ret = 0;
	char regulator_id[8];

	mmc = mmc_alloc_host(sizeof(struct dw_mci_slot), &host->pdev->dev);
	if (!mmc)
		return -ENOMEM;

	slot = mmc_priv(mmc);
	slot->id = pdata->id;
	slot->pdata = pdata;
	slot->mmc = mmc;
	slot->host = host;
	slot->power_mode = MMC_POWER_OFF;

	mmc->ops = &dw_mci_ops;
	mmc->f_min = DIV_ROUND_UP(host->bus_hz, 510);
	if (pdata->f_max)
		mmc->f_max = min(pdata->f_max, host->bus_hz);
	else
		mmc->f_max = host->bus_hz;

	if (pdata->caps)
		mmc->caps = pdata->caps;
	else
		mmc->caps = 0;

	if (host->pdata->quirks & DW_MCI_QUIRK_HIGHSPEED)
		mmc->caps |= MMC_CAP_SD_HIGHSPEED;

	mmc->pm_caps = MMC_PM_KEEP_POWER | MMC_PM_WAKE_SDIO_IRQ;

#ifdef CONFIG_MMC_DW_DSPG_IDMAC
	mmc->max_segs = host->ring_size;
	mmc->max_blk_size = 65536;
	mmc->max_blk_count = host->ring_size;
	mmc->max_seg_size = 0x1000;
	mmc->max_req_size = mmc->max_seg_size * mmc->max_blk_count;
#else
	if (host->pdata->blk_settings) {
		mmc->max_segs = host->pdata->blk_settings->max_segs;
		mmc->max_blk_size = host->pdata->blk_settings->max_blk_size;
		mmc->max_blk_count = host->pdata->blk_settings->max_blk_count;
		mmc->max_req_size = host->pdata->blk_settings->max_req_size;
		mmc->max_seg_size = host->pdata->blk_settings->max_seg_size;
	} else {
		/* Useful defaults if platform data is unset. */
		mmc->max_segs = 64;
		mmc->max_blk_size = 65536; /* BLKSIZ is 16 bits */
		mmc->max_blk_count = 512;
		mmc->max_req_size = mmc->max_blk_size * mmc->max_blk_count;
		mmc->max_seg_size = mmc->max_req_size;
	}
#endif /* CONFIG_MMC_DW_DSPG_IDMAC */

	sprintf(regulator_id, "vmmc%d", pdata->id);
	slot->vmmc = regulator_get(mmc_dev(mmc), regulator_id);
	if (IS_ERR(slot->vmmc)) {
		printk(KERN_INFO "%s: no %s regulator found\n",
		       mmc_hostname(mmc), regulator_id);
		slot->vmmc = NULL;
	}

	if (slot->vmmc)
		mmc->ocr_avail = mmc_regulator_get_ocrmask(slot->vmmc);
	else
		mmc->ocr_avail = pdata->ocr;

	if (!mmc->ocr_avail)
		mmc->ocr_avail = MMC_VDD_32_33 | MMC_VDD_33_34;

	if (gpio_is_valid(pdata->cd_gpio)) {
		ret = gpio_request(pdata->cd_gpio, "mmc_cd");
		if (ret)
			goto err_gpio_cd;
		gpio_direction_input(pdata->cd_gpio);
	}

	if (gpio_is_valid(pdata->wp_gpio)) {
		ret = gpio_request(pdata->wp_gpio, "mmc_wp");
		if (ret)
			goto err_gpio_wp;
		gpio_direction_input(pdata->wp_gpio);
	}

	if (dw_mci_get_cd(mmc))
		set_bit(DW_MMC_CARD_PRESENT, &slot->flags);
	else
		clear_bit(DW_MMC_CARD_PRESENT, &slot->flags);

	host->slot[i] = slot;
	ret = mmc_add_host(mmc);
	if (ret)
		goto err_add_host;

#if defined(CONFIG_DEBUG_FS)
	dw_mci_init_debugfs(slot);
#endif

	/* Card initially undetected */
	slot->last_detect_state = 0;

	if (gpio_is_valid(pdata->cd_gpio)) {
		slot->cd_irq = ret = gpio_to_irq(pdata->cd_gpio);
		if (ret < 0)
			goto err_request_irq;

		ret = request_irq(slot->cd_irq, dw_mci_cdetect_interrupt,
			IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,
			"dw-mci cdetect", host);
		if (ret)
			goto err_request_irq;
	} else
		slot->cd_irq = -1;

	/*
	 * Card may have been plugged in prior to boot so we
	 * need to run the detect tasklet
	 */
	queue_work(dw_mci_card_workqueue, &host->card_work);

	return 0;

err_request_irq:
	mmc_remove_host(mmc);
err_add_host:
	if (gpio_is_valid(pdata->wp_gpio))
		gpio_free(pdata->wp_gpio);
err_gpio_wp:
	if (gpio_is_valid(pdata->cd_gpio))
		gpio_free(pdata->cd_gpio);
err_gpio_cd:
	if (slot->vmmc)
		regulator_put(slot->vmmc);
	mmc_free_host(mmc);
	return ret;
}

static void dw_mci_cleanup_slot(struct dw_mci_slot *slot, unsigned int i)
{
	/* Shutdown detect IRQ */
	if (slot->cd_irq >= 0)
		free_irq(slot->cd_irq, slot->host);

	/* Debugfs stuff is cleaned up by mmc core */
	slot->host->slot[i] = NULL;
	mmc_remove_host(slot->mmc);

	if (gpio_is_valid(slot->pdata->wp_gpio))
		gpio_free(slot->pdata->wp_gpio);

	if (gpio_is_valid(slot->pdata->cd_gpio))
		gpio_free(slot->pdata->cd_gpio);

	if (slot->vmmc)
		regulator_put(slot->vmmc);

	mmc_free_host(slot->mmc);
}

static void __init dw_mci_init_dma(struct dw_mci *host)
{
	/* Alloc memory for sg translation */
	host->sg_cpu = dma_alloc_coherent(&host->pdev->dev, PAGE_SIZE,
					  &host->sg_dma, GFP_KERNEL);
	if (!host->sg_cpu) {
		dev_err(&host->pdev->dev, "%s: could not alloc DMA memory\n",
			__func__);
		goto no_dma;
	}

	/* Determine which DMA interface to use */
#ifdef CONFIG_MMC_DW_DSPG_IDMAC
	host->dma_ops = &dw_mci_idmac_ops;
	dev_info(&host->pdev->dev, "Using internal DMA controller.\n");
#endif

	if (!host->dma_ops)
		goto no_dma;

	if (host->dma_ops->init) {
		if (host->dma_ops->init(host)) {
			dev_err(&host->pdev->dev, "%s: Unable to initialize "
				"DMA Controller.\n", __func__);
			goto no_dma;
		}
	} else {
		dev_err(&host->pdev->dev, "DMA initialization not found.\n");
		goto no_dma;
	}

	host->use_dma = 1;
	return;

no_dma:
	dev_info(&host->pdev->dev, "Using PIO mode.\n");
	host->use_dma = 0;
	return;
}

#ifdef CONFIG_OF
static struct dw_mci_board * __init dw_mci_parse_dt(struct platform_device *pdev)
{
	struct dw_mci_board *pdata;
	struct device_node *np = pdev->dev.of_node, *child;
	u32 prop;
	int slots;
	struct dw_mci_board_slot *slot;

	if (!np)
		return NULL;

	pdata = kzalloc(sizeof(*pdata), GFP_KERNEL);
	if (!pdata)
		return NULL;

	pdata->quirks = DW_MCI_QUIRK_SBE_NEED_STOP | DW_MCI_QUIRK_HIGHSPEED;
	pdata->bus_hz = 50000000;
	pdata->detect_delay_ms = 200;
	if (of_property_read_u32(np, "snps,fifo-depth", &prop) == 0)
		pdata->fifo_depth = prop;

	/* count slots */
	slots = 0;
	for_each_child_of_node(np, child) {
		if (of_device_is_available(child))
			slots++;
	}

	pdata->num_slots = slots;
	pdata->slots = slot = kzalloc(sizeof(*slot) * slots, GFP_KERNEL);
	if (!slot)
		goto err_free;

	/* fill slots */
	for_each_child_of_node(np, child) {
		if (!of_device_is_available(child))
			continue;

		if (of_property_read_u32(child, "snps,slot", &prop)) {
			dev_err(&pdev->dev, "missing slot 'snps,slot' property\n");
			goto err_free_slots;
		}

		slot->id = prop;
		slot->caps = MMC_CAP_SDIO_IRQ;

		if (of_property_read_u32(child, "bus-width", &prop)) {
			dev_err(&pdev->dev, "missing 'bus-width'\n");
			goto err_free_slots;
		}

		switch (prop) {
		case 8:
			slot->caps |= MMC_CAP_8_BIT_DATA;
			/* fall through */
		case 4:
			slot->caps |= MMC_CAP_4_BIT_DATA;
			/* fall through */
		case 1:
			break;
		default:
			dev_err(&pdev->dev, "invalid 'bus-width'\n");
			goto err_free_slots;
		}

		if (of_find_property(child, "non-removable", NULL))
			slot->caps |= MMC_CAP_NONREMOVABLE;

		if (of_property_read_u32(child, "max-frequency", &prop) == 0)
			slot->f_max = prop;

		slot->cd_gpio = of_get_named_gpio(child, "cd-gpios", 0);
		if (of_find_property(child, "cd-inverted", NULL))
			slot->cd_inverted = 1;

		slot->wp_gpio = of_get_named_gpio(child, "wp-gpios", 0);
		if (of_find_property(child, "wp-inverted", NULL))
			slot->wp_inverted = 1;

		/* next slot */
		slot++;
	}

	return pdata;

err_free_slots:
	kfree(pdata->slots);
err_free:
	kfree(pdata);
	return NULL;
}

static const struct of_device_id dw_mci_of_match[] = {
	{ .compatible = "snps,dw-mmc", },
	{},
};
MODULE_DEVICE_TABLE(of, dw_mci_of_match);
#else
static struct dw_mci_board * __init dw_mci_parse_dt(struct platform_device *pdev)
{
	return NULL;
}
#endif

static int __init dw_mci_probe(struct platform_device *pdev)
{
	struct dw_mci *host;
	struct resource	*regs;
	struct dw_mci_board *pdata;
	int irq, ret, i, width, intmsk;
	u32 fifo_size;

	regs = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!regs)
		return -ENXIO;

	irq = platform_get_irq(pdev, 0);
	if (irq < 0)
		return irq;

	host = kzalloc(sizeof(struct dw_mci), GFP_KERNEL);
	if (!host)
		return -ENOMEM;

	if (pdev->dev.of_node)
		pdata = dw_mci_parse_dt(pdev);
	else
		pdata = pdev->dev.platform_data;
	if (!pdata) {
		dev_err(&pdev->dev, "platform data missing\n");
		ret = -ENODEV;
		goto err_freehost;
	}

	host->pdev = pdev;
	host->pdata = pdata;

	host->clk = clk_get(&pdev->dev, NULL);
	if (IS_ERR(host->clk)) {
		dev_err(&pdev->dev, "cannot get clock\n");
		ret = PTR_ERR(host->clk);
		goto err_freehost;
	}

	if (pdata->bus_hz) {
		long bus_hz = clk_round_rate(host->clk, pdata->bus_hz);
		if (bus_hz < 0) {
			dev_err(&pdev->dev, "cannot round clock\n");
			ret = bus_hz;
			goto err_freehost;
		}

		ret = clk_set_rate(host->clk, bus_hz);
		if (ret) {
			dev_err(&pdev->dev, "cannot set clock\n");
			goto err_freehost;
		}

		clk_enable(host->clk);
		host->bus_hz = bus_hz;
	} else {
		clk_enable(host->clk);
		host->bus_hz = clk_get_rate(host->clk);
	}

	init_timer(&host->clk_off_timer);
	host->clk_off_timer.function = dw_mci_clk_off;
	host->clk_off_timer.data = (unsigned long)host;

	host->quirks = pdata->quirks;
	host->sdio_irq_enabled = 0;

	spin_lock_init(&host->lock);
	INIT_LIST_HEAD(&host->queue);

	init_timer(&host->poll_timer);
	host->poll_timer.function = dw_mci_poll_queue;
	host->poll_timer.data = (unsigned long)host;

	ret = -ENOMEM;
	host->regs = ioremap(regs->start, regs->end - regs->start + 1);
	if (!host->regs)
		goto err_freehost;

	host->dma_ops = pdata->dma_ops;
	dw_mci_init_dma(host);

	/*
	 * Get the host data width - this assumes that HCON has been set with
	 * the correct values.
	 */
	i = (mci_readl(host, HCON) >> 7) & 0x7;
	if (!i) {
		host->push_data = dw_mci_push_data16;
		host->pull_data = dw_mci_pull_data16;
		width = 16;
		host->data_shift = 1;
	} else if (i == 2) {
		host->push_data = dw_mci_push_data64;
		host->pull_data = dw_mci_pull_data64;
		width = 64;
		host->data_shift = 3;
	} else {
		/* Check for a reserved value, and warn if it is */
		WARN((i != 1),
		     "HCON reports a reserved host data width!\n"
		     "Defaulting to 32-bit access.\n");
		host->push_data = dw_mci_push_data32;
		host->pull_data = dw_mci_pull_data32;
		width = 32;
		host->data_shift = 2;
	}

	/* Reset all blocks */
	if (!mci_wait_reset(host, SDMMC_CTRL_RESET | SDMMC_CTRL_FIFO_RESET |
			    SDMMC_CTRL_DMA_RESET)) {
		ret = -ENODEV;
		goto err_dmaunmap;
	}

	/* Clear the interrupts for the host controller */
	mci_writel(host, RINTSTS, 0xFFFFFFFF);
	mci_writel(host, INTMASK, 0); /* disable all mmc interrupt first */

	/* Put in max timeout */
	mci_writel(host, TMOUT, 0xFFFFFFFF);

	/*
	 * FIFO threshold settings  RxMark  = fifo_size / 2 - 1,
	 *                          Tx Mark = fifo_size / 2 DMA Size = 8
	 */
	if (!host->pdata->fifo_depth) {
		/*
		 * Power-on value of RX_WMark is FIFO_DEPTH-1, but this may
		 * have been overwritten by the bootloader, just like we're
		 * about to do, so if you know the value for your hardware, you
		 * should put it in the platform data.
		 */
		fifo_size = mci_readl(host, FIFOTH);
		fifo_size = 1 + ((fifo_size >> 16) & 0x7ff);
	} else {
		fifo_size = host->pdata->fifo_depth;
	}
	host->fifo_depth = fifo_size;
	host->fifoth_val = ((0x2 << 28) | ((fifo_size/2 - 1) << 16) |
			((fifo_size/2) << 0));
	mci_writel(host, FIFOTH, host->fifoth_val);

	/* disable clock to CIU */
	mci_writel(host, CLKENA, 0);
	mci_writel(host, CLKSRC, 0);

	tasklet_init(&host->tasklet, dw_mci_tasklet_func, (unsigned long)host);
	dw_mci_card_workqueue = create_singlethread_workqueue("dw-mci-card");
	if (!dw_mci_card_workqueue)
		goto err_dmaunmap;
	INIT_WORK(&host->card_work, dw_mci_work_routine_card);

	ret = request_irq(irq, dw_mci_interrupt, 0, "dw-mci", host);
	if (ret)
		goto err_workqueue;
	host->irq = irq;

	device_init_wakeup(&pdev->dev, 1);

	platform_set_drvdata(pdev, host);

	if (host->pdata->num_slots)
		host->num_slots = host->pdata->num_slots;
	else
		host->num_slots = ((mci_readl(host, HCON) >> 1) & 0x1F) + 1;

	/*
	 * Enable interrupts for command done, data over, data empty, card det,
	 * receive ready and error such as transmit, receive timeout, crc error
	 */
	mci_writel(host, RINTSTS, 0xFFFFFFFF);

	intmsk = SDMMC_INT_CMD_DONE | SDMMC_INT_DATA_OVER | DW_MCI_ERROR_FLAGS;

	/* If no custom card detect is used enable the built-in interrupt */
	for (i = 0; i < host->num_slots; i++)
		if (!gpio_is_valid(pdata->slots[i].cd_gpio) &&
		    !(pdata->slots[i].caps & MMC_CAP_NONREMOVABLE))
			intmsk |= SDMMC_INT_CD;

	host->intmsk = intmsk;
	mci_writel(host, INTMASK, intmsk);
	mci_writel(host, CTRL, SDMMC_CTRL_INT_ENABLE); /* Enable mci interrupt */

	/* We need at least one slot to succeed */
	for (i = 0; i < host->num_slots; i++) {
		ret = dw_mci_init_slot(host, i, pdata->slots + i);
		if (ret)
			goto err_init_slot;
	}

	dev_info(&pdev->dev, "DW MMC controller at irq %d, "
		 "%d bit host data width, "
		 "%u deep fifo\n",
		 irq, width, fifo_size);
	if (host->quirks & DW_MCI_QUIRK_IDMAC_DTO)
		dev_info(&pdev->dev, "Internal DMAC interrupt fix enabled.\n");

	/*
	 * If the internal card detect is not used we can disable the clock
	 * here
	 */
	if (!(intmsk & SDMMC_INT_CD))
		clk_disable(host->clk);

	return 0;

err_init_slot:
	/* De-init any initialized slots */
	while (i > 0) {
		if (host->slot[i])
			dw_mci_cleanup_slot(host->slot[i], i);
		i--;
	}
	free_irq(irq, host);

err_workqueue:
	destroy_workqueue(dw_mci_card_workqueue);

err_dmaunmap:
	if (host->use_dma && host->dma_ops->exit)
		host->dma_ops->exit(host);
	dma_free_coherent(&host->pdev->dev, PAGE_SIZE,
			  host->sg_cpu, host->sg_dma);
	iounmap(host->regs);

err_freehost:
	if (pdata && pdev->dev.of_node) {
		kfree(pdata->slots);
		kfree(pdata);
	}
	kfree(host);
	return ret;
}

static int __exit dw_mci_remove(struct platform_device *pdev)
{
	struct dw_mci *host = platform_get_drvdata(pdev);
	int i;

	mci_writel(host, RINTSTS, 0xFFFFFFFF);
	mci_writel(host, INTMASK, 0); /* disable all mmc interrupt first */

	platform_set_drvdata(pdev, NULL);

	for (i = 0; i < host->num_slots; i++) {
		dev_dbg(&pdev->dev, "remove slot %d\n", i);
		if (host->slot[i])
			dw_mci_cleanup_slot(host->slot[i], i);
	}

	/* disable clock to CIU */
	mci_writel(host, CLKENA, 0);
	mci_writel(host, CLKSRC, 0);

	free_irq(platform_get_irq(pdev, 0), host);
	destroy_workqueue(dw_mci_card_workqueue);
	dma_free_coherent(&pdev->dev, PAGE_SIZE, host->sg_cpu, host->sg_dma);

	if (host->use_dma && host->dma_ops->exit)
		host->dma_ops->exit(host);

	iounmap(host->regs);

	if (pdev->dev.of_node) {
		kfree(host->pdata->slots);
		kfree(host->pdata);
	}
	kfree(host);
	return 0;
}

#ifdef CONFIG_PM
static int dw_mci_suspend(struct platform_device *pdev, pm_message_t mesg)
{
	int i, ret;
	struct dw_mci *host = platform_get_drvdata(pdev);

	for (i = 0; i < host->num_slots; i++) {
		struct dw_mci_slot *slot = host->slot[i];
		if (!slot)
			continue;
		ret = mmc_suspend_host(slot->mmc);
		if (ret < 0) {
			while (--i >= 0) {
				slot = host->slot[i];
				if (slot)
					mmc_resume_host(host->slot[i]->mmc);
			}
			return ret;
		}
	}

	/* Cancel the clk_off timer, and if it was running, disable clock  */
	if (del_timer(&host->clk_off_timer))
		dw_mci_clk_off((unsigned long)host);

	/*
	 * If the internal card detect is used we have to disable the clock
	 * here since it was not done in _probe()
	 */
	if (host->intmsk & SDMMC_INT_CD)
		clk_disable(host->clk);

	if (device_may_wakeup(&pdev->dev))
		enable_irq_wake(host->irq);

	return 0;
}

static int dw_mci_resume(struct platform_device *pdev)
{
	int i, ret;
	struct dw_mci *host = platform_get_drvdata(pdev);

	if (device_may_wakeup(&pdev->dev))
		disable_irq_wake(host->irq);

	if (host->clk)
		clk_enable(host->clk);

#if 0
	if (host->dma_ops->init)
		host->dma_ops->init(host);

	if (!mci_wait_reset(host, SDMMC_CTRL_RESET | SDMMC_CTRL_FIFO_RESET | SDMMC_CTRL_DMA_RESET)) {
		ret = -ENODEV;
		return ret;
	}

	/* Restore the old value at FIFOTH register */
	mci_writel(host, FIFOTH, host->fifoth_val);

	mci_writel(host, RINTSTS, 0xFFFFFFFF);
	mci_writel(host, INTMASK, host->intmsk);
	mci_writel(host, CTRL, SDMMC_CTRL_INT_ENABLE);
#endif

	/*
	 * If the internal card detect is not used we can disable the clock
	 * here
	 */
	if (!(host->intmsk & SDMMC_INT_CD))
		clk_disable(host->clk);

	for (i = 0; i < host->num_slots; i++) {
		struct dw_mci_slot *slot = host->slot[i];
		if (!slot)
			continue;
		ret = mmc_resume_host(host->slot[i]->mmc);
		if (ret < 0)
			return ret;
	}

	return 0;
}

static void
dw_mci_shutdown(struct platform_device *pdev)
{
	pm_message_t mesg;

	mesg.event = PM_EVENT_SUSPEND;
	(void)dw_mci_suspend(pdev, mesg);
}
#else
#define dw_mci_suspend	NULL
#define dw_mci_resume	NULL
#define dw_mci_shutdown	NULL
#endif /* CONFIG_PM */

static struct platform_driver dw_mci_driver = {
	.remove		= __exit_p(dw_mci_remove),
	.suspend	= dw_mci_suspend,
	.resume		= dw_mci_resume,
	.shutdown	= dw_mci_shutdown,
	.driver		= {
		.name		= "dw_mmc_dspg",
		.of_match_table	= of_match_ptr(dw_mci_of_match),
	},
};

static int __init dw_mci_init(void)
{
	return platform_driver_probe(&dw_mci_driver, dw_mci_probe);
}

static void __exit dw_mci_exit(void)
{
	platform_driver_unregister(&dw_mci_driver);
}

module_init(dw_mci_init);
module_exit(dw_mci_exit);

MODULE_DESCRIPTION("DW Multimedia Card Interface driver");
MODULE_AUTHOR("NXP Semiconductor VietNam");
MODULE_AUTHOR("Imagination Technologies Ltd");
MODULE_LICENSE("GPL v2");
