/*
 * linux/drivers/staging/dspg/ocf-dspg-crypto.c - OCF fronted for DSPG crypto
 *
 *  Copyright (C) 2013 DSP Group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/crypto.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/platform_device.h>
#include <linux/of.h>
#include <linux/mfd/core.h>
#include <linux/mfd/dspg-crypto.h>
#include <cryptodev.h>
#include <uio.h>

#define DSPG_OCF_CRYPTO "dspg-ocf-crypto"

#define SW_TYPE_HMAC		0x1
#define SW_TYPE_HASH		0x2
#define SW_TYPE_BLKCIPHER	0x4

struct dspg_cr_data {
	int dspg_type;
	int dspg_alg; /* descriptors */
	int cri_alg;
	union {
		struct {
			char *dspg_mkey;
			int dspg_mlen;
			int dspg_olen;
		} hmac;
	} u;
	struct dspg_cr_data *dspg_next;
	struct crypto_op *operation;
};
static struct dspg_cr_data **dspg_sessions = NULL;

static DECLARE_BITMAP(channels, 32);

static int dspg_id;
static unsigned int dspg_sesnum = 0;

static DEFINE_SPINLOCK(ocf_dspg_crypto_lock);

static struct {
	softc_device_decl sc_dev;
	struct dspg_crypto_dev *dev;
	struct platform_device *pdev;
} dspg_dev;

static void
ocf_dspg_crypto_done(int status, void *context)
{
	struct cryptodesc *crda = NULL;
	struct cryptodesc *crde = NULL;
	struct cryptop *crp = NULL;
	struct dspg_cr_data *sw;
	struct crypto_op *op = (struct crypto_op*)context;
	unsigned long flags;

	u8* result_buf = (u8*)(op->data_out_virt);

	crp = op->dspg_crp;

	spin_lock_irqsave(&ocf_dspg_crypto_lock, flags);
	sw = dspg_sessions[crp->crp_sid];

	if ((sw->dspg_type == SW_TYPE_HMAC) || (sw->dspg_type == SW_TYPE_HASH)) {
		crda = crp->crp_desc;
		crypto_copyback(CRYPTO_F_IOV, crp->crp_buf, crda->crd_len, crp->crp_olen, result_buf);
	} else {
		if (sw->dspg_type == SW_TYPE_BLKCIPHER) {
			crde = crp->crp_desc;
			crypto_copyback(CRYPTO_F_IOV, crp->crp_buf, 0, crp->crp_olen, result_buf);
		} else {
			printk("algorithm type is not supported.\n");
		}
	}

	clear_bit_unlock(op->channel, channels);
	spin_unlock_irqrestore(&ocf_dspg_crypto_lock, flags);

	kfree(op->data_in_virt);
	kfree(op->data_out_virt);
	crypto_done(op->dspg_crp);
}

static int
ocf_dspg_crypto_freesession(device_t arg, unsigned long long tid)
{
	struct dspg_cr_data *swd;
	unsigned int sid = CRYPTO_SESID2LID(tid);
	unsigned long flags;

	if (sid > dspg_sesnum || dspg_sessions == NULL ||
	    dspg_sessions[sid] == NULL)
		return EINVAL;

	/* Silently accept and return */
	if (sid == 0)
		return 0;

	spin_lock_irqsave(&ocf_dspg_crypto_lock, flags);

	while ((swd = dspg_sessions[sid]) != NULL) {
		dspg_sessions[sid] = swd->dspg_next;

		if (swd->u.hmac.dspg_mkey != NULL) {
			kfree(swd->u.hmac.dspg_mkey);
			swd->u.hmac.dspg_mkey = NULL;
		}
		if (swd->operation != NULL)
			dspg_crypto_free_operation(swd->operation);

		kfree(swd);
		swd = NULL;
	}

	spin_unlock_irqrestore(&ocf_dspg_crypto_lock, flags);

	return 0;
}

static int
ocf_dspg_crypto_newsession(device_t dev, unsigned int *sid, struct cryptoini *cri)
{
	/* struct dspg_dev *dspg_dev = device_get_softc(dev); */
	struct dspg_cr_data **swd;
	unsigned int i;
	char *algo;
	unsigned long flags;

	if (sid == NULL || cri == NULL)
		return EINVAL;

	spin_lock_irqsave(&ocf_dspg_crypto_lock, flags);

	if (dspg_sessions) {
		for (i = 1; i < dspg_sesnum; i++)
			if (dspg_sessions[i] == NULL)
				break;
	} else {
		i = 1; /* silence compiler warning */
	}

	if (dspg_sessions == NULL || i == dspg_sesnum) {
		if (dspg_sessions == NULL) {
			i = 1; /* We leave dspg_sessions[0] empty */
			dspg_sesnum = CRYPTO_SW_SESSIONS;
		} else
			dspg_sesnum *= 2;

		swd = kmalloc(dspg_sesnum * sizeof(struct dspg_data *),
			      GFP_ATOMIC);
		
		if (swd == NULL) {
			/* Reset session number */
			if (dspg_sesnum == CRYPTO_SW_SESSIONS)
				dspg_sesnum = 0;
			else
				dspg_sesnum /= 2;
			spin_unlock_irqrestore(&ocf_dspg_crypto_lock, flags);
			return ENOBUFS;
		}
		memset(swd, 0, dspg_sesnum * sizeof(struct dspg_data *));

		/* Copy existing sessions */
		if (dspg_sessions) {
			memcpy(swd, dspg_sessions,
			       (dspg_sesnum / 2) * sizeof(struct dspg_data *));

			kfree(dspg_sessions);
		}

		dspg_sessions = swd;
	}

	swd = &dspg_sessions[i];
	*sid = i;

	spin_unlock_irqrestore(&ocf_dspg_crypto_lock, flags);

	while (cri) {
		*swd = (struct dspg_cr_data *)
		       kmalloc(sizeof(struct dspg_cr_data), GFP_ATOMIC);

		if (*swd == NULL) {
			ocf_dspg_crypto_freesession(NULL, i);
			return ENOBUFS;
		}
		memset(*swd, 0, sizeof(struct dspg_cr_data));

		algo = NULL;

		switch (cri->cri_alg) {
		case CRYPTO_MD5:
			algo = "md5";
			(*swd)->dspg_alg = DESC_CTL_MODE_MD5;
			(*swd)->dspg_type = SW_TYPE_HASH;

			(*swd)->u.hmac.dspg_olen = 16;
			break;

		case CRYPTO_SHA1:
			algo = "sha160";
			(*swd)->dspg_alg = DESC_CTL_MODE_SHA1_160;
			(*swd)->dspg_type = SW_TYPE_HASH;

			(*swd)->u.hmac.dspg_olen = 20;
			break;

		case CRYPTO_SHA2_256:
			algo = "sha2-256";
			(*swd)->dspg_alg = DESC_CTL_MODE_SHA2_256;
			(*swd)->dspg_type = SW_TYPE_HASH;

			(*swd)->u.hmac.dspg_olen = 32;
			break;

		case CRYPTO_MD5_HMAC:
			algo = "md5-hmac";
			(*swd)->dspg_alg = DESC_CTL_MODE_MD5_HMAC;
			(*swd)->dspg_type = SW_TYPE_HMAC;

			(*swd)->u.hmac.dspg_olen = 16;
			break;

		case CRYPTO_SHA1_HMAC:
			algo = "SHA1-hmac";
			(*swd)->dspg_alg = DESC_CTL_MODE_SHA1_160_HMAC;
			(*swd)->dspg_type = SW_TYPE_HMAC;

			(*swd)->u.hmac.dspg_olen = 20;
			break;

		case CRYPTO_SHA2_256_HMAC:
			algo = "SHA256-hmac";
			(*swd)->dspg_alg = DESC_CTL_MODE_SHA2_256_HMAC;
			(*swd)->dspg_type = SW_TYPE_HMAC;

			(*swd)->u.hmac.dspg_olen = 32;
			break;

		case CRYPTO_AES_CBC:
			algo = "aes-cbc";
			(*swd)->dspg_alg = DESC_CTL_MODE_AES_CBC_ENCRYPT;
			(*swd)->dspg_type = SW_TYPE_BLKCIPHER;
			break;

		case CRYPTO_AES_CTR:
			algo = "aes-ctr";
			(*swd)->dspg_alg = DESC_CTL_MODE_AES_CTR_ENCRYPT;
			(*swd)->dspg_type = SW_TYPE_BLKCIPHER;
			break;

		case CRYPTO_DES_CBC:
			algo = "des-cbc";
			(*swd)->dspg_alg = DESC_CTL_MODE_DES_CBC_ENCRYPT;
			(*swd)->dspg_type = SW_TYPE_BLKCIPHER;
			break;

		case CRYPTO_3DES_CBC:
			algo = "3des-cbc";
			(*swd)->dspg_alg = DESC_CTL_MODE_3DES_CBC_ENCRYPT;
			(*swd)->dspg_type = SW_TYPE_BLKCIPHER;
			break;

		default:
			printk("unsupported crypto algorithm: %d\n",
			       cri->cri_alg);
			ocf_dspg_crypto_freesession(NULL, *sid);
			return -EINVAL;
		}

		if (((*swd)->dspg_type == SW_TYPE_HMAC) ||
		    ((*swd)->dspg_type == SW_TYPE_BLKCIPHER))
			(*swd)->u.hmac.dspg_mlen = (cri->cri_klen + 7) / 8;

		if ((*swd)->u.hmac.dspg_mlen > 0) {
			(*swd)->u.hmac.dspg_mkey =
				(char *)kmalloc((*swd)->u.hmac.dspg_mlen,
						SLAB_ATOMIC);
			if ((*swd)->u.hmac.dspg_mkey == NULL) {
				ocf_dspg_crypto_freesession(NULL, i);
				return ENOBUFS;
			}
			memcpy((*swd)->u.hmac.dspg_mkey, cri->cri_key,
			       (*swd)->u.hmac.dspg_mlen);
		}
		(*swd)->cri_alg = cri->cri_alg;
		cri = cri->cri_next;
		swd = &((*swd)->dspg_next);
	}

	return 0;
}

static int
ocf_dspg_crypto_process(device_t arg, struct cryptop *crp, int hint)
{
	int ret, input_len = 0;
	unsigned int lid;
	struct cryptodesc *crda = NULL;
	struct cryptodesc *crde = NULL;
	struct dspg_cr_data *sw;
	u8* op_in_buf;
	u8* op_out_buf;
	u8* op_iv_buf = NULL;
	int channel = 0;
	unsigned long flags;

	/* Sanity check */
	if (crp == NULL)
		return EINVAL;

	crp->crp_etype = 0;

	if (crp->crp_desc == NULL || crp->crp_buf == NULL) {
		crp->crp_etype = EINVAL;
		goto done;
	}

	/* Find the session we are using */
	lid = crp->crp_sid & 0xffffffff;
	if (lid >= dspg_sesnum || lid == 0) {
		crp->crp_etype = ENOENT;
		goto done;
	}

	sw = dspg_sessions[lid];

	if ((sw->dspg_type == SW_TYPE_HMAC) || (sw->dspg_type == SW_TYPE_HASH)) {
		crda = crp->crp_desc;
		input_len = crda->crd_len;
		crp->crp_olen = sw->u.hmac.dspg_olen;
	} else {
		if (sw->dspg_type == SW_TYPE_BLKCIPHER) {
			crde = crp->crp_desc;
			/* input_len = crp->crp_ilen; */
			input_len = crde->crd_len;
			/* crp->crp_olen = crp->crp_ilen; */
			crp->crp_olen = crde->crd_len; /* incorrect should be fixed */
			/* crp->crp_olen = sw->u.hmac.dspg_olen; */

			if (crde->crd_iv) {
				op_iv_buf = kzalloc(16, GFP_KERNEL);
				memcpy(op_iv_buf, crde->crd_iv, 16);
				/* crypto_copydata(CRYPTO_F_IOV, crde->crd_iv, 0, 16, op_iv_buf); */
			}
		} else {
			printk("Unsupported algorithm type\n");
		}
	}
	/* crypto operator in/out memory must be contiguous! */
	op_in_buf = kzalloc(input_len, GFP_KERNEL);
	op_out_buf = kzalloc(crp->crp_olen, GFP_KERNEL);

	crypto_copydata(CRYPTO_F_IOV, crp->crp_buf, 0, input_len, op_in_buf);

	spin_lock_irqsave(&ocf_dspg_crypto_lock, flags);
	channel = find_first_zero_bit(channels, 32);
	if (channel >= 32) {
		crp->crp_etype = ENODEV;
		goto done;
	}
	test_and_set_bit_lock(channel, channels);
	spin_unlock_irqrestore(&ocf_dspg_crypto_lock, flags);

	if ((sw->dspg_type == SW_TYPE_HMAC) ||
	    (sw->dspg_type == SW_TYPE_HASH) ||
	    (sw->dspg_type == SW_TYPE_BLKCIPHER)) {
		sw->operation = dspg_crypto_alloc_operation(op_in_buf,
			input_len,
			op_out_buf,
			crp->crp_olen,
			sw->dspg_alg,
			channel,
			sw->u.hmac.dspg_mkey,
			sw->u.hmac.dspg_mlen,
			op_iv_buf, NULL, 0, 0);

		if (op_iv_buf != NULL) {
			kfree(op_iv_buf);
			op_iv_buf = NULL;
		}
	} else {
		printk("No alloc operation\n");
	}

	if (sw->operation) {
		sw->operation->dspg_crp = crp;
		sw->operation->done = ocf_dspg_crypto_done;
		sw->operation->context = sw->operation;
		spin_lock_irqsave(&ocf_dspg_crypto_lock, flags);
		ret = dspg_crypto_submit_operation(sw->operation);
		spin_unlock_irqrestore(&ocf_dspg_crypto_lock, flags);
		sw = sw->dspg_next;
	}

	return 0;

done:
	crypto_done(crp);

	return 0;
}

static device_method_t dspg_crypto_methods = {
	DEVMETHOD(cryptodev_newsession,  ocf_dspg_crypto_newsession),
	DEVMETHOD(cryptodev_freesession, ocf_dspg_crypto_freesession),
	DEVMETHOD(cryptodev_process,     ocf_dspg_crypto_process),
};

#ifdef CONFIG_OCF_RANDOMHARVEST
/*
 * We run the random number generator output through SHA so that it
 * is FIPS compliant.
 * Without runing SHA, rndtest will be failed sometimes
 */
static volatile int sha_done = 0;

static unsigned int sha_digest[8];
static unsigned int rnd_buff[8];

static void
ocf_dspg_crypto_hash_cb(int status, void *context)
{
	/* struct crypto_op *op = (struct crypto_op*)context;
	u8* result_buf = (u8*)(op->data_out_virt); */

	sha_done = 1;
}

static int
ocf_dspg_crypto_read_random(void *arg, unsigned int *buf, int maxwords)
{
	int num, ret;
	struct crypto_op *op = 0;
	int channel;
	unsigned long flags;

	spin_lock_irqsave(&ocf_dspg_crypto_lock, flags);
	channel = find_first_zero_bit(channels, 32);
	if (channel >= 32) {
		printk("No free channel\n");
		return 0;
	}
	test_and_set_bit_lock(channel, channels);
	spin_unlock_irqrestore(&ocf_dspg_crypto_lock, flags);

	memset(buf, 0, maxwords * sizeof(*buf));
	dspg_crypto_generate_trng();
	num = dspg_crypto_get_trng(rnd_buff);
	if (!num) {
		printk("get_trng failed\n");
		return 0;
	}

	sha_done = 0;
	op = dspg_crypto_alloc_operation(rnd_buff,
			sizeof(rnd_buff),
			sha_digest,
			sizeof(sha_digest),
			DESC_CTL_MODE_SHA2_256,
			channel,
			0, 0, 0, NULL, 0, 0);

	op->done = ocf_dspg_crypto_hash_cb;
	op->context = op;
	spin_lock_irqsave(&ocf_dspg_crypto_lock, flags);
	ret = dspg_crypto_submit_operation(op);
	spin_unlock_irqrestore(&ocf_dspg_crypto_lock, flags);

	while (!sha_done)
		schedule();
	if (sha_done < 0) {
		printk("hash failed: %d\n", -sha_done);
		return 0;
	}

	if (maxwords < NUM_OF_TRNG)
		memcpy(buf, sha_digest, (maxwords*4));
	else
		memcpy(buf, sha_digest, sizeof(sha_digest));

	spin_lock_irqsave(&ocf_dspg_crypto_lock, flags);
	clear_bit_unlock(op->channel, channels);
	dspg_crypto_free_operation(op);
	spin_unlock_irqrestore(&ocf_dspg_crypto_lock, flags);

	if (maxwords < NUM_OF_TRNG)
		return maxwords;
	else
		return num;
}
#endif /* CONFIG_OCF_RANDOMHARVEST */

static int
__init ocf_dspg_crypto_probe(struct platform_device *pdev)
{
	struct dspg_crypto_dev *dev = dev_get_drvdata(pdev->dev.parent);

	spin_lock_init(&ocf_dspg_crypto_lock);

	memset(&dspg_dev, 0, sizeof(dspg_dev));
	dspg_dev.dev = dev;
	dspg_dev.pdev = pdev;
	softc_device_init(&dspg_dev, DSPG_OCF_CRYPTO, 0, dspg_crypto_methods);

	dspg_id = crypto_get_driverid(softc_get_device(&dspg_dev),
				      CRYPTOCAP_F_HARDWARE);
	if (dspg_id < 0) {
		dev_err(&pdev->dev, "failed to initialize");
		return -ENODEV;
	}

	crypto_register(dspg_id, CRYPTO_DES_CBC, 0, 0);
	crypto_register(dspg_id, CRYPTO_3DES_CBC, 0, 0);
	crypto_register(dspg_id, CRYPTO_RIJNDAEL128_CBC, 0, 0);
	crypto_register(dspg_id, CRYPTO_MD5, 0, 0);
	crypto_register(dspg_id, CRYPTO_SHA1, 0, 0);
	crypto_register(dspg_id, CRYPTO_MD5_HMAC, 0, 0);
	crypto_register(dspg_id, CRYPTO_SHA1_HMAC, 0, 0);
	crypto_register(dspg_id, CRYPTO_SHA2_256, 0, 0),
	crypto_register(dspg_id, CRYPTO_SHA2_256_HMAC, 0, 0);
	crypto_register(dspg_id, CRYPTO_AES_CBC, 0, 0);
	crypto_register(dspg_id, CRYPTO_AES_CTR, 0, 0);
#ifdef CONFIG_OCF_RANDOMHARVEST
	crypto_rregister(dspg_id, ocf_dspg_crypto_read_random, NULL);
#endif

	dev_info(&pdev->dev, "OCF frontend for DSPG cryptographic security module\n");

	return 0;
}

static int
__exit ocf_dspg_crypto_remove(struct platform_device *pdev)
{
	crypto_unregister_all(dspg_id);

	return 0;
}

static const struct of_device_id ocf_dspg_crypto_of_ids[] = {
	{ .compatible = "dspg,ocf-crypto" },
	{ },
};

MODULE_DEVICE_TABLE(of, ocf_dspg_crypto_of_ids);

static struct platform_driver ocf_dspg_crypto_driver = {
	.driver = {
		.name = DSPG_OCF_CRYPTO,
		.owner = THIS_MODULE,
		.of_match_table = ocf_dspg_crypto_of_ids,
	},
	.remove = __exit_p(ocf_dspg_crypto_remove),
};

static int __init ocf_dspg_crypto_init(void)
{
	return platform_driver_probe(&ocf_dspg_crypto_driver,
				     ocf_dspg_crypto_probe);
}
module_init(ocf_dspg_crypto_init);

static void __exit ocf_dspg_crypto_exit(void)
{
	platform_driver_unregister(&ocf_dspg_crypto_driver);
}
module_exit(ocf_dspg_crypto_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("DSPG");
MODULE_DESCRIPTION("ocf_dspg_crypto - OCF fronted for DVF99 cryptographic security module");
