#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/mm.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/dma-mapping.h>
#include <linux/interrupt.h>
#include <linux/clk.h>
#include <linux/delay.h>
#include <linux/mfd/core.h>
#include <linux/of.h>
#include <linux/of_device.h>
#include <linux/of_gpio.h>
#include <asm/uaccess.h>
#include <video/dspgfb.h>

MODULE_DESCRIPTION("DSPG LCD Controller framebuffer driver");
MODULE_LICENSE("GPL");

static unsigned underruns = 0;
module_param(underruns, uint, 0644);
MODULE_PARM_DESC(underruns, "Number of underruns that happened.");

static unsigned unhandled_ints = 0;
module_param(unhandled_ints, uint, 0644);
MODULE_PARM_DESC(unhandled_ints, "Number of unhandled interrupts.");

#define DSPGFB "dspgfb"

#define INT_FRAME_START	1
#define INT_FRAME_DONE	2
#define INT_UNDERRUN	4

struct dspgfb {
	void __iomem *regs;
	struct fb_info *info;
	struct fb_videomode mode;
	int irq;
	struct clk *clk;

	void __iomem *vidmem;

	struct dspgfb_pdata pdata;
	bool					pending;
	wait_queue_head_t		wq;
};

unsigned long
dspgfb_readl(struct dspgfb *dspgfb, unsigned long addr)
{
	return readl(dspgfb->regs + addr);
}
EXPORT_SYMBOL(dspgfb_readl);

void
dspgfb_writel(struct dspgfb *dspgfb, unsigned long addr, unsigned long val)
{
	writel(val, dspgfb->regs + addr);
}
EXPORT_SYMBOL(dspgfb_writel);

static void
dspgfb_hw_enable(struct dspgfb *dspgfb)
{
	dspgfb_writel(dspgfb, LCDC_REG_LCDCCR, 1);
}

static void
dspgfb_hw_disable(struct dspgfb *dspgfb)
{
	dspgfb_writel(dspgfb, LCDC_REG_LCDCCR, 0);
}

/* If the panel supports only 16bits, the framebuffer will hold rgb565.
 * If the panel supports rgb666 or rgb888, the frame buffer will hold rgb888
 * (in case of panel which supports rgb666, the LCDC will drop the 2
 * least significat bits of each color component))
 */
static void
dspg_bpp_to_info(struct fb_var_screeninfo *var, int bits_per_pixel)
{
	if (bits_per_pixel == 16) {
		var->bits_per_pixel = 16;
		var->red.offset = 11;
		var->red.length = 5;
		var->green.offset = 5;
		var->green.length = 6;
		var->blue.offset = 0;
		var->blue.length = 5;
	} else if ((bits_per_pixel == 18) || (bits_per_pixel == 24)) {
		var->bits_per_pixel = 24;
		var->red.offset = 16;
		var->red.length = 8;
		var->green.offset = 8;
		var->green.length = 8;
		var->blue.offset = 0;
		var->blue.length = 8;
	}
}

static void
dspgfb_hw_param_update(struct dspgfb *dspgfb)
{
	dspgfb_writel(dspgfb, LCDC_REG_PARUP, 1);
	if (dspgfb->pdata.cputype)
		dspgfb_writel(dspgfb, LCDC_REG_CDISPUPR, 1);
}

static int
dspgfb_set_display_bpp(struct dspgfb *dspgfb, struct fb_info *info)
{
	unsigned int bpp = 0;

	switch (info->var.bits_per_pixel) {
	case 16:
		bpp = 0;
		break;
	case 18:
	case 24:
		bpp = 1;
		break;
	case 2:
		bpp = 3;
		break;
	default:
		return -EINVAL;
	}

	dspgfb_writel(dspgfb, LCDC_REG_INTMR, bpp);

	/* input data transfer sizes */
	dspgfb_writel(dspgfb, LCDC_REG_INDTR, ((info->var.xres * info->var.bits_per_pixel)/32) - 1);

	dspgfb_writel(dspgfb, LCDC_REG_OFFAR0,  info->var.xres * info->var.bits_per_pixel / 8);
	dspgfb_writel(dspgfb, LCDC_REG_OFFAR1,  info->var.xres * info->var.bits_per_pixel / 8);

	return 0;
}

static void
dspgfb_set_hw_init(struct dspgfb *dspgfb)
{
	int dispir = 0, pancsr;

	if (dspgfb->pdata.cputype)
		dspgfb_writel(dspgfb, LCDC_REG_LCDCCR, 2); /* FIF0 Reset */

	/* clear all pending interrupts (may exist from u-boot) */
	dspgfb_writel(dspgfb, LCDC_REG_INTSR, dspgfb_readl(dspgfb, LCDC_REG_INTSR));

	if (dspgfb->pdata.cputype)
		dspgfb_writel(dspgfb, LCDC_REG_INTER, INT_FRAME_DONE);
	else
		dspgfb_writel(dspgfb, LCDC_REG_INTER, INT_FRAME_START | INT_UNDERRUN);

	if (dspgfb->pdata.cputype)
		dispir = 4;
	else if (dspgfb->pdata.bits_per_pixel == 24)
		dispir = 1;
	else
		dispir = 0;

	dspgfb_writel(dspgfb, LCDC_REG_DISPIR, dispir);

	pancsr = 0x6 | (dspgfb->pdata.order << 4);
	if (dspgfb->pdata.mode.sync & FB_SYNC_VERT_HIGH_ACT)
		pancsr &= ~0x2;
	if (dspgfb->pdata.mode.sync & FB_SYNC_HOR_HIGH_ACT)
		pancsr &= ~0x4;

	dspgfb_writel(dspgfb, LCDC_REG_PANCSR, pancsr);
}

static void
dspgfb_set_display_background(struct dspgfb *dspgfb, unsigned short background_color)
{
	dspgfb_writel(dspgfb, LCDC_REG_BACKCPR, background_color);
}

static int
dspgfb_get_bpp(struct dspgfb *dspgfb)
{
	int bpp = 0;

	switch (dspgfb_readl(dspgfb, LCDC_REG_INTMR)) {
		case 0:
			bpp = 16;
			break;
		case 1:
			bpp = 24;
			break;
		/* YUV422 and 2bpp are not supported */
	}

	return bpp;
}

static void
dspgfb_convert_16bpp_to_24bpp(unsigned char *dest, unsigned short *src, int num_pixels)
{
	int r;
	int g;
	int b;

	while (num_pixels--) {
		b = *src & 0x001f;
		g = (*src & 0x07e0) >> 5;
		r = (*src & 0xf800) >> 11;

		dest[0] = b << 3;
		dest[1] = g << 2;
		dest[2] = r << 3;

		src++;
		dest += 3;
	}
}

static void
dspgfb_update_refresh_address(struct dspgfb *dspgfb, int index, dma_addr_t addr, int start_refresh_new_addr)
{
	struct fb_info *info = dspgfb->info;

	if (index == 0) {
		dspgfb_writel(dspgfb, LCDC_REG_MSBAHBA0R, addr >> 16);
		dspgfb_writel(dspgfb, LCDC_REG_LSBAHBA0R, addr & 0xffff);
		dspgfb_writel(dspgfb, LCDC_REG_OFFAR0,    info->fix.line_length);
	} else {
		dspgfb_writel(dspgfb, LCDC_REG_MSBAHBA1R, addr >> 16);
		dspgfb_writel(dspgfb, LCDC_REG_LSBAHBA1R, addr & 0xffff);
		dspgfb_writel(dspgfb, LCDC_REG_OFFAR1,    info->fix.line_length);
	}

	if (start_refresh_new_addr) {
		dspgfb_writel(dspgfb, LCDC_REG_DISPIDXR, index);
		dspgfb_hw_param_update(dspgfb);
	}
}

static int
dspgfb_hw_try_resume(struct dspgfb *dspgfb)
{
	struct fb_info *info = dspgfb->info;
	struct fb_videomode *mode = &dspgfb->pdata.mode;
	dma_addr_t addr1 = info->fix.smem_start + (info->var.yres * info->fix.line_length);
	dma_addr_t old;
	int running = dspgfb_readl(dspgfb, LCDC_REG_LCDCCR) & 1;
	void *old_virt;
	int old_bpp = dspgfb_get_bpp(dspgfb);
	int new_bpp = info->var.bits_per_pixel == 16 ? 16 : 24;
	int xres, yres;

	if (!running)
		return 0;

	/* same screen resolution? */
	xres = dspgfb_readl(dspgfb, LCDC_REG_HADSTR) + 1;
	yres = dspgfb_readl(dspgfb, LCDC_REG_VATR) + 1;
	if ((xres != mode->xres) || (yres != mode->yres))
		return 0;

	/* copy over the contents of the framebuffer to the new location */
	old = dspgfb_readl(dspgfb, LCDC_REG_LSBAHBA0R) |
	      dspgfb_readl(dspgfb, LCDC_REG_MSBAHBA0R) << 16;
	old_virt = ioremap(old, info->screen_size/2);
	if (!old_virt)
		return 0;

	if (old_bpp == new_bpp) {
		/* simple copy. u-boot has only one buffer, so we copy it to both our buffers */
		memcpy(info->screen_base, old_virt, info->screen_size/2);
		memcpy(info->screen_base + info->screen_size/2, old_virt, info->screen_size/2);
	} else if (new_bpp == 24 && old_bpp == 16) {
		/* convert the image while copying */
		dspgfb_convert_16bpp_to_24bpp(info->screen_base, old_virt, info->var.xres * info->var.yres);
		dspgfb_convert_16bpp_to_24bpp(info->screen_base + info->screen_size/2,
					      old_virt, mode->xres * mode->yres);
	} else {
		memset(info->screen_base, 0, info->screen_size);
	}

	iounmap(old_virt);

	/*
	 * The following code is ugly, but for a reason :-)
	 * The PARUP mechanism is not working on OFFARX, MSBAHBAXR and LSBAHBXR
	 * registers!
	 * Thus, writing to any of these registers will cause a flickering on
	 * the screen (while refreshing from address X).
	 * Solution:
	 *	If we are refreshing from address 0:
	 *		Update address1, and switch to it
	 *		Then update address0, and switch back to it
	 *	If we are refreshing from address 1:
	 *		Update address0, and switch to it
	 *		Then update address1, and don't switch back to it!
	 *
	 * Note: rest of the code assumes we are on address0 at boot time!
	 */
	dspgfb_writel(dspgfb, LCDC_REG_INTMR, new_bpp == 24 ? 1 : 0);
	dspgfb_writel(dspgfb, LCDC_REG_INDTR, ((info->var.xres * info->var.bits_per_pixel) / 32) - 1);

	if (dspgfb_readl(dspgfb, LCDC_REG_DISPIDXR) == 0) {
		dspgfb_update_refresh_address(dspgfb, 1, addr1, 1);
		dspgfb_update_refresh_address(dspgfb, 0, info->fix.smem_start, 1);
	} else {
		dspgfb_update_refresh_address(dspgfb, 0, info->fix.smem_start, 1);
		dspgfb_update_refresh_address(dspgfb, 1, addr1, 0);
	}

	return 1;
}

static int
dspgfb_pan_display(struct fb_var_screeninfo *var, struct fb_info *info)
{
	struct dspgfb *dspgfb = info->par;
	dma_addr_t addr;
	int res;

	if (dspgfb->pdata.cputype) {
		/* make sure last DMA towards cpu-type lcd is finished before starting new one */
		res = wait_event_interruptible_timeout( dspgfb->wq, !dspgfb->pending, HZ/10 );
		if (res <= 0)
			printk("%s: warning - couldn't get frame_end interrupt!\n", __func__);
		dspgfb->pending = true;
	}

	addr = info->fix.smem_start + (var->yoffset * info->fix.line_length);

	dspgfb_writel(dspgfb, LCDC_REG_MSBAHBA0R, addr >> 16);
	dspgfb_writel(dspgfb, LCDC_REG_LSBAHBA0R, addr & 0x0000ffff);

	/* kick lcdc hw: */
	/* for cpu-type lcd, DMA will be started */
	/* for rgb-interface lcd, ahb-address will be updated on next frame */
	dspgfb_hw_param_update(dspgfb);

	if (!dspgfb->pdata.cputype) {
		/* wait for the start of the next frame */
		/* after it - we are sure that LCDC is refreshing from the new address */
		dspgfb->pending = true;
		res = wait_event_interruptible_timeout( dspgfb->wq, !dspgfb->pending, HZ/10 );
		if (res <= 0)
			printk("%s: warning - couldn't get frame_start interrupt!\n", __func__);
	}

	return 0;
}

static u32 pseudo_palette[16];

static int
dspgfb_setcolreg(unsigned regno, unsigned red, unsigned green,
                          unsigned blue, unsigned transp, struct fb_info *info)
{
	u32 *palette = info->pseudo_palette;

	if (regno >= 16)
		return -EINVAL;

	/* only FB_VISUAL_TRUECOLOR supported */
	red    >>= 16 - info->var.red.length;
	green  >>= 16 - info->var.green.length;
	blue   >>= 16 - info->var.blue.length;
	transp >>= 16 - info->var.transp.length;

	palette[regno] = (red    << info->var.red.offset   ) |
			 (green  << info->var.green.offset ) |
			 (blue   << info->var.blue.offset  ) |
			 (transp << info->var.transp.offset) ;

	return 0;
}

static int
dspgfb_check_var(struct fb_var_screeninfo *var, struct fb_info *info)
{
	/* For now, don't check anything... */
	return 0;
}

static int
dspgfb_ioctl(struct fb_info *info, unsigned int cmd, unsigned long arg)
{
	void __user *argp = (void __user *)arg;
	struct dspgfb *dspgfb = info->par;
	unsigned short background_color;

	switch (cmd) {
		case FBIOSET_BACKGROUND:
			if (copy_from_user(&background_color, argp, sizeof(unsigned short)))
				return -EFAULT;

			dspgfb_set_display_background(dspgfb, background_color);
			break;
	}

	return 0;
}

static unsigned short
dspgfb_get_rgb565(unsigned short red, unsigned short green, unsigned short blue)
{
	return (((red & 0x1f) << 11) | ((green & 0x3f) << 5) | (blue & 0x1f));
}

static int
dspgfb_cmap(struct fb_cmap *cmap, struct fb_info *info)
{
	struct dspgfb *dspgfb = info->par;

	dspgfb_writel(dspgfb, LCDC_REG_PALR0,
		      dspgfb_get_rgb565(cmap->red[0], cmap->green[0], cmap->blue[0]));
	dspgfb_writel(dspgfb, LCDC_REG_PALR1,
		      dspgfb_get_rgb565(cmap->red[1], cmap->green[1], cmap->blue[1]));
	dspgfb_writel(dspgfb, LCDC_REG_PALR2,
		      dspgfb_get_rgb565(cmap->red[2], cmap->green[2], cmap->blue[2]));
	dspgfb_writel(dspgfb, LCDC_REG_PALR3,
		      dspgfb_get_rgb565(cmap->red[3], cmap->green[3], cmap->blue[3]));

	return 0;
}

static int
dspgfb_set_par(struct fb_info *info)
{
	struct dspgfb *dspgfb = info->par;
	unsigned long addr;
	int ret, xres, yres;
	long clk_rate = 0;

	dspgfb_hw_disable(dspgfb);
	dspgfb_writel(dspgfb, LCDC_REG_DISPCR, 0);

	xres = info->var.hsync_len + info->var.left_margin +
	       info->var.xres      + info->var.right_margin;
	yres = info->var.vsync_len + info->var.upper_margin +
	       info->var.yres      + info->var.lower_margin;
	clk_rate = xres * yres * info->mode->refresh;

	clk_rate = clk_round_rate(dspgfb->clk, clk_rate);
	if (clk_rate < 0)
		return clk_rate;
	clk_set_rate(dspgfb->clk, clk_rate);

	ret = dspgfb_set_display_bpp(dspgfb, info);
	if (ret < 0)
		return ret;

	dspgfb_writel(dspgfb, LCDC_REG_VSTR,   info->var.vsync_len - 1);
	dspgfb_writel(dspgfb, LCDC_REG_VFTR,   info->var.upper_margin - 1);
	dspgfb_writel(dspgfb, LCDC_REG_VATR,   info->var.yres - 1);
	dspgfb_writel(dspgfb, LCDC_REG_VETR,   info->var.lower_margin - 1);

	dspgfb_writel(dspgfb, LCDC_REG_HSTR,   info->var.hsync_len - 1);
	dspgfb_writel(dspgfb, LCDC_REG_HFTR,   info->var.left_margin - 1);
	dspgfb_writel(dspgfb, LCDC_REG_HADSTR, info->var.xres - 1);
	dspgfb_writel(dspgfb, LCDC_REG_HAPWR,  info->var.xres - 1);
	dspgfb_writel(dspgfb, LCDC_REG_HETR,   info->var.right_margin - 1);

	dspgfb_writel(dspgfb, LCDC_REG_INDTR, ((info->var.xres * info->var.bits_per_pixel) / 32 ) - 1);

	dspgfb_writel(dspgfb, LCDC_REG_INDXSR, info->var.xres - 1);
	dspgfb_writel(dspgfb, LCDC_REG_INDYSR, info->var.yres - 1);

	/* display position */
	dspgfb_writel(dspgfb, LCDC_REG_DISPXSPOSR,  0);
	dspgfb_writel(dspgfb, LCDC_REG_DISPXEPOSR,  info->var.xres - 1);
	dspgfb_writel(dspgfb, LCDC_REG_DISPYSPOS1R, 0);
	dspgfb_writel(dspgfb, LCDC_REG_DISPYEPOS1R, info->var.yres - 1);

	/* input buffer */
	addr = info->fix.smem_start;
	dspgfb_writel(dspgfb, LCDC_REG_MSBAHBA0R, addr >> 16);
	dspgfb_writel(dspgfb, LCDC_REG_LSBAHBA0R, addr & 0xffff);

	addr = info->fix.smem_start + (info->screen_size / 2);
	dspgfb_writel(dspgfb, LCDC_REG_MSBAHBA1R, addr >> 16);
	dspgfb_writel(dspgfb, LCDC_REG_LSBAHBA1R, addr & 0xffff);

	dspgfb_writel(dspgfb, LCDC_REG_OFFAR0, info->fix.line_length);
	dspgfb_writel(dspgfb, LCDC_REG_OFFAR1, info->fix.line_length);

	dspgfb_writel(dspgfb, LCDC_REG_DISPIDXR, 0);

	dspgfb_hw_param_update(dspgfb);

	dspgfb_writel(dspgfb, LCDC_REG_DISPCR, 1);

	dspgfb_hw_enable(dspgfb);

	return 0;
}

static struct fb_ops dspgfb_ops = {
	.owner          = THIS_MODULE,
	.fb_pan_display = dspgfb_pan_display,
	.fb_check_var   = dspgfb_check_var,
	.fb_ioctl       = dspgfb_ioctl,
	.fb_fillrect    = cfb_fillrect,
	.fb_copyarea    = cfb_copyarea,
	.fb_imageblit   = cfb_imageblit,
	.fb_setcolreg   = dspgfb_setcolreg,
	.fb_setcmap     = dspgfb_cmap,
	.fb_set_par     = dspgfb_set_par,
};

static irqreturn_t dspgfb_irq(int irq, void *priv)
{
	struct dspgfb *dspgfb = priv;
	int stat;

	/* read & clear active interrupts */
	stat = dspgfb_readl(dspgfb, LCDC_REG_INTSR);
	dspgfb_writel(dspgfb, LCDC_REG_INTSR, stat);

	if (stat & (INT_FRAME_START | INT_FRAME_DONE)) {
		dspgfb->pending = false;
		wake_up_interruptible( &dspgfb->wq );
		return IRQ_HANDLED;
	}

	/* handle error interrupts */
	if (stat & INT_UNDERRUN) {
		underruns++;
		return IRQ_HANDLED;
	}

	unhandled_ints++;

	return IRQ_HANDLED;
}

static int __devinit dspgfb_probe(struct platform_device *pdev)
{
	struct dspgfb *dspgfb;
	struct dspgfb_pdata *pdata = pdev->dev.platform_data;
	struct fb_info *info;
	struct resource *res;
	struct device_node *np = pdev->dev.of_node;
	int ret = 0, val;

	dspgfb = kzalloc(sizeof(*dspgfb), GFP_KERNEL);
	if (!dspgfb)
		return -ENOMEM;

	if (!pdata) {
		const char *tmp;

		ret = of_property_read_u32_array(np, "dspg,mode",
						 &(dspgfb->pdata.mode.refresh), 13);
		if (ret < 0) {
			dev_err(&pdev->dev, "invalid or missing 'dspg,mode' property!\n");
			goto err_free;
		}

		dspgfb->pdata.name = "unknown";
		ret = of_property_read_string(np, "dspg,panel_name", &dspgfb->pdata.name);
		if (ret && ret != -EINVAL) {
			dev_err(&pdev->dev, "invalid 'dspg,panel_name'\n");
			goto err_free;
		}

		val = 0;
		ret = of_property_read_u32(np, "dspg,cputype", &val);
		if (ret && ret != -EINVAL) {
			dev_err(&pdev->dev, "invalid 'cputype'\n");
			goto err_free;
		}
		dspgfb->pdata.cputype = val;

		ret = of_property_read_u32(np, "dspg,bits_per_pixel", &val);
		if (ret < 0 || (val != 16 && val != 18 && val != 24)) {
			dev_err(&pdev->dev, "invalid 'dspg,bits_per_pixel'\n");
			goto err_free;
		}
		dspgfb->pdata.bits_per_pixel = val;

		ret = of_property_read_u32(np, "dspg,fb_bits_per_pixel", &val);
		if ((ret && ret != -EINVAL) || (val != 16 && val != 18 && val != 24)) {
			dev_err(&pdev->dev, "invalid 'dspg,fb_bits_per_pixel'\n");
			goto err_free;
		}
		dspgfb->pdata.fb_bits_per_pixel = val;

		tmp = "rgb";
		ret = of_property_read_string(np, "dspg,data_order", &tmp);

		if (strcasecmp(tmp, "rgb") == 0)
			val = 0;
		else if (strcasecmp(tmp, "bgr") == 0)
			val = 1;
		else
			val = 2;

		if ((ret && ret != -EINVAL) || val > 1) {
			dev_err(&pdev->dev, "invalid 'dspg,data_order'\n");
			goto err_free;
		}
		dspgfb->pdata.order = val;
	} else {
		dspgfb->pdata = *pdata;
	}

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res) {
		dev_err(&pdev->dev, "need a memory resource\n");
		ret = -EINVAL;
		goto err_free;
	}

	dspgfb->irq = platform_get_irq(pdev, 0);
	if (dspgfb->irq < 0) {
		dev_err(&pdev->dev, "need an irq resource\n");
		ret = dspgfb->irq;
		goto err_free;
	}

	if (!request_mem_region(res->start, res->end - res->start, DSPGFB)) {
		dev_err(&pdev->dev, "failed to request memory region\n");
		ret = -EBUSY;
		goto err_free;
	}

	dspgfb->clk = clk_get(&pdev->dev, "lcdc");
	if (IS_ERR(dspgfb->clk)) {
		dev_err(&pdev->dev, "could not get clock\n");
		ret = PTR_ERR(dspgfb->clk);
		goto err_region;
	}
	clk_enable(dspgfb->clk);

	info = framebuffer_alloc(sizeof(dspgfb), &pdev->dev);
	if (!info) {
		dev_err(&pdev->dev, "out of memory\n");
		ret = -ENOMEM;
		goto err_clk;
	}
	dspgfb->info = info;

	info->pseudo_palette = pseudo_palette;
	info->par = dspgfb;
	info->flags = FBINFO_FLAG_DEFAULT;

	strcpy(info->fix.id, DSPGFB);
	info->fix.type = FB_TYPE_PACKED_PIXELS;
	info->fix.visual = FB_VISUAL_TRUECOLOR;
	info->fix.accel = FB_ACCEL_NONE;

	dspg_bpp_to_info(&info->var, dspgfb->pdata.fb_bits_per_pixel);
	info->fix.line_length = dspgfb->pdata.mode.xres * info->var.bits_per_pixel / 8;

	info->fbops = &dspgfb_ops;
	info->mode = &dspgfb->pdata.mode;
	fb_videomode_to_var(&info->var, &dspgfb->pdata.mode);

	/* double buffering */
	info->fix.ypanstep = 1;
	info->var.yres_virtual *= 3;

	info->screen_size = info->fix.line_length * info->var.yres_virtual;
	info->fix.smem_len = info->screen_size;
	info->var.activate = FB_ACTIVATE_FORCE | FB_ACTIVATE_NOW;

	ret = fb_alloc_cmap(&info->cmap, 256, 0);
	if (ret < 0) {
		dev_err(&pdev->dev, "unable to allocate color map\n");
		goto err_fb;
	}

	/* setup memory */
	dspgfb->vidmem = dma_alloc_coherent(&pdev->dev, info->screen_size,
					    (dma_addr_t *)&info->fix.smem_start,
					    GFP_DMA);
	if (!dspgfb->vidmem) {
		dev_err(&pdev->dev, "could not allocate framebuffer (%lu bytes)\n",
		        info->screen_size);
		ret = -ENOMEM;
		goto err_cmap;
	}
	info->screen_base = dspgfb->vidmem;

	/* init hardware */
	dspgfb->regs = ioremap(res->start, res->end - res->start);
	if (!dspgfb->regs) {
		dev_err(&pdev->dev, "unable to map registers\n");
		ret = -ENOMEM;
		goto err_dma;
	}

	dev_set_drvdata(&pdev->dev, dspgfb);
	init_waitqueue_head( &dspgfb->wq );

	ret = request_irq(dspgfb->irq, dspgfb_irq, 0, DSPGFB, dspgfb);
	if (ret < 0) {
		dev_err(&pdev->dev, "request irq %d failed\n", dspgfb->irq);
		goto err_ioremap;
	}

	dspgfb_set_hw_init(dspgfb);
	if (!dspgfb_hw_try_resume(dspgfb)) {
		memset(info->screen_base, 0, info->screen_size);

		/* fb_set_var must be called after all other things, since it calls fb_pan_display */
		/* which needs irq and wq to be initialized */
		ret = fb_set_var(info, &info->var);
		if (ret < 0)
			goto err_irq;
	}

	/* register framebuffer in userspace */
	ret = register_framebuffer(info);
	if (ret < 0)
		goto err_irq;

	ret = of_platform_populate(np, NULL, NULL, &pdev->dev);
	if (ret < 0) {
		dev_err(&pdev->dev, "error populating subdevices\n");
		goto err_framebuffer;
	}

	dev_info(&pdev->dev, "fb%d: %s frame buffer, %s (%dx%d, %d bpp, %luk)\n",
	       info->node,
	       info->fix.id,
	       dspgfb->pdata.name,
	       info->var.xres,
	       info->var.yres,
	       info->var.bits_per_pixel,
	       info->screen_size >> 10);

	return 0;

err_framebuffer:
	framebuffer_release(dspgfb->info);
err_irq:
	free_irq(dspgfb->irq, dspgfb);
err_ioremap:
	iounmap(dspgfb->regs);
err_dma:
	dma_free_coherent(&pdev->dev, info->screen_size, dspgfb->vidmem, (dma_addr_t)info->fix.smem_start);
err_cmap:
	fb_dealloc_cmap(&info->cmap);
err_fb:
	framebuffer_release(dspgfb->info);
err_clk:
	clk_disable(dspgfb->clk);
	clk_put(dspgfb->clk);
err_region:
	release_mem_region(res->start, res->end - res->start);
err_free:
	kfree(dspgfb);

	return ret;
}

static int __exit dspgfb_remove(struct platform_device *pdev)
{
	struct dspgfb *dspgfb = dev_get_drvdata(&pdev->dev);
	struct resource *res = platform_get_resource(pdev, IORESOURCE_MEM, 0);

	free_irq(dspgfb->irq, dspgfb);
	iounmap(dspgfb->regs);
	dma_free_coherent(&pdev->dev, dspgfb->info->screen_size, dspgfb->vidmem,
			  (dma_addr_t)dspgfb->info->fix.smem_start);
	fb_dealloc_cmap(&dspgfb->info->cmap);
	framebuffer_release(dspgfb->info);
	clk_disable(dspgfb->clk);
	clk_put(dspgfb->clk);
	release_mem_region(res->start, res->end - res->start);
	kfree(dspgfb);

	return 0;
}

#if defined(CONFIG_PM)
static int dspgfb_dev_suspend(struct platform_device *pdev, pm_message_t state)
{
	struct dspgfb *dspgfb = dev_get_drvdata(&pdev->dev);

	dspgfb_writel(dspgfb, LCDC_REG_DISPCR, 0);
	dspgfb_hw_param_update(dspgfb);
	dspgfb_hw_param_update(dspgfb);

	disable_irq(dspgfb->irq);

	clk_disable(dspgfb->clk);

	return 0;
}

static int dspgfb_dev_resume(struct platform_device *pdev)
{
	struct dspgfb *dspgfb = dev_get_drvdata(&pdev->dev);

	clk_enable(dspgfb->clk);

	enable_irq(dspgfb->irq);

	dspgfb_writel(dspgfb, LCDC_REG_DISPCR, 1);
	dspgfb_hw_param_update(dspgfb);

	return 0;
}
#else
#define dspgfb_dev_suspend NULL
#define dspgfb_dev_resume NULL
#endif

/* TODO: integrate CPU-type panels as "dspg,dspgfb-cpu". Can use .data to pass
 * extra information if required...
 */
static struct of_device_id dspgfb_of_ids[] = {
	{ .compatible = "dspg,dspgfb-lcd" },
	{ },
};

static struct platform_driver dspgfb_driver = {
	.remove = __exit_p(dspgfb_remove),
	.suspend = dspgfb_dev_suspend,
	.resume = dspgfb_dev_resume,
	.driver = {
		.name = DSPGFB,
		.owner = THIS_MODULE,
		.of_match_table = dspgfb_of_ids,
	},
};

static int __init dspgfb_init(void)
{
	return platform_driver_probe(&dspgfb_driver, dspgfb_probe);
}

static void __exit dspgfb_exit(void)
{
	platform_driver_unregister(&dspgfb_driver);
}

module_init(dspgfb_init);
module_exit(dspgfb_exit);
