/****************************************************************************
 *
 * FILENAME:        $RCSfile: kgui.c,v $
 *
 * LAST REVISION:   $Revision: 1.6 $
 * LAST MODIFIED:   $Date: 2014/05/16 20:29:11 $
 *
 * DESCRIPTION:
 *
 * vi: set ts=4:
 *
 * Copyright (c) 2005 by Grandstream Networks, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***************************************************************************/

//=======================
//  Includes
//=======================
#include <linux/module.h>
#include <linux/fb.h>
#include <linux/kthread.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/cdev.h>
#include <linux/kgui.h>
#include <linux/vmalloc.h>
#include <linux/mm.h>
#include <linux/bmp_layout.h>
#include <linux/xz.h>
#include <asm/uaccess.h>
#include "font.h"

//=======================
//  Defines
//=======================
#define KERNING 0
#define SPACING 6

#define BLACK 0xF0
#define GREY 0xF0

#define FONT_HEIGHT 16
#define THROBBER_BAR_HEIGHT 10
#define VERTICAL_SPACING 5
#define MESSAGE_AREA_HEIGHT (FONT_HEIGHT + VERTICAL_SPACING)
#define THROBBER_BAR_AREA_HEIGHT (THROBBER_BAR_HEIGHT + VERTICAL_SPACING)

#define lcd_par struct uc1701fb_par

//=======================
//  Globals
//=======================

//=======================
//  Typedefs
//=======================

//=======================
//  Locals
//=======================
    /*
     * Private Data
     */
static int throbber_location = -1;
static struct task_struct * render_thread = NULL;
static int progress_location = -1;
static int progress_enabled = 0;
static int progress_max = 100; /* initialize the denominator */
static int progress_now;

static character_t fallback_glyph =
    {
        0,
        {
            0xFFFF, /* **************** */
            0x8001, /* *              * */
            0x8001, /* *              * */
            0x8001, /* *              * */
            0x8001, /* *              * */
            0x8001, /* *              * */
            0x8281, /* *     * *      * */
            0x8101, /* *      *       * */
            0x8281, /* *     * *      * */
            0x8001, /* *              * */
            0x8001, /* *              * */
            0x8001, /* *              * */
            0x8001, /* *              * */
            0x8001, /* *              * */
            0xFFFF  /* **************** */
        },
        16
    };
static unsigned char * backbuffer_data;
static unsigned int screen_size_x;
static unsigned int screen_size_y;
static unsigned int viewable_x;
static unsigned int viewable_y;
static unsigned int start_x;
static unsigned int shift_bpp = 0;
static unsigned int pixel_bytes;
static int kgui_running;

static void (*activate)(void) = NULL;
static void (*deactivate)(void) = NULL;
static void (*flip)(int) = NULL;
static void (*update_screen)(void *, int, int, int, int) = NULL;

static int render_task( void * data );

static struct cdev cdev;
static struct class* kgui_class;
static dev_t dev_no;

struct uc1701fb_par
{
    unsigned char* fb_buffer;
    unsigned int fb_size;
    dma_addr_t dma_handle;
    struct spi_device * spi;
};

static unsigned char* background = NULL;
static unsigned int background_size;
static struct kgui_op kgui_ops;
static DEFINE_MUTEX(kgui_mutex);

    /*
     * Private Functions
     */

/* Utility Functions */
static character_t * find_glyph( unsigned int utf8 );
static unsigned int decode_utf8( unsigned char * buf, int bytes );
static unsigned int getchar_utf8( unsigned char * str, int * offset );
static void draw_glyphs( character_t ** glyphs, int len, int x, int y );
static void draw_glyph( character_t * glyph, int x, int y );
static void put_pixel( unsigned char color, int x, int y );
static void clear_block( int x, int y, int width, int height );
static void kgui_draw_throbber( void );
static int y_location( int location );

static long kgui_ioctl( struct file* filp, unsigned int cmd, unsigned long arg );
static int kgui_mmap( struct file* filp, struct vm_area_struct* vma );

static void kgui_draw_progress( void );
static void kgui_enable_progress( int max );
static void kgui_progress( int increment );
static void kgui_string( char * str );
static void kgui_clear( void );
static void kgui_logo( void );

static struct file_operations kgui_fops = {
    owner: THIS_MODULE,
    unlocked_ioctl: kgui_ioctl,
    mmap: kgui_mmap
};


//=======================
//  Functions
//=======================

static struct uc1701fb_par* par;
//=============================================================================
void
kgui_init(
    struct fb_info * info,
    void (*activate_fxn)(void),
    void (*deactivate_fxn)(void),
    void (*flip_fxn)(int),
    void (*update_screen_fxn)(void *, int, int, int, int)
)
//=============================================================================
{
    
    par = info->par;
    unsigned char * fb = par->fb_buffer;
    unsigned int depth = info->var.bits_per_pixel;

    screen_size_x = info->var.xres; 
    screen_size_y = info->var.yres;
    viewable_x = info->var.xres;
    viewable_y = info->var.yres;
    start_x = 0;
    pixel_bytes = 1; 

    printk("screen_size_x = %d, screen_size_y = %d, viewable_x =%d, viewable_y = %d, start_x = %d, pixel_bytes = %d\n", 
    screen_size_x, screen_size_y, viewable_x, viewable_y, start_x, pixel_bytes);

    backbuffer_data = fb;

    activate = activate_fxn;
    deactivate = deactivate_fxn;
    flip = flip_fxn;
    update_screen = update_screen_fxn;

    if ( alloc_chrdev_region( &dev_no, 0, 1, "kgui" ) )
    {
        printk("kgui_init: unable to allocate chrdev region\n");
        return;
    }

    cdev_init( &cdev, &kgui_fops );
    if ( cdev_add( &cdev, dev_no, 1 ) )
    {
        printk("kgui_init: unable to init cdev\n");
        unregister_chrdev_region( dev_no, 1 );
        return;
    }

    kgui_class = class_create( THIS_MODULE, "kgui" );
    if ( !kgui_class )
    {
        unregister_chrdev_region( dev_no, 1 );
        cdev_del( &cdev );
        printk("kgui_init: unable to create class\n");
        return;
    }

    device_create( kgui_class, NULL, MKDEV( MAJOR( dev_no ), 0 ), NULL, "kgui%d", 0 );

    memset( &kgui_ops, 0, sizeof ( kgui_ops ) );
    kgui_ops.logo = 0;
    kgui_running = 1;
    render_thread = kthread_run( render_task, 0, "kgui_render_task" );

    printk("kgui_init: done\n");
}

//=============================================================================
static long
kgui_ioctl(
    struct file* filp,
    unsigned int cmd,
    unsigned long arg
)
//=============================================================================
{
    int err = 0;
    int progress;
    struct kgui_op kgui_op;

    if ( _IOC_DIR( cmd ) & _IOC_WRITE )
        err = !access_ok( VERIFY_READ, (void __user *)arg, _IOC_SIZE( cmd )  );

    if ( err )
        return -EFAULT;

    switch ( cmd )
    {
        case KGUI_IOC_ACTIVATE:
            if ( activate && !kgui_running )
            {
                kgui_running = 1;
                render_thread = kthread_run( render_task, 0, "kgui_render_task" );
                activate();
            }
            break;
        case KGUI_IOC_DEACTIVATE:
            if ( deactivate && kgui_running )
            {
                kgui_running = 0;
                kthread_stop( render_thread );
                deactivate();
            }
            break;
        case KGUI_IOC_PROGRESS:
            if ( copy_from_user( &progress, (void*)arg, sizeof ( progress ) ) != 0 )
                return -EIO;
           
            mutex_lock( &kgui_mutex ); 
            kgui_progress( progress );
            mutex_unlock( &kgui_mutex );
            break;
        case KGUI_IOC_OP:
            if ( copy_from_user( &kgui_op, (void*)arg, sizeof( kgui_op ) ) != 0 )
                return -EIO;

            mutex_lock( &kgui_mutex );
            memcpy( &kgui_ops, &kgui_op, sizeof ( kgui_ops ) );

            if ( kgui_ops.progress.percent )
                kgui_enable_progress( kgui_ops.progress.percent );
            mutex_unlock( &kgui_mutex );

            break;
        default:
            return -EINVAL;
    }

    return 0;
}

//=============================================================================
static int
kgui_mmap(
    struct file* filp,
    struct vm_area_struct* vma
)
//=============================================================================
{
    unsigned long start = vma->vm_start;
    unsigned long size = vma->vm_end - vma->vm_start;
    unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;
    unsigned long page, pos;

    if ( offset + size > background_size )
    {
        return -EINVAL;
    }

    pos = ( unsigned long )background;

    while ( size > 0 )
    {
        page = vmalloc_to_pfn( (void*)pos );
        if ( remap_pfn_range(
                    vma,
                    start,
                    page,
                    PAGE_SIZE,
                    pgprot_noncached( vma->vm_page_prot ) ) )
        {
            return -EAGAIN;
        }

        start += PAGE_SIZE;
        pos += PAGE_SIZE;

        if ( size > PAGE_SIZE )
            size -= PAGE_SIZE;
        else
            size = 0;
    }

    return 0;
}

//=============================================================================
void
kgui_logo(
    void
)
//=============================================================================
{
#if defined( CONFIG_FB_LOGO )
    int i;
    int j;

    for ( i = 0; i < logo_height; i++ ) {
        for ( j = 0; j < logo_width; j++ ) {
            if ( logo_data[ i * logo_width + j ] ) {
                put_pixel( BLACK, j, i );
            }
        }
    }
#endif
}

//=============================================================================
static void
kgui_clear(
    void
    )
//=============================================================================
{

}

//=============================================================================
static void
kgui_progress(
    int increment
)
//=============================================================================
{
    progress_now += increment;

    if ( progress_now > progress_max )
        progress_now = progress_max;
}

//=============================================================================
static void
kgui_draw_progress(
    void
)
//=============================================================================
{
    int x, y;
    int i, j;

    y = viewable_y - THROBBER_BAR_AREA_HEIGHT;
    x = viewable_x / 2 - 52 + start_x;

    clear_block( x - 1, y, 106, 10 );

    for ( i = 0; i < 102; i++ )
    {
        put_pixel( BLACK, x + i, y );
    }
    put_pixel( BLACK, x - 1, y + 1 );
    put_pixel( BLACK, x + 102, y + 1 );
    for ( i = 0; i < 6; i++ )
    {
        put_pixel( BLACK, x - 1, y + 2 + i );
        put_pixel( BLACK, x + 102, y + 2 + i );
        for ( j = 0; j < ( 100 * progress_now / progress_max ); j++ )
        {
            put_pixel( GREY, x + 1 + j, y + 2 + i );
        }
    }
    put_pixel( BLACK, x - 1, y + 8 );
    put_pixel( BLACK, x + 102, y + 8 );
    for ( i = 0; i < 102; i++ )
    {
        put_pixel( BLACK, x + i, y + 9 );
    }
}

//=============================================================================
static void
kgui_enable_progress(
    int max
)
//=============================================================================
{
    progress_max = max;
    progress_now = 0;

    if ( progress_max == 0 )
    {
        progress_now = progress_max = 100;
    }
}

//=============================================================================
static int
render_task(
    void * data
)
//=============================================================================
{
    while ( !kthread_should_stop() )
    {
        memset( backbuffer_data, 0, screen_size_x * screen_size_y * pixel_bytes / 8);

        if ( kgui_ops.string.enabled )
            kgui_string( &kgui_ops.string.string[0] );

        if ( kgui_ops.progress.enabled )
            kgui_draw_progress( );
            
        if ( kgui_ops.throbber.enabled )
            kgui_draw_throbber();


        update_screen(par, 0, 0, screen_size_x, screen_size_y);

        msleep_interruptible( 20 );
    }

    return 0;
}

//=============================================================================
static void
kgui_draw_throbber(
    void
)
//=============================================================================
{
    int x, y;
    int i, j;
    static int throbber = 0;

    y = viewable_y - THROBBER_BAR_AREA_HEIGHT;
    x = viewable_x / 2 - 52 + start_x;

    //while ( !kthread_should_stop() )
    {
        clear_block( x - 1, y, 106, 10 );

        for ( i = 0; i < 102; i++ )
        {
            put_pixel( BLACK, x + i, y );
        }
        put_pixel( BLACK, x - 1, y + 1 );
        put_pixel( BLACK, x + 102, y + 1 );
        for ( i = 0; i < 6; i++ )
        {
            put_pixel( BLACK, x - 1, y + 2 + i );
            put_pixel( BLACK, x + 102, y + 2 + i );
            if ( throbber < 32 )
            {
                for ( j = 0; j < throbber; j++ )
                {
                    put_pixel( GREY, x + 1 + j, y + 2 + i );
                }
            }
            else
            {
                for ( j = 0; j < 32; j++ )
                {
                    if ( 1 + ( throbber - 32 ) + j <= 100 )
                        put_pixel( GREY, x + 1 + ( throbber - 32 ) + j, y + 2 + i );
                    else
                        break;
                }
            }
        }
        put_pixel( BLACK, x - 1, y + 8 );
        put_pixel( BLACK, x + 102, y + 8 );
        for ( i = 0; i < 102; i++ )
        {
            put_pixel( BLACK, x + i, y + 9 );
        }

        if ( throbber > 132 )
            throbber = 0;
        else
            throbber += 1;

    }
}

//=============================================================================
static void
kgui_string(
    char * str
)
//=============================================================================
{
    int x, y;
    int printable_width;
    int str_bytes;
    int i, j;
    character_t * utf8_char[256];
    int utf8_index;
    int utf8_len;
    //int split_points[16];
    int offset;

    y = viewable_y - (MESSAGE_AREA_HEIGHT + THROBBER_BAR_AREA_HEIGHT);
    
    str_bytes = strlen( str );
    utf8_index = 0;
    offset = 0;
    for ( offset = 0; offset < str_bytes; )
    {
        unsigned int utf8;
        utf8 = getchar_utf8( str, &offset );
        if ( utf8 == 0 )
        {
            printk("Invalid UTF8\n");
            return;
        }
        utf8_char[ utf8_index ] = find_glyph( utf8 );
        if ( utf8_char[ utf8_index ] == NULL )
        {
            utf8_char[ utf8_index ] = &fallback_glyph;
        }

        utf8_index++;
    }
    utf8_char[ utf8_index ] = 0;
    utf8_len = utf8_index;

    printable_width = 0;
    for ( i = 0; i < utf8_len; i++ )
    {
        printable_width += utf8_char[i]->width + KERNING;
    }

    if ( printable_width > viewable_x )
    {
        struct {
            int start;
            int width;
            int chars;
        } fragments[128];
        int fi = 0;
        int last_char = 0x20;
        int line_width;
        int line_start;

        memset( fragments, 0, sizeof ( fragments ) );

        for ( i = 0; i < utf8_len; i++ )
        {
            if ( last_char == 0x20 && utf8_char[i]->utf8 != 0x20 )
            {
                fragments[fi].start = i;
                for ( j = i; j < utf8_len; j++ )
                {
                    if ( utf8_char[j]->utf8 == 0x20 )
                        break;
                    fragments[fi].width += utf8_char[j]->width + KERNING;
                    fragments[fi].chars++;
                }
                fi++;
                i = j;
                last_char = 0x20;
                continue;
            }
            last_char = utf8_char[i]->utf8;
        }

        line_width = 0;
        line_start = 0;
        for ( i = 0; i < fi; i++ )
        {
            if ( fragments[i].width > viewable_x )
            {
                continue;
            }
            else if ( line_width + fragments[i].width + SPACING > viewable_x )
            {
                int next_line = 0;

                clear_block( start_x, y, viewable_x, 16 );

                x = ( viewable_x - line_width ) / 2;
                x += start_x;
                for ( j = line_start; j < i; j++ )
                {
                    draw_glyphs( &utf8_char[fragments[j].start], fragments[j].chars, x, y  );
                    x += fragments[j].width + SPACING;
                }
                y += 16;

                line_start = -1;
                for ( j = i; j < fi; j++ )
                {
                    if ( fragments[j].width <= viewable_x )
                    {
                        line_start = j;
                        line_width = fragments[j].width + SPACING;
                        i = j;
                        break;
                    }
                }
                if ( line_start < 0 )
                {
                    break;
                }
                for ( j = line_start; j < fi; j++ )
                {
                    next_line += fragments[j].width + SPACING;
                }
                if ( next_line < viewable_x )
                {
                    clear_block( start_x, y, viewable_x, 16 );

                    x = ( viewable_x - next_line ) / 2;
                    x += start_x;
                    for ( j = line_start; j < fi; j++ )
                    {
                        draw_glyphs( &utf8_char[fragments[j].start], fragments[j].chars, x, y  );
                        x += fragments[j].width + SPACING;
                    }
                    break;
                }
                continue;
            }
            else
            {
                line_width += ( fragments[i].width + SPACING );
            }
        }

    }
    else
    {
        clear_block( start_x, y, viewable_x, 16 );

        x = ( viewable_x - printable_width ) / 2;
        x += start_x;
        draw_glyphs( utf8_char, utf8_len, x, y  );
    }
}

//=============================================================================
static void
draw_glyphs(
    character_t ** glyphs,
    int len,
    int x,
    int y
)
//=============================================================================
{
    int i;

    for ( i = 0; i < len; i++ )
    {
        draw_glyph( glyphs[i], x, y );
        x += glyphs[i]->width + KERNING;
    }
}

//=============================================================================
static void
draw_glyph(
    character_t * glyph,
    int x,
    int y
)
//=============================================================================
{
    int i, j;
    int tx, ty;

    if ( !glyph )
    {
        glyph = &fallback_glyph;
    }

    tx = x;
    ty = y;
    for ( i = 0; i < 15; i++ )
    {
        for ( j = 15; j >= (signed)( 16 - glyph->width ); j-- )
        {
            if ( glyph->glyph[i] & ( 1 << j ) )
            {
                put_pixel( BLACK, tx, ty );
            }
            tx++;
        }
        tx = x;
        ty++;
    }
}

//=============================================================================
static character_t *
find_glyph(
    unsigned int utf8
)
//=============================================================================
{
    int i;

    for ( i = 0; i < unifont.num_chars; i++ )
    {
        if ( unifont.characters[i].utf8 == utf8 )
        {
            return &unifont.characters[i];
        }
    }

    return NULL;
}

//=============================================================================
static void
put_pixel(
    unsigned char color,
    int x,
    int y
)
//=============================================================================
{
    unsigned char * p_fb;


    p_fb = backbuffer_data + ( screen_size_x * ( y / 8 ) + x ); 
    *p_fb |= ( ( color != 0 ) << ( y % 8 ) ); 
}


//=============================================================================
static void
clear_block(
    int x,
    int y,
    int width,
    int height
)
//=============================================================================
{

    unsigned char * p_fb;
    int i;
    int j;

    for ( i = x; i < ( x + width ); i++ )
    {
        for ( j = y; j < ( y + height ); j++ )
        {
            p_fb = backbuffer_data + ( screen_size_x * ( j / 8 ) + i );
            *p_fb &= ~( 1 << ( j % 8 ) );
        }
    }
}

//=============================================================================
static int
y_location(
    int location
)
//=============================================================================
{
    int y;

    switch ( location )
    {
        case LOCATION_TOP:
            y = viewable_y / 4 - 8; /* top third */
            break;
        case LOCATION_MID:
            y = viewable_y / 2 - 8;
            break;
        case LOCATION_BOT:
            y = 3 * ( viewable_y / 4 ) - 8; /* bottom third */
            break;
        case LOCATION_TRAIL:
            y = viewable_y - 16;
            break;
        default:
            y = 0;
            break;
    }

    return y;
}

//=============================================================================
static unsigned int
decode_utf8(
    unsigned char * buf,
    int bytes
)
//=============================================================================
{
    switch ( bytes )
    {
        case 1:
            return (unsigned int)( buf[0] & ~0x80 );
        case 2:
            return (unsigned int)( ( ( buf[0] & ~0xE0 ) << 6 ) | ( buf[1] & ~0xC0 ) );
        case 3:
            return (unsigned int)( ( ( buf[0] & ~0xF0 ) << 12 ) | ( ( buf[1] & ~0xC0 ) << 6 ) | ( buf[2] & ~0xC0 ) );
        case 4:
            return (unsigned int)( ( ( buf[0] & ~0xF1 ) << 18 ) | ( ( buf[1] & ~0xC0 ) << 12 ) | ( ( buf[2] & ~0xC0 ) <<6 ) | ( buf[3] & ~0xC0 ) );
    }
    return 0;
}

//=============================================================================
static unsigned int
getchar_utf8(
    unsigned char * str,
    int * off
)
//=============================================================================
{
    unsigned char buf[4];

    buf[0] = str[*off]; *off = *off + 1;

    if ( ( buf[0] & 0x80 ) == 0 )
    {   /* 8-bit char */
        return decode_utf8( buf, 1 );
    }
    else if ( ( buf[0] & 0xE0 ) == 0xC0 )
    {   /* 16-bit char */
        buf[1] = str[*off]; *off = *off + 1;
        if ( ( buf[1] & 0xC0 ) != 0x80 )
        {
            goto parse_error;
        }
        return decode_utf8( buf, 2 );
    }
    else if ( ( buf[0] & 0xF0 ) == 0xE0 )
    {   /* 24-bit char */
        buf[1] = str[*off]; *off = *off + 1;
        buf[2] = str[*off]; *off = *off + 1;
        if ( ( ( buf[1] & 0xC0 ) != 0x80 ) || ( ( buf[2] & 0xC0 ) != 0x80 ) )
        {
            goto parse_error;
        }
        return decode_utf8( buf, 3 );
    }
    else if ( ( buf[0] & 0xF8 ) == 0xF0 )
    {   /* 32-bit char */
        buf[1] = str[*off]; *off = *off + 1;
        buf[2] = str[*off]; *off = *off + 1;
        buf[3] = str[*off]; *off = *off + 1;
        if ( ( ( buf[1] & 0xC0 ) != 0x80 ) || ( ( buf[2] & 0xC0 ) != 0x80 ) || ( ( buf[3] & 0xC0 ) != 0x80 ) )
        {
            goto parse_error;
        }
        return decode_utf8( buf, 4 );
    }

parse_error:
    printk("Parse Error\n");
    return 0;
}

