/*
 * dspg_wdt.c - DSPG watchdog userspace interface
 *
 * Copyright (c) 2011, 2012 DSPG Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <linux/fs.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/module.h>
#include <linux/clk.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>
#include <linux/watchdog.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <mach/hardware.h>

MODULE_DESCRIPTION("DSPG watchdog driver");
MODULE_LICENSE("GPL");
MODULE_ALIAS_MISCDEV(WATCHDOG_MINOR);

#define DSPG_WDT "dspg_wdt"

#define WD_CTRL     0x00
#define WD_ENABLE   0x04
#define WD_REFRESH  0x08
#define WD_INTCLR   0x1c
#define WD_INTSTAT  0x20

#define WD_CTRL_PERIOD_1s     (0 << 0)
#define WD_CTRL_PERIOD_10s    (1 << 0)
#define WD_CTRL_PERIOD_100s   (2 << 0)
#define WD_CTRL_PERIOD_1000s  (3 << 0)
#define WD_CTRL_INTEN         (1 << 2)

static void __iomem *dspg_wdt_base;
static int dspg_wdt_boot_status;
static unsigned long dspg_wdt_opened;
static struct miscdevice dspg_wdt_miscdev;
static DEFINE_SPINLOCK(dspg_wdt_lock);

static void
dspg_wdt_enable(void)
{
	unsigned long flags;

	spin_lock_irqsave(&dspg_wdt_lock, flags);
	writew(0xabcd, dspg_wdt_base + WD_ENABLE);
	writew(0xabcd, dspg_wdt_base + WD_ENABLE);
	spin_unlock_irqrestore(&dspg_wdt_lock, flags);
}

static void
dspg_wdt_refresh(void)
{
	unsigned long flags;

	spin_lock_irqsave(&dspg_wdt_lock, flags);
	writew(0xa5a5, dspg_wdt_base + WD_REFRESH);
	writew(0x5a5a, dspg_wdt_base + WD_REFRESH);
	spin_unlock_irqrestore(&dspg_wdt_lock, flags);
}

static int
dspg_wdt_open(struct inode *inode, struct file *file)
{
	if (test_and_set_bit(0, &dspg_wdt_opened))
		return -EBUSY;

	dspg_wdt_enable();
	dev_info(dspg_wdt_miscdev.parent, "timer started\n");

	return nonseekable_open(inode, file);
}

static int
dspg_wdt_release(struct inode *inode, struct file *file)
{
	/*
	 * once the watchdog is running it cannot be disabled, we can refresh
	 * it once more, though
	 */
	dspg_wdt_refresh();

	clear_bit(0, &dspg_wdt_opened);
	return 0;
}

static ssize_t
dspg_wdt_write(struct file *file, const char *data, size_t len, loff_t * ppos)
{
	if (len)
		dspg_wdt_refresh();

	return len;
}

static struct watchdog_info info = {
	.identity = DSPG_WDT,
	.options = WDIOF_CARDRESET,
};

static long
dspg_wdt_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	int ret = -EOPNOTSUPP;
	void __user *argp = (void __user *)arg;
	int __user *p = argp;

	switch (cmd) {
	case WDIOC_GETSUPPORT:
		ret = copy_to_user(argp, &info,
		                   sizeof(info)) ? -EFAULT : 0;
		break;

	case WDIOC_GETSTATUS:
		ret = put_user(0, p);
		break;

	case WDIOC_GETBOOTSTATUS:
		ret = put_user(dspg_wdt_boot_status, p);
		break;

	case WDIOC_KEEPALIVE:
		dspg_wdt_refresh();
		ret = 0;
		break;
	}

	return ret;
}

static const struct file_operations dspg_wdt_fops = {
	.owner		= THIS_MODULE,
	.llseek		= no_llseek,
	.write		= dspg_wdt_write,
	.unlocked_ioctl	= dspg_wdt_ioctl,
	.open		= dspg_wdt_open,
	.release	= dspg_wdt_release,
};

static struct miscdevice dspg_wdt_miscdev = {
	.minor		= WATCHDOG_MINOR,
	.name		= DSPG_WDT,
	.fops		= &dspg_wdt_fops,
};

static irqreturn_t
dspg_wdt_isr(int irq, void *unused)
{
	return IRQ_HANDLED;
}

static int __init
dspg_wdt_probe(struct platform_device *pdev)
{
	int irq, ret;
	struct resource *res;

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res)
		return -EINVAL;

	irq = platform_get_irq(pdev, 0);
	if (irq < 0)
		return irq;

	if (!request_mem_region(res->start, res->end - res->start + 1,
	    pdev->name)) {
		dev_err(&pdev->dev, "failed to request memory resource\n");
		return -EBUSY;
	}

	dspg_wdt_base = ioremap_nocache(res->start, res->end - res->start + 1);
	if (!dspg_wdt_base) {
		dev_err(&pdev->dev, "failed to map registers\n");
		return -ENOMEM;
	}

	ret = request_irq(irq, dspg_wdt_isr, 0, DSPG_WDT, NULL);
	if (ret) {
		dev_err(&pdev->dev, "error requesting irq\n");
		return -EIO;
	}

	/* setup with 10 seconds period */
	writew(WD_CTRL_PERIOD_10s, dspg_wdt_base + WD_CTRL);

	dspg_wdt_miscdev.parent = &pdev->dev;
	ret = misc_register(&dspg_wdt_miscdev);
	if (ret)
		dev_err(&pdev->dev, "error registering miscdev: %d\n", ret);

	dev_info(&pdev->dev, "initialized, not started\n");

	return ret;
}

static struct of_device_id dspg_wdt_of_ids[] = {
	{ .compatible = "dspg,dspg_wdt" },
	{ },
};

static struct platform_driver watchdog_driver = {
	.driver = {
		.name = "dspg_wdt",
		.owner = THIS_MODULE,
		.of_match_table = dspg_wdt_of_ids,
	},
};

static int __init
dspg_wdt_init(void)
{
	return platform_driver_probe(&watchdog_driver, dspg_wdt_probe);
}

module_init(dspg_wdt_init);

