/*
 * include/linux/coma/cfifo.h - cordless FIFOs
 *
 * Cordless FIFOs store variable-length messages in a ring buffer. This FIFO
 * implementation is safe for parallel accesses of not more than one reader and
 * one writer. All messages are stored in a consecutive memory area, i.e.,
 * they are never wrapped around at the end of the FIFO as in many other ring
 * buffer implementations.
 *
 * Copyright (C) 2007 NXP Semiconductors
 * Copyright (C) 2008 - 2012 DSP Group Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __COMA_CFIFO_H
#define __COMA_CFIFO_H

#include <linux/device.h>
#include <linux/dma-mapping.h>

#define CFIFO_MSGHEADER_SIZE (sizeof(unsigned int))
#define CFIFO_PAD(x)         (((x) + (CFIFO_MSGHEADER_SIZE-1)) & \
                              ~(CFIFO_MSGHEADER_SIZE-1))

#define CFIFO_MAGIC (((unsigned)'C' << 0) | ((unsigned)'F' << 8) | \
			((unsigned)'I' << 16) | ((unsigned)'F' << 24))

/*
 * CFIFO shared data structure between Linux and Cordless domains.
 */
struct cfifo_shm {
	unsigned int magic;	/* CFIFO magic 'CFIF' */
	unsigned int size;	/* the size of the allocated buffer */
	unsigned int in;	/* data is added at offset in */
	unsigned int out;	/* data is extracted from offset out */

	char buffer[0];
};

#ifdef __KERNEL__
#include <linux/kernel.h>
#include <linux/atomic.h>

struct cfifo {
	struct cfifo_shm *shm;	/* pointer to shared buffer */
	unsigned int size;	/* the size of the shared buffer payload */
	unsigned int lastin;	/* temporary offset in until message is
				   commited, i.e. can be read */
	unsigned int lastout;	/* temporary offset out until message is
				   processed, i.e. can be overwritten */
	int processed;		/* message processed */

	atomic_t refs;		/* users of the cfifo */

	int allocated;		/* shared buffer was not provided from the caller */
	unsigned int alloc_size;
	dma_addr_t phys;	/* physical address shared buffer */
	struct list_head node;	/* anchor for global cfifo list */
	struct device *dev;	/* for allocation from reserved memory area */
};

struct cfifo *cfifo_alloc(struct device *dev, unsigned int size, void *buffer, dma_addr_t phys);
void cfifo_ref(struct cfifo *fifo);
void cfifo_unref(struct cfifo *fifo);

static inline dma_addr_t cfifo_phys(struct cfifo *cfifo)
{
	return cfifo->phys;
}

static inline unsigned int cfifo_size(struct cfifo *cfifo)
{
	return cfifo->size;
}

int  cfifo_empty(struct cfifo *fifo);
void cfifo_reset(struct cfifo *fifo);

void *cfifo_request(struct cfifo *cfifo, unsigned int len);
int   cfifo_commit(struct cfifo *cfifo);
int   cfifo_get(struct cfifo *cfifo, void **buf);
void  cfifo_processed(struct cfifo *cfifo);

#endif /* __KERNEL__ */
#endif /* __COMA_CFIFO_H */
