/*
 * si3217x.c  --  Silicon Labs si3217x ALSA SoC Audio Codec driver
 *
 * Copyright (C) 2011 DSPG (DSP Group)
 *
 * Based on wm8978.c:
 *  Copyright (C) 2009-2010 Guennadi Liakhovetski <g.liakhovetski@gmx.de>
 *  Copyright (C) 2007 Carlos Munoz <carlos@kenati.com>
 *  Copyright 2006-2009 Wolfson Microelectronics PLC.
 *  Based on wm8974 and wm8990 by Liam Girdwood <lrg@slimlogic.co.uk>
 *
 * Adapted by Murali TD. Mohan <Murali.Mohan@dspg.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

//#define V_DEBUG
//#define DEBUG

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/pm.h>
#include <linux/spi/spi.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>
#include <sound/initval.h>
#include <sound/tlv.h>
#include <asm/div64.h>
#include <linux/gpio.h>
#include <linux/ioctl.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <linux/err.h>
#include <linux/list.h>
#include <linux/errno.h>
#include <linux/mutex.h>
#include <linux/of_gpio.h>

#include <linux/spi/spi.h>
#include <linux/spi/spidev.h>
#include <sound/si32178.h>
#include <asm/uaccess.h>


#ifdef DEBUG
        #define dbg_prt(fmt, ...) \
                pr_debug("Si3217x asoc Cdec-drv: %s(): " fmt, __func__, ##__VA_ARGS__)

        #define DIR_STR(stream_dir)  ((stream_dir == 0) ? "for Playback" : "for Capture")
#else
        #define dbg_prt(fmt, ...)
        #define DIR_SRT(dir)
#endif


/*** SPI USER SPACE INTERFACE *************************************************/

static DECLARE_BITMAP(minors, N_SPI_MINORS);
static LIST_HEAD(device_list);
static DEFINE_MUTEX(device_list_lock);

static unsigned bufsiz = 4096;

/*
 * We can't use the standard synchronous wrappers for file I/O; we
 * need to protect against async removal of the underlying spi_device.
 */
static void si3217x_spidev_complete(void *arg)
{
	complete(arg);
}

static ssize_t
si3217x_spidev_sync(struct si3217x_spidev_data *si3217x_spidev, struct spi_message *message)
{
	DECLARE_COMPLETION_ONSTACK(done);
	int status;

	message->complete = si3217x_spidev_complete;
	message->context = &done;

	spin_lock_irq(&si3217x_spidev->spi_lock);
	
	if (si3217x_spidev->spi == NULL)
		status = -ESHUTDOWN;
	else
		status = spi_async(si3217x_spidev->spi, message);

	spin_unlock_irq(&si3217x_spidev->spi_lock);

	if (status == 0) {
		wait_for_completion(&done);
		status = message->status;
		if (status == 0)
			status = message->actual_length;
	}

	return status;
}

static inline ssize_t
si3217x_spidev_sync_write(struct si3217x_spidev_data *si3217x_spidev, size_t len)
{
	struct spi_transfer t = {
		.tx_buf=si3217x_spidev->buffer,
		.len=len,
	};
	struct spi_message m;

	spi_message_init(&m);
	spi_message_add_tail(&t, &m);

	return si3217x_spidev_sync(si3217x_spidev, &m);
}

static inline ssize_t
si3217x_spidev_sync_read(struct si3217x_spidev_data *si3217x_spidev, size_t len)
{
	struct spi_transfer t = {
			.rx_buf	= si3217x_spidev->buffer,
			.len	= len,
	};
	struct spi_message m;

	spi_message_init(&m);
	spi_message_add_tail(&t, &m);

	return si3217x_spidev_sync(si3217x_spidev, &m);
}

/* Read-only message with current device setup */
static ssize_t
si3217x_spidev_read(struct file *filp, char __user *buf, size_t count, loff_t *f_pos)
{
	struct si3217x_spidev_data *si3217x_spidev;
	ssize_t status = 0;

	/* chipselect only toggles at start or end of operation */
	if (count > bufsiz)
		return -EMSGSIZE;

	si3217x_spidev = filp->private_data;

	mutex_lock(&si3217x_spidev->buf_lock);
	status = si3217x_spidev_sync_read(si3217x_spidev, count);
	if (status > 0) {
		unsigned long missing;

		missing = copy_to_user(buf, si3217x_spidev->buffer, status);
		if (missing == status) {
			status = -EFAULT;
		} else {
			status = status - missing;
		}
	}
	mutex_unlock(&si3217x_spidev->buf_lock);

	return status;
}

/* Write-only message with current device setup */
static ssize_t
si3217x_spidev_write(struct file *filp, const char __user *buf,
			size_t count, loff_t *f_pos)
{
	struct si3217x_spidev_data *si3217x_spidev;
	ssize_t status = 0;
	unsigned long missing;

	/* chipselect only toggles at start or end of operation */
	if (count > bufsiz)
		return -EMSGSIZE;

	si3217x_spidev = filp->private_data;

	mutex_lock(&si3217x_spidev->buf_lock);
	missing = copy_from_user(si3217x_spidev->buffer, buf, count);
	if (missing == 0)
		status = si3217x_spidev_sync_write(si3217x_spidev, count);
	else
		status = -EFAULT;
	mutex_unlock(&si3217x_spidev->buf_lock);

	return status;
}

static int si3217x_spidev_message(struct si3217x_spidev_data *si3217x_spidev,
				  struct spi_ioc_transfer *u_xfers,
				  unsigned n_xfers)
{
	struct spi_message msg;
	struct spi_transfer *k_xfers;
	struct spi_transfer *k_tmp;
	struct spi_ioc_transfer *u_tmp;
	unsigned int n, total;
	u8 *buf;
	int status = -EFAULT;

	spi_message_init(&msg);
	k_xfers = kcalloc(n_xfers, sizeof(*k_tmp), GFP_KERNEL);
	if (k_xfers == NULL)
		return -ENOMEM;

	/* Construct spi_message, copying any tx data to bounce buffer.
	* We walk the array of user-provided transfers, using each one
	* to initialize a kernel version of the same transfer.
	*/
	buf = si3217x_spidev->buffer;
	total = 0;
	for (n = n_xfers, k_tmp = k_xfers, u_tmp = u_xfers;
	     n;
	     n--, k_tmp++, u_tmp++) {

		k_tmp->len = u_tmp->len;

		total += k_tmp->len;
		if (total > bufsiz) {
			status = -EMSGSIZE;
			goto done;
		}

		if (u_tmp->rx_buf) {
			k_tmp->rx_buf = buf;
			if (!access_ok( VERIFY_WRITE,
					(u8 __user *)(uintptr_t)u_tmp->rx_buf,
					u_tmp->len )) {
				goto done;
			}
		}
		if (u_tmp->tx_buf) {
			k_tmp->tx_buf = buf;
			if (copy_from_user( buf,
					    (const u8 __user *)(uintptr_t)u_tmp->tx_buf,
					    u_tmp->len )) {
				goto done;
			}
		}
		buf += k_tmp->len;

		k_tmp->cs_change = !!u_tmp->cs_change;
		k_tmp->bits_per_word = u_tmp->bits_per_word;
		k_tmp->delay_usecs = u_tmp->delay_usecs;
		k_tmp->speed_hz = u_tmp->speed_hz;

		spi_message_add_tail(k_tmp, &msg);
	}

	status = si3217x_spidev_sync(si3217x_spidev, &msg);
	if (status < 0) {
		goto done;
	}

	/* copy any rx data out of bounce buffer */
	buf = si3217x_spidev->buffer;
	for (n = n_xfers, u_tmp = u_xfers; n; n--, u_tmp++) {
		if (u_tmp->rx_buf) {
			if (__copy_to_user( (u8 __user *)(uintptr_t)u_tmp->rx_buf,
					     buf,
					     u_tmp->len )) {
				status = -EFAULT;
				goto done;
			}
		}
		buf += u_tmp->len;
	}
	status = total;

done:
	kfree(k_xfers);
	return status;
}

static long
si3217x_spidev_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	int err = 0;
	int retval = 0;
	struct si3217x_spidev_data *si3217x_spidev;
	struct spi_device *spi;
	u32 tmp;
	unsigned int n_ioc;
	struct spi_ioc_transfer *ioc;
	struct si3217x_priv *si3217x;

	/* Check type and command number */
	if (_IOC_TYPE(cmd) != SPI_IOC_MAGIC)
		return -ENOTTY;

	/* Check access direction once here; don't repeat below.
	* IOC_DIR is from the user perspective, while access_ok is
	* from the kernel perspective; so they look reversed.
	*/
	if (_IOC_DIR(cmd) & _IOC_READ) {
		err = !access_ok(VERIFY_WRITE, (void __user *)arg,
					_IOC_SIZE(cmd));
	}
	
	if (err == 0 && (_IOC_DIR(cmd) & _IOC_WRITE)) {
		err = !access_ok(VERIFY_READ,(void __user *)arg,
					_IOC_SIZE(cmd));
	}
	
	if (err)
		return -EFAULT;

	/* guard against device removal before, or while,
	* we issue this ioctl.
	*/
	si3217x_spidev = filp->private_data;

	spin_lock_irq(&si3217x_spidev->spi_lock);

	spi = spi_dev_get(si3217x_spidev->spi);
	si3217x = spi_get_drvdata(spi);

	spin_unlock_irq(&si3217x_spidev->spi_lock);

	if (spi == NULL)
		return -ESHUTDOWN;

	/* use the buffer lock here for triple duty:
	*  - prevent I/O (from us) so calling spi_setup() is safe;
	*  - prevent concurrent SPI_IOC_WR_* from morphing
	*    data fields while SPI_IOC_RD_* reads them;
	*  - SPI_IOC_MESSAGE needs the buffer locked "normally".
	*/
	mutex_lock(&si3217x_spidev->buf_lock);

	switch (cmd) {
		/* read requests */
		case SPI_IOC_RD_MODE:
			retval = __put_user(spi->mode & SPI_MODE_MASK,
						(__u8 __user *)arg);
		break;
		case SPI_IOC_RD_LSB_FIRST:
			retval = __put_user((spi->mode & SPI_LSB_FIRST) ?  1 : 0,
						(__u8 __user *)arg);
		break;
		case SPI_IOC_RD_BITS_PER_WORD:
			retval = __put_user(spi->bits_per_word,
						(__u8 __user *)arg);
		break;
		case SPI_IOC_RD_MAX_SPEED_HZ:
			retval = __put_user(spi->max_speed_hz,
					(__u32 __user *)arg);
		break;

		/* write requests */
		case SPI_IOC_WR_MODE:
			retval = __get_user(tmp, (u8 __user *)arg);
			if (retval == 0) {
				u8  save = spi->mode;

				if (tmp & ~SPI_MODE_MASK) {
					retval = -EINVAL;
					break;
				}

				tmp |= spi->mode & ~SPI_MODE_MASK;
				spi->mode = (u8)tmp;
				retval = spi_setup(spi);
				if (retval < 0) {
					spi->mode = save;
				} else {
					dev_dbg(&spi->dev,
						"spi mode %02x\n", tmp);
				}
			}
		break;
		case SPI_IOC_WR_LSB_FIRST:
			retval = __get_user(tmp, (__u8 __user *)arg);
			if (retval == 0) {
				u8  save = spi->mode;

				if (tmp) {
					spi->mode |= SPI_LSB_FIRST;
				} else {
					spi->mode &= ~SPI_LSB_FIRST;
				}
				retval = spi_setup(spi);
				if (retval < 0) {
					spi->mode = save;
				} else {
					dev_dbg(&spi->dev, "%csb first\n",
						tmp ? 'l' : 'm');
				}
			}
		break;
		case SPI_IOC_WR_BITS_PER_WORD:
			retval = __get_user(tmp, (__u8 __user *)arg);
			if (retval == 0) {
				u8  save = spi->bits_per_word;

				spi->bits_per_word = tmp;
				retval = spi_setup(spi);
				if (retval < 0) {
					spi->bits_per_word = save;
				} else {
					dev_dbg(&spi->dev,
						"%d bits per word\n", tmp);
				}
			}
		break;
		case SPI_IOC_WR_MAX_SPEED_HZ:
			retval = __get_user(tmp, (__u32 __user *)arg);
			if (retval == 0) {
				u32 save = spi->max_speed_hz;

				spi->max_speed_hz = tmp;
				retval = spi_setup(spi);
				if (retval < 0) {
					spi->max_speed_hz = save;
				} else {
					dev_dbg(&spi->dev, "%d Hz (max)\n", tmp);
				}
			}
		break;

		case si3217x_IOC_RESET:
			si3217x_reset(si3217x->si3217x_reset);
		break;

		default:
			/* segmented and/or full-duplex I/O request */
			if ( (_IOC_NR(cmd) != _IOC_NR(SPI_IOC_MESSAGE(0))) ||
			     (_IOC_DIR(cmd) != _IOC_WRITE) ) {
				retval = -ENOTTY;
				break;
			}

			tmp = _IOC_SIZE(cmd);
			if ((tmp % sizeof(struct spi_ioc_transfer)) != 0) {
				retval = -EINVAL;
				break;
			}
			n_ioc = tmp / sizeof(struct spi_ioc_transfer);
			if (n_ioc == 0) {
				break;
			}

			/* copy into scratch area */
			ioc = kmalloc(tmp, GFP_KERNEL);
			if (!ioc) {
				retval = -ENOMEM;
				break;
			}
			if (__copy_from_user(ioc, (void __user *)arg, tmp)) {
				kfree(ioc);
				retval = -EFAULT;
				break;
			}

			/* translate to spi_message, execute */
			retval = si3217x_spidev_message(si3217x_spidev, ioc, n_ioc);
			kfree(ioc);
		break;
	}

	mutex_unlock(&si3217x_spidev->buf_lock);

	spi_dev_put(spi);

	return retval;
}

static int si3217x_spidev_open(struct inode *inode, struct file *filp)
{
	struct si3217x_spidev_data *si3217x_spidev;
	int status = -ENXIO;

	mutex_lock(&device_list_lock);

	list_for_each_entry(si3217x_spidev, &device_list, device_entry) {
		if (si3217x_spidev->devt == inode->i_rdev) {
			status = 0;
			break;
		}
	}

	if (status == 0) {
		if (!si3217x_spidev->buffer) {
			si3217x_spidev->buffer = kmalloc(bufsiz, GFP_KERNEL);
			if (!si3217x_spidev->buffer) {
				dev_dbg(&si3217x_spidev->spi->dev,
					"open/ENOMEM\n");
				status = -ENOMEM;
			}
		}
		if (status == 0) {
			si3217x_spidev->users++;
			filp->private_data = si3217x_spidev;
			nonseekable_open(inode, filp);
		}
	} else {
		pr_debug("si3217x_spidev: nothing for minor %d\n",
			 iminor(inode));
	}

	mutex_unlock(&device_list_lock);

	return status;
}

static int si3217x_spidev_release(struct inode *inode, struct file *filp)
{
	struct si3217x_spidev_data *si3217x_spidev;
	int status = 0;

	mutex_lock(&device_list_lock);

	si3217x_spidev = filp->private_data;
	filp->private_data = NULL;

	/* last close? */
	si3217x_spidev->users--;
	if (!si3217x_spidev->users) {
		int     dofree;

		kfree(si3217x_spidev->buffer);
		si3217x_spidev->buffer = NULL;

		/* ... after we unbound from the underlying device? */
		spin_lock_irq(&si3217x_spidev->spi_lock);
		dofree = (si3217x_spidev->spi == NULL);
		spin_unlock_irq(&si3217x_spidev->spi_lock);

		if (dofree)
			kfree(si3217x_spidev);
	}

	mutex_unlock(&device_list_lock);

	return status;
}

static const struct file_operations si3217x_spidev_fops = {
	.owner		= THIS_MODULE,
	/* REVISIT switch to aio primitives, so that userspace
	* gets more complete API coverage. It'll simplify things
	* too, except for the locking.
	*/
	.write		= si3217x_spidev_write,
	.read		= si3217x_spidev_read,
	.unlocked_ioctl	= si3217x_spidev_ioctl,
	.open		= si3217x_spidev_open,
	.release	= si3217x_spidev_release,
};

/*---------------------------------------------------------------------------*/

/* The main reason to have this class is to make mdev/udev create the
 * /dev/si3217x_spidevB.C character device nodes exposing our userspace API.
 * It also simplifies memory management.
 */

static struct class *si3217x_spidev_class;

int si3217x_spidev_probe(struct spi_device *spi)
{
	struct si3217x_spidev_data *si3217x_spidev;
	int status;
	unsigned long minor;
	
	/* Allocate driver data */
	si3217x_spidev = kzalloc(sizeof(*si3217x_spidev), GFP_KERNEL);
	if (!si3217x_spidev) {
		return -ENOMEM;
	}

	/* Initialize the driver data */
	si3217x_spidev->spi = spi;
	spin_lock_init(&si3217x_spidev->spi_lock);
	mutex_init(&si3217x_spidev->buf_lock);

	INIT_LIST_HEAD(&si3217x_spidev->device_entry);

	/* If we can allocate a minor number, hook up this device.
	 * Reusing minors is fine so long as udev or mdev is working.
	 */
	mutex_lock(&device_list_lock);

	minor = find_first_zero_bit(minors, N_SPI_MINORS);

	if (minor < N_SPI_MINORS) {
		struct device *dev;

		si3217x_spidev->devt = MKDEV(si3217x_SPIDEV_MAJOR, minor);
		dev = device_create(si3217x_spidev_class, &spi->dev,
				    si3217x_spidev->devt, si3217x_spidev,
				    "si3217xspi%d.%d", spi->master->bus_num,
				    spi->chip_select);
		status = IS_ERR(dev) ? PTR_ERR(dev) : 0;
	} else {
		dev_dbg(&spi->dev, "no minor number available!\n");
		status = -ENODEV;
	}

	if (status == 0) {
		set_bit(minor, minors);
		list_add(&si3217x_spidev->device_entry, &device_list);
	}

	mutex_unlock(&device_list_lock);

	if (status == 0)
		spi_set_drvdata(spi, si3217x_spidev);
	else
		kfree(si3217x_spidev);

	return status;
}

int si3217x_spidev_remove(struct spi_device *spi)
{
	struct si3217x_spidev_data *si3217x_spidev = spi_get_drvdata(spi);

	/* make sure ops on existing fds can abort cleanly */
	spin_lock_irq(&si3217x_spidev->spi_lock);

	si3217x_spidev->spi = NULL;
	spi_set_drvdata(spi, NULL);

	spin_unlock_irq(&si3217x_spidev->spi_lock);

	/* prevent new opens */
	mutex_lock(&device_list_lock);

	list_del(&si3217x_spidev->device_entry);
	device_destroy(si3217x_spidev_class, si3217x_spidev->devt);
	clear_bit(MINOR(si3217x_spidev->devt), minors);
	if (si3217x_spidev->users == 0)
		kfree(si3217x_spidev);

	mutex_unlock(&device_list_lock);

	return 0;
}

int si3217x_spidev_init(void)
{
	int status;

	/* Claim our 256 reserved device numbers.  Then register a class
	 * that will key udev/mdev to add/remove /dev nodes.  Last, register
	 * the driver which manages those device numbers.
	 */
	BUILD_BUG_ON(N_SPI_MINORS > 256);
	status = register_chrdev(si3217x_SPIDEV_MAJOR, "spi", &si3217x_spidev_fops);
	if (status < 0)
		return status;

	si3217x_spidev_class = class_create(THIS_MODULE, "si3217xspi");
	if (IS_ERR(si3217x_spidev_class)) {
		unregister_chrdev(si3217x_SPIDEV_MAJOR, "si3217xspi");
		return PTR_ERR(si3217x_spidev_class);
	}

	return status;
}

void si3217x_spidev_exit(void)
{
	class_destroy(si3217x_spidev_class);
	unregister_chrdev(si3217x_SPIDEV_MAJOR, "si3217xspi");
}

/******************* END OF SPI USER SPACE INTERFACE **************************/

static int
si3217x_dai_set_fmt(struct snd_soc_dai *codec_dai, unsigned int fmt)
{
	//struct snd_soc_codec *codec = codec_dai->codec;

	dbg_prt();

	/* Set Master/Slave */
	switch (fmt & SND_SOC_DAIFMT_MASTER_MASK) {
	case SND_SOC_DAIFMT_CBM_CFM:
		dbg_prt("%s\n", "si3217x is TDM Master");
		break;
	case SND_SOC_DAIFMT_CBS_CFS:
		dbg_prt("%s\n", "si3217x is TDM Slave");
		break;
	default:
		return -EINVAL;
	}

	/* Set Format */
	switch (fmt & SND_SOC_DAIFMT_FORMAT_MASK) {
	case SND_SOC_DAIFMT_I2S:
		dbg_prt("%s\n", "TDM format is I2S");
		break;
	case SND_SOC_DAIFMT_RIGHT_J:
		dbg_prt("%s\n", "TDM format is Right-Justified");
		break;
	case SND_SOC_DAIFMT_LEFT_J:
		dbg_prt("%s\n", "TDM format is Left-Justified");
		break;
	case SND_SOC_DAIFMT_DSP_B:
		dbg_prt("%s\n", "TDM format is L data MSB during FRM LRC");
		break;
	default:
		return -EINVAL;
	}

	/* Set TDM clocks Polarity */
	switch (fmt & SND_SOC_DAIFMT_INV_MASK) {
	case SND_SOC_DAIFMT_NB_NF:
		dbg_prt("%s\n", "TDM clocks polarity: normal BCLK(sclk); normal LRC(FSync)");
		break;
	case SND_SOC_DAIFMT_IB_IF:
		dbg_prt("%s\n", "TDM clocks polarity: inverted BCLK(sclk); inverted LRC(FSync)");
		break;
	case SND_SOC_DAIFMT_IB_NF:
		dbg_prt("%s\n", "TDM clocks polarity: inverted BCLK(sclk); normal LRC(FSync)");
		break;
	case SND_SOC_DAIFMT_NB_IF:
		dbg_prt("%s\n", "TDM clocks polarity: normal BCLK(sclk); inverted LRC(FSync)");
		break;
        default:
                return -EINVAL;
        }

        return 0;
}

static int
si3217x_dai_hw_params(struct snd_pcm_substream *substream,
                      struct snd_pcm_hw_params *params,
                      struct snd_soc_dai *dai)
{
        //struct snd_soc_pcm_runtime *rtd = substream->private_data;
        //struct snd_soc_device *socdev = rtd->socdev;
        //struct snd_soc_codec *codec = socdev->card->codec;
        //struct csr8811_priv *csr8811 = snd_soc_codec_get_drvdata(codec);

        dbg_prt("%s\n", DIR_STR(substream->stream));

        /* Bit width */
        switch (params_format(params)) {
        case SNDRV_PCM_FORMAT_S8:
                dbg_prt("%s\n", "PCM sample-length (vs. ALSA) = 8 bits");
                break;
        case SNDRV_PCM_FORMAT_S16_LE:
                dbg_prt("%s\n", "PCM sample-length (vs. ALSA) = 16 bits; LE");
                break;
        default:
                return -EINVAL;
        }

        /* Filter coefficient (per sample-rate) */
        switch (params_rate(params)) {
        case 8000:
                dbg_prt("%s\n", "fs = 8KHz");
                break;
        case 48000:
                dbg_prt("%s\n", "fs = 48KHz");
                break;
        case 11025:
        case 12000:
        case 16000:
        case 22050:
        case 24000:
        case 32000:
        case 44100:
        default:
                dbg_prt("%s %u %s\n",
                         "Unsupported fs:", (params_rate(params)) / 1000, "KHz");
                return -EINVAL;
        }

        return 0;
}

static struct snd_soc_dai_ops si3217x_dai_ops = {
        .set_fmt        = si3217x_dai_set_fmt,
        .hw_params      = si3217x_dai_hw_params,
};

static struct snd_soc_codec_driver soc_codec_dev_si3217x = {
	.set_bias_level	=	NULL,
	.reg_cache_size	=	0,
	.reg_word_size	=	0,
	.reg_cache_default =	0,
};

struct snd_soc_dai_driver si3217x_dai_drv = {
	.name = "si32178",
	.playback = {
		.stream_name	= "Playback",
		.channels_min	= 1,
		.channels_max	= 2,	/* si3217x has Single output-path */
		.rates		= SNDRV_PCM_RATE_8000, /* A.M. was 16KHz */
		.formats	= si3217x_FORMATS,
	},
	.capture = {
		.stream_name	= "Capture",
		.channels_min	= 1,
		.channels_max	= 2,    /* si3217x has Single input-path */
		.rates		= SNDRV_PCM_RATE_8000, /* A.M. was 16KHz */
		.formats	= si3217x_FORMATS,
	},
	.ops = &si3217x_dai_ops,
};

void si3217x_reset(unsigned int gpio)
{
	gpio_direction_output(gpio, 1);
	udelay(100);
	gpio_set_value(gpio, 0);
	udelay(1000);
	gpio_set_value(gpio, 1);
	udelay(1000);
}

/* Must keep this private pointer in the local scope of the file so 
   get_si3217x_priv() can correctly return it to the bsp driver of the slic  
*/
static struct si3217x_priv * ptr_si3217x[NUM_SPI_BANKS] = { 0, 0};

static int si3217x_spi_probe(struct spi_device *spi)
{
	struct si3217x_priv * si3217x = NULL;
	struct snd_soc_codec *codec;
	struct device_node *np;
	int reset_gpio, ret;
#if 0
	int dev_count, i;
#endif
	unsigned int id;

	np = spi->dev.of_node;
	if (np == NULL)
		dev_info(&spi->dev, "of node is null\n");

	/* Get the reset line associated wtih the devices on this SPI bank */
	if ((reset_gpio = of_get_named_gpio(np, "rst-gpio", 0)) < 0) {
		dev_err(&spi->dev, "Reset value not specified in devtree\n");
		return -ENODEV;
	}

	/* Read the SPI id and verify it before moving on */
	ret = of_property_read_u32(np, "id", &id);
	if (ret && ret != -EINVAL) {
		dev_err(&spi->dev, "Error parsing id value in devtree\n");
		return -ENODEV;
	}
	if ( id >= NUM_SPI_BANKS ) {
		printk("Driver support only %d SPI banks. Must increase NUM_SPI_BANKS macro\n", NUM_SPI_BANKS);
		return -ENODEV;
	}

	si3217x = kzalloc(sizeof(struct si3217x_priv), GFP_KERNEL);
	if (si3217x == NULL)
		return -ENOMEM;

	codec = &si3217x->codec;
	codec->dev = &spi->dev;

	si3217x->si3217x_reset = reset_gpio;
	si3217x->si3217x_id = id;

#if 0
	ret = of_property_read_u32(np, "nr_devices", &dev_count);
	if (ret && ret != -EINVAL) {
		dev_err(&spi->dev, "Number of spi devices not specified in devtree\n");
		goto err_si3217x;
	}

	if ( dev_count > IRQ_SILABS_LINES )
		printk("Found %d SPI devices on 0x%08p\n", dev_count, spi);

	si3217x->si3217x_nr_devices = dev_count;

	for (i=0; i < IRQ_SILABS_LINES; i++)
	{
		/* Read interrupt from devtree */
		si3217x->si3217x_int_gpio[i] = of_get_named_gpio(np, "int-gpio", i);
		if (gpio_is_valid( si3217x->si3217x_int_gpio[i] ))
		{
			gpio_request( si3217x->si3217x_int_gpio[i], "gs-si3050-int");
			gpio_direction_input( si3217x->si3217x_int_gpio[i] );
			/* Map GPIO to extirq */
			si3217x->si3217x_irq[i] = gpio_to_irq( si3217x->si3217x_int_gpio[i] );
		}

		printk("local_int_data[%d].int_gpio/irq = %d/%d\n", i, si3217x->si3217x_int_gpio[i], si3217x->si3217x_irq[i]);
	}
#endif

	spi_set_drvdata(spi, si3217x);
	printk("spi_dev[%d] = 0x%8p\n", id, spi);
	ptr_si3217x[id] = si3217x;
	printk("ptr_si3217x[%d] = 0x%8p\n", id, ptr_si3217x[id]);
	/*
	* setup spi parameters; this makes sure that parameters we request
	* are acceptable by the spi driver
	*/
	spi->mode = SPI_MODE_3; /* clk active high */
	spi->bits_per_word = 8;

	ret = spi_setup(spi);
	if (ret < 0) {
		dev_err(&spi->dev, "spi_setup() failed\n");
		goto err_si3217x;
	}

	si3217x->spi = spi;

	ret = gpio_request(si3217x->si3217x_reset, "si3217x reset");
	if (ret < 0) {
		dev_err(&spi->dev, "failed requesting reset gpio\n");
		goto err_si3217x;
	}

	/* Resetting the SPI devices will be done by the low level drivers  */
//	si3217x_reset(si3217x->si3217x_reset);

	si3217x_spidev_probe(spi);

	dev_set_name(&spi->dev, "%s.%u", "si3217x-daa", si3217x->si3217x_id);

	dev_info(&spi->dev, "successfully registered\n");

	ret = snd_soc_register_codec(&spi->dev, &soc_codec_dev_si3217x,
				     &si3217x_dai_drv, 1);
	if (ret < 0) {
		dev_err(&spi->dev, "failed registering soc codec\n");
		goto err_si3217x;
	}

	return ret;

err_si3217x:
	kfree(si3217x);
	return ret;
}

/* Helper function to allow low level BSP driver finish all the setups */
struct si3217x_priv * get_si3217x_priv(int spi_dev_id)
{
	printk("calling get_si3217x_priv reutrn spi_dev[%d] = 0x%8p\n", spi_dev_id, ptr_si3217x[spi_dev_id]);

	return ptr_si3217x[spi_dev_id];
}
EXPORT_SYMBOL(get_si3217x_priv);

static int __devexit si3217x_spi_remove(struct spi_device *spi)
{
	struct si3217x_priv *si3217x = spi_get_drvdata(spi);

	snd_soc_unregister_codec(&spi->dev);
	si3217x_spidev_remove(spi);

	kfree(si3217x);

	return 0;
}

static const struct of_device_id dspg_si32178_dt_match[] = {
	{ .compatible = "dspg,dvf_si32178_spi", },
	{}
};

static struct spi_driver dvf_si32178_spi_driver = {
	.driver = {
		.name	= "dvf_si32178_spi",
		.bus	= &spi_bus_type,
		.owner	= THIS_MODULE,
		.of_match_table = dspg_si32178_dt_match,
	},
	.probe	= si3217x_spi_probe,
	.remove	= __exit_p(si3217x_spi_remove),
};

static int __init si3217x_module_init(void)
{
	si3217x_spidev_init();
	return spi_register_driver(&dvf_si32178_spi_driver);
}

static __devexit void si3217x_module_exit(void)
{
	si3217x_spidev_exit();
	spi_unregister_driver(&dvf_si32178_spi_driver);
}

module_init(si3217x_module_init);
module_exit(si3217x_module_exit);

MODULE_DESCRIPTION("ASoC si3217x D.A.A. codec driver");
MODULE_AUTHOR("DSP Group");
MODULE_LICENSE("GPL");
