/*
 * css-codec.c --  css codec dai for creating a  dai link
 *
 * Copyright (C) 2012 DSPG (DSP Group)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/pm.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include <linux/mfd/dvf9a.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/initval.h>
#include <sound/tlv.h>
#include <asm/div64.h>
#include <linux/of.h>
#include "alsa-dvf9a.h"

struct css_priv {
	struct snd_soc_codec *codec;
};

static const char *dvf99_css_input_select1[] = {
	"SINGIN0", "SINGIN1", "DIFFIN0", "DIFFIN1",
	"SINGIN0-SINGIN1", "VREF-OFFSET", "MUTE-VREF", "MUTE-FLOAT"
};

static const char *dvf99_css_input_select[] = {
	"SINGIN0", "SINGIN1", "DIFFIN0", "DIFFIN1",
	"SINGIN0-SINGIN1", "DAC0", "MUTE-VREF", "MUTE-FLOAT"
};

static const char *dvf99_css_output_select1[] = {
	"DAC0", "DAC1", "DAC0_INV", "DAC1_INV", "SINGIN1",
	"DIFFIN0_INV", "DIFFIN1_INV", "MUTE"
};

static const char *dvf99_css_output_select[] = {
	"DAC0", "DAC1", "DAC0_INV", "DAC1_INV", "SINGIN0",
	"DIFFIN0", "DIFFIN1", "MUTE"
};

static const char *dvf99_css_mute_type[] = {
	"NO_MUTE", "MUTE1", "MUTE2", "MUTE3"
};

static const char *dvf99_css_lat_type[] = {
	"FALLING_EDGE", "LOW_PHASE", "RISING_EDGE", "HIGH_PHASE"
};

static const char *dvf99_css_amp_current[] = {
	"50%", "66%" , "100%", "130%"
};

static const char *dvf99_css_adc_current[] = {
	"22%", "26%", "40%", "52%", "32%", "40%", "133%", "30%",
	 "36%", "66%", "103%", "38%", "45%", "100%", "200%"
};

static const char *dvf99_css_vref_level[] = {
	"1.5V", "1.65V", "1.0V", "1.8V", "1.2V", "AUX_MUX", 
	"VCCAMPOUT/2", "VCCA/2"
};

static const char *dvf99_css_out_amp_ref [] = {
	"GLOBAL_AFE_REF", "VCCAMPOUT/2", 
};

static const char *dvf99_css_switch[] = {
	"off" , "on"
};

static const char *dvf99_css_dac_curr_scale [] = {
	"-6dB", "-3dB", "-2dB", "-1dB", "0dB", 
	"+1dB", "+2dB", "+3dB"
};

static const struct soc_enum dvf99_css_input_enum[] = {
	SOC_ENUM_SINGLE(DAIF_AFEIN1, 0, 8, dvf99_css_input_select),
	SOC_ENUM_SINGLE(DAIF_AFEIN1, 4, 8, dvf99_css_input_select1),
	SOC_ENUM_SINGLE(DAIF_AFEIN2, 0, 8, dvf99_css_input_select1),
};

static const struct soc_enum dvf99_css_output_enum[] = {
	SOC_ENUM_SINGLE(DAIF_AMPOUTSEL1, 0, 8, dvf99_css_output_select),
	SOC_ENUM_SINGLE(DAIF_AMPOUTSEL1, 4, 8, dvf99_css_output_select1),
	SOC_ENUM_SINGLE(DAIF_AMPOUTSEL2, 0, 8, dvf99_css_output_select),
	SOC_ENUM_SINGLE(DAIF_AMPOUTSEL2, 4, 8, dvf99_css_output_select1),
};

static const struct soc_enum dvf99_css_dac_mute_enum[] = {
	SOC_ENUM_SINGLE(DAIF_DACMUTE,0, 3, dvf99_css_mute_type),
	SOC_ENUM_SINGLE(DAIF_DACMUTE,3, 3, dvf99_css_mute_type)
};

static const struct soc_enum dvf99_css_dac_latency_enum[] = {
	SOC_ENUM_SINGLE(DAIF_LQSD_CFG2, 3, 3, dvf99_css_lat_type),
	SOC_ENUM_SINGLE(DAIF_LQSD_CFG2, 5, 3, dvf99_css_lat_type)
};

static const struct soc_enum dvf99_css_amp_current_enum[] = {
	SOC_ENUM_SINGLE(DAIF_AFEICTRL1, 0, 3, dvf99_css_amp_current),
	SOC_ENUM_SINGLE(DAIF_AFEICTRL1, 2, 3, dvf99_css_amp_current),
	SOC_ENUM_SINGLE(DAIF_AFEICTRL1, 4, 3, dvf99_css_amp_current),
	SOC_ENUM_SINGLE(DAIF_AFEICTRL1, 6, 3, dvf99_css_amp_current),
	SOC_ENUM_SINGLE(DAIF_AFEICTRL1, 8, 3, dvf99_css_amp_current),
	SOC_ENUM_SINGLE(DAIF_AFEICTRL1, 10, 3, dvf99_css_amp_current),
	SOC_ENUM_SINGLE(DAIF_AFEICTRL1, 12, 3, dvf99_css_amp_current),
	SOC_ENUM_SINGLE(DAIF_AFEICTRL1, 14, 3, dvf99_css_amp_current),
	SOC_ENUM_SINGLE(DAIF_AFEICTRL2, 12, 3, dvf99_css_amp_current),
	SOC_ENUM_SINGLE(DAIF_AFEICTRL2, 14, 3, dvf99_css_amp_current),
	SOC_ENUM_SINGLE(DAIF_AFEICTRL2, 0, 0xf, dvf99_css_adc_current),
	SOC_ENUM_SINGLE(DAIF_AFEICTRL2, 4, 0xf, dvf99_css_adc_current),
	SOC_ENUM_SINGLE(DAIF_AFEICTRL2, 8, 0xf, dvf99_css_adc_current)
};

static const struct soc_enum dvf99_css_vref_level_enum[] = {
	SOC_ENUM_SINGLE(DAIF_AFECFG1, 4 , 7, dvf99_css_vref_level)
};

static const struct soc_enum dvf99_css_out_amp_ref_enum [] = { 
	SOC_ENUM_SINGLE(DAIF_AFECFG2, 0 , 2 ,dvf99_css_out_amp_ref)
};


static const struct soc_enum dvf99_css_vref_to_aux_enum[] = { 
	SOC_ENUM_SINGLE(DAIF_AFECFG2, 3, 2, dvf99_css_switch)
};

static const struct soc_enum dvf99_css_low_imp_amp_enum = 
	SOC_ENUM_SINGLE(DAIF_AFECFG2, 3, 2, dvf99_css_switch);

static const struct soc_enum dvf99_css_vref_current_enum = 
	SOC_ENUM_SINGLE(DAIF_AFECFG2, 2, 4, dvf99_css_amp_current);

static const struct soc_enum dvf99_css_30db_att_enum [] = {
	SOC_ENUM_SINGLE(DAIF_AFEOUTGC0, 3, 2, dvf99_css_switch),
	SOC_ENUM_SINGLE(DAIF_AFEOUTGC0, 7, 2, dvf99_css_switch),
	SOC_ENUM_SINGLE(DAIF_AFEOUTGC1, 3, 2, dvf99_css_switch),
	SOC_ENUM_SINGLE(DAIF_AFEOUTGC1, 7, 2, dvf99_css_switch)
};

static const struct soc_enum dvf99_css_dac_curr_scale_enum[] = {
	SOC_ENUM_SINGLE(DAIF_DAC0GC0, 0, 8, dvf99_css_dac_curr_scale),
	SOC_ENUM_SINGLE(DAIF_DAC0GC0, 4, 8, dvf99_css_dac_curr_scale)
};

static const DECLARE_TLV_DB_SCALE(input_tlv,  -400,  200,  0);
static const DECLARE_TLV_DB_SCALE(output_single_ended_tlv,  0,  -600,  0);
static const DECLARE_TLV_DB_SCALE(output_differential_tlv,  600,  -600,  0);

/* ASoC Controls */
static const struct snd_kcontrol_new dvf99_css_snd_controls[] = {
	/* Input slect enum */
	SOC_DAPM_ENUM("Input Mux0", dvf99_css_input_enum[0]),
	SOC_DAPM_ENUM("Input Mux1", dvf99_css_input_enum[1]),
	SOC_DAPM_ENUM("Output Mux0", dvf99_css_output_enum[0]),
	SOC_DAPM_ENUM("Output Mux1", dvf99_css_output_enum[1]),
	SOC_DAPM_ENUM("Output Mux2", dvf99_css_output_enum[2]),
	SOC_DAPM_ENUM("Output Mux3", dvf99_css_output_enum[3]),
};

static const struct snd_kcontrol_new dvf99_css_snd_kcontrols[] = {
	SOC_SINGLE_TLV("DIFFIN0 Gain", DAIF_AFEDIGC1, 0, 28, 0, input_tlv),
	SOC_SINGLE_TLV("DIFFIN1 Gain", DAIF_AFEDIGC2, 0, 28, 0, input_tlv),
	SOC_SINGLE_TLV("SINGIN0 Gain", DAIF_AFESIGC1, 0, 28, 0, input_tlv),
	SOC_SINGLE_TLV("SINGIN1 Gain", DAIF_AFESIGC2, 0, 28, 0, input_tlv),
	SOC_SINGLE_TLV("Output0_sing Gain", DAIF_AFEOUTGC0, 0, 6, 0,
			 output_single_ended_tlv),
	SOC_SINGLE_TLV("Output1_sing Gain", DAIF_AFEOUTGC0, 4, 6, 0,
			 output_single_ended_tlv),
	SOC_SINGLE_TLV("Output2_sing Gain", DAIF_AFEOUTGC1, 0, 6, 0,
			 output_single_ended_tlv),
	SOC_SINGLE_TLV("Output3_sing Gain", DAIF_AFEOUTGC1, 4, 6, 0,
			 output_single_ended_tlv),
	SOC_SINGLE_TLV("Output0_diff Gain", DAIF_AFEOUTGC0, 0, 6, 0,
			 output_single_ended_tlv),
	SOC_SINGLE_TLV("Output1_diff Gain", DAIF_AFEOUTGC0, 4, 6, 0,
			 output_single_ended_tlv),
	SOC_SINGLE_TLV("Output2_diff Gain", DAIF_AFEOUTGC1, 0, 6, 0,
			 output_single_ended_tlv),
	SOC_SINGLE_TLV("Output3_diff Gain", DAIF_AFEOUTGC1, 4, 6, 0,
			 output_single_ended_tlv),

	SOC_ENUM("DAC0 MuteType", dvf99_css_dac_mute_enum[0]),
	SOC_ENUM("DAC1 MuteType", dvf99_css_dac_mute_enum[0]),
	SOC_ENUM("DIFFIN0_ICTRL", dvf99_css_amp_current_enum[0]),
	SOC_ENUM("DIFFIN1_ICTRL", dvf99_css_amp_current_enum[1]),
	SOC_ENUM("SINGIN0_ICTRL", dvf99_css_amp_current_enum[2]),
	SOC_ENUM("SINGIN1_ICTRL", dvf99_css_amp_current_enum[3]),
	SOC_ENUM("AMPOUT0_ICTRL", dvf99_css_amp_current_enum[4]),
	SOC_ENUM("AMPOUT1_ICTRL", dvf99_css_amp_current_enum[5]),
	SOC_ENUM("AMPOUT2_ICTRL", dvf99_css_amp_current_enum[6]),
	SOC_ENUM("AMPOUT3_ICTRL", dvf99_css_amp_current_enum[7]),
	SOC_ENUM("DAC0_ICTRL", dvf99_css_amp_current_enum[8]),
	SOC_ENUM("DAC1_ICTRL", dvf99_css_amp_current_enum[9]),
	SOC_ENUM("ADC0_ICTRL", dvf99_css_amp_current_enum[10]),
	SOC_ENUM("ADC1_ICTRL", dvf99_css_amp_current_enum[11]),
	SOC_ENUM("ADC2_ICTRL", dvf99_css_amp_current_enum[12]),
	SOC_ENUM("VREF_LVL", dvf99_css_vref_level_enum[0]),
	SOC_ENUM("AMPOUT_VREF_SEL", dvf99_css_out_amp_ref_enum[0]),
	SOC_ENUM("VREF_TO_AUX_EN", dvf99_css_vref_to_aux_enum[0]),
	SOC_ENUM("LOW_IMP_AMP_EN", dvf99_css_low_imp_amp_enum),
	SOC_ENUM("VREF_ICTRL", dvf99_css_vref_current_enum),
	SOC_ENUM("AMPOUT0_30DBATT", dvf99_css_30db_att_enum[0]),
	SOC_ENUM("AMPOUT1_30DBATT", dvf99_css_30db_att_enum[1]),
	SOC_ENUM("AMPOUT2_30DBATT", dvf99_css_30db_att_enum[2]),
	SOC_ENUM("AMPOUT3_30DBATT", dvf99_css_30db_att_enum[3]),
	SOC_ENUM("DAC0_CURRENT_SCALE", dvf99_css_dac_curr_scale_enum[0]),
	SOC_ENUM("DAC1_CURRENT_SCALE", dvf99_css_dac_curr_scale_enum[1]),
	SOC_ENUM("DAC_LATENCY", dvf99_css_dac_latency_enum[0]),
	SOC_ENUM("ADC_LATENCY", dvf99_css_dac_latency_enum[1]),

	SOC_SINGLE("VREF_EN", DAIF_AFECFG1, 7, 1, 0), 
	SOC_SINGLE("DCLASS_EN", DAIF_DCLS_ANA_CTRL, 0, 1, 0), 
	SOC_SINGLE("DCLASS_SUP_MODE", DAIF_DCLS_ANA_CTRL, 6, 1, 0), 
	SOC_SINGLE("ULTRA_WB_EN", DAIF_LQSD_CFG2, 2, 1, 0),
	SOC_SINGLE("AFE_CLK_SRC_CRYSTAL_OSC", DAIF_LQSD_CFG2, 1, 1, 0),
	SOC_SINGLE("AFE_SAMPLING_864K", DAIF_LQSD_CFG2, 0, 1, 0),
	SOC_SINGLE("MIC0_SHORT_WIN_SIZE", DAIF_AFEMIC0_SHORT, 0 , 3, 0),
	SOC_SINGLE("MIC0_SHORT_DET_EN", DAIF_AFEMIC0_SHORT, 2 , 1, 0),
	SOC_SINGLE("MIC0_SHORT_DET_PROTECTION_EN", DAIF_AFEMIC0_SHORT, 3 , 1,0),
	SOC_SINGLE("MIC1_SHORT_WIN_SIZE", DAIF_AFEMIC1_SHORT, 0 , 3, 0),
	SOC_SINGLE("MIC1_SHORT_DET_EN", DAIF_AFEMIC1_SHORT, 2 , 1, 0),
	SOC_SINGLE("MIC1_SHORT_DET_PROTECTION_EN", DAIF_AFEMIC1_SHORT, 3 , 1,0),
	SOC_SINGLE("AMPOUT0_LOW_THD", DAIF_AMPOUTTHD1, 0 , 0xf, 0),
	SOC_SINGLE("AMPOUT1_LOW_THD", DAIF_AMPOUTTHD1, 4 , 0xf, 0),
	SOC_SINGLE("AMPOUT2_LOW_THD", DAIF_AMPOUTTHD2, 0 , 0xf, 0),
	SOC_SINGLE("AMPOUT3_LOW_THD", DAIF_AMPOUTTHD2, 4 , 0xf, 0),
	SOC_SINGLE("MICPWR0_EN", DAIF_AFECFG1, 1, 1,0),
	SOC_SINGLE("MICPWR1_EN", DAIF_AFECFG1, 3, 1,0),
	SOC_SINGLE("MICPWR0_CURR_LIM_EN", DAIF_AFECFG1, 0, 1,0),
	SOC_SINGLE("MICPWR1_CURR_LIM_EN", DAIF_AFECFG1, 2, 1,0),
	SOC_SINGLE("AFE_DITHER_ON", DAIF_LQSD_CFG1, 0, 1,0),
	SOC_SINGLE("MUTE_REQ_INT_EN", DAIF_LQSD_CFG1, 1, 1,0),
	SOC_SINGLE("MUTE_STAT_EN", DAIF_LQSD_CFG1, 2, 1,0),
	SOC_SINGLE("ADC2_EN", DAIF_LQSD_CFG1, 3, 1,0),
	SOC_SINGLE("DAC0_EN", DAIF_LQSD_CFG1, 4, 1,0),
	SOC_SINGLE("DAC1_EN", DAIF_LQSD_CFG1, 5, 1,0),
	SOC_SINGLE("ADC0_EN", DAIF_LQSD_CFG1, 6, 1,0),
	SOC_SINGLE("ADC1_EN", DAIF_LQSD_CFG1, 7, 1,0),
	SOC_SINGLE("AMPOUT0_EN", DAIF_AFEAMPEN, 0, 1,0),
	SOC_SINGLE("AMPOUT1_EN", DAIF_AFEAMPEN, 1, 1,0),
	SOC_SINGLE("AMPOUT2_EN", DAIF_AFEAMPEN, 2, 1,0),
	SOC_SINGLE("AMPOUT3_EN", DAIF_AFEAMPEN, 3, 1,0),
	SOC_SINGLE("SINGIN0_EN", DAIF_AFEAMPEN, 4, 1,0),
	SOC_SINGLE("SINGIN1_EN", DAIF_AFEAMPEN, 5, 1,0),
	SOC_SINGLE("DIFFIN0_EN", DAIF_AFEAMPEN, 6, 1,0),
	SOC_SINGLE("DIFFIN1_EN", DAIF_AFEAMPEN, 7, 1,0),
};

/* ASoC DAPM (Dynamic Audio Power Management) Widgets */
static const struct snd_soc_dapm_widget dvf99_css_dapm_widgets[] = {
	/* Input Widgets */
	SND_SOC_DAPM_INPUT("SINGIN0"),
	SND_SOC_DAPM_INPUT("SINGIN1"),

	/* two input differential - each has two pins! */
	SND_SOC_DAPM_INPUT("DIFFIN0P"),
	SND_SOC_DAPM_INPUT("DIFFIN0N"),
	SND_SOC_DAPM_INPUT("DIFFIN1P"),
	SND_SOC_DAPM_INPUT("DIFFIN1N"),

	SND_SOC_DAPM_INPUT("SINGIN0-SINGIN1_INP"),
	SND_SOC_DAPM_INPUT("VREF-OFFSET_INP"),
	SND_SOC_DAPM_INPUT("MUTE-VREF_INP"),
	SND_SOC_DAPM_INPUT("MUTE-FLOAT_INP"),
	
	/* Input to AMPOUt Mux */
	SND_SOC_DAPM_INPUT("DAC0_INV"),
	SND_SOC_DAPM_INPUT("DAC1_INV"),
	SND_SOC_DAPM_INPUT("DIFFIN0_INV"),
	SND_SOC_DAPM_INPUT("DIFFIN1_INV"),
	SND_SOC_DAPM_INPUT("MUTE"),
	
	/* Output Widgets */
	SND_SOC_DAPM_OUTPUT("AMPOUT0"),
	SND_SOC_DAPM_OUTPUT("AMPOUT1"),
	SND_SOC_DAPM_OUTPUT("AMPOUT2"),
	SND_SOC_DAPM_OUTPUT("AMPOUT3"),

	/* MicBias Widget */
	SND_SOC_DAPM_MICBIAS("Mic0 Bias", DAIF_AFECFG1, 1, 0),
	SND_SOC_DAPM_MICBIAS("Mic1 Bias", DAIF_AFECFG1, 3, 0),

	/* PGA Widgets */
	SND_SOC_DAPM_PGA("SINGIN0_PGA", DAIF_AFEAMPEN, 4, 0, NULL, 0),
	SND_SOC_DAPM_PGA("SINGIN1_PGA", DAIF_AFEAMPEN, 5, 0, NULL, 0),
	SND_SOC_DAPM_PGA("DIFFIN0_PGA", DAIF_AFEAMPEN, 6, 0, NULL, 0),
	SND_SOC_DAPM_PGA("DIFFIN1_PGA", DAIF_AFEAMPEN, 7, 0, NULL, 0),
	SND_SOC_DAPM_PGA("AMPOUT0_PGA", DAIF_AFEAMPEN, 0, 0, NULL, 0),
	SND_SOC_DAPM_PGA("AMPOUT1_PGA", DAIF_AFEAMPEN, 1, 0, NULL, 0),
	SND_SOC_DAPM_PGA("AMPOUT2_PGA", DAIF_AFEAMPEN, 2, 0, NULL, 0),
	SND_SOC_DAPM_PGA("AMPOUT3_PGA", DAIF_AFEAMPEN, 3, 0, NULL, 0),

	/* Mux widgets */
	SND_SOC_DAPM_MUX("Input Mux0", SND_SOC_NOPM, 0, 0,
			 &dvf99_css_snd_controls[0]),
	SND_SOC_DAPM_MUX("Input Mux1", SND_SOC_NOPM, 0, 0,
			 &dvf99_css_snd_controls[1]),
	SND_SOC_DAPM_MUX("Output Mux0", SND_SOC_NOPM, 0, 0,
			 &dvf99_css_snd_controls[2]),
	SND_SOC_DAPM_MUX("Output Mux1", SND_SOC_NOPM, 0, 0,
			 &dvf99_css_snd_controls[3]),
	SND_SOC_DAPM_MUX("Output Mux2", SND_SOC_NOPM, 0, 0,
			 &dvf99_css_snd_controls[4]),
	SND_SOC_DAPM_MUX("Output Mux3", SND_SOC_NOPM, 0, 0,
			 &dvf99_css_snd_controls[5]),

	/* ADC Widgets */
	SND_SOC_DAPM_ADC("ADC0", "CaptureFD", DAIF_LQSD_CFG1, 6, 0),
	
	/* DAC Widgets */
	SND_SOC_DAPM_DAC("DAC0", "CSS Playback", DAIF_LQSD_CFG1, 4, 0),
	SND_SOC_DAPM_DAC("DAC1", "CSS Playback2", DAIF_LQSD_CFG1, 5, 0),
	SND_SOC_DAPM_DAC("DCLASS", "DClassSpk", DAIF_DCLS_ANA_CTRL,0,0),
};

/* ASoC audio-route Map */
static const struct snd_soc_dapm_route dvf99_css_audio_route[] = {
	/* output routes */
	{"Output Mux0", "DAC0", "DAC0"},
	{"Output Mux0", "DAC1", "DAC1"},
	{"Output Mux0", "DAC0_INV", "DAC0"},
	{"Output Mux0", "DAC1_INV", "DAC1"},
	{"Output Mux0", "SINGIN0", "SINGIN0_PGA"},
	{"Output Mux0", "DIFFIN0", "DIFFIN0_PGA"},
	{"Output Mux0", "DIFFIN1", "DIFFIN1_PGA"},
	{"Output Mux0", "MUTE", "MUTE"},
	
	{"Output Mux1", "DAC0", "DAC0"},
	{"Output Mux1", "DAC1", "DAC1"},
	{"Output Mux1", "DAC0_INV", "DAC0"},
	{"Output Mux1", "DAC1_INV", "DAC1"},
	{"Output Mux1", "SINGIN1", "SINGIN1_PGA"},
	{"Output Mux1", "DIFFIN1_INV", "DIFFIN1_INV"},
	{"Output Mux1", "DIFFIN0_INV", "DIFFIN0_INV"},
	{"Output Mux1", "MUTE", "MUTE"},
	
	{"Output Mux2", "DAC0", "DAC0"},
	{"Output Mux2", "DAC1", "DAC1"},
	{"Output Mux2", "DAC0_INV", "DAC0"},
	{"Output Mux2", "DAC1_INV", "DAC1"},
	{"Output Mux2", "SINGIN0", "SINGIN0_PGA"},
	{"Output Mux2", "DIFFIN0", "DIFFIN0_PGA"},
	{"Output Mux2", "DIFFIN1", "DIFFIN1_PGA"},
	{"Output Mux2", "MUTE", "MUTE"},
	
	{"Output Mux3", "DAC0", "DAC0"},
	{"Output Mux3", "DAC1", "DAC1"},
	{"Output Mux3", "DAC0_INV", "DAC0"},
	{"Output Mux3", "DAC1_INV", "DAC1"},
	{"Output Mux3", "SINGIN1", "SINGIN1_PGA"},
	{"Output Mux3", "DIFFIN1_INV", "DIFFIN1_INV"},
	{"Output Mux3", "DIFFIN0_INV", "DIFFIN0_INV"},
	{"Output Mux3", "MUTE", "MUTE"},
	
	/* un conditionaly */
	{"AMPOUT0_PGA",  NULL,  "Output Mux0"},
	{"AMPOUT0",  NULL,  "AMPOUT0_PGA"},

	{"AMPOUT1_PGA",  NULL,  "Output Mux1"},
	{"AMPOUT1",  NULL,  "AMPOUT1_PGA"},

	{"AMPOUT2_PGA",  NULL,  "Output Mux2"},
	{"AMPOUT2",  NULL,  "AMPOUT2_PGA"},

	{"AMPOUT3_PGA",  NULL,  "Output Mux3"},
	{"AMPOUT3",  NULL,  "AMPOUT3_PGA"},
	
	/* input path */
	/* zivh - mic bias should be set from other power management ? */
	{"SINGIN0_PGA",  NULL ,  "SINGIN0"},
	{"SINGIN1_PGA",  NULL ,  "SINGIN1"},

	{"DIFFIN0_PGA",  NULL ,  "DIFFIN0P"},
	{"DIFFIN0_PGA",  NULL ,  "DIFFIN0N"},

	{"DIFFIN1_PGA",  NULL ,  "DIFFIN1P"},
	{"DIFFIN1_PGA",  NULL ,  "DIFFIN1N"},
	/* inputs to mixer are from the pga inputs */
	{"Input Mux0",  "SINGIN0",  "SINGIN0_PGA"},
	{"Input Mux0",  "SINGIN1",  "SINGIN1_PGA"},
	{"Input Mux0",  "DIFFIN0",  "DIFFIN0_PGA"},
	{"Input Mux0",  "DIFFIN1",  "DIFFIN1_PGA"},

	{"Input Mux0",  "SINGIN0-SINGIN1",  "SINGIN0-SINGIN1_INP"},
	{"Input Mux0",  "DAC0",  "DAC0"},
	{"Input Mux0",  "MUTE-VREF",  "MUTE-VREF_INP"},
	{"Input Mux0",  "MUTE-FLOAT",  "MUTE-FLOAT_INP"},
	{"Input Mux1",  "SINGIN0",  "SINGIN0_PGA"},
	{"Input Mux1",  "SINGIN1",  "SINGIN1_PGA"},
	{"Input Mux1",  "DIFFIN0",  "DIFFIN0_PGA"},
	{"Input Mux1",  "DIFFIN1",  "DIFFIN1_PGA"},

	{"Input Mux1",  "SINGIN0-SINGIN1",  "SINGIN0-SINGIN1_INP"},
	{"Input Mux1",  "VREF-OFFSET",  "VREF-OFFSET_INP"},

	{"Input Mux1",  "MUTE-VREF",  "MUTE-VREF_INP"},
	{"Input Mux1",  "MUTE-FLOAT",  "MUTE-FLOAT_INP"},
	{"SINGIN0-SINGIN1_INP" , NULL, "SINGIN0_PGA"},
	{"SINGIN0-SINGIN1_INP" , NULL, "SINGIN1_PGA"},
	

	{"ADC0",  NULL, "Input Mux0"},
	{"DAC0",  NULL, "PlaybackFD"},
};

static int
css_dai_hw_params(struct snd_pcm_substream *substream,
		  struct snd_pcm_hw_params *params,
		  struct snd_soc_dai *dai)
{
	/* nothing to do, mode switching will be handled with dua */
	return 0;
}

static struct snd_soc_dai_ops css_dai_ops = {
	.hw_params = css_dai_hw_params,
};

/* CSS Codec DAI: */
struct snd_soc_dai_driver css_dais[] = {
	{
		.name = "PCMHD_codec_dai",
		.playback = {
			.stream_name	= "PlaybackHD",
			.channels_min	= 1,
			.channels_max	= 1,
			.rates		= SNDRV_PCM_RATE_16000,
			.formats	= SNDRV_PCM_FMTBIT_S16_LE,
		},
		.ops = &css_dai_ops,
	},
	{
		.name = "PCMFD_codec_dai",
		.playback = {
			.stream_name	= "PlaybackFD",
			.channels_min	= 1,
			.channels_max	= 1,
			.rates		= SNDRV_PCM_RATE_16000,
			.formats	= SNDRV_PCM_FMTBIT_S16_LE,
		},
		.capture = {
			.stream_name	= "CaptureFD",
			.channels_min	= 1,
			.channels_max	= 1,
			.rates		= SNDRV_PCM_RATE_16000,
			.formats	= SNDRV_PCM_FMTBIT_S16_LE,
		},
		.ops = &css_dai_ops,
	},
	{
		.name = "TRACE1_codec_dai",
		.capture = {
			.stream_name	= "CaptureTrace1",
			.channels_min	= 1,
			.channels_max	= 3,
			.rates		= SNDRV_PCM_RATE_32000 |
					  SNDRV_PCM_RATE_16000 |
					  SNDRV_PCM_RATE_8000,
			.formats	= SNDRV_PCM_FMTBIT_S16_LE,
		},
		.ops = &css_dai_ops,
	},
	{
		.name = "USB_plug",
		.playback = {
			.stream_name	= "PlaybackUSB",
			.channels_min	= 1,
			.channels_max	= 2,
			.rates		= SNDRV_PCM_RATE_16000,
			.formats	= SNDRV_PCM_FMTBIT_S16_LE,
		},
		.capture = {
			.stream_name	= "CaptureUSB",
			.channels_min	= 1,
			.channels_max	= 2,
			.rates		= SNDRV_PCM_RATE_16000,
			.formats	= SNDRV_PCM_FMTBIT_S16_LE,
		},
		.ops = &css_dai_ops,
	},
	{
		.name = "TRACE2_codec_dai",
		.capture = {
			.stream_name	= "CaptureTrace2",
			.channels_min	= 1,
			.channels_max	= 3,
			.rates		= SNDRV_PCM_RATE_32000 |
					  SNDRV_PCM_RATE_16000 |
					  SNDRV_PCM_RATE_8000,
			.formats	= SNDRV_PCM_FMTBIT_S16_LE,
		},
		.ops = &css_dai_ops,
	},
	{
		.name = "ANYFD_codec_dai",
		.playback = {
			.stream_name	= "PlaybackANY",
			.channels_min	= 1,
			.channels_max	= 1,
			.rates		= SNDRV_PCM_RATE_64000 |
					  SNDRV_PCM_RATE_32000 |
					  SNDRV_PCM_RATE_16000 |
					  SNDRV_PCM_RATE_8000,
			.formats	= SNDRV_PCM_FMTBIT_S16_LE,
		},
		.capture = {
			.stream_name	= "CaptureANY",
			.channels_min	= 1,
			.channels_max	= 1,
			.rates		= SNDRV_PCM_RATE_64000 |
					  SNDRV_PCM_RATE_32000 |
					  SNDRV_PCM_RATE_16000 |
					  SNDRV_PCM_RATE_8000,
			.formats	= SNDRV_PCM_FMTBIT_S16_LE,
		},
		.ops = &css_dai_ops,
	},
	{
		.name = "PCMHD_DClass_codec_dai",
		.playback = {
			.stream_name	= "PlaybackHD_DClass",
			.channels_min	= 1,
			.channels_max	= 1,
			.rates		= SNDRV_PCM_RATE_16000,
			.formats	= SNDRV_PCM_FMTBIT_S16_LE,
		},
		.ops = &css_dai_ops,
	},
	{
		.name = "PCMFD_DClass_codec_dai",
		.playback = {
			.stream_name	= "DClassSpk",
			.channels_min	= 1,
			.channels_max	= 1,
			.rates		= SNDRV_PCM_RATE_8000_48000,
			.formats	= SNDRV_PCM_FMTBIT_S16_LE,
		},
		.ops = &css_dai_ops,
	},
};

static int
css_set_bias_level(struct snd_soc_codec *codec,
		   enum snd_soc_bias_level level)
{
	switch (level) {
	case SND_SOC_BIAS_ON:
		break;
	case SND_SOC_BIAS_PREPARE:
		break;
	case SND_SOC_BIAS_STANDBY:
		break;
	case SND_SOC_BIAS_OFF:
		break;
	default:
		return -EINVAL;
	}
	/* Change to new state */
	codec->dapm.bias_level = level;

	return 0;
}

static int
css_dev_probe(struct snd_soc_codec *codec)
{
	codec->control_data = NULL;
	css_set_bias_level(codec, SND_SOC_BIAS_STANDBY);
	dvf9a_alsa_daif_write(DAIF_AFEAMPEN, 0x0);
	return 0;
}

static int
css_dev_remove(struct snd_soc_codec *codec)
{
	css_set_bias_level(codec,  SND_SOC_BIAS_OFF);
	return 0;
}

static int
css_dev_suspend(struct snd_soc_codec *codec)
{
	css_set_bias_level(codec, SND_SOC_BIAS_OFF);
	return 0;
}

static int
css_dev_resume(struct snd_soc_codec *codec)
{
	css_set_bias_level(codec, SND_SOC_BIAS_STANDBY);
	return 0;
}

static unsigned int
css_dev_read(struct snd_soc_codec *codec,  unsigned int reg)
{
	return dvf9a_alsa_daif_read(reg);
}

static int
css_dev_write(struct snd_soc_codec *codec,  unsigned int reg, unsigned int val)
{
    // Bug 36235 (false EHS off-hook event): always enable MICPWR0 block and current limit.
    if (reg == DAIF_AFECFG1)
        val = val | 0x03;
    
	dvf9a_alsa_daif_write(reg, val);
	return 0;
}

static struct snd_soc_codec_driver soc_codec_dev_css = {
	.probe   = css_dev_probe,
	.remove  = css_dev_remove,
	.suspend = css_dev_suspend,
	.resume  = css_dev_resume,
	.set_bias_level = css_set_bias_level,
	.read = css_dev_read,
	.write = css_dev_write,
	.controls = dvf99_css_snd_kcontrols,
	.num_controls = ARRAY_SIZE(dvf99_css_snd_kcontrols),
	.dapm_widgets = dvf99_css_dapm_widgets,
	.num_dapm_widgets = ARRAY_SIZE(dvf99_css_dapm_widgets),
	.dapm_routes = dvf99_css_audio_route,
	.num_dapm_routes = ARRAY_SIZE(dvf99_css_audio_route),
	/* No codec registers ! */
	.reg_cache_size = 0,
	.reg_word_size = 0,
	.reg_cache_default = NULL,
	.ignore_pmdown_time = true,
};

static __devinit int
css_codec_register(struct css_priv *css, struct platform_device *pdev)
{
	int ret = 0;

	BUG_ON(css->codec != NULL);

	css->codec = kzalloc(sizeof(*(css->codec)),  GFP_KERNEL);
	if (!css->codec)
		return -ENOMEM;

	css->codec->dev = &(pdev->dev);

	css->codec->name = "CSS";
	snd_soc_codec_set_drvdata(css->codec,  css);

	/* Register the codec */
	ret = snd_soc_register_codec(css->codec->dev,
				     &soc_codec_dev_css,
				     css_dais,
				     ARRAY_SIZE(css_dais));
	if (ret != 0) {
		dev_err(css->codec->dev,
			"Failed to register Codec and its Dai: %d\n",  ret);
		kfree(css->codec);
	}

	return 0;
}

static __devexit void
css_codec_unregister(struct css_priv *css)
{
	css_set_bias_level(css->codec,  SND_SOC_BIAS_OFF);
	snd_soc_unregister_codec(css->codec->dev);
}

static __devinit int
css_codec_platform_probe(struct platform_device *pdev)
{
	int ret;
	struct css_priv *css;

	css = kzalloc(sizeof(struct css_priv),  GFP_KERNEL);
	if (!css)
		return -ENOMEM;

	pdev->dev.platform_data = css;

	ret = css_codec_register(css, pdev);
	if (ret)
		kfree(css);

	return ret;
}

static int __devexit
css_codec_platform_remove(struct platform_device *pdev)
{
	struct css_priv *css = pdev->dev.platform_data;

	css_codec_unregister(css);
	kfree(css);

	return 0;
}

static struct of_device_id snd_css_of_ids[] = {
	{ .compatible = "dspg,dvf99_css_codec" },
	{ },
};

static struct platform_driver css_codec_platform_driver = {
	.driver = {
		.name	= "dvf99_css_codec",
		.owner	= THIS_MODULE,
		.of_match_table = snd_css_of_ids,
	},
	.probe		= css_codec_platform_probe,
	.remove		= __devexit_p(css_codec_platform_remove),
	.shutdown	= NULL,
	.suspend	= NULL,
	.resume		= NULL,
};

static int __init
css_codec_mod_Init(void)
{
	return platform_driver_register(&css_codec_platform_driver);
}
module_init(css_codec_mod_Init);

static void __exit
css_codec_mod_Exit(void)
{
	platform_driver_unregister(&css_codec_platform_driver);
}
module_exit(css_codec_mod_Exit);

MODULE_DESCRIPTION("ASoC css codec driver");
MODULE_AUTHOR("Santosh Kumar <Sanotsh.kumar@dspg.com>");
MODULE_LICENSE("GPL");
