/*
 * snd-dv99-pcm.c -- DVF99 ASoC platform driver
 *
 *  Copyright (C) 2011 DSPG Technologies GmbH
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/clk.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/spinlock.h>
#include <linux/vmalloc.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <linux/of.h>
#include <linux/dma-mapping.h>
#include <linux/coma/coma-alsa.h>

#define DRV_NAME "dvf99-snd-soc-platform"

/* defaults */
#define MAX_BUFFER_SIZE		(32*1020) /* full divide by 6 */
#define MIN_PERIOD_SIZE		64
#define MAX_PERIOD_SIZE		MAX_BUFFER_SIZE/4
#define USE_FORMATS		(SNDRV_PCM_FMTBIT_U8 | SNDRV_PCM_FMTBIT_S16_LE)
#define USE_RATE		SNDRV_PCM_RATE_CONTINUOUS | \
				SNDRV_PCM_RATE_8000_96000
#define USE_RATE_MIN		5500
#define USE_RATE_MAX		64000
#define USE_CHANNELS_MIN	1
#define USE_CHANNELS_MAX	3
#define USE_PERIODS_MIN		1
#define USE_PERIODS_MAX		1024
#define MAX_COMA_ALSA_DEVICE_ID 6

struct css_coma_alsa_pcm {
	ktime_t base_time;
	ktime_t period_time;
	atomic_t running;
	uint8_t id;
	u32 period;
	struct snd_pcm_substream *substream;
};

struct css_model {
	const char *name;
	int (*playback_constraints)(struct snd_pcm_runtime *runtime);
	int (*capture_constraints)(struct snd_pcm_runtime *runtime);
	u64 formats;
	size_t buffer_bytes_max;
	size_t period_bytes_min;
	size_t period_bytes_max;
	unsigned int periods_min;
	unsigned int periods_max;
	unsigned int rates;
	unsigned int rate_min;
	unsigned int rate_max;
	unsigned int channels_min;
	unsigned int channels_max;
};

struct snd_css {
	struct snd_soc_card *card;
	struct css_model *model;
	struct snd_pcm_hardware pcm_hw;
};

static struct snd_pcm_hardware css_pcm_hardware = {
	.info =			(SNDRV_PCM_INFO_MMAP |
				 SNDRV_PCM_INFO_INTERLEAVED |
				 SNDRV_PCM_INFO_RESUME |
				 SNDRV_PCM_INFO_MMAP_VALID |
				 SNDRV_PCM_INFO_BATCH),
	.formats =		USE_FORMATS,
	.rates =		USE_RATE,
	.rate_min =		USE_RATE_MIN,
	.rate_max =		USE_RATE_MAX,
	.channels_min =		USE_CHANNELS_MIN,
	.channels_max =		USE_CHANNELS_MAX,
	.buffer_bytes_max =	MAX_BUFFER_SIZE,
	.period_bytes_min =	MIN_PERIOD_SIZE,
	.period_bytes_max =	MAX_PERIOD_SIZE,
	.periods_min =		USE_PERIODS_MIN,
	.periods_max =		USE_PERIODS_MAX,
	.fifo_size =		0,
};

static int
dvf99_get_css_device_id(int device)
{
	return device % MAX_COMA_ALSA_DEVICE_ID ;
}

static void
css_coma_alsa_callback(unsigned long int data)
{
	struct css_coma_alsa_pcm *dpcm = (struct css_coma_alsa_pcm *)data;

	if (atomic_read(&dpcm->running)) {
		snd_pcm_period_elapsed(dpcm->substream);
	}
}

static int
css_coma_alsa_start(struct snd_pcm_substream *substream)
{
	struct css_coma_alsa_pcm *dpcm = substream->runtime->private_data;
	coma_alsa_start(dpcm->id, substream->stream);
	return 0;
}

static int
css_coma_alsa_stop(struct snd_pcm_substream *substream)
{
	struct css_coma_alsa_pcm *dpcm = substream->runtime->private_data;
	coma_alsa_stop(dpcm->id, substream->stream);
	return 0;
}

static int
dvf99_pcm_open(struct snd_pcm_substream *substream)
{
	struct css_coma_alsa_pcm *dpcm;
	struct snd_pcm_runtime *runtime = substream->runtime;

	dpcm = kzalloc(sizeof(*dpcm), GFP_KERNEL);
	if (!dpcm)
		return -ENOMEM;
	runtime->private_data = dpcm;
	dpcm->substream = substream;
	snd_soc_set_runtime_hwparams(substream,&css_pcm_hardware);

	atomic_set(&(dpcm->running), 1);

	return 0;
}

static int
dvf99_pcm_close(struct snd_pcm_substream *substream)
{
	struct css_coma_alsa_pcm *dpcm = substream->runtime->private_data;

	atomic_set(&(dpcm->running), 0);

	coma_alsa_deregister_elapsed_cb(dpcm->id, substream->stream);
	coma_alsa_close(dpcm->id, substream->stream);
	kfree(dpcm);

	return 0;
}

static int
dvf99_pcm_hw_params(struct snd_pcm_substream *substream,
		    struct snd_pcm_hw_params *hw_params)
{
	struct snd_pcm_runtime *runtime = substream->runtime;

	snd_pcm_set_runtime_buffer(substream, &substream->dma_buffer);	
	runtime->dma_bytes   = params_buffer_bytes(hw_params);
	runtime->buffer_size = params_buffer_size(hw_params);

	return 0;
}

static int
dvf99_pcm_prepare(struct snd_pcm_substream *substream)
{
	struct snd_pcm_runtime *runtime = substream->runtime;
	struct css_coma_alsa_pcm *dpcm = runtime->private_data;
	unsigned int period,period_bytes, rate,format,channels ,direction,size_bytes;
	long unsigned int address , physc;
	long sec;
	unsigned long nsecs;
	int ret = 0;

	period = runtime->period_size;
	period_bytes = frames_to_bytes(runtime, runtime->period_size);
	rate = runtime->rate;
	sec = period / rate;
	period %= rate;
	nsecs = div_u64((u64)period * 1000000000UL + rate - 1, rate);
	dpcm->id = dvf99_get_css_device_id(substream->pcm->device);
	format =runtime->sample_bits;
	channels = runtime->channels;
	address = (long unsigned int)runtime->dma_area;
	physc = runtime->dma_addr;
	size_bytes = snd_pcm_lib_buffer_bytes(substream);

	dpcm->period = period;
	direction = substream->stream;
	if (direction == SNDRV_PCM_STREAM_PLAYBACK)
		ret = coma_alsa_open(dpcm->id, rate, format, period_bytes,
				     channels, 0, 0, (void *)physc, size_bytes+4);
	else
		ret = coma_alsa_open(dpcm->id, rate, format, period_bytes,
				     channels, (void *)physc, size_bytes+4, 0, 0);

	/* if already open just reset */
	if (ret == -2)
		coma_alsa_reset(dpcm->id, substream->stream);

	if (ret != -2 && ret < 0) {
		pr_err("css_coma_alsa_prepare coma alsa open failed (%d)\n",ret);
		return ret;
	}

	dpcm->substream = substream;
	coma_alsa_register_elapsed_cb(dpcm->id,
				      direction,
				      css_coma_alsa_callback,
				      (unsigned long int)dpcm);

	return 0;
}

static int
dvf99_pcm_trigger(struct snd_pcm_substream *substream, int cmd)
{
	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
	case SNDRV_PCM_TRIGGER_RESUME:
		return css_coma_alsa_start(substream);
	case SNDRV_PCM_TRIGGER_STOP:
	case SNDRV_PCM_TRIGGER_SUSPEND:
		return css_coma_alsa_stop(substream);
	}

	return -EINVAL;
}

static snd_pcm_uframes_t
dvf99_pcm_pointer(struct snd_pcm_substream *substream)
{
	struct snd_pcm_runtime *runtime = substream->runtime;
	unsigned int size = snd_pcm_lib_buffer_bytes(substream);
	u32 pos;

	pos = *(volatile u32 *)&(runtime->dma_area[size]);

	return bytes_to_frames(substream->runtime, pos);
}

static struct snd_pcm_ops dvf99_pcm_ops = {
	.open		= dvf99_pcm_open,
	.close		= dvf99_pcm_close,
	.ioctl		= snd_pcm_lib_ioctl,
	.hw_params	= dvf99_pcm_hw_params,
	.prepare	= dvf99_pcm_prepare,
	.trigger	= dvf99_pcm_trigger,
	.pointer	= dvf99_pcm_pointer,
};

static int
css_pcm_preallocate_dma_buffer(struct snd_pcm *pcm, int stream)
{
	struct snd_pcm_substream *substream = pcm->streams[stream].substream;
	struct snd_dma_buffer *buf = &substream->dma_buffer;
	size_t size = MAX_BUFFER_SIZE;

	buf->dev.type = SNDRV_DMA_TYPE_DEV;
	buf->dev.dev = pcm->card->dev;
	buf->private_data = NULL;
	buf->area = dma_alloc_coherent(pcm->card->dev,
				       size + 4, /* + 4 bytes for position index */
				       &buf->addr,
				       GFP_KERNEL);
	if (!buf->area) {
		pr_err("Failed to allocate dma memory -\
			 Please increase uncached DMA memory region\n");
		return -ENOMEM;
	}
	buf->bytes = size;

	return 0;
}

static int
dvf99_pcm_probe(struct snd_soc_platform *pt)
{
	struct snd_css *css;

	css = kzalloc(sizeof(*css),GFP_KERNEL);
	if(!css)
		return -ENOMEM;
	css->card = pt->card;
	css->pcm_hw = css_pcm_hardware;
	snd_soc_platform_set_drvdata(pt, css);

	return 0;
}

static int
dvf99_pcm_remove(struct snd_soc_platform *pt)
{
	struct snd_css *css;

	css = snd_soc_platform_get_drvdata(pt);
	kfree(css);

	return 0;
}

static int
dvf99_pcm_new(struct snd_soc_pcm_runtime *runtime)
{
	struct snd_pcm *pcm;
	int ret = 0;

	pcm = runtime->pcm;
	if (pcm->streams[SNDRV_PCM_STREAM_PLAYBACK].substream) {
		ret = css_pcm_preallocate_dma_buffer(pcm,
			SNDRV_PCM_STREAM_PLAYBACK);
		if (ret)
			goto out;
	}

	if (pcm->streams[SNDRV_PCM_STREAM_CAPTURE].substream) {
		ret = css_pcm_preallocate_dma_buffer(pcm,
			SNDRV_PCM_STREAM_CAPTURE);
		if (ret)
			goto out;
	}
out:
	return ret;
}

static void
dvf99_pcm_free(struct snd_pcm *pcm)
{
	struct snd_pcm_substream *substream;
	struct snd_dma_buffer *buf;
	int stream;

	for (stream = 0; stream < 2; stream++) {
		substream = pcm->streams[stream].substream;
		if (!substream)
			continue;

		buf = &substream->dma_buffer;
		if (!buf->area)
			continue;
		dma_free_coherent(pcm->card->dev,
				  buf->bytes + 4, /* + 4 bytes for position index */
				  (void *)buf->area,
				  buf->addr);
		buf->area = NULL;
	}
}

static struct snd_soc_platform_driver dvf99_soc_platform = {
	.probe		= &dvf99_pcm_probe,
	.remove		= &dvf99_pcm_remove,
	.ops		= &dvf99_pcm_ops,
	.pcm_new	= dvf99_pcm_new,
	.pcm_free	= dvf99_pcm_free,
};

static int __devinit
dvf99_pcm_platform_probe(struct platform_device *pdev)
{
	int err;

	err = snd_soc_register_platform(&pdev->dev, &dvf99_soc_platform);
	if (err)
		printk("snd_soc_register_platform() failed");

	return err;
}

static int __devexit
dvf99_pcm_platform_remove(struct platform_device *pdev)
{
	snd_soc_unregister_platform(&pdev->dev);

	return 0;
}

static struct of_device_id snd_soc_platform_of_ids[] = {
	{ .compatible = "dspg,dvf99-snd-soc-platform" },
	{ },
};

static struct platform_driver dvf99_pcm_driver = {
	.driver = {
		.name = DRV_NAME,
		.owner = THIS_MODULE,
		.of_match_table = snd_soc_platform_of_ids,
	},
	.probe = dvf99_pcm_platform_probe,
	.remove = __devexit_p(dvf99_pcm_platform_remove),
};

static int __init
snd_dvf99_pcm_init(void)
{
	return platform_driver_register(&dvf99_pcm_driver);
}
module_init(snd_dvf99_pcm_init);

static void __exit
snd_dvf99_pcm_exit(void)
{
	platform_driver_unregister(&dvf99_pcm_driver);
}
module_exit(snd_dvf99_pcm_exit);

MODULE_DESCRIPTION("DVF99 ASoC platform driver");
MODULE_AUTHOR("DSP Group");
MODULE_LICENSE("GPL");
MODULE_ALIAS("platform:" DRV_NAME);
MODULE_AUTHOR("Santosh Kumar, DSP Group");
MODULE_SUPPORTED_DEVICE("{{ALSA,css soundcard}}");
