/*
 * snd-dvf99.c -- ASoC Machine Driver for DVF99
 *
 *  Copyright (C) 2011 DSPG Technologies GmbH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/clk.h>
#include <linux/io.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <mach/platform.h>
#include <mach/gpio.h>
#include <asm/atomic.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>
#include <linux/of.h>

#define DRIVER_NAME "snd-dvf99-mach-drv"
#define CODEC_NAME "dvf99_css_codec"
#define PLATFORM_DEV_NAME "dvf99-snd-soc-platform"

#define DUMMY_FORMATS   (SNDRV_PCM_FMTBIT_S8 | SNDRV_PCM_FMTBIT_S16_LE)
#define DUMMY_RATES     (SNDRV_PCM_RATE_8000_96000)

static int board_dai_init(struct snd_soc_pcm_runtime *rtd);

static struct snd_soc_dai_link board_dvf99_dai_link[] = {
	{
		.name = "dvf99_dai_link.1",
		.stream_name = "PCMHD",
		/* asoc Platform driver (PCM) node to be populated later
		   during run time */
		/* asoc Codec device node to be populated later */
		/* asoc Cpu-Dai  device name */
		.cpu_dai_name = "CSS",
		/* asoc Codec-Dai device name */
		.codec_dai_name = "PCMHD_codec_dai",
		.init = board_dai_init,
	},

	{
		.name = "dvf99_dai_link.2",
		.stream_name = "PCMFD",
		/* asoc Cpu-Dai  device name */
		.cpu_dai_name = "CSS",
		/* asoc Codec-Dai device name */
		.codec_dai_name = "PCMFD_codec_dai",
	},

	{
		.name = "dvf99_dai_link.3",
		.stream_name = "TRACE1",
		/* asoc Cpu-Dai  device name */
		.cpu_dai_name = "CSS",
		/* asoc Codec-Dai device name */
		.codec_dai_name = "TRACE1_codec_dai",
	},

	{
		.name = "dvf99_dai_link.4",
		.stream_name = "CSS_USB_plug",
		/* asoc Cpu-Dai  device name */
		.cpu_dai_name = "CSS",
		/* asoc Codec-Dai device name */
		.codec_dai_name = "USB_plug",
	},

	{
		.name = "dvf99_dai_link.5",
		.stream_name = "TRACE2",
		/* asoc Cpu-Dai  device name */
		.cpu_dai_name = "CSS",
		/* asoc Codec-Dai device name */
		.codec_dai_name = "TRACE2_codec_dai",
	},

	{
		.name = "dvf99_dai_link.6",
		.stream_name = "ANYFD",
		/* asoc Cpu-Dai  device name */
		.cpu_dai_name = "CSS",
		/* asoc Codec-Dai device name */
		.codec_dai_name = "ANYFD_codec_dai",
	},
	{
		.name = "dvf99_dai_link.7",
		.stream_name = "PCMHD_DCLASS",
		/* asoc Cpu-Dai  device name */
		.cpu_dai_name = "CSS",
		/* asoc Codec-Dai device name */
		.codec_dai_name = "PCMHD_DClass_codec_dai",
	},
	{
		.name = "dvf99_dai_link.8",
		.stream_name = "PCMFD_DCLASS",
		/* asoc Cpu-Dai  device name */
		.cpu_dai_name = "CSS",
		/* asoc Codec-Dai device name */
		.codec_dai_name = "PCMFD_DClass_codec_dai",
	},

};

static int
board_dai_init(struct snd_soc_pcm_runtime *rtd)
{
	int ret;
	struct snd_soc_codec *codec = rtd->codec;
	struct snd_soc_dapm_context *dapm = &codec->dapm;

	ret = snd_soc_dapm_new_controls(dapm, board_dapm_widgets,
	                                ARRAY_SIZE(board_dapm_widgets));
	if (ret) {
		printk(KERN_ERR "Dapm widgets adding failed\n");
		return ret;
	}

	ret = snd_soc_dapm_add_routes(dapm, board_audio_route,
	                              ARRAY_SIZE(board_audio_route));
	if (ret) {
		printk(KERN_ERR "Dapm routes adding failed\n");
		return ret;
	}

	ret = snd_soc_dapm_sync(dapm);
	if (ret) {
		printk(KERN_ERR "Dapm syncing failed\n");
		return ret;
	}

	return 0;
}

static struct snd_soc_card dspg_dvf99_css_card = {
	.name		= "dspg-dvf99-css",
	.dai_link	= board_dvf99_dai_link,  /* DAI-link Array(!) */
	.num_links	= ARRAY_SIZE(board_dvf99_dai_link),
	/* Note: new to ASoC:
	 * Card's (not Codec's !) set_bias_level().
	 * May be used for certain activities, e.g. diasbling
	 * card's power amplifiers or enabling them just upon BIAS_ON;
	 * this is if they are Not declared as card's DAPM Widgests,
	 * as part of the audio Routes.
	 */
	.set_bias_level		= NULL,
	.set_bias_level_post	= NULL,
};

static int
dspg_init_dai_link(struct snd_soc_card *card)
{
	int nCnt;
	struct snd_soc_dai_link *dai_link;
	struct device_node *codec_node, *platform_node;

	codec_node = of_find_node_by_name(0, CODEC_NAME);
	if (!codec_node) {
		printk(KERN_ERR "Codec node not found\n");
		return -1;
	}

	platform_node = of_find_node_by_name(0, PLATFORM_DEV_NAME);
	if (!platform_node) {
		printk(KERN_ERR "Platform node not found\n");
		return -1;
	}

	for (nCnt = 0; nCnt < card->num_links; nCnt++) {
		dai_link = &card->dai_link[nCnt];
		dai_link->codec_of_node = codec_node;
		dai_link->platform_of_node = platform_node;
	}

	return 0;
}

static int
dspg_snd_probe(struct platform_device *pdev)
{
	int ret = 0;
	struct snd_soc_card *card = &dspg_dvf99_css_card;

	/* note: platform_set_drvdata() here saves pointer to the card's data
	 * on the device's 'struct device_private *p'->driver_data
	 */
	card->dev = &pdev->dev;
	if (dspg_init_dai_link(card) < 0) {
		dev_err(&pdev->dev, "dspg_init_dai_link failed\n");
		ret = -1;
		goto ERR_CLEAR;
	}

	/* Register ASoC sound Card */
	ret = snd_soc_register_card(card);
	if (ret) {
		dev_err(&pdev->dev, "snd_soc_register_card failed\n");
		goto ERR_CLEAR;
	}

	dev_info(&pdev->dev, "DVF99 ASoC card registered\n");

	return 0;

ERR_CLEAR:
	return ret;
}

static int
dspg_snd_remove(struct platform_device *pdev)
{
	struct snd_soc_card *card = platform_get_drvdata(pdev);

	snd_soc_unregister_card(card);

	return 0;
}

static struct of_device_id snd_css_of_ids[] = {
	{ .compatible = "dspg,snd-dvf99-mach-drv" },
	{ },
};

static struct platform_driver board_dvf99_css_snd_drv = {
	.driver = {
		.name = DRIVER_NAME,
		.owner = THIS_MODULE,
		.of_match_table = snd_css_of_ids,
		.pm = &snd_soc_pm_ops,
	},
	.probe = dspg_snd_probe,
	.remove = __devexit_p(dspg_snd_remove),
};

int
dvf99_snd_init(void)
{
	return platform_driver_register(&board_dvf99_css_snd_drv);
}
EXPORT_SYMBOL(dvf99_snd_init);

void
dvf99_snd_exit(void)
{
	platform_driver_unregister(&board_dvf99_css_snd_drv);
}
EXPORT_SYMBOL(dvf99_snd_exit);

static int __init
board_dvf99_mod_init(void)
{
	return 0;
}

static void __exit
board_dvf99_mod_exit(void)
{
}

module_init(board_dvf99_mod_init);
module_exit(board_dvf99_mod_exit);

MODULE_DESCRIPTION("ASoC machine driver for DSPG DVF99 board");
MODULE_AUTHOR("DSP Group");
MODULE_LICENSE("GPL");
