/*
 * ntpclient.c - NTP client
 *
 * Copyright 1997, 1999, 2000, 2003  Larry Doolittle  <larry@doolittle.boa.org>
 * Last hack: July 5, 2003
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License (Version 2,
 *  June 1991) as published by the Free Software Foundation.  At the
 *  time of writing, that license was published by the FSF with the URL
 *  http://www.gnu.org/copyleft/gpl.html, and is incorporated herein by
 *  reference.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  Possible future improvements:
 *      - Double check that the originate timestamp in the received packet
 *        corresponds to what we sent.
 *      - Verify that the return packet came from the host we think
 *        we're talking to.  Not necessarily useful since UDP packets
 *        are so easy to forge.
 *      - Write more documentation  :-(
 *
 *  Compile with -D_PRECISION_SIOCGSTAMP if your machine really has it.
 *  There are patches floating around to add this to Linux, but
 *  usually you only get an answer to the nearest jiffy.
 *  Hint for Linux hacker wannabes: look at the usage of get_fast_time()
 *  in net/core/dev.c, and its definition in kernel/time.c .
 *
 *  If the compile gives you any flak, check below in the section
 *  labelled "XXXX fixme - non-automatic build configuration".
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>     /* gethostbyname */
#include <arpa/inet.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>
#ifdef _PRECISION_SIOCGSTAMP
#include <sys/ioctl.h>
#endif
#include <syslog.h>
#include <libgen.h>

#define ENABLE_DEBUG

#define DEVICE_MAC "/proc/gxp/dev_info/dev_mac"

extern char *optarg;

/* XXXX fixme - non-automatic build configuration */
#ifdef linux
#include <sys/utsname.h>
#include <sys/time.h>
typedef u_int32_t __u32;
#include <sys/timex.h>
#else
#define main ntpclient
extern struct hostent *gethostbyname(const char *name);
extern int h_errno;
#define herror(hostname) \
	fprintf(stderr,"Error %d looking up hostname %s\n", h_errno,hostname)
typedef uint32_t __u32;
#endif

#define JAN_1970        0x83aa7e80      /* 2208988800 1970 - 1900 in seconds */
#define NTP_PORT (123)

/* How to multiply by 4294.967296 quickly (and not quite exactly)
 * without using floating point or greater than 32-bit integers.
 * If you want to fix the last 12 microseconds of error, add in
 * (2911*(x))>>28)
 */
#define NTPFRAC(x) ( 4294*(x) + ( (1981*(x))>>11 ) )

/* The reverse of the above, needed if we want to set our microsecond
 * clock (via settimeofday) based on the incoming time in NTP format.
 * Basically exact.
 */
#define USEC(x) ( ( (x) >> 12 ) - 759 * ( ( ( (x) >> 10 ) + 32768 ) >> 16 ) )

/* Converts NTP delay and dispersion, apparently in seconds scaled
 * by 65536, to microseconds.  RFC1305 states this time is in seconds,
 * doesn't mention the scaling.
 * Should somehow be the same as 1000000 * x / 65536
 */
#define sec2u(x) ( (x) * 15.2587890625 )

struct ntptime {
	unsigned int coarse;
	unsigned int fine;
};

/* prototype for function defined in phaselock.c */
int contemplate_data(unsigned int absolute, double skew, double errorbar, int freq);

/* prototypes for some local routines */
void send_packet(int usd);
int rfc1305print(uint32_t *data, struct ntptime *arrival);
void udp_handle(int usd, char *data, int data_len, struct sockaddr *sa_source, int sa_len);
char* syslog_init( void ) ;
void syslog_close( char * ) ;
void print_time_stamp(  struct ntptime *, char *, int ) ;
size_t get_nvram_value( const char *, char *, int ) ;
void sys_log( const char *, int ) ;

/* variables with file scope
 * (I know, bad form, but this is a short program) */
static uint32_t incoming_word[325];
#define incoming ((char *) incoming_word)
#define sizeof_incoming (sizeof(incoming_word)*sizeof(uint32_t))
static struct timeval time_of_send;
static int live=0;
static int set_clock=0;   /* non-zero presumably needs root privs */
static int time_offset=0;   
char *syslog_ident = NULL ;
char tmp_syslog_str[2] ; /*for empty syslog ident*/ 

/* when present, debug is a true global, shared with phaselock.c */
#ifdef ENABLE_DEBUG
int debug=0;
#define DEBUG_OPTION "d"
#else
#define debug 0
#define DEBUG_OPTION
#endif

int sysloglevel = 0 ;
/*
*  EXTRA_DEBUG	 5
*  DEBUG		 1  as per the web interface
*  INFO    		 2
*  WARNING 		 3
*  ERROR   		 4
*  NONE    		 0
*/

int get_current_freq(void)
{
	/* OS dependent routine to get the current value of clock frequency.
	 */
#ifdef linux
	struct timex txc;
	txc.modes=0;
	if (adjtimex(&txc) < 0) {
		perror("adjtimex"); exit(1);
	}
	return txc.freq;
#else
	return 0;
#endif
}

int set_freq(int new_freq)
{
	/* OS dependent routine to set a new value of clock frequency.
	 */
#ifdef linux
	struct timex txc;
	txc.modes = ADJ_FREQUENCY;
	txc.freq = new_freq;
	if (adjtimex(&txc) < 0) {
		perror("adjtimex"); exit(1);
	}
	return txc.freq;
#else
	return 0;
#endif
}

void send_packet(int usd)
{
	__u32 data[12];
	struct timeval now;
#define LI 0
#define VN 3
#define MODE 3
#define STRATUM 0
#define POLL 4 
#define PREC -6

	if (debug) fprintf(stderr,"Sending ...\n");
	if (sizeof(data) != 48) {
		fprintf(stderr,"size error\n");
		return;
	}
	memset((char *) data, 0, sizeof(data));
	data[0] = htonl (
		( LI << 30 ) | ( VN << 27 ) | ( MODE << 24 ) |
		( STRATUM << 16) | ( POLL << 8 ) | ( PREC & 0xff ) );
	data[1] = htonl(1<<16);  /* Root Delay (seconds) */
	data[2] = htonl(1<<16);  /* Root Dispersion (seconds) */
	gettimeofday(&now,NULL);
	data[10] = htonl(now.tv_sec + JAN_1970); /* Transmit Timestamp coarse */
	data[11] = htonl(NTPFRAC(now.tv_usec));  /* Transmit Timestamp fine   */
	send(usd,data,48,0);
	time_of_send=now;
}

void get_packet_timestamp(int usd, struct ntptime *udp_arrival_ntp)
{
	struct timeval udp_arrival;
#ifdef _PRECISION_SIOCGSTAMP
	if ( ioctl(usd, SIOCGSTAMP, &udp_arrival) < 0 ) {
		perror("ioctl-SIOCGSTAMP");
		gettimeofday(&udp_arrival,NULL);
	}
#else
	gettimeofday(&udp_arrival,NULL);
#endif
	udp_arrival_ntp->coarse = udp_arrival.tv_sec + JAN_1970;
	udp_arrival_ntp->fine   = NTPFRAC(udp_arrival.tv_usec);
}

void check_source(int data_len, struct sockaddr *sa_source, int sa_len)
{
	/* This is where one could check that the source is the server we expect */
	if (debug) {
		struct sockaddr_in *sa_in=(struct sockaddr_in *)sa_source;
		char msg[56] ;
		snprintf( msg, 56, "packet of length %d received\n", data_len ) ;
		printf( msg ) ;
		sys_log( msg, LOG_DEBUG ) ;
		if (sa_source->sa_family==AF_INET) {
		    snprintf( msg, 56, "Source: INET Port %d host %s\n", ntohs(sa_in->sin_port),inet_ntoa(sa_in->sin_addr) ) ;
		    printf( msg ) ;
		    sys_log( msg, LOG_DEBUG ) ;
		} else {
		    snprintf( msg, 56, "Source: Address family %d\n", sa_source->sa_family ) ;
		    printf( msg ) ;
		    sys_log( msg, LOG_DEBUG ) ;
		}
	}
}

double ntpdiff( struct ntptime *start, struct ntptime *stop)
{
	int a;
	unsigned int b;
	a = stop->coarse - start->coarse;
	if (stop->fine >= start->fine) {
		b = stop->fine - start->fine;
	} else {
		b = start->fine - stop->fine;
		b = ~b;
		a -= 1;
	}
	
	return a*1.e6 + b * (1.e6/4294967296.0);
}

/* Does more than print, so this name is bogus.
 * It also makes time adjustments, both sudden (-s)
 * and phase-locking (-l).  */
/* return value is number of microseconds uncertainty in answer */
int rfc1305print(uint32_t *data, struct ntptime *arrival)
{
/* straight out of RFC-1305 Appendix A */
	int li, vn, mode, stratum, poll, prec;
	int delay, disp, refid;
	struct ntptime reftime, orgtime, rectime, xmttime;
	double el_time,st_time,skew1,skew2;
	int freq;

#define Data(i) ntohl(((uint32_t *)data)[i])
	li      = Data(0) >> 30 & 0x03;
	vn      = Data(0) >> 27 & 0x07;
	mode    = Data(0) >> 24 & 0x07;
	stratum = Data(0) >> 16 & 0xff;
	poll    = Data(0) >>  8 & 0xff;
	prec    = Data(0)       & 0xff;
	if (prec & 0x80) prec|=0xffffff00;
	delay   = Data(1);
	disp    = Data(2);
	refid   = Data(3);
	reftime.coarse = Data(4);
	reftime.fine   = Data(5);
	orgtime.coarse = Data(6);
	orgtime.fine   = Data(7);
	rectime.coarse = Data(8);
	rectime.fine   = Data(9);
	xmttime.coarse = Data(10);
	xmttime.fine   = Data(11);
#undef Data

	if (set_clock) {   /* you'd better be root, or ntpclient will crash! */
		struct timeval tv_set;
		/* it would be even better to subtract half the slop */
		tv_set.tv_sec  = xmttime.coarse - JAN_1970;
		/* offset time */
		tv_set.tv_sec  += time_offset*60;
		
		/* divide xmttime.fine by 4294.967296 */
		tv_set.tv_usec = USEC(xmttime.fine);
		if (settimeofday(&tv_set,NULL)<0) {
			perror("settimeofday");
			exit(1);
		}
		if (debug) {
			char msg[56] ;
			snprintf( msg, 56, "set time to %lu.%.6lu\n", tv_set.tv_sec, tv_set.tv_usec ) ;
			printf( msg ) ;
			sys_log( msg, LOG_DEBUG ) ; 
		}
	}

	if (debug) {
	printf("LI=%d  VN=%d  Mode=%d  Stratum=%d  Poll=%d  Precision=%d\n",
		li, vn, mode, stratum, poll, prec);
	printf("Delay=%.1f  Dispersion=%.1f  Refid=%u.%u.%u.%u\n",
		sec2u(delay),sec2u(disp),
		refid>>24&0xff, refid>>16&0xff, refid>>8&0xff, refid&0xff);
	printf("Reference %u.%.10u\n", reftime.coarse, reftime.fine);
	printf("Originate %u.%.10u\n", orgtime.coarse, orgtime.fine);
	printf("Receive   %u.%.10u\n", rectime.coarse, rectime.fine);
	printf("Transmit  %u.%.10u\n", xmttime.coarse, xmttime.fine);
	printf("Our recv  %u.%.10u\n", arrival->coarse, arrival->fine);
	}
	el_time=ntpdiff(&orgtime,arrival);   /* elapsed */
	st_time=ntpdiff(&rectime,&xmttime);  /* stall */
	skew1=ntpdiff(&orgtime,&rectime);
	skew2=ntpdiff(&xmttime,arrival);
	freq=get_current_freq();
	if (debug) {
		char msg[128] ;
	snprintf( msg, 128, "Total elapsed: %9.2f\n"
	                    "Server stall:  %9.2f\n"
	                    "Slop:          %9.2f\n",
		                el_time, st_time, el_time-st_time);
	printf( msg ) ;
	sys_log( msg, LOG_DEBUG ) ;
	snprintf(msg, 128, "Skew:          %9.2f\n"
	                   "Frequency:     %9d\n"
	                   " day   second     elapsed    stall     skew  dispersion  freq\n",
		               (skew1-skew2)/2, freq);
	printf( msg ) ;
	sys_log( msg, LOG_DEBUG ) ;
	}
	/* Not the ideal order for printing, but we want to be sure
	 * to do all the time-sensitive thinking (and time setting)
	 * before we start the output, especially fflush() (which
	 * could be slow).  Of course, if debug is turned on, speed
	 * has gone down the drain anyway. */
	if (live) {
		int new_freq;
		new_freq = contemplate_data(arrival->coarse, (skew1-skew2)/2,
			el_time+sec2u(disp), freq);
		if (!debug && new_freq != freq) set_freq(new_freq);
	}
	printf("%d %.5d.%.3d  %8.1f %8.1f  %8.1f %8.1f %9d\n",
		arrival->coarse/86400, arrival->coarse%86400,
		arrival->fine/4294967, el_time, st_time,
		(skew1-skew2)/2, sec2u(disp), freq);
	fflush(stdout);
	{
		char msg[200] ;
	    char time_str[50] ;
		snprintf( msg, 200, "Days:%d Seconds:%.5d.%.3d Elapsed:%8.1f sec Stall:%8.1f sec Skew:%8.1f sec Dispersion:%8.1f usec Frequency:%9d",
		                    arrival->coarse/86400, arrival->coarse%86400,
            		        arrival->fine/4294967, el_time, st_time,
		                    (skew1-skew2)/2, sec2u(disp), freq);
        sys_log( msg, LOG_INFO ) ;

	    snprintf( msg, 200, "Leap Indicator=%d Version Number=%d Mode=%d Peer Clock Stratum=%d Peer Polling Interval=%d Precision=%d",
		                    li, vn, mode, stratum, poll, prec ) ;
        sys_log( msg, LOG_DEBUG ) ;
	    snprintf( msg, 200, "Root Delay=%.1f usec Root Dispersion=%.1f usec Reference Clock ID=%u.%u.%u.%u",
		                    sec2u(delay),sec2u(disp),
        		            refid>>24&0xff, refid>>16&0xff, refid>>8&0xff, refid&0xff ) ;
        sys_log( msg, LOG_DEBUG ) ;
		print_time_stamp( &reftime, time_str, sizeof( time_str ) ) ;
        snprintf( msg, 200, "Reference Time Stamp %s", time_str ) ;
        sys_log( msg, LOG_DEBUG ) ;
		print_time_stamp( &orgtime, time_str, sizeof( time_str ) ) ;
        snprintf( msg, 200, "Originate Time Stamp %s", time_str ) ;
        sys_log( msg, LOG_DEBUG ) ;
		print_time_stamp( &rectime, time_str, sizeof( time_str ) ) ;
        snprintf( msg, 200, "Receive Time Stamp %s", time_str ) ;
        sys_log( msg, LOG_DEBUG ) ;
		print_time_stamp( &xmttime, time_str, sizeof( time_str ) ) ;
        snprintf( msg, 200, "Transmit Time Stamp %s", time_str ) ;
        sys_log( msg, LOG_DEBUG ) ;
		print_time_stamp( arrival, time_str, sizeof( time_str ) ) ;
        snprintf( msg, 200, "Our recieve Time Stamp %s", time_str ) ;
        sys_log( msg, LOG_DEBUG ) ;
    	snprintf( msg, 200, "Total time elapsed: %9.2f sec Server stall: %9.2f sec Slop: %9.2f sec",
	            el_time, st_time, el_time-st_time ) ;
        sys_log( msg, LOG_DEBUG ) ;
        snprintf( msg, 200, "Skew: %9.2f sec Frequency: %9d",
		        (skew1-skew2)/2, freq ) ;
        sys_log( msg, LOG_DEBUG ) ;
	}
	return(el_time-st_time);
}

int stuff_net_addr(struct in_addr *p, char *hostname)
{
	struct hostent *ntpserver;
	char msg[100] ;
	ntpserver=gethostbyname(hostname);
	if (ntpserver == NULL) {
		snprintf( msg, 100, "Error looking up hostname %s\n", hostname) ;
		sys_log( msg, LOG_ERR ) ;
		herror(hostname);
		return 1 ;
	}
	if (ntpserver->h_length != 4) {
		snprintf( msg, 100, "Incorrect address length %d\n", ntpserver->h_length ) ;
		sys_log( msg, LOG_ERR ) ;
		fprintf(stderr,"oops %d\n",ntpserver->h_length);
		return 1 ;
	}
	memcpy(&(p->s_addr),ntpserver->h_addr_list[0],4);
	return 0 ;
}

int setup_receive(int usd, unsigned int interface, short port)
{
	struct sockaddr_in sa_rcvr;
	memset((char *) &sa_rcvr, 0, sizeof(sa_rcvr));
	sa_rcvr.sin_family=AF_INET;
	sa_rcvr.sin_addr.s_addr=htonl(interface);
	sa_rcvr.sin_port=htons(port);
	if(bind(usd,(struct sockaddr *) &sa_rcvr,sizeof(sa_rcvr)) == -1) {
		char msg[32] ;
		perror("bind");
		snprintf( msg, 32, "Could not bind to port %d\n", port ) ;
		sys_log( msg, LOG_ERR ) ;
		fprintf(stderr,"could not bind to udp port %d\n",port);
		return 1 ;
	}
	listen(usd,3);
	return 0 ;
}

int setup_transmit(int usd, char *host, short port)
{
	struct sockaddr_in sa_dest;
	char msg[32] ;
	memset((char *) &sa_dest, 0, sizeof(sa_dest));
	sa_dest.sin_family=AF_INET;
	if( stuff_net_addr(&(sa_dest.sin_addr),host) )
	{
		return 1 ;
	}
	sa_dest.sin_port=htons(port);
	if (connect(usd,(struct sockaddr *)&sa_dest,sizeof(sa_dest))==-1)
	{
		perror("connect");
		snprintf( msg, 32, "Could not connect to port %d\n", port ) ;
		sys_log( msg, LOG_ERR ) ;
		return 1 ;
	}
	return 0 ;	
}

void primary_loop(int usd, int num_probes, int interval, int goodness)
{
	fd_set fds;
	struct sockaddr sa_xmit;
	int i, pack_len, sa_xmit_len, probes_sent, error;
	struct timeval to;
	struct ntptime udp_arrival_ntp;

	if (debug) printf("Listening...\n");

	probes_sent=0;
	sa_xmit_len=sizeof(sa_xmit);
	to.tv_sec=0;
	to.tv_usec=0;
	for (;;) {
		FD_ZERO(&fds);
		FD_SET(usd,&fds);
		i=select(usd+1,&fds,NULL,NULL,&to);  /* Wait on read or error */
		if ((i!=1)||(!FD_ISSET(usd,&fds))) {
			if (i==EINTR) continue;
			if (i<0) perror("select");
			if ((to.tv_sec == 0)|| (to.tv_sec == interval)) {
				if (probes_sent >= num_probes &&
					num_probes != 0) 
					break;
				send_packet(usd);
				++probes_sent;
				to.tv_sec=interval;
				to.tv_usec=0;
			}	
			continue;
		}
		pack_len=recvfrom(usd,incoming,sizeof_incoming,0,
		                  &sa_xmit,&sa_xmit_len);
		error = goodness+1;
		if (pack_len<0) {
			perror("recvfrom");
		} else if (pack_len>0 && (unsigned)pack_len<sizeof_incoming){
			get_packet_timestamp(usd, &udp_arrival_ntp);
			check_source(pack_len, &sa_xmit, sa_xmit_len);
			error = rfc1305print(incoming_word, &udp_arrival_ntp);
			/* udp_handle(usd,incoming,pack_len,&sa_xmit,sa_xmit_len); */
			/*if clock has been set, our work is done*/
			if( set_clock ) break ;
		} else {
			printf("Ooops.  pack_len=%d\n",pack_len);
			fflush(stdout);
		}
		if ( error < goodness && goodness != 0) break;
		if (probes_sent >= num_probes && num_probes != 0) break;
	}
}

void do_replay(void)
{
	char line[100];
	int n, day, freq, absolute;
	float sec, el_time, st_time, disp;
	double skew, errorbar;
	int simulated_freq = 0;
	unsigned int last_fake_time = 0;
	double fake_delta_time = 0.0;

	while (fgets(line,sizeof(line),stdin)) {
		n=sscanf(line,"%d %f %f %f %lf %f %d",
			&day, &sec, &el_time, &st_time, &skew, &disp, &freq);
		if (n==7) {
			fputs(line,stdout);
			absolute=day*86400+(int)sec;
			errorbar=el_time+disp;
			if (debug) printf("contemplate %u %.1f %.1f %d\n",
				absolute,skew,errorbar,freq);
			if (last_fake_time==0) simulated_freq=freq;
			fake_delta_time += (absolute-last_fake_time)*((double)(freq-simulated_freq))/65536;
			if (debug) printf("fake %f %d \n", fake_delta_time, simulated_freq);
			skew += fake_delta_time;
			freq = simulated_freq;
			last_fake_time=absolute;
			simulated_freq = contemplate_data(absolute, skew, errorbar, freq);
		} else {
			fprintf(stderr,"Replay input error\n");
			exit(2);
		}
	}
}

void usage(char *argv0)
{
	fprintf(stderr,
	"Usage: %s [-c count] [-d] [-g goodness] -h hostname [-i interval]\n"
	"\t[-l] [-p port] [-r] [-s] [-o timeoffset]\n",
	argv0);
}

int main(int argc, char *argv[]) {
	int usd;  /* socket */
	int c;
	/* These parameters are settable from the command line
	   the initializations here provide default behavior */
	short int udp_local_port=0;   /* default of 0 means kernel chooses */
	int cycle_time=3600;          /* seconds */
	int probe_count=0;            /* default of 0 means loop forever */
	/* int debug=0; is a global above */
	int goodness=0;
	char *hostname=NULL;          /* must be set */
	int replay=0;                 /* replay mode overrides everything */

	for (;;) {
		c = getopt( argc, argv, "c:" DEBUG_OPTION "g:h:i:lp:rso:");
		if (c == EOF) break;
		switch (c) {
			case 'c':
				probe_count = atoi(optarg);
				break;
#ifdef ENABLE_DEBUG
			case 'd':
				++debug;
				break;
#endif
			case 'g':
				goodness = atoi(optarg);
				break;
			case 'h':
				hostname = optarg;
				break;
			case 'i':
				cycle_time = atoi(optarg);
				break;
			case 'l':
				live++;
				break;
			case 'p':
				udp_local_port = atoi(optarg);
				break;
			case 'r':
				replay++;
				break;
			case 's':
				set_clock++;
				/*don't set probe_count to 1 in case of network problems*/
				//probe_count = 1;
				break;
			case 'o':
				time_offset = atoi(optarg);
				break;
			default:
				usage(argv[0]);
				exit(1);
		}
	}

    syslog_ident = syslog_init() ;
	if( !syslog_ident )
		fprintf(stderr,"Syslog initializing error\n");

#ifdef vxworks
	hostname = argv[0];
	set_clock++;
#endif

	if (replay) {
		do_replay();
		syslog_close( syslog_ident ) ;
		exit(0);
	}
	if (hostname == NULL) {
		usage(argv[0]);
		syslog_close( syslog_ident ) ;
		exit(1);
	}
	if (debug) {
		printf("Configuration:\n"
		"  -c probe_count %d\n"
		"  -d (debug)     %d\n"
		"  -g goodness    %d\n"
		"  -h hostname    %s\n"
		"  -i interval    %d\n"
		"  -l live        %d\n"
		"  -p local_port  %d\n"
		"  -s set_clock   %d\n"
		"  -o time_offset   %d\n",
		probe_count, debug, goodness, hostname, cycle_time,
		live, udp_local_port, set_clock, time_offset );
	}

	for( ;; )
	{
		/* Startup sequence */
		if ((usd=socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP))==-1)
		{
			perror ("socket");
			sys_log( "Could not open socket\n", LOG_ERR ) ;
			sleep( cycle_time ) ;
			continue ;
		}

		if( setup_receive(usd, INADDR_ANY, udp_local_port) )
		{
			close( usd ) ;
			sleep( cycle_time ) ;
			continue ;
		}

		if( setup_transmit(usd, hostname, NTP_PORT) )
		{
			close( usd ) ;
			sleep( cycle_time ) ;
			continue ;
		}
		break ;
	}

	primary_loop(usd, probe_count, cycle_time, goodness);

	close(usd);
	syslog_close( syslog_ident ) ;
	return 0;
}

void get_mac_addr( char *lower_mac, char*upper_mac )
{
	FILE *fp ;
	int i ;

	fp = fopen( DEVICE_MAC, "r" ) ;
	if ( fp == NULL )
		return ;

	fread( upper_mac, 17, 1, fp ) ;
	fclose( fp ) ;
	for( i = 0 ; i < 17 ; i++ )
	{
		lower_mac[i] = upper_mac[i] ;
		if( upper_mac[i] > 0x60 && upper_mac[i] < 0x7B )
			upper_mac[i] -= 0x20 ;
		if( lower_mac[i] > 0x40 && lower_mac[i] < 0x5B )
			lower_mac[i] += 0x20 ;
	}		
}

char* syslog_init()
{
	char buffer[2048], loglevel[3] ;
	int i = 0 ;
	size_t bytes_read ;
	char prog_name[6] = " NTP " ;
	char upper_mac_addr[18] = { 0 } ;
	char lower_mac_addr[18] = { 0 } ;
	int log_length = 0 ;
								
    memset( loglevel, 0, 3 ) ;
	get_nvram_value( "208", loglevel, 3 ) ;	
	          //208 sets the log level						
	sysloglevel=atoi( loglevel ) ;

	memset( upper_mac_addr, 0, 18 ) ;
	memset( lower_mac_addr, 0, 18 ) ;
	get_mac_addr( lower_mac_addr, upper_mac_addr ) ;
    memset( buffer, 0, 2048 ) ;
	bytes_read = get_nvram_value( "syslog_id", buffer, 2048 ) ; 
	             //syslog_id is needed to display log messages 						
 	if( bytes_read <= 0 )   			 
		/*set bytes_read to single byte as dummy syslog id*/
		bytes_read = 1 ;
	i = 0 ;
	while( buffer[i] != '\0' && i < 2048 )
	{
		if( buffer[i] == '\t' )
			buffer[i] = ' ' ;
		if( buffer[i] < 0x20 || buffer[i] > 0x7e )
		{
			if( (i + 1 ) < 2048 )
			{
				buffer[i] = ' ' ;
			    buffer[i+1] = 0 ;
			}
			else
				buffer[i] = 0 ;
			break ;
		}
		i++ ;	
	}
	/*include mac address in syslog if absent from syslog_id*/
	if( strlen( upper_mac_addr ) && !strstr( buffer, lower_mac_addr ) && !strstr( buffer, upper_mac_addr ) )
	{
		log_length =  strlen( buffer ) * sizeof( char ) + 4 + strlen( upper_mac_addr ) + strlen( prog_name ) + 1 ;
		syslog_ident = ( char *)malloc( log_length ) ;
		memset( syslog_ident, 0, log_length ) ;
		strncpy( syslog_ident, buffer, bytes_read - 1 ) ;
		syslog_ident[bytes_read - 1] = '\0' ;
		strcat( syslog_ident, ": [" ) ;
		strcat( syslog_ident, upper_mac_addr ) ;
		strcat( syslog_ident, "]" ) ;
		syslog_ident[bytes_read + 3 + strlen( upper_mac_addr ) ] = '\0' ;
	}
	else
	{
		log_length =  strlen( buffer ) * sizeof( char ) + strlen( prog_name ) + 1 ;
		syslog_ident = ( char *)malloc( log_length ) ;
		memset( syslog_ident, 0, log_length ) ;
		strncpy( syslog_ident, buffer, bytes_read - 1 ) ;
		syslog_ident[bytes_read - 1] = '\0' ;
	}
	strcat( syslog_ident, prog_name ) ;
	if( sysloglevel != 5 ) 
		openlog( syslog_ident, LOG_CONS|LOG_NDELAY, LOG_USER ) ;
	else
	{
		tmp_syslog_str[0] = '1' ;
		tmp_syslog_str[1] = '\0' ;
		openlog( tmp_syslog_str, LOG_CONS|LOG_NDELAY, LOG_USER ) ;
		tmp_syslog_str[0] = '\0' ;
	}
	//printf( "%s:%d %s\n", __FILE__, __LINE__, syslog_ident ) ;
	return syslog_ident ;
}

void syslog_close( char *syslog_ident )
{
	closelog() ;
	if( syslog_ident ) ;
	free( syslog_ident ) ;
}

void print_time_stamp(  struct ntptime *timestamp, char *time_str, int str_size )
{
    struct timeval tv_syslog ;
    struct tm *tm_ptr ;
	int no_of_chars ;

	tv_syslog.tv_sec  = timestamp->coarse - JAN_1970;
	tv_syslog.tv_usec = USEC(timestamp->fine);
	tm_ptr = localtime( &tv_syslog.tv_sec ) ;
    no_of_chars = strftime( time_str, str_size,"%Z %b %d, %Y %H:%M:%S", tm_ptr ) ;
	if( (no_of_chars + 5 ) < str_size ) 
	{ 
		sprintf( time_str + no_of_chars, ".%04ld", tv_syslog.tv_usec / 100 ) ;
	}
	time_str[ str_size - 1 ] = '\0' ;
}

size_t get_nvram_value( const char *var, char *buffer, int size )
{
	FILE* fp ;
	char cmd[256] ;
	size_t bytes_read = 0 ;

	snprintf( cmd, 256, "nvram get %s", var ) ;
	cmd[255] = '\0' ;
	fp = popen( cmd, "r" ) ;							
	if( !fp )
	{
		fprintf ( stderr, "incorrect parameters or too many files.\n" ) ;
		return 0 ;
	}
	bytes_read = fread( buffer, 1, size, fp ) ;
	if(( bytes_read == 0 ) || ( bytes_read == size ) )
	{
	    snprintf( cmd, 256, "Failed to read  %s", var ) ;
	    cmd[255] = '\0' ;
		perror( cmd ) ;
		return 0 ;
	}
	buffer[bytes_read] = '\0' ;
    if (pclose (fp) != 0)
		fprintf (stderr, "Could not run nvram or other error.\n");
	return bytes_read ;	
}

/*
*  Convert to levels as per the web interface
*/
void sys_log( const char *msg, int level )
{
	char priority[6] ;
	switch( level )
	{
	    case LOG_DEBUG:
		    if( sysloglevel == 1 )
			    syslog( LOG_MAKEPRI( LOG_USER, LOG_DEBUG ), "%s", msg ) ;
			else if( sysloglevel == 5 )
			{
				snprintf( priority, 6, "<%d>", LOG_USER + LOG_DEBUG ) ;	
			    syslog( LOG_MAKEPRI( LOG_USER, LOG_DEBUG ), "%s %s %s\n\n", priority, syslog_ident, msg ) ;
			}
		    break ;
	    case LOG_INFO:
		    if( sysloglevel > 0 && sysloglevel < 3 )
			    syslog( LOG_MAKEPRI( LOG_USER, LOG_INFO ), "%s", msg ) ;
			else if( sysloglevel == 5 )
			{
				snprintf( priority, 6, "<%d>", LOG_USER + LOG_INFO ) ;	
			    syslog( LOG_MAKEPRI( LOG_USER, LOG_INFO ), "%s %s %s\n\n", priority, syslog_ident, msg ) ;
			}
    		break ;
	    case LOG_NOTICE:
	    case LOG_WARNING:
		    if( sysloglevel > 0 && sysloglevel < 4 )
			    syslog( LOG_MAKEPRI( LOG_USER, LOG_WARNING ), "%s", msg ) ;
			else if( sysloglevel == 5 )
			{
				snprintf( priority, 6, "<%d>", LOG_USER + LOG_WARNING ) ;	
			    syslog( LOG_MAKEPRI( LOG_USER, LOG_WARNING ), "%s %s %s\n\n", priority, syslog_ident, msg ) ;
			}
        	break ;
	    case LOG_ERR:
	    case LOG_CRIT:
	    case LOG_ALERT:
	    case LOG_EMERG:
		    if( sysloglevel > 0 && sysloglevel < 5 )
			    syslog( LOG_MAKEPRI( LOG_USER, LOG_ERR ), "%s", msg ) ;
			else if( sysloglevel == 5 )
			{
				snprintf( priority, 6, "<%d>", LOG_USER + LOG_ERR ) ;	
			    syslog( LOG_MAKEPRI( LOG_USER, LOG_ERR ), "%s %s %s\n\n", priority, syslog_ident, msg ) ;
			}
	}
}
